/*! \mainpage Luiza

Welcome to the Luiza documentation page. 
This is the place where you can find some explanation and examples of how
to run GLORIA analysis programs within Luiza.
<br><br>

If you use Luiza in your analysis, please acknowledge it by using the reference to:<br>
<i> A.F. Zarnecki et al., Analysis framework for GLORIA,
<a href="http://dx.doi.org/10.1117/12.2000199">Proc. SPIE 8454, 845408, 2012.</a></i>

<br><br>

Luiza framework was based on the 
<a href="http://ilcsoft.desy.de/portal/software_packages/marlin/">Marlin</a> package
developed by  Frank Gaede and Jan Engels from DESY for the ILC data analysis. <br> 
Please acknowledge their effort by refering to: <br> 
<i>F. Gaede, Marlin and LCCD: Software tools for the ILC, 
<a href="http://dx.doi.org/10.1016/j.nima.2005.11.138">Nucl.Instrum.Meth.A559:177-180,2006.</a></i>
<br><br>

The idea (and large parts of code) taken from Marlin is that every computing task is implemented 
as a processor (module) that analyzes the data stored in a gloria::GloriaDataContainer 
structure and created additional output is also added to that collection. 
The advantage of such a modular approach is to keep things as
simple as possible. Every single step of the full analysis chain that
goes eg. from raw images to light curves can be processed step by step and the
output of each step is still self consistent and can be fed in to the
next step without any manipulation. Dedicated processors can also be included to read 
data from local file or external file server, access databases as well as to store analysis results. 
FIT image collections as well as tabular data collections can be stored.
The framework allows to define the processors (and their order) that are executed at runtime in a
simple XML steering file. Via the steering file you can also define named parameters 
(as a string, float, int - single and arrays) for every processor as well as for the global scope.
By using the framework users don't have to write any code that deals with the IO they simply
write processors with defined callbacks, i.e. init(), processData(), check(), end().
<br><br>


More information is available at the following pages:

@li @ref luiza_install
@li @ref how_it_works
@li @ref steering_file
@li @ref image_analysis
@li @ref object_analysis
@li @ref databases
@li @ref examples
@li @ref network_configuration
@li @ref new_processor
@li @ref test_data
@li @ref luiza_gui
@li @ref to_do


<b>Last versions release info:</b>

Version v03-01\n
New features and improvements
@li selection by header keywords possible when calculating average and median frame
@li dark matching by header keywords and exposure time
@li reading of raw Pi of the Sky frames (nonstandard file structure)
@li returne values for processors reading input files implemented

Version v03-00\n
New features and improvements
@li SQL database interface implemented
@li processor for searching new objects 
@li code restructuring

Version v02-00\n
Final version for GLORIA
@li improved aperture photometry implemented
@li processor for matching image object with sky catalog
@li frame calibration based on matched reference stars
@li storing tables in root TTree format

Version v01-01\n
Development Version with improvements based on feedback and test results
@li dedicated background map calculation and interpolation class implemented
@li pixel cluster shape analysis methods implemented
@li processors for data table reading and writting in FITS format
@li data table sorting implemented
@li analysis of compressed FITS images possible

Version v01-00\n
Final version for demonstrator experiment
@li processor for light curve reconstruction implemented
@li improved background estimate in clustering algorithm
@li aperture photometry implemented
@li Laplace transformation for image "sharpening" added

Version v00-06
@li astrometry processor based on Astrometry.Net added
@li XML file for storing network configuration and telescope specific analysis parameters
@li extended options for image manipulation
@li weighted subtraction of dark and bias frames based on exposure times

Version v00-05
@li First version stored in Gloria SVN
@li GUI for steering file editing added, based on QT4
@li Processor for writing tables to text files added
@li Image viewer processor based on CERN root libraries
@li Exceptions allowing for job control from processor level implemented

Version v00-04
@li First public version
@li Processors for image normalization and object finding added
@li Directories structure modified to reflect code modularity better
@li Processors for image analysis built as a separate module

Version v00-03
@li Processors for adding/averaging images and for writing images
    to output FITS file added. 

Version v00-02
@li Makefile added. Not all dependencies properly included, but it should help anyway
@li ExampleProcessor added, linked as shared library - example for users to start developing their own processors
@li Still no time for proper cleaning of the code...

Version v00-01
@li Class for handling FITS tables added
@li More documentation added
@li Still a lot of cleaning needed in the code...

Version v00-00
@li First running version
@li Includes only basic functionality
@li Under heavy development: version not stable

*/

*****************************************************************************

/*!

@page luiza_install  How to install Luiza frmework


@par Requirements

Luiza is written in C++ and internal data structures are based 
on Standard Template Library (STL) containers. It was developed
under gcc version 4.4.4, but is supposed to run in any modern 
C++ installation.  

Additional packages and/or libraries which are required to build Luiza 
and its packages are:

@li CFITSIO library for reading and writing data files in FITS 
  (Flexible Image Transport System) data format is obligatory. 
  It is available as precompiled library in some linux distributions, 
  can be also downloaded from 
  <a href="http://heasarc.gsfc.nasa.gov/docs/software/fitsio/fitsio.html">CFITSIO</a> web page.

@li Astrometry.net library is required to run luiza::Astrometry processor. 
    Current implementation is based on version 0.38, which can be downloaded from 
    <a href="http://astrometry.net/downloads/astrometry.net-0.38.tar.bz2"> here </a>.
    Path to the catalogue where the library sources are downloaded and compiled (make) 
    should be specified by ASTROMETRY_DIR environment variable.
    In addition, index files have to be downloaded from  dedicated
    <a href="http://broiler.astrometry.net/~dstn/4000/">repository</a>.
    Location of index files has to be specified in steering file.

@li CERN ROOT package is required to run processors stored in 
\a processors/luiza_root. It is used to provied graphic output from Luiza.
The source and binaries can be downloaded from <a href="http://root.cern.ch/root/">ROOT home page</a>.

@li PYTHON libraries and the Mahotas packages are required to compile and use
source finder processor from \a processors/image_python. Mahotas can be
downloaded from <a href="http://luispedro.org/software/mahotas">here</a>.

@li SOCI libraries are required to compile and use the SQL database interface.
Please refer to <a href="http://soci.sourceforge.net/doc.html">SOCI documentation</a>.

@li QT4 libraries are required to compile Luiza GUI interface

If you lack any of the required packages (but cfitsio), you can still compile 
and run Luiza (with reduced functionality): just delete catalogues with processors which can not be compiled (eg. \a processors/luiza_root or \a processors/image_python ).

@par Installation

Download archive with current luiza version: from 
Gloria <a href="http://sourceforge.net/projects/gloriaproject/">SVN repository</a>
at Sourceforge or from <a href="../../luiza_current.tgz"> here </a>.

Create main destination folder and unpack the code there. 
You should then set the environment variable LUIZA_ROOT 
pointing to this location. For example:
@code 
   setenv LUIZA_ROOT ${HOME}/astro/gloria/LuizaFramework
@endcode

To create Luiza binary, shared libraries and documentation use 'make'.
Current version of \a Makefile does not take all dependencies into 
account but it should work fine when building from scratch.
Otherwise you should use 'make clean' and 'make' to be sure all 
cross-package dependencies are taken into account.

Luiza shared libraries will be built in a subfolder so you have 
to add it to library path before running Luiza:
@code 
    setenv LD_LIBRARY_PATH ${LD_LIBRARY_PATH}:${LUIZA_ROOT}/luiza/lib
@endcode


\a Luiza itself is just a framework for I/O. 
The whole power of this solution is only visible when you consider 
additional processors, dedicated for special tasks. To make Luiza load 
different processor packages (stored and compiled in separate folders), 
one has to define LUIZA_DLL variable. To load all processors available 
in the current version:
@code 
   setenv LUIZA_DLL ${LUIZA_ROOT}/processors/image_analysis/lib/libimageanalysis.so:${LUIZA_ROOT}/processors/luiza_example/lib/libexample.so:${LUIZA_ROOT}/processors/image_python/lib/libSimpleSourceFinder.so:${LUIZA_ROOT}/processors/luiza_root/lib/libluizaroot.so:${LUIZA_ROOT}/processors/astrometry/lib/libAstrometry.so
@endcode

Python base processor requires also setting the variable pointing to the puthon sources:
@code 
   setenv LUIZA_PYTHON ${LUIZA_ROOT}/processors/image_python/src
@endcode


To run Luiza job:
@code 
    ${LUIZA_ROOT}/luiza/bin/Luiza {options} [steering file]
@endcode

All required variables can be also set by using scripts \a env_tcsh
or \a env_bash located in the main catalogue. You can modify these scripts
to match your instalation.

You can test Luiza by running prepared test job (in the main catalog):
@code 
    luiza/bin/Luiza test.xml 
@endcode

Small sample of fit files used in the test job should be stored in catalog linked as
\a gloriadata. You can download these data <a href="../../luiza_data.tgz"> here </a>

Larger samples of fits images from different telescopes, sent by partners,
which can be used for algorithm testing, are available from the dedicated
<a href="http://hep.fuw.edu.pl/u/zarnecki/gloria/gloriadata/GloriaSampleData.html"> web page </a>.

*/

*****************************************************************************

/*!

@page how_it_works  How Luiza processes data

Principle is very simple. The data processing task is divided in a series of
simple steps. The only requirement is that each step need to have a well defined 
input and output structure. In addition, separate processors are needed for reading 
data from file, database or net, and for storing analysis results.

User defines what the program will do by selecting the processors in the XML 
steering file. The idea is to develope a large number of processors in GLORIA, doing 
many different tasks, so user is always able to find a set which matches his/her needs.

The main "working horse" of Luiza is luiza::ProcessorMgr class. It is used by Luiza
to create a list of active processors (after parsing XML file), and setting values to 
parameters required by this processors (given in the same XML file).  Same processor type 
(eg. processor reading fits images luiza::FitsImageReader) can be used many times: the instances
are distinguished by a unique name given by user. Each instance has its own set of parameters 
(so one instance of image reader can be used to read dark frames used for calibration and 
another instance to read actual images).

After all processors are created and their parameters are set, luiza::ProcessorMgr is called to 
perform following analysis steps:
@li \b init() method is called for all active processors \n 
     Processors can do required initialization. 
     No data structures are initialized at this point yet. \n\n 
@li \b startAnalysis() method is called for all active processors \n 
     Pointer to GLORIA data structure is passed to the processor, so required data structures
     can be initialized or additional input read. \n\n 
@li \b processData() method is called in a loop for all active processors in an order specified by user.  \n
   One loop usually corresponds to analysis of one image, but it is also possible to analyze a set
   of images in one loop. The analysis is interrupted when all input files are closed (no more input) 
   or the specified number of loops was reached. \n
   After each \a processData() call \b check() method from the same processor is called to allow 
   additional diagnostics, if not disabled in the steering file. \n
   After data processing loop is finished, all data collections are cleared to allow reading
   and analyzing subsequent images. However, if image or table collection is declared \a permanent, 
   then its contents is kept and can be accessed in next loops. This can be the case e.g. for 
   calibration images (darks, flats) or for analysis results (reconstructed light curves etc.) \n\n 
@li \b concludeAnalysis() method is called for all active processors \n
     Processors can perform final calculations, write results to files or print running statistics. 
     In particular, results stored in permanent collections should be saved, as
     after this step all data stored in GLORIA data structure is lost.\n\n 
@li \b end() method is called for all active processors \n
     Processors can close open files, clean temporary storage etc.

These six methods should be defined for each luiza::Processor. Processors supplying input for analysis
(reading input files or downloading from the net) should be defined as luiza::DataSourceProcessor.
*/

*****************************************************************************

/*!

@page steering_file How to prepare steering file

In this page we would like to describe how one can configure Luiza
and processors selected for the analysis

All relevant input in the XML steering should be enclosed by \a luiza tag:
@code
<luiza>
  .
  .
  .
</luiza>
@endcode

The XML file consists of two general sections and one parameter section for each active
processor defined by user. Active processor list (and order of execution) is defined by
the \a execute section. It could look like:
@code
 <execute>
  <processor name="DarkImageReader"/>
  <processor name="RawImageReader"/>  
  <processor name="CalibrateImage"/>
  <processor name="FindObjects"/>  
  <processor name="DoPhotometry"/>  
  <processor name="DoAstrometry"/>  
  <processor name="StoreObjectList"/>  
  <processor name="StoreFinalImage"/>  
 </execute>
@endcode
where processor names given are unique names defined by user (note that same processor type 
can be used many times, e.g. with different parameter settings, under different names).

The second general section is used to set \a global parameters, which control performance 
of Luiza and all processors
@code
 <global>
  <parameter name="MaxRecordNumber" type="int"> 50  </parameter>  
  <parameter name="NetworkXMLFile" value="my_telescopes.xml"/>
  <parameter name="SupressCheck" type="bool"> false </parameter>  
  <parameter name="Verbosity" options="DEBUG,MESSAGE,WARNING,ERROR,SILENT"> MESSAGE  </parameter> 
 </global>
@endcode
Their functions are the following:
@li \a MaxRecordNumber parameter sets the maximum number of loops in data processing 
    (in most applications, when images are read one by one, is is equivalent to the number 
    of images to be processed).
@li \a NetworkXMLFile parameter is used to define the network configuration file name, 
     see @ref network_configuration 
@li \a SupressCheck is a flag to supress calling check() method of processors. 
@li \a Verbosity parameter allows user to set the global verbosity level for the Luiza job.
    This is used for all processors, for which verbosity level is not set in processor parameter section.

Following sections, tagged with processor names (as given by user and used in \a <execute> section), 
define parameters (and resulting performance) of subsequent processors. 
Order of processor sections in XML file is not relevant. For example,
parameters for \a DarkImageReader processor (as in example above) which is an instance of
luiza::FitsImageReader could look like that:
@code
 <processor name="DarkImageReader" type="FitsImageReader">
  <!--Processor for reading input FITS images. Reads images from given files-->
  <!--List of FITS files to be read-->
  <parameter name="FitsFileList" type="StringVec"> 
       data/darkimage1.fit  data/darkimage2.fit  data/darkimage3.fit
  </parameter> 
  <!--Name of the image collection to which images from file should be stored-->
  <parameter name="ImageCollectionName" type="string">DarkImages </parameter>
  <!--Number of images to be read per processing loop (0 for all))-->
  <parameter name="ImagesPerLoop" type="int">1 </parameter>
  <!--Name of file containing FITS file list (one per line)-->
  <parameter name="ListFileName" type="string"> </parameter>
  <!--Flag for collections, which should not be deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool">false </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string"> MESSAGE </parameter>
</processor>
@endcode

To get a template steering file, with list of all parameters defined in the code 
of all available processors you can use the dedicated option:
@code
${LUIZA_ROOT}/luiza/bin/Luiza -x > template.xml
@endcode
In the generated template, each parameter line will be preceded by a comment line
describing this parameter (as in the example above).

One can also use LuizaGUI to edit steering file or create a new one.
@code
${LUIZA_ROOT}/luiza/bin/LuizaGUI steering_file.xml
@endcode
LuizaGUI checks basic consistency of the steering file and the defined data flow.

*/

*****************************************************************************

/*!

@page image_analysis Image analysis chain

In this page processors available for image analysis are described.
All processors required for standard image reduction are already
available. Other, more advanced tools will be added in the future.

\li \b luiza::FitsImageReader 
\n
\n
This processor can be used to read images from fit files.
List of file names or name of the file containing file names should be specified.
By default one image is read for each analysis pass, but this can be changed by
setting \a ImagesPerLoop parameter. Read images are stored in an image collection
with name given by \a ImageCollectionName parameter.

\li \b luiza::SimpleImageAdder
\n
\n
This processor can be used to add images stored in input collection and store resulting
frames into output collection. User has to specify \a InputCollectionName and 
\a OutputCollectionName. The default behaviour is to add all images in input collection
(in given analysis loop) into one output image. This can be changed using
\a NumberAddImages and \a UseManyLoops parameters. User can also decide to average
images instead of adding them, by specifying \a AverageImages flag.

\li \b luiza::ImageMedian
\n
\n
Processor to calculate median image from images in the input
collection. It is an alternative to simple adding the images, eg. in
case of dark frames. All images from the given collection are used to
calculate median image. 

\li \b luiza::ImageNormalizationProcessor
\n
\n
This processor can be used to normalize raw images ie. subtract dark and bias images, 
and divide by flat frame. Corrected frames are stored in a new collection, if specified, 
otherwise old images are modified. Normalizaction of flat frames (if raw images are used) 
is also possible, as is the automatic weigth calculation for dark and bias images, based 
on raw image and dar image exposure times (see processor description for details).


\li \b luiza::SimpleImageRotator
\n
\n
Simple processor implementing basic image geometry manipulations. Includes image rotation
(by 90, 180 or 270 degree), image flipping (reflection with respect to horizontal or vertical 
axis), as well as cropping and rebinning.


\li \b luiza::PixelClusterFinder
\n
\n
Simple processor to find pixel clusters (ie. stars) on the CCD frame.
This algorithm searches for groups of pixels with signal above the defined threshold. 
The seed pixel is searched first (with tighter seed selection cut) and then the neighbouring pixels, 
with signal above pixel signal threshold (which can be looser) are added to form the cluster. 
The algroithm assumes that the cluster should have one maximum only, if two maxima 
(above seed threshold) are found, two clusters will be reconstructed. To define thresholds, 
background level and average noise level are calculated first. For images with significant
background variation over the frame, background level can be calculated for smaller subframes
and then interpolated (linear and 2nd order interpolation available).


\li \b luiza::SimpleSourceFinder 
\n
\n
Another simple processor to find stars on the frame. 
This processor is based on a Python library \a Mahotas, which
has to be installed first (<a href="http://luispedro.org/software/mahotas">Mahotas page</a>)


\li \b luiza::ImageLaplace
\n
\n
Processor to apply Laplace transformation to images. Laplace transformation 
can be used to "sharpen" the image and remove (variable) background contribution.
This can improve object finding efficiency in some cases. However, precision of
object brightness determination on the resulting image is largly
reduced. After finding objects on the transformed image one should
consider using aperture photometry on the oryginal image (see examples).

\li \b luiza::FastAperturePhotometry
\n
\n
Processor to calculate aperture photometry on CCD images. If object list 
is specified on input, photometry is calculated for all objects in the list and
results added to the table. For the targeted analysis, coordinates of the objects 
of interest can be specified. However, valid astrometry has to be found in
the image fits header in that case.


\li \b luiza::WeightedAperturePhotometry 
\n
\n
 Processor to calculate aperture photometry on CCD images in the input
 collection. If input object collection is specified, photometry is
 calculated for all objects in the object list. For the targeted
 analysis, coordinates of the objects of interest can be
 specified. However, valid astrometry has to be found in the image
 fits header in that case. Two approaches for background calculation
 are possible: either it is calculated from media of pixels in annulus
 around object position (when \a AnnulusRadii is specified) or from the
 background map calculated for the whole image (as in luiza::BackgroundImage
 processor). The algorithm used differs from the one implemented in
 luiza::FastAperturePhotometry by introducing weights for pixels which
 are at the aperture boundary. The weight depends also on the exact
 position of the object. 

\li \b luiza::BackgroundImage
\n
\n
 Processor calculates background map from given image. The frame is
 divided into \a FrameDivision x \a FrameDivision subframes, then background
 level and noise are calculated from median in each subframe. Finally,
 background and/or noise image is created by interpolating between
 subframe centers. New images have the same resolution as the input
 image. 



\li \b luiza::RootImageViewer
\n
\n
Processor allowing to view images during processing.
This processor uses CERN root classes to display images from collection 
given by \a ImageCollectionName parameter. Images can be displayed as
images or as 2D histograms. In the later case it is possible to overlay objects
from the give object list (eg. to check performance of object findind algorithm)

\li \b luiza::FitsImageWriter
\n
\n
This processor can be used to write modified images or new images resulting
from the analysis to the standard fit files. User should specify the file name 
root (\a FitsFileNameRoot parameter) - it will be appended by image number 
and ".fit" extension. All images read from an input image collection 
given by \a ImageCollectionName are stored to file. Format of the output file
can be set with \a BitsPerPixel parameter.

\li \b luiza::TestProcessor
\n
\n
This processor prints out all collections and dimensions of all images found.
Can be used to monitor performance of other processors.


\li \b luiza::ExampleProcessor
\n
\n
This processor prints out image statistics from given collection.
It can be compiled as a separate library loaded dynamicly by Luiza.
It is to be used as a template for developing user processors.

*/

*****************************************************************************

/*!

@page object_analysis Object analysis chain


\li \b luiza::SelectTableData
\n
\n
 This processor can be used to select rows of the table
 (eg. catalog entries or reconstricted objects in the object list) by
 specifying simple cuts on values stored in table columns. Cuts can be
 applied on integer or float columns only (no string nor vector
 columns can be used for selection). Additional flag allows for
 sorting the output table by selected column value. 

\li \b luiza::Astrometry
\n
\n
Processor to apply Astrometry.net algorithm for finding frame
orientation and transformation for calculating object positions
in the sky. Analyses object lists, but if image collection is specified 
as well, calculated astrometry parameters are added to the image FITS header
(and stored when writing image to the output file).
Additional options are included to improve astrometry efficiency, in particular
aiming at rejection of artificial objects due to blooming or halo. User can also limit
the number of objects considered in the algorithm. Brightest objects are always selected.

\li \b luiza::MatchSkyCatalog
\n
\n
This processor matches objects in the table of the input collection
 with sky objects stored in catalog table (gloria::GloriaSkyCatalog),
 based on their coordinates on the sky (RA and DEC from
 astrometry). Selected entries (columns) from the catalog can also be
 copied into the object table (for matched entries). Matched catalog
 brightness can be then used eg. for image calibration. 



\li \b luiza::CalibrateFrameObjects
\n
\n
This processor analyses the object list of the input collection and
 normalizes all brightness measurements based on the reference stars
 found. Reference stars can be searched for in the gloria::GloriaSkyCatalog
 collection or selected from the object table itself (if already
 matched with the catalog, eg. with luiza::MatchSkyCatalog). AverageMap is
 used to create calibration map for the frame, based on the identified
 reference stars. The map is then applied to all objects in the table,
 estimated calibration error is also calculated. 


\li \b luiza::ObjectLightCurve
\n
\n
Processor analyses the input object list and finds measurements of the selected 
object based on its coordinates (on the sky or on the CCD). The measurement is then
added to the light curve. It is also possible to normalize the brightness measurement 
to the reference stars by giving their positions and brightness.

\li \b luiza::SelectNewObjects
\n
\n
Processor matches objects from the input list with
objects from the previous frame and selects new ones. 
Additional cut can be applied on the signal to noise ratio. 
If astrometry is stored in the object list, it is also verified
that the object was within the previous field of view. 

\li \b luiza::TableTextReader
\n
\n
This processor reads data tables from simple text files. It is be used for
reading eg. reference star list object lists stored previously. It is
defined as luiza::DataSourceProcessor, so it can be the first
processor in the analysis chain. 


\li \b luiza::TableTextWriter
\n
\n
This processor can be used to write created tables to output TEXT files.
User should specify the file name root (\a TextFileNameRoot parameter) - it will 
and ".tbl" extension. All tables read from an input table collection 
given by \a TableCollectionName are stored to file. User can select table columns 
to be stored.

\li \b luiza::FitsTableReader 
\n
\n
This processor reads data tables from FITS files. It can be used for
reading eg. star catalog or previously stored object lists. It is
 defined as luiza::DataSourceProcessor, so it can be  the first
 processor in the analysis chain.  

\li \b luiza::FitsTableWriter
\n
\n
This processor writes tables to output FITS file It can be used to
store analysis results, eg. object lists. Binary as well as text table
formats are supported. 


\li \b luiza::RootTreeWriter
\n
\n
This processor converts tables to root tree and stores as binary root
 file. They can then loaded directly into root for final analysis.


*/

*****************************************************************************

/*!

@page databases Database access

\li \b luiza::SociSelect
\n
\n
This processor reads data from SQL database and puts them into
internal FITS table structure, so it can be used in subsequent
processing steps. SOCI library is used to implement interface to
different databases. This processor is designed to execute
SELECT queries only. For detailed description of command syntax
please refer to <a href="http://soci.sourceforge.net/doc.html">SOCI documentation</a>.

*/

*****************************************************************************

/*!
@page examples More examples

Here we present few "real life" examples of Luiza processing tasks.
The steering files presented were used in the analysis of 
the Pi of the Sky data. Three steps of the analysis are presented.

<b>Reconstructing flat image from dedicated frames</b>

This simple image processing task can be divided into the following steps:
@li Read flat images for each night considered
@li Average flat images for each night
@li Read dark images for each night
@li Calculate master dark (median) for each night
@li Subtract master dark from average flat for each night
@li Calculate average flat (for all nights)
@li Smooth resulting flat image
@li Write final image to file

@code
<luiza>

   <execute>
      <processor name="FlatReader"/>
      <processor name="FlatAverage"/>
      <processor name="DarkReader"/>
      <processor name="DarkMedian"/>
      <processor name="DarkSubtraction"/>
      <processor name="FlatAverage2"/>
      <processor name="FlatSmooth"/>
      <processor name="FlatViewer"/>
      <processor name="FlatWriter"/>
   </execute>

   <global>
      <parameter name="MaxRecordNumber" value="1000"/>
      <parameter name="NetworkXMLFile" value="gloria_network.xml"/>
      <parameter name="SupressCheck" value="false"/>
      <parameter name="Verbosity" value="MESSAGE"/>
   </global>

 <processor name="FlatReader" type="FitsImageReader">
 <!--Processor for reading input FITS imagesReads images from given files-->
  <!--Name of the image collection to which images from file should be stored-->
  <parameter name="ImageCollectionName" type="string" gloriaOutType="Image"> FlatImages </parameter>
  <!--List of FITS files to be read-->
  <parameter name="FitsFileList" type="StringVec"> </parameter>
  <!--Number of images to be read per processing loop (0 for all))-->
  <parameter name="ImagesPerLoop" type="int" value="4"/>
  <!--Names of files containing FITS file lists (one per line)-->
  <parameter name="ListFileList" type="StringVec" value="good_flats.list"/>
  <!--Flag for collections, which should not be deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool" value="false"/>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

 <processor name="FlatAverage" type="SimpleImageAdder">
 <!--Processor to add images in the input collection.No corrections for possible shifts are applied.-->
  <!--Flag to select adding (false) or averaging (true) of input images-->
  <parameter name="AverageImages" type="bool">true </parameter>
  <!--Name of the collection containing images to be added-->
  <parameter name="InputCollectionName" type="string" lcioInType="1">FlatImages  </parameter>
  <!--Number of images to be added each time (0 for all images in input collection)-->
  <!--parameter name="NumberAddImages" type="int">0 </parameter-->
  <!--Name of the collection where image sums will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="1">FlatAverage </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Wait for images in next loops, if not enough images in input collection-->
  <!--parameter name="UseManyLoops" type="bool">false </parameter-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string">MESSAGE </parameter-->
</processor>

 <processor name="DarkReader" type="FitsImageReader">
 <!--Processor for reading input FITS imagesReads images from given files-->
  <!--Name of the image collection to which images from file should be stored-->
  <parameter name="ImageCollectionName" type="string" gloriaOutType="Image"> DarkImages </parameter>
  <!--List of FITS files to be read-->
  <parameter name="FitsFileList" type="StringVec"> </parameter>
  <!--Number of images to be read per processing loop (0 for all))-->
  <parameter name="ImagesPerLoop" type="int" value="20"/>
  <!--Names of files containing FITS file lists (one per line)-->
  <parameter name="ListFileList" type="StringVec" value="good_darks.list"/>
  <!--Flag for collections, which should not be deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool" value="false"/>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

 <processor name="DarkMedian" type="ImageMedian">
 <!--Processor to calculate median image from images in the input collection.-->
  <!--Name of the collection containing input images-->
  <parameter name="InputCollectionName" type="string" gloriaInType="Image"> DarkImages </parameter>
  <!--Name of the collection where median image will be stored-->
  <parameter name="OutputCollectionName" type="string" gloriaOutType="Image"> DarkMedian </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool" value="false"/>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

 <processor name="DarkSubtraction" type="ImageNormalizationProcessor">
 <!--Simple processor to apply dark subtraction and flat correction-->
  <!--Flag for using exposure time information (from fits header) for calculating correction factors-->
  <parameter name="AutoNormalization" type="bool">false </parameter>
  <!--Name of the collection containing bias frames-->
  <parameter name="BiasCollectionName" type="string" lcioInType="1"> </parameter>
  <!--Factor used to scale bias image-->
  <!--parameter name="BiasScale" type="double">0 </parameter-->
  <!--Name of the collection containing dark images-->
  <parameter name="DarkCollectionName" type="string" lcioInType="1"> DarkMedian</parameter>
  <!--Factor used to scale dark image-->
  <!--parameter name="DarkScale" type="double">1 </parameter-->
  <!--Name of the collection containing flat frames-->
  <parameter name="FlatCollectionName" type="string" lcioInType="1"> </parameter>
  <!--Flag for applying dark/bias subtraction to the flat frame image-->
  <parameter name="FlatNormalization" type="bool">false </parameter>
  <!--Factor used to scale flat frame-->
  <!--parameter name="FlatScale" type="double">1 </parameter-->
  <!--Factor used to scale image pixels-->
  <!--parameter name="ImageScale" type="double">1 </parameter-->
  <!--Name of the collection containing images to be corrected-->
  <parameter name="InputCollectionName" type="string" lcioInType="1"> FlatAverage </parameter>
  <!--Name of the collection where image sums will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="1"> Flat </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string">MESSAGE </parameter-->
</processor>

 <processor name="FlatAverage2" type="SimpleImageAdder">
 <!--Processor to add images in the input collection.No corrections for possible shifts are applied.-->
  <!--Flag to select adding (false) or averaging (true) of input images-->
  <parameter name="AverageImages" type="bool">true </parameter>
  <!--Name of the collection containing images to be added-->
  <parameter name="InputCollectionName" type="string" lcioInType="1">Flat  </parameter>
  <!--Number of images to be added each time (0 for all images in input collection)-->
  <parameter name="NumberAddImages" type="int">2 </parameter>
  <!--Name of the collection where image sums will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="1">FlatSum </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Wait for images in next loops, if not enough images in input collection-->
  <parameter name="UseManyLoops" type="bool">true </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string">MESSAGE </parameter-->
</processor>

 <processor name="FlatSmooth" type="BackgroundImage">
 <!--Processor to create background/noise image from the frame.-->
  <!--Name of the collection where background image will be stored-->
  <parameter name="BackgroundCollectionName" type="string" lcioOutType="1"> FinalFlat</parameter>
  <!--Only every BackgroundSampling pixel is considered in background calculation-->
  <!--parameter name="BackgroundSampling" type="int">1 </parameter-->
  <!--Frame division for background calculation-->
  <parameter name="FrameDivision" type="int">16 </parameter>
  <!--Name of the collection containing input images-->
  <parameter name="InputCollectionName" type="string" lcioInType="1">FlatSum </parameter>
  <!--Polinomial order used for background map interpolation-->
  <parameter name="InterpolationOrder" type="int">2 </parameter>
  <!--Name of the collection where noise image will be stored-->
  <parameter name="NoiseCollectionName" type="string" lcioOutType="1"> </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Use network configuration file to update processor parameters on even by event basis-->
  <!--parameter name="UseTelescopeParameters" type="bool">false </parameter-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string">MESSAGE </parameter-->
</processor>

 <processor name="FlatWriter" type="FitsImageWriter">
 <!--Processor for writing FITS images to diskWrite images to specified files-->
  <!--Name of the image collection which should be stored to file-->
  <parameter name="ImageCollectionName" type="string" gloriaInType="Image"> FinalFlat </parameter>
  <!--Bits per pixel in the output fits file (0 to keep image format unchanged)-->
  <parameter name="BitsPerPixel" type="int" value="16"/>
  <!--Output file name root, which will be appended by the image number-->
  <parameter name="FitsFileNameRoot" type="string" value="luiza_output/flat_image"/>
  <!-- BZERO keyword value, used when storing pixel values to fits file (default is 0, use 32768 for unsigned short)-->
  <parameter name="PixelOffset" type="int" value="32768"/>
  <!-- BSCALE keyword value, used when storing pixel values to fits file (default is 1.0)-->
  <!--parameter name="PixelScale" type="double" value="1"/-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string" value="DEBUG"/>
</processor>

 <processor name="FlatViewer" type="RootImageViewer">
 <!--Simple image viewer for Luiza. Based on CERN root image class-->
  <!--Name of the image collection to view-->
  <parameter name="ImageCollectionName" type="string" gloriaInType="Image"> FinalFlat </parameter>
  <!--Name of the object list collection, corresponding to the displayed images-->
  <parameter name="ListCollectionName" type="string" gloriaInType="Table">  </parameter>
  <!--Sorry, no description for this parameter-->
  <parameter name="Table" type="string" gloriaMissing="Table">  </parameter>
  <!--Name of the root canvas. Default is the image collection name.-->
  <!--parameter name="CanvasName" type="string" value=""/-->
  <!--Title to be shown on the root canvas-->
  <!--parameter name="CanvasTitle" type="string" value=""/-->
  <!--Canvas geometry: width, height, x offset, y offset-->
  <!--parameter name="CanvasWH" type="IntVec"> </parameter-->
  <!--Time (in seconds) to pause after dispying image. Waits for <Enter> when set to -1-->
  <parameter name="PauseTime" type="int" value="-1"/>
  <!--When set to true image is shown as root 2D histogram (TH2D)-->
  <!--parameter name="ShowHistogram" type="bool" value="false"/-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

</luiza>

@endcode


<b>Normalizing and running astrometry for collected frames</b>

This analysis, preformed on the series of images from one night,
includes following steps: 
@li Read flat image (only once)
@li Read dark images (only once)
@li Calculate master dark image (median from collected images)
@li Read input raw image  (source of data for main analysis loop)
@li Normalize image (subtract dark, divide by flat)
@li Find stars on the image (pixel clusters)
@li Reject "bad" clusters (hotpixels etc.)
@li Run astrometry
@li Write final image to file

One should notice that by giving path name only as the file root
(\a FitsFileNameRoot parameter in \a ImageWriter processor)
the output images will have the same file names as the input images
(but will be placed in different catalog).  

@code
<luiza>

   <execute>
      <processor name="FlatReader"/>
      <processor name="DarkReader"/>
      <processor name="DarkMedian"/>
      <processor name="ImageReader"/>
      <processor name="NormalizeFrame"/>
      <processor name="ClusterFinder"/>
      <processor name="ClusterSelection"/>
      <processor name="Astrometry"/>
      <processor name="ImageWriter"/>  
   </execute>

   <global>
      <parameter name="MaxRecordNumber" value="10000"/>
      <parameter name="NetworkXMLFile" value="gloria_network.xml"/>
      <parameter name="SupressCheck" value="false"/>
      <parameter name="Verbosity" value="MESSAGE"/>
   </global>

 <processor name="FlatReader" type="FitsImageReader">
 <!--Processor for reading input FITS imagesReads images from given files-->
  <!--Name of the image collection to which images from file should be stored-->
  <parameter name="ImageCollectionName" type="string" gloriaOutType="Image"> FlatImages </parameter>
  <!--List of FITS files to be read-->
  <parameter name="FitsFileList" type="StringVec"> </parameter>
  <!--Number of images to be read per processing loop (0 for all))-->
  <parameter name="ImagesPerLoop" type="int" value="0"/>
  <!--Names of files containing FITS file lists (one per line)-->
  <parameter name="ListFileList" type="StringVec" value="flat_frames.list"/>
  <!--Flag for collections, which should not be deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool" value="true"/>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

 <processor name="DarkReader" type="FitsImageReader">
 <!--Processor for reading input FITS imagesReads images from given files-->
  <!--Name of the image collection to which images from file should be stored-->
  <parameter name="ImageCollectionName" type="string" gloriaOutType="Image"> DarkImages </parameter>
  <!--List of FITS files to be read-->
  <parameter name="FitsFileList" type="StringVec"> </parameter>
  <!--Number of images to be read per processing loop (0 for all))-->
  <parameter name="ImagesPerLoop" type="int" value="0"/>
  <!--Names of files containing FITS file lists (one per line)-->
  <parameter name="ListFileList" type="StringVec" value="dark_frames.list"/>
  <!--Flag for collections, which should not be deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool" value="false"/>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

 <processor name="DarkMedian" type="ImageMedian">
 <!--Processor to calculate median image from images in the input collection.-->
  <!--Name of the collection containing input images-->
  <parameter name="InputCollectionName" type="string" gloriaInType="Image"> DarkImages </parameter>
  <!--Name of the collection where median image will be stored-->
  <parameter name="OutputCollectionName" type="string" gloriaOutType="Image"> DarkMedian </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool" value="true"/>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

 <processor name="ImageReader" type="FitsImageReader">
 <!--Processor for reading input FITS imagesReads images from given files-->
  <!--Name of the image collection to which images from file should be stored-->
  <parameter name="ImageCollectionName" type="string" gloriaOutType="Image"> RawImages </parameter>
  <!--List of FITS files to be read-->
  <parameter name="FitsFileList" type="StringVec"> </parameter>
  <!--Number of images to be read per processing loop (0 for all))-->
  <parameter name="ImagesPerLoop" type="int" value="1"/>
  <!--Names of files containing FITS file lists (one per line)-->
  <parameter name="ListFileList" type="StringVec" value="sky_frames.list"/>
  <!--Flag for collections, which should not be deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool" value="false"/>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

 <processor name="NormalizeFrame" type="ImageNormalizationProcessor">
 <!--Simple processor to apply dark subtraction and flat correction-->
  <!--Flag for using exposure time information (from fits header) for calculating correction factors-->
  <parameter name="AutoNormalization" type="bool">false </parameter>
  <!--Name of the collection containing bias frames-->
  <parameter name="BiasCollectionName" type="string" lcioInType="1"> </parameter>
  <!--Factor used to scale bias image-->
  <!--parameter name="BiasScale" type="double">0 </parameter-->
  <!--Name of the collection containing dark images-->
  <parameter name="DarkCollectionName" type="string" lcioInType="1"> DarkMedian</parameter>
  <!--Factor used to scale dark image-->
  <!--parameter name="DarkScale" type="double">1 </parameter-->
  <!--Name of the collection containing flat frames-->
  <parameter name="FlatCollectionName" type="string" lcioInType="1"> FlatImages </parameter>
  <!--Flag for applying dark/bias subtraction to the flat frame image-->
  <parameter name="FlatNormalization" type="bool">false </parameter>
  <!--Factor used to scale flat frame-->
  <parameter name="FlatScale" type="double">0.00003 </parameter>
  <!--Factor used to scale image pixels-->
  <!--parameter name="ImageScale" type="double">1 </parameter-->
  <!--Name of the collection containing images to be corrected-->
  <parameter name="InputCollectionName" type="string" lcioInType="1"> RawImages </parameter>
  <!--Name of the collection where image sums will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="1"> Images </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string">MESSAGE </parameter-->
</processor>


<processor name="ClusterFinder" type="PixelClusterFinder">
 <!--Processor to find pixel clusters on the frame.-->
  <!--Name of the collection containing input images-->
  <parameter name="InputCollectionName" type="string" lcioInType="1">Images </parameter>
  <!--When set to true: search pixel neighbours in 5x5 square rather than 3x3-->
  <parameter name="MoreNeighbours" type="bool">true </parameter>
  <!--Maximum number of clusters to be reconstructed-->
  <parameter name="MaxClusters" type="int"> -1 </parameter>
  <!--Maximum number of pixels considered in cluster CoG and signal calculation-->
  <parameter name="MaxPixels" type="int"> 25 </parameter>
  <!--Name of the collection where cluster list will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2">  Objects</parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Threshold for the cluster pixel signal (in units of noise level)-->
  <parameter name="PixelThreshold" type="double">3 </parameter>
  <!--Threshold for the cluster seed signal (in units of noise level)-->
  <parameter name="SeedThreshold" type="double">8 </parameter>
  <!--Only every BackgroundSampling pixel is considered in background calculation-->
  <parameter name="BackgroundSampling" type="int">3 </parameter>
  <!--Frame division for background calculation-->
  <parameter name="FrameDivision" type="int">8 </parameter>
  <!--Flag to add charge distribution shape information to the output table.-->
  <parameter name="AddShape" type="bool">true </parameter>
  <!--Flag to add charge spread information to the output table.-->
  <parameter name="AddSpread" type="bool">true </parameter>
  <!--Use network configuration file to update processor parameters on even by event basis-->
  <parameter name="UseTelescopeParameters" type="bool">false</parameter>
  <!--verbosity level of this processor ("DEBUG0-4,MESSAGE0-4,WARNING0-4,ERROR0-4,SILENT")-->
  <parameter name="Verbosity" type="string">MESSAGE</parameter>
</processor>

 <processor name="ClusterSelection" type="SelectTableData">
 <!--Processor to select table data.-->
  <!--Choose selection based on OR of cuts (when true) or AND of cuts (when false).-->
  <!--parameter name="AcceptIfAny" type="bool">false </parameter-->
  <!--Names of the columns and allowed ranges of values (3 fields per cut)-->
  <parameter name="ColumnValueRange" type="StringVec"> 
    CCD_X 50. 2010.
    CCD_Y 50. 1990.
    Pixels 2     0
    ELONG  0.  0.9
  </parameter>
  <!--Name of the collection containing input tables-->
  <parameter name="InputCollectionName" type="string" lcioInType="2">Objects </parameter>
  <!--Name of the collection where output table will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2">Selected </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Flag to sort output table in reverse order.-->
  <parameter name="ReverseSort" type="bool">true </parameter>
  <!--Name of the column used for table sorting-->
  <parameter name="SortColumn" type="string"> Signal </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">DEBUG </parameter>
</processor>


<processor name="Astrometry" type="Astrometry">
  <!--Performs astrometry (astrometry.net based) on list of stars on the frame-->
  <!--Add RA DEC positions from astrometry to the input object list-->
  <parameter name="AddPosition" type="bool">true </parameter>
  <!--Verbosity of astrometry.net output-->
  <parameter name="AstVerbosity" type="int">0 </parameter>
  <!--Depth limit for the astrometry algorithm (default 0)-->
  <!--parameter name="AstroDepth" type="int">0 </parameter-->
  <!--Order of the SIP polynomial in astrometry fit-->
  <!--parameter name="AstroOrder" type="int">3 </parameter-->
  <!--Minimum distance between objects along vertical or horizontal line-->
  <parameter name="BloomingCut" type="int">100 </parameter>
  <!--Maximum number of objects to be used for astrometry.-->
  <parameter name="MaxObjects" type="int"> 1000 </parameter>
  <!--Solve using indexes parallely or not-->
  <parameter name="InParallel" type="bool">false</parameter>
  <!--Name of the collection containing images to be added-->
  <parameter name="ImageCollectionName" type="string" lcioInType="1">Images  </parameter>
  <!--Name of the object list collection, corresponding to the analysed images-->
  <parameter name="ListCollectionName" type="string" lcioInType="2"> Selected</parameter>
  <!--Name of the collection where object list will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2"> </parameter>
  <!--Recalculate and overwrite astrometry if exists-->
  <parameter name="Overwrite" type="bool"> true </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Upper boundary of possible pixel sizes in arcseconds-->
  <parameter name="PixscaleHigh" type="double">40 </parameter>
  <!--Lower boundary of possible pixel sizes in arcseconds-->
  <parameter name="PixscaleLow" type="double">30 </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">MESSAGE </parameter>
  <!--Path to astrometric (astrometry.net) indexes files-->
  <parameter name="IndexesPath" type="string">astrometry_indexes</parameter>
  <!--parameter name="IndexesPath" type="string">bootes_indexes </parameter-->
</processor>


 <processor name="ImageWriter" type="FitsImageWriter">
 <!--Processor for writing FITS images to diskWrite images to specified files-->
  <!--Bits per pixel in the output fits file (0 to keep image format unchanged)-->
  <parameter name="BitsPerPixel" type="int">16 </parameter>
  <!--Output file name root, which will be appended by the image number-->
  <parameter name="FitsFileNameRoot" type="string">astrometry_output/ </parameter>
  <!--Name of the image collection which should be stored to file-->
  <parameter name="ImageCollectionName" type="string" lcioInType="1">Images </parameter>
  <!-- BZERO keyword value, used when storing pixel values to fits file (default is 0, use 32768 for unsigned short)-->
  <parameter name="PixelOffset" type="int">32768 </parameter>
  <!-- BSCALE keyword value, used when storing pixel values to fits file (default is 1.0)-->
  <!--parameter name="PixelScale" type="double">1 </parameter-->
  <!--verbosity level of this processor ("DEBUG0-4,MESSAGE0-4,WARNING0-4,ERROR0-4,SILENT")-->
  <parameter name="Verbosity" type="string">DEBUG </parameter>
</processor>

</luiza>

@endcode


<b>Matching objects on the image with sky catalog and image calibration</b>

This analysis task can be divided into following steps
@li Read star catalogs (local copy of Tycho converted to FITS table format)
@li Select catalog objects in the required RA,Dec range
@li Read input images 
    (resulting from the previous analysis, i.e. normalized and with astrometry results stored) 
@li Find stars (pixel clusters)
@li  Reject "bad" clusters
@li Calculate aperture photometry for clusters
@li Apply astrometry results
@li Match objects with the catalog
@li Calibrate frame base on Tycho V magnitudo
@li Write calibrated objects to root tree file

@code
<luiza>
 <execute>
  <processor name="TychoReader"/>
  <processor name="TychoSelection"/>
  <processor name="ImageReader"/>
  <processor name="ClusterFinder"/>
  <processor name="ClusterSelection"/>
  <processor name="AperturePhotometry"/>
  <processor name="Astrometry"/>
  <processor name="MatchTycho"/>
  <processor name="CalibrateFrame"/>
  <processor name="ObjectWriter"/>
 </execute>

 <global>
  <parameter name="MaxRecordNumber" type="int">10000 </parameter>  
  <parameter name="NetworkXMLFile" value="gloria_network.xml"/>
  <parameter name="SupressCheck" type="bool"> false </parameter>  
  <parameter name="Verbosity" options="DEBUG,MESSAGE,WARNING,ERROR,SILENT">MESSAGE</parameter> 
 </global>

 <processor name="TychoReader" type="FitsTableReader">
  <parameter name="ListFileList" type="StringVec"> </parameter>
  <parameter name="TableCollectionName" type="string" lcioOutType="2"> TychoAll </parameter>
  <parameter name="TableFileList" type="StringVec"> tycho-2.tbl </parameter>
  <parameter name="PermanentCollection" type="bool">false </parameter>
  <!--Number of table files to be read per processing loop (0 for all))-->
  <!--parameter name="TablesPerLoop" type="int">1 </parameter-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <!--parameter name="Verbosity" type="string">MESSAGE </parameter-->
</processor>

 <processor name="TychoSelection" type="SelectTableData">
 <!--Processor to select table data.-->
  <!--Choose selection based on OR of cuts (when true) or AND of cuts (when false).-->
  <!--parameter name="AcceptIfAny" type="bool">false </parameter-->
  <!--Names of the columns and allowed ranges of values (3 fields per cut)-->
  <parameter name="ColumnValueRange" type="StringVec"> 
    RA 35. 65.
    DEC -35. -5.
    prox 998 1000
  </parameter>
  <!--Name of the collection containing input tables-->
  <parameter name="InputCollectionName" type="string" lcioInType="2">TychoAll </parameter>
  <!--Name of the collection where output table will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2">Tycho </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool">true </parameter>
  <!--Flag to sort output table in reverse order.-->
  <parameter name="ReverseSort" type="bool"> </parameter>
  <!--Name of the column used for table sorting-->
  <parameter name="SortColumn" type="string"> </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">DEBUG </parameter>
</processor>

 <processor name="ImageReader" type="FitsImageReader">
 <!--Processor for reading input FITS imagesReads images from given files-->
  <!--List of FITS files to be read-->
  <parameter name="FitsFileList" type="StringVec"> </parameter>
  <!--Name of the image collection to which images from file should be stored-->
  <parameter name="ImageCollectionName" type="string"> Images </parameter>
  <!--Number of images to be read per processing loop (0 for all))-->
  <parameter name="ImagesPerLoop" type="int">1 </parameter>
  <!--Name of file containing FITS file list (one per line)-->
  <parameter name="ListFileList" type="StringVec"> good_astro.list   </parameter>
  <!--Flag for collections, which should not be deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool">false </parameter>
  <!--verbosity level of this processor ("DEBUG0-4,MESSAGE0-4,WARNING0-4,ERROR0-4,SILENT")-->
  <!--parameter name="Verbosity" type="string">DEBUG </parameter-->
</processor>


<processor name="ClusterFinder" type="PixelClusterFinder">
 <!--Processor to find pixel clusters on the frame.-->
  <!--Name of the collection containing input images-->
  <parameter name="InputCollectionName" type="string" lcioInType="1">Images </parameter>
  <!--When set to true: search pixel neighbours in 5x5 square rather than 3x3-->
  <parameter name="MoreNeighbours" type="bool">true </parameter>
  <!--Maximum number of clusters to be reconstructed-->
  <parameter name="MaxClusters" type="int"> -1 </parameter>
  <!--Maximum number of pixels considered in cluster CoG and signal calculation-->
  <parameter name="MaxPixels" type="int"> 25 </parameter>
  <!--Name of the collection where cluster list will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2">  Objects</parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Threshold for the cluster pixel signal (in units of noise level)-->
  <parameter name="PixelThreshold" type="double">3 </parameter>
  <!--Threshold for the cluster seed signal (in units of noise level)-->
  <parameter name="SeedThreshold" type="double">6 </parameter>
  <!--Only every BackgroundSampling pixel is considered in background calculation-->
  <parameter name="BackgroundSampling" type="int">1 </parameter>
  <!--Frame division for background calculation-->
  <parameter name="FrameDivision" type="int">8 </parameter>
  <!--Flag to add charge distribution shape information to the output table.-->
  <parameter name="AddShape" type="bool">true </parameter>
  <!--Flag to add charge spread information to the output table.-->
  <parameter name="AddSpread" type="bool">true </parameter>
  <!--Use network configuration file to update processor parameters on even by event basis-->
  <parameter name="UseTelescopeParameters" type="bool">false</parameter>
  <!--verbosity level of this processor ("DEBUG0-4,MESSAGE0-4,WARNING0-4,ERROR0-4,SILENT")-->
  <parameter name="Verbosity" type="string">MESSAGE</parameter>
</processor>

 <processor name="ClusterSelection" type="SelectTableData">
 <!--Processor to select table data.-->
  <!--Choose selection based on OR of cuts (when true) or AND of cuts (when false).-->
  <!--parameter name="AcceptIfAny" type="bool">false </parameter-->
  <!--Names of the columns and allowed ranges of values (3 fields per cut)-->
  <parameter name="ColumnValueRange" type="StringVec"> 
    CCD_X 30. 2032.
    CCD_Y 20. 2028.
    Pixels 2     0
    ELONG  0.  0.95
  </parameter>
  <!--Name of the collection containing input tables-->
  <parameter name="InputCollectionName" type="string" lcioInType="2">Objects </parameter>
  <!--Name of the collection where output table will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2"></parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Flag to sort output table in reverse order.-->
  <parameter name="ReverseSort" type="bool">true </parameter>
  <!--Name of the column used for table sorting-->
  <parameter name="SortColumn" type="string"> Signal </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">DEBUG </parameter>
</processor>


 <processor name="AperturePhotometry" type="WeightedAperturePhotometry">
 <!--Processor to calculate aperture photometry on the frame.-->
  <!--Inner and outer radius of the annulus used for background subtraction-->
  <parameter name="AnnulusRadii" type="FloatVec">  </parameter>
  <!--Radius of circular aperture used for signal counting in pixels-->
  <parameter name="ApertureRadius" type="float">2.5 </parameter>
  <!--Name of the object brightness column-->
  <parameter name="BrightnessColumn" type="string"> Aper </parameter>
  <!--Name of the collection containing input images-->
  <parameter name="ImageCollectionName" type="string" lcioInType="1"> Images </parameter>
  <!--Name of the collection containing object lists-->
  <parameter name="ListCollectionName" type="string" lcioInType="2">Objects  </parameter>
  <!--Positions of objects of interest (RA,DEC in degrees)-->
  <parameter name="ObjectPositions" type="FloatVec"> </parameter>
  <!--Name of the collection where object list will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2"> </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Only every BackgroundSampling pixel is considered in background calculation-->
  <parameter name="BackgroundSampling" type="int">1 </parameter>
  <!--Frame division for background calculation-->
  <parameter name="FrameDivision" type="int">8 </parameter>
  <!--Polinomial order used for background map interpolation-->
  <parameter name="InterpolationOrder" type="int">1 </parameter>
  <!--Flag to remove clusters with signal below the threshold-->
  <parameter name="RemoveEmpty" type="bool">true </parameter>
  <!--Signal is stored only if above the threshold (in noise units)-->
  <parameter name="SignalThreshold" type="double">4. </parameter>
  <!--Use network configuration file to update processor parameters on even by event basis-->
  <!--parameter name="UseTelescopeParameters" type="bool">false </parameter-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">MESSAGE</parameter>
</processor>

<processor name="Astrometry" type="Astrometry">
  <!--Performs astrometry (astrometry.net based) on list of stars on the frame-->
  <!--Add RA DEC positions from astrometry to the input object list-->
  <parameter name="AddPosition" type="bool">true </parameter>
  <!--Verbosity of astrometry.net output-->
  <parameter name="AstVerbosity" type="int">0 </parameter>
  <!--Depth limit for the astrometry algorithm (default 0)-->
  <!--parameter name="AstroDepth" type="int">0 </parameter-->
  <!--Order of the SIP polynomial in astrometry fit-->
  <!--parameter name="AstroOrder" type="int">3 </parameter-->
  <!--Minimum distance between objects along vertical or horizontal line-->
  <parameter name="BloomingCut" type="int">100 </parameter>
  <!--Maximum number of objects to be used for astrometry.-->
  <parameter name="MaxObjects" type="int"> 1000 </parameter>
  <!--Solve using indexes parallely or not-->
  <parameter name="InParallel" type="bool">false</parameter>
  <!--Name of the collection containing images to be added-->
  <parameter name="ImageCollectionName" type="string" lcioInType="1"> Images  </parameter>
  <!--Name of the object list collection, corresponding to the analysed images-->
  <parameter name="ListCollectionName" type="string" lcioInType="2"> Objects</parameter>
  <!--Name of the collection where object list will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2"> </parameter>
  <!--Recalculate and overwrite astrometry if exists-->
  <parameter name="Overwrite" type="bool">false</parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <!--parameter name="PermanentCollection" type="bool">false </parameter-->
  <!--Upper boundary of possible pixel sizes in arcseconds-->
  <parameter name="PixscaleHigh" type="double">40 </parameter>
  <!--Lower boundary of possible pixel sizes in arcseconds-->
  <parameter name="PixscaleLow" type="double">30 </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">MESSAGE </parameter>
  <!--Path to astrometric (astrometry.net) indexes files-->
  <parameter name="IndexesPath" type="string">astrometry_indexes</parameter>
  <!--parameter name="IndexesPath" type="string">bootes_indexes </parameter-->
</processor>

 <processor name="MatchTycho" type="MatchSkyCatalog">
 <!--Processor to match objects with sky catalog.-->
  <!--Name of the collection containing input object lists-->
  <parameter name="InputCollectionName" type="string" lcioInType="2">Objects </parameter>
  <!--Name of the collection containing star catalog-->
  <parameter name="SkyCatalogName" type="string" lcioInType="2"> Tycho </parameter>
  <!--Position matching tolerance for object and reference stars-->
  <parameter name="MatchingTolerance" type="double">0.015 </parameter>
  <!--Name of the collection where output table will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2"> </parameter>
  <!--Flag output collection, as permanent, not deleted after loop is finished-->
  <parameter name="PermanentCollection" type="bool">false </parameter>
  <!--Names of the position columns in the table-->
  <!--parameter name="PositionColumns" type="StringVec">RA DEC  </parameter-->
  <!--Names of the catalog columns to be copied to the object table-->
  <parameter name="CopyColumns" type="StringVec"> BTmag VTmag</parameter>
  <!--Names which should be used for columns copied from the catalog-->
  <!--parameter name="NewColumnNames" type="StringVec"> </parameter-->
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">DEBUG </parameter>
</processor>



 <processor name="CalibrateFrame" type="CalibrateFrameObjects">
 <!--Processor to reconstruct object light curve.-->
  <!--Name of the object brightness column-->
  <parameter name="BrightnessColumn" type="string">Aper </parameter>
  <!--Name of the image collection to store calibration correction image-->
  <parameter name="CalibrationImageCollection" type="string" lcioOutType="1"> </parameter>
  <!--Name of the table collection to store calibration correction map-->
  <parameter name="CalibrationMapCollection" type="string" lcioOutType="2"> </parameter>
  <!--Order of the polynomial fit of correction dependence on position-->
  <parameter name="CorrectionFitOrder" type="int">0 </parameter>
  <!--Name of the image collection to store calibration correction image-->
  <parameter name="ErrorImageCollection" type="string" lcioOutType="1"> </parameter>
  <!--Frame division for background calculation-->
  <parameter name="FrameDivision" type="int">10 </parameter>
  <!--Flag to allow automatic adjustment of map binning for too few entries-->
  <parameter name="AutoAdjustDivision" type="bool">true </parameter>
  <!--Name of the collection containing input object lists-->
  <parameter name="InputCollectionName" type="string" lcioInType="2"> Objects </parameter>
  <!--Name of the input image collection - just to get size of images-->
  <parameter name="InputImageCollection" type="string" lcioInType="1"> </parameter>
  <!--Size of the image corresponding to the analysed object list (for map definition)-->
  <parameter name="InputImageSize" type="IntVec"> 2062 2048 </parameter>
  <!--Polinomial order used for background map interpolation-->
  <parameter name="InterpolationOrder" type="int">1 </parameter>
  <!--Flag to indicate that brightness is already given in magnitudo units (logarythmic scale)-->
  <!--parameter name="MagnitudoGiven" type="bool">false </parameter-->
  <!--Position matching tolerance for object and reference stars-->
  <parameter name="MatchingTolerance" type="double">0.02 </parameter>
  <!--Name of the collection  where output table will be stored-->
  <parameter name="OutputCollectionName" type="string" lcioOutType="2"> </parameter>
  <!--Names of columns for calibrated magnitudo and estimated uncertainty-->
  <parameter name="OutputColumnNames" type="StringVec">MagV ErrV AverrV </parameter>
  <!--Names of the position columns in the object table-->
  <!--parameter name="PositionColumns" type="StringVec">RA DEC  </parameter-->
  <!--Names of the column in the reference star catalog containing star magnitudo-->
  <parameter name="ReferenceMagColumn" type="string">VTmag </parameter>
  <!--Magnitudo range for reference stars used for calibration-->
  <parameter name="ReferenceMagRange" type="FloatVec">7 9  </parameter>
  <!--Name of the collection containing star catalog-->
  <parameter name="SkyCatalogName" type="string" lcioInType="2"> </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">DEBUG </parameter>
</processor>

<processor name="ObjectWriter" type="RootTreeWriter">
 <!--Processor for writing FITS images to diskWrite images to specified files-->
  <!--Flag to add table and row number as variables in the tree-->
  <parameter name="AddCounter" type="bool">true </parameter>
  <!--List of header records to be added as double variables-->
  <parameter name="FloatHeaderRecords" type="StringVec"> HJD </parameter>
  <!--List of header records to be added as long integer variables-->
  <!--parameter name="IntHeaderRecords" type="StringVec"> </parameter-->
  <!--Flag to store all tables in one big root file-->
  <parameter name="OneFile" type="bool">true </parameter>
  <!--Output file name, which will be used to store root tree-->
  <parameter name="RootFileName" type="string"> luiza_calibration </parameter>
  <!--List of table columns to be stored (empty for all)-->
  <!--parameter name="SelectedColumns" type="StringVec"> </parameter-->
  <!--Name of the table collection which should be stored to file-->
  <parameter name="TableCollectionName" type="string" lcioInType="2">Objects </parameter>
  <!--verbosity level of this processor ("DEBUG,MESSAGE,WARNING,ERROR,SILENT")-->
  <parameter name="Verbosity" type="string">DEBUG</parameter>
</processor>

</luiza>

@endcode


*/

*****************************************************************************

/*!
@page network_configuration Network configuration

Network configuration is stored in a dedicated XML file. Its name
should be specified in the global section of the steering file:

@code
 <global>
   ...
  <parameter name="NetworkXMLFile" value="gloria_network.xml"/>
   ...
 </global>
@endcode

Network configuration is loaded at the Luiza initialization stage
(before any processor is called) into gloria::GloriaNetworkMgr.
Configuration can include information about the observatories (telescopes) 
as well as databases and other network resources. It should include
all parameters, which can be required in the analysis, and should be rather
specified by GLORIA experts than the user. This includes telescope parameters 
(location, focal length, etc), names and addresses of GLORIA servers and 
databases, etc. Moreover, when storing parameters for all telescopes in the 
network, configuration file can also include processor parameters values, 
which should be used for given telescope. When specified, these parameters
values overwrite parameter defaults and user input (via steering file).
However, this behaviour has to be implemented in each processors
(by call to luiza::Processor::updateTelescopeParameters method).

Configuration file structure is shown in the following example:
@code
<gloria>

  <!-- Global section is used to identify the configuration -->
  <global configID="1" configDate="20120724"  configAuthor="AFZ" /> 
     
  <!-- Observatories section is used to define telescopes in the network -->
  <observatories>

    <telescope ID="9" name="Pi of the Sky prototype">

      <alias name="PiOfTheSky 0"/>

      <position latitude="-22.9533" longitude="-68.18" elevation="2300" />

      <parameter name="FocalLength" value="85" />
      <parameter name="Aperture" value="72" />

      <!-- Luiza processor parameters -->
      <parameter name="SeedThreshold"  value="5" />
      <parameter name="ClusterFinder:PixelThreshold" value="2" />
      <parameter name="MyClusterFinder:MaxPixels" value="25" />

      ...

    </telescope>
      ...

  </observatories>

      ...

</gloria>
@endcode

For processors which make use of luiza::Processor::updateTelescopeParameters 
method, when input image comes from Pi of the Sky prototype (recognized by
\a TELESCOP or \a INSTRUME keyword in the FITS file header), value of \a SeedThreshold 
is set to 5. As different processors can use same parameter name, it is also possible
to specify either processor type (as for \a PixelThreshold in example above) or processor 
name  (as for \a MaxPixels in example above) in front of parameter name.

*/

*****************************************************************************

/*!
@page new_processor  Writing your own processor

This is not very difficult. The best way is to take a running processor 
(eg. \a ExampleProcessor stored in \a processors/luiza_example/) 
and adopt it to your needs. If you start a new analysis it is best if you create
a dedicated catalogue for it, store all related code in this catalogue and compile it
as a separate shared library  (as is done for \a ExampleProcessor, see \a processors/luiza_example/Makefile).

Rules we tried to follow so far:
\li document your code while writing. \n
    Use <a href="http://www.doxygen.org/index.html">doxygen</a> tags to 
    describe all methods in header file. This will provide documentation for other users... 
    \n\n
\li Use initial '_' for names of class data members. This helps to distinguish them from
    local variables in the code.
    




*/

*****************************************************************************

/*!

@page test_data  GLORIA Sample Data 

Samples of fits images from different telescopes, sent by partners,
which can be used for algorithm testing, are available from the dedicated
<a href="http://hep.fuw.edu.pl/u/zarnecki/gloria/gloriadata/GloriaSampleData.html"> web page </a>.

Small sample of files used in the test job (should be stored in catalog linked as
\a gloriadata) can be downloaded from <a href="../../luiza_data.tgz"> here </a>

*/

*****************************************************************************

/*!

@page luiza_gui  Graphic User Interface for Luiza

The LuizaGUI  is a simple program that allows you to easily create
and modify Luiza steering files interactively in a consistent way.

To compile LuizaGUI you have to have QT 4 installed on your system.
The simplest solution is to download and use the precompiled version of QT4, 
if it is available for your system. If not, you can try to compile & install 
QT from source (available <a href="http://www.trolltech.com">here</a>).

Compiling LuizaGUI is not required to run Luiza. So if it fails, you can just
remove luiza_gui from your main Makefile and edit steering files with any 
text editor.


*/

*****************************************************************************

/*!

@page  to_do Development plans

Current version of \a Luiza includes all tools for a  a standard analysis 
of astronomical images. There are no plans for including more tools in the nearest future.
However, users can design and implement their own processors, using the framework.


*/
