/*
 * GloriaAstrometry.h
 *
 *  Created on: Sep 14, 2012
 *      Author: Lech Wiktor Piotrowski
 */

#ifndef GLORIAASTROMETRY_H_
#define GLORIAASTROMETRY_H_

#include "math.h"
#include "GloriaFitsTable.h"
#include "leblind.h"

namespace gloria
{

class GloriaAstrometry: public gloria::GloriaFitsTable
{
public:
	//! Default constructor
	GloriaAstrometry();

	//! Constructor from an input/output leblind_struct of the modified astrometry.net solver
	//! Currently we assume that all the polynomials are of the same order
	GloriaAstrometry(leblind_struct *lestr);

	//! Constructor of the GloriaAstrometry object from the FITS header containing astrometry in WCS format
	GloriaAstrometry(GloriaFitsHeader *header);

	//! Initialize from an input/output leblind_struct of the modified astrometry.net solver
	//! Currently we assume that all the polynomials are of the same order
	void InitFromLeblind(leblind_struct *lestr);

	//! Initialize from  the FITS header containing astrometry in WCS format
	void InitFromHeader(GloriaFitsHeader *header);

	//! Write astrometry in WCS format to existing header/image
	void WriteAstrometry(GloriaFitsHeader *header);

	//! Delete astrometry from given header/image
	void DeleteAstrometry(GloriaFitsHeader *header);

	//! Check if astrometry exists for the given header/image
	bool IsAstrometryPresent(GloriaFitsHeader *header);

	//! Print astrometry values
	void Print(std::ostream& os);

	//! Convert x,y pixels of the image to ra,dec
	void GetRaDecFromXY(double x, double y, double *ra, double *dec);

	inline void GetRaDecFromXY(double x, double y, double *radec)
	                        { GetRaDecFromXY( x, y, radec, radec+1); };

	//! Convert ra,dec coordinates (in degrees) to x,y pixels of the image
	void GetXYFromRaDec(double ra, double dec, double *x, double *y);

	inline void GetXYFromRaDec(double ra, double dec, double *xy)
                                { GetXYFromRaDec( ra, dec, xy, xy+1); };

	virtual ~GloriaAstrometry();

public:

	// WCS astrometry values
	//! Coordinates in pixels of the centre of WCS transformation for the image
	int crpix[2];
	//! Coordinates in ra,dec of the centre of WCS transformation for the image
	double crval[2];
	//! Rotation matrix x,y->ra,dec
	double cd[4];
	//! Order of the distortion polynomial for x
	int sip_order_a;
	//! Order of the distortion polynomial for y
	int sip_order_b;
	//! Order of the distortion polynomial for ra
	int sip_order_ap;
	//! Order of the distortion polynomial for dec
	int sip_order_bp;
	//! Distortion polynomial coefficients for x
	/*! Coefficients are defined as C_a_b, where a+b<=sip_order
	 *  For convienience of addressing we store them in a sip_order*sip_order matrix, the a+b>sip_order elements set to zero */
	vector< vector<double> > sip_a;
	//! Distortion polynomial coefficients for y
	/*! Coefficients are defined as C_a_b, where a+b<=sip_order
	 *  For convienience of addressing we store them in a sip_order*sip_order matrix, the a+b>sip_order elements set to zero */
	vector< vector<double> > sip_b;
	//! Distortion polynomial coefficients for ra
	/*! Coefficients are defined as C_a_b, where a+b<=sip_order
	 *  For convienience of addressing we store them in a sip_order*sip_order matrix, the a+b>sip_order elements set to zero */
	vector< vector<double> > sip_ap;
	//! Distortion polynomial coefficients for dec
	/*! Coefficients are defined as C_a_b, where a+b<=sip_order
	 *  For convienience of addressing we store them in a sip_order*sip_order matrix, the a+b>sip_order elements set to zero */
	vector< vector<double> > sip_bp;

 private:

	//! Last reference coordinates of the centre of WCS transformation
	double ref_ra, ref_dec;

	//! Initialization of vectors used for projection calculation
	void InitWCSVec();

        double wcs_ref_vec[3];
        double wcs_ra_vec[3];
        double wcs_dec_vec[3];

        double deg2rad;
        double rad2deg;
};

} /* namespace gloria */
#endif /* GLORIAASTROMETRY_H_ */
