// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaAverageMap_h
#define GloriaAverageMap_h 1

#include "GloriaValueMap.h"
#include "gloria/GloriaFitsImage.h"

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

namespace gloria {

  //! Class for calculating map of average values for any variable on the image.
  /*! This class can be use to calculate average value map over the
   *  image. For each considered variable/parameter two maps are allocated: one
   *  for storing arithmetic mean value of the parameter and the second one for
   *  storing dispersion (RMS) of the parameter values. Image area is divided
   *  into Nbx x Nby sectors, mean and spread are calculated for each
   *  sector separately. Single \a GloriaAverageMap instance can store mean+RMS map
   *  pairs for arbitrary number of variables (default is one
   *  variable). However, all maps need to have same binning.
   *
   *  \par
   *  The class is inheriting from \a GloriaValueMap class for value 
   *  calculation and creating average value or RMS images. Using the calculated
   *  average value and spread maps,  user can calculate expected
   *  values for any pixel in the image frame using different
   *  interpolation methods implemented for \a GloriaValueMap. 
   *
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaAverageMap : public GloriaValueMap {

  public:

    //! Dummy constructor - does nothing (!)
    GloriaAverageMap() : GloriaValueMap() { };

    //! Constructor for average value map
    /*! Two tables are allocated, for storing average value and value spread
     *  estimate, for any parameter corresponding to the image
     *  given. Tables should be then filled with a call to \a CalculateMap
     *  Note that image size is not defined at this moment yet!
     */
    GloriaAverageMap(int Nbx, int Nby, int Nvar = 1);


    //! Constructor for average value map corresponding to given image file
    /*! Two tables are allocated, for storing average value and value spread
     *  estimate, for any parameter corresponding to the image
     *  given. Tables should be then filled with a call to \a CalculateMap
     */
    GloriaAverageMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int Nvar=1);

    //! Constructor. Allocates map by importing it from fits table
    /*! Creates maps by reading them from fits table. The table has to
     *  created with \a GetTable() method previously, so all required
     *  parameters are stored in header keywords. Otherwise one should
     *  declare map first and read table contents with \a LoadFromTable.
     */
    GloriaAverageMap(GloriaFitsTable* mapTable) : GloriaValueMap(mapTable) { };

    //! Add another variable to average value map set
    /*! Can be used to add maps (one for average value, one for spread) for
     *  another parameter to the existing set of maps. Existing maps
     *  are not affected. 
     */
    void AddAverage();

    //! Add another variable to average value map set and give it a name
    /*! Can be used to add maps (one for average value, one for spread) for
     *  another parameter to the existing set of maps. Existing maps
     *  are not affected. 
     */
    void AddAverage(string name);


    //! Set name to the considered variable
    void SetAverageName(int ivar, string name);
 

    //! Return image filled with parameter values
    inline string GetAverageName(int ivar) {return _valueNames.at(2*ivar); } ;


    //! Calculate average value and spread maps for given variable
    /*! This method can be used for calculating average value and spread maps
     *  based on the table of parameter value measurements (\a Val array) 
     *  at different positions on the frame (specified by \a Xpos and
     *  \a Ypos arrays). All arrays contain \a nVal entries. \a iVar
     *  specifies the variable number (more than one variable can be
     *  considered in one GloriaAverageMap instance). Return value
     *  corresponds to minimal number of entries found in single map
     *  bin. Negative number is returned on error. If zero value is
     *  returned one or more map bins is not set. 
     */
    long CalculateMap(int iVar, int nVal, double* Xpos, double* Ypos, double* Val);

    //! Calculate average value and spread maps for given variable
    /*! This method can be used for calculating average value and spread maps
     *  based on the vector of parameter value measurements (\a Val vector) 
     *  at different positions on the frame (specified by \a Xpos and
     *  \a Ypos vectors). All vectors should contain same number of entries. \a iVar
     *  specifies the variable number (more than one variable can be
     *  considered in one GloriaAverageMap instance). Return value
     *  corresponds to minimal number of entries found in single map
     *  bin. Negative number is returned on error. If zero value is
     *  returned one or more map bins is not set. 
     */
    long CalculateMap(int iVar, vector<double>* Xpos, vector<double>* Ypos, vector<double>* Val);


    //! Calculate average value and spread maps using polynomial fit
    /*! This method can be used for calculating more accurate average and spread maps
     *  based on the table of parameter value measurements (\a Val array) 
     *  at different positions on the frame (specified by \a Xpos and
     *  \a Ypos arrays). Instead of calculating simple parameter average and RMS
     *  values in each map bin, polinomial dependence on the CCD
     *  position is fitted to reproduce better the parameter
     *  variation. Estimated value in the bin center is stored in the
     *  map as well as the average squared distance of the value from
     *  the fitted surface. The fitted average value is usually quite
     *  close to the arithmetic average, but the spread estimate can
     *  differ significantly and is more reliable. \a Npol parameter
     *  is the polynomial order used. 
     */
    long CalculateFitMap(int iVar, int nVal, double* Xpos, double* Ypos, double* Val, int Npol);

    //! Calculate average value and spread maps using polynomial fit
    /*! This method can be used for calculating more accurate average and spread maps
     *  based on the vector of parameter value measurements (\a Val vector) 
     *  at different positions on the frame (specified by \a Xpos and
     *  \a Ypos vectors). Instead of calculating simple parameter average and RMS
     *  values in each map bin, polinomial dependence on the CCD
     *  position is fitted to reproduce better the parameter
     *  variation. Estimated value in the bin center is stored in the
     *  map as well as the average squared distance of the value from
     *  the fitted surface. The fitted average value is usually quite
     *  close to the arithmetic average, but the spread estimate can
     *  differ significantly and is more reliable. \a Npol parameter
     *  is the polynomial order used. 
     */
    long CalculateFitMap(int iVar, vector<double>* Xpos, vector<double>* Ypos, vector<double>* Val, int Npol);

    //! Get expected average value for pixel (ix,iy) using calculated map
    /*! Calculates expected average value value in given pixel on the input frame
     *  by interpolating between map points. \a ivar parameter selects the
     *  variable (more than one map can be defined) and \a ipol defines
     *  interpolation order. Currently implemented  values are: 
     *     0 - no interpolation (take value from the bin center) and
     *     1 - linear interpolation between bin centers (default).
     *     2 - second order approximation (simplified).
     */
    inline double GetAverage(int ivar, int ix, int iy, int ipol=1)
                         { return GetMapValue(2*ivar,ix,iy,ipol); } ;

    //! Get expected average value level at position (x,y) using calculated map
    inline double GetAverage(int ivar, double x, double y, int ipol=1)
                         { return GetMapValue(2*ivar,x,y,ipol); } ;

    //! Get expected average value level for pixel ipix  using calculated map
    inline double GetAverage(int ivar, long ipix, int ipol=1)
                        { return GetMapValue(2*ivar,ipix,ipol); } ;

    //! Get spread of the variable for pixel (ix,iy) using calculated map
    inline double GetSpread(int ivar, int ix, int iy, int ipol=1)
                         { return GetMapValue(2*ivar+1,ix,iy,ipol); } ;

    //! Get spread of variable at position (x,y) using calculated map
    inline double GetSpread(int ivar, double x, double y, int ipol=1)
                         { return GetMapValue(2*ivar+1,x,y,ipol); } ;

    //! Get spread for pixel ipix  using calculated map
    inline double GetSpread(int ivar, long ipix, int ipol=1)
                        { return GetMapValue(2*ivar+1,ipix,ipol); } ;




    //! Returns pointer to the average value map array
    inline double* GetGloriaAverageMap(int ivar=0) {return _valueMaps.at(2*ivar); };

    //! Returns pointer to the spread map array
    inline double* GetSpreadMap(int ivar=0) {return _valueMaps.at(2*ivar+1); };


    //! Returns mean of background map
    inline  double GetAverageMean(int ivar=0){ return GetMapMean(2*ivar);} ;

    //! Returns mean of noise map
    inline double GetSpreadMean(int ivar=0){ return GetMapMean(2*ivar+1);} ;


    //! Return image filled with average value estimates
    inline GloriaFitsImage* GetAverageImage(int ivar=0, int ipol=1)
                    { return GetMapImage(2*ivar, ipol);}  ;

    //! Return image filled with spread values
    inline GloriaFitsImage* GetSpreadImage(int ivar=0, int ipol=1)
                    { return GetMapImage(2*ivar+1, ipol);}  ;

    //! Allocate new map sets with given binning. All data and image size information are lost.
    void InitializeMaps(int Nbx, int Nby, int Nvar);

  private:

    int  GaussjSolve(double *alfa,double *beta,int n);

  } ;


}

#endif



