// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaClusterStatistics_h
#define GloriaClusterStatistics_h 1

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

namespace gloria {

  //! Class for calculating basic properties of pixel cluster
  /*! This class can be use to calculate properties of the pixel
   *  cluster as signal (sum of pixel values), position (from Center
   *  of Gravity - CoG), spread in X, Y and X-Y correlation of charge
   *  distribution, and finally effective width (FWHM), elongation
   *  (elliptisity) as well as orientation of the primary axis.
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaClusterStatistics {

  public:


    //! Dummy constructor - creates empty cluster
    GloriaClusterStatistics();

    //! Destructor
    ~GloriaClusterStatistics();

    //! Clear cluster statistics
    void Clear();

    //! Add pixel to the cluster  
    /*! Pixels are not stored inside ClusterStatitics. Coordinates of
     *  the first pixel added (most probably the seed pixel) are
     *  remembered as the offset for shape calculation (for numerical
     *  stability). Otherwise, coordinates weighted by signal are
     *  added to different counters: weighted sums of x, y, x^2, y^2
     *  and xy. These sums can be then used to calculate statistical
     *  parameters of the charge distribution. 
     */
    void AddPixel(int ix, int iy, double signal);


    //! Get cluster multiplicity (number of pixels)
    inline int GetNpixel() { return _nPixel; };

    //! Get cluster signal (sum of pixel signals)
    inline double GetSignal() { return _signal; };

    //! Get cluster position (CoG) in X
    inline double GetXpos(){ return _sumX/_signal + _seedX; };

    //! Get cluster position (CoG) in Y
    inline double GetYpos(){ return _sumY/_signal + _seedY; };

    //! Get cluster spread (RMS) in X
    double GetXspread();

    //! Get cluster spread (RMS) in Y
    double GetYspread();

    //! Get X-Y correlation for the cluster
    double GetXYcorr();

    //! Get effective cluster width
    double GetFWHM();

    //! Get cluster elongation
    double GetElongation();

    //! Get cluster orientation (direction of primary axis)
    double GetOrientation();

    //! Get cluster length (length of primary axis)
    double GetLength();

    //! Get cluster width (length of secondary axis)
    double GetWidth();


private:

    void CalculateAverages();

    // Number of pixels in the cluster
    int _nPixel;

    // Position of first (seed?) pixel
    // Used as offset for sum calculations
    double _seedX, _seedY;

    // Total signal in the cluster
    double _signal;

    // Weighted sums
    double _sumX, _sumY, _sumXX, _sumYY, _sumXY;

    //Flag for doing basic statistics
    bool _averDone;
  
    //Results of calculations
    double _cogX, _cogY, _rmsX2, _rmsY2, _rmsX, _rmsY, _corrXY;
  } ;


}

#endif



