// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaDataContainer_h
#define GloriaDataContainer_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "GloriaFitsImage.h"
#include "GloriaFitsTable.h"

//! namespace for data storing and exchange formats
/*! gloria namespace includes classes defining formats for reading
 *  input data files, writing analysis results to files, as well as
 *  for internal data representation in memory for exchanging between
 *  processors 
 *
 * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
 *
 */
namespace gloria{

  typedef std::vector< GloriaFitsImage > ImageVec; 

  typedef std::vector< GloriaFitsTable > TableVec; 

  //! Enumerator for allowed collection types
  /*! Image and table collections are stored separated in the data
   *  container. However, some methods (eg. in luiza::Processor)
   *  require specifying collection type
   *
   * Two collection types are currently implemented:
   * \li \a ImageCollection - collection of GloriaFitsImage
   * \li \a TableCollection - collection of GloriaFitsTable
   * \li \a UndefinedCollection - used when collection type unknown
   */
  enum GloriaCollectionType
      { UndefinedCollection = 0,
        ImageCollection,
        TableCollection };


  static const char* GloriaCollectionTypeName[] = { "Undefined", "Image", "Table" };


  //! Class for storing all data processed in Luiza
  /*! This class allows to store collections of images or data tables,
   *  to be processed by Luiza processors. One instance of this
   *  class is initialized by luiza::ProcessorMgr and the pointer
   *  passed to subsequent processors (selected by user) in the
   *  Processor::processData call. Dedicated processors, defined as
   *  luiza::DataSourceProcessor are expected to fill the structure
   *  with actual data (images or tables).
   *
   *  Images and tables are always stored in so called
   *  collections. Collection is just a vector of images
   *  (gloria::GloriaFitsImage) or tables (gloria::GloriaFitsTable);
   *  in particular it can contain just a single image or
   *  table. However, collection also has a name, which is used to
   *  recognize different data sets in analysis. Collection name is
   *  usually specified as a parameter in the processor.
   *
   *  Each collection can be either defined as temporary or
   * permanent. Images or tables in a temporary collection are removed
   * after each data processing loop, so the new data can be read into
   * the collection (collection itself is not deleted, it is just
   * cleared). Permanent collections can be defined eg. to store
   * reference images (darks, flats) or reference data (eg. data from
   * star catalog). 
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaDataContainer {


  public:

    //! Default constructor
    /*! Default constructor does not allocate any space for storing data
     */
    GloriaDataContainer();

    //! Destructor
    ~GloriaDataContainer();

    //! Number of image collections
    inline int GetImageCollectionNumber() { return _nImageCollections;} ;

    //! Number of table collections
    inline int GetTableCollectionNumber() { return _nTableCollections;} ;

    //! Return image collection name
    inline std::string& GetImageCollectionName(int icl) { return _imageCollectionNames.at(icl);} ;

    //! Return table collection name
    inline std::string& GetTableCollectionName(int icl) { return _tableCollectionNames.at(icl);} ;

    //! Create new (empty) image collection
    ImageVec& CreateImageCollection(std::string name, bool isPermanent = false );   

    //! Check if image collection exists
    bool ImageCollectionExists(std::string name);

    //! Check if image collection is permanent
    bool ImageCollectionPermanent(std::string name);

    //! Flag table collection as permanent
    void SetImageCollectionPermanent(std::string name, bool isPermanent = true );

    //! Get stored image collection
    ImageVec& GetImageCollection(std::string name);   

    //! Remove stored image collection
    void RemoveImageCollection(std::string name);   

    //! Create new (empty) table collection
    TableVec& CreateTableCollection(std::string name, bool isPermanent = false );   

    //! Check if table collection exists
    bool TableCollectionExists(std::string name);

    //! Check if table collection is permanent
    bool TableCollectionPermanent(std::string name);

    //! Flag table collection as permanent
    void SetTableCollectionPermanent(std::string name, bool isPermanent = true );

    //! Get stored table collection
    TableVec& GetTableCollection(std::string name);   

    //! Remove stored table collection
    void RemoveTableCollection(std::string name);   

    //! Remove all collections not declared as permanent
    void RemoveTemporaryCollections();

    //! Clear all collections not declared as permanent
    void ClearTemporaryCollections();

    //! Return type of given collection 
    GloriaCollectionType GetCollectionType(std::string name);

    //! Return string describing type of given collection 
    std::string GetCollectionTypeName(std::string name);

private:

    std::map< std::string, ImageVec >  _imageCollectionMap;

    std::map< std::string, TableVec >  _tableCollectionMap;

    std::vector< std::string > _imageCollectionNames;

    std::vector< std::string > _tableCollectionNames;

    std::map< std::string, bool > _imageCollectionPermanent;

    std::map< std::string, bool > _tableCollectionPermanent;

    int _nImageCollections;
 
    int _nTableCollections;

  };


}

#endif



