// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaFitsHeader_h
#define GloriaFitsHeader_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include <string>
#include <vector>
#include <map>
#include <cfitsio/fitsio.h>
using namespace std;

namespace gloria{

  //! Class for handling fits file headers
  /*! This class defines methods to access header of fits files.
   *  It is used in definition of methods for accessing fits images
   *  and tables. 
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaFitsHeader {

    friend class GloriaFitsImage ;
    friend class GloriaSkyCatalog ;
    friend class GloriaObjectList ;
    friend class GloriaFitsTable ;

  public:

    //! Structure to store header records
    struct HeaderRecord {       
                        std::string Keyword;
                        std::string Value;
                        std::string Comment;
                        };


    //! Constructor
    GloriaFitsHeader();


    //! Destructor
    virtual ~GloriaFitsHeader();


    //! Read fits header from file 
    /*! Fits header is read from file and put into header
     *  structure. The file has to be opened with call to fitsio
     *  library. 
     */ 
    void ReadFitsHeader(fitsfile *fp);

    //! Write header to fits file
    /*! Header compliant to with FITS standard is created
     */ 
    void WriteFitsHeader(fitsfile *fp);


    // Header access methods
    // =====================


    //! Number of records defined in the header
    inline int GetKeywordNumber() {return _fitsHeaderRecords.size(); };

    
    //! Access header record by record index
    inline HeaderRecord& GetRecordAt(const int ikey) {return _fitsHeaderRecords.at(ikey); };   

    //! Access header keyword by record index
    inline std::string& GetKeywordAt(const int ikey) {return _fitsHeaderRecords.at(ikey).Keyword; };   

    //! Access header value (string) by record index
    inline std::string& GetValueAt(const int ikey) {return _fitsHeaderRecords.at(ikey).Value; };   

    //! Access header comment by record index
    inline std::string& GetCommentAt(const int ikey) {return _fitsHeaderRecords.at(ikey).Comment; };   

    //! Check if given keyword is set in the header 
    bool IsKeywordSet(const char *keyword);

    //! Access header record by record index
    HeaderRecord& GetKeywordRecord(const char *keyword);   

    //! Access header value (string) by keyword
    std::string& GetKeywordValue(const char *keyword);   

    //! Access header value (long int) by keyword
    long   GetKeywordValueInt(const char *keyword);   

    //! Access header value (double) by keyword
    double GetKeywordValueDouble(const char *keyword);   

    //! Add new keyword to the fits header
    void AddKeyword(const char *keyword, std::string value, std::string comment);   

    //! Add new keyword to the fits header, as double
    void AddKeywordDouble(const char *keyword, double value, std::string comment);   
    
    //! Add new keyword to the fits header, as int
    void AddKeywordInt(const char *keyword, int value, std::string comment);   
    
    //! Add new keyword record to the fits header
    void AddKeyword(HeaderRecord record);   

    //! Set new value for the existing keyword. Add it if not present.
    void SetKeyword(const char *keyword, std::string value, std::string comment);   
    
    //! Set new value for the existing keyword as double. Add it if not present.
    void SetKeywordDouble(const char *keyword, double value, std::string comment);
    
    //! Set new value for the existing keyword as int. Add it if not present.
    void SetKeywordInt(const char *keyword, int value, std::string comment);    
    
    //! Add comment to the header
    void AddComment(std::string comment);   
    
    //! Add history to the header
    void AddHistory(std::string history);   
    
    //! Remove given keyword from header.
    /*! Only one record is removed, even if the keyword is not
     *  unique. 
     */
    void RemoveKeyword(const char *keyword);   

    //! Remove given keyword from header.
    /*! All instances of the keyword are removed */
    void RemoveKeywords(const char *keyword);   



    //! Copy full header from given data object (table or image);
    void CopyHeader(GloriaFitsHeader* source);


    //! Print FITS header to output stream. Use maxRec=0 to print all records.
    void PrintHeader( std::ostream& os, int maxRec=100);

    // General status
    // ==============

    //! Method returning true if processing was successful
    inline bool OK() { return _statusOK; };

    //! Method returning true if there was an error in data processing
    inline bool Error() { return !_statusOK; };

    //! Returns error message string
    std::string ErrorMessage() { return _errorMessage; };


    //! Method returning true if there was an error in data processing
    inline void ClearError() { _statusOK=true; _errorMessage.clear();};

    //! Name of the last input file 
    //! Empty if image/table was not read from file
    inline std::string  InputFileName() { return _inputFileName; };

    //! Set name considered as an input file 
    //! Can be used to copy source file name to resulting image/table
    inline void  SetInputFileName(string name) {_inputFileName = name; };

    //! Name of the last output file. 
    //! Empty string if image/table was never written to file 
    inline std::string  OutputFileName() { return _outputFileName; };

private:


    //! Fits header 
    std::vector<HeaderRecord> _fitsHeaderRecords;

    //! Map for more efficient header access by keyword
    std::multimap<std::string,int> _headerKeywordMap;



    //! Name of the last input file. 
    /*! Empty string if image/table was not read from file */
    std::string   _inputFileName;

    //! Name of the last output file. 
    /*! Empty string if image/table was never written to file */
    std::string   _outputFileName;


    //! General status flag
    bool _statusOK;

    //! Error message, filled when status is not OK
    std::string _errorMessage;

  } ;


}

#endif



