// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaFitsImage_h
#define GloriaFitsImage_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include <string>
#include <vector>
#include <map>

#include "GloriaFitsHeader.h"

namespace gloria{

  //! Class for reading and processing fits images in Luiza
  /*! This class allows to read the fits image from file and process
   *  it in memory, without need to have the file opened. Virtual fits
   *  can be created (not stored on file) to pass images between
   *  processors. Current implementation allows simple 2D fits only.
   *
   *  Two kinds of images are possible in general, containing integer
   *  or floating point pixel values. To avoid possible information
   *  loss more precise types were selected i.e. \a long for integer
   *  pixels and \a double for floating point pixels. So the image
   *  data is stored either in long or  double  array. This has to be
   *  respected when addressing the image content directly (with
   *  GloriaFitsImage::GetIntImageData() or
   *  GloriaFitsImage::GetFloatImageData() method). 
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaFitsImage : public GloriaFitsHeader {

  public:

    //! Default constructor
    /*! Default constructor does not allocate any space for storing
        pixel values.
     */
    GloriaFitsImage();

    //! Constructor based on reading fits file
    /* Creates image and reads its contents from the fits file */
    GloriaFitsImage(std::string FitsFileName);

    //! Copy constructor
    GloriaFitsImage(const GloriaFitsImage& sourceFitsImage);


    //! Constructor allocating 2D image of given size and type
    /*! The double or long array will be allocated to store image 
     *  pixel values by default, depending on useFloatPixels
     *  parameter. 
     */
    GloriaFitsImage(int NpixelX, int NpixelY, bool useFloatPixels=true);


    //! Destructor
    ~GloriaFitsImage();


    //! Read fits image from file to the existing structure
    /*! Header and Image are read from file and put into existing
     *  structure. If image size or type is changed, new array will be
     *  allocated to store pixel values. Old header and data are lost.
     */ 
    void ReadFitsFile(std::string FitsFileName);

    //! Write image to fits file
    /*! Header compliant with FITS standard is created
     * \a cBitPix, \a cBScale and \a cBZero parameters, which should be set
     * accordingly to the FITS specifications, decide how the image
     * is stored. Following values are accepted for \a cBitPix
     *
     * \li 8 : 8-bit byte pixels
     *
     * \li 16 : 16 bit (short) integer pixels
     *
     * \li 32 : 32 bit (long) integer pixels
     *
     * \li -32 : 32 bit floating point (float) pixels
     *
     * \li -64 : 64 bit floating point (double) pixels
     *
     * If cBitPix is set to 0 image is stored in its current format
     *      (which is either long or double).
     */ 
    void WriteFitsFile(std::string FitsFileName, int cBitPix=0, int cBZero=0, double cBScale=1.0);


    //! Shift values of all pixels by given value
    /*! Image pixel type is not changed */
    GloriaFitsImage& operator+=(const long shift);


    //! Shift values of all pixels by given value
    /*! If image contains integer pixel values, they are converted to
     * floating point numbers
     */
    GloriaFitsImage& operator+=(const double shift);


    //! Add an image
    /*! If one of images contains integer pixel values, result contains
     * floating point numbers. Image is unchanged if sizes differ.
     */
    GloriaFitsImage& operator+=(const GloriaFitsImage& img);


    //! Subtract an image
    /*! If one of images contains integer pixel values, result contains
     * floating point numbers. Image is unchanged if sizes differ.
     */
    GloriaFitsImage& operator-=(const GloriaFitsImage& img);


    //! Scale values of all pixels by given value
    /*! Image pixel type is not changed */
    GloriaFitsImage& operator*=(const long fac);
 

    //! Scale values of all pixels by given value
    /*! If image contains integer pixel values, they are converted to
     * floating point numbers.
     */
    GloriaFitsImage& operator*=(const double fac);
 

    //! Multiply by image
    /*! If one of images contains integer pixel values, result contains
     * floating point numbers. Image is unchanged if sizes differ.
     */
    GloriaFitsImage& operator*=(const GloriaFitsImage& img);


    //! Divide by image
    /*! If one of images contains integer pixel values, result contains
     * floating point numbers. Image is unchanged if sizes differ.
     */
    GloriaFitsImage& operator/=(const GloriaFitsImage& img);


    // Data access methods
    // ===================

    //! Returns size of image in X
    inline int SizeX() const { return _sizeX; };

    //! Returns size of image in Y
    inline int SizeY() const { return _sizeY; };

    //! Returns number of bits per pixel
    /*! For floating point pixel values negative value is returned 
     *  (-64 for double)
     */
    inline int PixelBits() const { return _bitsPerPixel; };

    //! Returns \a true if image contains integer pixel values (long)
    inline bool IsIntImage() const {return _isIntImage; };

    //! Returns \a true if image contains floating point pixel values (double)
    inline bool IsFloatImage() const {return _isFloatImage; };

    //! Returns pointer to the array of integer pixel values
    inline long* GetIntImageData() const {return _intPixels; };

    //! Returns pointer to the array of floating point pixel values
    inline double* GetFloatImageData() const {return _floatPixels; };


    //! Convert image pixels to floating point numbers
    /*! Memory allocated for int pixels is released after copying.
     *  Returns without error if pixels are already of floating type
     */ 
    void ConvertToFloat();


    //! Convert image pixels to integer numbers
    /*! Memory allocated for floating point pixels is released after copying.
     *  Returns without error if pixels are already of integer type
     */ 
    void ConvertToInt();



private:

    void InitializeEmptyImage();
    void AllocateMemory(int NpixelX, int NpixelY, bool useFloatPixels=true);
    void ReleaseMemory();

   //! Image size. Only plain 2D images allowed in current implementation
    int _sizeX, _sizeY;

    //! Image resolution 
    int _bitsPerPixel;

    //! Image data
    bool _isIntImage;
    long* _intPixels;

    bool _isFloatImage;
    double* _floatPixels;
  
  } ;


}

#endif



