// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaFitsTable_h
#define GloriaFitsTable_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include <string>
#include <vector>
#include <map>

#include "GloriaFitsHeader.h"
#include "GloriaTableColumn.h"

using namespace std;

namespace gloria{

  //! Class for reading and processing fits tables in Luiza
  /*! This class allows to read the fits table from file and store
   *  it in memory. Tables can be also created to store
   *  different kind of data or analysis results.
   *
   *  \par Structure
   *  Table is defined as a collection of columns
   *  (gloria::GloriaTableColumn). Columns are stored in a vector, but
   *  each column is also given a name, so it can be accessed by its
   *  name. However, FITS standard allows for multiple columns with
   *  the same name. User is responsible for making sure that column
   *  names are unique if he wants to use them to address table
   *  contents. Only columns defined by gloria::TableColumnType are
   *  allowed in the table (see description of
   *  gloria::GloriaTableColumn for details). One also
   *  has to remember that rows and columns are numbered from 0 (as is
   *  the standard in C; FITS standard numbers columns and rows
   *  starting form 1). 
   * 
   *
   *  \par Defining the table
   *  Table can be created either as an empty table or a copy of
   *  existing table, or read from file. In any case user can then add
   *  and delete columns and rows in the table. 
   *    \li to create column of given type one can use
   *  GloriaFitsTable::AddColumn method. Column name is
   *  obligatory. Column type can be specified by
   *  gloria::TableColumnType or can be recognized from the type of
   *  "empty element" (value which is inserted in empty cell after it
   *  is created).
   *   \li to add row to the table one can use
   *  GloriaFitsTable::AddRow() method. One can also specify up to 5
   *  (in the current implementation) values to be put in the first
   *  columns of the new row (this is supposed to simplify the code).
   *   \li to remove row or column one can use
   *  GloriaFitsTable::DeleteColumn or GloriaFitsTable::DeleteRow. 
   *
   *  \par Accessing the table
   * There are few methods, which can be used to access (set
   * or get) the content of the table
   *     \li GloriaFitsTable::SetValue and  GloriaFitsTable::GetValue
   * are most safe to use, as content type and range of column and row
   * indexes are checked before accessing table element. Dedicated
   * GloriaFitsTable::SetLast method is available for setting values
   * in the last row of the table 
   *     \li where the efficiency of the code is crucial one can use
   * template method GloriaFitsTable::GetValueOf<T>. However, user is
   * then responsible for setting proper type (same as column type)
   * and checking index range. Otherwise the program will probably
   * crash.  GloriaFitsTable::GetValueOf<T> returns the actual element
   * in the table (by the reference) so it can be used both to set and
   * to get the value from the table.
   *     \li User can also access  gloria::GloriaTableColumn stored in
   * the table using  GloriaFitsTable::GetColumnOf<T> temlate method,
   * or even access directly the data vector stored in the column by 
   * GloriaFitsTable::GetVectorOf<T> temlate method.
   *
   * \par Utilities
   *  User is allowed to access table columns, so it is possible that
   *  numbers of elements in different columns will be different
   *  (eg. when new values are added to selected columns). Following
   *  are defined to help dealing with this problem:
   *     \li  GloriaFitsTable::CheckTable can be used to check the
   *  consistency between the number of rows stored in the table
   *  header and numbers of elements in subsequent columns
   *     \li  GloriaFitsTable::AdjustColumns can be used to adjust
   *  column lengths - all columns are extended to match the size of
   *  the longest column.
   *     \li  GloriaFitsTable::CreateSortIndex can be used to create
   *  an index array for accessing table rows in order of increasing
   *  (or decreasing, if reverse flag set to true) values of selected
   *  column. Table contents is not affected.
   *     \li  GloriaFitsTable::SortTable can be used to sort table
   *  rows in order of increasing (or decreasing, if reverse flag set
   *  to true) values of selected column. Table contents is changed.
   *     \li  GloriaFitsTable::Print method can be used to generate
   *  readable output of the table contents
   * 
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaFitsTable : public GloriaFitsHeader {

  public:

    //! Default constructor
    /*! Default constructor cretes empty table
     */
    GloriaFitsTable();

    //! Constructor based on reading fits file
    /* Creates empty table and reads its contents from the fits file */
    GloriaFitsTable(string FitsFileName);

    //! Copy constructor
    GloriaFitsTable(const GloriaFitsTable& sourceFitsTable);

    //! Destructor
    virtual ~GloriaFitsTable();

    //! Assignment operator
    GloriaFitsTable& operator= (const GloriaFitsTable& tab);

    //! Read fits table from file 
    /*! Header and content are read from file and put into existing
     *  table. Old header and all data are lost.
     */ 
    void ReadFitsFile(string FitsFileName);

    //! Read table from text file 
    /*! Content is read from text file and put into existing
     *  table. Header remains unchanged.
     */ 
    void ReadTextFile(string TextFileName);

    //! Write table to fits file
    /*! Header compliant to with FITS standard is created
     */ 
    void WriteFitsFile(string FitsFileName, bool binary_tbl=true);

    //! Write table to text file
    /*! Can be used to store data for direct analysis
     */ 
    void WriteTextFile(string TextFileName, int colWidth=10, char ColumnSeparator = ' ');

    //! Write table to output text stream
    /*! Can be used to store multiple tables to one stream
     */ 
    void WriteTextStream(std::ostream& outF, int colWidth=10, char ColumnSeparator = ' ');


    // Data access methods
    // ===================

    //! Returns number of defined columns
    inline int GetColumnNumber() {return _nColumn;};

    //! Returns number of filled rows
    inline int GetRowNumber() {return _nRow;};

    //! Returns name of column
    inline string GetColumnName(int iCol) 
               {return _columns.at(iCol)->Name();};

    //! Returns column Id
    inline int GetColumnId(string name) 
               {return _nameMap.find(name)->second; };

    //! Returns type of column
    inline TableColumnType GetColumnType(int iCol)
               {return _columns.at(iCol)->Type(); };

    //! Returns type for given column name
    inline TableColumnType GetColumnType(string name)
               {return _columns.at(_nameMap.find(name)->second)->Type(); };


    //! Check if given column exists
    inline bool ColumnExists(string name)
               {return _nameMap.find(name)!=_nameMap.end();};

     //! Returns base class pointer for given column index
     inline GloriaTableColumn* GetColumn(int iCol)
               {return _columns.at(iCol); } ;

     //! Returns base class pointer for named column stored in the table
     inline GloriaTableColumn* GetColumn(string name)
               {return _columns.at(_nameMap.find(name)->second); } ;

     //! Returns pointer to the column of given type, for given column index
     /* This method DOES NOT check consistency of types nor indexes - be careful */
     template <class T>
     inline GloriaTableColumnOf<T>* GetColumnOf(int iCol)
               {return (GloriaTableColumnOf<T> *)_columns.at(iCol); } ;

     //! Returns pointer to the column of given type, for given column name
     /* This method DOES NOT check consistency of types nor indexes - be careful */
     template <class T>
     inline GloriaTableColumnOf<T>* GetColumnOf(string name)
             {return (GloriaTableColumnOf<T> *)_columns.at(_nameMap.find(name)->second); } ;

     //! Returns data vector of given type, from given column index
     /* This method DOES NOT check consistency of types nor indexes - be careful */
     template <class T>
     inline vector<T>& GetVectorOf(int iCol) 
             { return ((GloriaTableColumnOf<T> *)_columns.at(iCol))->Data(); };


     //! Returns data vector of given type, from given column name
     /* This method DOES NOT check consistency of types nor indexes - be careful */
     template <class T>
     inline vector<T>& GetVectorOf(string name) 
             { return ((GloriaTableColumnOf<T> *)_columns.at(_nameMap.find(name)->second))->Data(); };


     //! Returns value of given type, from given column index and row
     /* This method DOES NOT check consistency of types nor indexes - be careful */
     template <class T>
     inline T& GetValueOf(int iCol, int iRow) 
             { return ((GloriaTableColumnOf<T> *)_columns.at(iCol))->At(iRow); };


     //! Returns value of given type, from given column name and row
     /* This method DOES NOT check consistency of types nor indexes - be careful */
     template <class T>
     inline T& GetValueOf(string name, int iRow) 
            { return ((GloriaTableColumnOf<T> *)_columns.at(_nameMap.find(name)->second))->At(iRow); };


    //! Add column of given type to the table
    /*! If any rows are already put in the table, new column is
     * initialized to the current number of rows (with empty string or
     * vector, or zero values, depending on type). Unit field is optional.
     */
    void AddColumn(string Name, TableColumnType Type, string Unit = "" , string Comment = "" );


    //! Add column to the table. Type recognized by type of empty cell given
    /*! If any rows are already put in the table, new column is
     * initialized to the current number of rows (with empty string or
     * vector, or zero values, depending on type). Unit field is optional.
     */
    template <class T>
    void AddColumn(string Name, T empty, string Unit = "" , string Comment = "" );

    //! Insert column in the table.
    /*! Source column is inserted as a new column in the table.
     *  Pointer to the column is stored so the data is not copied.    
     *  GloriaTableColumn::Clone() method should be used to avoid problems.
     */
    void InsertColumn(GloriaTableColumn* source);

    //! Insert column in the table.
    /*! Source column is inserted as a new column in the table, with
     *  new name (!). Pointer to the column is stored so the data is
     *  not copied.  GloriaTableColumn::Clone() method should be used
     *  to avoid problems. 
     */
    void InsertColumn(string NewName, GloriaTableColumn* source);

    //! Add new row in the table
    /*! Adding a row means increasing length of all columns by 1
     *  String values are set to empty string, numerical to zero,
     *  vectors to empty vectors.
     */ 
    void AddRow();


    //! Add new row and set value to the first column
    /*! Be careful to respect column type. Otherwise value will not be
     *  added.
     */ 
    template <class T>
    inline void AddRow(T val){ AddRow(); SetLast(0,val); };

    //! Add new row and set value to the first two columns
    /*! Be careful to respect column order and types. Otherwise value
     *  will not be added.
     */ 
    template <class T0, class T1>
    inline void AddRow(T0 val0, T1 val1){ AddRow(); SetLast(0,val0); SetLast(1,val1);};


    //! Add new row and set value to the first three columns
    /*! Be careful to respect column order and types. Otherwise value
     *  will not be added.
     */ 
    template <class T0, class T1, class T2>
    inline void AddRow(T0 val0, T1 val1, T2 val2){ AddRow(); SetLast(0,val0); SetLast(1,val1); SetLast(2,val2);};

    //! Add new row and set value to the first four columns
    /*! Be careful to respect column order and types. Otherwise value
     *  will not be added.
     */ 
    template <class T0, class T1, class T2, class T3>
    inline void AddRow(T0 val0, T1 val1, T2 val2, T3 val3)
                { AddRow(); SetLast(0,val0); SetLast(1,val1); SetLast(2,val2);SetLast(3,val3);};

    //! Add new row and set value to the first five columns
    /*! Be careful to respect column order and types. Otherwise value
     *  will not be added.
     */ 
    template <class T0, class T1, class T2, class T3, class T4>
    inline void AddRow(T0 val0, T1 val1, T2 val2, T3 val3, T4 val4)
                { AddRow(); SetLast(0,val0); SetLast(1,val1); SetLast(2,val2);SetLast(3,val3);SetLast(4,val4);};

    //! Set value of given cell
    /* This method checks consistency of variable and column types */
    template <class T>
    void SetValue(string Name, int iRow, T val);


    //! Set value of given cell
    /* This method checks consistency of variable and column types */
    template <class T>
    void SetValue(int iCol, int iRow, T val);


    //! Set value of last cell in given column
    /* This method checks consistency of variable and column types */
    template <class T>
    void SetLast(int iCol, T val);

    //! Set value of last cell in given column
    /* This method checks consistency of variable and column types */
    template <class T>
    void SetLast(string Name, T val);

    //! Get value of given cell
    /* This method checks consistency of variable and column types */
    template <class T>
    void GetValue(string Name, int iRow, T* val);

    //! Get value of given cell
    /* This method checks consistency of variable and column types */
    template <class T>
    void GetValue(int iCol, int iRow, T* val);

    //! Copy row elements from other table (assuming types match!)
    /*! New row is added if \a itarget is negative or beyond table length */
    void CopyTableRow(GloriaFitsTable* source, int isource, int itarget=-1);

    //! Copy table structure and header information. Data is not copied.
    GloriaFitsTable* EmptyClone();

    //! Checks sizes of columns and adjust to the longest one if needed 
    void AdjustColumns();

    //! Checks consistency of table structure
    bool CheckTable();


    //! Create an index array for accessing table rows in order of increasing column values
    long* CreateSortIndex(int iCol, bool reverse=false);

    //! Create an index array for accessing table rows in order of increasing column values
    long* CreateSortIndex(string name, bool reverse=false);

    //! Sort table rows in order of increasing  values of selected column.
    void  SortTable(int iCol, bool reverse=false);

    //! Sort table rows in order of increasing  values of selected column.
    void  SortTable(string name, bool reverse=false);


    //! Check if table is sorted
    inline bool IsSorted() { return  (_nSorted == _nRow); } ;

    //! Check if table is sorted by given column
    inline bool IsSortedBy(string name) 
              { return  (_nSorted == _nRow  && name == _sortName); } ;

    //! Return name of the column used for table sorting
    inline string SortName() { return  _sortName; } ;

    //! Return name of the column used for table sorting
    inline int SortColumn() { return  _sortColumn; } ;

    //! Delete given row
    void DeleteRow(int iRow);

    //! Return column pointer and remove it from the table structure
    GloriaTableColumn* RemoveColumn(int iCol);

    //! Return column pointer and remove it from the table structure
    GloriaTableColumn* RemoveColumn(string name);

    //! Delete given column
    void DeleteColumn(int iCol);

    //! Delete given column
    void DeleteColumn(string name);

    //! Clear the table (delete all rows, keep column definitions)
    void ClearTable();

    //! Delete all columns in the table and clear table header
    void EmptyTable();

    //! Print table content to the stream
    /* Column width can be set between 3 and 20 characters */

    void Print( std::ostream& os, int colWidth=10);

protected:

    void InitializeEmptyTable();

   
    //! Table size. 
    int _nRow, _nColumn;

    //! Column vector
    vector<GloriaTableColumn* > _columns;

    //! Column name map. Name is not always unique!
    multimap< string, int > _nameMap;

    //! Number of sorted rows
    int _nSorted;

    //! Name of column used for sorting
    string _sortName;

    //! ID of column used for sorting
    int _sortColumn;

  } ;


}

#endif
