// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaMedianMap_h
#define GloriaMedianMap_h 1

#include "GloriaValueMap.h"
#include "gloria/GloriaFitsImage.h"

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

namespace gloria {

  //! Class for calculating map of median values for any variable on the image.
  /*! This class can be use to calculate median value map over the
   *  image. For each considered variable/parameter two maps are allocated: one
   *  for storing median of the parameter value and the second one for
   *  storing the estimated parameter spread. Image area is divided
   *  into Nbx x Nby sectors, median and spread is calculated for each
   *  sector separately. Spread is calculated from the ordered value
   *  array, as half of the distance between 17% and 83% of
   *  entries. Single GloriaMedianMap instance can store median+spread map
   *  pairs for arbitrary number of variables (default is one
   *  variable). However, all maps need to have same binning.
   *
   *  \par
   *  The class is inheriting from GloriaValueMap class for value 
   *  calculation and creating median or spread images. Using the calculated
   *  median and spread maps,  user can calculate expected
   *  values for any pixel in the image frame using different
   *  interpolation methods implemented for \a GloriaValueMap. 
   *
   * \par
   *  In particular this class can be used for calculating background
   *  and noise level maps for sky images, where background is defined
   *  as the median of image pixels in the sector and noise as pixel
   *  value spread. Dedicated methods are defined for calculating
   *  background and noise maps directly from the image. It is assumed
   *  that background and noise are always stored in the first two
   *  maps of the map vector.
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaMedianMap : public GloriaValueMap {

  public:

    //! Dummy constructor - does nothing (!)
    GloriaMedianMap() : GloriaValueMap() { };

    //! Constructor for median map
    /*! Two tables are allocated, for storing median and value spread
     *  estimate, for any parameter corresponding to the image
     *  given. Tables should be then filled with a call to \a CalculateMap
     *  Note that image size is not defined at this moment yet!
     */
    GloriaMedianMap(int Nbx, int Nby, int Nvar = 1);


    //! Constructor for median map corresponding to given image file
    /*! Two tables are allocated, for storing median and value spread
     *  estimate, for any parameter corresponding to the image
     *  given. Tables should be then filled with a call to \a CalculateMap
     */
    GloriaMedianMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int Nvar=1);

    //! Constructor. Allocates map by importing it from fits table
    /*! Creates maps by reading them from fits table. The table has to
     *  created with \a GetTable() method previously, so all required
     *  parameters are stored in header keywords. Otherwise one should
     *  declare map first and read table contents with \a LoadFromTable.
     */
    GloriaMedianMap(GloriaFitsTable* mapTable) : GloriaValueMap(mapTable) { };

    //! Add another variable to median map set
    /*! Can be used to add maps (one for median, one for spread) for
     *  another parameter to the existing set of maps. Existing maps
     *  are not affected. 
     */
    void AddMedian();

    //! Add another variable to median map set and give it a name
    /*! Can be used to add maps (one for median, one for spread) for
     *  another parameter to the existing set of maps. Existing maps
     *  are not affected. 
     */
    void AddMedian(string name);


    //! Set name to the considered variable
    void SetMedianName(int ivar, string name);
 

    //! Return image filled with parameter values
    inline string GetMedianName(int ivar) {return _valueNames.at(2*ivar); } ;


    //! Calculate median and spread maps for given variable
    /*! This method can be used for calculating median and spread maps
     *  based on the table of parameter value measurements (\a Val array) 
     *  at different positions on the frame (specified by \a Xpos and
     *  \a Ypos arrays). All arrays contain \a nVal entries. \a iVar
     *  specifies the variable number (more than one variable can be
     *  considered in one GloriaMedianMap instance). Return value
     *  corresponds to minimal number of entries found in single map
     *  bin. Negative number is returned on error. If zero value is
     *  returned one or more map bins is not set. 
     */
    long CalculateMap(int iVar, int nVal, double* Xpos, double* Ypos, double* Val);

    //! Calculate median and spread maps for given variable
    /*! This method can be used for calculating median and spread maps
     *  based on the vector of parameter value measurements (\a Val vector) 
     *  at different positions on the frame (specified by \a Xpos and
     *  \a Ypos vectors). All vectors should contain same number of entries. \a iVar
     *  specifies the variable number (more than one variable can be
     *  considered in one GloriaMedianMap instance). Return value
     *  corresponds to minimal number of entries found in single map
     *  bin. Negative number is returned on error. If zero value is
     *  returned one or more map bins is not set. 
     */
    long CalculateMap(int iVar, vector<double>* Xpos, vector<double>* Ypos, vector<double>* Val);



    //! Calculate background and noise map for given image 
    /*! This method allocates map tables with Nbx x Nby bins and
     *   calculates background and noise levels in each bin. If number
     *   of bins is small (number of pixels in each bin very large)
     *   sampling factor can be used to speed the calculation
     *   up. Sampling N > 1 means than not all pixels are taken for
     *   calculation but only every N-th pixel. For each bin median of
     *   the pixel value distribution and noise as half of the
     *   distance between 17% and 83% of stars. 
     */
    void CalculateBgMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int sampling=1);

    //! Calculate map for given image (without changing map binning)
    void CalculateBgMap(const GloriaFitsImage* fitsImage, int sampling=1);


    //! Get expected median for pixel (ix,iy) using calculated map
    /*! Calculates expected median value in given pixel on the input frame
     *  by interpolating between map points. \a ivar parameter selects the
     *  variable (more than one map can be defined) and \a ipol defines
     *  interpolation order. Currently implemented  values are: 
     *     0 - no interpolation (take value from the bin center) and
     *     1 - linear interpolation between bin centers (default).
     *     2 - second order approximation (simplified).
     */
    inline double GetMedian(int ivar, int ix, int iy, int ipol=1)
                         { return GetMapValue(2*ivar,ix,iy,ipol); } ;

    //! Get expected median level at position (x,y) using calculated map
    inline double GetMedian(int ivar, double x, double y, int ipol=1)
                         { return GetMapValue(2*ivar,x,y,ipol); } ;

    //! Get expected median level for pixel ipix  using calculated map
    inline double GetMedian(int ivar, long ipix, int ipol=1)
                        { return GetMapValue(2*ivar,ipix,ipol); } ;

    //! Get spread of the variable for pixel (ix,iy) using calculated map
    inline double GetSpread(int ivar, int ix, int iy, int ipol=1)
                         { return GetMapValue(2*ivar+1,ix,iy,ipol); } ;

    //! Get spread of variable at position (x,y) using calculated map
    inline double GetSpread(int ivar, double x, double y, int ipol=1)
                         { return GetMapValue(2*ivar+1,x,y,ipol); } ;

    //! Get spread for pixel ipix  using calculated map
    inline double GetSpread(int ivar, long ipix, int ipol=1)
                        { return GetMapValue(2*ivar+1,ipix,ipol); } ;


    //! Get background level for pixel (ix,iy) using calculated map
    /*! Calculates background level in given pixel on the input frame
     *  by interpolating between map points. \a ipol parameter defines
     *  interpolation order. Currently implemented  values are: 
     *     0 - no interpolation (take value from the bin center) and
     *     1 - linear interpolation between bin centers (default).
     *     2 - second order approximation (simplified).
     */
    inline double GetBgLevel(int ix, int iy, int ipol=1)
                         { return GetMapValue(0,ix,iy,ipol); } ;

    //! Get background level at position (x,y) using calculated map
    inline double GetBgLevel(double x, double y, int ipol=1)
                         { return GetMapValue(0,x,y,ipol); } ;

    //! Get background level for pixel ipix  using calculated map
    inline double GetBgLevel(long ipix, int ipol=1)
                        { return GetMapValue(0,ipix,ipol); } ;

    //! Get noise level for pixel (ix,iy) using calculated map
    inline double GetNoise(int ix, int iy, int ipol=1)
                         { return GetMapValue(1,ix,iy,ipol); } ;

    //! Get noise level at position (x,y) using calculated map
    inline double GetNoise(double x, double y, int ipol=1)
                         { return GetMapValue(1,x,y,ipol); } ;

    //! Get noise level for pixel ipix  using calculated map
    inline double GetNoise(long ipix, int ipol=1)
                        { return GetMapValue(1,ipix,ipol); } ;



    //! Returns pointer to the median map array
    inline double* GetGloriaMedianMap(int ivar=0) {return _valueMaps.at(2*ivar); };

    //! Returns pointer to the spread map array
    inline double* GetSpreadMap(int ivar=0) {return _valueMaps.at(2*ivar+1); };

    //! Returns pointer to the background map array
    inline double* GetBgMap() {return _valueMaps.at(0); };

    //! Returns pointer to the noise map array
    inline double* GetNoiseMap() {return _valueMaps.at(1); };



    //! Returns mean of median map
    inline  double GetMedianMean(int ivar=0){ return GetMapMean(2*ivar);} ;

    //! Returns mean of spread map
    inline double GetSpreadMean(int ivar=0){ return GetMapMean(2*ivar+1);} ;

    //! Returns mean of background map
    inline double GetBgMean() { return GetMapMean(0); };

    //! Returns mean of noise map
    inline double GetNoiseMean() { return GetMapMean(1); } ;



    //! Return image filled with median values
    inline GloriaFitsImage* GetMedianImage(int ivar=0, int ipol=1)
                    { return GetMapImage(2*ivar, ipol);}  ;

    //! Return image filled with spread values
    inline GloriaFitsImage* GetSpreadImage(int ivar=0, int ipol=1)
                    { return GetMapImage(2*ivar+1, ipol);}  ;


    //! Return image filled with background values
    inline GloriaFitsImage* GetBgImage(int ipol=1)
                    { return GetMapImage(0, ipol);}  ;

    //! Return image filled with noise values
    inline GloriaFitsImage* GetNoiseImage(int ipol=1)
                    { return GetMapImage(1, ipol);}  ;

    //! Allocate new map sets with given binning. All data and image size information are lost.
    void InitializeMaps(int Nbx, int Nby, int Nvar);

  } ;


}

#endif



