// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaNetworkMgr_h
#define GloriaNetworkMgr_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "GloriaTelescope.h"

namespace gloria{

  //! Class for storing GLORIA network configuration
  /*! This class allows to store configuration of the GLORIA network,
   *  both the observatories as well as databases and other network
   *  resources. Configuration is read from XML file. This should
   *  reduce number of parameters which has to be passed to the
   *  processors via the steering file.
   *
   *  Current version implements only telescope parameters.
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaNetworkMgr {

  friend std::ostream& operator<< (  std::ostream& , GloriaNetworkMgr& ) ;

  public:

    //! Default constructor
    /*! Default constructor creates empty configuration
     */
    GloriaNetworkMgr();

    //! File constructor
    /*! Creates configuration based on input XML file
     */
    GloriaNetworkMgr(std::string fileName);

    //! Destructor
    ~GloriaNetworkMgr();

    //! Load network configuration from file
    void LoadConfigXML(std::string fileName);

    //! Get configuration ID
    inline long GetID() { return _id;} ;

    //! Get configuration date
    inline long GetDate() { return _date;} ;

    //! Get configuration author
    inline std::string GetAuthor() { return _author;} ;


    //! Number of telescopes defined in the configuration
    inline int GetTelescopeNumber() { return _nTelescopes;} ;

    //! Check if telescope configuration is defined
    inline bool IsTelescopeDefined(std::string name) 
         { return _telescopeNameMap.find(name)!=_telescopeNameMap.end();};

    //! Check if telescope configuration is defined
    inline bool IsTelescopeDefined(int id) 
         { return _telescopeIdMap.find(id)!=_telescopeIdMap.end();};

    //! Get telescope configuration by name or alias
    GloriaTelescope*  GetTelescope(std::string name);

    //! Get telescope configuration by telescope ID
    GloriaTelescope*  GetTelescope(int id);

protected:

    long _id;
    long _date;
    std::string _author;

    std::vector< GloriaTelescope > _telescopeVec; 

    std::map< std::string, int >  _telescopeNameMap;

    std::map< int, int >  _telescopeIdMap;

    int _nTelescopes;

  };


std::ostream& operator<< (  std::ostream& s, GloriaNetworkMgr& n ) ;

}

#endif



