// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaObjectList_h
#define GloriaObjectList_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */


#include "GloriaFitsTable.h"

using namespace std;

namespace gloria{

  //! Class for storing object (star images on CCD) lists 
  /*! This class defines GloriaFitsTable with predefined columns for
   *  storing object position on CCD ("CCD_X" and "CCD_Y") and object
   *  brightness ("Signal"). This makes sure that object lists will be
   *  exchangeable between processors. User can add additional columns
   *  if needed. 
   * 
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */

  class GloriaObjectList : public GloriaFitsTable {


  public:

    //! Default constructor
    /*! Default constructor cretes empty table
     */
    GloriaObjectList();


    // Data access methods
    // ===================

    //! Add new object to the list
    inline void AddObject(double ccdx, double ccdy, double signal) { AddRow(ccdx,ccdy,signal); };


    //! Add new object to the list
    inline void AddObject(double ccdx, double ccdy) { AddRow(ccdx,ccdy); };


    //! Add new object to the list
    inline void AddObject() { AddRow(); };


    //! Returns number of objects in the list
    inline int GetObjectNumber() {return _nRow;};

    //! Get object X position on CCD
    inline double GetObjectXpos( int iObject) 
             { return ((GloriaTableColumnOf<double> *)_columns.at(0))->At(iObject); };

    //! Get object Y position on CCD
    inline double GetObjectYpos( int iObject) 
             { return ((GloriaTableColumnOf<double> *)_columns.at(1))->At(iObject); };

    //! Get object signal
    inline double GetObjectSignal( int iObject) 
             { return ((GloriaTableColumnOf<double> *)_columns.at(2))->At(iObject); };

     //! Returns vector of object positions on CCD in X
     inline vector<double>& GetXposVector() 
             { return ((GloriaTableColumnOf<double> *)_columns.at(0))->Data(); };

     //! Returns vector of object positions on CCD in Y
     inline vector<double>& GetYposVector() 
             { return ((GloriaTableColumnOf<double> *)_columns.at(1))->Data(); };


     //! Returns vector of object signal values
     inline vector<double>& GetSignalVector() 
             { return ((GloriaTableColumnOf<double> *)_columns.at(2))->Data(); };

    //! Set object X position on CCD
    inline void SetObjectXpos( int iObject, double ccdx) 
             { ((GloriaTableColumnOf<double> *)_columns.at(0))->At(iObject) = ccdx; };

    //! Set object Y position on CCD
    inline void SetObjectYpos( int iObject, double ccdy) 
             { ((GloriaTableColumnOf<double> *)_columns.at(1))->At(iObject) = ccdy; };

    //! Set object signal
    inline void SetObjectSignal( int iObject, double signal) 
             { ((GloriaTableColumnOf<double> *)_columns.at(2))->At(iObject) = signal; };

    //! Returns the table itself
    inline GloriaFitsTable& GetTable() { return *(( GloriaFitsTable *) this ); };

  } ;


}

#endif
