// -*- mode: c++; mode: auto-fill; mode: flyspell-prig; -*-


#ifndef GloriaSkyCatalog_h
#define GloriaSkyCatalog_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */


#include "GloriaFitsTable.h"

using namespace std;

namespace gloria{

  //! Class for storing catalogs of sky objects (eg. stars) 
  /*! This class defines GloriaFitsTable with predefined columns for
   *  storing position of objects on the sky ("RA" and "DEC"). It is
   *  expected that RA and DEC values are given in degrees according
   *  to J2000. User can add additional columns if needed. This makes
   *  sure that catalogs will be exchangeable between processors and
   *  allows implementing general matching algorithms. 
   * 
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */

  class GloriaSkyCatalog : public GloriaFitsTable {


  public:

    //! Default constructor
    /*! Default constructor cretes empty table
     */
    GloriaSkyCatalog();


    // Data access methods
    // ===================

    //! Add new object to the list
    inline void AddObject(double ra, double dec) { AddRow(ra,dec); };


    //! Add new object to the list
    inline void AddObject() { AddRow(); };


    //! Returns number of objects in the list
    inline int GetObjectNumber() {return _nRow;};

    //! Get object RA position 
    inline double GetObjectRA( int iObject) 
          { return ((GloriaTableColumnOf<double> *)_columns.at(0))->At(iObject); };

    //! Get object DEC position
    inline double GetObjectDEC( int iObject) 
             { return ((GloriaTableColumnOf<double> *)_columns.at(1))->At(iObject); };

    //! Returns vector of object positions in RA
    inline vector<double>& GetRAVector() 
             { return ((GloriaTableColumnOf<double> *)_columns.at(0))->Data(); };

    //! Returns vector of object positions in DEC
    inline vector<double>& GetDECVector() 
             { return ((GloriaTableColumnOf<double> *)_columns.at(1))->Data(); };


    //! Set object RA position
    inline void SetObjectRA( int iObject, double ra) 
             { ((GloriaTableColumnOf<double> *)_columns.at(0))->At(iObject) = ra; };

    //! Set object DEC position
    inline void SetObjectDEC( int iObject, double dec) 
             { ((GloriaTableColumnOf<double> *)_columns.at(1))->At(iObject) = dec; };

    //! Check catalog table structure
    bool  CheckCatalog();


    //! Returns the table itself
    inline GloriaFitsTable& GetTable() { return *(( GloriaFitsTable *) this ); };

    //! Match object in catalog by its position on the sky
    /*! This method finds matching objects in the catalog for given
     *  values of RA and DEC (in degrees). Index of the matching raw
     *  is returned or -1 if no match found. \a cut parameter is used
     *  as the matching tolerance (maximum distance of the matched
     *  object in degrees). If \a cut==0, closes object is returned
     *  (without any constraints). 
     */
    int MatchObject(double ra, double dec, double cut);


    //! Find all matching object in catalog at given position on the sky
    /*! This method finds matching objects in the catalog for given
     *  values of RA and DEC (in degrees). Vector of index to the
     *  matching raws is returned (empty vector no match found). 
     *  \a distMax and \a distMin parameters determine the maximal and
     *  minimal matching radius (in degrees). The method can be used
     *  to search for neighbouring objects (eg. reference stars) for
     *  given star (in that case one should set \a distMin > 0 to
     *  avoid matching object to itself). Returned objects are sorted
     *  by the distance (from the smallest to the largest one) to the
     *  given position. 
     */
    vector<long> MatchAllObjects(double ra, double dec, double distMax, double distMin=0.);


    //! Match list of objects by their position on the sky
    /*! This method can be used to find matching objects for all
     *  entries in the object table. Pointers to RA and DEC value arrays
     *  need to be given. Pointer to array of matched sky objects (indexes in
     *  the catalog table) is returned. Procedure is faster if the
     *  catalog is prevuiously sorted in DEC (with \a SortTable("DEC") ).
     */
    long* MatchObjectArray(int nObject, double* RaArr, double* DecArr, double cut);

    
    //! Match list of objects by their position on the sky
    /*! This method can be used to find matching objects for all
     *  entries in the object table. Pointers to RA and DEC value vectors
     *  need to be given. Vector of matched sky objects (indexes in
     *  the catalog table) is returned. Procedure is faster if the
     *  catalog is prevuiously sorted in DEC (with \a SortTable("DEC") ).
     */
    vector<long> MatchObjectVector(vector<double>* RaVec, vector<double>* DecVec, double cut);

    
    //! Match objects in the table with the catalog, and copy catalog data
    /*! This method can be used to find matching objects for all
     *  entries in the object table. Pointers to the object table has to
     *  be given, as well as names of columns with RA and DEC values.
     *  Vector of matched sky objects (indexes in the catalog table)
     *  is returned. In addition, columns listed in \a CopyNames vector
     *  are copied from the catalog to the table being matched (can be
     *  used eg. to store catalog magnitudo, position, or star ID). If
     *  \a NewNames vector is specified (not empty) column names are
     *  changed accordingly. Procedure is faster if the
     *  catalog is prevuiously sorted in DEC (with \a SortTable("DEC") ).
     */
    vector<long> MatchObjectTable(GloriaFitsTable *table, double cut,
 vector<string> RaDecNames, vector<string> CopyNames, vector<string> NewNames );

    //! Match objects in the table with the catalog (using default column names)
    vector<long> MatchObjectTable(GloriaFitsTable *table, double cut);

    //! Match objects in the table with the catalog 
    vector<long> MatchObjectTable(GloriaFitsTable *table, double cut, vector<string> RaDecNames);

  private:

    //! Calculate distance in the sky between two objects
  double  SkyDistance(double ra1, double dec1, double ra2, double dec2);
    
  } ;


}

#endif
