// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaTableColumn_h
#define GloriaTableColumn_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include <string>
#include <vector>
#include <map>
#include <sstream>

#include "GloriaSTLTypes.h"

using namespace std;

namespace gloria{


  //! Enumerator for allowed column type classes
  /*! Table column with any content can be defined using template
   *  class gloria::GloriaTableColumnOf. However, for some methods
   *  implementation depends on the value type. To avoid possible
   *  information loss more precise types were selected for integer
   *  and floating point numbers i.e. \a long and \a double.
   * Five different column types are currently implemented:
   * \li \a LongColumn - column of long integer values (long)
   * \li \a DoubleColumn - column of double precision floating point values (double)
   * \li \a StringColumn - column of strings (string)
   * \li \a LongVectorColumn - each column cell is a vector 
   *                          of integers (vector<long>)
   * \li \a DoubleVectorColumn - each cell is a vector of floating
   *                            point numbers (vector<double>)
   * \li \a UndefinedColumn - returned if wrong or undefined data type
   *                          is detected  
   */
  enum TableColumnType
      { UndefinedColumn = 0,
      LongColumn,
      DoubleColumn,
      StringColumn,
      LongVectorColumn,
      DoubleVectorColumn
      };

  //! Helper function recognizing column type.
  /*! Helper function returning column type based on the element
   *  type. Returns \a UnidefinedColumn if not recognized.
   */
template<class T> 
TableColumnType GetValueType(T val);

  //! Helper function returning name of the column type.
  /*! Helper function returning column type as a string, based on the element
   *  type. Returns \a Unidefined if not recognized.
   */
template<class T> 
string GetValueTypeName(T val);


 //! Abstract class. Used for deriving classes with different column types
  /*! 
   *  This class defines only the column header and no data storage,
   *  as the data type is not defined yet. However, we know that the
   *  data are stored in a std::vector<> so some virtual methods can
   *  already be defined.
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
class GloriaTableColumn {

public: 

  GloriaTableColumn(string Name, string Unit = "", string Comment = "") : 
                         _name(Name), _unit(Unit), _comment(Comment){}

  inline virtual ~GloriaTableColumn() {}

  //! Column name. Used to address columns in the table
  inline string& Name() { return _name; }

  //! Units for values stored in this column
  inline string& Unit() { return _unit; }

  //! Column comment, can be used to store additional information
  inline string& Comment() { return _comment; }

  //! Column type defining type of elements stored
  inline TableColumnType& Type() { return _type; }

  //! Column type name (string) for elements stored
  string TypeName(); 

  //! Column size (size of vector storing column elements)
  virtual int Size() = 0;

  //! Clear column (clear vector of column elements)
  virtual void Clear() = 0;

  //! Add empty cell at the end of the column
  virtual void AddRow() = 0;

  //! Resize column to given number of rows
  virtual void Resize(int Nrow) = 0;

  //! Delete given cell 
  virtual void DeleteRow(int iRow) = 0;

  //! Returns size of given cell
  /*! Returns:
   * \li 0 - if row \a i is not found 
   * \li 1 - if cell is a single number or string
   * \li size() - if cell is a vector
   */
  virtual int ValSize(int i) = 0;

  //! Return value of element stored in given cell as long.
  /*! Can be used only for \a LongColumn or \a LongVectorColumn.
   *  For other column types 0 is returned. Second argument is used
   *  only if cell is a vector.
   */
  virtual long LongVal(int i, int j=0) = 0;

  //! Return value of element stored in given cell as double.
  /*! Can be used for all column types except \a StringColumn 
   *  (if this case 0. is returned). Second argument is used
   *  only if cell is a vector. Integers are converted to floats.
   */
  virtual double DoubleVal(int i, int j=0) = 0;

  //! Return element stored in given cell as string
  /*! If cell contains a vector all elements are included, separated
   *  by space (' '). If unit is defined it is appended at the end of
   *  the sting.  
   */
  virtual string StringVal(int i) = 0;

  //! Return element stored in given cell as string
  /*! If cell contains a vector only one vector element (indicated by
   * second argument) is returned. If unit is defined it is appended
   * at the end of the sting.
   */
  virtual string StringVal(int i, int j) = 0;


  //! Return pointer to column copy
  virtual GloriaTableColumn* Clone() = 0;


  //! Return pointer to empty column duplicate
  virtual GloriaTableColumn* EmptyClone() = 0;


  //! Copy cell content from other column
  virtual void CopyCellValue(GloriaTableColumn* source, int isource, int itarget=-1) = 0;

  
  //! Generate list of cell indexes ordered by cell value
  virtual long* CreateSortIndex(bool reverse=false) = 0;
  
  //! Sort column cells according to the index table
  virtual void SortByIndex(long* index) = 0;
  
protected:
  string _name;
  string _unit;
  string _comment;
  TableColumnType _type;
};      



//! Table column class for different content types
template <class T>
class GloriaTableColumnOf : public GloriaTableColumn {

public:
  GloriaTableColumnOf(string Name, T empty, string Unit = "", string Comment = "") : 
         GloriaTableColumn(Name,Unit,Comment), _emptyVal(empty) 
                                { _type = GetValueType(empty); }

  inline  ~GloriaTableColumnOf() {}

  //! Direct access to the data stored in the table 
  inline vector<T>& Data() { return _data;}

  //! Direct access to single cell
  inline T& At(int i) { return _data.at(i); }

  //! Element used to initialize empty cells when new row is added
  inline T& EmptyVal() { return _emptyVal; }

  //! Column size (size of vector storing column elements)
  inline int Size() { return _data.size(); }

  //! Clear column (clear vector of column elements)
  inline void Clear() {  _data.clear(); }

  //! Add element to the column
  inline void Push_Back(T val) { _data.push_back(val); }

  //! Add empty cell at the end of the column
  inline void AddRow() { _data.push_back(_emptyVal); }

  //! Resize column to given number of rows
  inline void Resize(int Nrow){_data.resize(Nrow,_emptyVal); }

  //! Delete given cell 
  inline void DeleteRow(int iRow) { _data.erase(_data.begin()+iRow); }


  //! Returns size of given cell
  /*! Returns:
   * \li 0 - if row \a i is not found 
   * \li 1 - if cell is a single number or string
   * \li size() - if cell is a vector
   */
  int ValSize(int i);


  //! Return value of element stored in given cell as long.
  /*! Can be used only for \a LongColumn or \a LongVectorColumn.
   *  For other column types 0 is returned. Second argument is used
   *  only if cell is a vector.
   */
  long LongVal(int i, int j=0);


  //! Return value of element stored in given cell as double.
  /*! Can be used for all column types except \a StringColumn 
   *  (if this case 0. is returned). Second argument is used
   *  only if cell is a vector. Integers are converted to floats.
   */
  double DoubleVal(int i, int j=0);


  //! Return element stored in given cell as string
  /*! If cell contains a vector all elements are included, separated
   *  by space (' '). If unit is defined it is appended at the end of
   *  the sting.  
   */
  string StringVal(int i);

  //! Return element stored in given cell as string
  /*! If cell contains a vector only one vector element (indicated by
   * second argument) is returned. If unit is defined it is appended
   * at the end of the sting.
   */
  string StringVal(int i, int j);

  //! Return pointer to column copy
  GloriaTableColumn* Clone();

  //! Return pointer to empty column duplicate
  GloriaTableColumn* EmptyClone();

  //! Copy cell content from other column
  void CopyCellValue(GloriaTableColumn* source, int isource, int itarget=-1);

  //! Generate list of cell indexes ordered by cell value
  long* CreateSortIndex(bool reverse=false);
  
  //! Sort column cells according to the index table
  void SortByIndex(long* index);
  
protected:

  T _emptyVal;
  vector<T> _data;
};



}

#endif



