// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaTelescope_h
#define GloriaTelescope_h 1


/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "luiza/StringParameters.h"
#include <iostream>
#include <vector>
#include <string>
#include <map>

namespace gloria{

  //! Class for storing GLORIA telescope parameters
  /*! This class allows to store telescope parameters, which can be
   *  used whene analysing data. Telescope parameters are a part of
   *  the GLORIA network configuration. For each telescope its network
   *  ID, name and position (longitude, latitude and elevation) should
   *  be defined, as well as an arbitrary number of additional
   *  parameters (name/value pairs). 
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaTelescope {

  friend std::ostream& operator<< (  std::ostream& , GloriaTelescope& ) ;

  public:

    //! Default constructor
    /*! Default constructor creates empty configuration
     */
    GloriaTelescope();

    //! Copy constructor
    /*! Copy constructor required eq. for creating vector of telescopes
     */
    GloriaTelescope(const GloriaTelescope &t);

    //! Named constructor
    /*! Creates configuration with given ID and name
     */
    GloriaTelescope(int id, std::string name);

    //! Destructor
    ~GloriaTelescope();

    //! Define telescope ID
    inline void SetID(int id) { _id = id; };

    //! Add telescope name/alias
    inline void SetName(std::string name) { _name.push_back(name); };

    //! Define telescope position
    inline void SetPosition(double lon, double lat, double elev) 
                  { _longitude=lon;  _latitude=lat; _elevation=elev;};

    //! Add parameter
    void AddParameter(std::string name, std::string value);



    //! Get telescope ID
    inline int GetID() { return _id;} ;

    //! Get number of telescope name/aliases
    inline int GetNameNumber() { return _name.size();} ;

    //! Get telescope name
    inline std::string GetName(int iname=0) { return _name.at(iname);} ;


    //! Check if telescope parameter is set
    inline bool IsParameterSet(std::string key) { return _parameters->isParameterSet(key);} ;

    //! Get telescope parameter as a string
    inline std::string GetStringParameter(std::string key) { return _parameters->getStringVal(key);} ;

    //! Get telescope parameter as int
    inline int GetIntParameter(std::string key) { return _parameters->getIntVal(key);} ;

    //! Get telescope parameter as float
    inline float GetFloatParameter(std::string key) { return _parameters->getFloatVal(key);} ;

    //! Return parameters defined for this telescope
    virtual luiza::StringParameters* parameters() { return _parameters ; } 

private:

    int _id;

    std::vector< std::string >  _name;

    double _longitude, _latitude, _elevation;

    // StringParameters class allow for manipulation of arbitrary parameters
    // parameters are stored in  std::map< std::string ,  std::vector< std::string > > 


    luiza::StringParameters* _parameters; 
  };



std::ostream& operator<< (  std::ostream& s, GloriaTelescope& t ) ;

}

#endif

