// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-


#ifndef GloriaValueMap_h
#define GloriaValueMap_h 1

#include "gloria/GloriaFitsTable.h"
#include "gloria/GloriaFitsImage.h"

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

namespace gloria {

  //! Class for constructing parameter/value maps for an image
  /*! This class can be use to calculate value of parameters for any
   *  position on a sky image (object of GloriaFitsImage class). 
   *  Arbitrary number of parameters can be stored. User has to define
   *  size of value grid which is used to set parameter values, and
   *  the method can be used to interpolate between these values.
   *  This class is also used to build more specialized classes as
   *  \a MedianMap (eg. for background and noise calculation).
   *
   * \author A.F.Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   *
   */
  class GloriaValueMap {

  public:


    //! Dummy constructor - does nothing (!)
    GloriaValueMap();

    //! Constructor. Allocates map tables for given map size.
    /*! Only one table is allocated by default (Nmap=1), but user can
     *  request allocating many maps at the same time. Note: all maps
     *  should correspond to the same image (image size). Image size
     *  has to be defined by call to SetImageSize().
     */
    GloriaValueMap(int Nbx, int Nby, int Nmap=1);


    //! Constructor. Allocates map tables for given image and map size.
    /*! Only one table is allocated by default (Nmap=1), but user can
     *  request allocating many maps at the same time. 
     */
     GloriaValueMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int Nmap=1);


    //! Constructor. Allocates map by importing it from fits table
    /*! Creates maps by reading them from fits table. The table has to
     *  created with \a GetTable() method previously, so all required
     *  parameters are stored in header keywords. Otherwise one should
     *  declare map first and read table contents with \a LoadFromTable.
     */
     GloriaValueMap(GloriaFitsTable* mapTable);


    //! Destructor
    virtual ~GloriaValueMap();

    //! Set size of image 
    void SetImageSize(const GloriaFitsImage* fitsImage);

    //! Set size of image 
    void SetImageSize(int SizeX, int SizeY);


    //! Add another map to current map set
    /*! Can be used to add map for another parameter to the existing
     *  set of maps. Existing maps are not affected.
     */
    void AddMap();

    //! Add another map to current map set, giving it a name
    /*! Can be used to add map for another parameter to the existing
     *  set of maps. Existing maps are not affected.
     */
    void AddMap(string name);

    //! Set name for existing map

    inline void SetMapName(int imap, string name) 
                                    { _valueNames.at(imap) = name ; } ;

    //! Get value from map \a imap for pixel (ix,iy) 
    /*! Calculates parameter value in given pixel of the frame
     *  by interpolating between map points. \a ipol parameter defines
     *  interpolation order. Currently implemented  values are: 
     *     0 - no interpolation (take value from the bin center) and
     *     1 - linear interpolation between bin centers (default).
     */
    double GetMapValue(int imap, int ix, int iy, int ipol=1);

    //! Get value from map \a imap for pixel \a ipix
    double GetMapValue(int imap, long ipix, int ipol=1);

    //! Get map value for arbitrary position on the image
    double GetMapValue(int imap, double x, double y, int ipol=1);

    //! Get number of maps
    inline int Nmap() { return _valueMaps.size(); };

    //! Get map binning in X
    inline int NMapX() { return _nMapX; };

    //! Get map binning in Y
    inline int NMapY() { return _nMapY; };

    //! Get image pixels in X
    inline int ImgSizeX() { return _sizeX; };

    //! Get image pixels in Y
    inline int ImgSizeY() { return _sizeY; };

    //! Returns pointer to the map array
    inline double* GetMap(int imap) {return _valueMaps.at(imap); };

    //! Return image filled with parameter values
    inline string GetMapName(int imap) {return _valueNames.at(imap); } ;

    //! Returns mean value of map entries
    double GetMapMean(int imap);

    //! Return image filled with parameter values
    GloriaFitsImage* GetMapImage(int imap, int ipol=1);

    //! Return table containing all defined maps as columns
    GloriaFitsTable* GetTable();

    //! Initialize value map structures from the table
    bool LoadFromTable(GloriaFitsTable* mapTable);

    //! Clear all defined maps. Clears also map binning and image size information.
    void ClearMaps();

    //! Allocate new, empty maps with given binning. All data and image size information are lost.
    void InitializeMaps(int Nbx, int Nby, int Nmap);

protected:

    // Different interpolation orders
    void FillMapImage0(double* map, GloriaFitsImage* mapImage);
    void FillMapImage1(double* map, GloriaFitsImage* mapImage);
    void FillMapImage2(double* map, GloriaFitsImage* mapImage);

    // Number of map bins in X,Y
    int _nMapX, _nMapY, _nMapXY;

    // Number of bins in the input image
    int _sizeX, _sizeY;
    long _sizeXY;

    // Size of map bin
    int _binX, _binY, _binXY;

    // vector of map arrays
    std::vector<double*> _valueMaps;

    // vector of map names
    std::vector<string> _valueNames;

  } ;


}

#endif



