// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef FitsImageReader_h
#define FitsImageReader_h 1

#include "DataSourceProcessor.h"
// #include "Processor.h"
// #include "data/GloriaDataContainer.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor for reading input FITS images 
  /*! This processor reads FITS images from given files.
   *  It is usually the first processor in the analysis
   *  chain. Multiple instances can be used (under different names) to
   *  allow reading sky images as well as dark, bias or flat frames.
   * 
   *  \par Input 
   *  Names of input FIT files can be given in two ways: directly, or
   *  by specifying names of list files (i.e. files containing image
   *  file names, one per line). 
   *    
   *  \param FitsFileList List of FITS files to be read
   * 
   *  \param ListFileList Names of files containing FITS file lists (one per line)
   *
   *  \param ImagesPerLoop Number of images to be read in each
   *   processor call. Default is one. If zero value is set, all images
   *   from the list are read at the first call (!)
   *
   *  \par Output
   *   Images read as gloria::GloriaFitsImage are added to specified
   *   image collection (named image vector) and put into global 
   *   gloria::GloriaDataContainer
   *
   *  \param ImageCollectionName Name of the created
   *    GloriaImageCollection (default is "FitsImages");
   *
   *  \param PermanentCollection Flag indicating that created
   *    collection should be kept in memory for the whole
   *    processing time. This is needed in case of eg. calibration frames
   *    (darks, flats) or collections containing analysis
   *    results. Default is \a false
   *
   * 
   *  \author A.F. Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   */

   class FitsImageReader : public DataSourceProcessor {
    
  public:
    
    virtual FitsImageReader*  newProcessor();
    
    
    FitsImageReader() ;
    
    /** Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /** Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /** Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! List of FITS files to be read
    std::vector< std::string > _fitsFileList ;

    //! Names of files containing FITS file lists (one per line) - alternative method
    std::vector< std::string > _listFileList ;

    //! Name of the image collection to which images from file should be stored
    std::string _imageCollectionName ;

    //! Flag for collections, which should not be deleted after loop is finished
    bool _permanentCollection;

    //! Number of images to be read per processing loop (0 for all)
    int _imagesPerLoop ;

    //! Loop counter
    int _nLoops ;

    //! Image counter
    int _nImages ;

    //! File counter
    int _nFiles ;

    /**Test method for const.*/
    void printEndMessage() const ;
  } ;

 } // end namespace luiza
#endif
