// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef FitsImageWriter_h
#define FitsImageWriter_h 1

// #include "DataSourceProcessor.h"
#include "Processor.h"
// #include "data/GloriaDataContainer.h"

using namespace gloria;

namespace luiza {


  //! Processor for writing FITS images to output files
  /*! This processor writes FITS images to output files.
   *  Can be used to store analysis results.
   *
   *  \par Input 
   *  Images are taken from the data collection.
   *  User has to specify in the steering file:
   *
   *  \param ImageCollectionName -
   *    Name of the image collection which should be stored to
   *    file. If image collection is marked as 'permanent', it will
   *    only be stored at 'concludeAnalysis' stage (not for every loop).
   *
   *  \par Output
   *  GloriaFitsImages from given collection 
   *  are stored as FITS files. Output file names consist of file name 
   *  root, image number (starting from 1) and \a .fit extension.
   *
   *  \param FitsFileNameRoot -
   *   Output file name root, which will be appended by the image number 
   *   and extension. 
   *
   *  \param BitsPerPixel - 
   *   Bits per pixel in the output fits file. Possible values are:
   *
   *  \li 8 : 8-bit byte pixels
   *
   *  \li 16 : 16 bit (short) integer pixels
   *
   *  \li 32 : 32 bit (long) integer pixels
   *
   *  \li -32 : 32 bit floating point (float) pixels
   *
   *  \li -64 : 64 bit floating point (double) pixels
   *
   *  \li 0 : to keep image format unchanged
   *
   *  \param PixelOffset - 
   *    BZERO keyword value, used when storing pixel values to fits
   *    file. Default is 0.
   *
   *  \param PixelScale - 
   *    BSCALE keyword value, used for scaling pixel values when
   *    storing to fits file. Default is 1.
   *
   *
   *  \author L.W. Piotrowski, University of Warsaw
   */

   class FitsImageWriter : public Processor {

  public:

    virtual FitsImageWriter*  newProcessor();


    FitsImageWriter();

    //! Called at the begin of the job before anything is read.
    /*! Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    //! Called before first data processing loop 
    /*! Can be used to initialize data collections required for
     *  analysis, eg. define permanent collections, or fill them with
     *  external data (from file, database or net). It is possible
     * as GloriaDataContainer is already defined, contrary to init().
     */
    virtual void startAnalysis( GloriaDataContainer* ) ; 

    //! Called for every data processing loop 
    /*! This is where most of the work (analysis) is done. 
     * Data is exchanged between processors via GloriaDataContainer 
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    //! Called for every loop - right after processData() call
    /*! 
     *  Can be used to check processing and/or produce check plots.
     */
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops.
    /*! This method is called to finalize the analysis, after all
     * input data was read (or required number of loops was
     * reached). Analysis of data collected in permanent data
     * collections is possible here.
     */
    virtual void concludeAnalysis( GloriaDataContainer* ) ;

    //! Called after data processing is finished
    /*! Called after all input sources are closed for clean up.
     *  Notice that processors are called in the inverse order of the init()
     *  method so that resources allocated in the first processor also will 
     *  be available for all following processors.
     */
    virtual void end() ;

     //! Routine to do the actual image storage to file
     /*! This routine is called either from processData or from
      *  concludeAnalysis, to do the job required. Default is to store
      *  images for every loop, but permanent collections are stored only
      *  once, at the very end.
      */
     void  processImages( GloriaDataContainer* data ) ;

  protected:

    //! Name of the image collection to which images from file should be stored
    string _imageCollectionName ;

    //! Output file name root
    string _fitsFileNameRoot;

    //! Output file pixel format
    int _bitPix;

    //! Output file pixel scale (BSCALE)
    double _bScale;

    //! Output file pixel offset (BZERO)
    int _bZero;

    //! Loop counter
    int _nLoops ;

     //! Image counter, used to number file name
    int _nImages ;

     //! Counter for images successfully written to file
    int _nWritten ;


    void printEndMessage() const ;
  } ;

 } // end namespace luiza
#endif
