// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef FitsTableReader_h
#define FitsTableReader_h 1

#include "DataSourceProcessor.h"
// #include "Processor.h"
// #include "data/GloriaDataContainer.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor for reading data tables from input FITS files 
  /*! This processor reads data tables from FITS files.
   *  It can be used for reading eg. reference star lists or object lists
   *  stored previously. It is defined as luiza::DataSourceProcessor, so it
   *  can be the first processor in the analysis chain. 
   * 
   *  \par Input 
   *  Names of input fits files can be given in two ways: directly, or
   *  by specifying names of list files (i.e. text files containing table
   *  file names, one per line). 
   *    
   *  \param TableFileList List of fits files to be read
   * 
   *  \param ListFileList Names of files containing lists of fits file names (one per line)
   *
   *  \param TablesPerLoop Number of tables to be read in each
   *   processor call. Default is one. If zero value is set, all tables
   *   from the list are read at the first call (!)
   *
   *
   *  \par Output
   *   Tables read as gloria::GloriaFitsTable are added to specified
   *   table collection (named table vector) and put into global 
   *   gloria::GloriaDataContainer.
   *
   *  \param TableCollectionName Name of the created
   *    GloriaTableCollection (default is "FitsTable");
   *
   *  \param PermanentCollection Flag indicating that created
   *    collection should be kept in memory for the whole
   *    processing time. This is needed in case of eg. tables
   *    containing reference stars.
   *
   * 
   *  \author A.F. Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   */

   class FitsTableReader : public DataSourceProcessor {
    
  public:
    
    virtual FitsTableReader*  newProcessor();
    
    
    FitsTableReader() ;
    
    /** Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    
    /** Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /** Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! List of fits table files to be read
    std::vector< std::string > _tableFileList ;

    //! Names of files containing lists of table file names (one per line)
    std::vector< std::string > _listFileList ;

    //! Name of the table collection to which tables from file should be stored
    std::string _tableCollectionName ;

    //! Flag for collections, which should not be deleted after loop is finished
    bool _permanentCollection;

    //! Number of tables to be read per processing loop (0 for all)
    int _tablesPerLoop ;

    //! Loop counter
    int _nLoops ;

    //! Table counter
    int _nTables ;

    //! File counter
    int _nFiles ;

    /**Test method for const.*/
    void printEndMessage() const ;

    GloriaFitsTable* _inputTable; 
  } ;

 } // end namespace luiza
#endif
