// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef FitsTableWriter_h
#define FitsTableWriter_h 1

#include "Processor.h"

#include <fstream>

using namespace gloria;

namespace luiza {


  //! Processor for writing tables to output TEXT files
  /*! This processor writes tables to output FITS file
   *  It can be used to store analysis results, eg. object lists.
   *
   *  \par Input 
   *  Tables are taken from the data collection.
   *  User has to specify in the steering file:
   *
   *  \param TableCollectionName -
   *    Name of the table collection which should be stored to file(s)
   *
   *  \par Output
   *  GloriaFitsTables from given collection 
   *  are stored as FITS files. Output file names consist of file name 
   *  root, table number (starting from 1) and \a .tbl
   *  extension. Tables are stored at each loop unless the collection
   *  is permanent - then they are stored only at concludeAnalysis()
   *  call. 
   *
   *  \param FitsFileNameRoot -
   * Output file name root, which will be appended by the table number 
   * and extension
   *
   *  \param BinaryTable - flag to select binary table format
   *
   *
   *  \author A.F. Zarnecki, University of Warsaw
   */

   class FitsTableWriter : public Processor {

  public:

    virtual FitsTableWriter*  newProcessor();


    FitsTableWriter();

    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    /*! Called for before first data loop
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;


    /*! Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;


    virtual void check( GloriaDataContainer* data ) ;


    /*! Called after all data loops to conclude analysis
     */
    virtual void concludeAnalysis( GloriaDataContainer* data ) ;


    /*! Called after data processing for clean up.
     */
    virtual void end() ;


  protected:

    //! Name of the table collection which should be stored
    string _tableCollectionName ;

    //! Output file name root
    string _fitsFileNameRoot;

     //! Flag for output to single file
     bool _binaryTable;

     //! Table columns to be stored
     std::vector< std::string > _selectedColumns;

     //! Temporary table to prepare text output
     GloriaFitsTable* _outputTable;

     //! Loop counter
     int _nLoops ;

     //! Table counter, used to number file name
     int _nTables ;

     //! Counter for images successfully written to file
    int _nWritten ;


    void printEndMessage() const ;
  } ;

 } // end namespace luiza
#endif
