// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef TableTextReader_h
#define TableTextReader_h 1

#include "DataSourceProcessor.h"
// #include "Processor.h"
// #include "data/GloriaDataContainer.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor for reading data tables from input text files 
  /*! This processor reads data tables from given files.
   *  It is be used for reading eg. reference star list object lists
   *  stored previously. It is defined as luiza::DataSourceProcessor, so it
   *  can be the first processor in the analysis chain. 
   * 
   *  \par Input 
   *  Names of input text files can be given in two ways: directly, or
   *  by specifying names of list files (i.e. text files containing table
   *  file names, one per line). 
   *    
   *  \param TableFileList List of table text files to be read
   * 
   *  \param ListFileList Names of files containing list of table file names (one per line)
   *
   *  \param TablesPerLoop Number of tables to be read in each
   *   processor call. Default is one. If zero value is set, all tables
   *   from the list are read at the first call (!)
   *
   *  \par Data format
   *   As only data are read from text
   *   file, other parameters required to define a table need to
   *   be given as processor parameters. User has to specify
   *   ColumnNames and ColumnTypes. Optional parameters are
   *   ColumnDefaults and ColumnUnits (zero value or empty string will
   *   be used if missing).
   *
   *  \param ColumnNames  Names which will be given to data columns,
   *  which are read from the text file. 
   *
   *  \param ColumnTypes  Types of data columns in the text
   *  file. Valid types are I or L for long integer, F or D for
   *  double, S or T for text string. As space is used as column
   *  separator no spaces inside strings are allowes. 
   *
   *  \param ColumnDefaults  Default values for subsequent columns.
   *
   *  \param ColumnUnits  Units for subsequent columns.
   *
   *
   *  \par Output
   *   Tables read as gloria::GloriaFitsTable are added to specified
   *   table collection (named table vector) and put into global 
   *   gloria::GloriaDataContainer.
   *
   *  \param TableCollectionName Name of the created
   *    GloriaTableCollection (default is "TextTable");
   *
   *  \param PermanentCollection Flag indicating that created
   *    collection should be kept in memory for the whole
   *    processing time. This is needed in case of eg. tables
   *    containing reference stars.
   *
   * 
   *  \author A.F. Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   */

   class TableTextReader : public DataSourceProcessor {
    
  public:
    
    virtual TableTextReader*  newProcessor();
    
    
    TableTextReader() ;
    
    /** Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /** Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /** Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! List of text table files to be read
    std::vector< std::string > _tableFileList ;

    //! Names of files containing lists of table file names (one per line)
    std::vector< std::string > _listFileList ;

    //! Names of columns in the input table
    std::vector< std::string > _columnNames ;

    //! Types of columns in the input table
    std::vector< std::string > _columnTypes ;

    //! Units of columns in the input table
    std::vector< std::string > _columnUnits ;

    //! Default values for columns in the input table
    std::vector< std::string > _columnDefaults ;

    //! Name of the table collection to which tables from file should be stored
    std::string _tableCollectionName ;

    //! Flag for collections, which should not be deleted after loop is finished
    bool _permanentCollection;

    //! Number of tables to be read per processing loop (0 for all)
    int _tablesPerLoop ;

    //! Loop counter
    int _nLoops ;

    //! Table counter
    int _nTables ;

    //! File counter
    int _nFiles ;

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Template of input table, as defined by user
    GloriaFitsTable* _inputTable; 
  } ;

 } // end namespace luiza
#endif
