// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef TableTextWriter_h
#define TableTextWriter_h 1

#include "Processor.h"

#include <fstream>

using namespace gloria;

namespace luiza {


  //! Processor for writing tables to output TEXT files
  /*! This processor writes tables to output file
   *  in the plain text format
   *
   *  \par Input 
   *  Tables are taken from the data collection.
   *  User has to specify in the steering file:
   *
   *  \param TableCollectionName -
   *    Name of the table collection which should be stored to file(s)
   *
   *  \par Output
   *  GloriaFitsTables from given collection 
   *  are stored as TEXT files. Output file names consist of file name 
   *  root, table number (starting from 1) and \a .tbl
   *  extension. Tables are stored at each loop unless the collection
   *  is permanent - then they are stored only at concludeAnalysis()
   *  call. User can also choose to write all tables in the collection
   *  to one big file.
   *
   *  \param TextFileNameRoot -
   * Output file name root, which will be appended by the table number 
   * and extension
   *
   *  \param OneFile - flag to force storing all tables in one big file
   *
   *  \param SelectedColumns - names of columns to be stored in the
   *         file. If not specified (empty list) all columns are written.
   *
   *  \param AddCounter - flag to add row number as the first column
   *                       in the output file. If OneFile option is
   *                       used, table number is also added.
   *
   *  \param HeaderRecords - list of FITS header fields to be added to the table.
   *
   *  \author A.F. Zarnecki, University of Warsaw
   */

   class TableTextWriter : public Processor {

  public:

    virtual TableTextWriter*  newProcessor();


    TableTextWriter();

    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    /*! Called for before first data loop
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;


    /*! Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;


    virtual void check( GloriaDataContainer* data ) ;


    /*! Called after all data loops to conclude analysis
     */
    virtual void concludeAnalysis( GloriaDataContainer* data ) ;


    /*! Called after data processing for clean up.
     */
    virtual void end() ;


  protected:

    //! Name of the table collection which should be stored
    string _tableCollectionName ;

    //! Output file name root
    string _textFileNameRoot;

     //! Flag for output to single file
     bool _oneFile;

     //! Flag for output to single file
     bool _addCounter;

     //! Table columns to be stored
     std::vector< std::string > _selectedColumns;

     //! Header records to be stored
     std::vector< std::string > _headerRecords;

     //! Temporary table to prepare text output
     GloriaFitsTable* _outputTable;

     //! Output file pointer
     ofstream _outputStream;

     //! Loop counter
     int _nLoops ;

     //! Table counter, used to number file name
     int _nTables ;

     //! Counter for images successfully written to file
    int _nWritten ;


    void printEndMessage() const ;
  } ;

 } // end namespace luiza
#endif
