/*
 * GloriaAstrometry.cpp
 *
 *  Created on: Sep 14, 2012
 *	  Author: Lech Wiktor Piotrowski
 */

#include "gloria/GloriaAstrometry.h"

namespace gloria
{

// Default constructor
// Initializes astrometric transformations to 0
GloriaAstrometry::GloriaAstrometry()
{
	crpix[0]=crpix[1]=0;
	crval[0]=crval[1]=0;
	cd[0]=cd[1]=cd[2]=cd[3]=0;
	sip_order_a=sip_order_b=sip_order_ap=sip_order_bp=0;

        ref_ra=ref_dec=-100.;
}

GloriaAstrometry::GloriaAstrometry(GloriaFitsHeader *header)
{
	InitFromHeader(header);
}

void GloriaAstrometry::InitFromHeader(GloriaFitsHeader *header)
{
	crpix[0]=header->GetKeywordValueDouble("CRPIX1");
	crpix[1]=header->GetKeywordValueDouble("CRPIX2");
	crval[0]=header->GetKeywordValueDouble("CRVAL1");
	crval[1]=header->GetKeywordValueDouble("CRVAL2");
	cd[0]=header->GetKeywordValueDouble("CD1_1");
	cd[1]=header->GetKeywordValueDouble("CD1_2");
	cd[2]=header->GetKeywordValueDouble("CD2_1");
	cd[3]=header->GetKeywordValueDouble("CD2_2");
	sip_order_a=header->GetKeywordValueDouble("A_ORDER");
	sip_order_b=header->GetKeywordValueDouble("B_ORDER");
	sip_order_ap=header->GetKeywordValueDouble("AP_ORDER");
	sip_order_bp=header->GetKeywordValueDouble("BP_ORDER");

	// Initialize sip distortion polynomials
	// there are 2*sip_order coefficients, but for convienence we use a sip_order*sip_order 2D vector
	for (int i=0; i<=sip_order_a; i++)
	{
		vector<double> row_a, row_b, row_ap, row_bp; // Create an empty row
		ostringstream sip_name;
		for (int j=0; j<=sip_order_b; j++)
		{
			if(i+j>sip_order_a)
			{
				row_a.push_back(0); // Add an empty element (column) to the row
				row_b.push_back(0); // Add an empty element (column) to the row
				row_ap.push_back(0); // Add an empty element (column) to the row
				row_bp.push_back(0); // Add an empty element (column) to the row
			}
			else
			{
				sip_name.str(""); sip_name << "A_" << i << "_" << j;
				row_a.push_back(header->GetKeywordValueDouble((sip_name.str()).c_str()));
				sip_name.str(""); sip_name << "B_" << i << "_" << j;
				row_b.push_back(header->GetKeywordValueDouble((sip_name.str()).c_str()));
				sip_name.str(""); sip_name << "AP_" << i << "_" << j;
				row_ap.push_back(header->GetKeywordValueDouble((sip_name.str()).c_str()));
				sip_name.str(""); sip_name << "BP_" << i << "_" << j;
				row_bp.push_back(header->GetKeywordValueDouble((sip_name.str()).c_str()));
			}
		}
		sip_a.push_back(row_a); // Add the row to the main vector
		sip_b.push_back(row_b); // Add the row to the main vector
		sip_ap.push_back(row_ap); // Add the row to the main vector
		sip_bp.push_back(row_bp); // Add the row to the main vector
	}


ref_ra=ref_dec=-100.;
}

// Constructor from an input/output leblind_struct of the modified astrometry.net solver
// Currently we assume that all the polynomials are of the same order
GloriaAstrometry::GloriaAstrometry(leblind_struct *lestr)
{
	InitFromLeblind(lestr);
}

//! Initialize from an input/output leblind_struct of the modified astrometry.net solver
//! Currently we assume that all the polynomials are of the same order
void GloriaAstrometry::InitFromLeblind(leblind_struct *lestr)
{
	crpix[0]=lestr->solution[0];
	crpix[1]=lestr->solution[1];
	crval[0]=lestr->solution[2];
	crval[1]=lestr->solution[3];
	cd[0]=lestr->solution[4];
	cd[1]=lestr->solution[5];
	cd[2]=lestr->solution[6];
	cd[3]=lestr->solution[7];
	sip_order_a=sip_order_b=sip_order_ap=sip_order_bp=lestr->sip_order;

	// Initialize sip distortion polynomials
	// there are 2*sip_order coefficients, but for convienence we use a sip_order*sip_order 2D vector
	for (int i=0; i<=lestr->sip_order; i++)
	{
		vector<double> row_a, row_b, row_ap, row_bp; // Create an empty row
		for (int j=0; j<=lestr->sip_order; j++)
		{
			if(i+j>lestr->sip_order)
			{
				row_a.push_back(0); // Add an empty element (column) to the row
				row_b.push_back(0); // Add an empty element (column) to the row
				row_ap.push_back(0); // Add an empty element (column) to the row
				row_bp.push_back(0); // Add an empty element (column) to the row
			}
			else
			{
				row_a.push_back(lestr->solution_sip[0][i][j]);
				row_b.push_back(lestr->solution_sip[1][i][j]);
				row_ap.push_back(lestr->solution_sip[2][i][j]);
				row_bp.push_back(lestr->solution_sip[3][i][j]);
			}
		}
		sip_a.push_back(row_a); // Add the row to the main vector
		sip_b.push_back(row_b); // Add the row to the main vector
		sip_ap.push_back(row_ap); // Add the row to the main vector
		sip_bp.push_back(row_bp); // Add the row to the main vector
	}

ref_ra=ref_dec=-100.;
}

// Fill the header with astrometry results
void GloriaAstrometry::WriteAstrometry(GloriaFitsHeader *cur_frame)
{
        // To keep everything in one place and avoid problems - delete old astrometry, if any

        DeleteAstrometry(cur_frame);

	// astrometry header
	cur_frame->SetKeyword("CTYPE1", "\'RA---TAN-SIP\'", "TAN (gnomic) projection + SIP distortions");
	cur_frame->SetKeyword("CTYPE2", "\'DEC---TAN-SIP\'", "TAN (gnomic) projection + SIP distortions");
	cur_frame->SetKeywordInt("WCSAXES", 2, "");
	cur_frame->SetKeywordDouble("EQUINOX", 2000.0, "Equatorial coordinates definition (yr)");
	cur_frame->SetKeywordDouble("LONPOLE", 180.0, "");
	cur_frame->SetKeywordDouble("LATPOLE", 0.0, "");

	// astrometry values
	cur_frame->SetKeywordDouble("CRVAL1", crval[0], "");
	cur_frame->SetKeywordDouble("CRVAL2", crval[1], "");
	cur_frame->SetKeywordInt("CRPIX1", crpix[0], "");
	cur_frame->SetKeywordInt("CRPIX2", crpix[1], "");
	cur_frame->SetKeyword("CUNIT1", "'deg     '", "X pixel scale units");
	cur_frame->SetKeyword("CUNIT2", "'deg     '", "Y pixel scale units");
	cur_frame->SetKeywordDouble("CD1_1", cd[0], "");
	cur_frame->SetKeywordDouble("CD1_2", cd[1], "");
	cur_frame->SetKeywordDouble("CD2_1", cd[2], "");
	cur_frame->SetKeywordDouble("CD2_2", cd[3], "");

	// sip values
	cur_frame->SetKeywordInt("A_ORDER", sip_order_a, "");
	for(int i=0; i<=sip_order_a; i++) for(int j=0; (i+j)<=sip_order_a; j++)
	{
		// drop linear terms - that's how it is done in astrometry.net
		if(i+j<2) continue;
		char sip_keyword[128];
		sprintf(sip_keyword, "A_%d_%d", i, j);
		cur_frame->SetKeywordDouble(sip_keyword, sip_a[i][j], "");
	}
	cur_frame->SetKeywordInt("B_ORDER", sip_order_b, "");
	for(int i=0; i<=sip_order_b; i++) for(int j=0; (i+j)<=sip_order_b; j++)
	{
		// drop linear terms
		if(i+j<2) continue;
		char sip_keyword[128];
		sprintf(sip_keyword, "B_%d_%d", i, j);
		cur_frame->SetKeywordDouble(sip_keyword, sip_b[i][j], "");
	}
	cur_frame->SetKeywordInt("AP_ORDER", sip_order_ap, "");
	for(int i=0; i<=sip_order_ap; i++) for(int j=0; (i+j)<=sip_order_ap; j++)
	{
		if(i+j<1) continue;
		char sip_keyword[128];
		sprintf(sip_keyword, "AP_%d_%d", i, j);
		cur_frame->SetKeywordDouble(sip_keyword, sip_ap[i][j], "");
	}
	cur_frame->SetKeywordInt("BP_ORDER", sip_order_bp, "");
	for(int i=0; i<=sip_order_bp; i++) for(int j=0; (i+j)<=sip_order_bp; j++)
	{
		if(i+j<1) continue;
		char sip_keyword[128];
		sprintf(sip_keyword, "BP_%d_%d", i, j);
		cur_frame->SetKeywordDouble(sip_keyword, sip_bp[i][j], "");
	}
}

void GloriaAstrometry::DeleteAstrometry(GloriaFitsHeader *header)
{
	// remove astrometry header
	header->RemoveKeywords("CTYPE1");
	header->RemoveKeywords("CTYPE2");
	header->RemoveKeywords("WCSAXES");
	header->RemoveKeywords("EQUINOX");
	header->RemoveKeywords("LONPOLE");
	header->RemoveKeywords("LATPOLE");
	header->RemoveKeywords("CUNIT1");
	header->RemoveKeywords("CUNIT2");
	header->RemoveKeywords("RADESYS");

	// remove astrometry values
	header->RemoveKeywords("CRPIX1");
	header->RemoveKeywords("CRPIX2");
	header->RemoveKeywords("CRVAL1");
	header->RemoveKeywords("CRVAL2");
	header->RemoveKeywords("CD1_1");
	header->RemoveKeywords("CD1_2");
	header->RemoveKeywords("CD2_1");
	header->RemoveKeywords("CD2_2");
	header->RemoveKeywords("A_ORDER");
	header->RemoveKeywords("B_ORDER");
	header->RemoveKeywords("AP_ORDER");
	header->RemoveKeywords("BP_ORDER");

	// remove distortion polynomials coefficients
	// Data present in the header can be from higher order fit
	// so remove up to 6th order (just in case)

	for (int i=0; i<=6; i++)
	{
		ostringstream sip_name;

		for (int j=0; j<=6; j++)
		{
			sip_name.str(""); sip_name << "A_" << i << "_" << j;
			header->RemoveKeywords((sip_name.str()).c_str());
			sip_name.str(""); sip_name << "B_" << i << "_" << j;
			header->RemoveKeywords((sip_name.str()).c_str());
			sip_name.str(""); sip_name << "AP_" << i << "_" << j;
			header->RemoveKeywords((sip_name.str()).c_str());
			sip_name.str(""); sip_name << "BP_" << i << "_" << j;
			header->RemoveKeywords((sip_name.str()).c_str());
		}
	}


}

// Check if astrometry exists in the header
// Checks if minimal amount of header keys is present
bool GloriaAstrometry::IsAstrometryPresent(GloriaFitsHeader *header)
{
	if(header->IsKeywordSet("CTYPE1") && header->IsKeywordSet("CTYPE2") && header->IsKeywordSet("CRVAL1") && header->IsKeywordSet("CRVAL2") && header->IsKeywordSet("CRPIX1") && header->IsKeywordSet("CRPIX2") && header->IsKeywordSet("CD1_1") && header->IsKeywordSet("CD1_2") && header->IsKeywordSet("CD2_1") && header->IsKeywordSet("CD2_2"))
		return true;
	return false;
}

void GloriaAstrometry::Print(std::ostream& os)
{
	os << "Astrometry:" << endl;
	os << "CRPIX " << crpix[0] << " " << crpix[1] << endl;
	os << "CRVAL " << crval[0] << " " << crval[1] << endl;
	os << "CD\t" << cd[0] << " " << cd[1] << endl;
	os << "\t" << cd[2] << " " << cd[3] << endl;

	for(int i=0; i<=sip_order_a; i++) for(int j=0; (i+j)<=sip_order_a; j++)
	{
		// drop linear terms - that's how it is done in astrometry.net
		if(i+j<2) continue;
		os << "A_" << i << "_" << j << ": " << sip_a[i][j] << endl;
	}
	for(int i=0; i<=sip_order_b; i++) for(int j=0; (i+j)<=sip_order_b; j++)
	{
		// drop linear terms
		if(i+j<2) continue;
		os << "B_" << i << "_" << j << ": " << sip_b[i][j] << endl;
	}
	for(int i=0; i<=sip_order_ap; i++) for(int j=0; (i+j)<=sip_order_ap; j++)
	{
		if(i+j<1) continue;
		os << "AP_" << i << "_" << j << ": " << sip_ap[i][j] << endl;
	}
	for(int i=0; i<=sip_order_bp; i++) for(int j=0; (i+j)<=sip_order_bp; j++)
	{
		if(i+j<1) continue;
		os << "BP_" << i << "_" << j << ": " << sip_bp[i][j] << endl;
	}
}

void GloriaAstrometry::GetRaDecFromXY(double x, double y,  double *ra, double *dec)
{
	double f, g;
	f=g=0;

	// Need to shift pixel coordinates, since transformation is centered in (crpix[0],crpix[1])
	x-=crpix[0];
	y-=crpix[1];

	// Calculate distortion polynomials
	for(int i=0; i<=sip_order_a; i++) 
	  {
          for(int j=0; (i+j)<=sip_order_a; j++)
	    {
		// drop linear terms - that's how it is done in astrometry.net
		if(i+j<2) continue;
		f += sip_a[i][j]*pow(x,i)*pow(y,j);
	    }
	  }

	for(int i=0; i<=sip_order_b; i++) 
	  {
           for(int j=0; (i+j)<=sip_order_b; j++)
	     {
		// drop linear terms
		if(i+j<2) continue;
		g += sip_b[i][j]*pow(x,i)*pow(y,j);
	     }
	  }

	double dra  = cd[0]*(f+x)+cd[1]*(g+y);
	double ddec = cd[2]*(f+x)+cd[3]*(g+y);


	// 20121002: flipped non-diagonal elements (AFZ) - to be confirmed 
	//  	*ra  = cd[0]*(f+x)+cd[1]*(g+y)+crval[0];
	//	*dec = cd[2]*(f+x)+cd[3]*(g+y)+crval[1];
	//
	// Changed back on 20121129 - still testing (?!)

        if(ref_ra != crval[0] ||  ref_dec !=crval[1]) InitWCSVec();

        double rpos[3];
        double rlen=0.;

        dra *= deg2rad;
        ddec *= deg2rad;

        for(int iax=0;iax<3;iax++)
	  {
	  rpos[iax]=wcs_ref_vec[iax] + dra*wcs_ra_vec[iax] + ddec*wcs_dec_vec[iax];
	  rlen+=rpos[iax]*rpos[iax];
          }

        rlen=sqrt(rlen);

        *dec = rad2deg * asin(rpos[2]/rlen);
        *ra =  rad2deg * atan2(rpos[1],rpos[0]);

        if(*ra < 0.0) *ra += 360.;

	return;
}

void GloriaAstrometry::InitWCSVec()
{
  // Return if current vectors can be used
 
  if(ref_ra == crval[0] &&  ref_dec == crval[1]) return;

  // Conversion constants (to multipy with, for given conversion)
  deg2rad=3.14159265358979323846/180.;
  rad2deg=180./3.14159265358979323846;

  // Reference point coordinates in radians

  double ra=deg2rad*crval[0];
  double dec=deg2rad*crval[1];

  // Reference point position

  wcs_ref_vec[0]= cos(dec)*cos(ra);
  wcs_ref_vec[1]= cos(dec)*sin(ra);
  wcs_ref_vec[2]= sin(dec);

  // Unit vector in RA direction from reference point (change per degree!)

  wcs_ra_vec[0]= -sin(ra);
  wcs_ra_vec[1]= cos(ra);
  wcs_ra_vec[2]= 0.;

  // Unit vector in DEC direction from reference point (change per degree!)

  wcs_dec_vec[0] = -sin(dec)*cos(ra);
  wcs_dec_vec[1] = -sin(dec)*sin(ra);
  wcs_dec_vec[2] = cos(dec);

  // Store current values

  ref_ra=crval[0];
  ref_dec=crval[1];

  return;
}

void GloriaAstrometry::GetXYFromRaDec(double ra, double dec, double *x, double *y)
{

	// Calculate inverse of the cd matrix
	double inv_cd[4];
	double d = 1./(cd[0]*cd[3]-cd[1]*cd[2]);
	inv_cd[0] = cd[3]*d;
	inv_cd[1] = -cd[1]*d;
	inv_cd[2] = -cd[2]*d;
	inv_cd[3] = cd[0]*d;

	// Initialize WCS base vectors

        if(ref_ra != crval[0] ||  ref_dec !=crval[1]) InitWCSVec();

        // Vector corresponding to current position
        
        double rpos[3];

	ra*=deg2rad;
	dec*=deg2rad;

        rpos[0]= cos(dec)*cos(ra);
        rpos[1]= cos(dec)*sin(ra);
        rpos[2]= sin(dec);

	// Angular distance betweed given position and reference point

        double cosang=0;
        for(int iax=0;iax<3;iax++)
  	  cosang += wcs_ref_vec[iax] * rpos[iax];
 

        // Calculate projection of the position difference on the tangential plane

        for(int iax=0;iax<3;iax++)
	  rpos[iax] = rpos[iax]/cosang - wcs_ref_vec[iax];
    
	// Calculate distances in RA and DEC from reference point

	double dra=0.;
	double ddec=0.;

        for(int iax=0;iax<3;iax++)
	  {
	    dra += rpos[iax] * wcs_ra_vec[iax];
	    ddec += rpos[iax] * wcs_dec_vec[iax];
	  }

        dra *= rad2deg;
        ddec *= rad2deg;
         
	// Apply local transformation now

	double U = inv_cd[0]*dra+inv_cd[1]*ddec;
	double V = inv_cd[2]*dra+inv_cd[3]*ddec;

	// Calculate distortion polynomials
	double F, G;
	F=G=0;

	for(int i=0; i<=sip_order_ap; i++) 
	  {
          for(int j=0; (i+j)<=sip_order_ap; j++)
	     {
		// drop linear terms - that's how it is done in astrometry.net
		if(i+j<1) continue;
		F += sip_ap[i][j]*pow(U,i)*pow(V,j);
	     }
	  }
	for(int i=0; i<=sip_order_bp; i++) 
	  {
          for(int j=0; (i+j)<=sip_order_bp; j++)
	     {
		// drop linear terms    (??? AFZ  ???  Shuldn't be i+j<1 ???)
		if(i+j<1) continue;
		G += sip_bp[i][j]*pow(U,i)*pow(V,j);
	     }
	  }

	*x = U+F+crpix[0];
	*y = V+G+crpix[1];

	return;
}


GloriaAstrometry::~GloriaAstrometry()
{
	// TODO Auto-generated destructor stub
}

} /* namespace gloria */
