// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaAverageMap.h"
#include <algorithm>
#include <cmath>

namespace gloria{

using namespace std;
  
// Constructor creating map tables for given image

  GloriaAverageMap::GloriaAverageMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int Nvar)
  {
    InitializeMaps(Nbx, Nby, Nvar);

    SetImageSize(fitsImage);
  }


// Constructor creating map tables without specifying the image

GloriaAverageMap::GloriaAverageMap(int Nbx, int Nby, int Nvar)
  {
    InitializeMaps(Nbx, Nby, Nvar);
  }

// Initialize maps

void GloriaAverageMap::InitializeMaps(int Nbx, int Nby, int Nvar)
{
  _nMapX = Nbx;
  _nMapY = Nby;
  _nMapXY = Nbx * Nby;
 
 
  for(int imap=0; imap<_valueMaps.size(); imap++)
      delete [] _valueMaps.at(imap);

  _valueMaps.clear();
  _valueNames.clear();
 
  for(int ivar=0; ivar<Nvar; ivar++)
    {
     double *averMap = new double[_nMapXY];
     double *spreadMap = new double[_nMapXY];

     stringstream averName, spreadName;
     averName << "Average" << ivar ;
     spreadName << "Spread" << ivar ;


    _valueMaps.push_back(averMap);
    _valueMaps.push_back(spreadMap);

    _valueNames.push_back(averName.str());
    _valueNames.push_back(spreadName.str());
    }

  _sizeX = _sizeY =  _sizeXY = 0;
  _binX = _binY = _binXY = 0;
}

// Add new variable

void GloriaAverageMap::AddAverage()
 {
   stringstream averName, spreadName;
   averName << "Average" << _valueMaps.size() ;
   spreadName << "Spread" << _valueMaps.size() ;

   AddMap(averName.str());
   AddMap(spreadName.str());
 }


// Add new variable with a name

void GloriaAverageMap::AddAverage(string name)
 {
   AddMap(name);
   AddMap("S_"+name);
 }

// Set name to the considered variable
void GloriaAverageMap::SetAverageName(int ivar, string name)
{
  _valueNames.at(2*ivar) = name; 
  _valueNames.at(2*ivar+1) = "S_" + name;
}

// Calculate map from input data arrays

long GloriaAverageMap::CalculateMap(int iVar, int nVal, double* Xpos, double* Ypos, double* Val)
  {
  // Wrong input data
  if(nVal <=0  || Xpos==NULL || Ypos==NULL || Val==NULL) return -1;

  // if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return -1;

  // variable index out of range
  if(iVar < 0 || 2*iVar + 2 > _valueMaps.size() )return -1;

  vector<double> Xvec(Xpos,Xpos+nVal);
  vector<double> Yvec(Ypos,Ypos+nVal);
  vector<double> Vvec(Val,Val+nVal);

  long nmin =  CalculateMap(iVar,&Xvec,&Yvec,&Vvec);

  return nmin;
  }


// Calculate fit map from input data arrays

long GloriaAverageMap::CalculateFitMap(int iVar, int nVal, double* Xpos, double* Ypos, double* Val, int Npol)
  {
  // Wrong input data
  if(nVal <=0  || Xpos==NULL || Ypos==NULL || Val==NULL) return -1;

  // if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return -1;

  // variable index out of range
  if(iVar < 0 || 2*iVar + 2 > _valueMaps.size() )return -1;

  if(Npol<0)return -1;

  vector<double> Xvec(Xpos,Xpos+nVal);
  vector<double> Yvec(Ypos,Ypos+nVal);
  vector<double> Vvec(Val,Val+nVal);

  long nmin =  CalculateFitMap(iVar,&Xvec,&Yvec,&Vvec,Npol);

  return nmin;
  }


long GloriaAverageMap::CalculateMap(int iVar, vector<double>* Xpos, vector<double>* Ypos, vector<double>* Val)
{
  // Wrong input data
 if(Xpos->size()<Val->size() || Ypos->size()<Val->size()) return -1;

  // if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return -1;

  // variable index out of range
 if(iVar < 0 || 2*iVar + 2 > _valueMaps.size() )return -1;

 double *averMap = _valueMaps.at(2*iVar);
 double *spreadMap =  _valueMaps.at(2*iVar+1);

 vector<double> *ValVec = new vector<double>[_nMapXY];

 for(int ival=0; ival< Val->size() ; ival++)
   {
     double x = Xpos->at(ival);
     double y = Ypos->at(ival);

     if(x < 0.5 || y < 0.5 || x >= _sizeX+0.5 || y >= _sizeY+0.5 ) continue;

     int ibx = (x-0.5)/_binX;
     int iby = (y-0.5)/_binY;

     int imap = ibx + iby*_nMapX;

     ValVec[imap].push_back(Val->at(ival));
   }


 long nmin = Val->size();

 for(int imap=0; imap< _nMapXY; imap++)
   {

     averMap[imap] = spreadMap[imap] = 0.;

     int nval = ValVec[imap].size();

     if(nval < nmin) nmin = nval;

     if(nval == 0) continue;

     double offset = ValVec[imap].at(0);

     double sumval = 0.;
     double sumval2 = 0.;

     for(int ival=0; ival < ValVec[imap].size() ; ival++)
       {
	double val = ValVec[imap].at(ival) - offset;
	sumval += val/nval;
	sumval2 += val*val/nval;
       }


     averMap[imap]= sumval + offset;

     sumval2 -= sumval*sumval;
     if(sumval2 > 0.)
           sumval2 = sqrt(sumval2); 
     else
       sumval2 = 0.;

     if(nval>1)sumval2*=sqrt((1.*nval)/(nval-1.));

     spreadMap[imap]= sumval2;
   }

 delete [] ValVec;

 return nmin;
}


long GloriaAverageMap::CalculateFitMap(int iVar, vector<double>* Xpos, vector<double>* Ypos, vector<double>* Val, int Npol)
 {
  // Wrong input data
  if(Xpos->size()<Val->size() || Ypos->size()<Val->size()) return -1;

  // if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return -1;

  // variable index out of range
  if(iVar < 0 || 2*iVar + 2 > _valueMaps.size() )return -1;

 double *averMap = _valueMaps.at(2*iVar);
 double *spreadMap =  _valueMaps.at(2*iVar+1);


 // Fill position shift and value vectors for subsequent map bins

 vector<double> *XVec = new vector<double>[_nMapXY];
 vector<double> *YVec = new vector<double>[_nMapXY];
 vector<double> *ValVec = new vector<double>[_nMapXY];

 for(int ival=0; ival< Val->size() ; ival++)
   {
     double x = Xpos->at(ival);
     double y = Ypos->at(ival);

     if(x < 0.5 || y < 0.5 || x >= _sizeX+0.5 || y >= _sizeY+0.5 ) continue;

     int ibx = (x-0.5)/_binX;
     int iby = (y-0.5)/_binY;

     double xbin = (ibx+0.5)*_binX + 0.5;
     double ybin = (iby+0.5)*_binY + 0.5;

     int imap = ibx + iby*_nMapX;

     XVec[imap].push_back(Xpos->at(ival) - xbin);
     YVec[imap].push_back(Ypos->at(ival) - ybin);
     ValVec[imap].push_back(Val->at(ival));
   }


 // Fit parameter distribution inside each bin

 long nmin = Val->size();

 for(int imap=0; imap< _nMapXY; imap++)
   {
     averMap[imap] = spreadMap[imap] = 0.;

     int nval = ValVec[imap].size();

     if(nval < nmin) nmin = nval;

     // Skip if no data
     if(nval == 0) continue;


     int nfit = Npol;
     int npolpar = (nfit+1)*(nfit+2)/2;

     // Reduce fit order if too few measurements

     while(nval < 2*npolpar && nfit > 0)
       {
	 nfit--;
	 npolpar = (nfit+1)*(nfit+2)/2;
       }

     int nfittab = npolpar*npolpar;
     double *mag2sum = new double[nfittab];
     double *magsum = new double[npolpar];
     double *magval = new double[npolpar];
     double val2sum=0.;

     for(int i1=0;i1<npolpar;i1++)
       {
       magsum[i1]=0.;
       for(int i2=0;i2<npolpar;i2++)
          mag2sum[i1+npolpar*i2]=0.;
       }

     for(int ival=0; ival < ValVec[imap].size() ; ival++)
       {
       int ipow=0;
       for(int i1=0;i1<=nfit;i1++)
       	  for(int i2=0;i2<=nfit-i1;i2++)
	    magval[ipow++]= pow( XVec[imap].at(ival),i1) * pow(YVec[imap].at(ival),i2);

       for(int i1=0;i1<npolpar;i1++)
          {
       	  magsum[i1]+=ValVec[imap].at(ival)*magval[i1];
       	  for(int i2=0;i2<npolpar;i2++)
       	       mag2sum[i1+npolpar*i2]+=magval[i1]*magval[i2];
       	  }

       val2sum+=ValVec[imap].at(ival) * ValVec[imap].at(ival);
       }

     //Duplicate sum tables for solving

     double *mag2inv = new double[nfittab];
     double *magpar = new double[npolpar];

     for(int i1=0;i1<npolpar;i1++)
                 magpar[i1] = magsum[i1];

     for(int i2=0;i2<nfittab;i2++)
    	         mag2inv[i2]=mag2sum[i2];

     GaussjSolve(mag2inv,magpar,npolpar);

     averMap[imap]= magpar[0];

     double serror = val2sum;

     for(int i1=0;i1<npolpar;i1++)
	{
	 serror -= 2*magpar[i1]*magsum[i1];

         for(int i2=0;i2<npolpar;i2++)
	    serror += magpar[i1]*magpar[i2]*mag2sum[i1+npolpar*i2];
	 }

     if(serror>0 && nval>npolpar )
         serror = sqrt(serror/(nval-npolpar));
     else
         serror = 0.;

     spreadMap[imap]= serror;


     delete [] mag2inv;
     delete [] magpar;
     delete [] mag2sum;
     delete [] magsum;
     delete [] magval;
   }
    // end of loop over map bins

 delete [] XVec;
 delete [] YVec;
 delete [] ValVec;

 return nmin;
 }

//
// Routine for solving matrix equations
//

int GloriaAverageMap::GaussjSolve(double *alfa,double *beta,int n)
{
  int *ipiv;
  int *indxr;
  int *indxc;
  int i,j,k;
  int irow=0;
  int icol=0;
  double abs,big,help,pivinv;

  ipiv = new int[n];
  indxr = new int[n];
  indxc = new int[n];

  for(i=0;i<n;ipiv[i++]=0);

  for(i=0;i<n;i++)
    {
      big=0.;
      for(j=0;j<n;j++)
	{
	  if(ipiv[j]==1)continue;
	  for(k=0;k<n;k++)
	    {
	      if(ipiv[k]!=0)continue;
	      abs=fabs(alfa[n*j+k]);
	      if(abs>big)
		{
		  big=abs;
		  irow=j;
		  icol=k;
		}
	    }
	}
      ipiv[icol]++;

      if(ipiv[icol]>1)
	return 1;

      if(irow!=icol)
	{
	  help=beta[irow];
	  beta[irow]=beta[icol];
	  beta[icol]=help;
	  for(j=0;j<n;j++)
	    {
	      help=alfa[n*irow+j];
	      alfa[n*irow+j]=alfa[n*icol+j];
	      alfa[n*icol+j]=help;
	    }
	}
      indxr[i]=irow;
      indxc[i]=icol;

      if(alfa[n*icol+icol]==0.)
	return 1;

      help=alfa[n*icol+icol];
      pivinv=1./help;
      alfa[n*icol+icol]=1.;
      for(j=0;j<n;alfa[n*icol+(j++)]*=pivinv);
      beta[icol]*=pivinv;

      for(j=0;j<n;j++)
	{
	  if(j==icol)continue;
	  help=alfa[n*j+icol];
	  alfa[n*j+icol]=0.;
	  for(k=0;k<n;k++)
	    alfa[n*j+k]-=alfa[n*icol+k]*help;
	  beta[j]-=beta[icol]*help;
	}
    }

  for(i=n-1;i>=0;i--)
    {
      if(indxr[i]==indxc[i])continue;
      for(j=0;j<n;j++)
	{
	  help=alfa[n*j+indxr[i]];
	  alfa[n*j+indxr[i]]=alfa[n*j+indxc[i]];
	  alfa[n*j+indxc[i]]=help;
	}
    }

  delete [] ipiv;
  delete [] indxr;
  delete [] indxc;

  return 0;
}


  
}// namespace gloria
