// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaClusterStatistics.h"
#include <cmath>

namespace gloria{

  
// Default constructor

GloriaClusterStatistics::GloriaClusterStatistics()
  {
    Clear();
  }


// Default destructor

GloriaClusterStatistics::~GloriaClusterStatistics()
  {
   return;
  }

// Clear cluster

void GloriaClusterStatistics::Clear()
  {
   _nPixel=0;

   _seedX = _seedY = 0.;

   _signal = 0.;

   _sumX = _sumY = _sumXX = _sumYY = _sumXY = 0.;

   _cogX = _cogY = _rmsX2 = _rmsY2 = _rmsX = _rmsY = _corrXY = 0.;

   _averDone = false;

   return;
  }

// Add pixel to the cluster

void GloriaClusterStatistics::AddPixel(int ix, int iy, double signal)
  {
    if(_nPixel==0)
      {
	_seedX = ix;
	_seedY = iy;
      }
    else
      {
      double dx=ix - _seedX;
      double dy=iy - _seedY;

      _sumX+= dx*signal;
      _sumY+= dy*signal;
      _sumXX+= dx*dx*signal;
      _sumYY+= dy*dy*signal;
      _sumXY+= dx*dy*signal;
      }

    _signal += signal;

    _nPixel++;

    _averDone = false;

    return;
  }

// Add pixel to the cluster

void GloriaClusterStatistics::CalculateAverages()
  {
    if(_averDone) return;

    if(_signal <= 0.) return;

    _cogX = _sumX/_signal;
    _cogY = _sumY/_signal;

    _rmsX2 = _sumXX/_signal - _cogX*_cogX;
    if(_rmsX2 > 0. ) 
      _rmsX = sqrt(_rmsX2);
    else
      _rmsX = _rmsX2 = 0.;

    _rmsY2 = _sumYY/_signal - _cogY*_cogY;
    if(_rmsY2 > 0. ) 
      _rmsY = sqrt(_rmsY2);
    else
      _rmsY = _rmsY2 = 0.;

    if( _rmsX * _rmsY > 0.)
      {
	_corrXY = _sumXY/_signal -  _cogX*_cogY;
	_corrXY /= _rmsX * _rmsY;
      }
    else
      _corrXY = 0.;

    // Can happen for weighted signals (with negative weights!)

    if(_corrXY > 1.0) _corrXY = 1.0;

    if(_corrXY < -1.0) _corrXY = -1.0;

    _averDone = true;

  }

// Get cluster spread (RMS) in X
double GloriaClusterStatistics::GetXspread()
{
  if(!_averDone)CalculateAverages();

  return _rmsX;
}

// Get cluster spread (RMS) in Y
double GloriaClusterStatistics::GetYspread()
{
  if(!_averDone)CalculateAverages();

  return _rmsY;
}


// Get X-Y correlation for the cluster
double GloriaClusterStatistics::GetXYcorr()
{
  if(!_averDone)CalculateAverages();

  return _corrXY;
}

// Get effective cluster width
double GloriaClusterStatistics::GetFWHM()
{
  if(!_averDone)CalculateAverages();

  double sum2 = _rmsX2 + _rmsY2;

  if(sum2 == 0.) return 0.;

  return 2.*sqrt( log(2.) * sum2 );
}


// Get cluster elongation
double GloriaClusterStatistics::GetElongation()
{
  if(!_averDone)CalculateAverages();

  double sum2 = _rmsX2 + _rmsY2;

  if(sum2 == 0.) return 0.;

  double diff2 =  _rmsX2 - _rmsY2;

  double ell = diff2*diff2 + 4*_corrXY*_corrXY*_rmsX2*_rmsY2;

  ell = sqrt(ell)/sum2;

  return ell;
}


// Get cluster orientation (direction of primary axis)
double GloriaClusterStatistics::GetOrientation()
{
  if(!_averDone)CalculateAverages();

  double ang = atan2( 2*_corrXY*_rmsX*_rmsY,_rmsX2-_rmsY2);

  return 90./3.1415926 * ang;
}

// Get cluster length (length of primary axis)
double GloriaClusterStatistics::GetLength()
{
  if(!_averDone)CalculateAverages();

  double sum2 = _rmsX2 + _rmsY2;

  if(sum2 == 0.) return 0.;

  double diff2 =  _rmsX2 - _rmsY2;

  double len = diff2*diff2 + 4*_corrXY*_corrXY*_rmsX2*_rmsY2;

  len = 0.5*(sum2 + sqrt(len));

  if(len <= 0.) return 0.;
  return sqrt(len);
}

// Get cluster width (length of secondary axis)
double GloriaClusterStatistics::GetWidth()
{
  if(!_averDone)CalculateAverages();

  double sum2 = _rmsX2 + _rmsY2;

  if(sum2 == 0.) return 0.;

  double diff2 =  _rmsX2 - _rmsY2;

  double wid = diff2*diff2 + 4*_corrXY*_corrXY*_rmsX2*_rmsY2;

  wid = 0.5*(sum2 - sqrt(wid));

  if(wid <= 0.) return 0.;
  return sqrt(wid);
}


  
}// namespace gloria
