// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaDataContainer.h"

#include <iostream>

//#include <stdlib.h>
//#include <string.h>
//#include <cfitsio/fitsio.h>

using namespace std;
using namespace gloria; 
  
// Default constructor

GloriaDataContainer::GloriaDataContainer()
{
  _nImageCollections=0;
  _nTableCollections=0;
}


// Default destructor

GloriaDataContainer::~GloriaDataContainer()
{
}

bool GloriaDataContainer::ImageCollectionExists(std::string name)
{
  return _imageCollectionPermanent.find(name)!=_imageCollectionPermanent.end();
}

bool GloriaDataContainer::ImageCollectionPermanent(std::string name)
{
  if(_imageCollectionPermanent.find(name)==_imageCollectionPermanent.end())
    return false;
  else
    return _imageCollectionPermanent[name];
}

void GloriaDataContainer::SetImageCollectionPermanent(std::string name, bool isPermanent)
{
  if(_imageCollectionPermanent.find(name)==_imageCollectionPermanent.end())
    return;
  else
    _imageCollectionPermanent[name] = isPermanent;
}

// Add image collection and return it

ImageVec& GloriaDataContainer::CreateImageCollection(std::string name, bool isPermanent)
{
  // Should not happen, but if collection already exists: clear it and
  // use it as if it was created

  ImageVec *imgvec;

  if(ImageCollectionExists(name))
    {
      std::cerr << "Problem in GloriaDataContainer::CreateImageCollection, name already in use: " 
                                                                             << name << std::endl;
      imgvec = &_imageCollectionMap[name];
      imgvec->clear();
      _imageCollectionPermanent[name]=isPermanent;
    }
       
  else
    {
    imgvec = new ImageVec;

    _imageCollectionMap.insert( pair< std::string, ImageVec >(name, *imgvec) );

    // Vector is copied to the map, so we have to delete it now and
    // return pointer to the stored vector
    delete imgvec;
    imgvec = &_imageCollectionMap[name];

    _imageCollectionPermanent.insert( pair< std::string, bool >(name, isPermanent) );
    _imageCollectionNames.push_back(name);
    _nImageCollections++;
    }

  return *imgvec;  
}

// Get existing image collection 

ImageVec& GloriaDataContainer::GetImageCollection(std::string name)
{
  ImageVec *imgvec;

  if(ImageCollectionExists(name))
    {
    imgvec = &_imageCollectionMap[name];
    }  
  else
    {
     std::cerr << "Problem in GloriaDataContainer::GetImageCollection, collection does not exist: " 
                                                                             << name << std::endl;
    imgvec = NULL;
    }

 return *imgvec;  
 
}

// Remove stored image collection

void GloriaDataContainer::RemoveImageCollection(std::string name)   
{

  // Return if collection does not exist

  if(!ImageCollectionExists(name))return;

   ImageVec *imgvec = &_imageCollectionMap[name];
  _imageCollectionMap.erase(name);
   delete imgvec;

  _imageCollectionPermanent.erase(name);

  // Find collection name in the list and delete it

  for(int iCol=0; iCol<_nImageCollections; iCol++)
    if(_imageCollectionNames.at(iCol) == name )
      {
        _imageCollectionNames.erase(_imageCollectionNames.begin()+iCol);
	break;
      }

  _nImageCollections--;

}

bool GloriaDataContainer::TableCollectionExists(std::string name)
{
  return _tableCollectionPermanent.find(name)!=_tableCollectionPermanent.end();
}

bool GloriaDataContainer::TableCollectionPermanent(std::string name)
{
  if(_tableCollectionPermanent.find(name)==_tableCollectionPermanent.end())
    return false;
  else
    return _tableCollectionPermanent[name];
}

void GloriaDataContainer::SetTableCollectionPermanent(std::string name, bool isPermanent)
{
  if(_tableCollectionPermanent.find(name)==_tableCollectionPermanent.end())
    return;
  else
    _tableCollectionPermanent[name] = isPermanent;
}

// Add table collection and return it

TableVec& GloriaDataContainer::CreateTableCollection(std::string name, bool isPermanent)
{
  // Should not happen, but if collection already exists: clear it and
  // use it as if it was created

  TableVec *tabvec;

  if(TableCollectionExists(name))
    {
      std::cerr << "Problem in GloriaDataContainer::CreateTableCollection, name already in use: " 
                                                                             << name << std::endl;
      tabvec = &_tableCollectionMap[name];
      tabvec->clear();
      _tableCollectionPermanent[name]=isPermanent;
    }
       
  else
    {
    tabvec = new TableVec;

    _tableCollectionMap.insert( pair< std::string, TableVec >(name, *tabvec) );

    // Vector is copied to the map, so we have to delete it now and
    // return pointer to the stored vector
    delete tabvec;
    tabvec = &_tableCollectionMap[name];

    _tableCollectionPermanent.insert( pair< std::string, bool >(name, isPermanent) );
    _tableCollectionNames.push_back(name);
    _nTableCollections++;
    }

  return *tabvec;  
}

// Get existing table collection 

TableVec& GloriaDataContainer::GetTableCollection(std::string name)
{
  TableVec *tabvec;

  if(TableCollectionExists(name))
    {
    tabvec = &_tableCollectionMap[name];
    }  
  else
    {
     std::cerr << "Problem in GloriaDataContainer::GetTableCollection, collection does not exist: " 
                                                                             << name << std::endl;
    tabvec = NULL;
    }

 return *tabvec;  
 
}

// Remove stored table collection

void GloriaDataContainer::RemoveTableCollection(std::string name)   
{

  // Return if collection does not exist

  if(!TableCollectionExists(name))return;

  TableVec *tabvec = &_tableCollectionMap[name];
  _tableCollectionMap.erase(name);
  delete tabvec;

  _tableCollectionPermanent.erase(name);

  // Find collection name in the list and delete it

  for(int iCol=0; iCol<_nTableCollections; iCol++)
    if(_tableCollectionNames.at(iCol) == name )
      {
        _tableCollectionNames.erase(_tableCollectionNames.begin()+iCol);
	break;
      }

  _nTableCollections--;

}


// Remove all collections which are not marked as permanent

void GloriaDataContainer::RemoveTemporaryCollections()
 {
   // Search for temporary images
   // Reverse search order to allow erasing vector elements

   for(int iCol=_nImageCollections-1; iCol>=0; iCol--)
     {
       string colName=_imageCollectionNames.at(iCol);

       if(! _imageCollectionPermanent[colName])
	 {
         _imageCollectionMap.erase(colName);
         _imageCollectionPermanent.erase(colName);
         _imageCollectionNames.erase(_imageCollectionNames.begin()+iCol);
         _nImageCollections--;
	 }
      }

   // Search for temporary tables

   for(int iCol=_nTableCollections-1; iCol>=0; iCol--)
     {
       string colName=_tableCollectionNames.at(iCol);

       if(! _tableCollectionPermanent[colName])
	 {
         _tableCollectionMap.erase(colName);
         _tableCollectionPermanent.erase(colName);
         _tableCollectionNames.erase(_tableCollectionNames.begin()+iCol);
         _nTableCollections--;
	 }
      }

 }


// Clear all collections which are not marked as permanent

void GloriaDataContainer::ClearTemporaryCollections()
 {
   // Search for temporary images

   for(int iCol=0; iCol<_nImageCollections;iCol++)
     {
       string colName=_imageCollectionNames.at(iCol);

       if(! _imageCollectionPermanent[colName])
                       _imageCollectionMap[colName].clear();
     }

   // Search for temporary tables

   for(int iCol=0; iCol<_nTableCollections;iCol++)
     {
       string colName=_tableCollectionNames.at(iCol);

       if(! _tableCollectionPermanent[colName])
                        _tableCollectionMap[colName].clear();
     }

 }
    
// Return type of given collection 

GloriaCollectionType GloriaDataContainer::GetCollectionType(std::string name)
{
  if(_imageCollectionPermanent.find(name)!=_imageCollectionPermanent.end())
           return ImageCollection;

  if(_tableCollectionPermanent.find(name)!=_tableCollectionPermanent.end())
           return TableCollection;

  return UndefinedCollection;
}
// Return string describing type of given collection 

std::string  GloriaDataContainer::GetCollectionTypeName(std::string name)
{
  GloriaCollectionType type = GetCollectionType(name);

  string typestring = GloriaCollectionTypeName[type];

  return typestring;
}

