// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaFitsHeader.h"

#include <iostream>
#include <iomanip>
#include <stdlib.h>
#include <string.h>
#include <cfitsio/fitsio.h>

using namespace std;
using namespace gloria; 
  
// Default constructor

GloriaFitsHeader::GloriaFitsHeader()
{
  _inputFileName.clear();
  _outputFileName.clear();

  _fitsHeaderRecords.clear();
  _headerKeywordMap.clear();

  _statusOK=true;
  _errorMessage.clear();
}


// Default destructor

GloriaFitsHeader::~GloriaFitsHeader()
{
}


// Read header from file

void GloriaFitsHeader::ReadFitsHeader(fitsfile *fp)
{
   int status = 0;

   _statusOK=true;
   _errorMessage.clear();

   // Remove all entries - header is overwritten (!)
  _fitsHeaderRecords.clear();
  _headerKeywordMap.clear();

   int nkeys=0;

   fits_get_hdrspace(fp, &nkeys, NULL, &status);

   char keyname[FLEN_KEYWORD+1];
   char keyvalue[FLEN_VALUE+1];
   char comment[FLEN_COMMENT+1];

   if(status)nkeys=0;

   for(int ikey=0; ikey<nkeys; ikey++)
      {
	HeaderRecord rec;

        fits_read_keyn(fp, ikey+1, keyname, keyvalue, comment, &status);

       if (status) break;

        rec.Keyword = keyname;
        rec.Value = keyvalue;
        rec.Comment = comment;

	_fitsHeaderRecords.push_back(rec);

        _headerKeywordMap.insert( pair<string,int>(rec.Keyword,ikey) ); 
     }

   if(status)
     {
       _statusOK=false;
       _errorMessage="Error reading header records from FITS file";
     }

   return;
}

bool GloriaFitsHeader::IsKeywordSet(const char *keyword)
{
   // Check if given keyword was used in a header.
   // Use map for searching

  string key = keyword;

  std::multimap<std::string,int>::iterator itHead = _headerKeywordMap.find(key); 
  return (itHead != _headerKeywordMap.end()); 
}

GloriaFitsHeader::HeaderRecord& GloriaFitsHeader::GetKeywordRecord(const char *keyword)
{
   // Get the value of a given keyword. Return "" if not found.
   // Use map for searching

  string key = keyword;

  std::multimap<std::string,int>::iterator itHead = _headerKeywordMap.find(key); 

  if (itHead == _headerKeywordMap.end()) 
     {
      HeaderRecord *empty = new HeaderRecord;
      empty->Keyword = keyword;
      empty->Value = "";
      empty->Comment = "";
      return *empty;
     } 
 else 
     {
      int ikey=itHead->second;
      return _fitsHeaderRecords[ikey];
    }
}


string& GloriaFitsHeader::GetKeywordValue(const char *keyword)
{
   // Get the value of a given keyword. Return "" if not found.
   // Use map for searching

  string key = keyword;

  std::multimap<std::string,int>::iterator itHead = _headerKeywordMap.find(key); 

  if (itHead == _headerKeywordMap.end()) 
     {
      string *empty = new string("");
      return *empty;
     } 
 else 
     {
      int ikey=itHead->second;
      return _fitsHeaderRecords[ikey].Value;
    }
}


long   GloriaFitsHeader::GetKeywordValueInt(const char *keyword)
{
  string valstr=GetKeywordValue(keyword);

  return atol(valstr.data());
}

double GloriaFitsHeader::GetKeywordValueDouble(const char *keyword)
{
  string valstr=GetKeywordValue(keyword);

  return atof(valstr.data());
}

//============================================================
// Still to be implemented

void GloriaFitsHeader::WriteFitsHeader(fitsfile *fp)
{
  int status = 0;

  // Itearate through and write all fits header entries
  for(std::vector<HeaderRecord>::iterator viter = _fitsHeaderRecords.begin(); viter != _fitsHeaderRecords.end(); viter++)
  {
	  // If comment, write comment
	  if((*viter).Keyword.find("COMMENT")!=string::npos)
		  fits_write_comment(fp, (char*)(*viter).Comment.c_str(), &status);
	  
	  // if history
	  else if((*viter).Keyword.find("HISTORY")!=string::npos)
		  fits_write_history(fp, (char*)(*viter).Comment.c_str(), &status);
	  
	  // if section
	  else if((*viter).Keyword.find("SECTIO")!=string::npos)
	  {
	          string secstring = (*viter).Keyword + string(" = ") + (*viter).Value + (*viter).Comment +
		    string("                                                                                ");
		  
		  char *tmpstring = (char*)(secstring.c_str());
		  tmpstring[80]=0;
			   
	          fits_write_record(fp, tmpstring, &status);
	  }
	  
	  // Try to guess data type
	 
	  // quoted string 
	  else if((*viter).Value.find("\'")!=string::npos)
	  {
		  // Remove apostrophes, as they are added automatically
		  string::size_type k = 0;
		  while((k=(*viter).Value.find("\'",k))!=(*viter).Value.npos)
		  {
			  (*viter).Value.erase(k, 1);
		  }
		  // Write the string
		  char *tmpstring = (char*)(*viter).Value.c_str();
	      fits_update_key(fp, TSTRING, (const char*)(*viter).Keyword.c_str(), tmpstring, (char*)(*viter).Comment.c_str(), &status);
	  }
	  
	  // bool
	  else if((*viter).Value.size()==1 && ((*viter).Value.find("T")!=string::npos || (*viter).Value.find("F")!=string::npos))
	  {
		  // TLOGICAL in fits equals int
		  int bval = (*viter).Value.find("T")!=string::npos ? true : false;
	      fits_update_key(fp, TLOGICAL, (const char*)(*viter).Keyword.c_str(), &bval, (char*)(*viter).Comment.c_str(), &status);
	  }
	  
	  // try float/double, otherwise use string
	  else 
	  {
	    double dval=0.;
	    int ival=0;
	    bool dok = false;
	    bool iok = false;
	    
	    if((*viter).Value.find(".")!=string::npos)
	      {
	      dval = atof((*viter).Value.c_str());
	      dok = dval!=0. || (*viter).Value==string("0.") ||  (*viter).Value==string("0.0");
	      }
	    else
	      {
	      ival = atoi((*viter).Value.c_str());
	      iok =  ival!=0 || (*viter).Value==string("0");
	      }
	    
	    if(dok)
	      fits_update_key(fp, TDOUBLE, (const char*)(*viter).Keyword.c_str(), &dval, (char*)(*viter).Comment.c_str(), &status);
	    else if(iok)
	      fits_update_key(fp, TINT, (const char*)(*viter).Keyword.c_str(), &ival, (char*)(*viter).Comment.c_str(), &status);
	    else
	      fits_update_key(fp, TSTRING, (const char*)(*viter).Keyword.c_str(),(char*)(*viter).Value.c_str(), (char*)(*viter).Comment.c_str(), &status);
	  }

  }
}

void GloriaFitsHeader::AddKeyword(HeaderRecord record)
{	
  _fitsHeaderRecords.push_back(record);
  
  // Update the map for searching in header
  _headerKeywordMap.insert( pair<string,int>(record.Keyword,_fitsHeaderRecords.size()-1) );  
  
}
   
// If keyword is to be string, you need to include it in ''
void GloriaFitsHeader::AddKeyword(const char *keyword, string value, string comment)
{	
  HeaderRecord tmp_record;
  tmp_record.Keyword = keyword;
  tmp_record.Value = value;
  tmp_record.Comment = comment;
  
  _fitsHeaderRecords.push_back(tmp_record);
  
  // Update the map for searching in header
  _headerKeywordMap.insert( pair<string,int>(tmp_record.Keyword,_fitsHeaderRecords.size()-1) );  
  
}

void GloriaFitsHeader::AddKeywordDouble(const char *keyword, double value, string comment)
{
	char tmpstring[128];
	sprintf(tmpstring, "%.10e", value);
	AddKeyword(keyword, tmpstring, comment);
}

void GloriaFitsHeader::AddKeywordInt(const char *keyword, int value, string comment)
{
	char tmpstring[128];
	sprintf(tmpstring, "%d", value);
	AddKeyword(keyword, tmpstring, comment);
}

// Adding comment to fits header
void GloriaFitsHeader::AddComment(string comment)
{	
  HeaderRecord tmp_record;
  tmp_record.Keyword = "COMMENT";
  tmp_record.Value.clear(); // no value
  tmp_record.Comment = comment;
  
  _fitsHeaderRecords.push_back(tmp_record);
  
  // Update the map for searching in header
  _headerKeywordMap.insert( pair<string,int>(tmp_record.Keyword,_fitsHeaderRecords.size()-1) );  
  
}


// Adding comment to fits header
void GloriaFitsHeader::AddHistory(string history)
{	
  HeaderRecord tmp_record;
  tmp_record.Keyword = "HISTORY";
  tmp_record.Value.clear(); // no value
  tmp_record.Comment = history;
  
  _fitsHeaderRecords.push_back(tmp_record);
  
  // Update the map for searching in header
  _headerKeywordMap.insert( pair<string,int>(tmp_record.Keyword,_fitsHeaderRecords.size()-1) );  
  
}


void GloriaFitsHeader::SetKeyword(const char *keyword, string value, string comment)   
{
	// Use map to check if the keyword is already set
        // If yes, change its value, otherwise add new keyword.

	string key = keyword;

	std::multimap<std::string,int>::iterator itHead = _headerKeywordMap.find(key);

	if (itHead != _headerKeywordMap.end())
	  {
		int ikey=itHead->second;
		_fitsHeaderRecords[ikey].Value=value;
		_fitsHeaderRecords[ikey].Comment=comment;
	  }
	else
	  AddKeyword(keyword,value,comment);

}

void GloriaFitsHeader::SetKeywordDouble(const char *keyword, double value, string comment)   
{
	char tmpstring[128];
	sprintf(tmpstring, "%.10e", value);
	SetKeyword(keyword, tmpstring, comment);
}

void GloriaFitsHeader::SetKeywordInt(const char *keyword, int value, string comment)   
{
	char tmpstring[128];
	sprintf(tmpstring, "%d", value);
	SetKeyword(keyword, tmpstring, comment);
}
 
void GloriaFitsHeader::RemoveKeyword(const char *keyword)
{
	string key = keyword;

	// Search for the keyword
	std::multimap<std::string,int>::iterator itHead = _headerKeywordMap.find(key);
	
	if (itHead == _headerKeywordMap.end())return;
	
	// If found, delete the keyword entry

	int ikey=itHead->second;
	_fitsHeaderRecords.erase(_fitsHeaderRecords.begin()+ikey);

        _headerKeywordMap.erase(itHead);
         
	// And change indexes to other keywords in the map

        for(itHead=_headerKeywordMap.begin();itHead !=_headerKeywordMap.end();itHead++)
          if(itHead->second > ikey) (itHead->second)--;
	
}

void GloriaFitsHeader::RemoveKeywords(const char *keyword)
{
	string key = keyword;

        int nkey = _fitsHeaderRecords.size();

        int ikey=0; 

	while(ikey < _fitsHeaderRecords.size())
	  {
	    if(_fitsHeaderRecords[ikey].Keyword == keyword)
                     _fitsHeaderRecords.erase(_fitsHeaderRecords.begin()+ikey);
            else
              ikey++; 
	  }

	// Return if no records removed

        if(_fitsHeaderRecords.size() == nkey) return;

	// Recreate keyword map

	_headerKeywordMap.clear();
        for(int ikey=0; ikey < _fitsHeaderRecords.size(); ikey++)
	  _headerKeywordMap.insert( pair<string,int>(_fitsHeaderRecords[ikey].Keyword,ikey));

}


void GloriaFitsHeader::CopyHeader(GloriaFitsHeader* source)
{
  // copy header record vector

  _fitsHeaderRecords = source->_fitsHeaderRecords;


  // Copy keyword map

  _headerKeywordMap = source->_headerKeywordMap;

}





void GloriaFitsHeader::PrintHeader( std::ostream& os, int maxRec)
{
  int nRec = _fitsHeaderRecords.size();

  if( nRec == 0 )
    { 
      os << "  Fits header is empty  " << endl;
    }

  
  if( nRec > maxRec && maxRec > 0 )
    { 
      os << "  Fits header truncated to " << maxRec << " lines" << endl;
      nRec = maxRec;
    }

  for(int iRec=0; iRec<nRec ; iRec++)
    {
      string keyword = _fitsHeaderRecords[iRec].Keyword;

      os << setw(8) << keyword << " ";

      if( keyword != "COMMENT" && keyword != "HISTORY")
	{ 
	  os << "= " << setw(20) << _fitsHeaderRecords[iRec].Value;
	  os << " / " <<  _fitsHeaderRecords[iRec].Comment;
	}
      else
         os << _fitsHeaderRecords[iRec].Value << "; " <<  _fitsHeaderRecords[iRec].Comment;

      os << endl;

    }

}

