// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaFitsImage.h"

#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <cfitsio/fitsio.h>

using namespace std;
using namespace gloria; 
  

void GloriaFitsImage::InitializeEmptyImage()
{
  _sizeX=0;
  _sizeY=0;

  _bitsPerPixel=0;

  _isIntImage=false;
  _intPixels=NULL;

  _isFloatImage=false;
  _floatPixels=NULL;

  _fitsHeaderRecords.clear();
  _headerKeywordMap.clear();

  _inputFileName.clear();
  _outputFileName.clear();

  _statusOK=true;
  _errorMessage.clear();
  
  return;
}

void GloriaFitsImage::ReleaseMemory()
{
  if(_intPixels) delete [] _intPixels;

  _isIntImage=false;
  _intPixels=NULL;

  if(_floatPixels) delete [] _floatPixels;

  _isFloatImage=false;
  _floatPixels=NULL;

  _sizeX=0;
  _sizeY=0;

  _bitsPerPixel=0;

  _statusOK=true;
  _errorMessage.clear();

  return;
}

void GloriaFitsImage::AllocateMemory(int NpixelX, int NpixelY, bool useFloatPixels)
{
  _sizeX=NpixelX;
  _sizeY=NpixelY;

  int sizeXY=_sizeX*_sizeY;

  _statusOK=true;
  _errorMessage.clear();

  if(useFloatPixels)
    {
    _floatPixels=new double[sizeXY];
     
    if(_floatPixels)
      {
      _isFloatImage=true;
      _bitsPerPixel=-64;
      }
    else
      {
      _statusOK=false;
      _errorMessage="Error allocating memory for GloriaFitsImage";
      }
    }
  else
    {
    _intPixels=new long[sizeXY];
     
    if(_intPixels)
      {
      _isIntImage=true;
      _bitsPerPixel=32;
      }
    else
      {
      _statusOK=false;
      _errorMessage="Error allocating memory for GloriaFitsImage";
      }
    }

  return;
}

// Default constructor: zero size image

GloriaFitsImage::GloriaFitsImage()
{
  InitializeEmptyImage();
}


// Default destructor

GloriaFitsImage::~GloriaFitsImage()
{
  ReleaseMemory();
}


// Empty image constructor (does memory allocation)

GloriaFitsImage::GloriaFitsImage(int NpixelX, int NpixelY, bool useFloatPixels)
{
  InitializeEmptyImage();

  AllocateMemory( NpixelX, NpixelY, useFloatPixels);
}



// Copy constructor 

GloriaFitsImage::GloriaFitsImage(const GloriaFitsImage& source)
{
  InitializeEmptyImage();

  if( !source._statusOK)
    {
      _statusOK=false;
      _errorMessage="Source image not OK";
      return;
    }

  if(source._isIntImage)
    {
      AllocateMemory(source._sizeX, source._sizeY, false);

      memcpy(_intPixels, source._intPixels, _sizeX*_sizeY*sizeof(long));
    }

  if(source._isFloatImage)
    {
      AllocateMemory(source._sizeX, source._sizeY, true);

      memcpy(_floatPixels, source._floatPixels, _sizeX*_sizeY*sizeof(double));
    }

  _fitsHeaderRecords=source._fitsHeaderRecords;
  _headerKeywordMap=source._headerKeywordMap;

  _inputFileName=source._inputFileName;
  _outputFileName=source._outputFileName;

  _statusOK=true;
  _errorMessage.clear();

}



// File constructor 

GloriaFitsImage::GloriaFitsImage(std::string FitsFileName)
{
  InitializeEmptyImage();

  ReadFitsFile(FitsFileName);
}

// Read image from file

void GloriaFitsImage::ReadFitsFile(std::string FitsFileName)
{
   fitsfile *fp=0;
   int status = 0;


   _statusOK=true;
   _errorMessage.clear();

   // Open fits file: look for image (!)

   fits_open_image(&fp, FitsFileName.data(), READONLY, &status);

   if (status)
     {
       _statusOK=false;
       _errorMessage="Error opening fits image file";
       return;
     }

    //Read image parameters
    int bitpix=0;
    int naxis=0;
    long axsizes[2];

    fits_get_img_param(fp, 2, &bitpix, &naxis, axsizes, &status);

    if(naxis!=2)
     {
       _statusOK=false;
       _errorMessage="Not supported FITS image structure";

       fits_close_file(fp, &status);

       return;

     }

    // Reallocate memory if size changed or different image type

    bool floatImage =  bitpix < 0;


    if(  _sizeX!=axsizes[0]  || _sizeY!=axsizes[1]
	 || ( _isIntImage && floatImage )
	 || ( _isFloatImage && !floatImage ) )
      { 
	ReleaseMemory();

        AllocateMemory( axsizes[0], axsizes[1], floatImage );
      }

    if(Error())
      {
       fits_close_file(fp, &status);
       return;
      }

    _bitsPerPixel=bitpix;

    // Read image data contents

    int sizeXY=_sizeX*_sizeY;

    long firstpixel[2] = {1, 1};

    if(floatImage)
        fits_read_pix(fp, TDOUBLE, firstpixel, sizeXY,
                     NULL, _floatPixels, NULL, &status);
    else
        fits_read_pix(fp, TLONG, firstpixel, sizeXY,
                     NULL, _intPixels, NULL, &status);

    if(status)
      {
	ReleaseMemory();

       _statusOK=false;
       _errorMessage="Error reading pixel values from FITS image";

       fits_close_file(fp, &status);

       return;
      }



  _inputFileName=FitsFileName;

   // Read header: use method of GloriaFitsHeader

   ReadFitsHeader(fp);

 
   fits_close_file(fp, &status);
   return;
}



// Convert image pixels to floating point numbers

void GloriaFitsImage::ConvertToFloat()
{
  if(_isIntImage)
    {
    AllocateMemory( _sizeX, _sizeY, true);

    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
         _floatPixels[ipix] = (double) _intPixels[ipix];

    delete [] _intPixels;
    _isIntImage=false;
    _intPixels=NULL;
    }

  return;
}

// Convert image pixels to integer numbers

void GloriaFitsImage::ConvertToInt()
{
  if(_isFloatImage)
    {
    AllocateMemory( _sizeX, _sizeY, false);

    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _intPixels[ipix] = (long)( _floatPixels[ipix] + (_floatPixels[ipix]>0.)?0.5:-0.5 );

    delete [] _floatPixels;
    _isFloatImage=false;
    _floatPixels=NULL;
    }

  return;
}


// Shift values of all pixels by given value

GloriaFitsImage& GloriaFitsImage::operator+=(const long shift)
{
if(_isIntImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
                   _intPixels[ipix] += shift;
    }


if(_isFloatImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] += (double) shift;
    }

 return *this;
}

// Shift values of all pixels by given value

GloriaFitsImage& GloriaFitsImage::operator+=(const double shift)
{
if(_isIntImage)ConvertToFloat();


if(_isFloatImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] += shift;
    }

 return *this;
}

// Scale values of all pixels by given value

GloriaFitsImage& GloriaFitsImage::operator*=(const long fac)
{
if(_isIntImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
                   _intPixels[ipix] *= fac;
    }


if(_isFloatImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] *= (double) fac;
    }

 return *this;
}

// Scale values of all pixels by given value

GloriaFitsImage& GloriaFitsImage::operator*=(const double fac)
{
if(_isIntImage)ConvertToFloat();


if(_isFloatImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] *= fac;
    }

 return *this;
}

// Add an image

GloriaFitsImage& GloriaFitsImage::operator+=(const GloriaFitsImage& img)
{
_statusOK=true;
_errorMessage.clear();

if (img._sizeX != _sizeX  || img._sizeY != _sizeY )
     {
       _statusOK=false;
       _errorMessage="Can not add images of different size";
       return *this;
     }

if(img._isFloatImage && _isIntImage) ConvertToFloat();


if(_isIntImage && img._isIntImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _intPixels[ipix] += img._intPixels[ipix];
    }

if(_isFloatImage && img._isFloatImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] += img._floatPixels[ipix];
    }

if(_isFloatImage && img._isIntImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] += (double) img._intPixels[ipix];
    }

 return *this;
}


// Subtract an image

GloriaFitsImage& GloriaFitsImage::operator-=(const GloriaFitsImage& img)
{
_statusOK=true;
_errorMessage.clear();

if (img._sizeX != _sizeX  || img._sizeY != _sizeY )
     {
       _statusOK=false;
       _errorMessage="Can not subtract images of different size";
       return *this;
     }

if(img._isFloatImage && _isIntImage) ConvertToFloat();


if(_isIntImage && img._isIntImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _intPixels[ipix] -= img._intPixels[ipix];
    }

if(_isFloatImage && img._isFloatImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] -= img._floatPixels[ipix];
    }

if(_isFloatImage && img._isIntImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] -= (double) img._intPixels[ipix];
    }

 return *this;
}

// Multiply by an image

GloriaFitsImage& GloriaFitsImage::operator*=(const GloriaFitsImage& img)
{
_statusOK=true;
_errorMessage.clear();

if (img._sizeX != _sizeX  || img._sizeY != _sizeY )
     {
       _statusOK=false;
       _errorMessage="Can not multiply images of different size";
       return *this;
     }

if(img._isFloatImage && _isIntImage) ConvertToFloat();


if(_isIntImage && img._isIntImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _intPixels[ipix] *= img._intPixels[ipix];
    }

if(_isFloatImage && img._isFloatImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] *= img._floatPixels[ipix];
    }

if(_isFloatImage && img._isIntImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      _floatPixels[ipix] *= (double) img._intPixels[ipix];
    }

 return *this;
}

// Divide by an image

GloriaFitsImage& GloriaFitsImage::operator/=(const GloriaFitsImage& img)
{
_statusOK=true;
_errorMessage.clear();

if (img._sizeX != _sizeX  || img._sizeY != _sizeY )
     {
       _statusOK=false;
       _errorMessage="Can not divide images of different size";
       return *this;
     }

if(img._isFloatImage && _isIntImage) ConvertToFloat();


if(_isIntImage && img._isIntImage)
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      if(img._intPixels[ipix]!=0)
        _intPixels[ipix] /= img._intPixels[ipix];
      else
	_intPixels[ipix] = 0;
    }

if(_isFloatImage && img._isFloatImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      if(img._floatPixels[ipix]!=0)
        _floatPixels[ipix] /= img._floatPixels[ipix];
      else
	_floatPixels[ipix]=0.;
    }

if(_isFloatImage && img._isIntImage )
    {
    for(long ipix=0; ipix<_sizeX*_sizeY; ipix++)
      if(img._intPixels[ipix]!=0)
         _floatPixels[ipix] /= (double) img._intPixels[ipix];
      else
 	_floatPixels[ipix]=0.;
    }

 return *this;
}


// Write table to FITS file

void GloriaFitsImage::WriteFitsFile(string FitsFileName, int cBitPix, int cBZero, double cBScale)
{
  long long sizeXY=_sizeX*_sizeY;
  bool newtable=false;

  if(cBitPix==0)
	  cBitPix = _isIntImage ? 32 : -64;

  fitsfile *fp=0;
  int status = 0;

  // Create FITS file
  fits_create_file(&fp, FitsFileName.data(), &status);

  if (status)
  {
      _statusOK=false;
      _errorMessage="Error opening fits image file for writing";
      return;
  }
  
  // Explicitly create new image, to support compression 
  // Note:
  // Datatype conversion/scaling is not supported 
  //     when writing to compressed images
  //
  // COMPRESSION still not available for output files!

  int naxis = 2;
  long naxes[2];
       naxes[0] =  _sizeX;
       naxes[1] =  _sizeY;

  fits_create_img(fp, cBitPix, naxis, naxes, &status);

  if (status)
  {
      _statusOK=false;
      _errorMessage="Error creating new image inside fits file";

      status=0;
      fits_close_file(fp, &status);

      return;
  }
  

  // Set transformation parameters (header not written yet)

  fits_set_bscale(fp, cBScale, cBZero, &status);

  if(status)
  {
     _statusOK=false;
     _errorMessage="Problem setting BSCALE and BZERO parameters";

     status=0;
     fits_close_file(fp, &status);

     return;
  }
  

  // Write the image to file, depending whhether it is int or float

  long firstpixel[2] = {1, 1};

  if(_isIntImage)
       	  fits_write_pix(fp, TLONG, firstpixel, sizeXY, _intPixels, &status);
  else
	  fits_write_pix(fp, TDOUBLE, firstpixel, sizeXY, _floatPixels, &status);

  if(status)
  {
     _statusOK=false;

      char tmpchar[16];
      sprintf(tmpchar, "%d", status);
      _errorMessage="Problem writing pixels to a FITS file, status = ";
      _errorMessage+= tmpchar;

      // Overflow in pixel value conversion results in error,
      // which is in most cases not fatal.  Ignore this error and
      // try to  write header information and close file properly 

      if( status != NUM_OVERFLOW )
	{
        status=0;
        fits_close_file(fp, &status);
        return;
	}

     _statusOK=true;
     status=0;
  }

  // Remove keywords which should be set by fitsio

  // Standard format

  RemoveKeyword("SIMPLE");
  RemoveKeyword("BITPIX");
  RemoveKeyword("NAXIS");
  RemoveKeyword("NAXIS1");
  RemoveKeyword("NAXIS2");
  RemoveKeyword("EXTEND");

  // This will be set by direct call to fitsio

  RemoveKeyword("BZERO");
  RemoveKeyword("BSCALE");

  // Keywords which could be read for compressed input file

  RemoveKeyword("XTENSION");
  RemoveKeyword("PCOUNT");
  RemoveKeyword("GCOUNT");
  RemoveKeyword("TFIELDS");
  RemoveKeyword("TTYPE1");
  RemoveKeyword("TFORM1");
  RemoveKeyword("ZIMAGE");
  RemoveKeyword("ZBITPIX");
  RemoveKeyword("ZNAXIS");
  RemoveKeyword("ZNAXIS1");
  RemoveKeyword("ZNAXIS2");
  RemoveKeyword("ZTILE1");
  RemoveKeyword("ZTILE2");
  RemoveKeyword("ZCMPTYPE");
  RemoveKeyword("ZNAME1");
  RemoveKeyword("ZVAL1");
  RemoveKeyword("ZNAME2");
  RemoveKeyword("ZVAL2");

  // Update header and with parameters defining output fits format 

  fits_update_key(fp, TINT,"BZERO", &cBZero,"value offset",&status);
  fits_update_key(fp, TDOUBLE,"BSCALE", &cBScale,"scaling factor",&status);

  // Write header: use method of GloriaFitsHeader
  //
  // Includes calls to cfitsio routines:
  // fits_write_comment, fits_write_history, fits_write_record and fits_update_key

  WriteFitsHeader(fp);


  // Done

  fits_close_file(fp, &status);
  _outputFileName=FitsFileName;

}

