// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaFitsTable.h"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <string.h>
#include <cfitsio/fitsio.h>

using namespace std;
  
namespace gloria{

void GloriaFitsTable::InitializeEmptyTable()
{
  _inputFileName.clear();
  _outputFileName.clear();

  _nRow=0;
  _nColumn=0;

  _columns.clear();

  _nameMap.clear();

  _nSorted=0;
  _sortName.clear();
  _sortColumn=0;

  _fitsHeaderRecords.clear();
  _headerKeywordMap.clear();

  _statusOK=true;
  _errorMessage.clear();
  
  return;
}

// Default constructor

GloriaFitsTable::GloriaFitsTable()
{
  InitializeEmptyTable();
}


// Default destructor

GloriaFitsTable::~GloriaFitsTable()
{
  // Need to delete all columns (only pointers are stored)

  for(int iCol = 0; iCol!=_columns.size(); iCol++)
    delete _columns.at(iCol);
}


// Copy constructor 

GloriaFitsTable::GloriaFitsTable(const GloriaFitsTable& source)
{
  InitializeEmptyTable();

  if( !source._statusOK)
    {
      _statusOK=false;
      _errorMessage="Source table not OK";
      return;
    }


  _nRow=source._nRow;
  _nColumn=source._nColumn;

  _nameMap=source._nameMap;

  // It is not sufficient to copy column pointers
  // We need to copy columns!

  for(int iCol = 0; iCol!=source._columns.size(); iCol++)
    {
      GloriaTableColumn* copy = source._columns.at(iCol)->Clone();

      _columns.push_back(copy);
    }


  _nSorted=source._nSorted;
  _sortName=source._sortName;
  _sortColumn=source._sortColumn;

  _fitsHeaderRecords=source._fitsHeaderRecords;
  _headerKeywordMap=source._headerKeywordMap;

  _inputFileName=source._inputFileName;
  _outputFileName=source._outputFileName;

  _statusOK=true;
  _errorMessage.clear();

}

// Assignment operator

GloriaFitsTable& GloriaFitsTable::operator= (const GloriaFitsTable& source)
{
 if( this != &source )
   {
  _nRow=source._nRow;
  _nColumn=source._nColumn;
  _nameMap=source._nameMap;

  for(int iCol = 0; iCol!=_columns.size(); iCol++)
    delete _columns.at(iCol);

  _columns.clear();

  for(int iCol = 0; iCol!=source._columns.size(); iCol++)
    {
      GloriaTableColumn* copy = source._columns.at(iCol)->Clone();

      _columns.push_back(copy);
    }


  _nSorted=source._nSorted;
  _sortName=source._sortName;
  _sortColumn=source._sortColumn;

  _fitsHeaderRecords=source._fitsHeaderRecords;
  _headerKeywordMap=source._headerKeywordMap;

  _inputFileName=source._inputFileName;
  _outputFileName=source._outputFileName;

   _statusOK=source._statusOK;
   _errorMessage=source._errorMessage;
  }

 return *this;
}

// Copy table structure and header only

GloriaFitsTable* GloriaFitsTable::EmptyClone()
{
  GloriaFitsTable* clone = new GloriaFitsTable(); 


  clone->_nRow = 0;
  clone->_nColumn = _nColumn;
  clone->_nameMap = _nameMap;

  // It is not sufficient to copy column pointers
  // We need to copy columns!

  for(int iCol = 0; iCol!=_columns.size(); iCol++)
    {
      GloriaTableColumn* copy = _columns.at(iCol)->EmptyClone();

      clone->_columns.push_back(copy);
    }

  clone->_fitsHeaderRecords = _fitsHeaderRecords;
  clone->_headerKeywordMap = _headerKeywordMap;

  clone->_inputFileName = _inputFileName;
  clone->_outputFileName = _outputFileName;

  clone->_statusOK = true;
  clone->_errorMessage.clear();

  return clone;
}


// Adding row to the table

void GloriaFitsTable::AddRow()
{
  for(int iCol = 0; iCol!=_columns.size(); iCol++)
               _columns.at(iCol)->AddRow();

  _nRow++;
}

// Add column of given type to the table

void GloriaFitsTable::AddColumn(string name, TableColumnType type, string units, string comment)
{
  ClearError();

  if(ColumnExists(name))
    {
     _statusOK=false;
     _errorMessage="String column " + name + " not created - name already in use";

     return;
    } 

  GloriaTableColumn* newcolumn;
  LongVec emptyLongVec;
  DoubleVec emptyDoubleVec;

  switch ( type )
    {
        case StringColumn:
           newcolumn = new GloriaTableColumnOf<string>(name, string(""), units, comment);
           break; 

        case LongColumn:
           newcolumn = new GloriaTableColumnOf<long>(name, long(0), units, comment);
           break; 

        case LongVectorColumn:
           newcolumn = new GloriaTableColumnOf<LongVec>(name, emptyLongVec, units, comment);
           break;

        case DoubleColumn:
           newcolumn = new GloriaTableColumnOf<double>(name, double(0.), units, comment);
           break;

        case DoubleVectorColumn:
           newcolumn = new GloriaTableColumnOf<DoubleVec>(name, emptyDoubleVec, units, comment);
           break;

        default:
          _statusOK=false;
          _errorMessage="Wrong column type - column not created";
          return;
	}

 _columns.push_back(newcolumn);

 _nameMap.insert( pair<string, int>(name,_nColumn));

 _nColumn++;

 AdjustColumns();
}

// Insert column of given type to the table, giving it new name

void GloriaFitsTable::InsertColumn(string name, GloriaTableColumn* source)
{
  source->Name() = name;

  InsertColumn(source);
}

// Insert column of given type to the table
  
void GloriaFitsTable::InsertColumn(GloriaTableColumn* source)
{
  ClearError();

  string name = source->Name();

  if(ColumnExists(name))
    {
     _statusOK=false;
     _errorMessage="String column " + name + " not created - name already in use";

     return;
    } 

 _columns.push_back(source);

 _nameMap.insert( pair<string, int>(name,_nColumn));

 _nColumn++;


 AdjustColumns();
}

// Add column of given type to the table

template <class T>
void GloriaFitsTable::AddColumn(string name,  T empty, string units, string comment)
{
  ClearError();

  if(GetValueType(empty)==UndefinedColumn)
    {
     _statusOK=false;
     _errorMessage="Element type does not match defined column types - column not created";

     return;
    } 


  GloriaTableColumn *newcolumn = new GloriaTableColumnOf<T> (name,empty,units,comment);

 _columns.push_back(newcolumn);

 _nameMap.insert( pair<string, int>(name,_nColumn));

 _nColumn++;

 AdjustColumns();
}

// Needed to avoid linker errors
// see: http://www.parashift.com/c++-faq-lite/templates.html#faq-35.13

template void GloriaFitsTable::AddColumn<long>(string name, long empty, string units, string comment);
template void GloriaFitsTable::AddColumn<double>(string name, double empty, string units, string comment);
template void GloriaFitsTable::AddColumn<string>(string name, string empty, string units, string comment);
template void GloriaFitsTable::AddColumn<LongVec>(string name, LongVec empty, string units, string comment);
template void GloriaFitsTable::AddColumn<DoubleVec>(string name, DoubleVec empty, string units, string comment);

// Set value of given cell
template <class T>
void GloriaFitsTable::SetValue(string colName, int iRow, T val)
{
ClearError();

if(!ColumnExists(colName))
    {
    _statusOK=false;
    _errorMessage="Column " + colName + " does not exist";

     return;
    } 

 SetValue(_nameMap.find(colName)->second, iRow, val);

}

// Set value of given cell

template <class T>
void GloriaFitsTable::SetValue(int iCol, int iRow, T val)
{
if(iCol<0 || iCol>=_nColumn)
  {
  _statusOK=false;
  _errorMessage="Column index out of range";

  return;
  } 

if(iRow<0 || iRow>=_nRow)
    {
    _statusOK=false;
    _errorMessage="Row index out of range";

     return;
    } 

 if( _columns.at(iCol)->Type() != GetValueType(val) )
    {
    _statusOK=false;
    _errorMessage="Value type does not match column type";

     return;
    } 

 ((GloriaTableColumnOf<T> *)_columns.at(iCol))->At(iRow) = val;

}

// Needed to avoid linker errors
// see: http://www.parashift.com/c++-faq-lite/templates.html#faq-35.13

template void  GloriaFitsTable::SetValue<long>(int iCol, int iRow, long val);
template void  GloriaFitsTable::SetValue<double>(int iCol, int iRow, double val);
template void  GloriaFitsTable::SetValue<string>(int iCol, int iRow, string val);
template void  GloriaFitsTable::SetValue<LongVec>(int iCol, int iRow, LongVec val);
template void  GloriaFitsTable::SetValue<DoubleVec>(int iCol, int iRow, DoubleVec val);

template void  GloriaFitsTable::SetValue<long>(string colName, int iRow, long val);
template void  GloriaFitsTable::SetValue<double>(string colName, int iRow, double val);
template void  GloriaFitsTable::SetValue<string>(string colName, int iRow, string val);
template void  GloriaFitsTable::SetValue<LongVec>(string colName, int iRow, LongVec val);
template void  GloriaFitsTable::SetValue<DoubleVec>(string colName, int iRow, DoubleVec val);

// Set value of last cell
template <class T>
void GloriaFitsTable::SetLast(string colName, T val)
{
ClearError();

if(!ColumnExists(colName))
    {
    _statusOK=false;
    _errorMessage="Column " + colName + " does not exist";

     return;
    } 

  SetLast(_nameMap.find(colName)->second, val);
}

// Set value of last cell
template <class T>
void GloriaFitsTable::SetLast(int iCol, T val)
{
ClearError();

 if(iCol<0 || iCol>=_nColumn)
    {
    _statusOK=false;
    _errorMessage="Column index out of range";

     return;
    } 

if(_nRow == 0)
    {
    _statusOK=false;
    _errorMessage="Table is empty";

     return;
    } 

 if( _columns.at(iCol)->Type() != GetValueType(val) )
    {
    _statusOK=false;
    _errorMessage="Value type does not match column type";

     return;
    } 

 ((GloriaTableColumnOf<T> *)_columns.at(iCol))->Data().back() = val;

}

// Needed to avoid linker errors
// see: http://www.parashift.com/c++-faq-lite/templates.html#faq-35.13

template void  GloriaFitsTable::SetLast<long>(int iCol,  long val);
template void  GloriaFitsTable::SetLast<double>(int iCol,  double val);
template void  GloriaFitsTable::SetLast<string>(int iCol,  string val);
template void  GloriaFitsTable::SetLast<LongVec>(int iCol,  LongVec val);
template void  GloriaFitsTable::SetLast<DoubleVec>(int iCol,  DoubleVec val);

template void  GloriaFitsTable::SetLast<long>(string colName,  long val);
template void  GloriaFitsTable::SetLast<double>(string colName,  double val);
template void  GloriaFitsTable::SetLast<string>(string colName,  string val);
template void  GloriaFitsTable::SetLast<LongVec>(string colName,  LongVec val);
template void  GloriaFitsTable::SetLast<DoubleVec>(string colName,  DoubleVec val);


// Get value of given cell
template <class T>
void GloriaFitsTable::GetValue(string colName, int iRow, T* val)
{
ClearError();

if(!ColumnExists(colName))
    {
    _statusOK=false;
    _errorMessage="Column " + colName + " does not exist";

     return;
    } 

 GetValue(_nameMap.find(colName)->second, iRow, val);

}

// Set value of given cell

template <class T>
void GloriaFitsTable::GetValue(int iCol, int iRow, T* val)
{
if(iCol<0 || iCol>=_nColumn)
  {
  _statusOK=false;
  _errorMessage="Column index out of range";

  return;
  } 

if(iRow<0 || iRow>=_nRow)
    {
    _statusOK=false;
    _errorMessage="Row index out of range";

     return;
    } 

 if( _columns.at(iCol)->Type() != GetValueType(*val) )
    {
    _statusOK=false;
    _errorMessage="Value type does not match column type";

     return;
    } 

*val = ((GloriaTableColumnOf<T> *)_columns.at(iCol))->At(iRow);

}

// Needed to avoid linker errors
// see: http://www.parashift.com/c++-faq-lite/templates.html#faq-35.13

template void  GloriaFitsTable::GetValue<long>(int iCol, int iRow, long* val);
template void  GloriaFitsTable::GetValue<double>(int iCol, int iRow, double* val);
template void  GloriaFitsTable::GetValue<string>(int iCol, int iRow, string* val);
template void  GloriaFitsTable::GetValue<LongVec>(int iCol, int iRow, LongVec* val);
template void  GloriaFitsTable::GetValue<DoubleVec>(int iCol, int iRow, DoubleVec* val);

template void  GloriaFitsTable::GetValue<long>(string colName, int iRow, long* val);
template void  GloriaFitsTable::GetValue<double>(string colName, int iRow, double* val);
template void  GloriaFitsTable::GetValue<string>(string colName, int iRow, string* val);
template void  GloriaFitsTable::GetValue<LongVec>(string colName, int iRow, LongVec* val);
template void  GloriaFitsTable::GetValue<DoubleVec>(string colName, int iRow, DoubleVec* val);

// Copy row elements from other table (assuming types match!)
void GloriaFitsTable::CopyTableRow(GloriaFitsTable* source, int isource, int itarget)
{
  // Just to be sure

  if(source == NULL)return;

  if(isource < 0 || isource >= source->GetRowNumber() )return;

  int idest = itarget;

  if(idest < 0 || idest >= _nRow) 
    {
      idest = _nRow;
      AddRow();
    }

  int nCol =_columns.size();

  // Check if the source table is not smaller !
  if(nCol>source->GetColumnNumber()) nCol=source->GetColumnNumber();

  for(int iCol = 0; iCol<nCol; iCol++)
      _columns.at(iCol)->CopyCellValue(source->GetColumn(iCol),isource,idest);

}





// Checks sizes of columns and adjust to the longest one if needed 
void GloriaFitsTable::AdjustColumns()
{
  int iMax=0;

  for(int iCol = 0; iCol!=_columns.size(); iCol++)
    if(_columns.at(iCol)->Size() > iMax) iMax = _columns.at(iCol)->Size();


  _nRow=iMax;


  for(int iCol = 0; iCol!=_columns.size(); iCol++)
    if(_columns.at(iCol)->Size() != iMax) _columns.at(iCol)->Resize(iMax);


  _nColumn = _columns.size();   // Should never be different, but...
}


// Clears all columns, i.e. removes all rows from the table
void GloriaFitsTable::ClearTable()
{
  for(int iCol = 0; iCol!=_columns.size(); iCol++)
           _columns.at(iCol)->Clear();


  _nRow=0;

  _nColumn = _columns.size();   // Should never be different, but...

  _nSorted=0;
  _sortName.clear();
  _sortColumn=0;

}


// Delete all columns in the table and clear table header
void GloriaFitsTable::EmptyTable()
{

  // Delete columns first (pointers are stored)

  for(int iCol=0; iCol<_nColumn; iCol++)
      delete _columns.at(iCol);
          
  // Clear all class data

  InitializeEmptyTable();
}

//! Checks consistency of table structure
bool GloriaFitsTable::CheckTable()
{

if(_nColumn != _columns.size())
    {
    _statusOK=false;
    _errorMessage="Column number mismatch !! ";

     return false;
    } 
 
if(_nColumn != _nameMap.size())
    {
    _statusOK=false;
    _errorMessage="Column name index size mismatch !! ";

     return false;
    } 
 
for(int iCol = 0; iCol!=_columns.size(); iCol++)
  if(_nRow != _columns.at(iCol)->Size())
        {
        _statusOK=false;
        _errorMessage="Row number mismatch: columns need adjustment !! ";

        return false;
        } 

 ClearError();

 return true;
}

//Create an index array for accessing table rows in order of increasing column values
long* GloriaFitsTable::CreateSortIndex(int iCol, bool reverse)
{
ClearError();

// Nothing to do for empty table

if(_nRow == 0) return NULL;

if(iCol<0 || iCol>=_nColumn)
  {
  _statusOK=false;
  _errorMessage="Column index out of range";

  return NULL;
  } 

if(  _columns.at(iCol)->Type() == LongVectorColumn ||
     _columns.at(iCol)->Type() == DoubleVectorColumn ||
     _columns.at(iCol)->Type() == UndefinedColumn ) 
    {
    _statusOK=false;
    _errorMessage="Column type can not be used for sorting";

     return NULL;
    } 

 return _columns.at(iCol)->CreateSortIndex(reverse);
}

// Create an index array for accessing table rows in order of increasing column values
long* GloriaFitsTable::CreateSortIndex(string name, bool reverse)
{
if(!ColumnExists(name))
    {
    _statusOK=false;
    _errorMessage="Column " + name + " does not exist";

     return NULL;
    } 

 int iCol = _nameMap.find(name)->second;

 return CreateSortIndex(iCol, reverse);
}

// Sort table rows in order of increasing  values of selected column.
void  GloriaFitsTable::SortTable(int iCol, bool reverse)
{
ClearError();

// Just in case...

AdjustColumns();

// Nothing to do for empty table

if(_nRow == 0) return;

if(iCol<0 || iCol>=_nColumn)
  {
  _statusOK=false;
  _errorMessage="Column index out of range";

  return;
  } 

long* index = CreateSortIndex(iCol, reverse);

if(index==NULL) return;

for(int icl=0; icl< _nColumn; icl++)
         _columns.at(icl)->SortByIndex(index);

delete [] index;

 _nSorted = _nRow;
 _sortColumn = iCol;
 _sortName = _columns.at(iCol)->Name();

return;
}

// Sort table rows in order of increasing  values of selected column.
void  GloriaFitsTable::SortTable(string name, bool reverse)
{
if(!ColumnExists(name))
    {
    _statusOK=false;
    _errorMessage="Column " + name + " does not exist";

     return;
    } 

 int iCol = _nameMap.find(name)->second;

 SortTable(iCol, reverse);

 return;
}


// Delete given row
void GloriaFitsTable::DeleteRow(int iRow)
{
  for(int iCol = 0; iCol!=_columns.size(); iCol++)
               _columns.at(iCol)->DeleteRow(iRow);

  _nRow--;
}

// Return column pointer and remove column from the table

GloriaTableColumn* GloriaFitsTable::RemoveColumn(string name)
{
  ClearError();

  // Return without error if column does not exist

  if(!ColumnExists(name))   return NULL;

  return RemoveColumn(_nameMap.find(name)->second);
}


// Return column pointer and remove column from the table


GloriaTableColumn* GloriaFitsTable::RemoveColumn(int iCol)
{
  ClearError();

  // Return without error if column does not exist

  if(iCol<0 || iCol>=_nColumn)   return NULL;

  // Delete column and remove pointer 

  GloriaTableColumn* col = _columns.at(iCol);

  _columns.erase(_columns.begin()+iCol);

 _nColumn--;

  // Refill name map - probably easier than correcting it

  _nameMap.clear();

  for(int icl=0;icl<_nColumn;icl++)
          _nameMap.insert( pair<string, int>(_columns.at(icl)->Name(),icl));

  return col;
}

// Delete column from the table

void GloriaFitsTable::DeleteColumn(string name)
{
  ClearError();

  // Return without error if column does not exist

  if(!ColumnExists(name))   return;

  DeleteColumn(_nameMap.find(name)->second);
}


// Delete column from the table


void GloriaFitsTable::DeleteColumn(int iCol)
{
  ClearError();

  // Return without error if column does not exist

  if(iCol<0 || iCol>=_nColumn)   return;

  // Delete column and remove pointer 

  delete _columns.at(iCol);

  _columns.erase(_columns.begin()+iCol);

 _nColumn--;

  // Refill name map - probably easier than correcting it

  _nameMap.clear();

  for(int icl=0;icl<_nColumn;icl++)
          _nameMap.insert( pair<string, int>(_columns.at(icl)->Name(),icl));
}

void GloriaFitsTable::Print( std::ostream& os, int colWidth)
{

  if( _nColumn == 0 )
    { 
      os << "  Table is empty  " << endl;
    }

  if(colWidth < 3)colWidth=3;
  if(colWidth > 20)colWidth=20;

  string Vsep = "|";
  string Xsep = "+";
  stringstream hline; 

  hline <<  Xsep;
  hline.fill('-'); 
  for(int icol=0;icol<_nColumn;icol++) hline <<  setw(colWidth+1) << Xsep ;


  // Table header: column names and units

  os << hline.str() << endl;

  os.fill(' '); 
  os << setw(0) << Vsep;

  for(int icol=0;icol<_nColumn;icol++) os <<  setw(colWidth) <<  left << GetColumnName(icol) << Vsep;
  os << endl;

  os << setw(0) << Vsep;
  for(int icol=0;icol<_nColumn;icol++) os <<  setw(colWidth) <<  right << GetColumn(icol)->Unit() << Vsep;
  os << endl;

  os << hline.str() << endl;


  // Table data - loop over rows
  
  for(int irow=0;irow<_nRow;irow++)
    {
    // First check what is the maximum size of element in this row
    int smax = 0;
    for(int icol=0;icol<_nColumn;icol++) 
        if( GetColumn(icol)->ValSize(irow) > smax)
                      smax =  GetColumn(icol)->ValSize(irow);

    // Then print subsequent values, one per row
    for(int ival=0;ival<smax;ival++)
      {
      os << setw(0) << Vsep;

      for(int icol=0;icol<_nColumn;icol++) 
          if( ival < GetColumn(icol)->ValSize(irow) )
	    os <<  setw(colWidth) <<  GetColumn(icol)->StringVal(irow,ival) << Vsep;
	  else
	    os <<  setw(colWidth + 1)  << Vsep;

      os << endl;
      }

    os << hline.str() << endl;
    }

  // Column comments, if present

  os.fill(' '); 
  for(int icol=0;icol<_nColumn;icol++) 
    if(GetColumn(icol)->Comment().size() > 0)
        os << GetColumnName(icol) << " : " << GetColumn(icol)->Comment() << endl;

}


// File access methods
//=====================

// File constructor 


GloriaFitsTable::GloriaFitsTable(string FitsFileName)
{
  InitializeEmptyTable();

  ReadFitsFile(FitsFileName);
}



void GloriaFitsTable::WriteTextFile(string TextFileName, int colWidth, char ColumnSeparator)
{

  // No output for empty table

  if( _nColumn == 0 )return;

  // Open output file
 
  ofstream outF;

  outF.open(TextFileName.c_str(),std::ios::out);

  if(outF.fail())
      {
      _statusOK=false;
      _errorMessage= "Error opening output file " + TextFileName;

       return;
      }

 
 WriteTextStream(outF,colWidth,ColumnSeparator);

 outF.close();
}


void GloriaFitsTable::WriteTextStream( std::ostream& outF, int colWidth, char ColumnSeparator)
{

  // No output for empty table

  if( _nColumn == 0 )return;

  // Table header: column names and units

  outF.fill(' '); 

  outF << "# ";
  for(int icol=0;icol<_nColumn;icol++)  outF <<  setw(colWidth) <<  left << GetColumnName(icol) << ColumnSeparator;
  outF << endl;


  outF << "# ";
  for(int icol=0;icol<_nColumn;icol++) outF <<  setw(colWidth) <<  right << GetColumn(icol)->Unit() << ColumnSeparator;
  outF << endl;

  // Table data - loop over rows
  
  for(int irow=0;irow<_nRow;irow++)
    {
    // Print subsequent values, one per row
    for(int icol=0;icol<_nColumn;icol++) 
      outF <<  setw(colWidth) <<  GetColumn(icol)->StringVal(irow) << ColumnSeparator;

    outF << endl;
    }

}


// Read table from text file

void GloriaFitsTable::ReadTextFile(string TextFileName)
{

  // No input for empty table

  if( _nColumn == 0 )return;

  // Open input file
 
  ifstream inF;

  inF.open(TextFileName.c_str(),std::ios::in);

  if(inF.fail())
      {
      _statusOK=false;
      _errorMessage= "Error opening input file " + TextFileName;

       return;
      }

  // Read input file line by line

 int maxlinelen = 1024;
 
 char *maxline = new char[maxlinelen];

 while(1)
   {
   inF.getline(maxline,maxlinelen);

   if(!inF.good())break;

   // Skip comment lines

   if(maxline[0] == '#') continue;

   // Add new empty row

   AddRow();
   
   stringstream inputline;

   inputline << maxline; 

   for(int icol=0; icol< _nColumn; icol++)
     {
       switch(GetColumnType(icol))
	 {
	 case LongColumn:
	   long lval;
           inputline >> lval;
           SetLast(icol,lval);
	   break;

	 case DoubleColumn:
	   double dval;
           inputline >> dval;
           SetLast(icol,dval);
	   break;

	 case StringColumn:
	   string sval;
           inputline >> sval;
           SetLast(icol,sval);
	   break;
	 }
     }
   }


 inF.close();
}

// Read table from FITS file

void GloriaFitsTable::ReadFitsFile(string FitsFileName)
{
   fitsfile *fp=0;
   int status = 0;


   _statusOK=true;
   _errorMessage.clear();

   // Open fits file: look for table (!)

   fits_open_table(&fp, FitsFileName.data(), READONLY, &status);

   if (status)
     {
       _statusOK=false;
       _errorMessage="Error opening fits table file";
       return;
     }

   // Clear table (all current contents is lost!)

   ClearTable();

   // Get input table parameters

   int ncols;
   long nrows;
   fits_get_num_rows(fp,&nrows,&status);
   fits_get_num_cols(fp,&ncols,&status);

// Retrieve column names and place them into name vector (temporary)

   vector<string> colnames;
   vector<int> colnums;

   char colname[80];     
   int colnum;
      
   fits_get_colname(fp, CASEINSEN, (char*) "*", colname, &colnum, &status);

   while (status == COL_NOT_UNIQUE)
      {
      string strname(colname);

      colnames.push_back(strname);
      colnums.push_back(colnum);

      fits_get_colname(fp, CASEINSEN, (char*) "*", colname, &colnum, &status);
      }

   if (status != COL_NOT_FOUND)
     {
       _statusOK=false;
       _errorMessage="Error reading fits table format information";
       return;
     }

   status = 0;  

   // Check if all columns found

   if(colnums.size() != ncols)
     {
       _statusOK=false;
       _errorMessage="Fits table format inconsistent !?";
       return;
     }

   // Loop over columns of the FITS table

   for(int icol=0; icol<ncols; icol++)
     {
       colnum = colnums[icol];
       string colname = colnames[icol];

       int typecode;
       long repeat, width;

      fits_get_coltype(fp, colnum, &typecode, &repeat, &width, &status);


      // Column with double or float values

      if ((typecode == TDOUBLE) || (typecode == TFLOAT)) 
        {
         double *array;
         array = new double [nrows * repeat];  

         double nulval = 0;
         int anynul=0;
               
         fits_read_col(fp, TDOUBLE, colnum, 1, 1, nrows * repeat, 
                                    &nulval, array, &anynul, &status);
                  
         if (status) 
             {
             delete [] array;

             _statusOK=false;
             _errorMessage="Error reading float/double column from fits table";

	     return;
             }


	 if(repeat == 1)
	   {
           AddColumn(colname, DoubleColumn);

	   //	   for(long irow=0; irow<nrows; irow++)
	   //          GetVectorOf<double>(icol).push_back(array[irow]);
	     
	   //
	   // Should be much more efficient

           GetVectorOf<double>(icol).assign(array, array+nrows);
	     
	   }
	 else
	   {
           AddColumn(colname, DoubleVectorColumn);

	   for(long irow=0; irow<nrows; irow++)
	     {
	     vector<double> vtmp(array + repeat*irow, array + repeat*(irow+1));
             GetVectorOf<DoubleVec>(icol).push_back(vtmp);
	     }
	   }

	 delete [] array;
	}
         // end of float/double column case
             
 
      // Column with byte, int or long values

      if ((typecode == TSHORT) || (typecode == TLONG) || (typecode == TBYTE))
        {
         long *array;
         array = new long [nrows * repeat];  

         long nulval = 0;
         int anynul=0;
               
         fits_read_col(fp, TLONG, colnum, 1, 1, nrows * repeat, 
                                    &nulval, array, &anynul, &status);
                  
         if (status) 
             {
             delete [] array;

             _statusOK=false;
             _errorMessage="Error reading byte/int/long column from fits table";

	     return;
             }


	 if(repeat == 1)
	   {
           AddColumn(colname, LongColumn);

	   //	   for(long irow=0; irow<nrows; irow++)
	   //             GetVectorOf<long>(icol).push_back(array[irow]);

           GetVectorOf<long>(icol).assign(array, array+nrows);
	     
	   }
	 else
	   {
           AddColumn(colname, LongVectorColumn);

	   for(long irow=0; irow<nrows; irow++)
	     {
	     vector<long> vtmp(array + repeat*irow, array + repeat*(irow+1));
             GetVectorOf<LongVec>(icol).push_back(vtmp);
	     }
	   }

	 delete [] array;
	}
         // end of byte/int/long column case
             


      // String columns

      if (typecode == TSTRING) 
        {
         int dispwidth=0;
         fits_get_col_display_width(fp, colnum, &dispwidth, &status);
         if (dispwidth <= 0)  dispwidth = 1;
               
         char *nulval = (char*) "";
         int anynul=0;
         char **array;
               
               
         array = new char* [nrows];
         for (long irow = 0; irow < nrows; irow++) 
             array[irow] = new char[dispwidth+1]; //also room for end null!!
               
         fits_read_col(fp, TSTRING, colnum, 1, 1, nrows, 
                                          nulval, array, &anynul, &status);

        if (status) 
           {
           for (long irow = 0; irow < nrows; irow++)  delete [] array[irow];
           delete [] array;

           _statusOK=false;
           _errorMessage="Error reading string column from fits table";

	   return;
           }

        AddColumn(colname, StringColumn);

	 for(long irow=0; irow<nrows; irow++)
	   {
           string stmp(array[irow]);
           GetVectorOf<string>(icol).push_back(stmp);

           delete [] array[irow]; 
	   } 


        delete [] array;
        }
        // End of string column case 
 
     }
     // End of loop over columns in input table


   AdjustColumns();

   // Read header: use method of GloriaFitsHeader

   ReadFitsHeader(fp);

 
   fits_close_file(fp, &status);
   return;
}


// Write table to FITS file

void GloriaFitsTable::WriteFitsFile(string FitsFileName, bool binary_tbl)
{
  AdjustColumns();
  ClearError();

  fitsfile *fp=0;
  int status = 0;

  // Create FITS file
  fits_create_file(&fp, FitsFileName.data(), &status);

  if (status)
  {
      _statusOK=false;
      _errorMessage="Error opening fits file for writing";
      return;
  }

  // Create table extension
  
  int tbltype = (binary_tbl)? BINARY_TBL : ASCII_TBL;

  int nrows = _nRow;
  int tfields = _nColumn;

  char **ttype = new char*[tfields];
  char **tform = new char*[tfields];
  char **tunit = new char*[tfields];

  int *colwidth = new int[tfields];

  for(int icol=0;icol<_nColumn;icol++)
    {
      ttype[icol] =  (char *) _columns.at(icol)->Name().c_str();
      tunit[icol] =  (char *) _columns.at(icol)->Unit().c_str();

      if(_columns.at(icol)->Type() == StringColumn)
	 {
	 colwidth[icol] = 10; // default width for string columns

         vector<string> *svec = &(GetVectorOf<string>(icol));
         for(long irow=0; irow < svec->size(); irow++)
	     if((svec->at(irow)).length() >= colwidth[icol]) colwidth[icol] = (svec->at(irow)).length()+1;

	 tform[icol] = new char[6];

         if(binary_tbl)
               sprintf(tform[icol],"%dA",colwidth[icol]);
         else
               sprintf(tform[icol],"A%d",colwidth[icol]);
         }
      else
      if(_columns.at(icol)->Type() ==  LongColumn)
	{
	tform[icol] = (char *)((binary_tbl)? "1J" : "I12");
        colwidth[icol] = 1;
        }
      else
      if(_columns.at(icol)->Type() ==   DoubleColumn)
	{
	tform[icol] = (char *)((binary_tbl)? "1D" : "D20.13");
        colwidth[icol] = 1;
        }
      else
      if(_columns.at(icol)->Type() ==  DoubleVectorColumn)
	{
	 colwidth[icol] = 1; // default width for numerical vector columns

         vector<DoubleVec> *dvec = &(GetVectorOf<DoubleVec>(icol));
         for(long irow=0; irow < dvec->size(); irow++)
	     if((dvec->at(irow)).size() > colwidth[icol]) colwidth[icol] = (dvec->at(irow)).size();

	tform[icol] = new char[6];

        sprintf(tform[icol],"%dD",colwidth[icol]);

	if(!binary_tbl)
	  {
          _statusOK=false;
          _errorMessage="Writing vector columns not possible for ASCII format!!!";
	  }
	}
      else
      if(_columns.at(icol)->Type() ==  LongVectorColumn)
	{
	 colwidth[icol] = 1; // default width for numerical vector columns

         vector<LongVec> *lvec = &(GetVectorOf<LongVec>(icol));
         for(long irow=0; irow < lvec->size(); irow++)
	     if((lvec->at(irow)).size() > colwidth[icol]) colwidth[icol] = (lvec->at(irow)).size();

	tform[icol] = new char[6];

        sprintf(tform[icol],"%dJ",colwidth[icol]);

	if(!binary_tbl)
	  {
          _statusOK=false;
          _errorMessage="Writing vector columns not possible for ASCII format!!!";
	  }
	} 
    }


  if(! _statusOK) return;

  char *extname = (char* )"GLORIA_TABLE";  

  fits_create_tbl(fp, tbltype, nrows, tfields, ttype, tform, tunit, extname, &status);

  delete [] ttype;
  delete [] tform;
  delete [] tunit;

  if (status)
  {
      _statusOK=false;
      _errorMessage="Error creating table extension in fits file";
      return;
  }


  // Write subsequent columns

  long firstrow  = 1;  /* first row in table to write   */
  long firstelem = 1;  /* first element in row  (ignored in ASCII tables) */

  

  for(int icol=0;icol<_nColumn;icol++)
    {
      if(_columns.at(icol)->Type() == StringColumn)
	{
          vector<string> *svec = &(GetVectorOf<string>(icol));

	  char **scont = new char*[_nRow];
          for(long irow=0;irow<_nRow;irow++)
	    scont[irow] = (char *)((svec->at(irow)).c_str());

           fits_write_col(fp, TSTRING, icol+1, firstrow, firstelem, nrows, scont, &status);

           delete [] scont;
	}
      else
      if(_columns.at(icol)->Type() == LongColumn)
	  {

          vector<long> *lvec = &(GetVectorOf<long>(icol));

	  long *lcont = new long[_nRow];
          for(long irow=0;irow<_nRow;irow++)
	         lcont[irow] = lvec->at(irow);

           fits_write_col(fp, TLONG, icol+1, firstrow, firstelem, nrows, lcont, &status);

           delete [] lcont;
	  }
      else
      if(_columns.at(icol)->Type() == DoubleColumn)
         {

          vector<double> *dvec = &(GetVectorOf<double>(icol));

	  double *dcont = new double[_nRow];
          for(long irow=0;irow<_nRow;irow++)
	         dcont[irow] = dvec->at(irow);

           fits_write_col(fp, TDOUBLE, icol+1, firstrow, firstelem, nrows, dcont, &status);

           delete [] dcont;
         }
      else
      if(_columns.at(icol)->Type() == DoubleVectorColumn)
         {

          vector<DoubleVec> *dvec = &(GetVectorOf<DoubleVec>(icol));

	  double *dcont = new double[_nRow*colwidth[icol]];

          for(long irow=0;irow<_nRow;irow++)
	    for(int iel=0;iel< colwidth[icol];iel++)
	      { 
		if(iel < (dvec->at(irow)).size())
 	            dcont[irow*colwidth[icol]+iel] = (dvec->at(irow)).at(iel);
		else
		    dcont[irow*colwidth[icol]+iel] = 0.;
	      }

           fits_write_col(fp, TDOUBLE, icol+1, firstrow, firstelem, nrows*colwidth[icol], dcont, &status);

           delete [] dcont;
         }
      else
      if(_columns.at(icol)->Type() == LongVectorColumn)
         {

          vector<LongVec> *lvec = &(GetVectorOf<LongVec>(icol));

	  long *lcont = new long[_nRow*colwidth[icol]];

          for(long irow=0;irow<_nRow;irow++)
	    for(int iel=0;iel< colwidth[icol];iel++)
	      { 
		if(iel < (lvec->at(irow)).size())
 	            lcont[irow*colwidth[icol]+iel] = (lvec->at(irow)).at(iel);
		else
		    lcont[irow*colwidth[icol]+iel] = 0;
	      }

           fits_write_col(fp, TLONG, icol+1, firstrow, firstelem, nrows*colwidth[icol], lcont, &status);

           delete [] lcont;
         }
  

     if (status)
        {
        _statusOK=false;
        _errorMessage="Error filling table column " + _columns.at(icol)->Name() + " of " + _columns.at(icol)->TypeName() + " in fits file";
        return;
        }

 
    }

  // Write header: use method of GloriaFitsHeader
  WriteFitsHeader(fp);

  fits_close_file(fp, &status);

  if (status)
        {
        _statusOK=false;
        _errorMessage="Error closing table fits file";
        return;
        }


}



}
