// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaMedianMap.h"
#include <algorithm>

namespace gloria{

using namespace std;
  
// Constructor creating map tables for given image

  GloriaMedianMap::GloriaMedianMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int Nvar)
  {
    InitializeMaps(Nbx, Nby, Nvar);

    SetImageSize(fitsImage);
  }


// Constructor creating map tables without specifying the image

GloriaMedianMap::GloriaMedianMap(int Nbx, int Nby, int Nvar)
  {
    InitializeMaps(Nbx, Nby, Nvar);
  }

// Initialize maps

void GloriaMedianMap::InitializeMaps(int Nbx, int Nby, int Nvar)
{
  _nMapX = Nbx;
  _nMapY = Nby;
  _nMapXY = Nbx * Nby;
 
 
  for(int imap=0; imap<_valueMaps.size(); imap++)
      delete [] _valueMaps.at(imap);

  _valueMaps.clear();
  _valueNames.clear();
 
  for(int ivar=0; ivar<Nvar; ivar++)
    {
     double *medMap = new double[_nMapXY];
     double *spreadMap = new double[_nMapXY];

     stringstream medName, spreadName;
     medName << "Median" << ivar ;
     spreadName << "Spread" << ivar ;


    _valueMaps.push_back(medMap);
    _valueMaps.push_back(spreadMap);

    _valueNames.push_back(medName.str());
    _valueNames.push_back(spreadName.str());
    }

  _sizeX = _sizeY =  _sizeXY = 0;
  _binX = _binY = _binXY = 0;
}

// Add new variable

void GloriaMedianMap::AddMedian()
 {
   stringstream medName, spreadName;
   medName << "Median" << _valueMaps.size() ;
   spreadName << "Spread" << _valueMaps.size() ;

   AddMap(medName.str());
   AddMap(spreadName.str());
 }


// Add new variable with a name

void GloriaMedianMap::AddMedian(string name)
 {
   AddMap(name);
   AddMap("S_"+name);
 }

// Set name to the considered variable
void GloriaMedianMap::SetMedianName(int ivar, string name)
{
  _valueNames.at(2*ivar) = name; 
  _valueNames.at(2*ivar+1) = "S_" + name;
}

// Calculate map from input data arrays

long GloriaMedianMap::CalculateMap(int iVar, int nVal, double* Xpos, double* Ypos, double* Val)
  {
  // Wrong input data
  if(nVal <=0  || Xpos==NULL || Ypos==NULL || Val==NULL) return -1;

  // if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return -1;

  // variable index out of range
  if(iVar < 0 || 2*iVar + 2 > _valueMaps.size() )return -1;

  vector<double> Xvec(Xpos,Xpos+nVal);
  vector<double> Yvec(Ypos,Ypos+nVal);
  vector<double> Vvec(Val,Val+nVal);

  long nmin =  CalculateMap(iVar,&Xvec,&Yvec,&Vvec);

  return nmin;
  }


long GloriaMedianMap::CalculateMap(int iVar, vector<double>* Xpos, vector<double>* Ypos, vector<double>* Val)
 {
  // Wrong input data
 if(Xpos->size()<Val->size() || Ypos->size()<Val->size()) return -1;

  // if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return -1;

  // variable index out of range
 if(2*iVar + 2 > _valueMaps.size() )return -1;

 double *medMap = _valueMaps.at(2*iVar);
 double *spreadMap =  _valueMaps.at(2*iVar+1);

 vector<double> *ValVec = new vector<double>[_nMapXY];

 for(int ival=0; ival< Val->size() ; ival++)
   {
     double x = Xpos->at(ival);
     double y = Ypos->at(ival);

     if(x < 0.5 || y < 0.5 || x >= _sizeX+0.5 || y >= _sizeY+0.5 ) continue;

     int ibx = (x-0.5)/_binX;
     int iby = (y-0.5)/_binY;

     int imap = ibx + iby*_nMapX;

     ValVec[imap].push_back(Val->at(ival));
   }

 long nmin = Val->size();

 for(int imap=0; imap< _nMapXY; imap++)
   {

     medMap[imap] = spreadMap[imap] = 0.;

     int nval = ValVec[imap].size();

     if(nval < nmin) nmin = nval;

     if(nval == 0) continue;

     sort(ValVec[imap].begin(),ValVec[imap].end());

     int imed=nval/2;

     if(nval%2 == 1)     
       medMap[imap]=ValVec[imap].at(imed);
     else
       medMap[imap]=0.5*(ValVec[imap].at(imed) + ValVec[imap].at(imed-1));


     int ilow  = (int)(0.17*nval);
     int ihigh = (int)(0.83*nval);

     spreadMap[imap]=0.5*(ValVec[imap].at(ihigh)-ValVec[imap].at(ilow));
   }

 delete [] ValVec;

 return nmin;
 }


// Calculate map, with possible adjustment of map size

void GloriaMedianMap::CalculateBgMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int sampling)
  {

   if(Nbx != _nMapX || Nby != _nMapY) InitializeMaps(Nbx, Nby, 1);

   CalculateBgMap(fitsImage, sampling);
  }

//
//    Calculate map from image
//

void GloriaMedianMap::CalculateBgMap(const GloriaFitsImage* fitsImage, int sampling)
  {
  if(_valueMaps.size()==0 || _nMapXY==0)return;
 
  double *bgMap = _valueMaps.at(0);
  double *noiseMap =  _valueMaps.at(1);

   _valueNames.at(0)="Bgnd";
   _valueNames.at(1)="Noise";

  // Calculate map bin sizes

    SetImageSize(fitsImage);

  // Images can be float or int

  double* floatPixels; 
  long* intPixels;

  if(fitsImage->IsFloatImage())
     floatPixels =  fitsImage->GetFloatImageData();
  else
     intPixels =  fitsImage->GetIntImageData();

  // Array for storing pixel values (for sorting)

  int nbg = _binXY / sampling + 1;
  double* subPixels = new double[nbg];

  // Calculate median and noise for sub-frames
  // Loop over sub-frames:

  for(int iSub=0;iSub<_nMapXY;iSub++)
    {
    int iSubX = iSub%_nMapX;
    int iSubY = iSub/_nMapX;
    
    long ibg=0;   // pixel counter
 
    // Loop over pixels in sub-frame

    for(long ipix=0;ipix<_binXY;ipix+=sampling)
      {
	int ipx = ipix%_binX + iSubX*_binX;
	int ipy = ipix/_binX + iSubY*_binY;

	if(ipx >= _sizeX || ipy >= _sizeY) continue;

        if(fitsImage->IsFloatImage())
           subPixels[ibg++] = floatPixels[ipx + _sizeX*ipy];
        else
	   subPixels[ibg++] = (double) intPixels[ipx + _sizeX*ipy];
      }

    // Do sorting of selected BG pixels

    std::sort(subPixels,subPixels+ibg);


    // Calculate background and noise levels

   int imed=ibg/2;

   if(ibg%2 == 1)     
      bgMap[iSub]=subPixels[imed];
   else
      bgMap[iSub]=0.5*(subPixels[imed]+subPixels[imed-1]);


   int ilow  = (int)(0.17*ibg);
   int ihigh = (int)(0.83*ibg);

   noiseMap[iSub]=0.5*(subPixels[ihigh]-subPixels[ilow]);
    }

  }
  
}// namespace gloria
