// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaNetworkMgr.h"
#include "gloria/GloriaTelescope.h"
#include "luiza/LuizaExceptions.h"
#include "tinyxml/tinyxml.h"

namespace gloria{

// helper function (from gearxml/XMLHandlerMgr.cc) 

std::string getXMLAttribute(const TiXmlNode* node , const std::string& name ) {

    const TiXmlElement* el = node->ToElement() ;
    if( el == 0 )
      throw luiza::ParseException("XMLParser::getAttribute not an XMLElement " ) ;

    const char* at = el->Attribute( name.c_str() ) ;
  
    if( at == 0 ){
    
      std::stringstream str ;
      str  << "XMLParser::getAttribute missing attribute \"" << name
           << "\" in element <" << el->Value() << "/> " ;
      throw luiza::ParseException( str.str() ) ;
    }

    return std::string( at )  ;

  }


// Default constructor

GloriaNetworkMgr::GloriaNetworkMgr()
{
  _telescopeVec.clear(); 
  _telescopeNameMap.clear();
  _nTelescopes=0;
}


// Named constructor

GloriaNetworkMgr::GloriaNetworkMgr(std::string fileName)
{
_telescopeVec.clear(); 
_telescopeNameMap.clear();
_nTelescopes=0;

LoadConfigXML(fileName);
}


// Default destructor

GloriaNetworkMgr::~GloriaNetworkMgr()
{
}


// Get pointer to telescope parameters using telescope name or alias

GloriaTelescope*  GloriaNetworkMgr::GetTelescope(std::string name)
 { 
  GloriaTelescope*  tel = NULL;

  if(_telescopeNameMap.find(name)!=_telescopeNameMap.end())
     tel  =  &(_telescopeVec.at(_telescopeNameMap[name]));   

  return tel;
 }


// Get pointer to telescope parameters using telescope name or alias

GloriaTelescope*  GloriaNetworkMgr::GetTelescope(int id)
 { 
  GloriaTelescope*  tel = NULL;

  if(_telescopeIdMap.find(id)!=_telescopeIdMap.end())
     tel  =  &(_telescopeVec.at(_telescopeIdMap[id]));   

  return tel;
 }


std::ostream& operator<< (  std::ostream& s,  GloriaNetworkMgr& n ) 
{
  s << "Network configuration ID = " << n._id <<  std::endl ;
  s << "Author:  " << n._author << "   date: " << n._date << std::endl;

  s << n._nTelescopes << " telescopes defined in the network" << std::endl << std::endl;

  for(int itel = 0 ; itel < n._nTelescopes ; itel++)
    s << n._telescopeVec[itel] << std::endl;


  return s ;
}

// Main parser of the configuration file


void GloriaNetworkMgr::LoadConfigXML(std::string fileName)
{
    TiXmlDocument doc ;
 
    bool loadOkay = doc.LoadFile( fileName  ) ;

    if( !loadOkay ) {

      std::stringstream str ;

      str  << "GloriaNetworkMgr::LoadConfigXML error in file: " << fileName
           << "[row: " << doc.ErrorRow() << ", col: " << doc.ErrorCol() << "] : "
           << doc.ErrorDesc() ;

      throw luiza::ParseException( str.str() ) ;
    }

    TiXmlElement* root = doc.RootElement();
      
    if( root == 0 ){
      throw luiza::ParseException( std::string( "GloriaNetworkMgr::LoadConfigXML: no root tag found in  ")
                           + fileName  ) ;
    }

    
    TiXmlNode* global =  root->FirstChild("global") ;
    if( global != 0 )
       {
	   _id  =  atoi (getXMLAttribute( global, "configID" ).c_str() ) ;
	 _date  =  atoi (getXMLAttribute( global, "configDate" ).c_str() ) ;
       _author  =  getXMLAttribute( global, "configAuthor" )  ;
       }


    TiXmlNode* observatories = root->FirstChild("observatories")  ;
    if( observatories == 0 ){
      throw luiza::ParseException( std::string( "GloriaNetworkMgr::LoadConfigXML: no observatories tag found in  ")
                           + fileName  ) ;
    }

    // Loop over telescopes

    TiXmlNode* tel = 0 ;

    while( ( tel = observatories->IterateChildren( "telescope", tel ) )  != 0  )
      {
      int tel_id  =  atoi( getXMLAttribute( tel, "ID" ).c_str() )  ;
      std::string tel_name  =  getXMLAttribute( tel, "name" )  ;

      // Create telescope instance

      GloriaTelescope* tel_conf = new GloriaTelescope(tel_id, tel_name); 

      // Add aliases if found

      TiXmlNode* alias = 0 ;
      while( ( alias = tel->IterateChildren( "alias", alias ) )  != 0  )
	 tel_conf->SetName( getXMLAttribute( alias, "name"));


      // Add position if found
      TiXmlNode* position = tel->FirstChild("position");

      if(position != 0)
	{
	double lon = atof( getXMLAttribute( position, "longitude" ).c_str() )  ;
	double lat = atof( getXMLAttribute( position, "latitude" ).c_str() )  ;
	double elev = atof( getXMLAttribute( position, "elevation" ).c_str() )  ;

	tel_conf->SetPosition(lon,lat,elev);
        }

      // Add string parameters defined

      TiXmlNode* param = 0 ;
      while( ( param = tel->IterateChildren( "parameter", param ) )  != 0  )
	{
	  std::string par_name = getXMLAttribute( param, "name");
	  std::string par_value = getXMLAttribute( param, "value");

          tel_conf->AddParameter(par_name,par_value);
	}


      // Store telescope data in configuration

      _telescopeVec.push_back(*tel_conf);
      _telescopeIdMap.insert(pair<int,int>(tel_id,_nTelescopes));

      //  Primary name only:
      //         _telescopeNameMap.insert(pair<std::string,int>(tel_name,_nTelescopes));

      // Telescope name: need to store aliases as well !

      for(int ialias=0; ialias< tel_conf->GetNameNumber(); ialias++)
	{
	 string telalias = tel_conf->GetName(ialias);
	_telescopeNameMap.insert(pair<std::string,int>(telalias,_nTelescopes));

	// Telescope name can be enclosed in apostrophes in the
	// FITS header:
        if(telalias.at(0)!='\'')
	  {
	   string  telalias2 = "'" + telalias + "'";
  	  _telescopeNameMap.insert(pair<std::string,int>(telalias2,_nTelescopes));
	  }
	}

      // Increase telescope counter
      _nTelescopes++;

      delete tel_conf;
      } // end of loop over telescopes

}

} // namespace gloria
