// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaSkyCatalog.h"

#include <cmath>

using namespace std;

namespace gloria{

GloriaSkyCatalog::GloriaSkyCatalog() 
: GloriaFitsTable() {
  AddColumn("RA",(double) -100., string("deg"),string("Object RA position (J2000)"));
  AddColumn("DEC",(double) -100., string("deg"),string("Object DEC position (J2000)"));
}

// Check catalog table structure

bool  GloriaSkyCatalog::CheckCatalog()
{
 if( ! CheckTable()) return false;

 if(_nColumn < 2)
    {
    _statusOK=false;
    _errorMessage="Too few columns defined ! ";

     return false;
    } 
 
 if( _columns.at(0)->Name() != string("RA")  || _columns.at(0)->Type() != DoubleColumn
  || _columns.at(1)->Name() != string("DEC") || _columns.at(1)->Type() != DoubleColumn)
    {
    _statusOK=false;
    _errorMessage="Required columns not defined ! ";

     return false;
    } 
 

 ClearError();

 return true;
}

// Match object in catalog by its position on the sky

int GloriaSkyCatalog::MatchObject(double ra, double dec, double cut)
{
  int imatch = -1;
  double dmin = cut;

  // Get pointers to catalog position vectors 
  vector<double>* RaVec = &(((GloriaTableColumnOf<double> *)_columns.at(0))->Data());
  vector<double>* DecVec = &(((GloriaTableColumnOf<double> *)_columns.at(1))->Data());

  if(RaVec->size() != DecVec->size() || RaVec->size() == 0 ) return imatch;

  double decmin = dec - cut;
  double decmax = dec + cut;

  if(cut == 0.)
    {
    decmin=-90.;
    decmax=90.;
    dmin=180.;
    }

  for(int irow=0; irow<_nRow; irow++)
    {
      double cdec = DecVec->at(irow);

      if(cdec < decmin || cdec > decmax) continue;

      double cra = RaVec->at(irow);

      double dist = SkyDistance(ra,dec,cra,cdec);

      if(dist < dmin)
	{
	  dmin=dist;
	  imatch = irow;
	}
    }

  return imatch;
}

// Match all objects in catalog at given position on the sky

vector<long>  GloriaSkyCatalog::MatchAllObjects(double ra, double dec, double distMax, double distMin)
{
  vector<long>  match;


  // Get pointers to catalog position vectors 

  vector<double>* RaVec = &(((GloriaTableColumnOf<double> *)_columns.at(0))->Data());
  vector<double>* DecVec = &(((GloriaTableColumnOf<double> *)_columns.at(1))->Data());

  if(RaVec->size() != DecVec->size() || RaVec->size() == 0 ) return match;

  double decmin = dec - distMax;
  double decmax = dec + distMax;

  multimap<double, long > distIndexMap;

  for(int irow=0; irow<_nRow; irow++)
    {
      double cdec = DecVec->at(irow);

      if(cdec < decmin || cdec > decmax) continue;

      double cra = RaVec->at(irow);

      double dist = SkyDistance(ra,dec,cra,cdec);

      if(dist >= distMin && dist <= distMax )
            distIndexMap.insert(make_pair( dist, irow ) );
    }

  multimap<double, long >::iterator idist;

  for(idist = distIndexMap.begin(); idist != distIndexMap.end(); idist++)
              match.push_back(idist->second);

  return match;
}

// Match list of objects by their position on the sky

long* GloriaSkyCatalog::MatchObjectArray(int nObject, double* RaArr, double* DecArr, double cut)
{
  long* match = NULL;

  if(nObject <=0  || RaArr==NULL || DecArr==NULL) return match;

  vector<double> RaVec(RaArr,RaArr+nObject);
  vector<double> DecVec(DecArr,DecArr+nObject);

  vector<long> idVec = MatchObjectVector(&RaVec,&DecVec,cut);

  if(idVec.size() == 0 || idVec.size() != nObject) return match;

  match = new long[nObject];

  for(int iobj=0; iobj<nObject; iobj++)
      match[iobj] = idVec.at(iobj);

  return match;
}

// Match list of objects by their position on the sky

vector<long> GloriaSkyCatalog::MatchObjectVector(vector<double>* RaVec, vector<double>* DecVec, double cut)
{
  // Initialize output vector

  vector<long> match;

  // Check input position vectors

  if(RaVec->size() != DecVec->size() || RaVec->size() == 0 ) return match;

  // Create input table

  GloriaTableColumnOf<double>* RaColumn = new GloriaTableColumnOf<double>("RA",(double)-100.);
  RaColumn->Data() = *RaVec;

  GloriaTableColumnOf<double>* DecColumn = new GloriaTableColumnOf<double>("DEC",(double)-100.);
  DecColumn->Data() = *DecVec;

  GloriaFitsTable InputTable;

  InputTable.InsertColumn("RA", RaColumn);
  InputTable.InsertColumn("DEC", DecColumn);

  return MatchObjectTable(&InputTable, cut);
}

// Match objects in the table with the catalog (using default column names)

vector<long> GloriaSkyCatalog::MatchObjectTable(GloriaFitsTable *table, double cut)
{
  vector<string> RaDecNames;
  vector<string> CopyNames;
  vector<string> NewNames;

  return MatchObjectTable(table, cut, RaDecNames, CopyNames, NewNames );
}

// Match objects in the table with the catalog 

vector<long> GloriaSkyCatalog::MatchObjectTable(GloriaFitsTable *table, double cut, vector<string> RaDecNames)
{
  vector<string> CopyNames;
  vector<string> NewNames;

  return MatchObjectTable(table, cut, RaDecNames, CopyNames, NewNames );
}

// Match objects in the table with the catalog, by their position on
// the sky - full version (actual implementation)

vector<long> GloriaSkyCatalog::MatchObjectTable(GloriaFitsTable *table, double cut, 
 vector<string> RaDecNames, vector<string> CopyNames, vector<string> TmpNames )
{

  // Check for columns to be copied
  // Suppress coping if any consistency found
  // This has to be done first!

  int nCopy = CopyNames.size();

  vector<string> NewNames = TmpNames;
  if(NewNames.size() == 0) NewNames = CopyNames;

  if(CopyNames.size() > 0 && NewNames.size() > 0 && 
     NewNames.size() != CopyNames.size() ) nCopy = 0;

  for(int icopy = 0 ; icopy < nCopy; icopy++)
     if(! ColumnExists(CopyNames.at(icopy)))nCopy = 0;

  // Create new columns if needed

  for(int icopy = 0 ; icopy < nCopy; icopy++)
    {
     string oldname = CopyNames.at(icopy);
     string newname = NewNames.at(icopy);

     if(! table->ColumnExists(newname))
       {
	 GloriaTableColumn*  newcolumn = GetColumn(oldname)->EmptyClone();
	 table->InsertColumn(newname,newcolumn);
       }
    }

  table->AdjustColumns();


  // Initialize output vector

  vector<long> match;


  // Check input table

  table->AdjustColumns();

  if(table->GetRowNumber() == 0 ) return match;

  // Get pointers to catalog position vectors 

  vector<double>* CatRa = &(((GloriaTableColumnOf<double> *)_columns.at(0))->Data());
  vector<double>* CatDec = &(((GloriaTableColumnOf<double> *)_columns.at(1))->Data());

  if(CatRa->size() != CatDec->size() || CatRa->size() == 0 ) return match;

  // Get pointers to input position vectors 

  string raName = "RA";
  string decName = "DEC";

  if(RaDecNames.size()==2)
    {
    raName = RaDecNames.at(0);
    decName = RaDecNames.at(1);
    }

  vector<double>* RaVec = &(table->GetVectorOf<double>(raName));
  vector<double>* DecVec = &(table->GetVectorOf<double>(decName));

  // Get sort index in DEC for input table

  long* DecInd = table->CreateSortIndex(decName);

  // Get sort index in DEC for the catalog 
  // (if not previously sorted in DEC)

  long* CatInd;

  if(!IsSortedBy("DEC"))
     CatInd = CreateSortIndex("DEC");
  else
    {
     CatInd = new long[_nRow];

     for(int icat=0; icat< _nRow; icat++)
 	  CatInd[icat]=icat;
    }
  
  int nVec = table->GetRowNumber();
  match.resize(nVec, -1);

  // Try to find starting row of the considered range in the catalog
  // (to make procedure slightly faster)

  double startdec = DecVec->at(DecInd[0]) - cut;

  int icat1=0;
  int icat2=_nRow;

  while(icat2-icat1 > 2)
    {
      int icat12 = (icat1+icat2)/2;
      double dec12 = CatDec->at(CatInd[icat12]);

      if(dec12 > startdec)
	  icat2 = icat12;
      else
	  icat1 = icat12;

    }

  int catFirst = icat1;


  for(int ivec = 0; ivec<nVec; ivec++)
    {
    int irow = DecInd[ivec];
    double dec=DecVec->at(irow);
    double ra = RaVec->at(irow);

    int imatch = -1;
    double dmin = cut;

    double decmin = dec - cut;
    double decmax = dec + cut;

    if(cut == 0.)
      {
      decmin=-90.;
      decmax=90.;
      dmin=180.;
      }

    for(int icat=catFirst; icat< _nRow; icat++)
      {
	int icatrow = CatInd[icat];
	double cdec = CatDec->at(icatrow);

      if(cdec < decmin)
	{
	  catFirst++;
	  continue;
	}

      if(cdec > decmax) break;

      double cra = CatRa->at(icatrow); 

      double dist = SkyDistance(ra,dec,cra,cdec);

      if(dist < dmin)
	{
	  dmin=dist;
	  imatch = icatrow; 
	}
      }

    if(imatch >= 0)
      {
      match[irow] = imatch;

      for(int icopy = 0 ; icopy < nCopy; icopy++)
        {
        string oldname = CopyNames.at(icopy);
        string newname = NewNames.at(icopy);
        table->GetColumn(newname)->CopyCellValue(GetColumn(oldname) ,imatch, irow);
        }
      }
    }


  delete [] DecInd;
  delete [] CatInd;

  return match;
}




// Calculate distance between sky objects

double GloriaSkyCatalog::SkyDistance(double ra1, double dec1, double ra2, double dec2)
{
 const double PI  = 3.141592653589793238462;
 const double deg2rad = PI/180.;
 const double rad2deg = 180./PI;

 double val = cos(deg2rad*dec1)*cos(deg2rad*dec2)*cos(deg2rad*(ra1-ra2))
            + sin(deg2rad*dec1)*sin(deg2rad*dec2);

 return rad2deg*acos(val);
}


} // namespace gloria
