// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

// Author: A.F.Zarnecki, University of Warsaw <mailto:zarnecki@fuw.edu.pl>

/*
 *   This source code is part of the LUIZA software package for GLORIA.
 *   You are free to use this source files for your own development as
 *   long as it stays in a public research context. You are not
 *   allowed to use it for commercial purpose. You must put this
 *   header with author names in all development based on this file.
 *
 */

#include "gloria/GloriaValueMap.h"
#include <algorithm>

namespace gloria{

using namespace std;
  
// Default constructor

GloriaValueMap::GloriaValueMap()
  {
    ClearMaps();
  }


void GloriaValueMap::ClearMaps()
  {
  _nMapX = _nMapY = _nMapXY = 0;
  _sizeX = _sizeY = _sizeXY = 0;
  _binX = _binY = _binXY = 0;

  for(int imap=0; imap<_valueMaps.size(); imap++)
      delete [] _valueMaps.at(imap);

  _valueMaps.clear();
  _valueNames.clear();
  }

// Constructor allocating map arrays

GloriaValueMap::GloriaValueMap(int Nbx, int Nby, int Nmap)
  {
    InitializeMaps(Nbx, Nby, Nmap);
  }

// Constructor including image size

GloriaValueMap::GloriaValueMap(const GloriaFitsImage* fitsImage, int Nbx, int Nby, int Nmap)
  {
    InitializeMaps(Nbx, Nby, Nmap);

    SetImageSize(fitsImage);
  }

// Constructor based on table file

GloriaValueMap::GloriaValueMap(GloriaFitsTable* mapTable)
  {
    ClearMaps();
    LoadFromTable(mapTable);
  }


// Initialize maps

void GloriaValueMap::InitializeMaps(int Nbx, int Nby, int Nmap)
  {
  _nMapX = Nbx;
  _nMapY = Nby;
  _nMapXY = Nbx * Nby;

  for(int imap=0; imap<_valueMaps.size(); imap++)
      delete [] _valueMaps.at(imap);

  _valueMaps.clear();
  _valueNames.clear();
 
  for(int imap=0; imap<Nmap; imap++)
    {
    double * valMap = new double[_nMapXY];
    stringstream valName;
    valName << "Val" << imap ;

    _valueMaps.push_back(valMap);
    _valueNames.push_back(valName.str());
    }

  _sizeX = _sizeY =  _sizeXY = 0;
  _binX = _binY = _binXY = 0;

  }


void GloriaValueMap::AddMap(string name)
  {
 double * tmpMap = new double[_nMapXY];
 _valueMaps.push_back(tmpMap);
 _valueNames.push_back(name);
  }


void GloriaValueMap::AddMap()
  {
 double * valMap = new double[_nMapXY];
 stringstream valName;
 valName << "Val" << _valueMaps.size() ;

 _valueMaps.push_back(valMap);
 _valueNames.push_back(valName.str());
 }


void GloriaValueMap::SetImageSize(int SizeX, int SizeY)
  {
  _sizeX = SizeX;
  _sizeY = SizeY;
  _sizeXY = _sizeX * _sizeY;

  _binX = ( _sizeX + _nMapX -1 )/_nMapX;
  _binY = ( _sizeY + _nMapY -1 )/_nMapY;
  _binXY = _binX * _binY;
  }

void GloriaValueMap::SetImageSize(const GloriaFitsImage* fitsImage)
  {
  _sizeX = fitsImage->SizeX();
  _sizeY = fitsImage->SizeY();
  _sizeXY = _sizeX * _sizeY;

  _binX = ( _sizeX + _nMapX -1 )/_nMapX;
  _binY = ( _sizeY + _nMapY -1 )/_nMapY;
  _binXY = _binX * _binY;
  }

// Default destructor

GloriaValueMap::~GloriaValueMap()
  {
    for(int imap=0; imap<_valueMaps.size(); imap++)
      delete [] _valueMaps.at(imap);
  }


// Return table containing all defined maps as columns
GloriaFitsTable* GloriaValueMap::GetTable()
  {
  GloriaFitsTable* mapTable = new GloriaFitsTable;

  // Set header information

  mapTable->AddComment("GloriaValueMap table created within Luiza framework");

  mapTable->SetKeywordInt("IMGSIZEX", _sizeX,"Image size in X");
  mapTable->SetKeywordInt("IMGSIZEY", _sizeY,"Image size in Y");

  mapTable->SetKeywordInt("MAPSIZEX", _nMapX,"Number of map bins in X");
  mapTable->SetKeywordInt("MAPSIZEY", _nMapY,"Number of map bins in Y");

  mapTable->SetKeywordInt("VALMAPS", _valueMaps.size(),"Number of maps stored");
  
  // Define and fill columns for stored maps
                            
  for(int imap=0; imap<_valueMaps.size(); imap++)
      mapTable->AddColumn(_valueNames.at(imap),DoubleColumn);

  for(int ibin=0; ibin<_nMapXY ; ibin++)
    {
    mapTable->AddRow();

    for(int imap=0; imap<_valueMaps.size(); imap++)
         mapTable->SetValue(imap, ibin, _valueMaps.at(imap)[ibin]);
    }  

  return mapTable;
  }


// Initialize value map structures from the table
bool GloriaValueMap::LoadFromTable(GloriaFitsTable* mapTable)
  {
    // check if input table is not empty

    if(mapTable->GetColumnNumber() == 0 ) return false;
    if(mapTable->GetRowNumber() == 0 ) return false;

    // If map structure is empty (not initialized), proper keywords
    // need to be present in the table header

    if(_nMapXY == 0 || _sizeXY == 0)
      {
	if(! mapTable->IsKeywordSet("VALMAPS") ) return false;
	if(! mapTable->IsKeywordSet("IMGSIZEX") ) return false;
	if(! mapTable->IsKeywordSet("IMGSIZEY") ) return false;
	if(! mapTable->IsKeywordSet("MAPSIZEX") ) return false;
	if(! mapTable->IsKeywordSet("MAPSIZEY") ) return false;
      }

    // Read header information if available

   if( mapTable->IsKeywordSet("MAPSIZEX") )
	_nMapX = mapTable->GetKeywordValueInt("MAPSIZEX");

   if( mapTable->IsKeywordSet("MAPSIZEY") )
	_nMapY = mapTable->GetKeywordValueInt("MAPSIZEY");

   _nMapXY = _nMapX * _nMapY;
  
   // Check if dimensions match

   if(mapTable->GetRowNumber() != _nMapXY) return false;


   if( mapTable->IsKeywordSet("IMGSIZEX") )
	_sizeX = mapTable->GetKeywordValueInt("IMGSIZEX");

   if( mapTable->IsKeywordSet("IMGSIZEY") )
	_sizeY = mapTable->GetKeywordValueInt("IMGSIZEY");

   _sizeXY = _sizeX * _sizeY;

   _binX = ( _sizeX + _nMapX -1 )/_nMapX;
   _binY = ( _sizeY + _nMapY -1 )/_nMapY;
   _binXY = _binX * _binY;
      

   // Remove old contents (if any)

   for(int imap=0; imap<_valueMaps.size(); imap++)
      delete [] _valueMaps.at(imap);

   _valueMaps.clear();
   _valueNames.clear();


   for(int imap=0; imap < mapTable->GetColumnNumber(); imap++)
     {
      double * valMap = new double[_nMapXY];

      for(int ibin=0; ibin<_nMapXY ; ibin++)
	valMap[ibin] = mapTable->GetValueOf<double>(imap,ibin);

     _valueMaps.push_back(valMap);
     _valueNames.push_back(mapTable->GetColumnName(imap));
     }


   // Success

   return true;
  }

// Get mean of map values
double GloriaValueMap::GetMapMean(int imap)
  {
  double val=0.;
  double *valMap = _valueMaps.at(imap);

  for(int ibin=0;ibin<_nMapXY;ibin++)
         val+=valMap[ibin];

  return val/_nMapXY;
  }

// Get map value for pixel ipix  using given map
double GloriaValueMap::GetMapValue(int imap, long ipix, int ipol)
  {
    int ix = ipix%_sizeX;
    int iy = ipix/_sizeX;

    return GetMapValue(imap, (double)ix + 1., (double)iy + 1., ipol);
  }


// Get background level for pixel ix, iy  using given map
double GloriaValueMap::GetMapValue(int imap, int ix, int iy, int ipol)
{
  return  GetMapValue(imap, (double)ix + 1., (double)iy + 1., ipol);
}

// Get background level for pixel ix, iy  using given map
double GloriaValueMap::GetMapValue(int imap, double x, double y, int ipol)
{
  // Return zero for wrong map index

  if(imap<0 || imap>=_valueMaps.size()) return 0.;

  // Also, if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return 0.;

  // Image corner corresponds to x=0.5, y=0.5
  // Center of pixel (0,0) corresponds to x=1., y=1. (!)
  // Return zero value for coordinates outside image

  if(x < 0.5 || y < 0.5 || x >= _sizeX+0.5 || y >= _sizeY+0.5 ) return 0.;

  // Pixel coordinates

  int ix = (int)(x-0.5);
  int iy = (int)(y-0.5);

  if(ipol==0 || _nMapX==1 || _nMapY==1 )  // No interpolation
    {
      int ibx = ix/_binX;
      int iby = iy/_binY;

      return _valueMaps.at(imap)[ibx + iby*_nMapX];
    }


  if(ipol == 1)  // Linear interpolation
      {
	double* vmap = _valueMaps.at(imap);

	int ibx0 =  (2*ix - _binX +1)/(2*_binX);
        if(ibx0 == _nMapX - 1) ibx0--;

	int iby0 =  (2*iy - _binY +1)/(2*_binY);
        if(iby0 == _nMapY - 1) iby0--;

	int ib0 = ibx0 + iby0 * _nMapX;

	double x0=  (ibx0 + 0.5)*_binX + 0.5;
	double y0=  (iby0 + 0.5)*_binY + 0.5;

	double fx = (x - x0)/_binX;
	double fy = (y - y0)/_binY;

	double val = vmap[ib0]*(1-fx)*(1-fy)    +  vmap[ib0+1]*fx*(1-fy) +
	             vmap[ib0+_nMapX]*(1-fx)*fy +  vmap[ib0+_nMapX+1]*fx*fy;

	return val; 
      }


  if(ipol == 2)  // Quadratic interpolation
      {
	double* vmap = _valueMaps.at(imap);

	int ibx0 =  (2*ix - _binX +1)/(2*_binX);
        if(ibx0 == _nMapX - 1) ibx0--;

	int iby0 =  (2*iy - _binY +1)/(2*_binY);
        if(iby0 == _nMapY - 1) iby0--;

	double x0=  (ibx0 + 0.5)*_binX + 0.5;
	double y0=  (iby0 + 0.5)*_binY + 0.5;

	double fx = (x - x0)/_binX;
	double fy = (y - y0)/_binY;


	// Now we have to calculate coefficients of quadratic fit. 
	// Simplified method: do fit in X first, fit parameter
	// dependence on Y next

	// Special cases - map edge

	bool left = ( ibx0 == 0 );
	bool right = ( ibx0 == _nMapX - 2 );
	bool bottom = ( iby0 == 0 );
	bool top = ( iby0 == _nMapY - 2 );


	double a[4],b[4],c[4];

	for(int irow=-1; irow < 3 ; irow++)
	  {
	    if(irow == -1 && bottom) continue;
	    if(irow == 2 && top) continue;

            int ib0 = ibx0 + (iby0+irow) * _nMapX;

  	    c[irow+1] = vmap[ib0];

	    if(left)
	      {
		a[irow+1] = 0.5*(vmap[ib0] - 2.*vmap[ib0+1] + vmap[ib0+2]);
		b[irow+1] = vmap[ib0+1] - a[irow+1] - c[irow+1];
	      }
	    else if(right)
	      {
		b[irow+1] = 0.5*(vmap[ib0+1] - vmap[ib0-1]);
		a[irow+1] = vmap[ib0+1] - b[irow+1] - c[irow+1];
	      }
	    else
	      {
                b[irow+1] = 1.25*vmap[ib0+1] - 0.75*vmap[ib0] 
                          - 0.25*vmap[ib0-1] - 0.25*vmap[ib0+2];

		a[irow+1] = vmap[ib0+1] - b[irow+1] - c[irow+1];
	      }
	  }

// values at reference pixel

        double ca = a[1];  
        double cb = b[1];  
        double cc = c[1]; 
 
        double aa,ab,ac,ba,bb,bc;

	if(bottom)
	  {
	    aa = 0.5*(a[1] - 2.*a[2] + a[3]);
            ba = a[2] - aa - ca;

	    ab = 0.5*(b[1] - 2.*b[2] + b[3]);
            bb = b[2] - ab - cb;

	    ac = 0.5*(c[1] - 2.*c[2] + c[3]);
            bc = c[2] - ac - cc;
	  }
	else if(top)
	  {
	    ba = 0.5*(a[2]-a[0]);
	    aa = a[2] - ba - ca;

	    bb = 0.5*(b[2]-b[0]);
	    ab = b[2] - bb - cb;

	    bc = 0.5*(c[2]-c[0]);
	    ac = c[2] - bc - cc;
	  }
	else
	  {
	    ba = 1.25*a[2] - 0.75*a[1] - 0.25*a[0] - 0.25*a[3];
	    aa = a[2] - ba - ca;


	    bb = 1.25*b[2] - 0.75*b[1] - 0.25*b[0] - 0.25*b[3];
	    ab = b[2] - bb - cb;


	    bc = 1.25*c[2] - 0.75*c[1] - 0.25*c[0] - 0.25*c[3];
	    ac = c[2] - bc - cc;
	  }


	double val = (aa*fy*fy + ba*fy + ca) *fx*fx 
                   + (ab*fy*fy + bb*fy + cb) *fx 
	           + (ac*fy*fy + bc*fy + cc);

	return val; 
      }

  // Higher orders not implemented !

  return 0.;

}


// Return image filled with interpolated values from map
GloriaFitsImage*  GloriaValueMap::GetMapImage(int imap, int ipol)
{
  // If imap is invalid
  if( imap<0 || imap >= _valueMaps.size() ) return NULL;

  // Also, if map is no initialized properly
  if(_nMapXY==0 || _sizeXY==0 || _binXY==0) return NULL;

  GloriaFitsImage*  mapImage = new  GloriaFitsImage(_sizeX, _sizeY, true);  // new float image

  if(ipol==0 || _nMapX==1 || _nMapY==1) FillMapImage0(_valueMaps.at(imap),mapImage);

  if(ipol==1) FillMapImage1(_valueMaps.at(imap),mapImage);

  if(ipol==2) FillMapImage2(_valueMaps.at(imap),mapImage);

return mapImage;
}


// Filled image without interpolation (just bin values)
void GloriaValueMap::FillMapImage0(double* vmap, GloriaFitsImage* mapImage)
{
 double* image = mapImage->GetFloatImageData();

 for(int ipix=0; ipix< _sizeXY; ipix++)
   {
   int ibx = (ipix%_sizeX)/_binX;
   int iby = (ipix/_sizeX)/_binY;

   image[ipix] = vmap[ibx + iby*_nMapX];
   }

 return;
}


// Filled image with linear interpolation (1st order polinomial)
void GloriaValueMap::FillMapImage1(double* vmap, GloriaFitsImage* mapImage)
{
 double* image = mapImage->GetFloatImageData();

 for(int iy=0;iy<_sizeY;iy++)
   {
   int iby0 =  (2*iy - _binY +1)/(2*_binY);
   if(iby0 == _nMapY - 1) iby0--;

   double y0=  (iby0 + 0.5)*_binY - 0.5;
   double fy = (iy - y0)/_binY;
   

   int ibx0 = 0;
   double x0 =  (ibx0 + 0.5)*_binX - 0.5;
   double fx = -x0/_binX;

   int ib0 = iby0 * _nMapX;
   double dfx = 1.0/_binX;

   for(int ix=0;ix<_sizeX;ix++)
      {
        //! Does not work for even _binX (!?)
	//	if((2*ix - _binX +1)%(2*_binX) == 0)

	if((2*ix - _binX +1) - ibx0*(2*_binX) >= (2*_binX) )
	  {
          ibx0 =  (2*ix - _binX +1)/(2*_binX);
          if(ibx0 == _nMapX - 1) ibx0--;
	  ib0 = ibx0 + iby0 * _nMapX;

	  x0=  (ibx0 + 0.5)*_binX - 0.5;
	  fx = (ix - x0)/_binX;
	  }

      image[ix + _sizeX*iy] = vmap[ib0]*(1-fx)*(1-fy)    +  vmap[ib0+1]*fx*(1-fy) +
	                      vmap[ib0+_nMapX]*(1-fx)*fy +  vmap[ib0+_nMapX+1]*fx*fy;

      fx+= dfx;
      }

   }

 return;
}


// Filled image with quadratic interpolation (2nd order polinomial)
void GloriaValueMap::FillMapImage2(double* vmap, GloriaFitsImage* mapImage)
{
 double* image = mapImage->GetFloatImageData();

 int ibx0_old = -1;
 int iby0_old = -1;

 double a[4],b[4],c[4];
 double aa,ab,ac,ba,bb,bc,ca,cb,cc;


 for(int iy=0;iy<_sizeY;iy++)
   {
   int iby0 =  (2*iy - _binY +1)/(2*_binY);
   if(iby0 == _nMapY - 1) iby0--;

   double y0=  (iby0 + 0.5)*_binY + 0.5;
   double fy = (iy + 1. - y0)/_binY;
   
   for(int ix=0;ix<_sizeX;ix++)
      {
      int ibx0 =  (2*ix - _binX +1)/(2*_binX);
      if(ibx0 == _nMapX - 1) ibx0--;

      double x0=  (ibx0 + 0.5)*_binX + 0.5;
      double fx = (ix + 1. - x0)/_binX;

      if(ibx0 != ibx0_old || iby0 != iby0_old)
	{
	// Now we have to calculate coefficients of quadratic fit. 
	// Simplified method: do fit in X first, fit parameter
	// dependence on Y next

	// Special cases - map edge

	bool left = ( ibx0 == 0 );
	bool right = ( ibx0 == _nMapX - 2 );
	bool bottom = ( iby0 == 0 );
	bool top = ( iby0 == _nMapY - 2 );


	for(int irow=-1; irow < 3 ; irow++)
	  {
	    if(irow == -1 && bottom) continue;
	    if(irow == 2 && top) continue;

            int ib0 = ibx0 + (iby0+irow) * _nMapX;

  	    c[irow+1] = vmap[ib0];

	    if(left)
	      {
		a[irow+1] = 0.5*(vmap[ib0] - 2.*vmap[ib0+1] + vmap[ib0+2]);
		b[irow+1] = vmap[ib0+1] - a[irow+1] - c[irow+1];
	      }
	    else if(right)
	      {
		b[irow+1] = 0.5*(vmap[ib0+1] - vmap[ib0-1]);
		a[irow+1] = vmap[ib0+1] - b[irow+1] - c[irow+1];
	      }
	    else
	      {
                b[irow+1] = 1.25*vmap[ib0+1] - 0.75*vmap[ib0] 
                          - 0.25*vmap[ib0-1] - 0.25*vmap[ib0+2];

		a[irow+1] = vmap[ib0+1] - b[irow+1] - c[irow+1];
	      }
	  }

// values at reference pixel

        ca = a[1];  
        cb = b[1];  
        cc = c[1];  

	if(bottom)
	  {
	    aa = 0.5*(a[1] - 2.*a[2] + a[3]);
            ba = a[2] - aa - ca;

	    ab = 0.5*(b[1] - 2.*b[2] + b[3]);
            bb = b[2] - ab - cb;

	    ac = 0.5*(c[1] - 2.*c[2] + c[3]);
            bc = c[2] - ac - cc;
	  }
	else if(top)
	  {
	    ba = 0.5*(a[2]-a[0]);
	    aa = a[2] - ba - ca;

	    bb = 0.5*(b[2]-b[0]);
	    ab = b[2] - bb - cb;

	    bc = 0.5*(c[2]-c[0]);
	    ac = c[2] - bc - cc;
	  }
	else
	  {
	    ba = 1.25*a[2] - 0.75*a[1] - 0.25*a[0] - 0.25*a[3];
	    aa = a[2] - ba - ca;


	    bb = 1.25*b[2] - 0.75*b[1] - 0.25*b[0] - 0.25*b[3];
	    ab = b[2] - bb - cb;


	    bc = 1.25*c[2] - 0.75*c[1] - 0.25*c[0] - 0.25*c[3];
	    ac = c[2] - bc - cc;
	  }

	}

	double val = (aa*fy*fy + ba*fy + ca) *fx*fx 
                   + (ab*fy*fy + bb*fy + cb) *fx 
	           + (ac*fy*fy + bc*fy + cc);



	image[ix + _sizeX*iy] = val;
      }

   }

 return;
}


  
}// namespace gloria
