#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "luiza/FitsImageReader.h"

namespace luiza {

  FitsImageReader aFitsImageReader ;


  // Constructor

  FitsImageReader::FitsImageReader() : DataSourceProcessor("FitsImageReader") {

    _description = "Processor for reading input FITS images"
                   "Reads images from given files" ;


  std::vector< std::string > emptyStringVec;

  registerProcessorParameter ("FitsFileList",
                              "List of FITS files to be read",
                              _fitsFileList,  emptyStringVec );


  registerProcessorParameter ("ListFileList",
                              "Names of files containing FITS file lists (one per line)",
                              _listFileList,  emptyStringVec );


  // Although this processor reads images from file it creates image collection as an output
  // This is why we register this collection as output collection
  registerOutputCollection (  ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection to which images from file should be stored",
                              _imageCollectionName,  std::string ("FitsImages") );


  // Two remaining parameters are optional - can be skiped in the steering file

  registerOptionalParameter ("PermanentCollection",
                              "Flag for collections, which should not be deleted after loop is finished",
                              _permanentCollection,  false );



  registerOptionalParameter ("ImagesPerLoop",
                              "Number of images to be read per processing loop (0 for all))",
                              _imagesPerLoop,  static_cast < int > (1) );


  }


  FitsImageReader*  FitsImageReader::newProcessor() {
    return new FitsImageReader ;
  }
  

  void FitsImageReader::init() { 

    streamlog_out( MESSAGE ) << "FitsImageReader::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    
    _nFiles = 0 ;
    _nImages = 0 ;
    _nLoops = 0 ;
 
    // Check if input file name given

    if(_fitsFileList.size()==0 && _listFileList.size()==0)
      {
	streamlog_out( ERROR ) << "No input files defined for FitsImageReader! " << std::endl;
	streamlog_out( ERROR ) << "Input NOT initialized! " << std::endl;

	//        throw StopProcessingException( this ) ;
	return;

      }

  // Open list files and copy file names to the fits file list

    for(int ilist=0; ilist < _listFileList.size(); ilist++)
      {
      std::ifstream listFile;
      listFile.open(_listFileList[ilist].c_str(),std::ios::in);

      if(listFile.fail())
	{
	  streamlog_out( ERROR ) << "Error opening input list file: " 
                                 << _listFileList[ilist] <<std::endl;
          continue;
	}

      std::string fitsFile;

      while(1)
	{
	  listFile >> fitsFile;

          if(!listFile.good())break;

	  _fitsFileList.push_back(fitsFile);  

	}

      listFile.close();
      }

  // As there are file names specified, we are ready to start processing

    _isInputInitialized = true;

    streamlog_out( MESSAGE ) << "Ready to fill "<< _imageCollectionName << " image collection" << std::endl;
    streamlog_out( MESSAGE ) << "Input file list has " << _fitsFileList.size() << " entries" <<std::endl;

 // Print list of files in DEBUG mode

    for(int ifile=0;ifile<_fitsFileList.size();ifile++)
	 streamlog_out( DEBUG )  <<  "      File " <<  std::setw ( 5 ) << ifile+1 << " : " 
                                 << _fitsFileList[ifile] <<std::endl;

    // Set return value to false (no input yet)
    
    setReturnValue( false ) ;
    
 }

  void FitsImageReader::processData( GloriaDataContainer* data ) { 

     bool readOK = false;
    
    // Set return value to false (no input)
    
    setReturnValue( false ) ;
    
   // Return without processing if no files left to read

    if(_nFiles >= _fitsFileList.size() ) return;


    // Count only loops where something was done...

    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    
    // Create output collection if not found

    if(! data -> ImageCollectionExists(_imageCollectionName))
         data -> CreateImageCollection(_imageCollectionName,_permanentCollection);

    // Read _imagesPerLoop in each call to processData()
    // If _imagesPerLoop is set to 0: all images are read at first call (!)

    for(int ifile=0; (ifile<_imagesPerLoop || _imagesPerLoop==0) && _nFiles<_fitsFileList.size() ;ifile++)
      {
      // Take next file name

      std::string fileName = _fitsFileList[_nFiles];

      _nFiles++;

      // When this is the last file, set the 'close' flag
    
      if(_nFiles == _fitsFileList.size()) _isInputClosed=true;


      // Create empty image

      GloriaFitsImage *myfit = new GloriaFitsImage;

      // read image data from file 

      myfit->ReadFitsFile(fileName);

      // Remove image if not properly read 
      if(myfit->Error())
         {
	 streamlog_out(ERROR) << "Error reading fits file " << fileName << std::endl ;
         streamlog_out(ERROR) << myfit->ErrorMessage() << std::endl ;
         continue;
         }

      // Add image to the collection

      (data->GetImageCollection(_imageCollectionName)).push_back(*myfit);
      delete myfit;

      readOK = true;
     _nImages++;
     }


    // return true if image read
    if(readOK) setReturnValue( true ) ;
    
  }
  
  void FitsImageReader::check(GloriaDataContainer* data) { 
    
    if(! data -> ImageCollectionExists(_imageCollectionName))
      {
         streamlog_out(ERROR) << " check(): Collection not found !? " << std::endl ;
	 return;
      }


    int nImg = data->GetImageCollection(_imageCollectionName).size();

    streamlog_out(MESSAGE) << " check(): Collection " << _imageCollectionName << " with " << nImg << " images" << std::endl ;
 
  }
  
  void FitsImageReader::end(){ 
    
    printEndMessage() ;

  }

  void FitsImageReader::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;

    streamlog_out(MESSAGE) << " processed "     << _nFiles << " files " << std::endl 
			   << std::endl ;
    
    streamlog_out(MESSAGE) << " read "     << _nImages << " images " << std::endl 
			   << std::endl ;
  }

  
}// namespace luiza
