#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "luiza/FitsImageWriter.h"

namespace luiza {

  FitsImageWriter aFitsImageWriter ;


  // Constructor

  FitsImageWriter::FitsImageWriter() : Processor("FitsImageWriter") {

    _description = "Processor for writing FITS images to disk"
                   "Write images to specified files" ;


  registerInputCollection   ( ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection which should be stored to file",
                              _imageCollectionName,  std::string ("FitsImages") );

  registerProcessorParameter ("FitsFileNameRoot",
                              "Output file name root, which will be appended by the image number",
                              _fitsFileNameRoot,  std::string("output/luiza_file_") );

  registerOptionalParameter ("BitsPerPixel",
"Bits per pixel in the output fits file (0 to keep image format unchanged)",
			     _bitPix,  int( 0 ) );

  registerOptionalParameter ("PixelOffset",
" BZERO keyword value, used when storing pixel values to fits file (default is 0, use 32768 for unsigned short)",
			     _bZero,  int( 0 ) );

  registerOptionalParameter ("PixelScale",
" BSCALE keyword value, used when storing pixel values to fits file (default is 1.0)",
			     _bScale,  double( 1.0 ) );

  }


  FitsImageWriter*  FitsImageWriter::newProcessor() {
    return new FitsImageWriter;
  }


  void FitsImageWriter::init() {

    streamlog_out( MESSAGE ) << "FitsImageWriter::init()  " << name()
			     << endl
			     << "  parameters: " << endl
			     << *parameters()
			     << endl ;

    _nLoops = 0 ;

    _nImages = 0 ;

    _nWritten = 0 ;

 }

  void FitsImageWriter::startAnalysis( GloriaDataContainer* data )
  {
	streamlog_out(MESSAGE) << "Start analysis "  << endl ;
  }

  void FitsImageWriter::processData( GloriaDataContainer* data )
  {
	_nLoops ++ ;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << endl ;


        if(! data -> ImageCollectionExists(_imageCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }


	// Skip permanent collection - will be written at conlcudeAnalysis


       if( data -> ImageCollectionPermanent(_imageCollectionName)) return;


       processImages(data);

 }


  void FitsImageWriter::concludeAnalysis( GloriaDataContainer* data )
  {
	streamlog_out(MESSAGE) << "Concluding analysis " << endl ;


        if(! data -> ImageCollectionExists(_imageCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }

	if(! data -> ImageCollectionPermanent(_imageCollectionName)) return;

	streamlog_out(DEBUG) << "Writing images from permanent collection " << _imageCollectionName << endl ;

        processImages(data);

  }


  void FitsImageWriter::processImages( GloriaDataContainer* data )
    {

      int nImg = data -> GetImageCollection(_imageCollectionName).size();

      if(nImg == 0)
	  {
          streamlog_out(DEBUG) << "No images in input collection ! " << endl ;

          return;
	  }
      
        streamlog_out(DEBUG) << "  Collection " << _imageCollectionName << " with " << nImg << " images " << endl ;


        for(int ifm=0; ifm<nImg; ifm++)
	  {
	   GloriaFitsImage *frame =  &(data -> GetImageCollection(_imageCollectionName).at(ifm));  

          _nImages++;


	  // Construct output file name from given root, image number and .fit extension

          stringstream outname;

          outname << _fitsFileNameRoot;


	  // If the file name root ends with '/' append it with input file name

          int fnlen=_fitsFileNameRoot.size();

          if( _fitsFileNameRoot.at(fnlen-1) == '/' && frame->InputFileName().size() > 0 )
             {
              int flen = frame->InputFileName().size();

	      int ilpos = frame->InputFileName().rfind('.');

	      if (ilpos != std::string::npos) flen = ilpos;

	      int ifpos = frame->InputFileName().rfind('/');

              if (ifpos == std::string::npos)
		outname << frame->InputFileName().substr(0,flen) << ".fit" ;
	      else
		outname << frame->InputFileName().substr(ifpos+1,flen-ifpos-1) << ".fit"; 
	     }
	  else
	     {
              outname.fill('0'); 
              outname << setw(4) << _nImages;

              outname << ".fit" ;
	     }

          frame->WriteFitsFile(outname.str(),_bitPix,_bZero,_bScale);


	  if(frame->Error())
		{
		streamlog_out(ERROR) << "Error writing fits file " << outname.str() << endl ;
		streamlog_out(ERROR) << frame->ErrorMessage() << endl ;
		continue;
		}
		
	  streamlog_out(DEBUG) << "  Image (" << _bitPix << ") written to file: " << outname.str() << endl ;

          _nWritten++;
	  }
  

    // always return true  for ProcessorName
    setReturnValue( true ) ;

  }

  void FitsImageWriter::check(GloriaDataContainer* data)
  {
	    streamlog_out(DEBUG) << " check() " << endl;
  }

  void FitsImageWriter::end()
  {
	  printEndMessage() ;

  }

  void FitsImageWriter::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "
			   << " after " << _nLoops << " calls to processData() " << endl ;

    streamlog_out(MESSAGE) << _nImages << " images processed " << endl;

    streamlog_out(MESSAGE) << _nWritten << " images written to output file " << endl
			   << endl ;
  }


}// namespace luiza
