#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "luiza/FitsTableReader.h"

namespace luiza {

  FitsTableReader aFitsTableReader ;


  // Constructor

  FitsTableReader::FitsTableReader() : DataSourceProcessor("FitsTableReader") {

    _description = "Processor for reading input FITS images"
                   "Reads images from given files" ;


  std::vector< std::string > emptyStringVec;

  registerProcessorParameter ("TableFileList",
                              "List of text table files to be read",
                              _tableFileList,  emptyStringVec );


  registerProcessorParameter ("ListFileList",
                              "Names of files containing list of table file names (one per line)",
                              _listFileList,  emptyStringVec );


  // Although this processor reads tables from file it creates table collection as an output
  // This is why we register this collection as output collection

  registerOutputCollection (  TableCollection,
                              "TableCollectionName",
                              "Name of the image collection to which images from file should be stored",
                              _tableCollectionName,  std::string ("TextTable") );

  

  // Remaining parameters are optional - can be skiped in the steering file


  registerOptionalParameter ("PermanentCollection",
                              "Flag for collections, which should not be deleted after loop is finished",
                              _permanentCollection,  false );

  registerOptionalParameter ("TablesPerLoop",
                              "Number of table files to be read per processing loop (0 for all))",
                              _tablesPerLoop,  static_cast < int > (1) );


  }


  FitsTableReader*  FitsTableReader::newProcessor() {
    return new FitsTableReader ;
  }
  

  void FitsTableReader::init() { 

    streamlog_out( MESSAGE ) << "FitsTableReader::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    
    _nFiles = 0 ;
    _nTables = 0 ;
    _nLoops = 0 ;

    _isInputInitialized = false;

    // Check if input file name given

    if(_tableFileList.size()==0 && _listFileList.size()==0)
      {
	streamlog_out( ERROR ) << "No input files defined for FitsTableReader! " << std::endl;
	streamlog_out( ERROR ) << "Input NOT initialized! " << std::endl;

	//        throw StopProcessingException( this ) ;
	return;

      }

  // Open list files and copy file names to the fits file list

    for(int ilist=0; ilist < _listFileList.size(); ilist++)
      {
      std::ifstream listFile;
      listFile.open(_listFileList[ilist].c_str(),std::ios::in);

      if(listFile.fail())
	{
	  streamlog_out( ERROR ) << "Error opening input list file: " 
                                 << _listFileList[ilist] <<std::endl;
          continue;
	}

      std::string fitsFile;

      while(1)
	{
	  listFile >> fitsFile;

          if(!listFile.good())break;

	  _tableFileList.push_back(fitsFile);  

	}

      listFile.close();
      }


 // Construct empty table, as defined by user

    _inputTable = new GloriaFitsTable;

 // As there are file names specified, we are ready to start processing

    _isInputInitialized = true;

    streamlog_out( MESSAGE ) << "Ready to fill "<< _tableCollectionName << " table collection" << std::endl;
    streamlog_out( MESSAGE ) << "Input file list has " << _tableFileList.size() << " entries" <<std::endl;

 // Print list of files in DEBUG mode

    for(int ifile=0;ifile<_tableFileList.size();ifile++)
	 streamlog_out( DEBUG )  <<  "      File " <<  std::setw ( 5 ) << ifile+1 << " : " 
                                 << _tableFileList[ifile] <<std::endl;

 
    // Set return value to false (no input yet)
    
    setReturnValue( false ) ;
    
}

  void FitsTableReader::processData( GloriaDataContainer* data ) { 

    bool readOK = false;
    
    // Set return value to false (no input)
    
    setReturnValue( false ) ;
    
   // Return without processing if no files left to read

    if(_nFiles >= _tableFileList.size() ) return;


    // Count only loops where something was done...

    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    
    // Create output collection if not found

    if(! data -> TableCollectionExists(_tableCollectionName))
         data -> CreateTableCollection(_tableCollectionName,_permanentCollection);

    // Read _tablesPerLoop in each call to processData()
    // If _tablesPerLoop is set to 0: all files are read at first call (!)

    for(int ifile=0; (ifile<_tablesPerLoop || _tablesPerLoop==0) && _nFiles<_tableFileList.size() ;ifile++)
      {
      // Take next file name

      std::string fileName = _tableFileList[_nFiles];

      _nFiles++;

      // When this is the last file, set the 'close' flag
    
      if(_nFiles == _tableFileList.size()) _isInputClosed=true;


      // Clear input table (removing table structure as well)

      _inputTable->EmptyTable();

      // read  data from file 

      _inputTable->ReadFitsFile(fileName);

      // Skip table if not properly read 
      if(_inputTable-> Error())
         {
	 streamlog_out(ERROR) << "Error reading table from fits file " << fileName << std::endl ;
	 streamlog_out(ERROR) << _inputTable-> ErrorMessage() << fileName << std::endl ;
         continue;
         }

      // Add table to the collection

      (data->GetTableCollection(_tableCollectionName)).push_back(*_inputTable);

       readOK = true;
      _nTables++;
      }


    // return true if table read
    if(readOK) setReturnValue( true ) ;
    
  }
  
  void FitsTableReader::check(GloriaDataContainer* data) { 
    
    if(! data -> TableCollectionExists(_tableCollectionName))
      {
         streamlog_out(ERROR) << " check(): Collection not found !? " << std::endl ;
	 return;
      }


    int nTbl = data->GetTableCollection(_tableCollectionName).size();

    streamlog_out(MESSAGE) << " check(): Collection " << _tableCollectionName << " with " << nTbl << " tables" << std::endl ;
 
  }
  
  void FitsTableReader::end(){ 
    
    printEndMessage() ;

  }

  void FitsTableReader::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;

    streamlog_out(MESSAGE) << " processed "     << _nFiles << " files " << std::endl 
			   << std::endl ;
    
    streamlog_out(MESSAGE) << " read "     << _nTables << " tables " << std::endl 
			   << std::endl ;
  }

  
}// namespace luiza
