#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "luiza/FitsTableWriter.h"

namespace luiza {

  FitsTableWriter aFitsTableWriter ;


  // Constructor

  FitsTableWriter::FitsTableWriter() : Processor("FitsTableWriter") {

    _description = "Processor for writing FITS tables to disk"
                   "Write tables to specified ASCII or binary files" ;


  registerInputCollection   ( TableCollection,
                              "TableCollectionName",
                              "Name of the table collection which should be stored to file",
                              _tableCollectionName,  std::string ("Objects") );

  registerProcessorParameter ("FitsFileNameRoot",
                              "Output file name root, which will be appended by the table number",
                              _fitsFileNameRoot,  std::string("output/luiza_table_") );

  registerOptionalParameter ("BinaryTable",
                             "Flag to select binary table format",
			     _binaryTable,  true );

  std::vector< std::string > emptyStringVec;

  registerOptionalParameter ("SelectedColumns",
                             "List of table columns to be stored (empty for all)",
			     _selectedColumns, emptyStringVec );
  }


  FitsTableWriter*  FitsTableWriter::newProcessor() {
    return new FitsTableWriter;
  }


  void FitsTableWriter::init() {

    streamlog_out( MESSAGE ) << "FitsTableWriter::init()  " << name()
			     << endl
			     << "  parameters: " << endl
			     << *parameters()
			     << endl ;
    _outputTable = NULL;

    _nLoops = 0 ;

    _nTables = 0 ;

    _nWritten = 0 ;

 }

  void FitsTableWriter::startAnalysis( GloriaDataContainer* data )
  {
   streamlog_out(MESSAGE) << "Starting data processing" << endl ;
  }


  void FitsTableWriter::processData( GloriaDataContainer* data )
  {
	_nLoops ++ ;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << endl ;

        if(! data -> TableCollectionExists(_tableCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      

        if( data -> TableCollectionPermanent(_tableCollectionName))
	  {
         if(_isFirstEvent)
             streamlog_out(MESSAGE) << "Permanent collection - skiped ! " << endl ;

          return;
	  }

	// This is not very nice, but effective
	// Use concludeAnalysis() method also to process 'normal' data 

	concludeAnalysis( data );

}


 void FitsTableWriter::concludeAnalysis( GloriaDataContainer* data )
  {
       int nTab = data -> GetTableCollection(_tableCollectionName).size();


        if(data -> TableCollectionPermanent(_tableCollectionName) && nTab == 0)
	  {
          streamlog_out(DEBUG) << "No tables in permanent collection ! " << endl ;

          return;
	  }
      

        streamlog_out(DEBUG) << "  Collection " << _tableCollectionName << " with " << nTab << " tables " << endl ;


        for(int itb=0; itb<nTab; itb++)
	  {
	   GloriaFitsTable *source =  &(data -> GetTableCollection(_tableCollectionName).at(itb));  

          _nTables++;

           
	  // If list of columns is empty, all columns will be stored,
	  // as found in the processed table 

	  bool is_temporary;

          if(_selectedColumns.size() == 0)
	    {
	      _outputTable = source; 
	      is_temporary = false;
	    }
	  else
	    {
             // Create output table 

             _outputTable = new GloriaFitsTable;
	      is_temporary = true;

	     _outputTable->CopyHeader(source);

   	     // Copy existing column
             // If column is missing, empty column is inserted

             for(int icol=0; icol< _selectedColumns.size();icol++)
	        {
  	        if(source->ColumnExists(_selectedColumns.at(icol)))
	            _outputTable->InsertColumn(_selectedColumns.at(icol), 
                            (source->GetColumn(_selectedColumns.at(icol)))->Clone());
	        else
	            _outputTable->AddColumn(_selectedColumns.at(icol),long(0));
	        }

	     // This filles constant/empty columns

	     _outputTable->AdjustColumns();
	    }


	  // Construct output file name from given root, table number and .fit extension

          stringstream outname;

          outname << _fitsFileNameRoot;

	  // If the file name root ends with '/' append it with input file name (if present)

          int fnlen=_fitsFileNameRoot.size();

          if( _fitsFileNameRoot.at(fnlen-1) == '/' && source->InputFileName().size() > 0 )
             {
              int flen = source->InputFileName().size();

	      int ilpos = source->InputFileName().rfind('.');

	      if (ilpos != std::string::npos) flen = ilpos;

	      int ifpos = source->InputFileName().rfind('/');

              if (ifpos == std::string::npos)
		outname << source->InputFileName().substr(0,flen) ;
	      else
		outname << source->InputFileName().substr(ifpos+1,flen-ifpos-1) ; 
	     }
	  else
          if(!data->TableCollectionPermanent(_tableCollectionName) || nTab!=1)
	      {
              outname.fill('0'); 
              outname << setw(4) << _nTables;
	      }

          outname << ".tbl" ;

          _outputTable->WriteFitsFile(outname.str(),_binaryTable);

	  if(_outputTable->Error())
		{
		streamlog_out(ERROR) << "Error writing table to output file " << endl ;
		streamlog_out(ERROR) << _outputTable->ErrorMessage() << endl ;

                if(is_temporary)delete _outputTable;

		return;
		}
		

          _nWritten++;

          if(is_temporary)delete _outputTable;
	  }
  

    // always return true  for ProcessorName
    setReturnValue( true ) ;

  }

  void FitsTableWriter::check(GloriaDataContainer* data)
  {
	    streamlog_out(DEBUG) << " check() " << endl;
  }

  void FitsTableWriter::end()
  {
	  printEndMessage() ;

  }

  void FitsTableWriter::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "
			   << " after " << _nLoops << " calls to processData() " << endl ;

    streamlog_out(MESSAGE) << _nTables << " tables processed " << endl;

    streamlog_out(MESSAGE) << _nWritten << " tables written to output file " << endl
			   << endl ;
  }


}// namespace luiza
