#include "luiza/Processor.h"
#include "luiza/ProcessorMgr.h" 
#include "luiza/Global.h"
#include "luiza/VerbosityLevels.h"
#include "gloria/GloriaNetworkMgr.h"

using namespace gloria ;

namespace luiza{

  // set default verbosity level to MESSAGE
  //int Processor::Verbosity=Processor::MESSAGE;

  Processor::Processor(const std::string& typeName) :
    _description(" description not set by author ") ,
    _typeName( typeName ) ,
    _parameters(0) ,
    _isFirstEvent( true ),
    _logLevelName(""),
    _str(0) {
  
    //register processor in map
    ProcessorMgr::instance()->registerProcessor( this ) ;


    registerOptionalParameter( "Verbosity" , 
			       "verbosity level of this processor (\"DEBUG,MESSAGE,WARNING,ERROR,SILENT\")"  ,
			       _logLevelName ,  
			       std::string("MESSAGE") ) ;
  
  }


  Processor::Processor(){}

  Processor::~Processor() {

    if( _parameters != 0 ){
      delete _parameters ;
    }

    if( _str !=0 )
      delete _str ;
  
    typedef ProcParamMap::iterator PMI ;
  
    for( PMI i = _map.begin() ; i != _map.end() ; i ++ ) {
    
      //     streamlog_out(DEBUG) << " deleting processor parameter " << i->second->name() 
      // 			 << " of processor [" << name() << "]" 
      // 			 << std::endl ;

      delete i->second ;
    }
  }


  void Processor::setParameters( StringParameters* parameters) {

    if( _parameters != 0 ){
      delete _parameters ;
      _parameters = 0 ;
    }
    _parameters = parameters ;

    updateParameters();

    // need to reset the log level name in case it has not been set by the user 
    if( !parameterSet("Verbosity") ){
      _logLevelName = "" ;
    }
  
    //  streamlog_out( MESSAGE)  << " setParameters :  Verbosity : " << _parameters->getStringVal("Verbosity") 
    // 			  << " parameterSet : " << parameterSet("Verbosity") 
    // 			  << " log level name : " << _logLevelName << std::endl ; 
    
  }
  
  std::stringstream& Processor::log() const {
    
    if( _str !=0 )
      delete _str ;

    _str = new std::stringstream ;

    return *_str ;
  }



  void Processor::printParameters() { printParameters<MESSAGE>() ;  }

  void Processor::printDescription() {
    
    std::cout << "ProcessorType: "   <<  type() << std::endl ;
    
    std::cout << "  Description: " << description() << std::endl ;
    
    typedef ProcParamMap::iterator PMI ;
    
    std::cout << std::endl  
     	      << "  Processor parameters:" << std::endl << std::endl ;
    
    for( PMI i = _map.begin() ; i != _map.end() ; i ++ ) {
      
      ProcessorParameter* p = i->second ;

      if( p->isOptional() )    {
	  std::cout << "    Optional: " << p->name()   << std::endl ;
  	
      }else{

	std::cout << "    Required: " << p->name() << std::endl ;
	

      }


      std::cout << "      Contains: "  << p->description() << std::endl ;

      std::cout << "      Type: " << " [" <<  p->type() << "]" << endl ; 
      std::cout << "      Default: " << p->defaultValue() << std::endl 
		  << std::endl ;
    }

    std::cout << std::endl 
	      << std::endl ;


  }

  void Processor::printDescriptionXML(std::ostream& stream) {

    if(stream == std::cout){
      stream << " <processor name=\"My" <<  type()  << "\"" 
	     << " type=\"" <<  type() << "\">" 
	     << std::endl ;
    }
    else{
      stream << " <processor name=\"" <<  name()  << "\"" 
	     << " type=\"" <<  type() << "\">" 
	     << std::endl ;

    }
    
    stream << " <!--" << description() << "-->" << std::endl ;
    
    typedef ProcParamMap::iterator PMI ;
    
    for( PMI i = _map.begin() ; i != _map.end() ; i ++ ) {
      
      ProcessorParameter* p = i->second ;

      stream << "  <!--" << p->description() << "-->" << std::endl ;

      if( p->isOptional() ) {
	stream << "  <!--parameter name=\"" << p->name() << "\" " 
	       << "type=\"" << p->type() ;

	if ( isInputCollectionName( p->name() ) )
	  stream << "\" lcioInType=\"" << _inTypeMap[ p->name() ]  ;

	if ( isOutputCollectionName( p->name() ) )
	  stream << "\" lcioOutType=\"" << _outTypeMap[ p->name() ]  ;

	stream << "\">"
	       << p->defaultValue() 
	       << " </parameter-->"
	       << std::endl ;
      } else {
	stream << "  <parameter name=\"" << p->name() << "\" " 
	       << "type=\"" << p->type() ;

	if ( isInputCollectionName( p->name() ) )
	  stream << "\" lcioInType=\"" << _inTypeMap[ p->name() ]  ;

	if ( isOutputCollectionName( p->name() ) )
	  stream << "\" lcioOutType=\"" << _outTypeMap[ p->name() ]  ;

	stream  << "\">"
		<< p->defaultValue() 
		<< " </parameter>"
		<< std::endl ;
      }
    }
    
    stream << "</processor>" 
	   << std::endl 
	   << std::endl ;
    
  }

  //   ProcessorParameter* Processor::getProcessorParameter( const std::string name) {
  //     ProcParamMap::iterator it = _map.find(name) ;
  //     if( it != _map.end() )
  //       return it->second ;    
  //     else
  //       return 0 ;
  //   }

  bool Processor::parameterSet( const std::string& name ) {

    ProcParamMap::iterator it = _map.find(name) ;

    if( it != _map.end() )
      return it->second->valueSet() ;    
    else
      return false ;
  }
  
  void Processor::baseInit() {
    
    //fg: now in setParameters 
    // updateParameters();

    init() ;
  }
  
  void Processor::updateParameters() {

    typedef ProcParamMap::iterator PMI ;

    for( PMI i = _map.begin() ; i != _map.end() ; i ++ ) {

      i->second->setValue( _parameters ) ;
    }
  }

  void Processor::updateTelescopeParameters(GloriaFitsHeader* header) 
  {
  if (Global::Network == NULL)
      {
        streamlog_out( MESSAGE ) << "  No network configuration available to read parameters " << std::endl ;
	return;
      }

  string instrument, telescope;

  if(header->IsKeywordSet("INSTRUME"))
      {
      instrument = header->GetKeywordValue("INSTRUME");

      if( Global::Network->IsTelescopeDefined(instrument) )
         {
           streamlog_out( MESSAGE ) << "Updating processor parameters based on network configuration" << std::endl ;
           streamlog_out( MESSAGE ) << "Instrument name used: " << instrument << std::endl ;
	   updateTelescopeParameters(instrument);
	   return;
         }
      }


  if(header->IsKeywordSet("TELESCOP"))
      {
      telescope = header->GetKeywordValue("TELESCOP");

      if( Global::Network->IsTelescopeDefined(telescope) )
         {
           streamlog_out( MESSAGE ) << "Updating processor parameters based on network configuration" << std::endl ;
           streamlog_out( MESSAGE ) << "Telescope name used: " << telescope << std::endl ;
	   updateTelescopeParameters(telescope);
	   return;
         }
      }

  if(telescope.size())
    {
     streamlog_out( MESSAGE ) << "Telescope " << telescope << " not found in network configuration" << std::endl ;
    }
  else
     streamlog_out( MESSAGE ) << "TELESCOP keyword not found in FITS header" << std::endl ;

  if(instrument.size())
    {
     streamlog_out( MESSAGE ) << "Instrument " << instrument << " not found in network configuration" << std::endl ;
    }
  else
     streamlog_out( MESSAGE ) << "INSTRUME keyword name not found in FITS header" << std::endl ;


  streamlog_out( MESSAGE ) << "Reseting to user/default parameters" << std::endl ;

  updateParameters();

  return;

  }


  void Processor::updateTelescopeParameters(std::string name ) 
  {
   
  if (Global::Network == NULL)
      {
        streamlog_out( MESSAGE ) << "  No network configuration available to read parameters " << std::endl ;
	return;
      }

  if( Global::Network->IsTelescopeDefined(name) == false )
      {
        streamlog_out( MESSAGE ) << "  Telescope " << name 
               << " not defined in network configuration (" << Global::Network->GetID() << ")" << std::endl ;
	return;
      }

    GloriaTelescope *telescope = Global::Network->GetTelescope(name);

    // Need a local copy of parameters

    StringParameters *TelPar =  telescope->parameters();

    // Parameters have to be copied into the local structure to
    // modify parameter names: remove processor type if found
    // This allows to remove ambiguity, if same parameter name 
    // is used in different processors

    StringParameters NewPar;

    for(ParametersMap::iterator m = TelPar->_map.begin() ; m != TelPar->_map.end() ; m++)
        {
	string key = m->first; 

	if(key.find(_typeName+":")==0)
  	       key = key.substr(_typeName.length()+1);

	if(key.find(_processorName+":")==0)
  	       key = key.substr(_processorName.length()+1);

	NewPar.add(key, m->second);        
      }

    // Debug

    streamlog_out( DEBUG ) << "Telescope  parameters from configuration file: "  << std::endl << *TelPar << std::endl ;


    streamlog_out( DEBUG ) << "Telescope  parameters for parameter update: "  << std::endl << NewPar << std::endl ;


    typedef ProcParamMap::iterator PMI ;

    for( PMI i = _map.begin() ; i != _map.end() ; i ++ ) {

      i->second->setValue( &NewPar ) ;
    }
  }




  void Processor::setReturnValue( bool val) {
    
    ProcessorMgr::instance()->setProcessorReturnValue(  this , val ) ;
  }
  
  void Processor::setGloriaInType(const std::string& collectionName,  const  GloriaCollectionType& gloriaInType) {
    _inTypeMap[ collectionName ] = gloriaInType ;
  }

  GloriaCollectionType Processor::getGloriaInType( const std::string& colName ) {

    if( isInputCollectionName( colName )  )
      return _inTypeMap[ colName ] ;
    else
      return UndefinedCollection ;
  }


   GloriaCollectionType Processor::getGloriaOutType( const std::string& colName ) {

    if( isOutputCollectionName( colName )  )
      return _outTypeMap[ colName ] ;
    else
      return UndefinedCollection ;
  }


  bool Processor::isInputCollectionName( const std::string& pName  ) {
    return _inTypeMap.find( pName  ) != _inTypeMap.end() ;
  }
  
  void Processor::setGloriaOutType(const std::string& collectionName,  const  GloriaCollectionType& gloriaOutType) {
    _outTypeMap[ collectionName ] = gloriaOutType ;
  }

  bool Processor::isOutputCollectionName( const std::string& pName  ) {
    return _outTypeMap.find( pName ) != _outTypeMap.end() ;
  }

  void Processor::setReturnValue( const std::string& name, bool val ){
  
    ProcessorMgr::instance()->setProcessorReturnValue(  this , val , name ) ;
  }


} // namespace luiza

