#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "luiza/TableTextReader.h"

namespace luiza {

  TableTextReader aTableTextReader ;


  // Constructor

  TableTextReader::TableTextReader() : DataSourceProcessor("TableTextReader") {

    _description = "Processor for reading input FITS images"
                   "Reads images from given files" ;


  std::vector< std::string > emptyStringVec;

  registerProcessorParameter ("TableFileList",
                              "List of text table files to be read",
                              _tableFileList,  emptyStringVec );


  registerProcessorParameter ("ListFileList",
                              "Names of files containing list of table file names (one per line)",
                              _listFileList,  emptyStringVec );


  registerProcessorParameter ("ColumnNames",
                              "Names which will be given to data columns read from input file",
                              _columnNames,  emptyStringVec );


  registerProcessorParameter ("ColumnTypes",
                              "Types of data columns in the input table file",
                              _columnTypes,  emptyStringVec );



  // Although this processor reads tables from file it creates table collection as an output
  // This is why we register this collection as output collection

  registerOutputCollection (  TableCollection,
                              "TableCollectionName",
                              "Name of the image collection to which images from file should be stored",
                              _tableCollectionName,  std::string ("TextTable") );

  

  // Remaining parameters are optional - can be skiped in the steering file

  registerOptionalParameter ("ColumnUnits",
                              "Units of data columns in the input table file",
                              _columnUnits,  emptyStringVec );


  registerOptionalParameter ("ColumnDefaults",
                              "Default values for empty data  in the input table file",
                              _columnDefaults,  emptyStringVec );


  registerOptionalParameter ("PermanentCollection",
                              "Flag for collections, which should not be deleted after loop is finished",
                              _permanentCollection,  false );

  registerOptionalParameter ("TablesPerLoop",
                              "Number of table files to be read per processing loop (0 for all))",
                              _tablesPerLoop,  static_cast < int > (1) );


  }


  TableTextReader*  TableTextReader::newProcessor() {
    return new TableTextReader ;
  }
  

  void TableTextReader::init() { 

    streamlog_out( MESSAGE ) << "TableTextReader::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    
    _nFiles = 0 ;
    _nTables = 0 ;
    _nLoops = 0 ;

    _inputTable = NULL;
    _isInputInitialized = false;

    // Check if input file name given

    if(_tableFileList.size()==0 && _listFileList.size()==0)
      {
	streamlog_out( ERROR ) << "No input files defined for TableTextReader! " << std::endl;
	streamlog_out( ERROR ) << "Input NOT initialized! " << std::endl;

	//        throw StopProcessingException( this ) ;
	return;

      }

  // Open list files and copy file names to the fits file list

    for(int ilist=0; ilist < _listFileList.size(); ilist++)
      {
      std::ifstream listFile;
      listFile.open(_listFileList[ilist].c_str(),std::ios::in);

      if(listFile.fail())
	{
	  streamlog_out( ERROR ) << "Error opening input list file: " 
                                 << _listFileList[ilist] <<std::endl;
          continue;
	}

      std::string fitsFile;

      while(1)
	{
	  listFile >> fitsFile;

          if(!listFile.good())break;

	  _tableFileList.push_back(fitsFile);  

	}

      listFile.close();
      }

    // Check column names and types - size should be the same

    if(_columnNames.size()==0 && _columnNames.size()!=_columnTypes.size() )
      {
	streamlog_out( ERROR ) << "Error in column name/type definition !" << std::endl;
	streamlog_out( ERROR ) << "Input NOT initialized! " << std::endl;

	//        throw StopProcessingException( this ) ;
	return;
      }


    // Construct empty table, as defined by user

    _inputTable = new GloriaFitsTable;

    for(int icol=0; icol<_columnNames.size(); icol++)
      {
        string name = _columnNames[icol];
        TableColumnType type = UndefinedColumn;


        string unit = "";
        if(_columnUnits.size() == _columnNames.size())
 	       unit = _columnUnits[icol];                

        string comment = "";

	double ddefault = 0.;
	long ldefault = 0;
        string sdefault = "";
        
        const char *tstring = _columnTypes[icol].c_str();

        switch (tstring[0])
	{
	case 'i':
	case 'I':
	case 'l':
	case 'L':
	  type = LongColumn;
          if(_columnDefaults.size() == _columnNames.size())
 	       ldefault = atoi(_columnDefaults[icol].c_str());                

          _inputTable->AddColumn(name,ldefault,unit);

	  break;


	case 'f':
	case 'F':
	case 'd':
	case 'D':
	  type = DoubleColumn;
          if(_columnDefaults.size() == _columnNames.size())
	    ddefault = atof(_columnDefaults[icol].c_str());                

          _inputTable->AddColumn(name,ddefault,unit);

	  break;


	case 's':
	case 'S':
	case 't':
	case 'T':
	  type = StringColumn;
          if(_columnDefaults.size() == _columnNames.size())
 	       sdefault = _columnDefaults[icol];                

          _inputTable->AddColumn(name,sdefault,unit);

	  break;

	default:
	  streamlog_out( ERROR ) << "Unrecognized column type !" << std::endl;
	  streamlog_out( ERROR ) << "Input NOT initialized! " << std::endl;
  	  //        throw StopProcessingException( this ) ;

          delete _inputTable;
	  return;

	}

      }



  // As there are file names specified, we are ready to start processing

    _isInputInitialized = true;

    streamlog_out( MESSAGE ) << "Ready to fill "<< _tableCollectionName << " table collection" << std::endl;
    streamlog_out( MESSAGE ) << "Input file list has " << _tableFileList.size() << " entries" <<std::endl;

 // Print list of files in DEBUG mode

    for(int ifile=0;ifile<_tableFileList.size();ifile++)
	 streamlog_out( DEBUG )  <<  "      File " <<  std::setw ( 5 ) << ifile+1 << " : " 
                                 << _tableFileList[ifile] <<std::endl;

    // Set return value to false (no input yet)
    
    setReturnValue( false ) ;
    
}

  void TableTextReader::processData( GloriaDataContainer* data ) { 

 
    bool readOK = false;
    
    // Set return value to false (no input)
    
    setReturnValue( false ) ;
    
    // Return without processing if no files left to read

    if(_nFiles >= _tableFileList.size() ) return;


    // Count only loops where something was done...

    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    
    // Create output collection if not found

    if(! data -> TableCollectionExists(_tableCollectionName))
         data -> CreateTableCollection(_tableCollectionName,_permanentCollection);

    // Read _tablesPerLoop in each call to processData()
    // If _tablesPerLoop is set to 0: all files are read at first call (!)

    for(int ifile=0; (ifile<_tablesPerLoop || _tablesPerLoop==0) && _nFiles<_tableFileList.size() ;ifile++)
      {
      // Take next file name

      std::string fileName = _tableFileList[_nFiles];

      _nFiles++;

      // When this is the last file, set the 'close' flag
    
      if(_nFiles == _tableFileList.size()) _isInputClosed=true;


      // Clear input table (structure remains unchanged)

      _inputTable->ClearTable();

      // read  data from file 

      _inputTable->ReadTextFile(fileName);

      // Skip table if not properly read 
      if(_inputTable->GetRowNumber() == 0)
         {
	 streamlog_out(ERROR) << "Error reading input file " << fileName << std::endl ;
         continue;
         }

      // Add table to the collection

      (data->GetTableCollection(_tableCollectionName)).push_back(*_inputTable);

       readOK = true;
      _nTables++;
      }


    // return true if table read
    if(readOK) setReturnValue( true ) ;
    
  }
  
  void TableTextReader::check(GloriaDataContainer* data) { 
    
    if(! data -> TableCollectionExists(_tableCollectionName))
      {
         streamlog_out(ERROR) << " check(): Collection not found !? " << std::endl ;
	 return;
      }


    int nTbl = data->GetTableCollection(_tableCollectionName).size();

    streamlog_out(MESSAGE) << " check(): Collection " << _tableCollectionName << " with " << nTbl << " tables" << std::endl ;
 
  }
  
  void TableTextReader::end(){ 
    
    printEndMessage() ;

  }

  void TableTextReader::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;

    streamlog_out(MESSAGE) << " processed "     << _nFiles << " files " << std::endl 
			   << std::endl ;
    
    streamlog_out(MESSAGE) << " read "     << _nTables << " tables " << std::endl 
			   << std::endl ;
  }

  
}// namespace luiza
