#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "luiza/TableTextWriter.h"

namespace luiza {

  TableTextWriter aTableTextWriter ;


  // Constructor

  TableTextWriter::TableTextWriter() : Processor("TableTextWriter") {

    _description = "Processor for writing FITS images to disk"
                   "Write images to specified files" ;


  registerInputCollection   ( TableCollection,
                              "TableCollectionName",
                              "Name of the table collection which should be stored to file",
                              _tableCollectionName,  std::string ("Objects") );

  registerProcessorParameter ("TextFileNameRoot",
                              "Output file name root, which will be appended by the image number",
                              _textFileNameRoot,  std::string("output/luiza_table_") );

  registerOptionalParameter ("OneFile",
                             "Flag  to force storing all tables in one big file",
			     _oneFile,  false );

  registerOptionalParameter ("AddCounter",
                             "Flag to add row number as the first column",
			     _addCounter,  false );

  std::vector< std::string > emptyStringVec;

  registerOptionalParameter ("SelectedColumns",
                             "List of table columns to be stored (empty for all)",
			     _selectedColumns, emptyStringVec );

  registerOptionalParameter ("HeaderRecords",
                             "List of header records to be added as a column",
			     _headerRecords, emptyStringVec );
  }


  TableTextWriter*  TableTextWriter::newProcessor() {
    return new TableTextWriter;
  }


  void TableTextWriter::init() {

    streamlog_out( MESSAGE ) << "TableTextWriter::init()  " << name()
			     << endl
			     << "  parameters: " << endl
			     << *parameters()
			     << endl ;
    _outputTable = NULL;

    _nLoops = 0 ;

    _nTables = 0 ;

    _nWritten = 0 ;

 }

  void TableTextWriter::startAnalysis( GloriaDataContainer* data )
  {
   streamlog_out(MESSAGE) << "Starting data processing data" << endl ;

   if(!_oneFile)return;
	    
   stringstream outname;

   outname << _textFileNameRoot <<  ".tbl" ;

   _outputStream.open(outname.str().c_str(),std::ios::out);

   if(_outputStream.fail())
         {
         streamlog_out(ERROR) << "Error opening output file:  " << outname.str().c_str() << endl ;

         return;
         }

   // throw StopProcessingException( this );
  }


  void TableTextWriter::processData( GloriaDataContainer* data )
  {
	_nLoops ++ ;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << endl ;

        if(! data -> TableCollectionExists(_tableCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      

        if( data -> TableCollectionPermanent(_tableCollectionName))
	  {
         if(_isFirstEvent)
             streamlog_out(MESSAGE) << "Permanent collection - skiped ! " << endl ;

          return;
	  }

	// This is not very nice, but effective
	// Use concludeAnalysis() method also to process 'normal' data 

	concludeAnalysis( data );

}


 void TableTextWriter::concludeAnalysis( GloriaDataContainer* data )
  {
       int nTab = data -> GetTableCollection(_tableCollectionName).size();


        if(data -> TableCollectionPermanent(_tableCollectionName) && nTab == 0)
	  {
          streamlog_out(DEBUG) << "No tables in permanent collection ! " << endl ;

          return;
	  }
      

        streamlog_out(DEBUG) << "  Collection " << _tableCollectionName << " with " << nTab << " tables " << endl ;


        for(int itb=0; itb<nTab; itb++)
	  {
	   GloriaFitsTable *source =  &(data -> GetTableCollection(_tableCollectionName).at(itb));  

          _nTables++;

           
          // Create output table 

         _outputTable = new GloriaFitsTable;

	  // Counters

          if( _addCounter && _oneFile )
 	          _outputTable->AddColumn("I_table", (long)_nTables);


          if( _addCounter )
 	          _outputTable->AddColumn("I_row", (long)0);

	  for(int irec=0; irec< _headerRecords.size(); irec++)
	    {
	      string keyword = _headerRecords.at(irec);
              string value = source->GetKeywordValue(keyword.c_str());

               streamlog_out(DEBUG) << "  Adding header column " << keyword << " with value of " << value << endl ;

              _outputTable->AddColumn(keyword,value);
	    }

	  // If list of columns is empty, all columns will be stored,
	  // as found in the first processed table (!)

          if(_selectedColumns.size() == 0)
 	      for(int icol=0; icol < source->GetColumnNumber(); icol++)
 		   _selectedColumns.push_back(source->GetColumnName(icol));

	  // Copy existing column
          // If column is missing, empty column is inserted

          for(int icol=0; icol< _selectedColumns.size();icol++)
	    {
  	    if(source->ColumnExists(_selectedColumns.at(icol)))
	        _outputTable->InsertColumn(_selectedColumns.at(icol), 
                          (source->GetColumn(_selectedColumns.at(icol)))->Clone());
	    else
	        _outputTable->AddColumn(_selectedColumns.at(icol),long(0));
	    }

	  // This filles constant/empty columns

	  _outputTable->AdjustColumns();
	    

          if( _addCounter )
             for(int irow=0;irow< _outputTable->GetRowNumber();irow++)
	         _outputTable->SetValue("I_row",irow,(long)irow);


	  // Construct output file name from given root, table number and .fit extension

	  if(!_oneFile)
	    {
            stringstream outname;

            outname << _textFileNameRoot;

	  // If the file name root ends with '/' append it with input file name (if present)

            int fnlen=_textFileNameRoot.size();

            if( _textFileNameRoot.at(fnlen-1) == '/' && source->InputFileName().size() > 0 )
             {
              int flen = source->InputFileName().size();

	      int ilpos = source->InputFileName().rfind('.');

	      if (ilpos != std::string::npos) flen = ilpos;

	      int ifpos = source->InputFileName().rfind('/');

              if (ifpos == std::string::npos)
		outname << source->InputFileName().substr(0,flen) ;
	      else
		outname << source->InputFileName().substr(ifpos+1,flen-ifpos-1) ; 
	     }
	    else
            if(!data->TableCollectionPermanent(_tableCollectionName) || nTab!=1)
	      {
              outname.fill('0'); 
              outname << setw(4) << _nTables;
	      }
           
            outname << ".tbl" ;

            _outputStream.open(outname.str().c_str(),std::ios::out);

            if(_outputStream.fail())
               {
 	       streamlog_out(ERROR) << "Error opening output file:  " << outname.str() << endl ;

               delete _outputTable;

               return;
               }

            }

          _outputTable->WriteTextStream(_outputStream);

	  if(_outputTable->Error())
		{
		streamlog_out(ERROR) << "Error writing table to output file " << endl ;
		streamlog_out(ERROR) << _outputTable->ErrorMessage() << endl ;

                delete _outputTable;
		if(!_oneFile)_outputStream.close();

		return;
		}
		

          _nWritten++;

          delete _outputTable;
 	  if(!_oneFile)_outputStream.close();
	  }
  

    // always return true  for ProcessorName
    setReturnValue( true ) ;

  }

  void TableTextWriter::check(GloriaDataContainer* data)
  {
	    streamlog_out(DEBUG) << " check() " << endl;
  }

  void TableTextWriter::end()
  {
        if(_oneFile)_outputStream.close();

	  printEndMessage() ;

  }

  void TableTextWriter::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "
			   << " after " << _nLoops << " calls to processData() " << endl ;

    streamlog_out(MESSAGE) << _nTables << " tables processed " << endl;

    streamlog_out(MESSAGE) << _nWritten << " tables written to output file " << endl
			   << endl ;
  }


}// namespace luiza
