#include "luiza/TestProcessor.h"
#include "luiza/Exceptions.h"
#include "luiza/Global.h"
#include <iostream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  TestProcessor aTestProcessor ;


  TestProcessor::TestProcessor() : Processor("TestProcessor") {

    _description = "Simple processor to test the luiza application."
      "Prints the data collection statistics" ;

   registerInputCollection (  ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection to watch, empty string for all collections",
                              _imageCollectionName,  std::string ("") );

   registerInputCollection (  TableCollection,
                              "TableCollectionName",
                              "Name of the table collection to watch, empty string for all collections",
                              _tableCollectionName,  std::string ("") );


 }

  void TestProcessor::init() { 

    streamlog_out( MESSAGE ) << "  TestProcessor::init()" << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nImages = 0 ;
    _nImgCollections = 0 ;
    _nTblCollections =0;
    _nTables =0;
    _nLoops = 0 ;
 
  }


 void TestProcessor::startAnalysis( GloriaDataContainer* data ) { 

    streamlog_out( MESSAGE ) << "  TestProcessor::startAnalysis()" << std::endl ;
    
   // Allocate table for statistics

    _imgTable = new GloriaFitsTable;
    _tblTable = new GloriaFitsTable;

    // Define columns

    _imgTable->AddColumn("Loop",LongColumn);
    _imgTable->AddColumn("Collection",StringColumn);
    _imgTable->AddColumn("Images",LongColumn);
    _imgTable->AddColumn("Size X",LongVectorColumn);
    _imgTable->AddColumn("Size Y",LongVectorColumn);

    _tblTable->AddColumn("Loop",LongColumn);
    _tblTable->AddColumn("Collection",StringColumn);
    _tblTable->AddColumn("Tables",LongColumn);
    _tblTable->AddColumn("Columns",LongVectorColumn);
    _tblTable->AddColumn("Rows",LongVectorColumn);


    // Check if network configuration was loaded

    if (Global::Network != NULL)
      {
	int id = Global::Network->GetID();
	int ntel = Global::Network->GetTelescopeNumber();

        streamlog_out( MESSAGE ) << "  Network configuration " << id 
                                 << " with " << ntel << " telescopes" << std::endl ;
     
      }
    else
        streamlog_out( MESSAGE ) << "  No network configuration available " << std::endl ;
      
  }



  void TestProcessor::processData( GloriaDataContainer* data ) { 

    _nLoops ++ ;

    streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops << std::endl ;

    int nImgCol=0;
    int nTabCol=0;

    if(_imageCollectionName.size()==0 && _tableCollectionName.size()==0)
      {
      nImgCol = data -> GetImageCollectionNumber();
      nTabCol = data -> GetTableCollectionNumber();
      }

    if(_imageCollectionName.size() > 0 && 
         data -> ImageCollectionExists(_imageCollectionName)) nImgCol=1;
      

    if( _tableCollectionName.size() > 0 &&
         data -> TableCollectionExists(_tableCollectionName)) nTabCol=1;
      

    if(nImgCol==0 && nTabCol==0)
      {
	streamlog_out(DEBUG) << "  no data in input collections" << std::endl ;
        return;
      }

    _nImgCollections+=nImgCol;
    _nTblCollections+=nTabCol;

    // Loop over image collections

    for(int icl=0; icl<nImgCol; icl++)
      {
	std::string ColName = 
         (_imageCollectionName.size()==0)?data -> GetImageCollectionName(icl):_imageCollectionName;

        int nImg = data -> GetImageCollection(ColName).size();

        streamlog_out(DEBUG) << "  Collection " << ColName << " with " << nImg << " images " << std::endl ;
        _nImages+=nImg;

	LongVec sizX;
	LongVec sizY;

        for(int ifm=0; ifm<nImg; ifm++)
	  {
	    GloriaFitsImage *frame =  &(data -> GetImageCollection(ColName).at(ifm));  

	    streamlog_out(DEBUG) << "     Image " << ifm+1 << " is " << frame->SizeX() 
                                   << "x" << frame->SizeY() << std::endl ;

	    sizX.push_back(frame->SizeX());
	    sizY.push_back(frame->SizeY());
	  }

	_imgTable->AddRow((long)_nLoops, ColName, (long)nImg, sizX, sizY );
      }

    // Loop over table collections

    for(int icl=0; icl<nTabCol; icl++)
      {
	std::string ColName = 
         (_tableCollectionName.size()==0)?data -> GetTableCollectionName(icl):_tableCollectionName;

        int nTab = data -> GetTableCollection(ColName).size();

        streamlog_out(DEBUG) << "  Collection " << ColName << " with " << nTab << " tables " << std::endl ;
        _nTables+=nTab;

	LongVec Cols;
	LongVec Rows;

        for(int itb=0; itb<nTab; itb++)
	  {
	    GloriaFitsTable *table =  &(data -> GetTableCollection(ColName).at(itb));  

	    streamlog_out(DEBUG) << "     Table " << itb+1 << " has " << table->GetColumnNumber() 
				 << " columns and " << table->GetRowNumber() << " rows" << std::endl ;

	    Cols.push_back(table->GetColumnNumber());
	    Rows.push_back(table->GetRowNumber());
	  }

	_tblTable->AddRow((long)_nLoops, ColName, (long)nTab, Cols, Rows );
      }


    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
  // set ProcessorName.EvenNumberOfEvents == true if this processor has been called 2n (n=0,1,2,...) times 
    setReturnValue("EvenNumberOfEvents",  !( _nLoops % 2 )   ) ;

    
  }
  
  void TestProcessor::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << "   TestProcessor::check() " << std::endl ;
    
  }
  
 void TestProcessor::concludeAnalysis( GloriaDataContainer* data ) { 

    
    streamlog_out( DEBUG ) << "   TestProcessor::concludeAnalysis()  " << std::endl ;
    
    printEndMessage() ;

    delete _imgTable;
    delete _tblTable;
 }


  void TestProcessor::end(){ 

    streamlog_out( DEBUG ) << "   TestProcessor::end()  " << std::endl ;

  }



 void TestProcessor::printEndMessage() const {

    streamlog_out(MESSAGE) << " Closing after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " processed "     << _nImages << " images in " 
			   << _nImgCollections << " collections " 
			   << std::endl ;
    
    streamlog_out(MESSAGE) << " processed "     << _nTables << " tables in " 
			   << _nTblCollections << " collections " << std::endl 
			   << std::endl ;
    
    if(_nImages>0)
      {
      stringstream imgss;

      _imgTable->Print(imgss,20);

      streamlog_out(MESSAGE) << " Image statistics collected during processing " << endl;

      if(_imageCollectionName.size()!=0)
        {
        streamlog_out(MESSAGE) << "   for image collection '" << _imageCollectionName << "':" << endl;
        }
      else
        streamlog_out(MESSAGE) << "   for all image collections found: " << endl;

      streamlog_out(MESSAGE) <<  imgss.str() << endl;
      }

    if(_nTables>0)
      {

      stringstream tabss;

      _tblTable->Print(tabss,20);

      streamlog_out(MESSAGE) << " Table statistics collected during processing " << endl;

      if(_tableCollectionName.size()!=0)
        {
        streamlog_out(MESSAGE) << "   for table collection '" << _tableCollectionName << "':" << endl;
        }
      else
        streamlog_out(MESSAGE) << "   for all table collections found: " << endl;

      streamlog_out(MESSAGE) <<  tabss.str() << endl;
      }

    // test deprecated method message 
    message<DEBUG>( " and this is really the final DEBUG message ....") ;
  }
  
}// namespace luiza
