// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef Astrometry_h
#define Astrometry_h 1

#include "luiza/Processor.h"
#include "GloriaAstrometry.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor implementing Astrometry.net algorithms
  /*! Processor to apply Astrometry.net algorithm for finding image
   *  orientation and transformation for calculating object position
   *  in the sky. As an input to the algorithm
   *  gloria::GloriaObjectList collection should be given. If image
   *  collection is specified as well, calculated astrometry
   *  parameters are added to the image FITS header.
   *  
   *
   *  \par Input 
   *  Processor takes object lists and images from Gloria data collections
   *
   *  \param ListCollectionName -
   *             Name of the collection containing object collection
   *             (corresponding to the object list)
   *
   *  \param ImageCollectionName -
   *             Name of the collection containing images
   *             (corresponding to the object list)
   *
   *  \param  PositionColumns
   *    Names of the object table columns from which object position should
   *    be taken. These are 'CCD_X' and 'CCD_Y' by default (default column
   *    names for Luiza photometry processors), but can be changed by
   *    user. 
   *
   *  \param BrightnessColumn
   *    Name of the object list column from which object brightness should
   *    be taken ('Signal' by default). 
   *
   *  \param MagnitudoGiven
   *    Flag to indicate that the brightness stored in the input table
   *    is already in magnitudo scale (logarythmic). Default is
   *    'false'.
   *
   *  \par Output
   *
   *  Object list with astrometry can be stored in the new collection
   *
   *  \param OutputCollectionName - 
   *            Name of the collection where object table will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   *  \param Overwrite - 
   *      Flag to recalculate and overwrite astrometry, even if
   *      it exists in the file header
   *
   *  \param AddPosition -
   *     Flag to add RA DEC positions from astrometry to the input
   *     object list
   *
   *  \param AddToHeader -
   *     Flag to add astrometry parameters to the object list header
   *
   *  \par Algorithm parameters
   *
   *  \param IndexesPath -
   *    Path to astrometric (astrometry.net) indexes files
   *    This files are required for the algorithm to run.
   *    They can be downloaded from dedicated
   *    <a href="http://broiler.astrometry.net/~dstn/4000/">astrometry.net</a> 
   *    web page.
   *
   *  \param PixscaleHigh  - 
   *      Upper boundary of possible pixel sizes in arcseconds, used
   *      when selecting the index files for analysis.
   *
   *  \param PixscaleLow  -
   *      Lower boundary of possible pixel sizes in arcseconds
   *
   *  \param AstroDepth - 
   *      Depth limit for the astrometry algorithm (default 0)
   *
   *  \param AstroOrder - 
   *      Order of the SIP polynomial in astrometry fit (default 3)
   *
   *  \param MaxObjects - 
   *      Maximum number of objects to be used for astrometry.
   *      Brightest objects are selected.
   *
   *  \param MinDistance - 
   *      Minimum distance (in pixels) between two objects to be used
   *      for astrometry. Brightest objects are selected. This can
   *      help to remove some artefacts.
   *
   *  \param BloomingCut - 
   *      Minimum distance between objects along vertical or
   *      horizontal line. To remove artificial sets of objects along
   *      vertical or horizontal lines which can appear due to
   *      blooming etc.
   *
   *  \param AstVerbosity -
   *      Verbosity level of astrometry.net routines (default 0)
   *
   *  \param InParallel -
   *      Solve using indexes parallely or not (default true)
   */

  class Astrometry : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new Astrometry; }
    
    
    Astrometry();
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init();
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data );
    
    
    virtual void check( GloriaDataContainer* data ); 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end();
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;
    
    //! Return list of astrometry.net indexes proper for the image angular size
    //! Only those indexes physically accessible are returned
    //! Need to be a char **, since astrometry.net code is in plain C
    int GetAstrometricIndexesForImage(char**, GloriaFitsImage*);
    
    //! Name of the image collection containing images to be added
    string _imageCollectionName;
    
    string _listCollectionName;    

    //! Names of the position columns in the object table
    std::vector<std::string> _positionColumns;

    //! Name of the object brightness column
    std::string _brightnessColumn;

    //! Flag for brightness given in magintudo
    bool _magnitudoGiven;

    //! Upper boundary of possible pixel sizes in arcseconds
    double _PixscaleHigh;    
    //! Lower boundary of possible pixel sizes in arcseconds
    double _PixscaleLow;
    
    //! Path to astrometric (astrometry.net) indexes files
    string _Indexes_path;
    
    //! Depth limit for astrometry.net algorithm
    int astrodepth;

    //! Order of the SIP polynomial in astrometry fit
    int astroorder;

    //! Maximum number of objects to be used in astrometry algorithm
    int _maxObjects;

    //! Minimum distance between two objects to be used in astrometry algorithm
    float _minDistance;

    //! Minimum distance between objects along vertical or horizontal line
    float _bloomingCut;

    //! Verbosity of astrometry.net output
    int verbosity;

    //! Solve using indexes parallely or not
    bool inparallel;

    //! Recalculate and overwrite astrometry if exists in a file
    bool overwrite;

    //! Add object positions to the object list table
    bool addposition;


    //! Add astrometry output to the object list table header
    bool addtoheader;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputLists ;

  } ;
  
} // end namespace luiza

extern "C"{
int main1(leblind_struct *lestr);
}

#endif



