#include "Astrometry.h"
#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <sstream>
#include <string.h>
#include <unistd.h>

#include "gloria/GloriaObjectList.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

Astrometry aAstrometry;

Astrometry::Astrometry() :
		Processor("Astrometry")
{

	_description = "Performs astrometry (astrometry.net based) on list of stars on the frame";

	registerInputCollection(ImageCollection, "ImageCollectionName",
			"Name of the collection containing images,  corresponding to the analysed object lists",
			_imageCollectionName, std::string(""));

	registerInputCollection (  TableCollection,
	                           "ListCollectionName",
	                           "Name of the object list collection, to do astrometry on",
	                           _listCollectionName,  std::string ("") );
	
	std::vector<std::string > initColumns;

	initColumns.push_back(std::string("CCD_X"));
	initColumns.push_back(std::string("CCD_Y"));

	registerOptionalParameter(
			"PositionColumns",
			"Names of the table columns containing X and Y positions on the CCD",
			_positionColumns, initColumns);


	registerOptionalParameter(
			"BrightnessColumn",
			"Name of the object brightness column",
			_brightnessColumn, std::string("Signal"));


	registerOptionalParameter(
			"MagnitudoGiven",
			"Flag to indicate that brightness is already given in magnitudo units (logarythmic scale)",
		        _magnitudoGiven, false);


	registerOptionalParameter (
                        "PixscaleHigh", 
                        "Upper boundary of possible pixel sizes in arcseconds", 
                        _PixscaleHigh,  static_cast < double > (0));
	
	registerOptionalParameter (
                        "PixscaleLow", 
                        "Lower boundary of possible pixel sizes in arcseconds", 
                        _PixscaleLow,  static_cast < double > (0));
 	 	
	registerOptionalParameter (
                         "IndexesPath", 
                         "Path to astrometric (astrometry.net) indexes files", 
                         _Indexes_path,  static_cast < string > ("/opt/astrometry/data"));

	registerOptionalParameter (
                         "InParallel",
                         "Solve using indexes parallely or not",
                          inparallel,  static_cast < bool > (true));

	registerOptionalParameter (
                         "AstroDepth", 
                         "Depth limit for the astrometry algorithm (default 0)", 
                          astrodepth,  static_cast < int > (0));

	registerOptionalParameter (
                         "AstroOrder", 
                         "Order of the SIP polynomial in astrometry fit", 
                          astroorder,  static_cast < int > (3));

	registerOptionalParameter (
                         "MaxObjects", 
                         "Maximum number of objects to be used for astrometry.", 
                          _maxObjects,  static_cast < int > (0));

	registerOptionalParameter (
                         "MinDistance", 
                         "Minimum distance between two objects to be used for astrometry.", 
                          _minDistance,  static_cast < float > (0.));

	registerOptionalParameter (
                         "BloomingCut", 
                         "Minimum distance between objects along vertical or horizontal line", 
			 _bloomingCut,  static_cast < float > (0.));

	registerOptionalParameter (
                         "AstVerbosity", 
                         "Verbosity of astrometry.net output", 
                          verbosity,  static_cast < int > (0));

	registerOptionalParameter (
                         "Overwrite", 
                         "Recalculate and overwrite astrometry if exists", 
                          overwrite,  static_cast < bool > (false));

	registerOptionalParameter (
                         "AddPosition", 
                         "Add RA DEC positions from astrometry to the input object list", 
                          addposition,  static_cast < bool > (true));

	registerOptionalParameter (
                         "AddToHeader", 
                         "Add astrometry parameters to the object list table header", 
                          addtoheader,  static_cast < bool > (false));

}

void Astrometry::init()
{

	streamlog_out(MESSAGE) << "Astrometry::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputLists = 0;
	_nLoops = 0;


	// Check input parameters

	if (_listCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No object list collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do!  " << std::endl;

		throw StopProcessingException( this ) ;
	}

	if (_imageCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No image collection name given! " << std::endl;
		streamlog_out(ERROR) << "Astrometry will not be added to the header! " << std::endl;

		//	 throw StopProcessingException( this ) ;
	}

    // Set return value to false (no astrometry yet)
    
  	setReturnValue(false);

}

void Astrometry::processData(GloriaDataContainer* data)
{
        bool anyOK = false;
  	setReturnValue(false);

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_imageCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collections exists

	if (!data->TableCollectionExists(_listCollectionName))
	{
		streamlog_out(DEBUG) << "Input object list collection not found ! " << endl;

		return;
	}

	if (!data->ImageCollectionExists(_imageCollectionName))
	{
		streamlog_out(DEBUG) << "Input image collection not found ! " << endl;

		return;
	}

        int nTab = data->GetTableCollection(_listCollectionName).size();

	if (nTab == 0)
	{
		streamlog_out(DEBUG) << "No tables in input collection ! " << endl;
		return;
	}

	int nImg = data->GetImageCollection(_imageCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;
		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _imageCollectionName << " with "
			<< nImg << " images " << endl;

	streamlog_out(DEBUG) << "  Collection " << _listCollectionName << " with "
			<< nImg << " object lists " << endl;

	if (nImg != nTab)
	{
		streamlog_out(MESSAGE) << "Number of images and object lists missmatch! " << endl;
		return;
	}


	// Loop over images in input collection

	for (int ifm = 0; ifm < nImg; ifm++)
	   {

	   GloriaFitsTable *objects =  &(data -> GetTableCollection(_listCollectionName).at(ifm));
	   GloriaFitsImage *cur_frame =  &(data -> GetImageCollection(_imageCollectionName).at(ifm));

	   GloriaAstrometry ga;

   	   streamlog_out(DEBUG) << "Is astrometry present in the file header: " << ga.IsAstrometryPresent(cur_frame) << endl;
	   if(!overwrite && ga.IsAstrometryPresent(cur_frame))
		{
		streamlog_out(MESSAGE) << "Astrometry found for " << cur_frame->InputFileName() 
                                       << " - will NOT be overwritten." << endl;

		}
	   else
		{
	        streamlog_out(MESSAGE) << "Astrometry will be called for " << cur_frame->InputFileName()  << endl;

		vector<double>* Xvec = &(objects->GetVectorOf<double>(_positionColumns.at(0)));
		vector<double>* Yvec = &(objects->GetVectorOf<double>(_positionColumns.at(1)));
		vector<double>* Svec = &(objects->GetVectorOf<double>(_brightnessColumn));
		int Nobj = Xvec->size();

		char *indexes[512];
		int index_cnt = GetAstrometricIndexesForImage(indexes, cur_frame);
		// If no index was found, astrometry cannot continue
		if(index_cnt==0)
		{
			streamlog_out(ERROR) << "No index found for given frame. Indexes can be downloaded from http://broiler.astrometry.net/~dstn/4000/" << endl;
			// return;
		}
		
		// Allowed range of pixel size in arcseconds
		// If not set, then try to find pixscales in the header

		double pixscale[] = {_PixscaleLow, _PixscaleHigh};

		if(_PixscaleLow==0 && _PixscaleHigh==0)
		{
			if(cur_frame->IsKeywordSet("XPLATE"))
			{
				pixscale[0] = 0.1*cur_frame->GetKeywordValueDouble("XPLATE");
				pixscale[1] = 2*cur_frame->GetKeywordValueDouble("XPLATE");
			}
			else if(cur_frame->IsKeywordSet("PIXSCALE"))
			{
				pixscale[0] = 0.1*cur_frame->GetKeywordValueDouble("PIXSCALE");
				pixscale[1] = 2*cur_frame->GetKeywordValueDouble("PIXSCALE");
			}
		}
		// Else, set the default values of pixscale as in backend.c of astrometry.net sources
		else
		{
			if(_PixscaleLow==0)  pixscale[0]=0.1*3600/cur_frame->SizeX();
			if(_PixscaleHigh==0) pixscale[1]=180*3600./cur_frame->SizeX();
		}
		
		// Centre of the frame

		// THIS IS CURRENTLY NOT USED - overwritten in astrometry, see leblind.c

		double crpix[] = {0.5*cur_frame->SizeX(), 0.5*cur_frame->SizeY()};

                if(cur_frame->IsKeywordSet("CRPIX1"))
		  crpix[0]=cur_frame->GetKeywordValueDouble("CRPIX1");

                if(cur_frame->IsKeywordSet("CRPIX2"))
		  crpix[1]=cur_frame->GetKeywordValueDouble("CRPIX2");

		
		// Frame dimensions, min x, max x, min y, max y
		double frame_dim[] = {0, cur_frame->SizeX(), 0, cur_frame->SizeY()};
		
		// Initialize structure with data sent to astrometric procedures
		leblind_struct lestr;

		//		lestr.solution = (double*)(malloc(100*sizeof(double)));
		lestr.solution = new double[100];

		// This could work, but does not follow STL rules
		//
		//		lestr.X = &Xvec->front();
		//		lestr.Y = &Yvec->front();
		//		lestr.signal = &Svec->front();

		// To avoid this problem, copy STL vectors to C arrays
		// Sort objects according to their brightness

                multimap<double, int , greater<double> > objectListMap;
                multimap<double, int , greater<double> >::iterator  objectListIterator;

                for(int iobj=0; iobj<Nobj; iobj++)
		  {
		    double signal =  Svec->at(iobj);
		    if(_magnitudoGiven)signal = pow(10.,5. - 0.4*signal);
		  objectListMap.insert(make_pair( signal, iobj ) );
		  }
		                  
                               
		// Limit number of objects which will be given to astrometry routine

                if(_maxObjects > 1 && _maxObjects < Nobj) Nobj = _maxObjects;

		// Initialize astrometry input tables

	        lestr.X = new double[Nobj];
		lestr.Y = new double[Nobj];
		lestr.signal = new double[Nobj];

           
                int iobj=0;

		for(objectListIterator=objectListMap.begin();
                    objectListIterator!=objectListMap.end() && iobj<Nobj;objectListIterator++)
		  {
		    int ilist=objectListIterator->second;

		    double xobj = Xvec->at(ilist);
		    double yobj = Yvec->at(ilist);

                    // To remove artificial patterns due to blooming etc
		    // Sets of objects along vertical or horizontal lines
                    
                    bool online = false;

                    if( _bloomingCut > 0.)
		      {
			for(int io2=0; io2<iobj;io2++)
			  {
			    if(fabs(lestr.X[io2]-xobj)<= 1. && 
                               fabs(lestr.Y[io2]-yobj)<= _bloomingCut) online = true;

			    if(fabs(lestr.Y[io2]-yobj)<= 1. && 
                               fabs(lestr.X[io2]-xobj)<= _bloomingCut) online = true;


  		         if(online)
			     {
		             streamlog_out(DEBUG) << "Object removed by blooming cut: " << iobj 
                                         << "  at " << xobj << ", " << yobj << endl;
			     break;
			     }
			  }
		      }

		    if(online)continue;


                    // To remove artificial clusters of objects, eg. from large PSF of bery bright star
                    
                    bool tooclose = false;

                    if( _minDistance > 0.)
		      {
			double distcut =  _minDistance* _minDistance;

			for(int io2=0; io2<iobj;io2++)
			  {
                           double dist = (lestr.X[io2]-xobj)*(lestr.X[io2]-xobj) +
			     (lestr.Y[io2]-yobj)*(lestr.Y[io2]-yobj);

                           if(dist < distcut) tooclose=true;

  		           if(tooclose)
			     {
		             streamlog_out(DEBUG) << "Object removed by distance cut: " << iobj 
                                         << "  at " << xobj << ", " << yobj << endl;
			     break;
			     }
			  }
		      }

		    if(tooclose)continue;

		    lestr.X[iobj] = xobj;
		    lestr.Y[iobj] = yobj;
		    lestr.signal[iobj] = objectListIterator->first;

		    iobj++;
		  }

		streamlog_out(MESSAGE) << "Number of objects in input object list:" << Xvec->size() << endl;
		streamlog_out(MESSAGE) << "Number of objects selected for astrometry:" << iobj << endl;

		lestr.star_cnt = iobj;

		// Order of the SIP polynomial
		lestr.sip_order = astroorder;
		lestr.pixscale = pixscale;
		streamlog_out(DEBUG) << "Pixscale used:" << pixscale[0] << " " << pixscale[1] << endl;

		// Number of stars in index to match with
		// I don't know yet why on bootes frame astrometry does not converge properly with depth<=20 (except 0)
		lestr.depth = astrodepth;
		lestr.crpix = crpix;
		lestr.frame_dim = frame_dim;
		lestr.verbosity = verbosity;
		lestr.index_list = indexes;
		lestr.index_cnt = index_cnt;
		lestr.inparallel = inparallel;
		
		// Call the astrometry!
		streamlog_out(DEBUG) << "About to call main astrometry.net routine..." << endl;
		int is_solved = main1(&lestr);
                streamlog_out(DEBUG) << "Astrometry main routine returns " << is_solved << endl;
		
		// Add history record to the FITS header

		cur_frame->AddComment(string(" ------------- Luiza ------------- "));
                stringstream hiss;
                hiss << "LUIZA " << _processorName << " (" << _typeName << ")";
		cur_frame->AddHistory(hiss.str());
                cur_frame->AddHistory(string("Using astrometry.net algorithm"));

                if( is_solved != 0)
		  {
                  streamlog_out(MESSAGE) << "Astrometry done" << endl;

 		  ga.InitFromLeblind(&lestr);

		  // Write the astrometry results to the header of the frame

		  cur_frame->SetKeywordInt("ASTROOK", 1, string("Astrometry converged"));

	          double RA_h= ga.crval[0]/15.;
		      if(RA_h < 0. ) RA_h+=24;

	          double DEC_deg= ga.crval[1];

		  cur_frame->SetKeywordDouble("RA", RA_h, string("RA of the centre of WCS transf. [h]"));
		  cur_frame->SetKeywordDouble("DEC", DEC_deg, string("DEC of the centre of WCS transf. [deg]"));

		  ga.WriteAstrometry(cur_frame);
		  }
		else
		  {
		  cur_frame->SetKeywordInt("ASTROOK", 0, string("Astrometry failed"));

		  // Delete old results, if found
                  if(ga.IsAstrometryPresent(cur_frame))ga.DeleteAstrometry(cur_frame);

                  streamlog_out(MESSAGE) << "Astrometry failed !!!" << endl;
		  }

                 
	        delete [] lestr.X;
		delete [] lestr.Y;
		delete [] lestr.signal;
		delete [] lestr.solution;

		} // End of  'else' case from  'if(!overwrite && ga.IsAstrometryPresent(cur_frame))'


        // Print results of the astrometry

        if(ga.IsAstrometryPresent(cur_frame))
	  {
	  ga.InitFromHeader(cur_frame);  // Read astrometry from header (also needed for position calculation)

          stringstream output;
          ga.Print(output);
	  streamlog_out(DEBUG) << output.str() << endl;

	  anyOK = true;
	  }

	// Add astrometry to the table header

	if(addtoheader && ga.IsAstrometryPresent(cur_frame))
	  {
	    objects->SetKeywordInt("ASTROOK", 1, string("Astrometry converged"));
	    
            double RA_h= ga.crval[0]/15.;
	    if(RA_h < 0. ) RA_h+=24;

	    double DEC_deg= ga.crval[1];

	    objects->SetKeywordDouble("RA", RA_h, string("RA of the centre of WCS transf. [h]"));
	    objects->SetKeywordDouble("DEC", DEC_deg, string("DEC of the centre of WCS transf. [deg]"));

            ga.WriteAstrometry(objects);
	  }

	// Add columns with position in RA DEC for each object in the table

	if(addposition && ga.IsAstrometryPresent(cur_frame))
	  {
  	  streamlog_out(DEBUG) << "About to add astrometry results to object table" << endl;

	  // Just in case: if columns already exist, delete them and create new
          if(objects->ColumnExists("RA"))objects->DeleteColumn("RA");
          if(objects->ColumnExists("DEC"))objects->DeleteColumn("DEC");

          objects->AddColumn("RA",(double) 0., string("deg"), string("Object RA from astrometry"));
          objects->AddColumn("DEC",(double) 0., string("deg"), string("Object DEC from astrometry"));
          objects->AdjustColumns();

	  streamlog_out(DEBUG) << " Astrometry results for input object list " << endl;

	  vector<double>*  Xvec = &(objects->GetVectorOf<double>("CCD_X"));
	  vector<double>*  Yvec = &(objects->GetVectorOf<double>("CCD_Y"));

         for(int iobj=0; iobj < Xvec->size(); iobj++)
	    {
	      double radec[2];
              ga.GetRaDecFromXY(Xvec->at(iobj),Yvec->at(iobj),radec);

              objects->SetValue("RA", iobj, radec[0]);

              objects->SetValue("DEC", iobj, radec[1]);

	      streamlog_out(DEBUG) << "Object " << iobj 
				   << ":  CCD_X = " << Xvec->at(iobj) << "  CCD_Y = " << Yvec->at(iobj)
                                   << "   >  RA = " << radec[0] << " DEC = " << radec[1] << endl;
	    }

	  } // End of  if(addposition)
		
	}  // End of loop over frames

	// return true if at least one astrometry OK
	
	if(anyOK)
 	  setReturnValue(true);
	else
	  setReturnValue(false);
	
}

void Astrometry::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void Astrometry::end()
{

	printEndMessage();

}

void Astrometry::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

}

int Astrometry::GetAstrometricIndexesForImage(char **index_list, GloriaFitsImage *cur_frame)
{	
	// Get the image field of view

        double x_fov=0.;
        double y_fov=0.;

	if(cur_frame->IsKeywordSet("XPLATE")) x_fov = cur_frame->GetKeywordValueDouble("XPLATE")*cur_frame->SizeX(); 
	if(cur_frame->IsKeywordSet("YPLATE")) y_fov = cur_frame->GetKeywordValueDouble("YPLATE")*cur_frame->SizeY();

	if(cur_frame->IsKeywordSet("PIXSCALE"))
	{
	  double pixscale = cur_frame->GetKeywordValueDouble("PIXSCALE");

	  if(pixscale < 0.01)
	    {
	    streamlog_out(WARNING) << " Pixel scale from FITS header: " <<  pixscale 
                       << " assumed to be given in degrees ! " << std::endl;

	    pixscale *= 3600; // convert to arcsec
	    }

	x_fov = pixscale*cur_frame->SizeX();
	y_fov = pixscale*cur_frame->SizeY();
	}

	// Otherwise try to guess from the pixel size range specified by user

	if( (x_fov<0.0001 || y_fov<0.0001) && _PixscaleHigh >0)
	    {
	    x_fov = _PixscaleHigh*cur_frame->SizeX();
	    y_fov = _PixscaleHigh*cur_frame->SizeY();
	    }

	double fov, low_fov;

	if(x_fov<0.0001 || y_fov<0.0001)
	{
		streamlog_out(ERROR) << "Cannot detemine anugular size of the frame (X/YPLATE or PIXSCALE probably not set in header). Setting FoV limits to 0.1-180 deg." << endl;
		low_fov = 0.1*3600;
		fov = 180*3600;
	}
	else
	{
		// According to the astrometry.net readme, indexes containing constelations from 0.1 to 1 of FoV should be used.
		// TODO: distinguish between x and y fov
		fov=x_fov;
		low_fov = 0.1*fov;
	}
	int in_cnt = 0;

        streamlog_out(DEBUG) << "Selecting index files for FoV from " << low_fov 
                             << " to " << fov << " arcsec " << endl;

	double low = 2, high=2.8;
	char file_path[1024];
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=47; j++) 
		{
			sprintf(file_path, "%s/index-4000-%2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';				
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
		}


	low = 2.8; high=4;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=47; j++) 
		{
			sprintf(file_path, "%s/index-4001-%0.2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';				
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
		}


	low = 4; high=5.6;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=47; j++) 
		{
			sprintf(file_path, "%s/index-4002-%0.2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';				
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;		
		}


	low = 5.6; high=8;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=47; j++) 
		{
			sprintf(file_path, "%s/index-4003-%0.2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';				
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;		
		}

	low = 8; high=11;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=47; j++) 
		{
			sprintf(file_path, "%s/index-4004-%0.2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';				
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;		
		}	

	low = 11; high=16;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=11; j++) 
		{
			sprintf(file_path, "%s/index-4005-%0.2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path); 
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
		}

	low = 16; high=22;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=11; j++) 
		{
			sprintf(file_path, "%s/index-4006-%0.2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';				
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;		
		}

	low = 22; high=30;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
		for(int j=0; j<=11; j++) 
		{
			sprintf(file_path, "%s/index-4007-%0.2d.fits", _Indexes_path.c_str(), j);
			if(access(file_path, R_OK)==0)
			{
				index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
				// Astrometry.net does not want to get extensions in the index path
				*(strstr(index_list[in_cnt], ".fits"))='\0';				
				in_cnt++;
			}
			else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;		
		}

	low = 30; high=42;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4008.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;	
	}

	low = 42; high=60;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4009.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 60; high=85;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4010.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 85; high=120;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4011.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 120; high=170;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4012.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 170; high=240;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4013.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 240; high=5.67*60;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4014.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 5.67*60; high=8*60;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4015.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 8*60; high=11.3*60;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4016.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 11.3*60; high=16.7*60;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4017.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 16.7*60; high=23.3*60;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4018.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}

	low = 23.3*60; high=33.3*60;
	if((low_fov>low*60 && low_fov<=high*60) || (fov>low*60 && fov<=high*60) || (low_fov<low*60 && fov>=high*60) )
	{
		sprintf(file_path, "%s/index-4019.fits", _Indexes_path.c_str());
		if(access(file_path, R_OK)==0)
		{
			index_list[in_cnt]=(char*)malloc(1024); strcpy(index_list[in_cnt], file_path);
			// Astrometry.net does not want to get extensions in the index path
			*(strstr(index_list[in_cnt], ".fits"))='\0';				
			in_cnt++;
		}
		else streamlog_out(MESSAGE) << "Astrometric index file " << file_path << " could not be accessed. Astrometry may not be optimal.\n" << endl;
	}	

	// Add older astrometry.net indexes if exist
	
        streamlog_out(DEBUG) << in_cnt << " index files selected" << endl;

	return in_cnt;			
}

} // namespace luiza
