// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef SociSelect_h
#define SociSelect_h 1

#include "soci.h"
#include "postgresql/soci-postgresql.h"

#include "luiza/DataSourceProcessor.h"


using namespace gloria;

namespace luiza {
  
  
  //! Processor for reading tables from SQL database with SOCI library
  /*! This processor reads data from SQL database and puts them into
   *  internal FITS table structure, so it can be used in subsequent
   *  processing steps. SOCI library is used to implement interface to
   *  different databases. This processor is designed to execute
   *  SELECT queries only. For detailed description of command syntax
   *  please refer to <a href="http://soci.sourceforge.net/doc.html">
   *  SOCI documentation</a>. 
   * 
   *  \par Input 
   *  User has to specify all connection information required by SOCI
   *  and the select query string. It is also possible to specify an
   *  input index table (and column name), from which the query parameter
   *  values can be taken.  
   *    
   *  \param ConnectString Single line containing all connection
   *  parameters, as required for SOCI session opening
   *  (eg. "postgresql://host=test.db port=1111 user=db_guest dbname=test_db") 
   * 
   *  \param SelectString Single line containing select statement, to
   *  be used as SOCI query. Parameter names (to be taken from index
   *  table) are to be preceded by a colon 
   *  (eg. "select * from table where index=:ID"). 
   *
   *  \param RowsPerLoop Number of rows to be read from the SQL query
   *   result in each processor call. Default is all (parameter value
   *   set to zero). This parameter is only considered when index
   *   table is not used.
   *
   *  \param IndexCollection   Name of the collection containing
   *   index table, i.e. table where SQL select statement parameter
   *   is stored. 
   *
   *  \param IndexColumn   Name of the column in the index table, from
   *   which the SQL select statement parameter is to be read. 
   *
   *  \param IndexesPerLoop   Number of indexes, which should be used
   *   for each loop. Default is one, i.e. select statement is
   *   executed only once for each processor call. To process data for
   *   all indexes, index table has to be defined as permanent or the
   *   parameter has to be set to zero (process all indexes at one call).
   *
   *
   *  \par Output
   *   Tables read are added to specified table collection (named
   *   table vector) and put into global gloria::GloriaDataContainer.
   *
   *  \param TableCollectionName Name of the created
   *    GloriaTableCollection (default is "DbTable");
   *
   *  \param PermanentCollection Flag indicating that created
   *    collection should be kept in memory for the whole
   *    processing time. 
   *
   *
   *  \author A.F. Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   */

   class SociSelect : public DataSourceProcessor {
    
  public:
    
    virtual SociSelect*  newProcessor();
    
    
    SociSelect() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! Connection parameters, as required for SOCI session opening 
    std::vector< std::string > _connectString;

    //! Connection parameters merged to single string
    std::string _connection;

    //! Select statement to be executed as SQL query
    std::vector< std::string > _selectString ;

    //! Select statement merged to single string
    std::string _select;

    //! Number of tables to be read per processing loop (0 for all)
    int _rowsPerLoop ;


    //! Name of the table collection containing index table
    std::string _indexCollection ;

    //! Name of the index table column containing select parameter
    std::string _indexColumn ;

    //! Number of indexes to be used per processing loop (0 for all)
    int _indexesPerLoop ;


    //! Name of the table collection to which SQL tables should be stored
    std::string _tableCollectionName ;

    //! Flag for collections, which should not be deleted after loop is finished
    bool _permanentCollection;


    //! Loop counter
    int _nLoops ;

    //! Table counter
    int _nTables ;

    //! SQL query counter
    int _nSQL ;


    void printEndMessage() const ;

     //! Table template created from SQL output
     GloriaFitsTable* _outputTemplate; 

     //! Data types of SQL output columns
     std::vector<soci::data_type> _sqlTypes;

     //! Names of SQL output columns
     std::vector<std::string> _sqlNames;

     //! SOCI session pointer
     soci::session *_sociSession;

     //! SOCI query statement pointer
     soci::statement *_sociStatement;

     //! SOCI query output row 
     soci::row *_sociRow;

     //! Flag for using indexes in query
     bool _useIndex;

     //! Index counter
     int _indexRow;

     //! Row counter
     int _sqlRow;
  } ;

 } // end namespace luiza
#endif
