#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "SociSelect.h"

namespace luiza {

  SociSelect aSociSelect ;


  // Constructor

  SociSelect::SociSelect() : DataSourceProcessor("SociSelect") {

    _description = "Processor for reading tables from SQL database with SOCI library";


  std::vector< std::string > emptyStringVec;

  registerProcessorParameter ("ConnectString",
                              "Connection parameters required for SOCI session opening",
                              _connectString,  emptyStringVec );


  registerProcessorParameter ("SellectString",
                              "Select statement, to be used as SOCI query",
                              _selectString,  emptyStringVec );


  // Although this processor reads tables from file it creates table collection as an output
  // This is why we register this collection as output collection

  registerOutputCollection (  TableCollection,
                              "TableCollectionName",
                              "Name of the table collection to which data should be stored",
                              _tableCollectionName,  std::string ("DbTable") );

  
  registerInputCollection (  TableCollection,
                              "IndexCollection",
                              "Name of the collection containing index table",
                              _indexCollection,  std::string ("") );

  
  registerProcessorParameter ("IndexColumn",
                              "Name of the column in the index table containing select statement index",
                              _indexColumn,  std::string("") );

  
  // Remaining parameters are optional - can be skiped in the steering file


  registerOptionalParameter ("PermanentCollection",
                              "Flag for collections, which should not be deleted after loop is finished",
                              _permanentCollection,  false );

  registerOptionalParameter ("RowsPerLoop",
                              "Number of rows to be read per processing loop (0 for all))",
                              _rowsPerLoop,  static_cast < int > (0) );


  registerOptionalParameter ("IndexesPerLoop",
                              "Number of indexes to be used per processing loop (0 for all))",
                              _indexesPerLoop,  static_cast < int > (1) );


  }


  SociSelect*  SociSelect::newProcessor() {
    return new SociSelect ;
  }
  

  void SociSelect::init() { 

    streamlog_out( MESSAGE ) << "SociSelect::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    
    _nSQL = 0 ;
    _nTables = 0 ;
    _nLoops = 0 ;

    _indexRow = 0;
    _sqlRow = 0;

    _sociSession = NULL;
    _sociStatement = NULL;
    _sociRow = NULL;

    _outputTemplate = new GloriaFitsTable();

    _sqlTypes.clear();
    _sqlNames.clear();

    // Check if all required parameters are given

    if(_connectString.size()==0)
      {
	streamlog_out( ERROR ) << "No connection parameters specified in SociSelect! " << std::endl;
	streamlog_out( ERROR ) << "Can not proceed! " << std::endl;

	throw StopProcessingException( this ) ;
	return;
      }

    _connection=string("");

    for(int istr=0; istr<_connectString.size();istr++)
      _connection+= _connectString.at(istr) + string(" ");

    streamlog_out( DEBUG ) << "Connection string is: " << _connection << std::endl;
  

    if(_selectString.size()==0)
      {
	streamlog_out( ERROR ) << "No select query specified in SociSelect! " << std::endl;
	streamlog_out( ERROR ) << "Can not proceed! " << std::endl;

	throw StopProcessingException( this ) ;
	return;
      }

    _select=string("");

    for(int istr=0; istr<_selectString.size();istr++)
         _select+= _selectString.at(istr) + string(" ");

    streamlog_out( DEBUG ) << "Query string is: " << _select << std::endl;
  

    _useIndex = false;

    if(_indexCollection.size()>0  && _indexColumn.size()==0)
	streamlog_out( WARNING ) << "Index column not specified, will not be used! " << std::endl;

    if(_indexCollection.size()>0  && _indexColumn.size()>0)_useIndex = true;


 }


void SociSelect::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(DEBUG) << " startAnalysis() " << endl;

  // Open SQL session

    try
    {
      _sociSession = new soci::session(_connection);

     streamlog_out(DEBUG) << " SOCI session established " << endl;
    }
    catch (exception const &e)
    {
      streamlog_out(ERROR) << "SOCI session opening failed: " << e.what() << endl;
      throw StopProcessingException( this ) ;
    }

    _sociRow = new soci::row;

    _isInputInitialized = true;

}


void SociSelect::processData( GloriaDataContainer* data ) 
{ 
    // Return without processing if no data left to read

    if(_isInputClosed ) return;

    // Count only loops where something was done...

    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    
    // Create output collection if not found

    if(_tableCollectionName.size() > 0  &&
	!data -> TableCollectionExists(_tableCollectionName))
      {
         data -> CreateTableCollection(_tableCollectionName,_permanentCollection);
         streamlog_out(DEBUG) << "  Output collection " << _tableCollectionName << " created " << endl;
      }

    // Look for index table

    GloriaFitsTable *index;
    int nIndexes =  0;

    if (_useIndex)
      {
      if(!data->TableCollectionExists(_indexCollection))
	{
	streamlog_out(WARNING) << "Index collection not found ! " << endl;
	return;
	}

      int nInd = data->GetTableCollection(_indexCollection).size();

      if (nInd == 0)
	{
	streamlog_out(WARNING) << "No table found in index collection ! " << endl;
	return;
	}


      if(nInd>1)
        {
        streamlog_out(WARNING) << "Multiple index tables found in index collection! " << endl;
        streamlog_out(WARNING) << "Only the first index table will be used !!!" << endl;
        }

      index = &(data -> GetTableCollection(_indexCollection).at(0));

      if(!data -> TableCollectionPermanent(_indexCollection))_indexRow=0;

       nIndexes =  index->GetRowNumber();
      }

    int nIndUse = 1;

    if(_useIndex)
      {
	if(_indexesPerLoop>0) nIndUse = _indexesPerLoop;
 
	if(_indexesPerLoop<=0 || _indexRow+nIndUse > nIndexes) nIndUse = nIndexes - _indexRow;
      }

    // Main loop over indexes (also executed once when indexes not used)


    for(int iind=0; iind<nIndUse ; iind++)
      {

       string indstr;
       if(_useIndex)
           indstr = (index->GetColumn(_indexColumn))->StringVal(_indexRow);   // take index value as a string

       if(_useIndex || _rowsPerLoop<=0 || _sqlRow==0)
	 {
         streamlog_out(DEBUG) << "Preparing query for index: " << indstr  << endl;

	 if(_sociStatement)delete _sociStatement;

         _sociStatement = new soci::statement(*_sociSession);

	 if(_useIndex)
	    *_sociStatement = (_sociSession->prepare << _select, soci::use(indstr), soci::into(*_sociRow));
          else
            *_sociStatement = (_sociSession->prepare << _select, soci::into(*_sociRow));


          _sociStatement->execute();
	  _nSQL++;
	  }

         // Check first row

         bool newdata = _sociStatement->fetch();

         if(!newdata)
	   {
	     streamlog_out(WARNING) << "SOCI reports no data returned from select, index: "<< indstr << endl;
	     if(!_useIndex)
	       {
               _isInputClosed = true;
	       }
	     else
	       {
	       _indexRow++;
	       if(_indexRow == nIndexes && data -> TableCollectionPermanent(_indexCollection))_isInputClosed = true;
	       break;
	       }

	     continue;
	   }

	 soci::row const& firstrow = *_sociRow;

         int ncol = firstrow.size();

	 if(_sqlTypes.size() > 0 && ncol != _sqlTypes.size())
	   {
	     streamlog_out(ERROR) <<  "Number of SQL output columns changed for index "<< indstr << endl;
	     streamlog_out(ERROR) <<  "Can not continue!!! "<< endl;

	     throw StopProcessingException( this ) ;
	     return;
	   }



	 // Define columns for first index only (if nonempty select output)

	 if(_sqlTypes.size()==0)
	   {
           for(std::size_t icol = 0; icol < firstrow.size(); ++icol)
             {
	      const soci::column_properties & props = firstrow.get_properties(icol);

	      // Add column definition

 	      string colname = props.get_name();
	      soci::data_type coltype = props.get_data_type();

	      _sqlNames.push_back(colname);
	      _sqlTypes.push_back(coltype);

	      string empty_string = "[NULL]";
	      double empty_double = -1.0;
	      long empty_long = -1;

              switch(coltype)
                {
                case soci::dt_string:
                  _outputTemplate->AddColumn(colname,empty_string);
                  break;

                case soci::dt_double:
                  _outputTemplate->AddColumn(colname,empty_double);
                  break;

                case soci::dt_integer:
                  _outputTemplate->AddColumn(colname,empty_long);
                  break;

                case soci::dt_date:
                  _outputTemplate->AddColumn(colname,empty_long,"s","Date-Time structure converted to Unix time");
                  break;

	        default:
		  streamlog_out(WARNING) << " Column " << colname << " will not be copied: unknown type " << endl;
                  break;
	        }
              }
	   }

       // Initialize output table


       (data->GetTableCollection(_tableCollectionName)).push_back(*_outputTemplate);

       int nOut = (data->GetTableCollection(_tableCollectionName)).size();

       GloriaFitsTable *outtab = &(data->GetTableCollection(_tableCollectionName).at(nOut-1));

       _nTables++;

       // Copy SQL output
        

       int nCopied = 0;

        while (newdata)
	  {
	    soci::row const& tabrow = *_sociRow;

	    outtab->AddRow();

          for(std::size_t icol = 0; icol < tabrow.size(); ++icol)
            {
            const soci::column_properties & props = tabrow.get_properties(icol);
 	    string colname = props.get_name();
	    soci::data_type coltype = props.get_data_type();

	    if(colname!=_sqlNames.at(icol) || coltype!=_sqlTypes.at(icol))
	      {
	       streamlog_out(ERROR) << " Column definition changed during data processing !?! "<< endl;
	       streamlog_out(ERROR) <<  "Can not continue!!! "<< endl;

	       throw StopProcessingException( this ) ;
	       return;
	      }

            if (tabrow.get_indicator(icol) != soci::i_null) // Non-empty columns only
	      {
              tm timestruc;
              time_t timesec;

              switch(props.get_data_type())
                {
                case soci::dt_string:
		  outtab->SetLast(colname, tabrow.get<std::string>(icol));
                  break;

                case soci::dt_double:
                  outtab->SetLast(colname, tabrow.get<double>(icol));
                  break;

                case soci::dt_integer:
	          outtab->SetLast(colname, (long)tabrow.get<int>(icol));
                  break;

                case soci::dt_date:
                  timestruc = tabrow.get<tm>(icol);
                  timesec = mktime(&timestruc);
                  outtab->SetLast(colname, timesec);
                  break;

	        default:
                  break;
	        }
	      }
	    }

	   _sqlRow++;
	   nCopied++;

	  if(!_useIndex &&  nCopied ==  _rowsPerLoop)
	     newdata = false;
          else
	    {
	     newdata = _sociStatement->fetch();
	     if(!_useIndex && !newdata)_isInputClosed = true;
	    }
        }



       int nout = outtab->GetRowNumber();
       streamlog_out(DEBUG) << nout << " rows added to output table for index " << indstr << endl;


      if(_useIndex)
	{
	_indexRow++;
	if(_indexRow == nIndexes && data -> TableCollectionPermanent(_indexCollection))_isInputClosed = true;
	}
      }

    // always return true  for ProcessorName

    setReturnValue( true ) ;
    
  }
  
  void SociSelect::check(GloriaDataContainer* data) { 
    
    if(! data -> TableCollectionExists(_tableCollectionName))
      {
         streamlog_out(ERROR) << " check(): Collection not found !? " << std::endl ;
	 return;
      }


    int nTbl = data->GetTableCollection(_tableCollectionName).size();

    streamlog_out(MESSAGE) << " check(): Collection " << _tableCollectionName << " with " << nTbl << " tables" << std::endl ;
 
  }

void SociSelect::concludeAnalysis( GloriaDataContainer* data )
{
   streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;

    if(_outputTemplate)delete _outputTemplate;

    if(_sociStatement)delete _sociStatement;

   if(_sociRow)delete _sociRow;

    _sociSession->close();

    streamlog_out(DEBUG) << " SOCI session closed " << endl;
}


  
void SociSelect::end()
{ 
    printEndMessage() ;
}

  void SociSelect::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;

    streamlog_out(MESSAGE) << " processed "     << _nSQL << " SQL queries " << std::endl ;
    
    streamlog_out(MESSAGE) << _nTables << " tables stored " << std::endl 
			   << std::endl ;
  }

  
}// namespace luiza
