// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef ApplyGausKernel_h
#define ApplyGausKernel_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to calculate Gaus Kernel and apply it to the image frame
  /*! This processor can be used to convolut sky image with  Gaus
   *  Kernel based on the input PSF shape parameters.  Shape
   *  parameters can be loaded from the table collection prepared in
   *  CalculateGausPSF. Kernel size can be set by user or calculated
   *  automatically from the shape parameters. After convoluting the
   *  sky frame with the kernel, the processor searches for the
   *  objects in an image and calculates their position (from signal
   *  weighted pixel positions).  
   *
   *  \par Input 
   *  Processor loads images from image collection and uses shape
   *  parameter tables loaded from table collection
   *
   *  \param ImageCollectionName -
   *             Name of the collection containing input images
   *  
   *  \param InputTableOfParameters - Name of the table collection
   *             with shape parameters (calculated in CalculateGausPSF)
   *  
   *  \par Algorithm parameters
   *  User can modify processor performance by specify number of sectors
   *  on the analysed frame. Moreover user can choose, whether want only
   *  brigher stars or also darker by defining parameter to distinguish
   *  the stars from the sky (in units of noise level)
   *
   *  \param VariableKernelFlag - flag to decide whether kernel size
   *              should be fixed by user (when set to false) or
   *              variable (calculated from PSF size; default)
   *
   *  \param FixedKernelSize - size of kernel, to be used when
   *                   VariableKernelSize is set to false
   *
   *
   *  \param PixelThresholdKernel - Threshold for the pixel convoluted
   *                        frame (in units of noise level), to be
   *                        considered star seed.
   *
   *  \param PixelCut - parameter to decide which pixels from convoluted
   *                  image are included in the star position
   *                  calculation. Only pixels with signal above
   *                  pixCut x seed signal on the original image (!) are included. 
   *
   *  \param MinDistBetweenStars - if two maximal pixels are closer than "MinDistBetweenStars", 
   *                     only the higher one is used to reconstruct the star
   *
   *  \param StarCogRadius - radius (around seed pixel) for star
   *                     position calculation with Center-Of-Gravity
   *                     method (CoG). The algorithm is run twice:
   *                     first using the seed pixel and then using the
   *                     position from first iteration.
   *
   *  \param BackgroundSampling  - Background level calculation for
   *               the inlut image can be sped up by considering only
   *               1/BackgroundSampling of pixels.
   *
   *  \param BackgroundFrameDivision - Background for the image is calculated for
   *              BackgroundFrameDivision x BackgroundFrameDivision subimages
   *                         and then interpolated between them.
   *
   *  \param InterpolationOrder - polinomial order used to interpolate
   *                         between centers of subframes when
   *                         calculating background values on the image
   *
   *
   *  \par Output
   *  Object lists are stored in the output collection
   *
   * 
   *  \param OutputListCollectionName - 
   *            Name of the collection where object list table will be stored
   * 
   *  \param OutputFrameCollectionNam - 
   *            Name of the collection where convoluted frames will be stored
   * 
   *  \param HeaderRecords -
   *            Keywords of image fits header, which should be copied
   *            to the fits header of the object list table. Default list
   *            consists of following observation describing keywords
   *            (as suggested in FITS standard): DATE-OBS, TELESCOP,
   *            INSTRUME, OBSERVER, OBJECT. 
   *
   *
   *  \param PermanentList  -  Flag output object list collection, as
   *           permanent, not deleted after loop is finished (default is false)
   *
   *  \param PermanentFrame  -  Flag output image collection, as
   *           permanent, not deleted after loop is finished (default is false)
   *
   */

  class ApplyGausKernel : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new ApplyGausKernel ; }
    
    
    ApplyGausKernel() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;

  //  virtual void createTable( GloriaDataContainer* data ) ;

    virtual void startAnalysis(GloriaDataContainer* data);
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images 
    string _imageCollectionName;

    //! Name of the table collection containing tables 
    string _inputTableOfParameters;

    //! Name of the output object list collection 
    string _outputListCollectionName;

    //! Name of the output collection for convoluted frames
    string _outputFrameCollectionName;

    //! Flag the output object list collection to be permanent
    bool _permanentList;

    //! Flag the output collection of convoluted frames to be permanent
    bool _permanentFrame;

    //! Background sampling factor
    /*! Parameter used to speed up background level calculation.
     *  Only every BackgroundSampling pixel is used.
     */
    int _backgroundSampling;

    //! Frame division for background calculation
    /*! Background is calculated for BackgroundFrameDivision x BackgroundFrameDivision 
     *  subimages and then interpolated between them
     */
    int _backgroundFrameDivision;

    //! Flag to decide, whether kernel size is constant or variable. If it's is true, size is variable, else constant.

    bool _variableKernelFlag;

    //! If kernel have got constant size, user set the value of size of kernel
    int _fixedKernelSize;

    //! Polinomial order used to interpolate background map
    int _interpolationOrder; 

    //! Threshold for the cluster pixel signal (in units of noise level)
    double _pixThresholdKernel;

    //! Decide wchich pixel is included to star (in units of the brightest pixel in the star)
    double _pixCut;

    //! Minimum distance between two stars
    double _minDistBetweenStars;

    //! Radius for calculating star position with CoG method
    double _starCogRadius;

    //! Keywords of image fits header to copied to the table header
    std::vector<std::string> _copykey;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputLists ;

  } ;
  
} // end namespace luiza
#endif


