// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef AstroImageAdder_h
#define AstroImageAdder_h 1

#include "luiza/Processor.h"
#include "gloria/GloriaFitsImage.h"
#include "gloria/GloriaAstrometry.h"
#include "gloria/GloriaValueMap.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to add images, based on astrometry (taking into account possible shifts)
  /*! Processor can be used to add images in the input collection.
   *  Corrections for possible shifts between images (eg. due to mount
   *  slide) are applied based on astrometry results stored in image
   *  headers. Float image stored in the output colection is an average of
   *  input images.
   *
   *  \par Input 
   *  Processor takes images from Gloria data collection.
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be added
   *
   *  \param NumberAddImages  -
   *  Number of images to be added each time, 0 for all images in
   *  input collection (default is 0) 
   *
   *  \param UseManyLoops  -
   *  Wait for images in next loops, if not enough images in input
   *  collection (default is false) 
   *
   *  \par Algorithm parameters
   *  User has to specify shift map binning
   *
   *  \param FrameDivision 
   *     Shift maps in X and Y art calculated only in bin centers, with
   *     frame divided into FrameDivision x FrameDivision bins. Shifts 
   *                         are then interpolating between bin centers.
   *
   *  \param InterpolationOrder 
   *     polinomial order used to interpolate between centers of
   *     bins when calculating X/Y position corrections. 
   *   
   *  \par Output
   *
   *  Averaged frames are stored in the output collection
   *
   *  \param OutputCollectionName - 
   *            Name of the collection where image sums will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   */

  class AstroImageAdder : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new AstroImageAdder ; }
    
    
    AstroImageAdder() ;
    
    //! Called at the begin of the job before anything is read.
    /*! Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    //! Called before first data processing loop 
    /*! Can be used to initialize data collections required for
     *  analysis, eg. define permanent collections, or fill them with
     *  external data (from file, database or net). It is possible
     * as GloriaDataContainer is already defined, contrary to init().
     */
    virtual void startAnalysis( GloriaDataContainer* ) ; 

    //! Called for every data processing loop 
    /*! This is where most of the work (analysis) is done. 
     * Data is exchanged between processors via GloriaDataContainer 
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    //! Called for every loop - right after processData() call
    /*! 
     *  Can be used to check processing and/or produce check plots.
     */
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops.
    /*! This method is called to finalize the analysis, after all
     * input data was read (or required number of loops was
     * reached). Analysis of data collected in permanent data
     * collections is possible here.
     */
    virtual void concludeAnalysis( GloriaDataContainer* ) ;

    //! Called after data processing is finished
    /*! Called after all input sources are closed for clean up.
     *  Notice that processors are called in the inverse order of the init()
     *  method so that resources allocated in the first processor also will 
     *  be available for all following processors.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images to be added
    string _inputCollectionName;

    //! Name of the output image collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Number of images in the sum (0 for all in the input collection)
    int _nAddImages;

    //! Allow adding images from different loops.
    bool _useManyLoops;

    //! Frame division for shift map calculation
    /*! Position correction is calculated for FrameDivision x FrameDivision 
     *  bins and then interpolated between them
     */
    int _frameDivision;

    //! Polinomial order used to interpolate shift correction maps
    int _interpolationOrder; 

    int _nLoops ;
    int _nInputImages ;
    int _nOutputImages ;
    int _nAdded;

    GloriaFitsImage* _sumImage;
    GloriaFitsImage* _sumCount;

    double *_sumImageData;
    long *_sumCountData;

    GloriaAstrometry _sumAstro;

    GloriaValueMap* _shiftMaps;

    int _sumSizeX;
    int _sumSizeY;

  } ;
  
} // end namespace luiza
#endif



