// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef BackgroundImage_h
#define BackgroundImage_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to generate background/noise map from CCD frame
  /*! Processor calculates background map from given image. The frame
   *  is divided into FrameDivision x FrameDivision subframes, then
   *  background level and noise are calculated from median in each
   *  subframe. Finally, background and/or noise image are created by
   *  interpolating between subframe centers. New images have the same
   *  resolution as the input image.
   *
   *  \par Input 
   *  Processor takes images from Gloria data collection
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be added
   *
   *  \par Algorithm parameters
   *  User can modify processor performance by setting following
   *  parameters: 
   *
   *  \param BackgroundSampling  - Background level calculation can be
   *                       sped up by considering only
   *                       1/BackgroundSampling of pixels.
   *
   *  \param FrameDivision - Background is calculated for
   *                         FrameDivision x FrameDivision subimages
   *                         and then interpolated between them.
   *
   *  \param InterpolationOrder - polinomial order used to interpolate
   *                         between centers of subframes when
   *                         calculating background image
   *
   *  \par Output
   *
   *  Background and/or noise images can be stored in data collection
   *
   *  \param BackgroundCollectionName - 
   *            Name of the collection where background image will be stored
   * 
   *  \param NoiseCollectionName - 
   *            Name of the collection where noise image will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection(s), as permanent, not deleted after loop is
   *   finished (default is false)
   *
   */

  class BackgroundImage : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new BackgroundImage ; }
    
    
    BackgroundImage() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images to be added
    string _inputCollectionName;

    //! Name of the output background image collection
    string _backgroundCollectionName;

    //! Name of the output noise image collection
    string _noiseCollectionName;

    //! Flag the output collections to be permanent
    bool _permanentCollection;

    //! Background sampling factor
    /*! Parameter used to speed up background level calculation.
     *  Only every BackgroundSampling pixel is used.
     */
    int _backgroundSampling;

    //! Frame division for background calculation
    /*! Background is calculated for FrameDivision x FrameDivision 
     *  subimages and then interpolated between them
     */
    int _frameDivision;

    //! Polinomial order used to interpolate background map
    int _interpolationOrder; 

    int sizeX;
    int sizeY;
    long sizeXY;

    //! Use network configuration file to update processor parameters 
    /*! When true, network configuration file is checked for each
     *  analysed image and processor parameters are changed if default
     *  parameter values are defined for given instrument/telescope.
     */
    bool _useTelescopeParameters;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputImages ;

  } ;
  
} // end namespace luiza
#endif



