// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef CalculateGausPSF_h
#define CalculateGausPSF_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to calculate Gaus profile parameters for stars on the CCD frame
  /*! Processor calculates Gaus profile parameters for stars found on
   *  CCD images from the input collection. 
   *  For each star found in the image (after background subtraction
   *  is applied) following parameters of its profile are
   *  calculated from the selected pixels: RMS in X, RMS in Y,
   *  XY-correlation, FWHM, Elongation and orientation.
   *  Median of the obtained parameter value is calculated in given
   *  number of frame sectors and the resulting map is stored in the
   *  output table. It can then be used to calculating Gauss kernel
   *  for given image. 
   *
   *  \par Input 
   *  Processor reads images from the Gloria data collection
   *
   *  \param ImageCollectionName -
   *             Name of the collection containing input images
   *
   *
   *  \par Algorithm parameters
   *  User can modify processor performance by specify number of sectors
   *  on the analysed frame. Moreover user can choose, whether want only
   *  brigher stars or also darker by defining parameter to distinguish
   *  the stars from the sky (in units of noise level)
   *
   *
   *  \param BackgroundSampling  - Background level calculation for
   *                       the whole image can be
   *                       sped up by considering only
   *                       1/BackgroundSampling of pixels.
   *
   *  \param BackgroundFrameDivision - Background for the image is calculated for
   *              BackgroundFrameDivision x BackgroundFrameDivision subimages
   *                         and then interpolated between them.
   *
   *  \param FrameDivision - Shape parameter map is calculated for 
   *                   FrameDivision x FrameDivision subimages. 
   *
   *  \param InterpolationOrder - polinomial order used to interpolate
   *                         between centers of subframes when
   *                         calculating background values on the
   *                         image
   *
   *  \param pixThreshold - signal threshold for the pixel to be
   *                         considered as a star maximum (if higher
   *                         than all its neighbours; in units of noise level)
   *
   *  \param medianThreshold - cut on the signal median of all pixels used
   *                         for star shape calculation (in units of
   *                         noise level). New pixels are added to the
   *                         star until this threshold is met.
   *                         
   *
   * \param MaxPixelValue - signal threshold for the saturated pixels used 
   *                        to eliminate saturated stars and shape analysis (in value of pixel)
   *  \par Output
   *  Object lists are stored in the output collection
   *
   * 
   *  \param OutputCollectionName - 
   *            Name of the collection where parameter map will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output table collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   */

  class CalculateGausPSF : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new CalculateGausPSF ; }
    
    
    CalculateGausPSF() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;

  //  virtual void createTable( GloriaDataContainer* data ) ;

    virtual void startAnalysis(GloriaDataContainer* data);
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images 
    string _imageCollectionName;

    //! Name of the output object list collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;


    //! Background sampling factor
    /*! Parameter used to speed up background level calculation.
     *  Only every BackgroundSampling pixel is used.
     */
    int _backgroundSampling;

    //! Frame division for background calculation
    /*! Background is calculated for BackgroundFrameDivision x BackgroundFrameDivision 
     *  subimages and then interpolated between them
     */
    int _backgroundFrameDivision;

    //! Polinomial order used to interpolate background map
    int _interpolationOrder; 

    //! Frame division for shape map calculation
    /*! Shape parameter maps are calculated for FrameDivision x FrameDivision 
     *  subimages 
     */
    int _frameDivision;

    //! Threshold for the highest pixel signal (in units of noise level)
    double _pixThreshold;

    //! Cut on the pixel median (in units of noise level) for star definition
    double _medianThreshold;

    //! Threshold using to eliminate the saturated stars
    double _MaxPixelValue;

    //! Use network configuration file to update processor parameters 
    /*! When true, network configuration file is checked for each
     *  analysed image and processor parameters are changed if default
     *  parameter values are defined for given instrument/telescope.
     */
    bool _useTelescopeParameters;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputMaps ;

  } ;
  
} // end namespace luiza
#endif
