// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef FastAperturePhotometry_h
#define FastAperturePhotometry_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Simple processor to calculate aperture photometry on the CCD frame
  /*! Processor to calculate aperture photometry on CCD images in the
   *  input collection. If input object collection is specified,
   *  photometry is calculated for all objects in the object list. For
   *  the targeted analysis, coordinates of the objects of interest
   *  can be specified. However, valid astrometry has to be found in
   *  the image fits header in that case. Two approaches for
   *  background calculation are possible: either it is calculated
   *  from media of pixels in annulus around object position (when
   *  \a AnnulusRadii is specified) or from the background map
   *  calculated for the whole image (as in \a BackgroundImage processor).
   *
   *  \par Input 
   *  Processor uses images from Gloria data collection
   *
   *  \param ImageCollectionName -
   *             Name of the collection containing input images
   *
   *  \param ListCollectionName -
   *             Name of the collection containing  GloriaObjectList
   *
   *  \param  ObjectPositions 
   *    Positions of the considered objects in (Ra,Dec), should be
   *    given in degrees (N x 2 numbers).
   *
   *  \par Algorithm parameters
   *  User can modify processor performance by settin the aperture
   *  radius (for signal counting) and annulus radii (background
   *  subtraction). Background is calculated as signal median from the
   *  pixels inside annulus.
   *
   *  \param  ApertureRadius -
   *           Radius of circular aperture used for signal
   *           counting in pixels. Distance is always calculated from/to the center
   *           of the pixel, to reduce systematic effects due to
   *           changing number of pixels considered.
   *
   *  \param  AnnulusRadii -
   *           Inner and outer radius of the annulus used for
   *           background subtraction.
   *
   *  \param BackgroundSampling  - Background level calculation for
   *                       the whole image can be
   *                       sped up by considering only
   *                       1/BackgroundSampling of pixels.
   *
   *  \param FrameDivision - Background for the image is calculated for
   *                         FrameDivision x FrameDivision subimages
   *                         and then interpolated between them.
   *
   *  \param InterpolationOrder - polinomial order used to interpolate
   *                         between centers of subframes when
   *                         calculating background values on the image
   *
   *  \par Output
   *  Object lists are stored in the output collection
   *
   *  \param BrightnessColumn - 
   *            Name of the column in which object brightness should
   *            be stored ('Signal' by default). 
   * 
   *  \param OutputCollectionName - 
   *            Name of the collection where object table will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   */

  class FastAperturePhotometry : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new FastAperturePhotometry ; }
    
    
    FastAperturePhotometry() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images 
    string _imageCollectionName;

    //! Name of the table collection containing object lists
    string _listCollectionName;

    //! Name of the output object list collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;


    //! Radius of circular aperture used for signal counting in pixels
    float _apertureRadius;

    //! Inner and outer radius of the annulus used for background subtraction.
    std::vector<float> _annulusRadii;

    //! Background sampling factor
    /*! Parameter used to speed up background level calculation.
     *  Only every BackgroundSampling pixel is used.
     */
    int _backgroundSampling;

    //! Frame division for background calculation
    /*! Background is calculated for FrameDivision x FrameDivision 
     *  subimages and then interpolated between them
     */
    int _frameDivision;

    //! Polinomial order used to interpolate background map
    int _interpolationOrder; 

    //! Name of the column to store object signal 
    std::string _brightnessColumn;

    //! Object positions (two numbers, RA,DEC in degree, per object)
    std::vector<float> _objectPositions;

    //! Use network configuration file to update processor parameters 
    /*! When true, network configuration file is checked for each
     *  analysed image and processor parameters are changed if default
     *  parameter values are defined for given instrument/telescope.
     */
    bool _useTelescopeParameters;

    //! Flag indicating input method used
    bool _usingObjectList;


    //! Pixel shifts for aperture definition (to speed up the code)
    std::vector<int> _apShiftX;
    std::vector<int> _apShiftY;
    std::vector<int> _anShiftX;
    std::vector<int> _anShiftY;
    
    int _nLoops ;
    int _nInputImages ;
    int _nOutputLists ;

  } ;
  
} // end namespace luiza
#endif



