// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef ImageLaplace_h
#define ImageLaplace_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Simple processor to produce laplace images from the input collection.
  /*! Processor produce laplace images from the input collection.
   *  
   *  \par Input 
   *  Processor takes images from Gloria data collection and transforms them.
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be transformed
   *
   *  \par Output
   *  Added (or averaged) frames are stored in the output collection
   *
   *  \param OutputCollectionName - 
   *            Name of the collection where laplace image will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   *  \par Transformation
   *  Laplace transformation replaces each pixel with the value
   *  calculated from the pixel and its neighbors. Close neighbors are
   *  added (signal), more distant ones are subtracted
   *  (background). This transformation increases efficiency for
   *  object identification on the frame. 
   *
   *  \param LaplaceType
   *   Allows to choose from implemented Laplace transformation
   *   matrices. Three transformation types are currently implemented,
   *   using 9, 13 or 16 pixels (type 1, 2 and 3 respectively).
   */

  class ImageLaplace : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new ImageLaplace ; }
    
    
    ImageLaplace() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images to be added
    string _inputCollectionName;

    //! Name of the output image collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Allow choosing Laplace transformation matrix
    int _laplaceType;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputImages ;

    GloriaFitsImage *_origImage;
    GloriaFitsImage *_LapImage;

  } ;
  
} // end namespace luiza
#endif



