// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef ImageNormalizationProcessor_h
#define ImageNormalizationProcessor_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Simple processor to apply dark subtraction and flat correction
  /*! Processor to normalize raw images ie. subtract dark and bias
   *  images, and divide by flat frame.  For each pixel the formula is:
   *  \f[   I_{out} = i \cdot \frac{ I_{in} - d \cdot D - b \cdot B }{f \cdot F} \f]
   *  where \f$I, D, B,  F \f$ are the pixel values in image, dark,
   *  bias and flat frames, and  \f$i, d, b,  f \f$ are the
   *  corresponding scale factors.
   *
   *  \par Input 
   *  Processor takes images from Gloria data collection. If sizes of
   *  collections agree, subsequent darks, flats or biases are used for each
   *  image. Otherwise first dark, flat and/or bias is used.
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be corrected
   *
   *  \param DarkCollectionName -
   *             Name of the collection containing darks - frames
   *             which should be subtracted from data images.
   *
   *  \param FlatCollectionName -
   *             Name of the collection containing flats - data images
   *             are divided by this image. 
   *
   *  \param BiasCollectionName -
   *             Name of the collection containing bias images - 
   *             used to correct for readout noise.
   *
   *  \par Output
   *  Corrected frames can be stored in the output collection, if specified
   *
   *  \param OutputCollectionName - 
   *            Name of the collection where corrected images will be
   *            stored. If name is omitted, images in the input
   *            collection will be modified.
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   *  \par Options
   *  Allow for more automatic correction calculation
   *
   *  \param AutoNormalization - if true: use exposure time
   *           information (as extracted from fits header) to
   *           calculated correction factors (for darks and biases)
   *
   *  \param FlatNormalization - apply dark/bias subtraction procedure 
   *       to the flat frame as well (before applying flat to the actual
   *       image). If set to true, the flat frame is also normalized
   *       to the average value of 1).
   *
   *  \param MatchKeywords - Names for header keywords which should match for
   *             selected dark/bias/flat frames. If no frame with
   *             matching keywords is found, first frame in
   *             corresponding collection is used.
   *
   *  \param MatchExpTime - Flag for using exposure time information
   *             (EXPTIME keyword) to find the best dark frame for the
   *             considered image.
   *
   *  \par Correction parameters
   *  Optional parameters allow to scale considered images (used only if
   *  AutoNormalization flag is set to false)
   *
   *  \param ImageScale - Factor used to scale input image pixels
   *                                           (default = 1.)
   *
   *  \param DarkScale - Factor used to scale dark image
   *                                           (default = 1.)
   *
   *  \param BiasScale - Factor used to scale bias image
   *                                           (default = 0.)
   *
   *  \param FlatScale - Factor used to scale flat frame
   *                                           (default = 1.)
   *
   */

  class ImageNormalizationProcessor : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new ImageNormalizationProcessor ; }
    
    
    ImageNormalizationProcessor() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images to be corrected
    string _inputCollectionName;

    //! Name of the image collection containing dark images
    string _darkCollectionName;

    //! Name of the image collection containing flat images
    string _flatCollectionName;

    //! Name of the image collection containing bias images
    string _biasCollectionName;

    //! Name of the output image collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

     //! Flag for automatic calculation of correction factors
    bool _AutoNormalization;

     //! Flag for subtracting dark/bias from flat frame
    bool _FlatNormalization;

    //! Factor used to scale input image
    double _imageScale;

    //! Factor used to scale dark image
    double _darkScale;

    //! Factor used to scale flat frame
    double _flatScale;

    //! Factor used to scale bias image
    double _biasScale;

    //! Keywords which should match for selected dark/bias/flat frame
    std::vector<std::string> _matchKeywords;

    //! Values of keywords being matched for considered image
    std::vector<std::string> _matchValues;

    //! Flag for using dark frame selection based on exposure time
    bool _matchExpTime;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputImages ;
    int _nCorrected;

    int _nKey;
    std::string _initKeyVal;
    
  } ;
  
} // end namespace luiza
#endif



