// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef PixelClusterFinder_h
#define PixelClusterFinder_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Simple processor to find pixel clusters (stars) on the CCD frame
  /*! Processor to find pixel clusters (stars) on CCD images in the
   *  input collection. This algorithm searches for groups of pixels
   *  with signal above the defined threshold. The seed pixel is
   *  searched first (with tighter seed selection cut) and then the
   *  neighbouring pixels, with signal above pixel signal threshold
   *  (which can be looser) are added to form the cluster. The
   *  algroithm assumes that the cluster should have one maximum only,
   *  if two maxima (above seed threshold) are found two clusters
   *  will be reconstructed. To define thresholds, mean background
   *  level and average noise level are calculated first.
   *
   *  \par Input 
   *  Processor takes images from Gloria data collection
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be added
   *
   *  \par Algorithm parameters
   *  User can modify processor performance by settin the thresholds
   *  and the cluster multiplicity cuts
   *
   *  \param  SeedThreshold -
   *           Threshold for the cluster seed signal, given in units of
   *           noise level (above background level) 
   *
   *  \param  PixelThreshold -
   *           Threshold for the cluster pixel signal, given in units of
   *           noise level (above background level) 
   *
   *  \param  MoreNeighbours - When set to true: search neighbours in 5x5 square
   *                       rather than 3x3 (default)
   *
   *  \param  MaxPixels -  Maximum number of pixels considered in cluster CoG and signal calculation 
   *
   *  \param  MaxClusters - Maximum number of clusters to be reconstructed  
   *
   *
   *  \param BackgroundSampling  - Background level calculation can be
   *                       sped up by considering only
   *                       1/BackgroundSampling of pixels.
   *
   *  \param FrameDivision - Background and Noise is calculated for
   *                         FrameDivision x FrameDivision subimages
   *                         and then interpolated between them.
   *
   *  \param BackgroundInterpolation - Flag for interpolation between
   *                       background levels calculated in different subframes: 
   *                          0 - no interpolation, 
   *                          1 -linear linterpolation (default).
   *
   *  \param NoiseInterpolation - Flag for interpolation between
   *                        noise levels calculated in different subframes: 
   *                          0 - no interpolation, 
   *                          1 -linear linterpolation (default).
   *
   *  \par Output
   *
   *  Object lists are stored in the output collection
   *
   *  \param OutputCollectionName - 
   *            Name of the collection where object table will be stored
   * 
   *  \param AddNoise
   *            Flag to add noise information to the output table.
   * 
   *  \param NoiseColumn
   *            Name for the noise column. Default name is "Noise" 
   * 
   *  \param AddSpread
   *            Flag to add charge spread information to the output
   *            table: RMS in X, RMS in Y and X-Y correlation.
   * 
   *  \param SpreadColumns
   *            Names for columns containing charge spread
   *            information. Default names are "RMS_X", "RMS_Y" and
   *            "CORR_XY" 
   * 
   *  \param AddShape
   *            Flag to add charge distribution shape information to
   *            the output table: object size(FWHM), elongation
   *            (ellipticity), spread along primary (longer) and
   *            secondary (shorter) axis, and orientation (direction
   *            of the primary axis in degrees).
   * 
   *  \param ShapeColumns
   *            Names for columns containing charge distribution shape
   *            information. Default names are "FWHM", "ELONG",
   *            "LENGTH", "WIDTH" and "ANGLE" 
   * 
   *  \param CopyImageHeader 
   *            Flag to copy the whole image header as a header of the
   *            object list table, if created. The default is FALSE (see
   *            also HeaderRecord parameter below).
   *
   *  \param HeaderRecords -
   *            Keywords of image fits header, which should be copied
   *            to the fits header of the object list. Default list
   *            consists of following records: DATE-OBS, JD, HJD, OBJECT,
   *                     TELESCOP, INSTRUME, OBSERVAT, CAMID, RA, DEC.
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   */

  class PixelClusterFinder : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new PixelClusterFinder ; }
    
    
    PixelClusterFinder() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images to be added
    string _inputCollectionName;

    //! Name of the output image collection
    string _outputCollectionName;

    //! Flag to copy the whole image header as a header of the object list.
    bool _copyImageHeader;
    
    //! Keywords of image fits header to copied to the table header
    std::vector<std::string> _copykey;

    //! Flag the output collection to be permanent
    bool _permanentCollection;


    //! Threshold for the cluster seed signal (in units of noise level)
    double _seedThreshold;

    //! Threshold for the cluster pixel signal (in units of noise level)
    double _pixThreshold;

    //! When set to true: search neighbours in 5x5 square rather than 3x3
    bool _moreNeighbours;

    //! Maximum number of pixels in the cluster
    /*! If more pixels are found in the cluster, they are not used in
     *  signal or CoG calculation. Nevertheless, they are not allowed to
     *  form separate clusters.
     */
    int _maxPixels;

    //! Maximum number of clusters. 
    /*! Cluster finding is interrupted if this number of clusters is
     *  found. As cluster seeds are sorted according to the signal,
     *  clusters with higher signal are likely (but not guaranteed) to
     *  be found first.
     */
    int _maxClusters;

    //! Background sampling factor
    /*! Parameter used to speed up background level calculation.
     *  Only every BackgroundSampling pixel is used.
     */
    int _backgroundSampling;

    //! Frame division for background calculation
    /*! Background is calculated for FrameDivision x FrameDivision 
     *  subimages and then interpolated between them
     */
    int _frameDivision;

    //! Background level interpolation flag
    //  Flag for interpolation between background levels calculated in
    //  different subframes: 
    //    0 - no interpolation, 
    //    1 -linear linterpolation (default).
    
    int _bgPol;

    //! Noise level interpolation flag
    //  Flag for interpolation between noise levels calculated in
    //  different subframes: 
    //    0 - no interpolation, 
    //    1 -linear linterpolation (default).
    
    int _noisePol;

    double BG_mean_av;
    double BG_rms_av;

    int sizeX;
    int sizeY;
    long sizeXY;

    //! Use network configuration file to update processor parameters 
    /*! When true, network configuration file is checked for each
     *  analysed image and processor parameters are changed if default
     *  parameter values are defined for given instrument/telescope.
     */
    bool _useTelescopeParameters;


    //! Flag for writing noise information to output table
    bool _addNoise;

    //! Column name for noise information
    string _noiseColumn;

    //! Flag for writing spread information to output table
    bool _addSpread;

    //! Column names for spread information
    std::vector<std::string> _spreadColumns;

    //! Flag for writing shape information to output table
    bool _addShape;

    //! Column names for shape information
    std::vector<std::string> _shapeColumns;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputLists ;

  } ;
  
} // end namespace luiza
#endif



