// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef SimpleImageAdder_h
#define SimpleImageAdder_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Simple processor to add images in the input collection.
  /*! Processor to add images in the input collection.
   *  No corrections for possible shifts or any other corrections are
   *  applied. 
   *
   *  \par Input 
   *  Processor takes images from Gloria data collection and adds them.
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be added
   *
   *  \param NumberAddImages  -
   *  Number of images to be added each time, 0 for all images in
   *  input collection (default is 0) 
   *
   *  \param UseManyLoops  -
   *  Wait for images in next loops, if not enough images in input
   *  collection (default is false) 
   *
   *  \param AverageImages  -
   *  Flag to select adding (when set to false) or averaging (when set
   *  to true) of input images (default is adding ie. false)
   *   
   *  \param MatchKeywords -
   *             Names for header keywords which should match for
   *             added frames. If any of these keywords changes, new 
   *             sum image is created, even if the required number of
   *             images is not processed.
   *
   *  \par Output
   *
   *  Added (or averaged) frames are stored in the output collection
   *
   *  \param OutputCollectionName - 
   *            Name of the collection where image sums will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   */

  class SimpleImageAdder : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new SimpleImageAdder ; }
    
    
    SimpleImageAdder() ;
    
    //! Called at the begin of the job before anything is read.
    /*! Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    //! Called before first data processing loop 
    /*! Can be used to initialize data collections required for
     *  analysis, eg. define permanent collections, or fill them with
     *  external data (from file, database or net). It is possible
     * as GloriaDataContainer is already defined, contrary to init().
     */
    virtual void startAnalysis( GloriaDataContainer* ) ; 

    //! Called for every data processing loop 
    /*! This is where most of the work (analysis) is done. 
     * Data is exchanged between processors via GloriaDataContainer 
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    //! Called for every loop - right after processData() call
    /*! 
     *  Can be used to check processing and/or produce check plots.
     */
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops.
    /*! This method is called to finalize the analysis, after all
     * input data was read (or required number of loops was
     * reached). Analysis of data collected in permanent data
     * collections is possible here.
     */
    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    //! Called after data processing is finished
    /*! Called after all input sources are closed for clean up.
     *  Notice that processors are called in the inverse order of the init()
     *  method so that resources allocated in the first processor also will 
     *  be available for all following processors.
     */
    virtual void end() ;
    
    
  protected:

    //! Storing summed image to output collection 
    void StoreSumImage( GloriaDataContainer* data );

    //! Final printout
    void printEndMessage() const ;

    //! Name of the image collection containing images to be added
    string _inputCollectionName;

    //! Name of the output image collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Number of images in the sum (0 for all in the input collection)
    int _nAddImages;

    //! Allow adding images from different loops.
    bool _useManyLoops;

    //! Flag to select adding (false) or averaging (true) of input images
    bool _averageImage;

    //! Keywords which should match for added frames
    std::vector<std::string> _matchKeywords;

    //! Values of keywords being matched
    std::vector<std::string> _matchValues;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputImages ;
    int _nAdded;
    GloriaFitsImage *_sumImage;

    int _nKey;
    std::string _initKeyVal;
    
  } ;
  
} // end namespace luiza
#endif



