// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef SimpleImageRotator_h
#define SimpleImageRotator_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Simple processor to rotate images in the input collection.
  /*! Processor to rotate images in the input collection and to
   *   preform some other basic operations like cropping or flipping.
   *  
   *  \par Input 
   *  Processor reads images from Gloria data collection and modifies
   *  them. Folowing order of operations is assumed: cropping,
   *  rebinning, rotating, flipping.
   *
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be added
   *
   *  \param WindowXY - 
   *         Image cropping window given as a vector of four numbers:
   *         X1 Y1 X2 Y2 (each changing from 1 to frame size)
   *
   *  \param RebinXY  -
   *         Parameter defining possible image rebinning of
   *         image. When two numbers (eg. N M) are specified, pixels
   *         are summed in NxM rectangle. If one number is given, same
   *         rebinning is assumed in X and Y.
   *
   *  \param RotDegree  -
   *        Flag to select set image rotation in degree. Allowe values
   *        are: 0 (default), 90, 180 and 270. 
   *
   *  \param FlipAxis -
   *         Flip the image along given axis: 1 for X, 2 for Y, 0 for
   *         no flip (default). Flip in X is equivalent to the change
   *         of vertical (Y) axis orientation.
   *   
   *  \par Output
   *
   *  Modified frames are stored in the output collection
   *
   *  \param OutputCollectionName - 
   *            Name of the collection where image sums will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   */

  class SimpleImageRotator : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new SimpleImageRotator ; }
    
    
    SimpleImageRotator() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images to be added
    string _inputCollectionName;

    //! Name of the output image collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Vector for optional crop region definition
    std::vector<int > _WindowXY;

    //! Vector for defining image rebinning 
    //! (different rebinning in X and Y possible)

    std::vector<int > _rebinXY;

    //! Flip the image along given axis
    int _flipAxis;

    //! Allow setting rotation (in degree): 90 (default), 180, 270 clockwise
    int _rotDegree;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputImages ;

    GloriaFitsImage *_origImage;
    GloriaFitsImage *_rotImage;


  } ;
  
} // end namespace luiza
#endif



