// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef WeightedAperturePhotometry_h
#define WeightedAperturePhotometry_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to calculate aperture photometry on the CCD frame
  /*! Processor to calculate aperture photometry on CCD images in the
   *  input collection. If input object collection is specified,
   *  photometry is calculated for all objects in the object list. For
   *  the targeted analysis, coordinates of the objects of interest
   *  can be specified. However, valid astrometry has to be found in
   *  the image fits header in that case. Two approaches for
   *  background calculation are possible: either it is calculated
   *  from media of pixels in annulus around object position (when
   *  \a AnnulusRadii is specified) or from the background map
   *  calculated for the whole image (as in \a BackgroundImage
   *  processor). The algorithm used differs from the one implemented
   *  in \a FastAperturePhotometry by introducing weights for pixels
   *  which are at the aperture boundary. The weight depends also on
   *  the exact position of the object.   
   *
   *  \par Input 
   *  Processor uses images from Gloria data collection
   *
   *  \param ImageCollectionName -
   *             Name of the collection containing input images
   *
   *  \param ListCollectionName -
   *             Name of the collection containing  GloriaObjectList
   *
   *  \param  ObjectPositions 
   *    Positions of the considered objects in (Ra,Dec), should be
   *    given in degrees (N x 2 numbers).
   *
   *  \par Algorithm parameters
   *  User can modify processor performance by settin the aperture
   *  radius (for signal counting) and annulus radii (background
   *  subtraction). Background is calculated as signal median from the
   *  pixels inside annulus.
   *
   *  \param  ApertureRadius -
   *           Radius of circular aperture used for signal
   *           counting in pixels. Distance is calculated from actual
   *           object position and appropriate weights are applied to
   *           pixels which are only partly contained in the defined
   *           aperture. Multiple values can be given resulting in
   *           calculation of multiple signal values.
   *
   *  \param  AnnulusRadii -
   *           Inner and outer radius of the annulus used for
   *           background subtraction (same for all apertures used!)
   *
   *  \param BackgroundSampling  - Background level calculation for
   *                       the whole image can be
   *                       sped up by considering only
   *                       1/BackgroundSampling of pixels.
   *
   *  \param FrameDivision - Background for the image is calculated for
   *                         FrameDivision x FrameDivision subimages
   *                         and then interpolated between them.
   *
   *  \param InterpolationOrder - polinomial order used to interpolate
   *                         between centers of subframes when
   *                         calculating background values on the image
   *
   *  \param  SignalThreshold -
   *           Threshold for the signal value, given in units of noise
   *           level (above background level). For signal below this
   *           threshold zero value is stored (or the cluster is
   *           removed from the list, if RemoveEmpty flag is set).
   *
   *  \param RemoveEmpty -
   *           Flag which indicates if clusters with signal below
   *           SignalThreshold should be removed from ObjectList or not.
   *
   *
   *  \par Output
   *  Object lists are stored in the output collection
   *
   *  \param BrightnessColumns - 
   *            Names of the columns in which object brightness should
   *            be stored ('Signal' by default). If one name is given
   *            for multiple apertures, aperture number is appended.
   * 
   *  \param MagnitudoColumns - 
   *            Names of the columns in which object instrumental magnitudo should
   *            be stored. If not given (default) magnitudo values are
   *            not calculated. 
   *
   *  \param MagnitudoOffset -
   *            Value of offset applied when calculating the
   *            instrumental magnitudo. Corresponds to the magnitudo
   *            of the object with signal of 1.
   * 
   *  \param OutputCollectionName - 
   *            Name of the collection where object table will be stored
   * 
   *  \param PermanentCollection  -
   *   Flag output collection, as permanent, not deleted after loop is
   *   finished (default is false)
   *
   *  \param CorrectPosition
   *            Flag to enable cluster position recalculation, based
   *            on the CoG of pixels within the aperture.
   * 
   *  \param PositionColumns
   *            Names for columns containing corrected
   *            position. Default is "CCD_X" and  "CCD_Y", which means
   *            that positions in the input object list will be corrected. 
   * 
   *  \param AddNoise
   *            Flag to add noise information to the output table.
   * 
   *  \param NoiseColumn
   *            Name for the noise column. Default name is "Noise" 
   * 
   *  \param AddSpread
   *            Flag to add charge spread information to the output
   *            table: RMS in X, RMS in Y and X-Y correlation
   *            calculated from pixels within aperture.
   * 
   *  \param SpreadColumns
   *            Names for columns containing charge spread
   *            information. Default names are "RMS_X", "RMS_Y" and
   *            "CORR_XY" 
   * 
   *  \param AddShape
   *            Flag to add charge distribution shape information to
   *            the output table: object size (FWHM), elongation
   *            (ellipticity), spread along primary (longer) and
   *            secondary (shorter) axis, and orientation (direction
   *            of the primary axis in degrees).
   * 
   *  \param ShapeColumns
   *            Names for columns containing charge distribution shape
   *            information. Default names are "FWHM", "ELONG",
   *            "LENGTH", "WIDTH" and "ANGLE" 
   * 
   *  \param CopyImageHeader 
   *            Flag to copy the whole image header as a header of the
   *            object list table, if created. The default is FALSE (see
   *            also HeaderRecord parameter below).
   *
   *  \param HeaderRecords 
   *            Keywords of image fits header, which should be copied
   *            to the fits header of the object list. This is done
   *            only if the new list is created. If brightness is
   *            added to the existing list, no records are
   *            copied. Default keyword list consists of following
   *            records: 
   *                  DATE-OBS, JD, HJD, OBJECT, TIME_UT
   *                  TELESCOP, INSTRUME, OBSERVAT, CAMID, RA, DEC.
   * 
   */

  class WeightedAperturePhotometry : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new WeightedAperturePhotometry ; }
    
    
    WeightedAperturePhotometry() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images 
    string _imageCollectionName;

    //! Name of the table collection containing object lists
    string _listCollectionName;

    //! Name of the output object list collection
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Radiuses of circular apertures used for signal counting in pixels
    std::vector<float> _apertureRadius;

    //! Inner and outer radius of the annulus used for background subtraction.
    std::vector<float> _annulusRadii;

    //! Background sampling factor
    /*! Parameter used to speed up background level calculation.
     *  Only every BackgroundSampling pixel is used.
     */
    int _backgroundSampling;

    //! Frame division for background calculation
    /*! Background is calculated for FrameDivision x FrameDivision 
     *  subimages and then interpolated between them
     */
    int _frameDivision;

    //! Polinomial order used to interpolate background map
    int _interpolationOrder; 

    //! Threshold for the cluster signal (in units of noise level)
    double _signalThreshold;

    //! Flags if clusters with signal below SignalThreshold should be removed
    bool _removeEmpty;

    //! Name of the column to store object signal 
    std::vector<std::string> _brightnessColumns;

    //! Column name for noise information
    std::vector<string> _noiseColumns;

    //! Name of the column to store instrumental magnitudo
    std::vector<std::string> _magnitudoColumns;

    //! Flag set to true, if magnitudo column name is given
    bool _addMagnitudo;
    
    //! Value of offset applied when calculating the instrumental magnitudo.
    double _magnitudoOffset;

    //! Object positions (two numbers, RA,DEC in degree, per object)
    std::vector<float> _objectPositions;

    //! Use network configuration file to update processor parameters 
    /*! When true, network configuration file is checked for each
     *  analysed image and processor parameters are changed if default
     *  parameter values are defined for given instrument/telescope.
     */
    bool _useTelescopeParameters;

    //! Flag indicating input method used
    bool _usingObjectList;


    //! Flag for calculating CoG position from pixels within aperture
    bool _correctPosition;

    //! Column names for corrected positions
    std::vector<std::string> _positionColumns;

    //! Flag for writing noise information to output table
    bool _addNoise;

    //! Flag for writing spread information to output table
    bool _addSpread;

    //! Column names for spread information
    std::vector<std::string> _spreadColumns;

    //! Flag for writing shape information to output table
    bool _addShape;

    //! Column names for shape information
    std::vector<std::string> _shapeColumns;

    //! Flag to copy the whole image header as a header of the object list.
    bool _copyImageHeader;
    
    //! Keywords of image fits header to copied to the table header
    std::vector<std::string> _copykey;

    //! Flag for including statistics for all radii
    bool _fillAllRadii;

    //! Flag indicating calculation of statistics for pixel within aperture
    bool _usingStatistics;

    //! Number of aperure radii used for signal calculation
    int _nAper;
    
    //! Pixel shifts for aperture definition (to speed up the code)
    std::vector<int> *_apShiftX;
    std::vector<int> *_apShiftY;
    std::vector<int> *_awShiftX;
    std::vector<int> *_awShiftY;

    //! Weights attributed to pixels at the aperture boundary
    /*! Weights are calculated for 9 object positions on the central
     *   pixel (3x3 grid: 0 for (0,0),  1 for (0.5,0) ... 8 for (1.,1))
     *   and then interpolated.
     */
    std::vector<double> *_awWeights;

    //! Pixel shifts for background annulus definition (to speed up the code)
    std::vector<int> _anShiftX;
    std::vector<int> _anShiftY;
    

    int _nLoops ;
    int _nInputImages ;
    int _nOutputLists ;

  } ;
  
} // end namespace luiza
#endif



