#include "ApplyGausKernel.h"
#include "gloria/GloriaMedianMap.h"
#include "gloria/GloriaClusterStatistics.h"


#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"
#include "gloria/GloriaAstrometry.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

using namespace std;
using namespace gloria;

// Some useful values

const double PI  = 3.14159265358979312;
const double LN2 = 0.693147180559945286;


namespace luiza
{

ApplyGausKernel aApplyGausKernel;

ApplyGausKernel::ApplyGausKernel() :
		Processor("ApplyGausKernel")
{
	_description = "Processor to calculate Gaus kernel, make convolution and search the stars on the frame.";

	// Input collections

	registerInputCollection(ImageCollection, "ImageCollectionName",
			"Name of the collection containing input images",
			_imageCollectionName, std::string(""));


	registerInputCollection(TableCollection, "InputTableOfParameters",
			"Name of the collection containing input PSF parameter table",
			_inputTableOfParameters, std::string(""));


	// Output collections

	registerOutputCollection(TableCollection, "OutputListCollectionName",
			"Name of the collection where object list will be stored",
			_outputListCollectionName, std::string(""));

	registerOutputCollection(ImageCollection, "OutputFrameCollectionName",
			"Name of the collection where convoluted frames will be stored",
			_outputFrameCollectionName, std::string(""));

	// Kernel parameters

	registerProcessorParameter(
			"VariableKernelFlag",
			"Flag to decide, whether kernel size is fixed by user or variable (true is variable)",
			_variableKernelFlag, true);

	registerOptionalParameter(
			"FixedKernelSize",
			"Kernel size, if fixed by user",
			_fixedKernelSize, (int)11 );

	// Object search parameters

	registerProcessorParameter(
			"PixelThresholdKernel",
			"Threshold for the star seed pixel signal on colvoluted frame (in units of noise level)",
			_pixThresholdKernel, (double)4);

	registerProcessorParameter(
			"PixelCut",
			"Cut for pixels included in star position calculation (in units of the seed pixel signal)",
			_pixCut, (double)0.05);

	registerProcessorParameter(
			"MinDistBetweenStars",
			"Minimum distance between seed pixels of two stars",
			_minDistBetweenStars, (double)3.0);

	registerProcessorParameter(
			"StarsCogRadius",
			"Radius for selecting pixels used for star position calculation",
			_starCogRadius, (double)3);

	registerOptionalParameter(
			"BackgroundSampling",
			"Only every BackgroundSampling pixel is considered in background calculation",
			_backgroundSampling, (int)1 );

	registerOptionalParameter(
			"BackgroundFrameDivision",
			"Frame division for background and noise calculation",
			_backgroundFrameDivision, (int)7 );

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for background map interpolation",
			_interpolationOrder, (int)1 );

	// Keywords to be copied to object list structure (follow FITS standard)

         std::vector<std::string> initkey;
         initkey.push_back(string("DATE-OBS"));
         initkey.push_back(string("TELESCOP"));
         initkey.push_back(string("INSTRUME"));
         initkey.push_back(string("OBSERVER"));
         initkey.push_back(string("OBJECT"));
 
        registerOptionalParameter(
               "HeaderRecords",
               "Keywords of image fits header, which should be copied to the fits header of the object list.",
               _copykey, initkey);
 
}//create kernel - register processor's parameters

void ApplyGausKernel::init()
{
	streamlog_out(MESSAGE) << "ApplyGausKernel::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputLists = 0;
	_nLoops = 0;
	// Check input and output parameters

	if (_imageCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input image collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if (_inputTableOfParameters.size() == 0)
	{
		streamlog_out(ERROR) << "No input image collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if ( _outputListCollectionName.size() == 0)
	{
	        streamlog_out(ERROR) << "No output object list collection name given! " << std::endl;
		streamlog_out(ERROR) << "Created object lists will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}

	if ( _outputFrameCollectionName.size() == 0)
	{
	        streamlog_out(ERROR) << "No output image collection name given! " << std::endl;
		streamlog_out(ERROR) << "Created convoluted frames will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}
}//end init()


void ApplyGausKernel::startAnalysis(GloriaDataContainer* data)
{
streamlog_out(DEBUG) << "Starting processor to convolution frame" << std::endl;
}//end startanalysis

void ApplyGausKernel::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_imageCollectionName.size() == 0)
		return;


	_nLoops++;


	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl;

	// Check if input collections exists

	if (!data->ImageCollectionExists(_imageCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_imageCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}


	streamlog_out(DEBUG) << "  Collection " << _imageCollectionName << " with "
			<< nImg << " images " << endl;

        if (!data->TableCollectionExists(_inputTableOfParameters))
	{
		streamlog_out(DEBUG) << "PSF parameter table collection not found ! " << endl;

		return;
	}

	int nTab = data->GetTableCollection(_inputTableOfParameters).size();

	if (nTab == 0)
	{
		streamlog_out(DEBUG) << "No PSF parameter tables in the input collection ! " << endl;

		return;
	}


	streamlog_out(DEBUG) << "  Collection " << _inputTableOfParameters << " with "
			<< nTab << " tables " << endl;



	if(nTab != nImg)
	  {
	        streamlog_out(WARNING) << "Number of PSF parameter tables does not match image number! " << std::endl;
		streamlog_out(WARNING) << "First parameter table will be used for all images! " << std::endl;
	  }


	// Create output collection if not found (and name is specified)

	if (_outputListCollectionName.size() > 0 && !data->TableCollectionExists(_outputListCollectionName) )
	  {
	   data->CreateTableCollection(_outputListCollectionName,_permanentList);
          }

        if(_outputFrameCollectionName.size() > 0 && !data->ImageCollectionExists(_outputFrameCollectionName)  ) 
          {
	    data->CreateImageCollection(_outputFrameCollectionName,_permanentFrame);
          }


        int ii, jj;
        vector<double> kernel;
        int size_ker, half_size_ker;
        int iii, jjj, kk, imax;


       //begin the loop over input frames    

       for (int ifm = 0; ifm < nImg; ifm++)
         {
          GloriaFitsImage *cur_img = &(data->GetImageCollection(_imageCollectionName).at(ifm));

	  // Skip empty images

	  int sizeX = cur_img->SizeX();
	  int sizeY = cur_img->SizeY();
	  long sizeXY = sizeX * sizeY;

	  if(sizeXY == 0) continue;


	  // Prepare the array of pixel values for input image

          double* floatPixels =  new double[sizeXY];

          if(cur_img->IsFloatImage())
           {
            double* oldPixels =  cur_img->GetFloatImageData();

            for(long ipix = 0; ipix < sizeXY; ipix++ )
              floatPixels[ipix]=oldPixels[ipix];                  
           }
          else
           {
            long* intPixels =  cur_img->GetIntImageData();

            for(long ipix = 0; ipix < sizeXY; ipix++ )
              floatPixels[ipix]=(double) intPixels[ipix];
           }
 

	  // Subtract background
	  // (Is it realy needed in the kernel based approach!?)

          GloriaMedianMap* NewMap = new GloriaMedianMap(_backgroundFrameDivision,_backgroundFrameDivision);
          NewMap->CalculateBgMap( cur_img, _backgroundSampling ); 

          GloriaFitsImage *BgImage = NewMap->GetBgImage();
          double* background = BgImage->GetFloatImageData();

          for(long ipix = 0; ipix < sizeXY; ipix++ )
             floatPixels[ipix] -= background[ipix];  // Subtract background here!!!


          //Open table with PSF parameters:

          GloriaFitsTable *source;

	  int itab = ifm;
	  if(nTab != nImg)itab=0;

	  source =  &(data -> GetTableCollection(_inputTableOfParameters).at(itab));  


          vector<double>* rms_x = &(source->GetVectorOf<double>("RMS_X"));// "RMS_X"
          vector<double>* rms_y = &(source->GetVectorOf<double>("RMS_Y"));//"RMS_Y" 
          vector<double>* cor_xy = &(source->GetVectorOf<double>("CORR_XY"));//"CORR_XY"


	  // Guess frame division for kernel calculation from the size of PSF parameter table

	  int _frameDivision = sqrt(rms_x->size()+0.5);


          // Create the convoluted frame

          GloriaFitsImage *conv = new GloriaFitsImage(sizeX,sizeY,true);
          double* work =  conv->GetFloatImageData();
         
	  // Subframe sizes
	  // Be careful for integer division!
	  // We need to cover the whole frame for all possible sizes...

          int dl_x = (sizeX+_frameDivision-1)/_frameDivision;
          int dl_y = (sizeY+_frameDivision-1)/_frameDivision;

	  // Kernel table

          vector<double> ker;

          for(int i=0; i<rms_x->size(); i++)
            {
             ker.push_back( (*rms_x)[i] );
             ker.push_back( (*rms_y)[i] );
            }
          sort( ker.begin(), ker.end() );
          double smax = ker[ ker.size()-1 ];
        
          int n_sigma = 2;
 
          if(_variableKernelFlag == true)
           {
            size_ker = 2*n_sigma*smax+1.5;
           }
          else
           {
            size_ker = _fixedKernelSize;
           }

	  // Do not allow for kernel size smaller than 7

          if(size_ker<7) size_ker=7;


	  // Kernel size has to be odd

          if(size_ker % 2 == 0) size_ker++; 


          half_size_ker = size_ker/2;

          streamlog_out(DEBUG)<< " Kernel size used = "<<size_ker<<endl;


          // Calculate convoluted frame, in _frameDivision x _frameDivision parts

          for(int a = 0; a < _frameDivision; a++)
            {
             for(int b = 0; b < _frameDivision; b++)
               {
                double norm, c1, c2, c3, cz, sx, sy, sxy, gaus, gaus_sum, n_gaus, fwhm;

                kernel.clear();

                n_gaus = gaus_sum = 0.;

                int m = a+_frameDivision*b;

                sx = (*rms_x)[m];
                sy = (*rms_y)[m];
                sxy = (*cor_xy)[m];

                fwhm = 2*sqrt(LN2*(sx+sy));

                norm = 1/(2*PI*sx*sy*sqrt( 1-pow(sxy, 2)));
                cz = -1/(2*(1-pow(sxy, 2)));

                for(int d1 = -half_size_ker; d1 <= half_size_ker; d1++)
                  {
                   for(int d2 = -half_size_ker; d2 <= half_size_ker; d2++)
                     {
                      double c1, c2, c3, gaus;
 
                      c1 = d1*d1/(sx*sx);
                      c3 = d2*d2/(sy*sy);
                      c2 = -2*sxy*d1*d2/(sx*sy);

		      // This cut seems to be a little bit to tight (?!)

                      if(  cz*(c1+c2+c3) > -4.0 )
                        {
                        gaus = norm*exp( cz*(c1+c2+c3) );
                        kernel.push_back( gaus );             
                        gaus_sum+=gaus;
                        n_gaus++;
                        }
		      else
                        kernel.push_back(0); 
                 
                      }
                   }    

		// Shift gauss values so the sum is zero

                gaus_sum/=n_gaus;

                for(int tt=0; tt<kernel.size(); tt++)
                    if(kernel[tt]>0) kernel[tt]-=gaus_sum; 
               
               // actual convolution of the (sub)frame
	       // Loop over subframe pixels

               for(int i = 0; i <=dl_x; i++)
                  {
                   for(int j = 0; j<=dl_y; j++)
                     {
		      // Position on the frame
                      ii = i + dl_x*a;
                      jj = j + dl_y*b;

		      // Check range

                      if(ii <0 || jj < 0 || ii > (sizeX-1) || jj > (sizeY-1) ) continue;

                      double suma = 0.;

                      for(int ker_x = -half_size_ker; ker_x <= half_size_ker; ker_x++)
                        {
			 if(ii+ker_x < 0 || ii+ker_x > (sizeX-1)) continue;

                         for(int ker_y = -half_size_ker; ker_y <= half_size_ker; ker_y++)
                           {
			   if(jj+ker_y < 0 || jj+ker_y > (sizeY-1)) continue;

                            suma+=floatPixels[(ii+ker_x)+sizeX*(jj+ker_y)]*kernel[ker_x+half_size_ker+size_ker*(ker_y+half_size_ker)];
                           }
                        }
                      work[ii+sizeX*jj] = suma;
                     }//end loop for( int dl_x)
                  }//end loop for( int dl_y)

             }//end loop for( int b)
          }  //end loop for( int a)


	  // Image 'conv' and table 'work' contain convoluted frame now !
	  //-------------------------------------------------------------

	  // Look for stars in the convoluted frame, if the output object list name specified

         if( _outputListCollectionName.size() > 0 )
           {

	   // Estimate background (should be consistent with zero) and noise on convoluted frame

           GloriaMedianMap* NewMap2 = new GloriaMedianMap(_backgroundFrameDivision,_backgroundFrameDivision);
           NewMap2->CalculateBgMap( conv, _backgroundSampling  );

           GloriaFitsImage *NoiseImage = NewMap2->GetNoiseImage();
           double* noise = NoiseImage->GetFloatImageData();

	  // Object table, to be filled with stars found on the convoluted frame

           GloriaObjectList *stars = new GloriaObjectList();

          // Copy selected header information

           int ncopy=_copykey.size();
 
           for(int icopy=0; icopy<ncopy; icopy++)
             if(cur_img->IsKeywordSet(_copykey[icopy].c_str()))
                stars->AddKeyword(cur_img->GetKeywordRecord(_copykey[icopy].c_str()));

	  // Pixel cluster statistics - to calculated star position

          GloriaClusterStatistics stat;

          //loop over the convoluted image to search the stars       

           long nclust = 0;

           for(long i = 10; i < sizeX-10; i++)
             {
              for(long j=10; j<sizeY-10; j++)
                {
                 long m=i+sizeX*j;

                 if( (work[m] > _pixThresholdKernel * noise[m])  && //value of pixel is higher than THR sigma 
                        (work[m]>work[m-1]) && (work[m]>work[m+1]) && (work[m]>work[m-sizeX]) && 
                        (work[m]>work[m+sizeX]) && //pixel value is higher than the neighbors
                            (work[m]>work[m-sizeX-1]) && (work[m]>work[m-sizeX+1]) && (work[m]>work[m+sizeX-1]) && 
                            (work[m]>work[m+sizeX+1]) ) //pixel value is higher than the diagonal neighbors
                      {

                       //check if there are no higher maxima nearby

		       double max = work[m]; 
                       imax = m;
                       int ir = _minDistBetweenStars+0.5;
                       for(int ii = -ir; ii <=ir; ii++)
                         {
                          iii = ii+i;
                          if(iii<0 || iii >= sizeX) continue;
                          for(int jj = -ir; jj <=ir; jj++)
                            {                   
                             if(ii==0 && jj ==0 )continue;
                             jjj = jj+j;
                             if(jjj<0 || jjj >= sizeY) continue;
                             kk = iii + jjj*sizeX;
                             if(work[kk] > max)
                              {
                               max = work[kk];
                               imax = kk;
                               break;
                              }
                            }  
                         }
                       if(m != imax) continue;

		       // Set position of the seed pixel as the initial cluster position

                       double x_c = i;
                       double y_c = j;

                       double thr = _pixCut*floatPixels[m];

                       for(int it=0; it<2; it++)
                         {
                          stat.Clear();
                          for(int ii=(x_c-_starCogRadius); ii<=x_c+_starCogRadius; ii++)
                            {
                             for(int jj=(y_c-_starCogRadius); jj<=y_c+_starCogRadius; jj++)
                               {
                                if(ii>=sizeX || jj>=sizeY) continue;
				if(ii<0 || jj<0) continue;

                                double dd = sqrt( pow(x_c-ii, 2)+ pow( y_c-jj, 2) );

                                if(dd > _starCogRadius) continue;

				if(floatPixels[ii+sizeX*jj]<thr && dd > 1) continue;

                                if(work[ii+sizeX*jj]>0 )
                                       stat.AddPixel(ii, jj, work[ii+sizeX*jj] );

                              }
                            }//loop ii jj

                          x_c = stat.GetXpos();
                          y_c = stat.GetYpos();
                         }
                          // loop it : CoG iterations

                       stars->AddObject( stat.GetXpos()+1, stat.GetYpos()+1, stat.GetSignal() );
                       nclust++;
                    } // seed pixel cuts if

                }//end loop for long j
             }//end loop for long i

           streamlog_out(MESSAGE) << " Number of stars found = "<< nclust << endl;

           (data->GetTableCollection(_outputListCollectionName)).push_back(*stars);

           delete stars;
           delete NewMap2;
           delete NoiseImage;
  	   }

         
	 if(_outputFrameCollectionName.size()>0)
             (data->GetImageCollection(_outputFrameCollectionName)).push_back(*conv);

         delete conv;
         delete floatPixels;          
         delete NewMap;
	 delete BgImage;

         }//end loop of images
 setReturnValue(true);
 //     stars->Print(cout,16);

}//end processData

void ApplyGausKernel::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void ApplyGausKernel::end()
{

	printEndMessage();

}

void ApplyGausKernel::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << _nInputImages << " images processed, " 
			       << _nOutputLists << " object lists stored " << std::endl;

}

}//namespace luiza

