#include "AstroImageAdder.h"


#include "luiza/Exceptions.h"
#include <iostream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  AstroImageAdder aAstroImageAdder ;


  AstroImageAdder::AstroImageAdder() : Processor("AstroImageAdder") {

    _description = "Processor to add images in the input collection."
                   "Astrometry is used to correct for possible shifts." ;

   registerInputCollection (  ImageCollection,
                     "InputCollectionName",
                     "Name of the collection containing images to be added",
                     _inputCollectionName,  std::string ("") );


   registerOutputCollection (  ImageCollection,
                     "OutputCollectionName",
                     "Name of the collection where image sums will be stored",
                     _outputCollectionName,  std::string ("") );


   registerOptionalParameter ("PermanentCollection",
   "Flag output collection, as permanent, not deleted after loop is finished",
                              _permanentCollection,  false );


   registerOptionalParameter ("NumberAddImages",
"Number of images to be added each time (0 for all images in input collection)",
                              _nAddImages,  static_cast < int > (0) );

   registerOptionalParameter ("UseManyLoops",
   "Wait for images in next loops, if not enough images in input collection",
                              _useManyLoops,  false );

	registerOptionalParameter(
			"FrameDivision",
			"Frame division for shift map calculation",
			_frameDivision, (int)1 );

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for shift map interpolation",
			_interpolationOrder, (int)1 );

 }


 void AstroImageAdder::init() { 

    streamlog_out( MESSAGE ) << "AstroImageAdder::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nInputImages = 0 ;
    _nOutputImages = 0 ;
    _nAdded = 0;
    _nLoops = 0 ;

    _sumImage = NULL;
    _sumCount = NULL;

   // Check input parameters

    if(_inputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No input collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Nothing to do! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    if(_outputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No output collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Adding images not possible! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

    if(_useManyLoops && _nAddImages<=0 )
      {
      streamlog_out( ERROR ) << "UseManyLoops not allowed when NumberAddImages not specified! " << std::endl;
      streamlog_out( ERROR ) << "UseManyLoops option changed to false! " << std::endl;

      _useManyLoops = false;
      }

 
 
    return;
 
  }


void AstroImageAdder::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "AstroImageAdder::startAnalysis()" << endl;
}




  void AstroImageAdder::processData( GloriaDataContainer* data ) 
{ 

  // Return if input or output collection not specified
  // Error already printed in init()

    if(_inputCollectionName.size()==0 )return;

    if(_outputCollectionName.size()==0 )return;



    _nLoops ++ ;

    streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops << std::endl ;

    // Check if input collection exists
 
     if(! data -> ImageCollectionExists(_inputCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      
     int nImg = data -> GetImageCollection(_inputCollectionName).size();

     if(nImg == 0)
	  {
          streamlog_out(DEBUG) << "No images in input collection ! " << endl ;

          return;
	  }
      
     streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with " << nImg << " images " << endl ;


    // Create output collection if not found

    if(! data -> ImageCollectionExists(_outputCollectionName))
         data -> CreateImageCollection(_outputCollectionName,_permanentCollection);

    // Loop over images in input collection

     for(int ifm=0; ifm<nImg; ifm++)
        {
          GloriaFitsImage *cur_image = &(data->GetImageCollection(_inputCollectionName).at(ifm));

	  // Skip empty images

	  int sizeX = cur_image->SizeX();
	  int sizeY = cur_image->SizeY();

	  int sizeXY = sizeX * sizeY;

	  if(sizeXY == 0) continue;


	  // Check if astrometry is available for this image

	  GloriaAstrometry ga;

   	  if(!ga.IsAstrometryPresent(cur_image))
	    {
	    streamlog_out(WARNING) << "Astrometry not found in the file header: " << cur_image->InputFileName() << endl;
	    continue;
	    }

         // Read astrometry from header 

	 ga.InitFromHeader(cur_image);


         if(_nAdded == 0)
	   {
	     _sumSizeX = sizeX;
	     _sumSizeY = sizeY;

	     // Starting image

             _sumImage = new GloriaFitsImage(*cur_image);

  	     if(_sumImage == NULL)
	       {
   	       streamlog_out(ERROR) << "Error in sum image allocation " << endl ;
	       streamlog_out(ERROR) << _sumImage->ErrorMessage() << endl ;
	       return;
	       }

	     // Force sum image to be float

	     if(_sumImage->IsIntImage())_sumImage->ConvertToFloat();

	     _sumImageData = _sumImage->GetFloatImageData();

	     // Additional image to store pixel counts

	     _sumCount = new GloriaFitsImage(sizeX,sizeY, false);  

  	     if(_sumCount == NULL)
	       {
   	       streamlog_out(ERROR) << "Error in count image allocation " << endl ;
	       streamlog_out(ERROR) << _sumCount->ErrorMessage() << endl ;
	       return;
	       }


	     _sumCountData = _sumCount->GetIntImageData();

	     for(int ipix=0; ipix<sizeXY; ipix++)
	       _sumCountData[ipix]=1;

   	    _sumAstro.InitFromHeader(cur_image);

            streamlog_out(DEBUG) << "New average started from image " << cur_image->InputFileName() << endl;
	   }


         if(_nAdded > 0)
	   {

           streamlog_out(DEBUG) << "   new image added: " << cur_image->InputFileName() << endl;

           // Prepare shift maps

	   _shiftMaps = new GloriaValueMap(cur_image,_frameDivision,_frameDivision,2);

	   double* shiftX = _shiftMaps -> GetMap(0);
	   double* shiftY = _shiftMaps -> GetMap(1);
           
	   for(int ibx=0; ibx < _frameDivision; ibx++)
  	     for(int iby=0; iby < _frameDivision; iby++)
	       {
		 int ibin = ibx + _frameDivision*iby;

		 double ipx = sizeX * (ibx+0.5)/_frameDivision;
		 double ipy = sizeY * (iby+0.5)/_frameDivision;

		 double ra,dec;
		 ga.GetRaDecFromXY(ipx, ipy, &ra, &dec);

                 double isx,isy;
		 _sumAstro.GetXYFromRaDec(ra, dec, &isx, &isy);

		 shiftX[ibin] = isx-ipx;
		 shiftY[ibin] = isy-ipy;

		 //		 streamlog_out(DEBUG) << "  bin "<< ibx << "," << iby << " => " <<
		 //		  ipx << "," << ipy << " => " << ra << "," << dec << " => " << isx << "," << isy 
		 //                      << " => Shift: " << shiftX[ibin] << ", "<< shiftY[ibin] << endl;
	       }

	   int icent = _frameDivision*_frameDivision/2;
	   streamlog_out(DEBUG) << "   with shift (in central bin):  " << shiftX[icent] << ", "<< shiftY[icent] << endl;

	   // Create shift images

           GloriaFitsImage *shiftXimage =  _shiftMaps->GetMapImage(0,_interpolationOrder);
           GloriaFitsImage *shiftYimage =  _shiftMaps->GetMapImage(1,_interpolationOrder);

           double *shiftXdata  = shiftXimage->GetFloatImageData();
           double *shiftYdata  = shiftYimage->GetFloatImageData();

	   // Add images

	   long *intData = NULL;
	   double *floatData = NULL;
           if(cur_image->IsIntImage())
                intData = cur_image->GetIntImageData();
	   else
                floatData = cur_image->GetFloatImageData();

	   for(int ipx = 0; ipx<sizeX; ipx++)
  	    for(int ipy = 0; ipy<sizeY; ipy++)
	     {
	       long ipix = ipx + sizeX*ipy;

	       double dx = shiftXdata[ipix];
	       double dy = shiftYdata[ipix];

               if(ipx + dx + 0.5 < 0.) continue;
               if(ipy + dy + 0.5 < 0.) continue;

               int isx = ipx + dx +0.5;
               int isy = ipy + dy +0.5;

	       if(isx >= _sumSizeX ) continue;
	       if(isy >= _sumSizeY ) continue;

	       long isum = isx + _sumSizeX*isy;

               if(intData)
                   _sumImageData[isum]+=intData[ipix];
	       else
                   _sumImageData[isum]+=floatData[ipix];

	       _sumCountData[isum]++;
	     }


	   delete shiftXimage;
           delete shiftYimage;

	   delete _shiftMaps;

	   } // end of:  if(_nAdded > 0)

          _nInputImages++;
          _nAdded++;

	  // Take next image if too few images summed so far

	  if(_nAdded != _nAddImages && 
	     (ifm < nImg-1 || _useManyLoops))continue; 
             
	  // Normalize 

	  for(long isum = 0 ; isum < _sumSizeX*_sumSizeY; isum++)
	    if(_sumCountData[isum])_sumImageData[isum]/=_sumCountData[isum];

	  // Add summed image to output collestion

          (data->GetImageCollection(_outputCollectionName)).push_back(*_sumImage);

          delete _sumImage;
          delete _sumCount;

          _sumImage = NULL;
          _sumCount = NULL;

          _nAdded = 0;


          _nOutputImages++;
	  }
  

    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
  }
  
  void AstroImageAdder::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
void AstroImageAdder::concludeAnalysis( GloriaDataContainer* data )
{

	streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;

	if(_nAdded == 0 ) return;

	// Last (partial) sum has not been written yet!

        streamlog_out(MESSAGE) << " Writing last, uncomplete sum of " << _nAdded << " images " << endl;

        // Normalize 

	  for(long isum = 0 ; isum < _sumSizeX*_sumSizeY; isum++)
	    if(_sumCountData[isum])_sumImageData[isum]/=_sumCountData[isum];

	  // Add summed image to output collestion

          (data->GetImageCollection(_outputCollectionName)).push_back(*_sumImage);

	  // Flag collection as permanent, so file is written out

	  data->SetImageCollectionPermanent(_outputCollectionName);

          delete _sumImage;
          _sumImage = NULL;

          _nAdded = 0;

          _nOutputImages++;
}

void AstroImageAdder::end(){ 
    
    printEndMessage() ;

  }

  void AstroImageAdder::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " Added "     << _nInputImages << " images "<< std::endl ; 

    streamlog_out(MESSAGE) << " Stored "    << _nOutputImages << " summed frames " << std::endl << std::endl;
    
  }
  
}// namespace luiza
