#include "BackgroundImage.h"
#include "gloria/GloriaMedianMap.h"


#include "luiza/Exceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

BackgroundImage aBackgroundImage;

BackgroundImage::BackgroundImage() :
		Processor("BackgroundImage")
{

	_description = "Processor to create background/noise image from the frame.";

	registerInputCollection(ImageCollection, "InputCollectionName",
			"Name of the collection containing input images",
			_inputCollectionName, std::string(""));

	registerOutputCollection(ImageCollection, "BackgroundCollectionName",
			"Name of the collection where background image will be stored",
			_backgroundCollectionName, std::string(""));

	registerOutputCollection(ImageCollection, "NoiseCollectionName",
			"Name of the collection where noise image will be stored",
			_noiseCollectionName, std::string(""));

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);

	registerOptionalParameter(
			"BackgroundSampling",
			"Only every BackgroundSampling pixel is considered in background calculation",
			_backgroundSampling, (int)1 );

	registerOptionalParameter(
			"FrameDivision",
			"Frame division for background calculation",
			_frameDivision, (int)1 );

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for background map interpolation",
			_interpolationOrder, (int)1 );

	registerOptionalParameter(
			"UseTelescopeParameters",
			"Use network configuration file to update processor parameters on even by event basis",
			_useTelescopeParameters, false );

}


void BackgroundImage::init()
{

	streamlog_out(MESSAGE) << "BackgroundImage::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputImages = 0;
	_nLoops = 0;


	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}

	if (_backgroundCollectionName.size() == 0 && _noiseCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No output collection name given! "
				<< std::endl;
		streamlog_out(ERROR) << "Output images will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}

	return;

}

void BackgroundImage::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->ImageCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_inputCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nImg << " images " << endl;

	// Create output collection if not found (and name is specified)
        // Background image

	if (_backgroundCollectionName.size() > 0  &&
	   !data->ImageCollectionExists(_backgroundCollectionName))
	  {
	  data->CreateImageCollection(_backgroundCollectionName,
	  			_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _backgroundCollectionName << " created " << endl;

	  }

	// Noise image

	if (_noiseCollectionName.size() > 0  &&
	   !data->ImageCollectionExists(_noiseCollectionName))
	  {
	  data->CreateImageCollection(_noiseCollectionName,
	  			_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _noiseCollectionName << " created " << endl;

	  }

	// Loop over images in input collection
	//=====================================

	for (int ifm = 0; ifm < nImg; ifm++)
	  {
          GloriaFitsImage *cur_img = &(data->GetImageCollection(_inputCollectionName).at(ifm));

	  // Skip empty images

	  sizeX = cur_img->SizeX();
	  sizeY = cur_img->SizeY();

	  sizeXY = sizeX * sizeY;

	  if(sizeXY == 0) continue;

	  _nInputImages++;

	  // Update parameters for this telescope

	  if(_useTelescopeParameters)
                          updateTelescopeParameters(cur_img);


	  // Calculate background and noise images
	  // Use MediamMap class

	  GloriaMedianMap* BgMap = new GloriaMedianMap( _frameDivision, _frameDivision);
	  BgMap->CalculateBgMap( cur_img, _backgroundSampling);

	  if (_backgroundCollectionName.size() > 0  && data->ImageCollectionExists(_backgroundCollectionName))
	     {
	      GloriaFitsImage *BgImage = BgMap->GetBgImage(_interpolationOrder);

	      (data->GetImageCollection(_backgroundCollectionName)).push_back(*BgImage);
	     _nOutputImages++;

              delete BgImage;
	     }

	  if (_noiseCollectionName.size() > 0  && data->ImageCollectionExists(_noiseCollectionName))
	     {
	      GloriaFitsImage *NoiseImage = BgMap->GetNoiseImage(_interpolationOrder);

	      (data->GetImageCollection(_noiseCollectionName)).push_back(*NoiseImage);
	     _nOutputImages++;

              delete NoiseImage;
	     }

	  delete BgMap;   
    

    } //	for (int ifm = 0; ifm < nImg; ifm++) - loop over images in input collection


  // always return true  for ProcessorName
  setReturnValue(true);

}

void BackgroundImage::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void BackgroundImage::end()
{

	printEndMessage();

}

void BackgroundImage::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << _nInputImages << " images processed " 
			       << _nOutputImages << " stored to file " << std::endl;

}



} // namespace luiza
