#include "CalculateGausPSF.h"
#include "gloria/GloriaMedianMap.h"
#include "gloria/GloriaClusterStatistics.h"

#include <string>
#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"
#include "gloria/GloriaAstrometry.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

using namespace std;
using namespace gloria;
namespace luiza
{

CalculateGausPSF aCalculateGausPSF;

CalculateGausPSF::CalculateGausPSF() :
		Processor("CalculateGausPSF")
{

	_description = "Processor to calculate star profile parameters on the frame.";

	registerInputCollection(ImageCollection, "ImageCollectionName",
			"Name of the collection containing input images",
			_imageCollectionName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where parameter map will be stored",
			_outputCollectionName, std::string(""));


	registerProcessorParameter(
			"PixelThreshold",
			"Signal threshold for the pixel to be considered as a star center (in units of noise level)",
			_pixThreshold, (double)5.);

	registerProcessorParameter(
			"MedianThreshold",
			"Cut on the pixel signal median (in units of noise level) for defining star size",
			_medianThreshold, (double)1.5);

	registerProcessorParameter(
			"MaxPixelValue",
			"Maximum signal for the pixel used in shape analysis (to avoid saturation effects)",
			_MaxPixelValue, (double)48000.);

	registerOptionalParameter(
			"BackgroundSampling",
			"Only every BackgroundSampling pixel is considered in background calculation",
			_backgroundSampling, (int)1 );//ok

	registerOptionalParameter(
			"BackgroundFrameDivision",
			"Frame division for background and noise calculation",
			_backgroundFrameDivision, (int)8 );

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for background map interpolation",
			_interpolationOrder, (int)1 );

	registerOptionalParameter(
			"FrameDivision",
			"Frame division for PSF parameter map definition",
			_frameDivision, (int)5 );

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output table collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);

	registerOptionalParameter(
			"UseTelescopeParameters",
			"Use network configuration file to update processor parameters on even by event basis",
			_useTelescopeParameters, false );

}


void CalculateGausPSF::init()
{

	streamlog_out(MESSAGE) << "CalculateGausPSF::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputMaps = 0;
	_nLoops = 0;


	// Check input and output parameters

	if (_imageCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input image collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}



	if ( _outputCollectionName.size() == 0)
	{
	        streamlog_out(ERROR) << "No output collection name given! " << std::endl;
		streamlog_out(ERROR) << "Created parameter maps will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}


 
}


void CalculateGausPSF::startAnalysis(GloriaDataContainer* data)
{
streamlog_out(DEBUG) << "Starting data analysis..." << std::endl;
}



void CalculateGausPSF::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_imageCollectionName.size() == 0)
		return;


	_nLoops++;


	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collections exists

	if (!data->ImageCollectionExists(_imageCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_imageCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}


	streamlog_out(DEBUG) << "  Collection " << _imageCollectionName << " with "
			<< nImg << " images " << endl;


	// Create output collection if not found (and name is specified)

	if ( _outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
	  data->CreateTableCollection(_outputCollectionName,
	  			_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;

	  }

// Definition of considered pixel configurations - apertures of incrising radius

 long nbShiftR[11] = {1,8,16,20,24,40,36,48,56,56,68};
 long nbShiftX[11][70] = {
  {0},
  {-1,0,1,-1,1,-1,0,1},
  {-2,-1,0,1,2,-2,2,-2,2,-2,2,-2,-1,0,1,2},
  {-2,-1,0,1,2,-3,3,-3,3,-3,3,-3,3,-3,3,-2,-1,0,1,2},
  {-2,-1,0,1,2,-3,3,-4,4,-4,4,-4,4,-4,4,-4,4,-3,3,-2,-1,0,1,2},
  {-3,-2,-1,0,1,2,3,-4,-3,3,4,-5,-4,4,5,-5,5,-5,5,-5,5,-5,5,-5,5,-5,-4,4,5,-4,-3,3,4,-3,-2,-1,0,1,2,3},
  {-3,-2,-1,0,1,2,3,-4,4,-5,5,-6,6,-6,6,-6,6,-6,6,-6,6,-6,6,-6,6,-5,5,-4,4,-3,-2,-1,0,1,2,3},
  {-3,-2,-1,0,1,2,3,-5,-4,4,5,-6,-5,5,6,-6,6,-7,7,-7,7,-7,7,-7,7,-7,7,-7,7,-7,7,-6,6,-6,-5,5,6,-5,-4,4,5,-3,-2,-1,0,1,2,3},
  {-4,-3,-2,-1,0,1,2,3,4,-5,-4,4,5,-6,6,-7,7,-8,-7,7,8,-8,8,-8,8,-8,8,-8,8,-8,8,-8,8,-8,8,-8,-7,7,8,-7,7,-6,6,-5,-4,4,5,-4,-3,-2,-1,0,1,2,3,4},
  {-4,-3,-2,-1,0,1,2,3,4,-5,5,-7,-6,6,7,-7,7,-8,8,-9,9,-9,9,-9,9,-9,9,-9,9,-9,9,-9,9,-9,9,-9,9,-8,8,-7,7,-7,-6,6,7,-5,5,-4,-3,-2,-1,0,1,2,3,4},
  {-4,-3,-2,-1,0,1,2,3,4,-6,-5,5,6,-7,-6,6,7,-8,8,-9,-8,8,9,-9,9,-10,10,-10,10,-10,10,-10,10,-10,10,-10,10,-10,10,-10,10,-10,10,-9,9,-9,-8,8,9,-8,8,-7,-6,6,7,-6,-5,5,6,-4,-3,-2,-1,0,1,2,3,4}
};
  long nbShiftY[11][70] = {
  {0},
  {-1,-1,-1,0,0,1,1,1},
  {-2,-2,-2,-2,-2,-1,-1,0,0,1,1,2,2,2,2,2},
  {-3,-3,-3,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,3,3,3},
  {-4,-4,-4,-4,-4,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,4,4,4,4,4},
  {-5,-5,-5,-5,-5,-5,-5,-4,-4,-4,-4,-3,-3,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,3,3,4,4,4,4,5,5,5,5,5,5,5},
  {-6,-6,-6,-6,-6,-6,-6,-5,-5,-4,-4,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,4,4,5,5,6,6,6,6,6,6,6},
  {-7,-7,-7,-7,-7,-7,-7,-6,-6,-6,-6,-5,-5,-5,-5,-4,-4,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,4,4,5,5,5,5,6,6,6,6,7,7,7,7,7,7,7},
  {-8,-8,-8,-8,-8,-8,-8,-8,-8,-7,-7,-7,-7,-6,-6,-5,-5,-4,-4,-4,-4,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,4,4,4,4,5,5,6,6,7,7,7,7,8,8,8,8,8,8,8,8,8},
  {-9,-9,-9,-9,-9,-9,-9,-9,-9,-8,-8,-7,-7,-7,-7,-6,-6,-5,-5,-4,-4,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,7,7,8,8,9,9,9,9,9,9,9,9,9},
  {-10,-10,-10,-10,-10,-10,-10,-10,-10,-9,-9,-9,-9,-8,-8,-8,-8,-7,-7,-6,-6,-6,-6,-5,-5,-4,-4,-3,-3,-2,-2,-1,-1,0,0,1,1,2,2,3,3,4,4,5,5,6,6,6,6,7,7,8,8,8,8,9,9,9,9,10,10,10,10,10,10,10,10,10}
};
 long Nnb = 11;

 // Table to store shape parameters from all stars found

 GloriaFitsTable* gaus_parameters = new GloriaFitsTable();

 TableColumnType coltype = DoubleColumn; 
 gaus_parameters->AddColumn("X_C", coltype);
 gaus_parameters->AddColumn("Y_C", coltype);
 gaus_parameters->AddColumn("SSXX", coltype);
 gaus_parameters->AddColumn("SSYY", coltype);
 gaus_parameters->AddColumn("SSXY", coltype);
 gaus_parameters->AddColumn("XFWHM", coltype);
 gaus_parameters->AddColumn("ORIENTATION", coltype);
 gaus_parameters->AddColumn("ELONGATION", coltype);


 // Methods to calculate shape parameters from star pixels

 GloriaClusterStatistics stat;

 // Other variables used

 long npix=0;
 double x_c, y_c, sxx, sxy, syy, fwhm;
 vector<double>vec_pix;

 long SizeX, SizeY;


// Loop over images in input collection
//=====================================

        
	for (int ifm = 0; ifm < nImg; ifm++)
	  {
          GloriaFitsImage *cur_img = &(data->GetImageCollection(_imageCollectionName).at(ifm));

	  int sizeX = cur_img->SizeX();
	  int sizeY = cur_img->SizeY();

	  if(sizeX < 41 || sizeY < 41)
	    {
	    streamlog_out(WARNING) << "Image in input collection too small (" << sizeX << " x " << sizeY << "). Skiping image " << ifm << endl;

	    continue;
	    }

          if(ifm == 0)
           {
            SizeX = sizeX;
            SizeY = sizeY;
           }

	  if(sizeX != SizeX || sizeY != SizeY)
	    {
            streamlog_out(WARNING) << "Images in input collection are of different size ! Skiping image " << ifm << endl;

	    continue;
	    }

	  long sizeXY = sizeX * sizeY;
         
          double* floatPixels =  new double[sizeXY];

          if(cur_img->IsFloatImage())
           {
            double* oldPixels =  cur_img->GetFloatImageData();
            for(long ipix = 0; ipix < sizeXY; ipix++ )
              floatPixels[ipix]=oldPixels[ipix];                  
           }
          else
           {
            long* intPixels =  cur_img->GetIntImageData();
            for(long ipix = 0; ipix < sizeXY; ipix++ )
            floatPixels[ipix]=(double) intPixels[ipix];
           }

      // Calculate background and noise
      // Use MediamMap class

          GloriaMedianMap* BgMap = new GloriaMedianMap(_backgroundFrameDivision,_backgroundFrameDivision);
          BgMap->CalculateBgMap( cur_img,  _backgroundSampling); 

          GloriaFitsImage *BgImage = BgMap->GetBgImage();
          GloriaFitsImage *NoiseImage = BgMap->GetNoiseImage();

      // Calculated background and noise tables

          double* background = BgImage->GetFloatImageData();
          double* noise = NoiseImage->GetFloatImageData();

          for(long ipix = 0; ipix < sizeXY; ipix++ )
            {
             floatPixels[ipix] -= background[ipix];  // Subtract background here!!!
            }


	  // Search for stars in the frame (neglecting edges of 20 pixel)
          
 	  for(long i = 20; i < sizeX-20; i++)
    	    {
    	     for(long j=20; j<sizeY-20; j++)
       	       {
                long m=i+sizeX*j;

                if( (floatPixels[m] > (_pixThreshold) * noise[m]) && (floatPixels[m]+background[m] < _MaxPixelValue) ) // whether pixels Thr sigma above background.

                 {
                  if( (floatPixels[m]>floatPixels[m-1]) && (floatPixels[m]>floatPixels[m+1]) && (floatPixels[m]>floatPixels[m-sizeX]) && 
                      (floatPixels[m]>floatPixels[m+sizeX]) &&     // whether this pixel is higher than the neighbors
                      (floatPixels[m]>floatPixels[m-sizeX-1]) && (floatPixels[m]>floatPixels[m-sizeX+1]) && (floatPixels[m]>floatPixels[m+sizeX-1]) && 
                      (floatPixels[m]>floatPixels[m+sizeX+1])    //pixel value is higher than the diagonal neighbors
		      )
                   {
            	    x_c = i;
            	    y_c = j;
                    sxx = sxy = syy = 0;
                    for(int it=0; it<2; it++)
                      {
                       stat.Clear();
                       vec_pix.clear();
            	       npix=0;

                       for(int nn=0; nn<Nnb; nn++)
                         {             
                          for(int kk=0; kk< nbShiftR[nn]; kk++)
                            {
                             long xx, yy;
                             xx = x_c+ nbShiftX[nn][kk]+0.5;
                             yy = y_c+ nbShiftY[nn][kk]+0.5;

                             if( xx>=sizeX || xx<0 || yy>=sizeY || yy<0 ) continue;
                             
                             if(floatPixels[xx+sizeX*yy]>0)
                                  stat.AddPixel(xx, yy, (double)floatPixels[xx+sizeX*yy]);

                             vec_pix.push_back( floatPixels[xx+sizeX*yy] );
                             npix++;
                            }//loop kk

                         if( nn>0 && vec_pix.size()>1)
                          {
                           sort( vec_pix.begin(), vec_pix.end() );  
                           if( vec_pix[vec_pix.size()/2+1] < _medianThreshold*noise[m] ) break;     
                          }
                        }//loop nn
                       if(stat.GetSignal()>0)
                        {
                         x_c = stat.GetXpos();
                         y_c = stat.GetYpos();
                        } 
                     }//loop it

 	           if(stat.GetSignal()<=0)continue;

                   if(stat.GetNpixel() < 3 || stat.GetElongation() >= 0.95 ) continue;          

		   // Check if the seed pixel is still the highest pixel in the cluster (!) 

                   if( vec_pix[npix-1] > floatPixels[m]+0.5) continue;

		  gaus_parameters->AddRow(x_c, y_c, stat.GetXspread(), stat.GetYspread(),  stat.GetXYcorr());

                  gaus_parameters->SetLast("XFWHM",stat.GetFWHM() );
                  gaus_parameters->SetLast("ORIENTATION",stat.GetOrientation() );
                  gaus_parameters->SetLast("ELONGATION",stat.GetElongation() );

                  }//end of condition: whether this pixel is higher than the neighbors

                 }//end of condition: pixels Thr sigma above background

              }// end loop searching the clusters j

           }// end loop searching the clusters i

          _nInputImages++;
          delete [] floatPixels;
          delete BgMap;
          delete BgImage;
          delete NoiseImage;

         } //	for (int ifm = 0; ifm < nImg; ifm++) - loop over images in input collection



      vector<double>* xxx = &(gaus_parameters->GetVectorOf<double>("X_C"));
      vector<double>* yyy = &(gaus_parameters->GetVectorOf<double>("Y_C"));

      vector<double>* sx = &(gaus_parameters->GetVectorOf<double>("SSXX"));
      vector<double>* sy = &(gaus_parameters->GetVectorOf<double>("SSYY"));
      vector<double>* s_xy = &(gaus_parameters->GetVectorOf<double>("SSXY"));

      vector<double> *ffwhm = &(gaus_parameters->GetVectorOf<double>("XFWHM"));
      vector<double> *orient = &(gaus_parameters->GetVectorOf<double>("ORIENTATION"));
      vector<double> *elong = &(gaus_parameters->GetVectorOf<double>("ELONGATION"));

 
      GloriaMedianMap* NewMap2 = new GloriaMedianMap( _frameDivision, _frameDivision ,6);

      NewMap2->SetImageSize(SizeX,SizeY);
      NewMap2->SetMedianName(0,"RMS_X");
      NewMap2->SetMedianName(1,"RMS_Y");
      NewMap2->SetMedianName(2,"CORR_XY");
      NewMap2->SetMedianName(3,"FWHM");
      NewMap2->SetMedianName(4,"ORIENTATION");
      NewMap2->SetMedianName(5,"ELONGATION");

      int nmin = NewMap2->CalculateMap(0, xxx, yyy, sx);

      streamlog_out(DEBUG) << " Calculation of PSF shape map returns " << nmin << std::endl;


      NewMap2->CalculateMap(1, xxx, yyy, sy);
      NewMap2->CalculateMap(2, xxx, yyy, s_xy);
      NewMap2->CalculateMap(3, xxx, yyy, ffwhm);
      NewMap2->CalculateMap(4, xxx, yyy, orient);
      NewMap2->CalculateMap(5, xxx, yyy, elong); 
   
      // Generate and store output parameter table

      GloriaFitsTable *maptab = NewMap2->GetTable();
      (data->GetTableCollection(_outputCollectionName)).push_back(*maptab);

      _nOutputMaps++;

      delete NewMap2;
      delete gaus_parameters;
  // always return true  for ProcessorName
  setReturnValue(true);

}


void CalculateGausPSF::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void CalculateGausPSF::end()
{

	printEndMessage();

}

void CalculateGausPSF::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << _nInputImages << " images processed, " 
			       << _nOutputMaps << " parameter maps stored " << std::endl;

}

} // namespace luiza
