#include "FastAperturePhotometry.h"
#include "gloria/GloriaMedianMap.h"


#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"
#include "gloria/GloriaAstrometry.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

FastAperturePhotometry aFastAperturePhotometry;

FastAperturePhotometry::FastAperturePhotometry() :
		Processor("FastAperturePhotometry")
{

	_description = "Processor to calculate aperture photometry on the frame.";

	registerInputCollection(ImageCollection, "ImageCollectionName",
			"Name of the collection containing input images",
			_imageCollectionName, std::string(""));

	registerInputCollection(TableCollection, "ListCollectionName",
			"Name of the collection containing object lists",
			_listCollectionName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where object list will be stored",
			_outputCollectionName, std::string(""));

  std::vector<float > initPosition;

	registerProcessorParameter(
			"ObjectPositions",
			"Positions of objects of interest (RA,DEC in degrees)",
			_objectPositions, initPosition);



	registerProcessorParameter(
			"ApertureRadius",
			"Radius of circular aperture used for signal counting in pixels",
			_apertureRadius, (float)2.01);

  std::vector<float > initRadii;
  //
  //  If default value given, there is no way to clear it ?! 
  //
  //  initRadii.push_back(5.99);
  //  initRadii.push_back(7.01);

	registerProcessorParameter(
			"AnnulusRadii",
			"Inner and outer radius of the annulus used for background subtraction",
			_annulusRadii,initRadii);

	registerOptionalParameter(
			"BackgroundSampling",
			"Only every BackgroundSampling pixel is considered in background calculation",
			_backgroundSampling, (int)1 );

	registerOptionalParameter(
			"FrameDivision",
			"Frame division for background calculation",
			_frameDivision, (int)1 );

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for background map interpolation",
			_interpolationOrder, (int)1 );

	registerOptionalParameter(
			"BrightnessColumn",
			"Name of the object brightness column",
			_brightnessColumn, std::string("Signal"));

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);

	registerOptionalParameter(
			"UseTelescopeParameters",
			"Use network configuration file to update processor parameters on even by event basis",
			_useTelescopeParameters, false );

}


void FastAperturePhotometry::init()
{

	streamlog_out(MESSAGE) << "FastAperturePhotometry::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputLists = 0;
	_nLoops = 0;


	// Check input parameters

	if (_imageCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input image collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if (_listCollectionName.size() == 0 && _objectPositions.size() == 0)
	{
		streamlog_out(ERROR) << "No input object list nor positions given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if (_listCollectionName.size() != 0 && _objectPositions.size() != 0)
	{
		streamlog_out(ERROR) << "Can not both input methods at the same time! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stoped! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

        // Flag indicating input method used
        _usingObjectList = _listCollectionName.size() != 0 ;

 

	if ( (_annulusRadii.size() !=0 && _annulusRadii.size() !=2) ||
           _annulusRadii.size() ==2 && 
	     (_annulusRadii[0] < _apertureRadius ||  _annulusRadii[0] >= _annulusRadii[1])  )
	{
		streamlog_out(ERROR) << "Annulus radii vector not properly defined! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stoped! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if ( _annulusRadii.size() ==0 )
	{
		streamlog_out(MESSAGE) << "Background annulus radii not defined! " << std::endl;
		streamlog_out(MESSAGE) << "Interpolated image background will be used! " << std::endl;
	}

	if (!_usingObjectList && _outputCollectionName.size() == 0)
	{
	        streamlog_out(ERROR) << "No output collection name given! " << std::endl;
		streamlog_out(ERROR) << "Created object lists will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}

	// Initialize aperture calculation vectors


        int shiftmax = _apertureRadius+0.5;

	if(_annulusRadii.size()==2)shiftmax = _annulusRadii[1]+0.5;


        for(int ix=-shiftmax; ix<=shiftmax; ix++)
        for(int iy=-shiftmax; iy<=shiftmax; iy++)
	  {
	    float x=ix;
            float y=iy;
 
            float r=sqrt(x*x+y*y);

            if(r <= _apertureRadius)
	      {
		_apShiftX.push_back(ix);
		_apShiftY.push_back(iy);
	      }

	    if(_annulusRadii.size()!=2)continue;

            if(r >= _annulusRadii[0] && r <= _annulusRadii[1])
	      {
		_anShiftX.push_back(ix);
		_anShiftY.push_back(iy);
	      }
	  }

	streamlog_out(MESSAGE) << "Aperture with " << _apShiftX.size() << " signal pixels and "
                               << _anShiftX.size() << " background pixels defined " << endl;


	return;

}

void FastAperturePhotometry::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_imageCollectionName.size() == 0)
		return;

	if (_listCollectionName.size() == 0 && _objectPositions.size() == 0)
	        return;


	_nLoops++;


	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collections exists

	if (!data->ImageCollectionExists(_imageCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_imageCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}

	if (_usingObjectList && !data->TableCollectionExists(_listCollectionName))
	{
		streamlog_out(DEBUG) << "Input object list collection not found ! " << endl;

		return;
	}


        int nTab = 1;
	if (_usingObjectList)
	  {
          nTab = data->GetTableCollection(_listCollectionName).size();
	
          if (nTab == 0)
  	     {
	     streamlog_out(DEBUG) << "No object list tables in input collection ! " << endl;
	     return;
	     }
	  }


	streamlog_out(DEBUG) << "  Collection " << _imageCollectionName << " with "
			<< nImg << " images " << endl;


	if (_usingObjectList)
	streamlog_out(DEBUG) << "  Collection " << _listCollectionName << " with "
			<< nTab << " object lists " << endl;


	if(_usingObjectList && nTab !=nImg)
	  {
	  streamlog_out(ERROR) << "  Number of object lists (" << nTab 
			       << ") does not match number of images (" << nImg 
                               << ") - skip event! " << endl;

          throw SkipEventException( this ) ;
	  }

        nTab = nImg;

	// Create output collection if not found (and name is specified)

	if (!_usingObjectList && _outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
	  data->CreateTableCollection(_outputCollectionName,
	  			_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;

	  }


	// Loop over images in input collection
	//=====================================

	for (int ifm = 0; ifm < nImg; ifm++)
	  {
          GloriaFitsImage *cur_img = &(data->GetImageCollection(_imageCollectionName).at(ifm));

	  // Skip empty images

	  int sizeX = cur_img->SizeX();
	  int sizeY = cur_img->SizeY();

	  long sizeXY = sizeX * sizeY;

	  if(sizeXY == 0) continue;

          // Skip if no astrometry in the header

          GloriaAstrometry ga;

	  if(!_usingObjectList && !ga.IsAstrometryPresent(cur_img)) 
	    {
	    streamlog_out(MESSAGE) << "No astrometry found: image is skipped! " <<  endl;
	    continue;
	    }

          _nInputImages++;

	  if(!_usingObjectList)
              ga.InitFromHeader(cur_img);  // Read astrometry from header 


	  // Update parameters for this telescope

	  if(_useTelescopeParameters)
                          updateTelescopeParameters(cur_img);

	  // Create output object list 

          GloriaObjectList *out_list;

	  if(_usingObjectList)
	    {
	    out_list = (GloriaObjectList *) &(data->GetTableCollection(_listCollectionName).at(ifm));
	    }
	  else
	    {
            out_list = new GloriaObjectList;

	    // Copy some header information

            int ncopy=9;
            string copykey[] = {"DATE-OBS","JD","HJD","OBJECT",
                                "TELESCOP","INSTRUME","OBSERVAT","RA","DEC"};

            for(int icopy=0; icopy<ncopy; icopy++)
	      if(cur_img->IsKeywordSet(copykey[icopy].c_str()))
		 out_list->AddKeyword(cur_img->GetKeywordRecord(copykey[icopy].c_str()));

	    }

	  // Add columns with position and signal

	 if(!_usingObjectList && !out_list->ColumnExists("RA"))
             out_list->AddColumn("RA",(double) 0., string("deg"), string("Object RA from astrometry"));
         if(!_usingObjectList && !out_list->ColumnExists("DEC"))
             out_list->AddColumn("DEC",(double) 0., string("deg"), string("Object DEC from astrometry"));

         if(!out_list->ColumnExists(_brightnessColumn))
             out_list->AddColumn(_brightnessColumn,(double) 0.,string(""),
                                               string("Signal from FastAperturePhotometry"));

          out_list->AdjustColumns();

  // Calculate background map for the image, if annulus radii not given

  GloriaMedianMap* BgMap =NULL;

  if(_annulusRadii.size()!=2)
    {
    BgMap = new GloriaMedianMap( _frameDivision, _frameDivision);
    BgMap->CalculateBgMap( cur_img, _backgroundSampling);
    }

  // Images can be float or int

  double* floatPixels = NULL;
  long* intPixels = NULL;

  if(cur_img->IsFloatImage())
      floatPixels =  cur_img->GetFloatImageData();
  else
      intPixels =  cur_img->GetIntImageData();


  // Number of objects to consider

  int nObj = _objectPositions.size()/2;

  if(_usingObjectList) nObj = out_list->GetRowNumber();


  for(int iObj=0; iObj<nObj; iObj++)
    {
      // Get object position 

      double ccdx,ccdy;
      double ra,dec;

    if(_usingObjectList)
      {
	ccdx  = out_list->GetValueOf<double>("CCD_X",iObj);
	ccdy  = out_list->GetValueOf<double>("CCD_Y",iObj);
      }
    else
      {
      ra=_objectPositions[2*iObj];
      dec=_objectPositions[2*iObj+1];

      // Transform to CCD coordinates

      ga.GetXYFromRaDec(ra,dec,&ccdx,&ccdy);

      // Skip this object if outside CCD frame

      if(ccdx<1 || ccdy<1 || ccdx>sizeX || ccdy>sizeY) continue;
      }

   // Sum signal

   int ix0 = ccdx-0.5;
   int iy0 = ccdy-0.5;

   double apSum=0.;
   int apN=0;

   for(int ipix=0;ipix< _apShiftX.size(); ipix++)
	{
	  int ix=ix0+_apShiftX[ipix];
	  int iy=iy0+_apShiftY[ipix];

          if(ix<0 || iy<0 || ix>=sizeX || iy>=sizeY )continue;

          if(cur_img->IsFloatImage())
            apSum+=floatPixels[iy*sizeX+ix];
          else
            apSum+=(double)intPixels[iy*sizeX+ix];

          apN++;         
	}

      if(apN < 1) continue;

  // Calculate background

  double medBG = 0.;
  int anN=0;

  // From annulus (when defined)

  if(_annulusRadii.size()==2)
    {
   std::vector<double> anVec;

   for(int ipix=0;ipix< _anShiftX.size(); ipix++)
        {
        int ix=ix0+_anShiftX[ipix];
        int iy=iy0+_anShiftY[ipix];

         if(ix<0 || iy<0 || ix>=sizeX || iy>=sizeY )continue;

         if(cur_img->IsFloatImage())
            anVec.push_back(floatPixels[iy*sizeX+ix]);
         else
            anVec.push_back((double)intPixels[iy*sizeX+ix]);

         anN++;         
	}

    if(anN < 1) continue;

    std::sort(anVec.begin(),anVec.end());

    int imed=anN/2;

    if(anN%2 == 1)     
      medBG=anVec[imed];
    else
      medBG=0.5*(anVec[imed]+anVec[imed-1]);
    }
  else
    medBG=BgMap->GetBgLevel(ccdx,ccdy,_interpolationOrder);


    double signal = apSum - apN * medBG;

    streamlog_out(DEBUG) << "Object " << iObj << " at " << ccdx << "," 
                         << ccdy << " (on CCD): " << signal << endl;


    streamlog_out(DEBUG) << "  from " << apSum << " signal in " << apN << " pixels with "
                         << medBG << " median background level" <<  endl;


    if(_usingObjectList)
      {
       out_list->SetValue(_brightnessColumn,iObj,signal);
      }
    else
      {
      out_list->AddObject( ccdx, ccdy, signal);	

      out_list->SetLast("RA",ra);
      out_list->SetLast("DEC",dec);

      out_list->SetLast(_brightnessColumn,signal);
      }

    } // end of loop over objects in the list

   // Print created table if not stored (DEBUG only)
  
    if (!_usingObjectList && _outputCollectionName.size() == 0)
	      {
              stringstream tabss;
              out_list->PrintHeader(tabss,20);  // Header
              out_list->Print(tabss,20);  // Table itself

              streamlog_out(DEBUG) << " Object list for image " << ifm 
                                     << " in collection " << _imageCollectionName<< endl;

              streamlog_out(DEBUG) << " Total of " << out_list->GetObjectNumber()
                                     << " objects measured " <<  endl;

              streamlog_out(DEBUG) << tabss.str() << endl;
	      }


       if (!_usingObjectList && _outputCollectionName.size() > 0  
                     && data->TableCollectionExists(_outputCollectionName))
	     {
	      (data->GetTableCollection(_outputCollectionName)).push_back(*out_list);
	     _nOutputLists++;
	     }

       if (!_usingObjectList)
                delete out_list;
            
    } //	for (int ifm = 0; ifm < nImg; ifm++) - loop over images in input collection


  // always return true  for ProcessorName
  setReturnValue(true);

}

void FastAperturePhotometry::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void FastAperturePhotometry::end()
{

	printEndMessage();

}

void FastAperturePhotometry::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << _nInputImages << " images processed, " 
			       << _nOutputLists << " object lists stored " << std::endl;

}

} // namespace luiza
