#include "ImageLaplace.h"
#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  ImageLaplace aImageLaplace ;


  ImageLaplace::ImageLaplace() : Processor("ImageLaplace") {

   _description = "Processor create laplace images from the input collection." ;


   registerInputCollection (  ImageCollection,
                     "InputCollectionName",
                     "Name of the collection containing images to be transformed",
                     _inputCollectionName,  std::string ("") );


   registerOutputCollection (  ImageCollection,
                     "OutputCollectionName",
                     "Name of the collection where laplace image will be stored",
                     _outputCollectionName,  std::string ("") );


   registerOptionalParameter ("PermanentCollection",
   "Flag output collection, as permanent, not deleted after loop is finished",
                              _permanentCollection,  false );

   registerOptionalParameter ("LaplaceType",
   "Flag to set type of laplace transformation",
                              _laplaceType,  static_cast < int > (1) );
 }

  void ImageLaplace::init() { 

    streamlog_out( MESSAGE ) << "ImageLaplace::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nInputImages = 0 ;
    _nOutputImages = 0 ;
    _nLoops = 0 ;

  
   // Check input parameters

    if(_inputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No input collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Nothing to do! " << std::endl;

        throw StopProcessingException( this ) ;
	return;
      }

 
    if(_outputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No output collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Transforming images not possible! " << std::endl;

        throw StopProcessingException( this ) ;
	return;
      }


    if(_laplaceType<1 || _laplaceType>3) 
       {
     streamlog_out(WARNING) << "Wrong LaplaceType parameter value: " << _laplaceType << endl ;

     streamlog_out(WARNING) << "Changing to default value of 1" << endl ;

       }


    return;
 
  }



  void ImageLaplace::processData( GloriaDataContainer* data ) 
{ 

  // Return if input or output collection not specified
  // Should not happen as exception is trown in init()

    if(_inputCollectionName.size()==0 )return;

    if(_outputCollectionName.size()==0 )return;

    _origImage = NULL;
    _LapImage = NULL;

    _nLoops ++ ;

    // Check if input collection exists
 
     if(! data -> ImageCollectionExists(_inputCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      
     int nImg = data -> GetImageCollection(_inputCollectionName).size();

     if(nImg == 0)
	  {
          streamlog_out(DEBUG) << "No images in input collection ! " << endl ;

          return;
	  }
      
     streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with " << nImg << " images " << endl ;


    // Create output collection if not found

    if(! data -> ImageCollectionExists(_outputCollectionName))
         data -> CreateImageCollection(_outputCollectionName,_permanentCollection);

     // Loop over images in input collection

   
    for(int ifm=0; ifm<nImg; ifm++)
        {

	 _nInputImages++;
            
	 _origImage = &(data->GetImageCollection(_inputCollectionName).at(ifm));  


       long sizeX=0;
       long sizeY=0;


        sizeX = _origImage->SizeX();
        sizeY = _origImage->SizeY();


	  long sizeXY = sizeX * sizeY;

       streamlog_out(DEBUG) << "     Image " << _origImage->InputFileName()  << " is " << _origImage->SizeX() 
                                   << "x" << _origImage->SizeY() << std::endl ;
       streamlog_out(DEBUG) << "     Image " << _origImage->InputFileName()  << " is Int: " <<_origImage->IsIntImage() << std::endl ;


      
       // Allocate laplace image

       _LapImage = new GloriaFitsImage(sizeX,sizeY,_origImage->IsFloatImage() );  
	 if(_LapImage == NULL)
	   {
   	   streamlog_out(ERROR) << "Error in image allocation " << endl ;
	   streamlog_out(ERROR) << _LapImage->ErrorMessage() << endl ;
	   return;
	   }


       // Copy header

       _LapImage->CopyHeader(_origImage);

       // Copy input file name

       _LapImage->SetInputFileName(_origImage->InputFileName());


// Prepare Laplace Matrix

//laplace nr 4 l- 4, l+ 5
long nbShift4[] = {  -sizeX, -1,0, 1, sizeX, 
		     -2*sizeX-2, -2*sizeX+2,
      		      2*sizeX-2,    2*sizeX+2  };

//laplace nr 9 l- 8, l+ 5
long nbShift9[] = {  -sizeX, -1,0, 1, sizeX, 
		     -2*sizeX-2, -3*sizeX, -2*sizeX+2,
			-3,3,
      		      2*sizeX-2,  3*sizeX,  2*sizeX+2  };

//laplace nr 11 l- 12, l+ 4
long nbShift11[] = {  0, 1, sizeX, sizeX+1, 
		     -sizeX-1, -sizeX, -sizeX+1,-sizeX+2,
			-1,2,
			-sizeX-1,-sizeX+2,
      		      2*sizeX-1,  2*sizeX,  2*sizeX+1, 2*sizeX+2  };

long *nbShift;
int L_plus;
int L_minus;

if(_laplaceType==1)
   {
     nbShift = nbShift4; 
     L_plus=5; 
     L_minus=4;
   }
else
if(_laplaceType==2)
   {
     nbShift = nbShift9; 
     L_plus=5; 
     L_minus=8;
   }
else
if(_laplaceType==3)
   {
     nbShift = nbShift11; 
     L_plus=4; 
     L_minus=12;
   }

int Nnb=L_plus+L_minus; 


if(_origImage->IsIntImage())
{
  long *pixels = _origImage->GetIntImageData();
  long *pixels_new = _LapImage->GetIntImageData();


for(long npix=0; npix<sizeXY;npix++)
	{

	double nbCharge=0.;
	double nbCharge_bg=0.;

        int L_plus_tmp=0;
        int L_minus_tmp=0;

	for (int inb=0 ; inb<Nnb; inb++)
          {
           long nbPixel = npix + nbShift[inb];   //SM sprawdza sasiadow

           if(nbPixel<0 || nbPixel >=sizeXY) continue;

           int xnb = npix%sizeX +  (nbShift[inb]+10*sizeX+10)%sizeX - 10;

           if( xnb<0 || xnb>=sizeX ) continue;  // changed line

           //  if(pixels[nbPixel] == -1 ) continue;  // Why is it here?

	//mgr lecha str 22 wzor 5.1

	if(inb<L_plus)
		{
		nbCharge = nbCharge+pixels[nbPixel];
		L_plus_tmp++;
		}
	else
		{		
		nbCharge_bg = nbCharge_bg+pixels[nbPixel]; //wciaz ni ejest jasne czy tlo oejmowac czy dodawac
		L_minus_tmp++;		
		}
   	}

	//	pixels_new[npix]=nbCharge-((double)L_plus_tmp/L_minus_tmp)*nbCharge_bg;

	pixels_new[npix]=nbCharge/(double)L_plus_tmp - nbCharge_bg/(double)L_minus_tmp;
        }

}
else
{
//float image
	
	double *pixels = _origImage->GetFloatImageData();
	double *pixels_new = _LapImage->GetFloatImageData();


for(long npix=0; npix<sizeXY;npix++)
	{
	double nbCharge=0.;
	double nbCharge_bg=0.;

        int L_plus_tmp=0;
        int L_minus_tmp=0;

	for (int inb=0 ; inb<Nnb; inb++)
          {
           long nbPixel = npix + nbShift[inb];   // neighbors

           if(nbPixel<0 || nbPixel >=sizeXY) continue;

           int xnb = npix%sizeX +  (nbShift[inb]+10*sizeX+10)%sizeX - 10;

           if( xnb<0 || xnb>=sizeX ) continue;  // changed line

//           if(pixels[nbPixel] == -1 ) continue;

	if(inb<L_plus)
		{
		nbCharge = nbCharge+pixels[nbPixel];
		L_plus_tmp++;
		}
	else
		{		
		nbCharge_bg = nbCharge_bg+pixels[nbPixel]; 
		L_minus_tmp++;		
		}
   	}

	//	pixels_new[npix]=nbCharge-((double)L_plus_tmp/L_minus_tmp)*nbCharge_bg;

	pixels_new[npix]=nbCharge/(double)L_plus_tmp - nbCharge_bg/(double)L_minus_tmp;
        }
}
  


          (data->GetImageCollection(_outputCollectionName)).push_back(*_LapImage);

          delete _LapImage;

          _nOutputImages++;

    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
	}
}
  
  void ImageLaplace::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
  void ImageLaplace::end(){ 
    
    printEndMessage() ;

  }

  void ImageLaplace::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " Read "     << _nInputImages << " images "<< std::endl ; 

    streamlog_out(MESSAGE) << " Stored "    << _nOutputImages << " transformed frames " << std::endl << std::endl;
    
  }
  
}// namespace luiza
