#include "ImageMedian.h"


#include "luiza/Exceptions.h"
#include <iostream>
#include <sstream>
#include <algorithm>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  ImageMedian aImageMedian ;


  ImageMedian::ImageMedian() : Processor("ImageMedian") {

    _description = "Processor to calculate median image from images in the input collection.";

   registerInputCollection (  ImageCollection,
                     "InputCollectionName",
                     "Name of the collection containing input images",
                     _inputCollectionName,  std::string ("") );


   registerOutputCollection (  ImageCollection,
                     "OutputCollectionName",
                     "Name of the collection where median image will be stored",
                     _outputCollectionName,  std::string ("") );


   registerOptionalParameter ("PermanentCollection",
   "Flag output collection, as permanent, not deleted after loop is finished",
                              _permanentCollection,  false );


  std::vector<std::string> initkeys;
 
  registerOptionalParameter(
			"MatchKeywords",
			"Names for header keywords which should match for considered frames",
		        _matchKeywords, initkeys);
  }

  void ImageMedian::init() { 

    streamlog_out( MESSAGE ) << "ImageMedian::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nInputImages = 0 ;
    _nOutputImages = 0 ;
    _nLoops = 0 ;

    _medianImage = NULL;

   // Check input parameters

    if(_inputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No input collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Nothing to do! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    if(_outputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No output collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Calculating median image not possible! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    return;
 
  }



  void ImageMedian::processData( GloriaDataContainer* data ) 
{ 

  // Return if input or output collection not specified
  // Error already printed in init()

    if(_inputCollectionName.size()==0 )return;

    if(_outputCollectionName.size()==0 )return;



    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    // Check if input collection exists
 
     if(! data -> ImageCollectionExists(_inputCollectionName))
	  {
          streamlog_out(MESSAGE) << "Input collection not found ! " << endl ;

          return;
	  }
      
     int nImgAll = data -> GetImageCollection(_inputCollectionName).size();

     if(nImgAll == 0)
	  {
          streamlog_out(MESSAGE) << "No images in input collection ! " << endl ;

          return;
	  }
      
     streamlog_out(MESSAGE) << "  Collection " << _inputCollectionName << " with " << nImgAll << " images " << endl ;


    // Create output collection if not found

    if(! data -> ImageCollectionExists(_outputCollectionName))
         data -> CreateImageCollection(_outputCollectionName,_permanentCollection);



    // Check header keywords to define frame sets (eg. for different cameras)

    int nKey = _matchKeywords.size();
    int nSet = 1;

    std::vector<int> nKeyVal;
    std::vector<std::string> keyStrings;

    int ifirst=0;
    
    for(int ikey = 0; ikey < nKey; ikey++)
      {
	int nVal = 0;

	const char *keyName = _matchKeywords.at(ikey).c_str();

	for(int ifm=0; ifm<nImgAll; ifm++)
	  {
          GloriaFitsImage *inputImage = 
               &((data->GetImageCollection(_inputCollectionName)).at(ifm));

	  string keyVal = string("Not set");
	  if(inputImage->IsKeywordSet(keyName))
	                      keyVal = inputImage->GetKeywordValue(keyName);

	  bool isNew = true;

	  for(int iold = 0; iold < nVal; iold++)
	    if(keyVal == keyStrings[ifirst+iold]) isNew = false;

	  if(isNew)
	    {
	      keyStrings.push_back(keyVal);
	      nVal++;
	    }
				    
	  }

	nKeyVal.push_back(nVal);
	ifirst += nVal;
	nSet*=nVal;
      }

    streamlog_out(DEBUG) << nSet << " sets considered in image collection " << endl ;

    // Loop over image sets

    for(int iset=0; iset<nSet ; iset++)
      {
	std::vector<int> keyID;

	int setTmp = iset;
	int ifirst = 0;

	// Decode set number
	
      streamlog_out(DEBUG) << "Looking at set " << iset << " :" ;

      for(int ikey = 0; ikey < nKey; ikey++)
	  {
	    int nVal = nKeyVal[ikey];
	    int iVal = setTmp%nVal;

	    keyID.push_back(iVal+ifirst);

            streamlog_out(DEBUG) << keyStrings.at(keyID[ikey]) << ":" ;

	    ifirst += nVal;

	    setTmp/=nVal;
	  }
	
      streamlog_out(DEBUG) << endl;
      
      // Find images in the set, determine size and type from the first one


     bool isFirst = true;
     GloriaFitsImage *firstImage;

     int sizeX;
     int sizeY;
     bool isIntImage;
     bool isFloatImage;

     std::vector<double*> floatImageDataVec;
     std::vector<long*>   intImageDataVec;
     
     
     for(int ifm=0; ifm<nImgAll; ifm++)
       {
       GloriaFitsImage *inputImage = &((data->GetImageCollection(_inputCollectionName)).at(ifm));

       bool isMatch = true;
     
       for(int ikey = 0; ikey < nKey; ikey++)
	  {
	  const char *keyName = _matchKeywords.at(ikey).c_str();

	  string keyVal = string("Not set");
	  if(inputImage->IsKeywordSet(keyName))
	                      keyVal = inputImage->GetKeywordValue(keyName);

	  if(keyVal != keyStrings.at(keyID[ikey])) isMatch = false;
	  
//         streamlog_out(DEBUG) << inputImage->InputFileName() << " " << keyVal << " " << isMatch << endl;
	  }

       if(! isMatch) continue;

       if(isFirst)
         {
	  firstImage = inputImage;
	  sizeX = firstImage->SizeX();
          sizeY = firstImage->SizeY();
          isIntImage = firstImage->IsIntImage();
          isFloatImage = firstImage->IsFloatImage();

	  isFirst = false;
         }
       else     
         if(inputImage->SizeX()!=sizeX || inputImage->SizeY()!=sizeY || 
           inputImage->IsIntImage()!=isIntImage )
	   {
	    streamlog_out(ERROR) << "Image " << ifm << " in " 
				 << _inputCollectionName 
				 << " skipped due to wrong size/type!" << endl ;
	    continue;
	   } 


       if(isIntImage)
	  intImageDataVec.push_back(inputImage->GetIntImageData());
        else
	  floatImageDataVec.push_back(inputImage->GetFloatImageData());

          _nInputImages++;

	}

     // Verify number of images after size/type check

     int nImg;
     
     if(isIntImage)
        nImg = intImageDataVec.size();
     else
	nImg = floatImageDataVec.size();

     streamlog_out(DEBUG) << nImg 
          << " images in input collection will be used to calculate median frame" << endl; 

    // Special case: only one image left in input collection - copy to output

    if(nImg == 1)
      {
        
       streamlog_out(DEBUG) 
          << "  ==>> Directly coping input image to output collection" << endl; 

	// push_back() copies the image !

       (data->GetImageCollection(_outputCollectionName)).push_back(
              (data->GetImageCollection(_inputCollectionName)).at(0)   
                                                                  );
       _nInputImages++;
       _nOutputImages++;
       
	continue;
      }

    // Create new image

    _medianImage = new GloriaFitsImage(sizeX,sizeY,isFloatImage);  

    if(_medianImage == NULL)
	   {
   	   streamlog_out(ERROR) << "Error in image allocation " << endl ;
	   streamlog_out(ERROR) << _medianImage->ErrorMessage() << endl ;
	   return;
	   }

    // Copy header from first image

    _medianImage->CopyHeader(firstImage);

    // Copy input file name

    if(firstImage->InputFileName().size() > 0 )
        _medianImage->SetInputFileName(firstImage->InputFileName());

    // Add note to image history

    stringstream hiss;
    hiss << "LUIZA " << _processorName << " (" << _typeName << ")";
	_medianImage->AddHistory(hiss.str());

    stringstream hiss1;
    hiss1 << "  Median image calculated from  " << nImg << " images";
	_medianImage->AddHistory(hiss1.str());


    // Vector index(es) to get median value

    int imed1,imed2;

    imed2=nImg/2;

    if(nImg%2 == 1)     
       imed1 = imed2;
    else
       imed1 = imed2-1;


    // Main calculation is here
    // Loop over all pixels and calculate median for each of them
    // Two cases, for two content types

    if(isIntImage)
      {
        long*   intMedianData;
	intMedianData = _medianImage->GetIntImageData();

	long* pixVals = new long[nImg];

        for(long ipix=0; ipix < sizeX*sizeY; ipix++)
           {
           for(int ifm=0; ifm<nImg; ifm++)
	      pixVals[ifm]=intImageDataVec[ifm][ipix];

	   std::sort(pixVals,pixVals+nImg);

           intMedianData[ipix] = (pixVals[imed1] + pixVals[imed2])/2;
	   }

	delete[] pixVals;
      }

    else
      {
        double* floatMedianData;
	floatMedianData = _medianImage->GetFloatImageData();

        double* pixVals = new double[nImg];

        for(long ipix=0; ipix < sizeX*sizeY; ipix++)
           {
           for(int ifm=0; ifm<nImg; ifm++)
	      pixVals[ifm]=floatImageDataVec[ifm][ipix];

	   std::sort(pixVals,pixVals+nImg);

           floatMedianData[ipix] = (pixVals[imed1] + pixVals[imed2])/2.;
	   }

	delete[] pixVals;
      }
    

    // Add median image to output collestion

    (data->GetImageCollection(_outputCollectionName)).push_back(*_medianImage);
    delete _medianImage;

     _nOutputImages++;

    // End of loop over image sets
    }

    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
  }
  
  void ImageMedian::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
  void ImageMedian::end(){ 
    
    printEndMessage() ;

  }

  void ImageMedian::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " Added "     << _nInputImages << " images "<< std::endl ; 

    streamlog_out(MESSAGE) << " Stored "    << _nOutputImages << " median frames " << std::endl << std::endl;
    
  }
  
}// namespace luiza
