#include "ImageNormalizationProcessor.h"


#include "luiza/Exceptions.h"
#include <iostream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  ImageNormalizationProcessor aImageNormalizationProcessor ;


  ImageNormalizationProcessor::ImageNormalizationProcessor() : Processor("ImageNormalizationProcessor") {

    _description = "Simple processor to apply dark subtraction and flat correction" ;

   registerInputCollection (  ImageCollection,
                     "InputCollectionName",
                     "Name of the collection containing images to be corrected",
                     _inputCollectionName,  std::string ("") );


   registerInputCollection (  ImageCollection,
                     "DarkCollectionName",
                     "Name of the collection containing dark images",
                     _darkCollectionName,  std::string ("") );


   registerInputCollection (  ImageCollection,
                     "BiasCollectionName",
                     "Name of the collection containing bias frames",
                     _biasCollectionName,  std::string ("") );


   registerInputCollection (  ImageCollection,
                     "FlatCollectionName",
                     "Name of the collection containing flat frames",
                     _flatCollectionName,  std::string ("") );


   registerOutputCollection (  ImageCollection,
                     "OutputCollectionName",
                     "Name of the collection where image sums will be stored",
                     _outputCollectionName,  std::string ("") );


   registerProcessorParameter ("AutoNormalization",
   "Flag for using exposure time information (from fits header) for calculating correction factors",
                              _AutoNormalization,  false );


   registerProcessorParameter ("FlatNormalization",
   "Flag for applying dark/bias subtraction to the flat frame image",
                              _FlatNormalization,  false );


   registerOptionalParameter ("PermanentCollection",
   "Flag output collection, as permanent, not deleted after loop is finished",
                              _permanentCollection,  false );


   registerOptionalParameter ("ImageScale",
                              "Factor used to scale image pixels",
                              _imageScale, (double) 1. );

   registerOptionalParameter ("DarkScale",
                              "Factor used to scale dark image",
                              _darkScale, (double) 1. );

   registerOptionalParameter ("BiasScale",
                              "Factor used to scale bias image",
                              _biasScale, (double) 0. );

   registerOptionalParameter ("FlatScale",
                              "Factor used to scale flat frame",
                              _flatScale, (double) 1. );

  std::vector<std::string> initkeys;
 
  registerOptionalParameter(
			"MatchKeywords",
			"Names for header keywords used to select proper dark/flat/bias frames",
		        _matchKeywords, initkeys);

  registerOptionalParameter(
			"MatchExpTime",
			"Flag for using dark frame selection based on exposure time",
		        _matchExpTime, true);
 }

  void ImageNormalizationProcessor::init() { 

    streamlog_out( MESSAGE ) << "ImageNormalizationProcessor::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nInputImages = 0 ;
    _nOutputImages = 0 ;
    _nCorrected = 0;
    _nLoops = 0 ;

   // Check input parameters

    if(_inputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No input collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Nothing to do! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    if(_outputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No output collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Input images will be modified! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    _nKey = _matchKeywords.size();
    
    _initKeyVal = string("Not set");
    
    for(int ikey = 0; ikey < _nKey; ikey++)
        _matchValues.push_back(_initKeyVal);
 
    return;
 
  }



  void ImageNormalizationProcessor::processData( GloriaDataContainer* data ) 
{ 

  // Return if input or output collection not specified
  // Message already printed in init()

    if(_inputCollectionName.size()==0 )return;


    _nLoops ++ ;

    streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops << std::endl ;

    // Check if input collection exists
 
     if(! data -> ImageCollectionExists(_inputCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      
     int nImg = data -> GetImageCollection(_inputCollectionName).size();

     if(nImg == 0)
	  {
          streamlog_out(DEBUG) << "No images in input collection ! " << endl ;

          return;
	  }
      
     streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with " << nImg << " images " << endl ;

     // Check dark collection

     int nDark = 0;

     if(_AutoNormalization || _darkScale != 0.)
       {

       if(_darkCollectionName.size()!=0 &&  data -> ImageCollectionExists(_darkCollectionName))
               nDark = data -> GetImageCollection(_darkCollectionName).size();

       if( nDark > 0 && nDark != nImg && _nKey == 0 && ! _matchExpTime ) nDark=1;

       if(nDark > 0 )
            streamlog_out(DEBUG) << "         using " << nDark << " dark frames from  " 
                                 << _darkCollectionName << endl ;
       }

     // Check bias collection

     int nBias = 0;

     if(_AutoNormalization || _biasScale != 0.)
       {

       if(_biasCollectionName.size()!=0 &&  data -> ImageCollectionExists(_biasCollectionName))
               nBias = data -> GetImageCollection(_biasCollectionName).size();

       if( nBias > 0 && nBias != nImg  && _nKey == 0 ) nBias=1;


       if(nBias > 0 )
            streamlog_out(DEBUG) << "         using " << nBias << " bias frames from  " 
                                 << _biasCollectionName << endl ;
       }


     // Check flat collection

     int nFlat = 0;

     if( _flatScale != 0.)
       {
       if(_flatCollectionName.size()!=0 &&  data -> ImageCollectionExists(_flatCollectionName))
               nFlat = data -> GetImageCollection(_flatCollectionName).size();

       if( nFlat > 0 && nFlat != nImg  && _nKey == 0 ) nFlat=1;


       if(nFlat > 0 )
            streamlog_out(DEBUG) << "         using " << nFlat << " flat frames from  " 
                                 << _flatCollectionName << endl ;
       }

    // Create output collection if not found

    if(_outputCollectionName.size()!=0 && data->ImageCollectionExists(_outputCollectionName)==false )
                 data -> CreateImageCollection(_outputCollectionName,_permanentCollection);


    // Pointers to images considered in processor

    GloriaFitsImage *newImage=NULL;
    GloriaFitsImage *darkImage=NULL;
    GloriaFitsImage *biasImage=NULL;
    GloriaFitsImage *flatImage=NULL;

    int iDarkOld = -1;
    int iBiasOld = -1;
    int iFlatOld = -1;
    
    // Scale factors of images considered;

    double imageNorm, darkNorm, biasNorm, flatNorm;

    // Exposure times of images

    double ExpTime, DarkTime, BiasTime, FlatTime;
    
    // Flag set after flat normalization

    bool flatNormalized;

    // Loop over images in input collection

    for(int ifm=0; ifm<nImg; ifm++)
        {

        // Load input image: make a local copy
	//------------------------------------

        newImage = new GloriaFitsImage((data->GetImageCollection(_inputCollectionName)).at(ifm));  

        _nInputImages++;

        newImage->ClearError();

        imageNorm = 1.;

	// Check exposure time

        if(newImage->IsKeywordSet("EXPTIME"))
	    {
            ExpTime = newImage->GetKeywordValueDouble("EXPTIME");   
            streamlog_out(DEBUG) << " Normalizing frame with exposure time = " << ExpTime << " s" << endl ;
	    }
        else
	   {
	    ExpTime = -1.;

	    if(_AutoNormalization)
                 streamlog_out(ERROR) << " Exposure time not set in fits header - can not normalize image !" << endl ;
            else 
                 streamlog_out(DEBUG) << " Exposure time not set for image being normalized (?)" << endl ;
	   }
       

	// Store keyword values for dark/bias/flat matching
	
        for(int ikey = 0; ikey < _nKey; ikey++)
	   {
	   const char *keyName = _matchKeywords.at(ikey).c_str();

	   string keyVal = _initKeyVal;
	   if(newImage->IsKeywordSet(keyName))
	                      keyVal = newImage->GetKeywordValue(keyName);

	   _matchValues[ikey] = keyVal;
	   }
	     

	
        // Load dark image (find the best one and make a local copy)
	//----------------------------------------------------------

        if(nDark > 0)
	  {
	  int iMatchKey = -1;
	  int iMatch = -1;

	  for(int idfm = 0; idfm < nDark ; idfm ++)
	    {
             GloriaFitsImage *inputImage = &((data->GetImageCollection(_darkCollectionName)).at(idfm));

	    bool isMatch = true;
            for(int ikey = 0; ikey < _nKey; ikey++)
  	      {
	      const char *keyName = _matchKeywords.at(ikey).c_str();

	      string keyVal = _initKeyVal;
	      if(inputImage->IsKeywordSet(keyName))
	                      keyVal = inputImage->GetKeywordValue(keyName);

	      if(keyVal != _matchValues.at(ikey)) isMatch = false;

	      //   streamlog_out(DEBUG) << ikey << " " << keyVal << " " << _matchValues.at(ikey) << endl ;
	      
	      }

	    double eTime = 0.;
	    
            if(inputImage->IsKeywordSet("EXPTIME"))
	       {
               eTime = inputImage->GetKeywordValueDouble("EXPTIME");   

               if(eTime < 0.) eTime = -eTime;  // Negative values indicate dark frame...
	       }


	    if(isMatch && iMatchKey < 0) iMatchKey = idfm;

	    if(isMatch && eTime == ExpTime && iMatch < 0) iMatch = idfm;
	    }


	  int iDark = 0;
	  if(nDark == nImg) iDark = ifm;
	  
	  if(iMatch >=0)
	     {
	      iDark = iMatch;
	      
              streamlog_out(DEBUG) << " Dark frame " << iDark << " matched with keywords and EXPTIME" << endl ;
	     }
	  else
	    if(iMatchKey >=0 && _nKey>0)
	      {
	      iDark = iMatchKey;
	      
              streamlog_out(DEBUG) << " Dark frame " << iDark << " matched with keywords" << endl ;
	      }

	  
	  if(iDark != iDarkOld)
	    {
            if(darkImage)delete darkImage;

	    darkImage = new GloriaFitsImage((data->GetImageCollection(_darkCollectionName)).at(iDark));  

            darkImage->ClearError();

            darkNorm=1.;

 	    // Check exposure time

            if(darkImage->IsKeywordSet("EXPTIME"))
	     {
              DarkTime = darkImage->GetKeywordValueDouble("EXPTIME");   

              if(DarkTime < 0.) DarkTime = -DarkTime;  // Negative values indicate dark frame...

              streamlog_out(DEBUG) << " Using dark image with exposure time = " << DarkTime << " s" << endl ;
	     }
           else
	     {
	      DarkTime = -1.;

	      if(_AutoNormalization)
                 streamlog_out(ERROR) << " Exposure time not set in dark fits header - can not normalize image !" << endl ;
              else 
                 streamlog_out(DEBUG) << " Exposure time not set for dark image used (?)" << endl ;
	     }

	    iDarkOld = iDark; 
	   }
	  }

	//


        // Select and load bias image (also make a local copy)
	//-----------------------------------------

        if(nBias > 0)
	  {

	  int iMatch = -1;

	  for(int ibfm = 0; ibfm < nBias ; ibfm ++)
	    {
             GloriaFitsImage *inputImage = &((data->GetImageCollection(_biasCollectionName)).at(ibfm));

	    bool isMatch = true;
            for(int ikey = 0; ikey < _nKey; ikey++)
  	      {
	      const char *keyName = _matchKeywords.at(ikey).c_str();

	      string keyVal = _initKeyVal;
	      if(inputImage->IsKeywordSet(keyName))
	                      keyVal = inputImage->GetKeywordValue(keyName);

	      if(keyVal != _matchValues.at(ikey)) isMatch = false;
	      }


	    if(isMatch && iMatch < 0) iMatch = ibfm;
	    }

	  int iBias = 0;
	  if(nBias == nImg) iBias = ifm;
	  if(iMatch >=0 && _nKey>0)
	      {
	      iBias = iMatch;
	      
              streamlog_out(DEBUG) << " Bias frame " << iBias << " matched with keywords" << endl ;
	      }
	  
	  if(iBias != iBiasOld)
	    {
            if(biasImage)delete biasImage;

	    
            biasImage = new GloriaFitsImage((data->GetImageCollection(_biasCollectionName)).at(iBias));  

            biasImage->ClearError();

            biasNorm=1.;

 	    // Check exposure time

            if(biasImage->IsKeywordSet("EXPTIME"))
	      {
              BiasTime = biasImage->GetKeywordValueDouble("EXPTIME");   

              if(BiasTime < 0.) BiasTime = -BiasTime;  // Negative values indicate dark frame...

              streamlog_out(DEBUG) << " Using bias image with exposure time = " << BiasTime << " s" << endl ;
	      }
            else
	      {
	      BiasTime = 0.;  // assumed for bias (!)

	      if(_AutoNormalization)
                 streamlog_out(ERROR) << " Exposure time not set in bias fits header - can not normalize image !" << endl ;
              else 
                 streamlog_out(DEBUG) << " Exposure time not set for bias image used (?)" << endl ;
	      }
	    
	    iBiasOld = iBias; 
	    }
	  }

        // Select and load flat image (also make a local copy)
	//-----------------------------------------

        if(nFlat > 0)
	  {

	  int iMatch = -1;

	  for(int iffm = 0; iffm < nFlat ; iffm ++)
	    {
             GloriaFitsImage *inputImage = &((data->GetImageCollection(_flatCollectionName)).at(iffm));

	    bool isMatch = true;
            for(int ikey = 0; ikey < _nKey; ikey++)
  	      {
	      const char *keyName = _matchKeywords.at(ikey).c_str();

	      string keyVal = _initKeyVal;
	      if(inputImage->IsKeywordSet(keyName))
	                      keyVal = inputImage->GetKeywordValue(keyName);

	      if(keyVal != _matchValues.at(ikey)) isMatch = false;

	      // streamlog_out(DEBUG) << ikey << " " << keyVal << " " << _matchValues.at(ikey) << endl ;
	      
	      }


	    if(isMatch && iMatch < 0) iMatch = iffm;
	    }

	  int iFlat = 0;
	  
	  if(nFlat == nImg) iFlat = ifm;
	  if(iMatch >=0 && _nKey>0)
	      {
	      iFlat = iMatch;
	      
              streamlog_out(DEBUG) << " Flat frame " << iFlat << " matched with keywords" << endl ;
	      }

	  
	  if(iFlat != iFlatOld)
	    {
            if(flatImage)delete flatImage;
	    
            flatImage = new GloriaFitsImage((data->GetImageCollection(_flatCollectionName)).at(iFlat));  

            flatImage->ClearError();

            flatNorm=1.;

            flatNormalized = false;

     	    // Check exposure time

            if(flatImage->IsKeywordSet("EXPTIME"))
	      {
              FlatTime = flatImage->GetKeywordValueDouble("EXPTIME");   
              streamlog_out(DEBUG) << " Using flat image with exposure time = " << FlatTime << " s" << endl ;
	      }
            else
	      {
	      FlatTime = -1.;

	      if(_FlatNormalization)
                 streamlog_out(ERROR) << " Exposure time not set in flat fits header - can not normalize flat !" << endl ;
              else 
                 streamlog_out(DEBUG) << " Exposure time not set for flat image used (?)" << endl ;
	      }
	    
	    iFlatOld = iFlat; 
	    }


	// Flat normalization
	//--------------------

        if(flatImage!=NULL && !flatNormalized)
	  {

          if(_FlatNormalization)
	    {
	    double fDark,fBias;

	    if(_AutoNormalization && FlatTime>0. && DarkTime>0.)
	      {
	        fDark = (FlatTime-BiasTime)/(DarkTime-BiasTime);
  	        fBias = 1. - fDark;
	      }
	    else 
	      {
		fDark=_darkScale;
		fBias=_biasScale;
	      }

            if(darkImage==NULL)fDark=0.;
            if(biasImage==NULL)fBias=0.;

            streamlog_out(DEBUG) << " Dark and bias scaling factors used for flat normalization: " 
                                 <<  fDark << " , " << fBias << std::endl ;

                        
	    if(fDark!=0. && fDark != darkNorm)
	      {
		*darkImage *= fDark/darkNorm;
		darkNorm=fDark;
	      }

            if(fDark!=0.) *flatImage -= *darkImage;

	    if(fBias!=0. && fBias != biasNorm)
	      {
		*biasImage *= fBias/biasNorm;
		biasNorm=fBias;
	      }

            if(fBias!=0.) *flatImage -= *biasImage;

	    // Check average flat value

	    double avFlat=0.;
            long nFlat = flatImage->SizeX() * flatImage->SizeY();

	      if(flatImage->IsIntImage())
		{
		long *pixels = flatImage->GetIntImageData();
                for(long ipix=0; ipix < nFlat; ipix++)avFlat+=pixels[ipix];
		}
	      else
		{
		double *pixels = flatImage->GetFloatImageData();
                for(long ipix=0; ipix < nFlat; ipix++)avFlat+=pixels[ipix];
		}

	    avFlat/=(double)nFlat;

            streamlog_out(DEBUG) << " Average value on flat frame after dark and bias subtraction:" 
                                 <<  avFlat << std::endl ;

                        
            if(avFlat > 2.) *flatImage *= 1./avFlat;
	    }


          // final image is multiplied by _imageScale and divided by _flatScale

	  if(_flatScale != _imageScale)   
	    {
	     *flatImage *= _flatScale/_imageScale;
	     flatNorm = _flatScale/_imageScale;
	    }

  	   // Check for possible errors

	  if(flatImage->Error())
	    streamlog_out(ERROR) << " Errorr in flat normalization: " <<  flatImage->ErrorMessage() << std::endl ;


	  flatNormalized=true;
	  }
	}

	// Normalize image itself
	//-----------------------


	// Calculate correction factors for this image

	double fDark, fBias;

        if(_AutoNormalization && ExpTime>0. && DarkTime>0.)
	   {
	   fDark = (ExpTime-BiasTime)/(DarkTime-BiasTime);
  	   fBias = 1. - fDark;
	   }
	else 
	   {
	   fDark=_darkScale;
	   fBias=_biasScale;
	   }

        if(darkImage==NULL)fDark=0.;
        if(biasImage==NULL)fBias=0.;

        streamlog_out(DEBUG) << " Dark and bias scaling factors used for image normalization: " 
                             <<  fDark << " , " << fBias << std::endl ;

        // Add note to image history
        stringstream hiss;
        hiss << "LUIZA " << _processorName << " (" << _typeName << ")";
	newImage->AddHistory(hiss.str());

	if(fDark !=0. || fBias!=0.)
	  { 
          stringstream hiss2;
  	  hiss2 << "  Dark and bias subtracted with scales: " 
	        <<  fDark << " " << fBias;
	  newImage->AddHistory(hiss2.str());
	  }
            

	if(fDark!=0. && fDark != darkNorm)
	    {
	    *darkImage *= fDark/darkNorm;
	    darkNorm=fDark;
	    }

        if(fDark!=0.) *newImage -= *darkImage;

	// Check for possible errors

	if(newImage->Error())
	  {
	    streamlog_out(ERROR) << " Errorr in dark subtraction: " <<  newImage->ErrorMessage() << std::endl ;

	    delete newImage;
	    continue;
	  }


        if(fBias!=0. && fBias != biasNorm)
           {
	    *biasImage *= fBias/biasNorm;
	    biasNorm=fBias;
	   }

        if(fBias!=0.) *newImage -= *biasImage;

	// Check for possible errors

	if(newImage->Error())
	  {
	    streamlog_out(ERROR) << " Errorr in bias subtraction: " <<  newImage->ErrorMessage() << std::endl ;

	    delete newImage;
	    continue;
	  }


	// Divide by flat

        if(flatImage!=NULL)  *newImage /= *flatImage;


	// Check for possible errors

	if(newImage->Error())
	  {
	    streamlog_out(ERROR) << " Errorr in flat division: " <<  newImage->ErrorMessage() << std::endl ;

	    delete newImage;
	    continue;
	  }


	// Image scale factor have to be applied if flat was not used 
	// (otherwise it was already included in flat scaling)

  	 if(flatImage==NULL && _imageScale != 1.0)
	   *newImage *= _imageScale;


	 // Add history record

	if(flatImage!=NULL)
	  {
          stringstream hiss3;
          hiss3 << "  Flat image correction applied"; 
	  newImage->AddHistory(hiss3.str());
	  }
            



	// Add corrected image to output collestion, if defined

        if(_outputCollectionName.size()!=0 && data->ImageCollectionExists(_outputCollectionName))
	  {
          (data->GetImageCollection(_outputCollectionName)).push_back(*newImage);
          _nOutputImages++;
	  }

	// Otherwise, replace input image (!)
	else
           (data->GetImageCollection(_inputCollectionName)).at(ifm) = *newImage;  


         delete newImage;
        _nCorrected++;
	}  

    if(darkImage)delete darkImage;
    if(biasImage)delete biasImage;
    if(flatImage)delete flatImage;

    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
  }
  
  void ImageNormalizationProcessor::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
  void ImageNormalizationProcessor::end(){ 
    
    printEndMessage() ;

  }

  void ImageNormalizationProcessor::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " Read "     << _nInputImages << " images "<< std::endl ; 

    streamlog_out(MESSAGE) << " Corrected "     << _nCorrected << " images "<< std::endl ; 

    streamlog_out(MESSAGE) << " Stored "    << _nOutputImages << " new frames " << std::endl << std::endl;
    
  }
  
}// namespace luiza
