#include "PixelClusterFinder.h"
#include "gloria/GloriaMedianMap.h"
#include "gloria/GloriaClusterStatistics.h"


#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

PixelClusterFinder aPixelClusterFinder;

PixelClusterFinder::PixelClusterFinder() :
		Processor("PixelClusterFinder")
{

	_description = "Processor to find pixel clusters on the frame.";

	registerInputCollection(ImageCollection, "InputCollectionName",
			"Name of the collection containing input images",
			_inputCollectionName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where cluster list will be stored",
			_outputCollectionName, std::string(""));

	registerProcessorParameter(
			"SeedThreshold",
			"Threshold for the cluster seed signal (in units of noise level)",
			_seedThreshold, (double)6.);

	registerProcessorParameter(
			"PixelThreshold",
			"Threshold for the cluster pixel signal (in units of noise level)",
			_pixThreshold, (double)3.);


	registerProcessorParameter(
			"MaxPixels",
			"Maximum number of pixels considered in cluster CoG and signal calculation",
			_maxPixels, (int)25);


	registerProcessorParameter(
			"MaxClusters",
			"Maximum number of clusters to be reconstructed",
			_maxClusters, (int)100);


	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);

	registerOptionalParameter(
			"CopyImageHeader",
			"Flag to copy the whole image header as a header of the object list.",
  		        _copyImageHeader, false);
	

        std::vector<std::string> initkey;
        initkey.push_back(string("DATE-OBS"));
        initkey.push_back(string("JD"));
        initkey.push_back(string("HJD"));
        initkey.push_back(string("OBJECT"));
        initkey.push_back(string("TELESCOP"));
        initkey.push_back(string("INSTRUME"));
        initkey.push_back(string("OBSERVAT"));
        initkey.push_back(string("CAMID"));
        initkey.push_back(string("RA"));
        initkey.push_back(string("DEC"));

	registerOptionalParameter(
			"HeaderRecords",
			"Keywords of image fits header, which should be copied to the fits header of the object list.",
		_copykey, initkey);

	registerOptionalParameter(
			"MoreNeighbours",
			"When set to true: search pixel neighbours in 5x5 square rather than 3x3",
		_moreNeighbours, false);

	registerOptionalParameter(
			"BackgroundSampling",
			"Only every BackgroundSampling pixel is considered in background calculation",
			_backgroundSampling, (int)1 );

	registerOptionalParameter(
			"FrameDivision",
			"Frame division for background calculation",
			_frameDivision, (int)1 );

	registerOptionalParameter(
			"BackgroundInterpolation",
			"Flag for interpolation between background levels calculated in different subframes",
			_bgPol, (int)1 );

	registerOptionalParameter(
			"NoiseInterpolation",
			"Flag for interpolation between noise levels calculated in different subframes",
			_noisePol, (int)1 );

	registerOptionalParameter(
			"AddNoise",
			"Flag to add noise information to the output table.",
  		        _addNoise, false);

	registerOptionalParameter(
			"NoiseColumn",
			"Name for column containing noise information.",
		        _noiseColumn, std::string("Noise"));

	registerOptionalParameter(
			"AddSpread",
			"Flag to add charge spread information to the output table.",
  		        _addSpread, false);

std::vector<std::string> initspread;
initspread.push_back(string("RMS_X"));
initspread.push_back(string("RMS_Y"));
initspread.push_back(string("CORR_XY"));

	registerOptionalParameter(
			"SpreadColumns",
			"Names for columns containing charge spread information.",
		        _spreadColumns, initspread);


	registerOptionalParameter(
			"AddShape",
			"Flag to add charge distribution shape information to the output table.",
  		        _addShape, false);

std::vector<std::string> initshape;
initshape.push_back(string("FWHM"));
initshape.push_back(string("ELONG"));
initshape.push_back(string("LENGTH"));
initshape.push_back(string("WIDTH"));
initshape.push_back(string("ANGLE"));

	registerOptionalParameter(
			"ShapeColumns",
			"Names for columns containing charge distribution shape information.",
		        _shapeColumns, initshape);


	registerOptionalParameter(
			"UseTelescopeParameters",
			"Use network configuration file to update processor parameters on even by event basis",
			_useTelescopeParameters, false );

}


void PixelClusterFinder::init()
{

	streamlog_out(MESSAGE) << "PixelClusterFinder::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputLists = 0;
	_nLoops = 0;


	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if (_outputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No output collection name given! "
				<< std::endl;
		streamlog_out(ERROR) << "Object lists will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}

	if(_addNoise && _noiseColumn.size() == 0)
	      {
		streamlog_out(ERROR) << "Noise column name not defined! " << std::endl;
		streamlog_out(ERROR) << "Noise information will not be stored! " << std::endl;

		_addNoise = false;
	      }


	if(_addSpread && _spreadColumns.size() < 3)
	      {
		streamlog_out(ERROR) << "Spread column name list too short! " << std::endl;
		streamlog_out(ERROR) << "Charge spread information will not be stored! " << std::endl;

		_addSpread = false;
	      }


	if(_addShape && _shapeColumns.size() < 5)
	      {
		streamlog_out(ERROR) << "Shape column name list too short! " << std::endl;
		streamlog_out(ERROR) << "Cluster shape information will not be stored! " << std::endl;

		_addShape = false;
	      }


	return;

}

void PixelClusterFinder::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->ImageCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_inputCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nImg << " images " << endl;

	// Create output collection if not found (and name is specified)

	if (_outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
	  data->CreateTableCollection(_outputCollectionName,
	  			_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;

	  }

	// Loop over images in input collection
	//=====================================

	for (int ifm = 0; ifm < nImg; ifm++)
	  {
          GloriaFitsImage *cur_img = &(data->GetImageCollection(_inputCollectionName).at(ifm));

	  // Skip empty images

	  sizeX = cur_img->SizeX();
	  sizeY = cur_img->SizeY();

	  sizeXY = sizeX * sizeY;

	  if(sizeXY == 0) continue;

	  // Update parameters for this telescope

	  if(_useTelescopeParameters)
                          updateTelescopeParameters(cur_img);

	  // Create new object list 

          GloriaObjectList *new_list = new GloriaObjectList;
 
	  // Add column with number of pixels in the cluster

          new_list->AddColumn("Pixels",(long) 0, string(""), string("Number of pixels in the cluster"));


	  // Add optional columns

	  if(_addNoise)
	      new_list->AddColumn(_noiseColumn,(double) 0., string(""), 
                                                       string("Noise level for the cluster"));

	  if(_addSpread)
	    {
	      new_list->AddColumn(_spreadColumns.at(0),(double) 0., string(""), 
                                                       string("Cluster charge spread in X"));

	      new_list->AddColumn(_spreadColumns.at(1),(double) 0., string(""), 
                                                       string("Cluster charge spread in Y"));

	      new_list->AddColumn(_spreadColumns.at(2),(double) 0., string(""), 
                                                       string("Cluster charge X-Y correlation"));
	    }

	  if(_addShape)
	    {
	      new_list->AddColumn(_shapeColumns.at(0),(double) 0., string(""), 
                                                       string("Cluster size"));

	      new_list->AddColumn(_shapeColumns.at(1),(double) 0., string(""), 
                                                       string("Cluster elongation"));

	      new_list->AddColumn(_shapeColumns.at(2),(double) 0., string(""), 
                                                       string("Cluster lenght"));

	      new_list->AddColumn(_shapeColumns.at(3),(double) 0., string(""), 
                                                       string("Cluster width"));

	      new_list->AddColumn(_shapeColumns.at(4),(double) 0., string("deg"), 
                                                       string("Cluster orientation"));
	    }

	  // Copy selected header information

	  if(_copyImageHeader)
	      new_list-> CopyHeader(cur_img);
	  else
	    {
            int ncopy=_copykey.size();

             for(int icopy=0; icopy<ncopy; icopy++)
	       if(cur_img->IsKeywordSet(_copykey[icopy].c_str()))
		  new_list->AddKeyword(cur_img->GetKeywordRecord(_copykey[icopy].c_str()));
	    }
	  
	  // Copy input file name

	  if(cur_img->InputFileName().size() > 0 )
      	       new_list->SetInputFileName(cur_img->InputFileName());


  //=============================================================================
  // Here goes the algorithm adopted from TrackDetailedSimulation package for ILC
  //=============================================================================

  // Calculate background and noise
  // Use MediamMap class

  // This constructor was removed. Use two step initialization instead(!)
  // GloriaMedianMap* BgMap = new GloriaMedianMap( *cur_img, _frameDivision, _frameDivision, _backgroundSampling);

  GloriaMedianMap* BgMap = new GloriaMedianMap( _frameDivision, _frameDivision);
  BgMap->CalculateBgMap( cur_img, _backgroundSampling);


  GloriaFitsImage *BgImage = BgMap->GetBgImage(_bgPol);
  GloriaFitsImage *NoiseImage = BgMap->GetNoiseImage(_noisePol);

  double* bgtab = BgMap->GetBgMap();
  double* noisetab = BgMap->GetNoiseMap();

  streamlog_out(DEBUG) << "BG calculated for collection " 
 		       << _inputCollectionName  << " image " << ifm << endl;

  for(int iSub=0; iSub<  _frameDivision * _frameDivision; iSub++)
    {
      streamlog_out(DEBUG) << "  For subimage " << iSub << " : "  
                           << bgtab[iSub] << " +/-" << noisetab[iSub] << endl;
    }


  BG_mean_av = BgMap->GetBgMean();
  BG_rms_av  = BgMap->GetNoiseMean();

  streamlog_out(DEBUG) << "  Average over whole image: " 
                       << BG_mean_av << " +/-" << BG_rms_av << endl;


  // Images can be float or int
  // In any case, copy pixel table to local float array

  double* floatPixels =  new double[sizeXY];

  if(cur_img->IsFloatImage())
    {
     double* oldPixels =  cur_img->GetFloatImageData();
   
     for(long ipix = 0; ipix < sizeXY; ipix++ )
             floatPixels[ipix]=oldPixels[ipix];                   
     }
  else
    {
     long* intPixels =  cur_img->GetIntImageData();
   
     for(long ipix = 0; ipix < sizeXY; ipix++ )
         floatPixels[ipix]=(double) intPixels[ipix];                   
    }


  // Get background and noise table (for cut)

  double* background = BgImage->GetFloatImageData();
  double* noise = NoiseImage->GetFloatImageData();

  // Map to sort pixels above threshold according to the charge

  std::multimap<double, long , std::greater<double> > pixelIndexMap;

  std::multimap<double, long , std::greater<double> >::iterator  pixelIndexIterator;

  // Vector of vectors, containing pixels belogning to given cluster

  std::vector< std::vector<long> > vectorClusterPixels;

  // Array to store cluster index for given pixel

  long* pixelClusterIndex = new long[sizeXY];

  // Now the actual clustering can start
  // Fill map with pixels above pixel cut

  for(long ipix = 0; ipix < sizeXY; ipix++ )
    {
     pixelClusterIndex[ipix] = -1;

     floatPixels[ipix] -= background[ipix];  // Subtract background here!!!

     // Use <= condition to avoid problems when noise=0

     if(floatPixels[ipix]<= _pixThreshold * noise[ipix] ) continue;

     pixelIndexMap.insert(make_pair( floatPixels[ipix], ipix ) );
    }

  // relative neighbour pixel positions on CCD

  long nbShiftX[] = { -1, 0, +1, -1,  1, -1, 0, +1,
		     -2, -1,  0,  +1,  +2,
                     -2, +2, -2,  +2,  -2, +2,
		     -2, -1,  0,  +1,  +2  };
  long nbShiftY[] = { -1, -1, -1, 0,  0,  1, 1,  1,
		      -2, -2, -2, -2, -2,
                       -1,   -1,  0,   0,     1,  1,
		      2,  2,  2,  2,   2  };
  long Nnb = 8;
  if(_moreNeighbours) Nnb=24;

  pixelIndexIterator=pixelIndexMap.begin(); 

  while(pixelIndexIterator != pixelIndexMap.end())
    {
      long pixID = pixelIndexIterator->second;
      double pixCharge=pixelIndexIterator->first;

      // First check if any of the neighbours is already in any cluster,
      // out of these, select neighbour with highest charge

      int imax=-1;
      double smax = 0.;

       for (int inb=0 ; inb<Nnb; inb++)
          {
	    int pixX = pixID%sizeX; 
	    int pixY = pixID/sizeX; 

           if(pixX + nbShiftX[inb]<0 || pixX + nbShiftX[inb] >=sizeX) continue;
           if(pixY + nbShiftY[inb]<0 || pixY + nbShiftY[inb] >=sizeY) continue;

           long nbPixel = pixID + nbShiftX[inb] + sizeX*nbShiftY[inb];

           if(pixelClusterIndex[nbPixel] == -1 ) continue;

  	   double nbCharge = floatPixels[nbPixel];

           if(nbCharge> smax)
	      {
    	      smax = nbCharge;
              imax = pixelClusterIndex[nbPixel]; 
	      }
	    
	   }

       // If cluster found - add pixel to it

      if(imax >=0)
	   {
	   vectorClusterPixels[imax].push_back(pixID);
           pixelClusterIndex[pixID] = imax;   
	   }


      // Otherwise check if pixel is above seed threshold
      // If so, create a new cluster with one pixel and add it to cluster list

      if( imax < 0 && pixCharge >= _seedThreshold * noise[pixID]  )
	   {
	  std::vector<long> newCluster; 
          newCluster.push_back(pixID);

          pixelClusterIndex[pixID] =  vectorClusterPixels.size() ;   

         // Control output
         //streamlog_out(DEBUG) <<  "Pixel with " << pixCharge << " ADU  at (" 
         //            << pixID%sizeX << "," << pixID/sizeX   
         //            <<  ")  used as a seed for cluster " << vectorClusterPixels.size() << endl;

          vectorClusterPixels.push_back(newCluster);
	   }
    	    
      pixelIndexIterator++;

   }  //   while(pixelIndexIterator != pixelIndexMap.end() && pixelIndexIterator->first >= pixThreshold )

  // Once clusters are reconstructed, calculated CoG position and signal, 
  // and add them to the object list

  for(int icluster=0; icluster < vectorClusterPixels.size() && icluster != _maxClusters ; icluster++)
   {
   GloriaClusterStatistics stat;
   double bgSum = 0.;

   for(int ipix=0; ipix< vectorClusterPixels[icluster].size() && ipix != _maxPixels; ipix++)
     {
       long ixy =  vectorClusterPixels[icluster].at(ipix);
       int ix = ixy % sizeX;
       int iy = ixy / sizeX;  

       double charge= floatPixels[ixy];

       // Position of the center of the first pixel i.e. (0,0) 
       // is defined to be (1.,1.) - that is why we have to add 1 here!!!

       stat.AddPixel(ix+1,iy+1,charge);

       // Sum noise level in quadrature (uncorrelated noise assumed)

       bgSum += noise[ipix]*noise[ipix];
     }

    new_list->AddObject( stat.GetXpos(), stat.GetYpos(), stat.GetSignal() );	

    new_list->SetLast("Pixels",(long) vectorClusterPixels[icluster].size() );


    if(_addNoise && bgSum > 0)
          new_list->SetLast(_noiseColumn,(double) sqrt(bgSum));


    if(_addSpread)
	    {
	      new_list->SetLast(_spreadColumns.at(0),(double) stat.GetXspread() );

	      new_list->SetLast(_spreadColumns.at(1),(double) stat.GetYspread() );

	      new_list->SetLast(_spreadColumns.at(2),(double) stat.GetXYcorr() );
	    }

    if(_addShape)
	    {
	      new_list->SetLast(_shapeColumns.at(0),(double) stat.GetFWHM() );

	      new_list->SetLast(_shapeColumns.at(1),(double) stat.GetElongation() );

	      new_list->SetLast(_shapeColumns.at(2),(double) stat.GetLength() );

 	      new_list->SetLast(_shapeColumns.at(3),(double) stat.GetWidth() );

	      new_list->SetLast(_shapeColumns.at(4),(double) stat.GetOrientation() );
	    }


    // End of loop over clusters. Done !
   }

  
  // Store additional keywords

  new_list->SetKeyword("PHOTOMTR","PIX_CLUS","Luiza PixelClusterFinder");

  int nstars = new_list->GetObjectNumber();
  new_list->SetKeywordInt("NSTARS",nstars,"Number of stars from PixelClusterFinder");

  new_list->SetKeywordInt("NX",sizeX,"Frame size (x)");
  new_list->SetKeywordInt("NY",sizeY,"Frame size (y)");
  
  new_list->SetKeywordDouble("BGMEAN",BG_mean_av,"Mean background level on the frame");
  new_list->SetKeywordDouble("BGRMS",BG_rms_av,"Mean background RMS on the frame");

  if(_frameDivision==1)
    {
    double seedCut = _seedThreshold * BG_rms_av + BG_mean_av;
    new_list->AddKeywordDouble("SEEDCUT",seedCut,"Seed pixel cut in PixelClusterFinder");

    double pixCut = _pixThreshold * BG_rms_av + BG_mean_av;
    new_list->AddKeywordDouble("PIXCUT",pixCut,"Other pixel cut in PixelClusterFinder");
    }

  
  delete [] pixelClusterIndex;
 



  // ==========================================================================================

        // Print created table if not permanent
  
        if(! (_outputCollectionName.size() > 0  && _permanentCollection))
	      {
              stringstream tabss;
              new_list->PrintHeader(tabss,20);  // Header
              new_list->Print(tabss,20);  // Table itself

              streamlog_out(MESSAGE) << " Object list for image " << ifm 
                                     << " in collection " << _inputCollectionName<< endl;

              streamlog_out(MESSAGE) << " Total of " << new_list->GetObjectNumber()
                                     << " objects found " <<  endl;

              streamlog_out(DEBUG) << tabss.str() << endl;
	      }


       if (_outputCollectionName.size() > 0  && data->TableCollectionExists(_outputCollectionName))
	     {
	      (data->GetTableCollection(_outputCollectionName)).push_back(*new_list);
	     _nOutputLists++;
	     }

    delete new_list;
    delete [] floatPixels;

    delete BgMap;   
    delete BgImage;
    delete NoiseImage;


    } //	for (int ifm = 0; ifm < nImg; ifm++) - loop over images in input collection


  // always return true  for ProcessorName
  setReturnValue(true);

}

void PixelClusterFinder::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void PixelClusterFinder::end()
{

	printEndMessage();

}

void PixelClusterFinder::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

}



} // namespace luiza
