#include "SimpleImageAdder.h"


#include "luiza/Exceptions.h"
#include <iostream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  SimpleImageAdder aSimpleImageAdder ;


  SimpleImageAdder::SimpleImageAdder() : Processor("SimpleImageAdder") {

    _description = "Processor to add images in the input collection."
                   "No corrections for possible shifts are applied." ;

   registerInputCollection (  ImageCollection,
                     "InputCollectionName",
                     "Name of the collection containing images to be added",
                     _inputCollectionName,  std::string ("") );


   registerOutputCollection (  ImageCollection,
                     "OutputCollectionName",
                     "Name of the collection where image sums will be stored",
                     _outputCollectionName,  std::string ("") );


   registerOptionalParameter ("PermanentCollection",
   "Flag output collection, as permanent, not deleted after loop is finished",
                              _permanentCollection,  false );


   registerOptionalParameter ("NumberAddImages",
"Number of images to be added each time (0 for all images in input collection)",
                              _nAddImages,  static_cast < int > (0) );

   registerOptionalParameter ("UseManyLoops",
   "Wait for images in next loops, if not enough images in input collection",
                              _useManyLoops,  false );

   registerOptionalParameter ("AverageImages",
   "Flag to select adding (false) or averaging (true) of input images",
                              _averageImage,  true );
  std::vector<std::string> initkeys;
 
  registerOptionalParameter(
			"MatchKeywords",
			"Names for header keywords which should match for added frames",
		        _matchKeywords, initkeys);
 }


 void SimpleImageAdder::init() { 

    streamlog_out( MESSAGE ) << "SimpleImageAdder::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nInputImages = 0 ;
    _nOutputImages = 0 ;
    _nAdded = 0;
    _nLoops = 0 ;

    _sumImage = NULL;

   // Check input parameters

    if(_inputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No input collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Nothing to do! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    if(_outputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No output collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Adding images not possible! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    _nKey = _matchKeywords.size();
    
    _initKeyVal = string("Not set");
    
    for(int ikey = 0; ikey < _nKey; ikey++)
        _matchValues.push_back(_initKeyVal);

    
    // Set return value to false (no sum written)
    
   setReturnValue( false ) ;
  }


void SimpleImageAdder::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "SimpleImageAdder::startAnalysis()" << endl;
}




  void SimpleImageAdder::processData( GloriaDataContainer* data ) 
{ 

    bool sumOK = false;
    
    // Set return value to false (no sum)
    
    setReturnValue( false ) ;
    

  // Return if input or output collection not specified
  // Error already printed in init()

    if(_inputCollectionName.size()==0 )return;

    if(_outputCollectionName.size()==0 )return;



    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    // Check if input collection exists
 
     if(! data -> ImageCollectionExists(_inputCollectionName))
	  {
          streamlog_out(MESSAGE) << "Input collection not found ! " << endl ;

          return;
	  }
      
     int nImg = data -> GetImageCollection(_inputCollectionName).size();

     if(nImg == 0)
	  {
          streamlog_out(MESSAGE) << "No images in input collection ! " << endl ;

          return;
	  }
      
     streamlog_out(MESSAGE) << "  Collection " << _inputCollectionName << " with " << nImg << " images " << endl ;


    // Create output collection if not found

    if(! data -> ImageCollectionExists(_outputCollectionName))
         data -> CreateImageCollection(_outputCollectionName,_permanentCollection);

     // Loop over images in input collection

     for(int ifm=0; ifm<nImg; ifm++)
        {

        GloriaFitsImage *inputImage = &((data->GetImageCollection(_inputCollectionName)).at(ifm));

	bool isMatch = true;
        for(int ikey = 0; ikey < _nKey; ikey++)
	  {
	  const char *keyName = _matchKeywords.at(ikey).c_str();

	  string keyVal = _initKeyVal;
	  if(inputImage->IsKeywordSet(keyName))
	                      keyVal = inputImage->GetKeywordValue(keyName);

	  if(keyVal != _matchValues.at(ikey)) isMatch = false;
	  }

	// Store current sum, if keywords do not match
	
	if(! isMatch && _nAdded > 0 )
	  {
            streamlog_out(DEBUG) << "Keywords do not match. Writing sum of " << _nAdded << " images " << endl;
	    
	    StoreSumImage(data);
            sumOK = true;
	  }
	  
        if(_nAdded == 0)
	  {
          _sumImage = new GloriaFitsImage(*inputImage);

          for(int ikey = 0; ikey < _nKey; ikey++)
	     {
	     const char *keyName = _matchKeywords.at(ikey).c_str();

	     string keyVal = _initKeyVal;
	     if(inputImage->IsKeywordSet(keyName))
	                      keyVal = inputImage->GetKeywordValue(keyName);

	     _matchValues[ikey] = keyVal;
	     }
	     
	  }
	
	if(_sumImage == NULL)
	   {
   	   streamlog_out(ERROR) << "Error in image sum allocation " << endl ;
	   streamlog_out(ERROR) << _sumImage->ErrorMessage() << endl ;
           _nAdded=0;
	   return;
	   }

         if(_nAdded > 0)
	   *_sumImage += *inputImage;
	 
         _nInputImages++;
         _nAdded++;

	 if(_sumImage->Error())
		{
		streamlog_out(ERROR) << "Error adding fits images from " << _inputCollectionName << endl ;
		streamlog_out(ERROR) << _sumImage->ErrorMessage() << endl ;
		continue;
		}

	  // Store the image if required number reached

	  if(_nAdded == _nAddImages)
	    {
              streamlog_out(DEBUG) << "Requested number of frames processed. Writing sum of " << _nAdded << " images " << endl;
	    	      
	      StoreSumImage(data);
              sumOK = true;
	    }
	  }
  
     // Store what is left, if in single loop mode

     if(_nAdded > 0 && ! _useManyLoops)
       {
        streamlog_out(DEBUG) << "No more frames in loop. Writing sum of " << _nAdded << " images " << endl;

	 StoreSumImage(data);
         sumOK = true;
       }
     
    // always return true  for this processor
     
     if(sumOK) setReturnValue( true ) ;
    
  }
  
  void SimpleImageAdder::StoreSumImage(GloriaDataContainer* data)
    { 
      if(_nAdded == 0) return;
      
     // Normalize if requested

     if(_averageImage && _nAdded>1) *_sumImage *= 1./(double)_nAdded;

     // Add summed image to output collestion

     (data->GetImageCollection(_outputCollectionName)).push_back(*_sumImage);

     delete _sumImage;
     _sumImage = NULL;

     _nAdded = 0;

     _nOutputImages++;

     return;
    }
  
  void SimpleImageAdder::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
void SimpleImageAdder::concludeAnalysis( GloriaDataContainer* data )
{

	streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;

	if(_nAdded == 0 ) return;

	// Last (partial) sum has not been written yet!

        streamlog_out(DEBUG) << " Writing last, uncomplete sum of " << _nAdded << " images " << endl;

        StoreSumImage(data);
 }

void SimpleImageAdder::end(){ 
    
    printEndMessage() ;

  }

  void SimpleImageAdder::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " Added "     << _nInputImages << " images "<< std::endl ; 

    streamlog_out(MESSAGE) << " Stored "    << _nOutputImages << " summed frames " << std::endl << std::endl;
    
  }
  
}// namespace luiza
