#include "SimpleImageRotator.h"
//#include "luiza/Exceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  SimpleImageRotator aSimpleImageRotator ;


  SimpleImageRotator::SimpleImageRotator() : Processor("SimpleImageRotator") {

   _description = "Processor to Rotate images in the input collection." ;


   registerInputCollection (  ImageCollection,
                     "InputCollectionName",
                     "Name of the collection containing images to be modified",
                     _inputCollectionName,  std::string ("") );


   registerOutputCollection (  ImageCollection,
                     "OutputCollectionName",
                     "Name of the collection where image sums will be stored",
                     _outputCollectionName,  std::string ("") );


   // Optional parameter to define crop window

  std::vector<int > initWindow;

  registerOptionalParameter ("WindowXY", 
              "Crop input image to given window (pixels numbered from 1 to N): X1 Y1 X2 Y2",
                             _WindowXY, initWindow);
                             

  std::vector<int > initRebin;

  registerOptionalParameter ("RebinXY", 
          "Rebin input image in X and Y (give one number for same rebinning)",
                             _rebinXY, initRebin);
                             

    registerOptionalParameter ("PermanentCollection",
   "Flag output collection, as permanent, not deleted after loop is finished",
                              _permanentCollection,  false );


   registerOptionalParameter ("RotDegree",
   "Flag to set rotation degree: 0 (default), 90, 180, 270",
                              _rotDegree,  static_cast < int > (0) );

   registerOptionalParameter ("FlipAxis",
   "Flip the image along given axis: 1 for X, 2 for Y, 0 for no flip (default)",
                              _flipAxis,  static_cast < int > (0) );
 }

  void SimpleImageRotator::init() { 

    streamlog_out( MESSAGE ) << "SimpleImageRotator::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nInputImages = 0 ;
    _nOutputImages = 0 ;
    _nLoops = 0 ;

  
   // Check input parameters

    if(_inputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No input collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Nothing to do! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }

 
    if(_outputCollectionName.size()==0 )
      {
      streamlog_out( ERROR ) << "No output collection name given! " << std::endl;
      streamlog_out( ERROR ) << "Rotating images not possible! " << std::endl;

       //        throw StopProcessingException( this ) ;
	return;
      }


    return;
 
  }



  void SimpleImageRotator::processData( GloriaDataContainer* data ) 
{ 

  // Return if input or output collection not specified
  // Error already printed in init()

    if(_inputCollectionName.size()==0 )return;

    if(_outputCollectionName.size()==0 )return;

    _origImage = NULL;
    _rotImage = NULL;

    _nLoops ++ ;

    // Check if input collection exists
 
     if(! data -> ImageCollectionExists(_inputCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      
     int nImg = data -> GetImageCollection(_inputCollectionName).size();

     if(nImg == 0)
	  {
          streamlog_out(DEBUG) << "No images in input collection ! " << endl ;

          return;
	  }
      
     streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with " << nImg << " images " << endl ;


    // Create output collection if not found

    if(! data -> ImageCollectionExists(_outputCollectionName))
         data -> CreateImageCollection(_outputCollectionName,_permanentCollection);

     // Loop over images in input collection

   
    for(int ifm=0; ifm<nImg; ifm++)
        {

	 _nInputImages++;
            
	 _origImage = &(data->GetImageCollection(_inputCollectionName).at(ifm));  


       long sizeX=0;
       long sizeY=0;

       sizeX = _origImage->SizeX();
       sizeY = _origImage->SizeY();

       streamlog_out(DEBUG) << "     Image " << _origImage->InputFileName()  << " is " 
                            << sizeX << "x" << sizeY << std::endl ;
       streamlog_out(DEBUG) << "     Image " << _origImage->InputFileName()  << " is Int: " 
                            <<_origImage->IsIntImage() << std::endl ;


       // Range of pixels considered

       long Xfirst,Xlast,Yfirst,Ylast;

       if(_WindowXY.size() == 4)
	     {
             Xfirst=_WindowXY[0] - 1;
	     Xlast=_WindowXY[2] - 1;

             sizeX=Xlast-Xfirst+1;

             Yfirst=_WindowXY[1] - 1;
  	     Ylast=_WindowXY[3] - 1;

             sizeY=Ylast-Yfirst+1;
	     }
           else
	     {
             Xfirst=0;
	     Xlast=sizeX - 1;

             Yfirst=0;
  	     Ylast=sizeY - 1;
	     }

       // Rebinning

       int rebinX=1, rebinY=1;
    
       if(_rebinXY.size() > 0)
             rebinX=rebinY=_rebinXY[0];
        
       if(_rebinXY.size() == 2)
	     rebinY=_rebinXY[1];
        

       if(sizeX%rebinX != 0)
	 {
         streamlog_out(ERROR) << "     Rebin factor does not match frame size in X: " << sizeX << " / " << rebinX << endl;
         rebinX=1;
	 }

       if(sizeY%rebinY != 0)
	 {
         streamlog_out(ERROR) << "     Rebin factor does not match frame size in Y: " << sizeY << " / " << rebinY << endl;
         rebinY=1;
	 }

       long sizeX_new= sizeX/rebinX;
       long sizeY_new= sizeY/rebinY;

       if( _rotDegree==90 ||  _rotDegree==270)
	 {
         sizeY_new= sizeX/rebinX;
         sizeX_new= sizeY/rebinY;
	 }

       streamlog_out(DEBUG) << "     Rotated image will be " << sizeX_new << " x " << sizeY_new << endl;

       // Allocate new image (same type)

       _rotImage = new GloriaFitsImage(sizeX_new,sizeY_new,_origImage->IsFloatImage() );  

        if(_rotImage == NULL)
	   {
   	   streamlog_out(ERROR) << "Error in image allocation " << endl ;
	   streamlog_out(ERROR) << _rotImage->ErrorMessage() << endl ;
	   return;
	   }


       // Copy header

       _rotImage->CopyHeader(_origImage);

       // Copy input file name

       _rotImage->SetInputFileName(_origImage->InputFileName());

       // Correct size description in the header

       if( sizeX_new !=  _origImage->SizeX() || sizeY_new !=  _origImage->SizeY())
	 {
          stringstream xpixS; xpixS << sizeX_new;
	  _rotImage->SetKeyword("NAXIS1",xpixS.str(),"Size changed in Luiza (!)" );

          stringstream ypixS; ypixS << sizeY_new;
	  _rotImage->SetKeyword("NAXIS2",ypixS.str(),"Size changed in Luiza (!)" );
	 }

        // Add note to image history

        stringstream hiss;
        hiss << "LUIZA " << _processorName << " (" << _typeName << ")";
	_rotImage->AddHistory(hiss.str());

	if(sizeX_new !=  _origImage->SizeX() || sizeY_new !=  _origImage->SizeY())
	  { 
          stringstream hiss1;
  	  hiss1 << "  Original image size  " << sizeX << " x "<< sizeY;
	  _rotImage->AddHistory(hiss1.str());
	  }
            

	if(rebinX !=1 || rebinY!=1)
	  { 
          stringstream hiss2;
  	  hiss2 << "  Rebined by factors (X,Y): " << rebinX << ", "<< rebinY;
	  _rotImage->AddHistory(hiss2.str());
	  }
            

	if(_rotDegree !=0)
	  { 
          stringstream hiss3;
  	  hiss3 << "  Rotated by " <<  _rotDegree;
	  _rotImage->AddHistory(hiss3.str());
	  }
            

	if(_flipAxis !=0)
	  { 
          stringstream hiss4;
  	  hiss4 << "  Flipped along axis " <<  _flipAxis;
	  _rotImage->AddHistory(hiss4.str());
	  }
            

        // Actual transformation
	//======================

	// Set rotation parameters

	long Xfirst_new = 0;
        long XstepX = 1, XstepY = 0;

	long Yfirst_new = 0;
        long YstepX = 0, YstepY = 1;


	if( _rotDegree==90)
	  {
	    Xfirst_new = 0;
	    XstepX = 0;  XstepY = 1;

            Yfirst_new = sizeY_new -1;
	    YstepX = -1;  YstepY = 0;
	  }

	if( _rotDegree==180)
	  {
	    Xfirst_new = sizeX_new-1;
	    XstepX = -1;

	    Yfirst_new = sizeY_new-1;
	    YstepY = -1;
	  }

	if( _rotDegree==270)
	  {
	    Xfirst_new = sizeX_new-1;
	    XstepX = 0;  XstepY = -1;

            Yfirst_new = 0;
	    YstepX = 1;  YstepY = 0;
	  }


	// Add image flip

	if( _flipAxis==2)
	  {
	    Xfirst_new = sizeX_new - 1 - Xfirst_new;
	    XstepX = -XstepX;  
            XstepY = -XstepY;
	  }

	if( _flipAxis==1)
	  {
	    Yfirst_new = sizeY_new - 1 - Yfirst_new;
	    YstepX = -YstepX;  
            YstepY = -YstepY;
	  }




       streamlog_out(DEBUG) << " New image loop parameters X: " << Xfirst_new << ", " 
			    << XstepX << ", " << XstepY << endl;

       streamlog_out(DEBUG) << " New image loop parameters Y: " << Yfirst_new << ", " 
			    << YstepX << ", " << YstepY << endl;


       // Do the transformation: loop is over pixels of the oryginal image 


	   if(_origImage->IsIntImage())
	     {

              long *pixels = _origImage->GetIntImageData();
	      long *pixels_new = _rotImage->GetIntImageData();

     
              long ipix_new= Xfirst_new;
              long jpix_new= Yfirst_new;

              for(long ipix=Xfirst; ipix<=Xlast;ipix+=rebinX)
		{
                for(long jpix=Yfirst; jpix<=Ylast;jpix+=rebinY)
                   {

		     //		    streamlog_out(DEBUG) << ipix << ", " << jpix << " => " 
		     //                  << ipix_new << ", " << jpix_new << endl;


                    long newpix = 0;

                    for(long i=0;i<rebinX;i++)
                      for(long j=0;j<rebinY;j++)
 			newpix += pixels[(jpix+j)*_origImage->SizeX()+ipix+i];


                    if(ipix_new<0 || ipix_new >=sizeX_new) ipix_new= Xfirst_new;
                    if(jpix_new<0 || jpix_new >=sizeY_new) jpix_new= Yfirst_new;

                    long npix_new=jpix_new*sizeX_new+ipix_new;
 	            pixels_new[npix_new]=newpix;



                    ipix_new += XstepY;
                    jpix_new += YstepY;
                    }
 
                 ipix_new += XstepX;
                 jpix_new += YstepX;
                }
	     }
	   else
	     {
             double *pixels = _origImage->GetFloatImageData();
	     double *pixels_new = _rotImage->GetFloatImageData();


             long ipix_new= Xfirst_new;
             long jpix_new= Yfirst_new;

             for(long ipix=Xfirst; ipix<=Xlast;ipix+=rebinX)
		{
                for(long jpix=Yfirst; jpix<=Ylast;jpix+=rebinY)
                   {
                    double newpix = 0.;

                    for(long i=0;i<rebinX;i++)
                      for(long j=0;j<rebinY;j++)
 			newpix += pixels[(jpix+j)*_origImage->SizeX()+ipix+i];



                    if(ipix_new<0 || ipix_new >=sizeX_new) ipix_new= Xfirst_new;
                    if(jpix_new<0 || jpix_new >=sizeY_new) jpix_new= Yfirst_new;

                    long npix_new=jpix_new*sizeX_new+ipix_new;
 	            pixels_new[npix_new]=newpix;


                    ipix_new += XstepY;
                    jpix_new += YstepY;
                    }

                 ipix_new += XstepX;
                 jpix_new += YstepX;
                 }
	     }
  


          (data->GetImageCollection(_outputCollectionName)).push_back(*_rotImage);

          delete _rotImage;

          _nOutputImages++;

    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
	}
}
  
  void SimpleImageRotator::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
  void SimpleImageRotator::end(){ 
    
    printEndMessage() ;

  }

  void SimpleImageRotator::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " Read "     << _nInputImages << " images "<< std::endl ; 

    streamlog_out(MESSAGE) << " Stored "    << _nOutputImages << " rotated frames " << std::endl << std::endl;
    
  }
  
}// namespace luiza
