#include "WeightedAperturePhotometry.h"
#include "gloria/GloriaMedianMap.h"
#include "gloria/GloriaClusterStatistics.h"


#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"
#include "gloria/GloriaAstrometry.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

WeightedAperturePhotometry aWeightedAperturePhotometry;

WeightedAperturePhotometry::WeightedAperturePhotometry() :
		Processor("WeightedAperturePhotometry")
{

	_description = "Processor to calculate aperture photometry on the frame.";

	registerInputCollection(ImageCollection, "ImageCollectionName",
			"Name of the collection containing input images",
			_imageCollectionName, std::string(""));

	registerInputCollection(TableCollection, "ListCollectionName",
			"Name of the collection containing object lists",
			_listCollectionName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where object list will be stored",
			_outputCollectionName, std::string(""));

  std::vector<float > initPosition;

	registerProcessorParameter(
			"ObjectPositions",
			"Positions of objects of interest (RA,DEC in degrees)",
			_objectPositions, initPosition);

  std::vector<float > initRadius;
  initRadius.push_back(2.5);

	registerProcessorParameter(
			"ApertureRadius",
			"Radius of circular aperture used for signal counting in pixels",
			_apertureRadius, initRadius);

  std::vector<float > initRadii;
  //
  //  If default value given, there is no way to clear it ?! 
  //
  //  initRadii.push_back(5.99);
  //  initRadii.push_back(7.01);

	registerProcessorParameter(
			"AnnulusRadii",
			"Inner and outer radius of the annulus used for background subtraction",
			_annulusRadii,initRadii);

	registerOptionalParameter(
			"BackgroundSampling",
			"Only every BackgroundSampling pixel is considered in background calculation",
			_backgroundSampling, (int)1 );

	registerOptionalParameter(
			"FrameDivision",
			"Frame division for background calculation",
			_frameDivision, (int)1 );

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for background map interpolation",
			_interpolationOrder, (int)1 );

	registerOptionalParameter(
			"SignalThreshold",
			"Signal is stored only if above the threshold (in noise units)",
			_signalThreshold, (double) 5. );

	registerOptionalParameter(
			"RemoveEmpty",
			"Flag to remove clusters with signal below the threshold",
			_removeEmpty, false );

        std::vector<std::string> initsig;
        initsig.push_back(string("Signal"));

        registerOptionalParameter(
			"BrightnessColumns",
			"Names of the object brightness columns",
			_brightnessColumns, initsig);

        std::vector<std::string> initmag;

	registerOptionalParameter(
			"NoiseColumns",
			"Names for column containing noise information.",
		        _noiseColumns, initmag);

	registerOptionalParameter(
			"MagnitudoColumns",
			"Names of the columns for instrumental magnitudo",
			_magnitudoColumns, initmag);

	registerOptionalParameter(
			"MagnitudoOffset",
			"Value of offset applied when calculating the instrumental magnitudo.",
			_magnitudoOffset, (double) 17. );

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);

	registerOptionalParameter(
			"UseTelescopeParameters",
			"Use network configuration file to update processor parameters on even by event basis",
			_useTelescopeParameters, false );

	registerOptionalParameter(
			"CorrectPosition",
			"Flag  for calculating CoG position from pixels within aperture.",
  		        _correctPosition, false);

std::vector<std::string> initpos;
initpos.push_back(string("CCD_X"));
initpos.push_back(string("CCD_Y"));

	registerOptionalParameter(
			"PositionColumns",
			"Names for columns containing corrected position.",
		        _positionColumns, initpos);
	
	registerOptionalParameter(
			"AddNoise",
			"Flag to add noise information to the output table.",
  		        _addNoise, false);

	registerOptionalParameter(
			"AddSpread",
			"Flag to add charge spread information to the output table.",
  		        _addSpread, false);

std::vector<std::string> initspread;
initspread.push_back(string("RMS_X"));
initspread.push_back(string("RMS_Y"));
initspread.push_back(string("CORR_XY"));

	registerOptionalParameter(
			"SpreadColumns",
			"Names for columns containing charge spread information.",
		        _spreadColumns, initspread);


	registerOptionalParameter(
			"AddShape",
			"Flag to add charge distribution shape information to the output table.",
  		        _addShape, false);

std::vector<std::string> initshape;
initshape.push_back(string("FWHM"));
initshape.push_back(string("ELONG"));
initshape.push_back(string("LENGTH"));
initshape.push_back(string("WIDTH"));
initshape.push_back(string("ANGLE"));

	registerOptionalParameter(
			"ShapeColumns",
			"Names for columns containing charge distribution shape information.",
		        _shapeColumns, initshape);

	registerOptionalParameter(
			"FillAllRadii",
			"Flag to fill position/spead/shape for all aperture radii.",
  		        _fillAllRadii, false);
	
	registerOptionalParameter(
			"CopyImageHeader",
			"Flag to copy the whole image header as a header of the object list.",
  		        _copyImageHeader, false);
	
        std::vector<std::string> initkey;
        initkey.push_back(string("DATE-OBS"));
        initkey.push_back(string("JD"));
        initkey.push_back(string("HJD"));
        initkey.push_back(string("OBJECT"));
        initkey.push_back(string("TIME_UT"));
        initkey.push_back(string("TELESCOP"));
        initkey.push_back(string("INSTRUME"));
        initkey.push_back(string("OBSERVAT"));
        initkey.push_back(string("CAMID"));
        initkey.push_back(string("RA"));
        initkey.push_back(string("DEC"));

	registerOptionalParameter(
			"HeaderRecords",
			"Keywords of image fits header, which should be copied to the fits header of the object list.",
		_copykey, initkey);



}


void WeightedAperturePhotometry::init()
{

	streamlog_out(MESSAGE) << "WeightedAperturePhotometry::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputLists = 0;
	_nLoops = 0;


	// Check input parameters

	if (_imageCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input image collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if (_listCollectionName.size() == 0 && _objectPositions.size() == 0)
	{
		streamlog_out(ERROR) << "No input object list nor positions given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if (_listCollectionName.size() != 0 && _objectPositions.size() != 0)
	{
		streamlog_out(ERROR) << "Can not use both input methods at the same time! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stoped! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

        // Flag indicating input method used
        _usingObjectList = _listCollectionName.size() != 0 ;

 
        // Flag indicating need for calculating pixel cluster statistics
        _usingStatistics = _correctPosition || _addSpread || _addShape ;

	
	// Number of aperure radii given
	
	_nAper = _apertureRadius.size();
	
	if ( _nAper == 0)
	{
		streamlog_out(ERROR) << "Aperture radius not specified! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stoped! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if (_brightnessColumns.size() > 1 && _brightnessColumns.size() < _nAper)
	{
		streamlog_out(ERROR) << "Brightness column names not specified properly! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stoped! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}


	if(_brightnessColumns.size() == 1 && _nAper > 1)
	  {
	    string basename = _brightnessColumns.at(0);

	    for(int iap = 1; iap<_nAper ; iap++)
	      {
              stringstream apername;

	      apername << basename << iap+1 ;

	      _brightnessColumns.push_back(apername.str());
	      }
	      
	  }
	
	if(_addNoise && _noiseColumns.size() == 0)
	      {
		streamlog_out(ERROR) << "Noise column name not defined! " << std::endl;
		streamlog_out(ERROR) << "Noise information will not be stored! " << std::endl;

		_addNoise = false;
	      }


	if(_addNoise && _noiseColumns.size() > 1 && _noiseColumns.size() < _nAper )
	      {
		streamlog_out(ERROR) << "Noise column names not specified properly! " << std::endl;
		streamlog_out(ERROR) << "Noise information will not be stored! " << std::endl;

		_addNoise = false;
	      }


	if(_addNoise && _noiseColumns.size() == 1 && _nAper > 1)
	  {
	    string basename = _noiseColumns.at(0);

	    for(int iap = 1; iap<_nAper ; iap++)
	      {
              stringstream apername;

	      apername << basename << iap+1 ;

	      _noiseColumns.push_back(apername.str());
	      }
	      
	  }
	
	

	
	if (_magnitudoColumns.size() > 1 && _magnitudoColumns.size() < _nAper)
	{
		streamlog_out(ERROR) << "Magnitudo column names not specified properly! " << std::endl;
		streamlog_out(ERROR) << "Magnitudo will not be calculated! " << std::endl;

		_magnitudoColumns.clear();
	}

	if(_magnitudoColumns.size() == 1 && _nAper > 1)
	  {
	    string basename = _magnitudoColumns.at(0);

	    for(int iap = 1; iap<_nAper ; iap++)
	      {
              stringstream apername;

	      apername << basename << iap+1 ;

	      _magnitudoColumns.push_back(apername.str());
	      }
	      
	  }
	
	
	if ( (_annulusRadii.size() !=0 && _annulusRadii.size() !=2) ||
	     (_annulusRadii.size() ==2 && _annulusRadii[0] >= _annulusRadii[1]) )
	{
		streamlog_out(ERROR) << "Annulus radii vector not properly defined! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stoped! " << std::endl;

		throw StopProcessingException( this ) ;
		return;
	}

	if ( _annulusRadii.size() ==0 )
	{
		streamlog_out(MESSAGE) << "Background annulus radii not defined! " << std::endl;
		streamlog_out(MESSAGE) << "Interpolated image background will be used! " << std::endl;
	}

	if (!_usingObjectList && _outputCollectionName.size() == 0)
	{
	        streamlog_out(ERROR) << "No output collection name given! " << std::endl;
		streamlog_out(ERROR) << "Created object lists will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}


	if(_addSpread && _spreadColumns.size() < 3)
	      {
		streamlog_out(ERROR) << "Spread column name list too short! " << std::endl;
		streamlog_out(ERROR) << "Charge spread information will not be stored! " << std::endl;

		_addSpread = false;
	      }


	if(_addShape && _shapeColumns.size() < 5)
	      {
		streamlog_out(ERROR) << "Shape column name list too short! " << std::endl;
		streamlog_out(ERROR) << "Cluster shape information will not be stored! " << std::endl;

		_addShape = false;
	      }


	_addMagnitudo = _magnitudoColumns.size() > 0;

	// Considered positions inside central pixel (for weights)

	int npos=9;
	double obPosX[9] = { -0.5, 0., 0.5, -0.5, 0., 0.5, -0.5, 0., 0.5};
	double obPosY[9] = { -0.5, -0.5, -0.5, 0., 0., 0., 0.5, 0.5, 0.5};
	double obw[9];

	// Considered pixel corners

	int npx=4;
	double pxPosX[4] = { -0.5,  0.5, -0.5, 0.5};
	double pxPosY[4] = { -0.5, -0.5,  0.5, 0.5};
	double rpx[4];

	// Initialize aperture calculation vectors

	_apShiftX = new std::vector<int>[_nAper];
	_apShiftY = new std::vector<int>[_nAper];
	_awShiftX = new std::vector<int>[_nAper];
	_awShiftY = new std::vector<int>[_nAper];
	_awWeights = new std::vector<double>[_nAper];
	
	// Main loop over aperture radii

	for(int iaper=0; iaper<_nAper; iaper++)
	  {
	
	// Center of the pixel where the object is found is (0.,0.)
	
        int shiftmax = _apertureRadius[iaper]+1.;

	// Loop over aperture pixels

        streamlog_out(DEBUG) << "Selecting pixels for aperture with R = " << _apertureRadius[iaper] << endl;

        for(int ix=-shiftmax; ix<=shiftmax; ix++)
        for(int iy=-shiftmax; iy<=shiftmax; iy++)
	  {
	    // Pixel center position
	    double x=ix;  
            double y=iy;

  	    double wmin=1.;
	    double wmax=0.;

	    for(int iobp=0;iobp<npos;iobp++)
	      {

	      for(int ipx=0;ipx<npx;ipx++)
		{
		  double dx= x+pxPosX[ipx]-obPosX[iobp];
		  double dy= y+pxPosY[ipx]-obPosY[iobp];

		  double r=sqrt(dx*dx+dy*dy);

		  rpx[ipx]=r;
		}

	      // Sort corner distances

	      std::sort(rpx,rpx+npx);

	      // Pixel outside aperture

	      if(rpx[0] >= _apertureRadius[iaper])
		{
		  obw[iobp]=0.;
		  wmin=0.;
		  continue;
		}

	      // Pixel inside aperture

	      if(rpx[3] <= _apertureRadius[iaper])
		{
		  obw[iobp]=1.;
		  wmax=1.;
		  continue;
		}

	      // Pixel at the boundary
              // Define direction angle in 0 - pi/4 range

	      double adx = abs(x-obPosX[iobp]);
	      double ady = abs(y-obPosY[iobp]);

              double phi = atan2(ady,adx);
	      if(ady>adx) phi = atan2(adx,ady);              

	      double delta = 0.5 * tan(phi);

	      if(rpx[1] >= _apertureRadius[iaper] && phi>0.)
		{
		  double h=_apertureRadius[iaper]-rpx[0];

		  obw[iobp]=h*h/sin(2*phi);
		}
	      else
	      if(rpx[2] <= _apertureRadius[iaper] && phi>0.)
		{
		  double h=rpx[3]-_apertureRadius[iaper];

		  obw[iobp]=1 - h*h/sin(2*phi);
		}
	      else
	      if(rpx[2] > _apertureRadius[iaper] && rpx[1]<_apertureRadius[iaper])
		{
		  obw[iobp] = delta + (1-2*delta)*(_apertureRadius[iaper]-rpx[1])/(rpx[2]-rpx[1]);
		}
	      else
		obw[iobp]=0.5;  // just in case - should never be used


	      if(obw[iobp]<wmin)wmin=obw[iobp];
	      if(obw[iobp]>wmax)wmax=obw[iobp];

	      }

	    // Pixel always outside aperture

	    if(wmax == 0.) continue;


	    // Pixel always inside aperture

            if(wmin == 1.)
	      {
		_apShiftX[iaper].push_back(ix);
		_apShiftY[iaper].push_back(iy);
                streamlog_out(DEBUG) << "Unweighted  pixel at (" << ix << "," << iy << ")"<< endl;
		continue;
	      }

	    // For weighted pixels

  	   _awShiftX[iaper].push_back(ix);
	   _awShiftY[iaper].push_back(iy);

	    for(int iobp=0;iobp<npos;iobp++)
	      _awWeights[iaper].push_back(obw[iobp]);

           streamlog_out(DEBUG) << "Weighted  pixel at (" << ix << "," << iy << ")"<< endl;
           streamlog_out(DEBUG)  << "    Weights: ";
           for(int iobp=0;iobp<npos;iobp++)
	     streamlog_out(DEBUG)  << obw[iobp] << "  " ;

           streamlog_out(DEBUG)  << endl;
	  }

	streamlog_out(MESSAGE) << "Aperture " << iaper <<  " with R = "
			       << _apertureRadius[iaper] << " has " 
                               << _apShiftX[iaper].size() << " unweighted pixels and "
                               << _awShiftX[iaper].size() << " weighted pixels" << endl;



	  }  // End of loop over apertures

// Annulus for background

	if(_annulusRadii.size()==2)
	  {
          int shiftmax = _annulusRadii[1]+0.5;

          for(int ix=-shiftmax; ix<=shiftmax; ix++)
          for(int iy=-shiftmax; iy<=shiftmax; iy++)
  	    {
	    double x=ix;
            double y=iy;
 
            double r=sqrt(x*x+y*y);

	    if(r >= _annulusRadii[0] && r <= _annulusRadii[1])
	      {
		_anShiftX.push_back(ix);
		_anShiftY.push_back(iy);
	      }
	    }


	   }

	streamlog_out(MESSAGE) << "Annulus with " 
               << _anShiftX.size() << " background pixels defined " << endl;


	return;

}

void WeightedAperturePhotometry::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_imageCollectionName.size() == 0)
		return;

	if (_listCollectionName.size() == 0 && _objectPositions.size() == 0)
	        return;


	_nLoops++;


	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collections exists

	if (!data->ImageCollectionExists(_imageCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_imageCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}

	if (_usingObjectList && !data->TableCollectionExists(_listCollectionName))
	{
		streamlog_out(DEBUG) << "Input object list collection not found ! " << endl;

		return;
	}


        int nTab = 1;
	if (_usingObjectList)
	  {
          nTab = data->GetTableCollection(_listCollectionName).size();
	
          if (nTab == 0)
  	     {
	     streamlog_out(DEBUG) << "No object list tables in input collection ! " << endl;
	     return;
	     }
	  }


	streamlog_out(DEBUG) << "  Collection " << _imageCollectionName << " with "
			<< nImg << " images " << endl;


	if (_usingObjectList)
	streamlog_out(DEBUG) << "  Collection " << _listCollectionName << " with "
			<< nTab << " object lists " << endl;


	if(_usingObjectList && nTab !=nImg)
	  {
	  streamlog_out(ERROR) << "  Number of object lists (" << nTab 
			       << ") does not match number of images (" << nImg 
                               << ") - skip event! " << endl;

          throw SkipEventException( this ) ;
	  }

        nTab = nImg;

	// Create output collection if not found (and name is specified)

	if (!_usingObjectList && _outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
	  data->CreateTableCollection(_outputCollectionName,
	  			_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;

	  }

	// Flag for creating new output tables

	bool newOutputCollection = true;

	// No new tables created if using object list

	if(_usingObjectList)newOutputCollection =false;

	// If output collection already exists and has same number of elements as input
	// then append objects to existing tables, no new tables created

	if(!_usingObjectList && _outputCollectionName.size() > 0)
	  {
           int nOut = data->GetTableCollection(_outputCollectionName).size();

  	   if(nOut == nImg) newOutputCollection = false;
	  }


	// Loop over images in input collection
	//=====================================

	for (int ifm = 0; ifm < nImg; ifm++)
	  {
          GloriaFitsImage *cur_img = &(data->GetImageCollection(_imageCollectionName).at(ifm));

	  // Skip empty images

	  int sizeX = cur_img->SizeX();
	  int sizeY = cur_img->SizeY();

	  long sizeXY = sizeX * sizeY;

	  if(sizeXY == 0) continue;

          // Skip if no astrometry in the header

          GloriaAstrometry ga;

	  if(!_usingObjectList && !ga.IsAstrometryPresent(cur_img)) 
	    {
	    streamlog_out(MESSAGE) << "No astrometry found: image is skipped! " <<  endl;
	    continue;
	    }

          _nInputImages++;

	  if(!_usingObjectList)
              ga.InitFromHeader(cur_img);  // Read astrometry from header 


	  // Update parameters for this telescope

	  if(_useTelescopeParameters)
                          updateTelescopeParameters(cur_img);

	  // Create output object list 

          GloriaObjectList *out_list;

	  if(_usingObjectList)
	    {
	    out_list = (GloriaObjectList *) &(data->GetTableCollection(_listCollectionName).at(ifm));
	    }
	  else
	    if(newOutputCollection)
	    {
            out_list = new GloriaObjectList;

	    // Copy some header information

	    if(_copyImageHeader)
	        out_list-> CopyHeader(cur_img);
	    else
	      {
              int ncopy=_copykey.size();

              for(int icopy=0; icopy<ncopy; icopy++)
	        if(cur_img->IsKeywordSet(_copykey[icopy].c_str()))
		   out_list->AddKeyword(cur_img->GetKeywordRecord(_copykey[icopy].c_str()));
	      }

	    }
	    else
	    {
	    out_list = (GloriaObjectList *) &(data->GetTableCollection(_outputCollectionName).at(ifm));
	    }


	  // Add columns with position and signal

	 if(!_usingObjectList && !out_list->ColumnExists("RA"))
             out_list->AddColumn("RA",(double) 0., string("deg"), string("Object RA from astrometry"));
         if(!_usingObjectList && !out_list->ColumnExists("DEC"))
             out_list->AddColumn("DEC",(double) 0., string("deg"), string("Object DEC from astrometry"));


	 // Loop over apertures
	 
	 for(int iaper=0; iaper<_nAper; iaper++)
	   {
           stringstream apernum;
           apernum << iaper+1;

           string apername = "";
           if(iaper>0) apername = apernum.str();

           stringstream apersize;
           apersize << _apertureRadius[iaper];

  
	 if(!out_list->ColumnExists(_brightnessColumns.at(iaper)))
             out_list->AddColumn(_brightnessColumns.at(iaper),(double) 0.,string(""),
				 string("Signal from WeightedAperturePhotometry, aperture ")+apersize.str());

         if(_addMagnitudo && !out_list->ColumnExists(_magnitudoColumns.at(iaper)))
              out_list->AddColumn(_magnitudoColumns.at(iaper),(double) 0.,string(""),
                                 string("Instrumental magnitudo for aperture ")+apersize.str());

         if(_addNoise && !out_list->ColumnExists(_noiseColumns.at(iaper)))
              out_list->AddColumn(_noiseColumns.at(iaper),(double) 0.,string(""),
                                 string("Noise level for aperture ")+apersize.str());

         if(iaper==0 || _fillAllRadii)
           {
	   if(_correctPosition)
	      {
	      if(!out_list->ColumnExists(_positionColumns.at(0)+apername))
	        out_list->AddColumn(_positionColumns.at(0)+apername,(double) -1., string("pixel"), 
                                                       string("CoG position in X"));

              if(!out_list->ColumnExists(_positionColumns.at(1)+apername))
	        out_list->AddColumn(_positionColumns.at(1)+apername,(double) -1., string("pixel"), 
                                                       string("CoG position in Y"));
	      }

	   if(_addSpread)
	      {
	      if(!out_list->ColumnExists(_spreadColumns.at(0)+apername))
	        out_list->AddColumn(_spreadColumns.at(0)+apername,(double) 0., string(""), 
                                                       string("Cluster charge spread in X"));

	      if(!out_list->ColumnExists(_spreadColumns.at(1)+apername))
	        out_list->AddColumn(_spreadColumns.at(1)+apername,(double) 0., string(""), 
                                                       string("Cluster charge spread in Y"));

	      if(!out_list->ColumnExists(_spreadColumns.at(2)+apername))
	        out_list->AddColumn(_spreadColumns.at(2)+apername,(double) 0., string(""), 
                                                       string("Cluster charge X-Y correlation"));
	      }

	   if(_addShape)
	      {
	      if(!out_list->ColumnExists(_shapeColumns.at(0)+apername))
	        out_list->AddColumn(_shapeColumns.at(0)+apername,(double) 0., string(""), 
                                                       string("Cluster size"));

	      if(!out_list->ColumnExists(_shapeColumns.at(1)+apername))
	        out_list->AddColumn(_shapeColumns.at(1)+apername,(double) 0., string(""), 
                                                       string("Cluster elongation"));

	      if(!out_list->ColumnExists(_shapeColumns.at(2)+apername))
	        out_list->AddColumn(_shapeColumns.at(2)+apername,(double) 0., string(""), 
                                                       string("Cluster lenght"));

	      if(!out_list->ColumnExists(_shapeColumns.at(3)+apername))
	        out_list->AddColumn(_shapeColumns.at(3)+apername,(double) 0., string(""), 
                                                       string("Cluster width"));

	      if(!out_list->ColumnExists(_shapeColumns.at(4)+apername))
	        out_list->AddColumn(_shapeColumns.at(4)+apername,(double) 0., string("deg"), 
                                                       string("Cluster orientation"));
	      }
	   }
	 
	   } // End of loop over apertures
	 
          out_list->AdjustColumns();

  // Calculate background map for the image, if annulus radii not given

  GloriaMedianMap* BgMap =NULL;

  if(_annulusRadii.size()!=2)
    {
    BgMap = new GloriaMedianMap( _frameDivision, _frameDivision);
    BgMap->CalculateBgMap( cur_img, _backgroundSampling);
    }

  // Images can be float or int

  double* floatPixels = NULL;
  long* intPixels = NULL;

  if(cur_img->IsFloatImage())
      floatPixels =  cur_img->GetFloatImageData();
  else
      intPixels =  cur_img->GetIntImageData();


  // Number of objects to consider

  int nObj = _objectPositions.size()/2;

  if(_usingObjectList) nObj = out_list->GetRowNumber();


  // Main loop: over considered objects

  for(int iObj=0; iObj<nObj; iObj++)
    {
      // Get object position 

      double ccdx,ccdy;
      double ra,dec;

    if(_usingObjectList)
      {
	ccdx  = out_list->GetValueOf<double>("CCD_X",iObj);
	ccdy  = out_list->GetValueOf<double>("CCD_Y",iObj);
      }
    else
      {
      ra=_objectPositions[2*iObj];
      dec=_objectPositions[2*iObj+1];

      // Transform to CCD coordinates

      ga.GetXYFromRaDec(ra,dec,&ccdx,&ccdy);

      // Skip this object if outside CCD frame

      if(ccdx<1 || ccdy<1 || ccdx>sizeX || ccdy>sizeY)
	{
	  streamlog_out(DEBUG) << "Specified position for object " << iObj << " outside CCD frame" << endl;
          continue;
	}

      }


  // Calculate background first

  int ix0 = ccdx-0.5;
  int iy0 = ccdy-0.5;

  double medBG = 0.;
  double sigBG = 0.;
  int anN=0;

  // From annulus (when defined)

  if(_annulusRadii.size()==2)
    {
   std::vector<double> anVec;

   for(int ipix=0;ipix< _anShiftX.size(); ipix++)
        {
        int ix=ix0+_anShiftX[ipix];
        int iy=iy0+_anShiftY[ipix];

         if(ix<0 || iy<0 || ix>=sizeX || iy>=sizeY )continue;

         if(cur_img->IsFloatImage())
            anVec.push_back(floatPixels[iy*sizeX+ix]);
         else
            anVec.push_back((double)intPixels[iy*sizeX+ix]);

         anN++;         
	}

    if(anN < 1)
     { 
      streamlog_out(WARNING) << "Object " << iObj << " at " << ccdx << "," << ccdy 
                    << " (on CCD):  no pixels in background annulus ?!?!" << endl;

      continue;
     }

    std::sort(anVec.begin(),anVec.end());

    int imed=anN/2;

    if(anN%2 == 1)     
      medBG=anVec[imed];
    else
      medBG=0.5*(anVec[imed]+anVec[imed-1]);

    int ilow  = (int)(0.17*anN);
    int ihigh = (int)(0.83*anN);

    sigBG=0.5*(anVec[ihigh]-anVec[ilow]);
    }
  else
    {
    medBG=BgMap->GetBgLevel(ccdx,ccdy,_interpolationOrder);

    // No interpolation for noise map (interpolation order forced to 0!)

    sigBG=BgMap->GetNoise(ccdx,ccdy,0);
    }


   // Loop over apertures

   std::vector<double> signal;
   std::vector<double> background;
   GloriaClusterStatistics *apStat = new GloriaClusterStatistics[_nAper];


   for(int iaper=0;iaper<_nAper;iaper++)
     {

   // Sum signal and calculate statistics (if requested)

   double apSig=0.;
   double apSum=0.;
   double apW=0;
   double apW2=0;
   
   // Sum unweighted pixels first

   for(int ipix=0;ipix< _apShiftX[iaper].size(); ipix++)
	{
	  int ix=ix0+_apShiftX[iaper][ipix];
	  int iy=iy0+_apShiftY[iaper][ipix];

          if(ix<0 || iy<0 || ix>=sizeX || iy>=sizeY )continue;

          if(cur_img->IsFloatImage())
            apSig=floatPixels[iy*sizeX+ix] - medBG;
          else
            apSig=(double)intPixels[iy*sizeX+ix] - medBG;

          if(_usingStatistics)apStat[iaper].AddPixel(ix+1,iy+1,apSig);

	  apSum+=apSig;
          apW++;         
          apW2++;         
	}


   // Sum weighted pixels 

   // Center of pixel is at integer position values, 
   // for first pixel it is (1.,1.) - that is why we have to subtract 1.

   double dx = ccdx-ix0-1.;
   double dy = ccdy-iy0-1.;

   int ivec;
   double fracx, fracy;

   if(dx<0)
     {
       ivec=0;
       fracx=1+2*dx;
     }
   else
     {
       ivec=1;
       fracx=2*dx;
     }


   if(dy<0)
     {
       fracy=1+2*dy;
     }
   else
     {
       ivec+=3;
       fracy=2*dy;
     }

   int npos=9;

   for(int ipix=0;ipix< _awShiftX[iaper].size(); ipix++)
	{
	  int ix=ix0+_awShiftX[iaper][ipix];
	  int iy=iy0+_awShiftY[iaper][ipix];

          if(ix<0 || iy<0 || ix>=sizeX || iy>=sizeY )continue;

	  double w = (1-fracx)*(1-fracy)*_awWeights[iaper][ivec   + npos*ipix] 
	           +  fracx   *(1-fracy)*_awWeights[iaper][ivec+1 + npos*ipix] 
	           + (1-fracx)*  fracy  *_awWeights[iaper][ivec+3 + npos*ipix] 
 	           +  fracx   *  fracy  *_awWeights[iaper][ivec+4 + npos*ipix];

          if(cur_img->IsFloatImage())
            apSig=floatPixels[iy*sizeX+ix] - medBG;
          else
            apSig=(double)intPixels[iy*sizeX+ix] - medBG;

          if(_usingStatistics)apStat[iaper].AddPixel(ix+1,iy+1,w*apSig);


          apSum+=w*apSig;
          apW+=w;         
          apW2+=w*w;         
	}


   if(iaper==0 && apW < 1.)
     { 
      streamlog_out(WARNING) << "Object " << iObj << " at " << ccdx << "," << ccdy 
                    << " (on CCD):  less than one weighted pixel (" << apW << ") ?!?!" << endl;

      continue;
     }

   signal.push_back(apSum);

   background.push_back(sqrt(apW2) * sigBG);


   if(iaper==0)
     { 
     streamlog_out(DEBUG) << "Object " << iObj << " at " << ccdx << "," 
                          << ccdy << " (on CCD) with signal = " << apSum << endl;


     streamlog_out(DEBUG) << "  from " << apSum + apW*medBG<< " counts in " << apW << " pixels with "
    		          << apW*medBG << "+/-" << sqrt(apW2) * sigBG << " background level" <<  endl;
     }
   
   } // End of loop over apertures


   // Use primary aperture radius for threshold cut
   
  double sigcut = _signalThreshold * background[0];
  
  if(!_usingObjectList && signal[0] < sigcut)
    streamlog_out(DEBUG) << "Signal at position " << iObj << ": "<< signal[0] << " below threshold: " << sigcut << endl;

  if(signal[0] < sigcut) continue;


  if(_correctPosition)
	    {
	      streamlog_out(DEBUG) << "     corrected position: " 
                                   << apStat[0].GetXpos() << ", " << apStat[0].GetYpos() 
                                   << " (" <<   apStat[0].GetSignal() << ")" << endl;
	    }

  int iSet = iObj;


  if(!_usingObjectList)
     {
     out_list->AddObject( ccdx, ccdy, signal[0]);	

     out_list->SetLast("RA",ra);
     out_list->SetLast("DEC",dec);

     iSet = out_list->GetObjectNumber() - 1;
     }


  // Fill result table
  // Loop over apertures (again :-)

  for(int iaper=0; iaper<_nAper; iaper++)
    {

    string apername = "";

    if(iaper>0)
      {
      stringstream apernum;
   
      apernum << iaper+1;

      apername = apernum.str();
      }
    
    out_list->SetValue(_brightnessColumns.at(iaper),iSet,signal[iaper]);

    if(_addMagnitudo && signal[iaper]>0.)
       {
	double magins = _magnitudoOffset - 2.5 * log10(signal[iaper]);

        out_list->SetValue(_magnitudoColumns.at(iaper),iSet,magins);
       }
    
    if(_addNoise)
      out_list->SetValue(_noiseColumns.at(iaper),iSet, background[iaper] );


    if(iaper==0 || _fillAllRadii)
      {
      if(_correctPosition)
	    {
	      out_list->SetValue(_positionColumns.at(0)+apername,iSet,(double) apStat[iaper].GetXpos() );

	      out_list->SetValue(_positionColumns.at(1)+apername,iSet,(double) apStat[iaper].GetYpos() );
	    }

      if(_addSpread)
	    {
	      out_list->SetValue(_spreadColumns.at(0)+apername,iSet,(double) apStat[iaper].GetXspread() );

	      out_list->SetValue(_spreadColumns.at(1)+apername,iSet,(double) apStat[iaper].GetYspread() );

	      out_list->SetValue(_spreadColumns.at(2)+apername,iSet,(double) apStat[iaper].GetXYcorr() );
	    }

      if(_addShape)
	    {
	      out_list->SetValue(_shapeColumns.at(0)+apername,iSet,(double) apStat[iaper].GetFWHM() );

	      out_list->SetValue(_shapeColumns.at(1)+apername,iSet,(double) apStat[iaper].GetElongation() );

	      out_list->SetValue(_shapeColumns.at(2)+apername,iSet,(double) apStat[iaper].GetLength() );

 	      out_list->SetValue(_shapeColumns.at(3)+apername,iSet,(double) apStat[iaper].GetWidth() );

	      out_list->SetValue(_shapeColumns.at(4)+apername,iSet,(double) apStat[iaper].GetOrientation() );
	    }
      }
    
    } //end of loop over apertures

  } // end of loop over objects in the list


  // Clean object table if requested

  if(_usingObjectList && _removeEmpty)
    {
      int nstart = out_list->GetRowNumber();
      int nrem=0;
      for(int iRow=nstart-1; iRow>=0; iRow--)
	if(out_list->GetValueOf<double>(_brightnessColumns.at(0),iRow) == (double) 0.)
	   {
	     out_list->DeleteRow(iRow);
	     nrem++;
	   }

      if(nrem>0)
         {
	   int nleft = out_list->GetRowNumber();
           streamlog_out(MESSAGE) << nrem 
                  << " objects with signal below threshold removed from object list " << endl;

           streamlog_out(MESSAGE) << nleft
		  << " objects remain on the list (out of " << nstart << ")" << endl;

         }

    }


   // Set additional keywords

  out_list->SetKeyword("PHOTOMTR","LUI_WAPP","Luiza WeightedAperturePhotometry");

  int nstars = out_list->GetObjectNumber();
  out_list->SetKeywordInt("NSTARS",nstars,"Number of stars from WeightedAperturePhotometry");


  out_list->SetKeywordInt("N_APERT",_nAper,"Number of apetures in WeightedAperturePhotometry");

  string rcomment = "Primary aperture radius";
  out_list->SetKeywordDouble("R_APERT",(double)_apertureRadius[0],rcomment);

  for(int iaper=0;iaper<_nAper;iaper++)
    {
    stringstream keyname;
    keyname << "R_APERT" << iaper+1;

    out_list->SetKeywordDouble(keyname.str().c_str(),(double)_apertureRadius[iaper],rcomment);
    rcomment="Other aperture radius";
    }
  
  
   // Print created table if not stored (DEBUG only)
 
   if (!_usingObjectList && _outputCollectionName.size() == 0)
	      {
              stringstream tabss;
              out_list->PrintHeader(tabss,20);  // Header
              out_list->Print(tabss,20);  // Table itself

              streamlog_out(DEBUG) << " Object list for image " << ifm 
                                     << " in collection " << _imageCollectionName<< endl;

              streamlog_out(DEBUG) << " Total of " << out_list->GetObjectNumber()
                                     << " objects measured " <<  endl;

              streamlog_out(DEBUG) << tabss.str() << endl;
	      }
 

       if (!_usingObjectList && _outputCollectionName.size() > 0  
                     && newOutputCollection)
	     {
	      (data->GetTableCollection(_outputCollectionName)).push_back(*out_list);
	     _nOutputLists++;
	     }

       if (newOutputCollection)
                delete out_list;
            
    } //	for (int ifm = 0; ifm < nImg; ifm++) - loop over images in input collection


  // always return true  for ProcessorName
  setReturnValue(true);

}

void WeightedAperturePhotometry::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void WeightedAperturePhotometry::end()
{

	printEndMessage();

}

void WeightedAperturePhotometry::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << _nInputImages << " images processed, " 
			       << _nOutputLists << " object lists stored " << std::endl;

}

} // namespace luiza
