#include "SimpleSourceFinder.h"
#include "luiza/Exceptions.h"
#include <iostream>
#include <sstream>

#include "gloria/GloriaObjectList.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

SimpleSourceFinder aSimpleSourceFinder;

SimpleSourceFinder::SimpleSourceFinder() :
		Processor("SimpleSourceFinder")
{

	_description = "Processor to find star centres on the frame.";

	registerInputCollection(ImageCollection, "InputCollectionName",
			"Name of the collection containing images to be added",
			_inputCollectionName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where object list will be stored",
			_outputCollectionName, std::string(""));

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);
}

void SimpleSourceFinder::init()
{

	streamlog_out(MESSAGE) << "SimpleSourceFinder::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputLists = 0;
	_nLoops = 0;


	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}

	if (_outputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No output collection name given! "
				<< std::endl;
		streamlog_out(ERROR) << "Object lists will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}

	return;

}

void SimpleSourceFinder::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->ImageCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_inputCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nImg << " images " << endl;

	// Create output collection if not found (and name is specified)

	if (_outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
	  data->CreateTableCollection(_outputCollectionName,
	  			_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;

	  }

	// Loop over images in input collection

	for (int ifm = 0; ifm < nImg; ifm++)
	{
 	// *** Initialize python and numpy ***
            PyObject *pName, *pModule, *pDict, *pFunc, *pValue, *pArgs;
	    Py_Initialize();	    
	    import_array();
	    pName = PyString_FromString("SimpleSourceFinder");

	    // Add source dir to search dir
	    char *luiza_python = getenv("LUIZA_PYTHON");
	    string pypath("sys.path.append(\"");
	    pypath+=string(luiza_python)+string("\")");
	    PyRun_SimpleString("import sys");
	    PyRun_SimpleString(pypath.c_str());

	    // Load the module object	    
	    pModule = PyImport_Import(pName);
	    pDict = PyModule_GetDict(pModule);
	    pFunc = PyDict_GetItemString(pDict, "find_sources");

	    // *** Prepare frame matrix for python ***

	    pArgs = PyTuple_New(1);
	    GloriaFitsImage *cur_img = new GloriaFitsImage((data->GetImageCollection(_inputCollectionName)).at(ifm));
	    PyObject *frame_mat;
	    npy_intp ndims[] = {cur_img->SizeY(), cur_img->SizeX()};

	    if(cur_img->IsIntImage())
	    	frame_mat = PyArray_SimpleNewFromData(2, ndims, NPY_INT, cur_img->GetIntImageData());
	    else
	    	frame_mat = PyArray_SimpleNewFromData(2, ndims, NPY_DOUBLE, cur_img->GetFloatImageData());
	    PyTuple_SetItem(pArgs, 0, frame_mat);

	    // *** Call the python module ***

	    pValue = PyObject_CallObject(pFunc, pArgs);
            if (pArgs != NULL)
              {
              Py_DECREF(pArgs);
              }
	    
            // *** Analyse the output ***

            if (pValue != NULL) 
             {
                //            printf("Return of call : %d\n", PyInt_AsLong(pValue));

        	int scnt = PyArray_SIZE(pValue)/3;

        	// Print list of sources
		cout << "Found " << scnt << " sources" << endl;

	       GloriaObjectList *new_list = new GloriaObjectList;

               for(int snum=0; snum<scnt; snum++)
        	{
        	cout << *(double*)(PyArray_GETPTR2(pValue, snum, 0)) << ", " << *(double*)(PyArray_GETPTR2(pValue, snum, 1)) << " " << *(double*)(PyArray_GETPTR2(pValue, snum, 2)) << endl;	

               new_list->AddObject( *(double*)(PyArray_GETPTR2(pValue, snum, 0)),
                                    *(double*)(PyArray_GETPTR2(pValue, snum, 1)),
                                    *(double*)(PyArray_GETPTR2(pValue, snum, 2)));	
               
        	}
            Py_DECREF(pValue);

	    // Print created table if not permanent
            if(! (_outputCollectionName.size() > 0  && _permanentCollection))
	      {
              stringstream tabss;
              new_list->Print(tabss,20);

              streamlog_out(MESSAGE) << " Object list for image " << ifm 
                                     << " in collection " << _inputCollectionName<< endl;

              streamlog_out(MESSAGE) << tabss.str() << endl;
	      }

  	   if (_outputCollectionName.size() > 0  && data->TableCollectionExists(_outputCollectionName))
	     {
	      (data->GetTableCollection(_outputCollectionName)).push_back(*new_list);
	     _nOutputLists++;
	     }

  	    delete new_list;
            }
          else 
            {
            PyErr_Print();
            streamlog_out(ERROR) << "Error in searching for sources for " << _inputCollectionName << endl;
            }


	}

	// always return true  for ProcessorName
	setReturnValue(true);

}

void SimpleSourceFinder::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void SimpleSourceFinder::end()
{

	printEndMessage();

}

void SimpleSourceFinder::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

}

} // namespace luiza
