#include "ExampleProcessor.h"
#include <iostream>
#include <sstream>
#include <cmath>

#include "luiza/VerbosityLevels.h"
#include "luiza/LuizaExceptions.h"
#include "streamlog/streamlog.h"

namespace luiza{



  ExampleProcessor aExampleProcessor ;


  ExampleProcessor::ExampleProcessor() : Processor("ExampleProcessor") {

    _description = "Simple example of a luiza processor."
      "Prints simple statistics (mean,RMS) for images found in given collection" ;

   registerInputCollection (  ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection, empty string for all collections",
                              _imageCollectionName,  std::string ("") );

   // Optional parameter to define window for mean and RMS calculation

  std::vector<int > initWindow;

  registerOptionalParameter ("WindowXY",
                      "Window for average and RMS calculation: X1 Y1 X2 Y2",
                             _WindowXY, initWindow);
                             

  registerOptionalParameter ("OutputTextFile",
                      "Name of the output text file to store the table",
                             _outputTextFile, std::string(""));
                             

 }

  void ExampleProcessor::init() { 

    streamlog_out( MESSAGE ) << "ExampleProcessor::init()  " << std::endl 
                             << "  processor name: " << name() << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

    _nImages = 0 ;
    _nImgCollections = 0 ;
    _nLoops = 0 ;
    _imgTable = NULL;
  }


 void ExampleProcessor::startAnalysis( GloriaDataContainer* data ) { 

    streamlog_out( MESSAGE ) << "ExampleProcessor::startAnalysis()  " << std::endl ;
    
    // Allocate table for statistics

    _imgTable = new GloriaFitsTable;

    // Define columns

    _imgTable->AddColumn("Num",LongColumn);
    _imgTable->AddColumn("Size X",LongColumn);
    _imgTable->AddColumn("Size Y",LongColumn);
    _imgTable->AddColumn("Average",DoubleColumn);
    _imgTable->AddColumn("RMS",DoubleColumn);
  }



  void ExampleProcessor::processData( GloriaDataContainer* data ) { 

    _nLoops ++ ;

    streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops << std::endl ;

    //------------ example code that tests/demonstrates ------------
    //------------ the usage of Exceptions to steer program flow ------------
    //
    //  --- skip every 7th event --------------

    if( !(  _nLoops % 7 )  )
               throw SkipEventException( this ) ;
    
    //  ---  stop afteer processing 20 events -----
  
    //if( _nLoops == 20   )
    //           throw StopProcessingException( this ) ;
    
    //  ---  abort if more than 20 events -----
  
    if( _nLoops > 20   )
               throw AbortProcessingException( this ) ;
    
    //---------end example code  ----------------------------------------
                             
                             
    int nCol=0;

    if(_imageCollectionName.size()==0)
      nCol = data -> GetImageCollectionNumber();
   
    else
      if(data -> ImageCollectionExists(_imageCollectionName)) nCol=1;
      

    if(nCol==0)
      {
	streamlog_out(DEBUG) << "  no images in data collection " << _imageCollectionName << std::endl ;
        return;
      }

    _nImgCollections+=nCol;

    // Loop over image collections (only one collection, if collection name given)

    for(int icl=0; icl<nCol; icl++)
      {
	std::string ColName = 
         (_imageCollectionName.size()==0)?data -> GetImageCollectionName(icl):_imageCollectionName;


	// Number of images in the collection

        int nImg = data -> GetImageCollection(ColName).size();

        streamlog_out(DEBUG) << "  Collection " << ColName << " with " << nImg << " images " << std::endl ;

	// Loop over images in the collection

        for(int ifm=0; ifm<nImg; ifm++)
	  {
	    GloriaFitsImage *frame =  &(data -> GetImageCollection(ColName).at(ifm));  

	    streamlog_out(DEBUG) << "     Image " << frame->InputFileName()  << " is " << frame->SizeX() 
                                   << "x" << frame->SizeY() << std::endl ;

           _nImages++;

           _imgTable->AddRow((long)_nImages, (long)frame->SizeX(), (long)frame->SizeY());

	   long double aver=0, rms=0;
           long npix=0;
           int sizeX = frame->SizeX();
           int sizeY = frame->SizeY();

	   if(_WindowXY.size() == 4)
	     {
             _Xmin=_WindowXY[0] - 1;
	     _Xmax=_WindowXY[2] - 1;

             _Ymin=_WindowXY[1] - 1;
  	     _Ymax=_WindowXY[3] - 1;
	     }
           else
	     {
             _Xmin=0;
	     _Xmax=sizeX - 1;

             _Ymin=0;
  	     _Ymax=sizeY - 1;
	     }
           
           int firstpix = _Xmin + _Ymin*sizeX;
           int lastpix  = _Xmax + _Ymax*sizeX;

	   if(lastpix >= sizeX*sizeY) lastpix=sizeX*sizeY;

           if(frame->IsIntImage())
	     {
              long *pixels = frame->GetIntImageData();

              for(long ipix=firstpix; ipix<=lastpix;ipix++)
		{
                  if(ipix%sizeX < _Xmin || ipix%sizeX > _Xmax)continue;

		  aver+=(double)pixels[ipix];
		  rms+=(double)pixels[ipix] * (double)pixels[ipix] ;
                  npix++;
		}
	     }

           if(frame->IsFloatImage())
	     {
              double *pixels = frame->GetFloatImageData();

              for(long ipix=firstpix; ipix<lastpix;ipix++)
		{
                  if(ipix%sizeX < _Xmin || ipix%sizeX > _Xmax)continue;

		  aver+= pixels[ipix];
		  rms+=  pixels[ipix] * pixels[ipix] ;
                  npix++;
		}
	     }


	   if(npix>0)
	     {
	       aver/=(double)npix;
	       rms/=(double)npix;
               rms = sqrt(rms - aver*aver);

	       _imgTable->SetLast("Average",(double)aver);
	       _imgTable->SetLast("RMS",(double)rms);
	     }
	  }

      }


    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
  // set ProcessorName.EvenNumberOfEvents == true if this processor has been called 2n (n=0,1,2,...) times 
    setReturnValue("EvenNumberOfEvents",  !( _nLoops % 2 )   ) ;

    
  }
  
  void ExampleProcessor::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
  void ExampleProcessor::concludeAnalysis( GloriaDataContainer* data ){ 
    
    printEndMessage() ;


    if(_outputTextFile.size() > 0 )
         _imgTable->WriteTextFile(_outputTextFile);
  }

  void ExampleProcessor::printEndMessage() const {

    streamlog_out(MESSAGE) << " concludeAnalysis() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) << " processed "     << _nImages << " images in " 
			   << _nImgCollections << " collections " << std::endl 
			   << std::endl ;
    
    stringstream tabss;

    _imgTable->Print(tabss,20);

    streamlog_out(MESSAGE) << " Image statistics collected during processing " << endl;

    if(_imageCollectionName.size()!=0)
      {
      streamlog_out(MESSAGE) << "   for image collection '" << _imageCollectionName << "':" << endl;
      }
    else
      streamlog_out(MESSAGE) << "   for all image collection found: " << endl;

    streamlog_out(MESSAGE) <<  tabss.str() << endl;


    // test deprecated method message 
    message<DEBUG>( " and this is really the final DEBUG message ....") ;
  }
  
  void ExampleProcessor::end() {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;

  }

}// namespace luiza
