// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef RootTreeWriter_h
#define RootTreeWriter_h 1

#include "RootProcessor.h"

// root includes

#include "TFile.h"
#include "TTree.h"

#include <fstream>

using namespace gloria;

namespace luiza {


  //! Processor for writing numerical tables to root tree structure
  /*! This processor converts tables to root tree and stores as
   *  binary root file. They can then loaded directly into root.
   *
   *  \par Input 
   *  Tables are taken from the data collection.
   *  User has to specify in the steering file:
   *
   *  \param TableCollectionName -
   *    Name of the table collection which should be stored to file(s)
   *
   *  \par Output
   *  GloriaFitsTables from given collection are stored to root
   *  tree(s). Root file names consist of file name root, table number
   *  (if multiple files should be created; starting from 1) and 
   *  \a .root extension. Tables are stored at each loop unless the
   *  collection is permanent - then they are stored only at
   *  concludeAnalysis() call. 
   *
   *  \param RootFileName -
   *        Output file name root, which will be appended by the table
   *        number, if multiple root files are created, and
   *        extension. If the path name is given (ending with '/'),
   *        input file name is used.
   *
   *  \param OneFile - flag for storing all tables in one big tree
   *                   (default is true).
   *
   *  \param MemoryResident - flag for filling created tree in memory
   *                   and writing to the file only when completed 
   *                   (memory resident tree) or writing it to disk
   *                   while filling (disk resident tree). Default is
   *                   true, i.e. memory resident, which can however
   *                   consume significant amount of memory.
   *
   *  \param SelectedColumns - names of columns to be stored in the
   *         tree. If not specified (empty list) all columns are written.
   *
   *  \param AddCounter - flag to add table number (if one file is
   *                     created) and row number as the variables in
   *                     the generated tree.
   *
   *  \param IntHeaderRecords - list of FITS header fields to be added
   *                            as long integer numbers to the root tree.
   *
   *  \param FloatHeaderRecords - list of FITS header fields to be added
   *                            as double numbers to the root tree.
   *
   *  \author A.F. Zarnecki, University of Warsaw
   */

   class RootTreeWriter : public RootProcessor {

  public:

    virtual RootTreeWriter*  newProcessor();


    RootTreeWriter();

    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    /*! Called for before first data loop
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;


    /*! Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;


    virtual void check( GloriaDataContainer* data ) ;


    /*! Called after all data loops to conclude analysis
     */
    virtual void concludeAnalysis( GloriaDataContainer* data ) ;


    /*! Called after data processing for clean up.
     */
    virtual void end() ;


  protected:

     //! Tree definition based on table
     void InitializeRootTree(string treeName, GloriaFitsTable *source);

     //! Tree definition based on table
     void FillRootTree(GloriaFitsTable *source);

     //! Write tree to file and close root file
     void WriteRootTree();

    //! Name of the table collection which should be stored
    string _tableCollectionName ;

    //! Output file name root
     string _rootFileName;

     //! Flag for output to single file
     bool _oneFile;

     //! Flag for keeping output tree in memory
     bool _memoryResident;

     //! Flag for output to single file
     bool _addCounter;

     //! Table columns to be stored
     std::vector< std::string > _selectedColumns;

     //! Header records to be stored as long integers
     std::vector< std::string > _intHeaderRecords;

     //! Header records to be stored as doubles
     std::vector< std::string > _floatHeaderRecords;

     //! Main root application instance 
     static TApplication* rootApl;

     //! Temporary table to prepare text output
     TTree* _outputTree;

     //! Output file name
     string _outputFileName;

     //! Output file pointer
     TFile* _outputFile;

     //! Loop counter
     int _nLoops ;

     //! Table counter, used to number file name
     int _nTables ;

     //! Counter for tables successfully written to file
     int _nWritten ;

     //! Row counter
     int _iRow;

     //! Pointer to table of long values
     long *_intBranch;

     //! Pointer to table of long values
     double *_floatBranch;

     //! Flag to recognize event loop
     bool _inLoop;

    void printEndMessage() const ;
  } ;

 } // end namespace luiza
#endif
