#include "RootImageViewer.h"
#include <iostream>
#include <sstream>
#include <cmath>
#include <unistd.h>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "Getline.h"

// root includes

#include "TSystem.h"
#include "TROOT.h"
#include "TCanvas.h"


namespace luiza{

  TTimer*  RootImageViewer::rootTimer = NULL;

  RootImageViewer aRootImageViewer ;

  RootImageViewer::RootImageViewer() : RootProcessor("RootImageViewer") {

    _description = "Simple image viewer for Luiza. Based on CERN root image class" ;

   registerInputCollection (  ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection to view",
                              _imageCollectionName,  std::string ("") );


   registerInputCollection (  TableCollection,
                              "ListCollectionName",
                              "Name of the object list collection, corresponding to the displayed images",
                              _listCollectionName,  std::string ("") );


  registerOptionalParameter(
			"ShowHistogram",
			"When set to true image is shown as root 2D histogram (TH2D)",
			_showHistogram, false);

  registerOptionalParameter(
			"PauseTime",
			"Time (in seconds) to pause after dispying image. Waits for <Enter> when set to -1",
			_pauseTime, (int)0);

  std::vector<int > initCanvas;

  registerOptionalParameter ("CanvasWH",
                      "Canvas geometry: width, height, x offset, y offset",
                             _CanvasGeom, initCanvas);
                             

   registerOptionalParameter("CanvasName",
                              "Name of the root canvas. Default is the image collection name.",
                              _canvasName,  std::string ("") );

   registerOptionalParameter("CanvasTitle",
                              "Title to be shown on the root canvas",
                              _canvasTitle,  std::string ("") );

}

  void RootImageViewer::init() { 

    streamlog_out( MESSAGE ) << "RootImageViewer::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;

   rootImage = NULL;
   rootHistogram = NULL;

   _nImages = 0 ;
   _nLoops = 0 ;
 
   if(_canvasName.size() == 0) _canvasName=_imageCollectionName;

 }



  void RootImageViewer::processData( GloriaDataContainer* data ) { 

    _nLoops ++ ;

    streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops << std::endl ;

    if(! data -> ImageCollectionExists(_imageCollectionName))
      {
	streamlog_out(DEBUG) << "  Input data collection not found: " << _imageCollectionName << std::endl ;
        return;
      }


    int nImg = data -> GetImageCollection(_imageCollectionName).size();


    if(nImg == 0)
      {
	streamlog_out(DEBUG) << "  No images in input collection: " << _imageCollectionName << std::endl ;
        return;
      }


    streamlog_out(DEBUG) << "  Collection " << _imageCollectionName << " with " << nImg << " images " << std::endl ;


// Create root application (if needed)

    InitializeRoot();

// Create image canvas if not found

    TCanvas * viewerCanvas;

    viewerCanvas = (TCanvas *) gROOT->GetListOfCanvases()->FindObject(_canvasName.c_str());

    if(!viewerCanvas)
      {
      int Cwid=600, Chig=600;
      int Cx=0, Cy=0;
 
      if(_CanvasGeom.size()>=2)
        {
	Cwid=_CanvasGeom.at(0);
	Chig=_CanvasGeom.at(1);
        }

      if(_CanvasGeom.size()>=4)
        {
	Cx=_CanvasGeom.at(2);
	Cy=_CanvasGeom.at(3);
        }

      viewerCanvas = new TCanvas(_canvasName.c_str(),_canvasTitle.c_str(),Cx,Cy,Cwid,Chig);
      }

    viewerCanvas->cd();


// Loop limitted to 1 image (!!!)

    int nShow=1;
    
    for(int ifm=0; ifm<nShow; ifm++)
       {
       GloriaFitsImage *frame =  &(data -> GetImageCollection(_imageCollectionName).at(ifm));  

       _nImages++;

       int sizeX = frame->SizeX();
       int sizeY = frame->SizeY();
       int sizeXY = sizeX*sizeY;

       double* pixels;

       if(frame->IsFloatImage())
          {
          pixels = frame->GetFloatImageData();

	  }
       else
          {
           long *lpixels = frame->GetIntImageData();

           pixels = new double[sizeXY];

           for(int ipix=0;ipix<sizeXY;ipix++)
	     pixels[ipix] = (double)lpixels[ipix];

          }

       if(!_showHistogram)
	 {
         if(rootImage) delete rootImage;

         rootImage = new TASImage("rootImage",pixels, sizeX, sizeY);

         if(_canvasTitle.size()>0)rootImage->SetTitle(_canvasTitle.c_str());

         rootImage->Draw();
	 }
       else
	 {
         if(rootHistogram) delete rootHistogram;
       

	   // This is too slow (?)
	   //
	   //	   rootHistogram = new TH2D("rootHistogram","Fits image",
	   //		    sizeX,0.5,sizeX+0.5,sizeY,0.5,sizeY+0.5);
	   //
	   //	   for(int ix=0;ix<sizeX;ix++)
	   //             for(int iy=0;iy<sizeY;iy++)
	   //	       rootHistogram->Fill(ix+1.,iy+1.,pixels[ix+iy*sizeX]);

	   rootHistogram = new TH2D("rootHistogram",_canvasTitle.c_str(),
			    sizeX-2,1.5,sizeX-0.5,sizeY-2,1.5,sizeY-0.5);


           rootHistogram->SetContent(pixels);

           rootHistogram->SetStats(kFALSE);

	   rootHistogram->Draw("zcol");
	 }


       viewerCanvas->Update();
       gSystem->ProcessEvents();

       if(!frame->IsFloatImage())delete [] pixels;


       // If histogram shown and event list available -> overlay objects

       if( _showHistogram &&

           data -> TableCollectionExists(_listCollectionName)  &&

           data -> GetTableCollection(_listCollectionName).size() == nImg)
	 {

         GloriaFitsTable *objects =  &(data -> GetTableCollection(_listCollectionName).at(ifm));  


          if(objects->IsKeywordSet("PIXCUT"))rootHistogram->SetMinimum(objects->GetKeywordValueDouble("PIXCUT"));


          vector<double>* Xvec = &(objects->GetVectorOf<double>("CCD_X"));
          vector<double>* Yvec = &(objects->GetVectorOf<double>("CCD_Y"));
          int Nobj = Xvec->size();

          rootGraph = new TGraph(Nobj,&(Xvec->front()),&(Yvec->front()));

          rootGraph->SetMarkerStyle(24);
          rootGraph->SetMarkerColor(3);
          rootGraph->SetMarkerSize(0.7);

          rootGraph->Draw("P");
	 }

       viewerCanvas->Update();
       gSystem->ProcessEvents();

   
    if(_pauseTime != 0)
      {
      if(!rootTimer)
         {
         streamlog_out(MESSAGE) << "Initializing root TTimer " << std::endl ;

         rootTimer = new TTimer("gSystem->ProcessEvents();", 50, kFALSE);
         rootTimer->TurnOn();
         }

      if(rootTimer)
        rootTimer->Start(-1,kFALSE);

      if(_pauseTime < 0 )
          Getline("\n Type <return> to continue:\n"); 
      else
	{
	 int lefttime = 1000*_pauseTime; // Time in us
	 while(lefttime>0)
	    {
  	      sleep(1);  // Will be interupted by timer after 50us anyway...
              lefttime-=50;
	    }
	}

      if(rootTimer)
        rootTimer->Stop();

      }

    } // End of loop over frames in input collection

    // always return true  for ProcessorName
    setReturnValue( true ) ;
    
  }
  
  void RootImageViewer::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << " check() " << std::endl ;
    
  }
  
  void RootImageViewer::end(){ 
    
    if(rootTimer) 
      {
       rootTimer->TurnOff();
       delete rootTimer;
       rootTimer=NULL;
      }

    if(rootImage) delete rootImage;
    if(rootHistogram) delete rootHistogram;

    TCanvas* viewerCanvas = (TCanvas *) gROOT->GetListOfCanvases()->FindObject(_imageCollectionName.c_str());

    if(viewerCanvas)
           viewerCanvas->Close();


    printEndMessage() ;

  }

  void RootImageViewer::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


    streamlog_out(MESSAGE) <<  _nImages << " images shown "  << std::endl ;
    
  }
  
}// namespace luiza
