#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <iomanip>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "RootTreeWriter.h"

namespace luiza {

  RootTreeWriter aRootTreeWriter ;


  // Constructor

  RootTreeWriter::RootTreeWriter() : RootProcessor("RootTreeWriter") {

    _description = "Processor for writing FITS images to disk"
                   "Write images to specified files" ;


  registerInputCollection   ( TableCollection,
                              "TableCollectionName",
                              "Name of the table collection which should be stored to file",
                              _tableCollectionName,  std::string ("Objects") );

  registerProcessorParameter ("RootFileName",
                              "Output file name, which will be used to store root tree",
                              _rootFileName,  std::string("output/luiza_tree") );

  registerOptionalParameter ("OneFile",
                             "Flag to store all tables in one big root file",
			     _oneFile,  true );

  registerOptionalParameter ("MemoryResident",
                             "Flag to keep tree in memory while filling. When false, tree is filled to disk",
			     _memoryResident,  true );

  registerOptionalParameter ("AddCounter",
                             "Flag to add table and row number as variables in the tree",
			     _addCounter,  false );

  std::vector< std::string > emptyStringVec;

  registerOptionalParameter ("SelectedColumns",
                             "List of table columns to be stored (empty for all)",
			     _selectedColumns, emptyStringVec );

  registerOptionalParameter ("IntHeaderRecords",
                             "List of header records to be added as long integer variables",
			     _intHeaderRecords, emptyStringVec );

  registerOptionalParameter ("FloatHeaderRecords",
                             "List of header records to be added as double variables",
			     _floatHeaderRecords, emptyStringVec );
  }


  RootTreeWriter*  RootTreeWriter::newProcessor() {
    return new RootTreeWriter;
  }


  void RootTreeWriter::init() {

    streamlog_out( MESSAGE ) << "RootTreeWriter::init()  " << name()
			     << endl
			     << "  parameters: " << endl
			     << *parameters()
			     << endl ;

    _outputTree = NULL;
    _outputFile = NULL;

    _intBranch = NULL;
    _floatBranch = NULL;

    _nLoops = 0 ;

    _nTables = 0 ;

    _nWritten = 0 ;


    _inLoop = false;

    InitializeRoot();

 }

  void RootTreeWriter::startAnalysis( GloriaDataContainer* data )
  {
   streamlog_out(MESSAGE) << "Starting data processing data" << endl ;
  }


  void RootTreeWriter::processData( GloriaDataContainer* data )
  {
	_nLoops ++ ;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << endl ;

        if(! data -> TableCollectionExists(_tableCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      

        if( data -> TableCollectionPermanent(_tableCollectionName))
	  {
          if(_isFirstEvent)
              streamlog_out(MESSAGE) << "Permanent collection - skiped ! " << endl ;

          return;
	  }

	// This is not very nice, but effective
	// Use concludeAnalysis() method also to process 'normal' data 

	_inLoop=true;

	concludeAnalysis( data );

	_inLoop=false;
}


 void RootTreeWriter::concludeAnalysis( GloriaDataContainer* data )
  {
     int nTab = 0;

     if(data -> TableCollectionExists(_tableCollectionName))
       { 
        nTab = data -> GetTableCollection(_tableCollectionName).size();


        if(data -> TableCollectionPermanent(_tableCollectionName) && nTab == 0)
	   {
           streamlog_out(DEBUG) << "No tables in permanent collection ! " << endl ;

           return;
	   }
       }
      
      
      streamlog_out(DEBUG) << "  Collection " << _tableCollectionName << " with " << nTab << " tables " << endl ;


      for(int itb=0; itb<nTab; itb++)
	  {
	   GloriaFitsTable *source =  &(data -> GetTableCollection(_tableCollectionName).at(itb));  

          _nTables++;


          // Create output tree

	  if(!_outputTree  || !_oneFile)
 	                   InitializeRootTree(_tableCollectionName,source);


	  FillRootTree(source);

          _nWritten++;

 	  if(!_oneFile)WriteRootTree();
	  }


    // Write only at the end, if single file  

    if(_oneFile && !_inLoop)WriteRootTree();

    // always return true  for ProcessorName
    setReturnValue( true ) ;

  }


  void RootTreeWriter::InitializeRootTree(string treeName, GloriaFitsTable *source)
  {
  // Prepare output file name first
  // Root file has to be opened first to allow disk resident TTree file...

   stringstream outname;

   outname << _rootFileName;

  // If the file name root ends with '/' append it with input file name (if present)

   int fnlen=_rootFileName.size();

   if( _rootFileName.at(fnlen-1) == '/' && source->InputFileName().size() > 0 )
        {
        int flen = source->InputFileName().size();

        int ilpos = source->InputFileName().rfind('.');

        if (ilpos != std::string::npos) flen = ilpos;

        int ifpos = source->InputFileName().rfind('/');

        if (ifpos == std::string::npos)
	  outname << source->InputFileName().substr(0,flen) ;
        else
  	  outname << source->InputFileName().substr(ifpos+1,flen-ifpos-1) ; 
	}

   if(!_oneFile)
      {
      outname.fill('0'); 
      outname << setw(4) << _nTables;
      }

   outname << ".root" ;

   _outputFileName =  outname.str();

  // Open output file (in not memory resident tree)

   if(!_memoryResident  && _oneFile)
     {
     _outputFile = new TFile( _outputFileName.c_str(), "CREATE" );

     if(!_outputFile)
         {
         streamlog_out(ERROR) << "Error opening output root file:  " << _outputFileName.c_str() << endl ;

         StopProcessingException( this );
         }


     streamlog_out(DEBUG) << "Root tree file " << _outputFileName.c_str() << " created "  << endl;
     }

  // Create TTree now

    _outputTree = new TTree(treeName.c_str() , "Root tree created with Luiza" ) ;

   int nint=0;
   int nfloat=0;

   if( _addCounter && _oneFile )
     {
      _outputTree->Branch( "Itable" , &_nTables , "Itable/I" ) ;
      nint++;
     }

   if( _addCounter )
     {
     _outputTree->Branch( "Irow" , &_iRow , "Irow/I" ) ;
      nint++;
     }


   // If list of columns is empty, all columns will be stored,
   // as found in the first processed table (!)

   if(_selectedColumns.size() == 0)
       for(int icol=0; icol < source->GetColumnNumber(); icol++)
 		   _selectedColumns.push_back(source->GetColumnName(icol));


   // Allocate data tables for tree filling

   int nCol = _selectedColumns.size();          

   int nIntRec = _intHeaderRecords.size();
   int nFloatRec = _floatHeaderRecords.size();
   int nRec = nIntRec + nFloatRec;

   int nAll = nCol + nRec;

  _intBranch = new long[nAll];
  _floatBranch = new double [nAll];

   // Set branches for header records

    for(int irec=0; irec< nIntRec; irec++)
       {
       string keyword = _intHeaderRecords.at(irec);

       string keydesc = keyword + "/L";
       _outputTree->Branch(keyword.c_str(),_intBranch+irec,keydesc.c_str()) ;

        nint++;
       }


    for(int irec=0; irec< nFloatRec; irec++)
       {
       string keyword = _floatHeaderRecords.at(irec);

       string keydesc = keyword + "/D";
       _outputTree->Branch(keyword.c_str(),_floatBranch+nIntRec+irec,keydesc.c_str()) ;

        nfloat++;
       }


   // Loop over table columns and add branches for each column

  for(int icol=0; icol<nCol; icol++)
    {
      string varname = _selectedColumns.at(icol);

      if(source->GetColumnType(varname) == LongColumn)
	{
	 string vardesc = varname + "/L";
         _outputTree->Branch(varname.c_str(),_intBranch+nRec+icol,vardesc.c_str()) ;
	 nint++;
	}

      if(source->GetColumnType(varname) == DoubleColumn)
	{
	 string vardesc = varname + "/D";
         _outputTree->Branch(varname.c_str(),_floatBranch+nRec+icol,vardesc.c_str()) ;
	 nfloat++;
	}
    }



   streamlog_out(MESSAGE) << "Root tree with " << nint << " integer and " 
			 << nfloat << " float variables defined" << endl;


  }


  void RootTreeWriter::FillRootTree(GloriaFitsTable *source)
  {

  int nRow = source -> GetRowNumber();
  int nCol = _selectedColumns.size();

  int nIntRec = _intHeaderRecords.size();
  int nFloatRec = _floatHeaderRecords.size();
  int nRec = nIntRec + nFloatRec;

  // Store header values (same for all rows)

  for(int irec=0; irec< nIntRec; irec++)
       {
       _intBranch[irec]=-1;

       string keyword = _intHeaderRecords.at(irec);
       if(!source->IsKeywordSet(keyword.c_str()))continue;

       long value = source->GetKeywordValueInt(keyword.c_str());
       _intBranch[irec]=value;
       }


  for(int irec=0; irec< nFloatRec; irec++)
       {
       _floatBranch[nIntRec+irec] = -1.;

       string keyword = _floatHeaderRecords.at(irec);
       if(!source->IsKeywordSet(keyword.c_str()))continue;

       double value = source->GetKeywordValueDouble(keyword.c_str());
       _floatBranch[nIntRec+irec] = value;
       }


  // Try to optimize table addressing

  int *selID = new int[nCol];

  for(int icol=0; icol<nCol; icol++)
     if(source->ColumnExists(_selectedColumns.at(icol)))
        selID[icol] = source->GetColumnId(_selectedColumns.at(icol));
     else 
        selID[icol] = -1;

  
  for(_iRow=0; _iRow < nRow; _iRow++)
    {

   // Loop over table columns store values for each column

    for(int isel=0; isel<nCol; isel++)
      {
      int icol = selID[isel];

      _intBranch[nRec+isel]=-1;
      _floatBranch[nRec+isel]=-1.;

      if(icol<0)continue;

      if(source->GetColumnType(icol) == LongColumn)
	_intBranch[nRec+isel] = source -> GetValueOf<long>(icol,_iRow);

      if(source->GetColumnType(icol) == DoubleColumn)
	_floatBranch[nRec+isel] = source -> GetValueOf<double>(icol,_iRow);
      }

    _outputTree->Fill();
    }
  }


  void RootTreeWriter::WriteRootTree()
  {
  // No tree to write
    
    if(! _outputTree)
      {
         streamlog_out(ERROR) << "Error: to tree to write...  " << endl ;
	 return;
      }

    
 // Open output file (for memory resident tree)

   if(_memoryResident || !_oneFile)
     {
     _outputFile = new TFile( _outputFileName.c_str(), "CREATE" );

     if(!_outputFile)
         {
         streamlog_out(ERROR) << "Error opening output root file:  " << _outputFileName.c_str() << endl ;

         StopProcessingException( this );
         }


     streamlog_out(DEBUG) << "Root tree file " << _outputFileName.c_str() << " created "  << endl;
     }


  // Write tree and close file

    _outputTree->Write();
    _outputFile->Close();

    delete[] _intBranch;
    delete[] _floatBranch;

    if(_memoryResident || !_oneFile)  delete _outputTree;

    _outputTree=NULL;
    _outputFile=NULL;
  }


  void RootTreeWriter::check(GloriaDataContainer* data)
  {
	    streamlog_out(DEBUG) << " check() " << endl;
  }

  void RootTreeWriter::end()
  {
	  printEndMessage() ;

  }

  void RootTreeWriter::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "
			   << " after " << _nLoops << " calls to processData() " << endl ;

    streamlog_out(MESSAGE) << _nTables << " tables processed " << endl;

    streamlog_out(MESSAGE) << _nWritten << " tables written to root tree " << endl
			   << endl ;
  }


}// namespace luiza
