// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef CalibrateFrameObjects_h
#define CalibrateFrameObjects_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to normalize brightness measurements on the whole frame
  /*! This processor analyses the object list of the input collection
   *  and normalizes all brightness measurements based on the
   *  reference stars found. Reference stars can be searched for in
   *  the \a GloriaSkyCatalog collection or selected from the object
   *  table itself (if already matched with the catalog, eg. with
   *  \a MatchSkyCatalog. \a AverageMap is used to create calibration
   *  map for the frame, based on the identified reference stars. The
   *  map is then applied to all objects in the table, estimated
   *  calibration error is also calculated.
   *
   *  \par Input 
   *  Processor reads object lists from input collection
   *
   *  \param InputCollectionName 
   *             Name of the collection containing \a GloriaObjectList
   *
   *  \param SkyCatalogName 
   *       Name of the collection containing \a GloriaSkyCatalog. If
   *       missing, reference magnitudo will be searched for in the
   *       input object table (assuming it was already matched to the
   *       catalog). 
   *
   *  \param InputImageCollection
   *       Name of the collection containing sky images, which were
   *       used to reconstruct analysed object lists. Image size has
   *       to be known to define calibration map. 
   *
   *  \param InputImageSize 
   *       Size of the image corresponding to the analysed
   *       object list. Image size has to be known to define
   *       calibration map. If image collection is not
   *       available, size can be specified by hand using this parameter.
   *
   *  \par Algorithm parameters
   *  User has to specify matching tolerance and other parameters
   *  related to calibration procedure
   *
   *  \param  MatchingTolerance
   *    Tolerance for object position matching, given in degrees
   *    
   *  \param  PositionColumns
   *    Names of the object table columns from which object position should
   *    be taken. These are 'RA' and 'DEC' by default (default column
   *    names for Luiza astrometry processor), but can be changed by
   *    user. 
   *
   *  \param BrightnessColumns
   *    Name of the object list column from which object brightness should
   *    be taken ('Signal' by default). Multiple columns can be given
   *    - independent calibration will be then performed for each
   *    column (from the same reference star set matched).
   *
   *  \param MagnitudoGiven
   *    Flag to indicate that the brightness stored in the input table
   *    is already in magnitudo scale (logarythmic). Default is
   *    'false'.
   *
   *  \param ReferenceMagColumn
   *    Names of the column in catalog table (or input object table,
   *    if \a SkyCatalogName is not given) which contains reference
   *    star magnitudo. Default is "Mag". 
   *
   *  \param  ReferenceMagRange 
   *     Range of the reference star magnitodo, which should be used
   *     for frame calibration
   *
   *  \param FrameDivision 
   *     Calibration map is calculated using \a AverageMap class, for
   *     frame divided into FrameDivision x FrameDivision subframes
   *                         and then interpolating between them.
   *
   *  \param AutoAdjustDivision 
   *     When this flag is set \a true, and there are too few entries
   *     to calculate (fit) calibration correction for each map bin,
   *     frame division is reduced to get more reliable calibration
   *     (number of entries per bin has to be larger than twice the
   *     number of parameters in the fit).
   *
   *  \param CorrectionFitOrder
   *    Order of the polynomial used to fit calibration correction
   *    dependence on the reference star position. The default is 0
   *    i.e. constant correction (averaged over selected reference
   *    stars) for each subframe. Polynomial fit is only performed if
   *    the sufficient number of reference stars is found (at least
   *    twice the number of parameters) in given subframe. Only
   *    correction value for the subframe center is stored in the
   *    calibration map. For polynomial order set to zero plain
   *    average is calculated, when set to negative value median
   *    from calibration corrections for reference stars is calculated
   *    in each bin.
   *
   *  \param InterpolationOrder 
   *     polinomial order used to interpolate between centers of
   *     subframes when calculating calibration correction. No
   *     interpolation is used for calibration uncertainty (!).
   *
   *  \par Output
   *  Corrected object magnitudos are stored in the output table
   *
   *  \param OutputCollectionName 
   *       Name of the collection where the output table will be
   *       stored. If missing, corrected magnitudo will be added to
   *       the input table.
   * 
   *  \param PermanentCollection 
   *         Flag output collection, as permanent, not deleted after
   *         loop is finished (default is false)
   *
   *  \param OutputMagnitudoColumns
   *          Names of columns to which calibrated object
   *          magnitudo should be stored. Default is 'Mag'. 
   *          Number of names has to match number of elements in
   *          BrightnessColumns parameter.
   *
   *  \param OutputErrorColumns
   *          Names of columns to which estimated calibration
   *          uncertainty should be stored. Default is 'err_Mag'.
   *          Number of names has to match number of elements in 
   *          BrightnessColumns parameter. If two names are given per
   *          input column, average calibration uncertainty is also stored.
   *
   *  \param CalibrationImageCollection
   *       Name of the image collection where the calibration correction
   *       image (created from map stored in \a AverageMap) will be
   *       stored. If missing, no image is stored. Image can be stored
   *       only for the first magnitudo column.
   * 
   *  \param ErrorImageCollection
   *       Name of the image collection where the calibration uncertainty
   *       image (created from map stored in \a AverageMap) will be
   *       stored. If missing, no image is stored.  Image can be stored
   *       only for the first magnitudo column.
   * 
   *  \param CalibrationMapCollection
   *       Name of the table collection where the calibration correction
   *       map (created from map stored in \a AverageMap) will be
   *       saved. If missing, no table is stored. Map can be stored
   *       only for the first magnitudo column.
   * 
   */

  class CalibrateFrameObjects : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new CalibrateFrameObjects ; }
    
    
    CalibrateFrameObjects() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the collection containing object lists
    string _inputCollectionName;

    //! Name of the collection containing sky catalog to be matched
    string _skyCatalogName;

    //! Name of the output table collection, with calibration results
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Name of the input image collection - just to get size of images
    string _inputImageCollection;

    //! Name of the image collection to store calibration correction image
    string _calibrationImageCollection;

    //! Name of the image collection to store calibration uncertainty image
    string _errorImageCollection;

    //! Name of the table collection to store calibration correction map
    string _calibrationMapCollection;


    //! Position matching tolerance for object and reference stars
    double _matchingTolerance;

    //! Names of the position columns in the table
    std::vector<std::string> _positionColumns;


    //! Names of the object brightness column
    std::vector<std::string> _brightnessColumns;


    //! Flag for brightness given in magintudo
    bool _magnitudoGiven;

    //! Name of the column containing reference magnitudo
    string _referenceMagColumn;


    //! Magnitudo range for reference stars used for calibration
    vector<float> _referenceMagRange;

    //! Frame division for calibration map calculation
    /*! Calibration correction is calculated for FrameDivision x FrameDivision 
     *  subimages and then interpolated between them
     */
    int _frameDivision;

    //! Flag to allow automatic adjustment of map binning for too few entries
    bool _autoAdjustDivision;

    //! Order of the polynomial fit of correction dependence on position
    int _correctionFitOrder;

    //! Polinomial order used to interpolate calibration map
    int _interpolationOrder; 

    //! Size of the image corresponding to the analysed object list (for map definition)
    vector<int> _inputImageSize;


    //! Names of columns for calibrated magnitudo
    vector<std::string> _outputMagnitudoColumns;

    //! Names of columns for estimated uncertainty
    vector<std::string> _outputErrorColumns;

     //! Flag to add calibration uncertainty
    bool _addError;

     //! Flag to add average calibration uncertainty
    bool _addAverageError;


    //! Number of calibrated signal columns
    int _nSignalColumn;
    
    int _nLoops ;
    int _nInputLists ;
    int _nObjectsMatched;

  } ;
  
} // end namespace luiza
#endif



