// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef MatchSkyCatalog_h
#define MatchSkyCatalog_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to match object table with sky catalog
  /*! This processor matches objects in the table of the input collection
   *  with sky objects stored in catalog table (gloria::GloriaSkyCatalog),
   *  based on their coordinates on the sky (RA and DEC from astrometry). 
   *  Selected entries (columns) from the catalog can also be copied
   *  into the object table (for matched entries). Matched catalog
   *  brightness can be then used eg. for image calibration.
   *
   *  \par Input 
   *  Processor reads object lists and catalog from input collections
   *
   *  \param InputCollectionName -
   *             Name of the collection containing object list with RA
   *             and DEC positions reconstructed (eg. after astrometry)
   *
   *  \param SkyCatalogName -
   *             Name of the collection containing \a GloriaSkyCatalog
   *
   *  \par Algorithm parameters
   *  User has to specify matching tolerance. 
   *
   *  \param  MatchingTolerance
   *    Tolerance for object position matching, given in degrees
   *    
   *  \param  PositionColumns
   *    Names of the object table columns from which object position should
   *    be taken. These are 'RA' and 'DEC' by default (default column
   *    names for Luiza astrometry processor), but can be changed by
   *    user. 
   *
   *  \par Output
   *  Object list is stored in the output table
   *
   *  \param OutputCollectionName 
   *          Name of the collection where the resulting objects table
   *          should be stored. If missing, input collection table
   *          will be modified.
   * 
   *  \param PermanentCollection 
   *         Flag output collection, as permanent, not deleted after
   *         loop is finished (default is false)
   *
   *  \param CopyColumns
   *          Names of the columns in the catalog, which should be
   *          copied to the output table 
   *
   *  \param NewColumnNames
   *          Names which should be used for copied columns. This 
   *          allows user to avoid conflicts of column names between
   *          two tables. If missing, same column name will be used
   *          for a copied column. 
   */

  class MatchSkyCatalog : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new MatchSkyCatalog ; }
    
    
    MatchSkyCatalog() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the collection containing object lists
    string _inputCollectionName;

    //! Name of the collection containing sky catalog to be matched
    string _skyCatalogName;

    //! Name of the output  collection (for modified object list)
    string _outputCollectionName;

    //! Position matching tolerance for object and reference stars
    double _matchingTolerance;

    //! Names of the position columns in the object table
    std::vector<std::string> _positionColumns;


    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Names of catalog columns to be copied in the output object table
    std::vector<std::string> _copyColumns;


    //! New column names, for columns copied from the catalog
    std::vector<std::string> _newColumnNames;

    int _nLoops ;
    int _nInputLists ;
  } ;
  
} // end namespace luiza
#endif



