// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef ObjectLightCurve_h
#define ObjectLightCurve_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to reconstruct light curve of given object
  /*! This processor analyses the object list of the input collection
   *  and finds measurements of the selected object based on its
   *  coordinates (on the sky or on the CCD). The measurement is then
   *  added to the new, permanent output collection. It is also
   *  possible to normalize the brightness measurement to the
   *  reference stars, by giving their positions and brightness.
   *  User can either select reference star(s) by himself 
   *  (using \b ReferenceStars parameter) or specify the name of a
   *  table collection, where reference stars are stored. 
   *  In the latter case, reference stars are selected based on 
   *  \b ReferenceMaxDist and \b MaxReferenceCount parameters. When
   *  multiple reference stars are selected, photometric calibration
   *  can be based on average magnitudo correction, correction median
   *  or polinomial fit of correction dependence on reference star
   *  position. With multiple reference stars it is also possible to
   *  estimate photometry uncertainty.
   *
   *  \par Input 
   *  Processor reads object lists from input collection
   *
   *  \param InputCollectionName -
   *             Name of the collection containing GloriaObjectList
   *
   *  \par Algorithm parameters
   *  User has to specify object position on the CCD or on the sky (if
   *  the astrometry is available) and the matching tolerance. 
   *
   *  \param  ObjectPosition 
   *    Expected object position. If (Ra,Dec) is used, it should be
   *    given in degrees.
   *
   *  \param  MatchingTolerance
   *    Tolerance for object position matching, given in same units as
   *    object position. 
   *    
   *  \param  PositionColumns
   *    Names of the table columns from which object position should
   *    be taken. These are 'RA' and 'DEC' by default (default column
   *    names for Luiza astrometry processor), but can be changed by
   *    user. 
   * 
   *  \param BrightnessColumn
   *    Name of the table column from which object brightness should
   *    be taken ('Signal' by default). 
   *
   *  \param MagnitudoGiven
   *    Flag to indicate that the brightness stored in the input table
   *    is already in magnitudo scale (logarythmic). Default is
   *    'false'.
   *
   *  \param  PolarAngleColumn
   *    When object position on the sky is given, distance has to be
   *    calculated differently than in case of CCD position. Namely,
   *    distance in RA has to be scaled by cosine of polar
   *    angle. Default value is 'DEC'.
   *
   *  \param  ReferenceStars
   *    When given, object brightness is normalized to this
   *    star. Multiple reference stars can also be given. 
   *    Position in same units as for object of interest and magnitudo
   *    should be given for each reference star (eg. RA,DEC,Mag). 
   *
   *  \param  ReferenceStarCatalog
   *    It is also possible to user reference star catalog stored in
   *    internal \a GloriaSkyCatalog table (it can be read from file 
   *    using eg. TableTextReader processors). \a GloriaSkyCatalog
   *    should include RA and DEC positions of the reference stars in
   *    the first two columns (by definition), one of the other
   *    columns need to specify object magnitudo (column name to be
   *    specified by \a ReferenceMagColumn parameter). Collection name
   *    should be specified in this parameter.
   *
   *  \param ReferenceMagColumn
   *    Names of the column in ReferenceStarCatalog table which
   *    contains reference star magnitudo. Default is "Mag".
   *
   *  \param  ReferenceStarFile
   *    No longer supported! If the list of reference stars is
   *    available as a text file, it should be first read into the
   *    data table using  TableTextReader processors.
   *
   *  \param ReferenceMaxDist
   *   Maximum allowed distance to reference star. When reference
   *   stars are read from file, value of this parameter is used to
   *   select a subset of reference stars for analysis. This allows
   *   user to have one reference file defined for given field, and
   *   then select proper stars for each object of interest
   *   automatically. If zero value given, all stars are used.
   *
   *  \param MaxReferenceCount
   *   Maximum number of reference stars. Stars closest to the object
   *   position are selected, within ReferenceMaxDist.
   *
   *  \param CorrectionFitOrder
   *  Order of the polynomial used to fit calibration correction
   *  dependence on the reference star position. The default is 0
   *  i.e. constant correction (averaged over selected reference
   *  stars). Polynomial fit is only performed if the sufficient
   *  number of reference stars is selected (at least twice the 
   *  number of parameters).
   *
   *  \param ReferenceUseMedian
   *  Flag for using median correction value instead of mean value for
   *  calibration with multiple reference stars. May help to avoid
   *  large correction fluctuations due to single reference stars.
   *
   *  \param EstimateError
   *  When multiple reference stars are used for photometry
   *  calibration and this flag is set to true, correction spread is
   *  used to get photometry error estimate. 
   *  
   *  \par Output
   *  Object ligth curve is stored in the permanent output table
   *
   *  \param OutputCollectionName 
   *            Name of the collection where light curve table will be stored
   * 
   *  \param CopyColumns
   *          Names of the columns from the input object list which
   *          should be copied to the light curve
   *
   *  \param HeaderRecords
   *          List of FITS header fields from the input table to be copied
   *          to the light curve table
   */

  class ObjectLightCurve : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new ObjectLightCurve ; }
    
    
    ObjectLightCurve() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    void initializeLCTable( GloriaDataContainer* data );

    void initializeRefStars( GloriaDataContainer* data );

    //! Name of the collection containing object lists
    string _inputCollectionName;

    //! Name of the output light curve collection
    string _outputCollectionName;

    //! Object position
    std::vector<float> _objectPosition;

    //! Position matching tolerance for object and reference stars
    double _matchingTolerance;

    //! Names of the position columns in the table
    std::vector<std::string> _positionColumns;


    //! Name of the object brightness column
    std::string _brightnessColumn;


    //! Flag for brightness given in magintudo
    bool _magnitudoGiven;

    //! Name of DEC column (for proper distance calculation)
    std::string _polarAngleColumn;


    //! Reference star positions and brightnesses 
    std::vector<float> _referenceStars;

    // Reference star file name(s) - no longer supported
    //    std::vector< std::string> _noninterference;

    //! Name of the collection containing reference stars
    string _referenceStarCatalog;


    //! Names of table columns containing reference star positions and magintudo
    string _referenceMagColumn;


    //! Maximum distance to reference star from file
    double _referenceMaxDist;

    //! Maximum number of reference stars to be used
    int _maxReferenceCount;

    //! Order of the polynomial fit of correction dependence on position
    int _correctionFitOrder;

    //! Flag for using median instead of mean for calibration to reference stars
    bool _referenceUseMedian;

    //! Flag  to do photometry error estimate, based on reference star spread
    bool _estimateError;

    //! Names of table columns to be copied in the light curve
    std::vector<std::string> _copyColumns;

    //! Names of table header records to be added in the light curve
    std::vector<std::string> _headerRecords;


    int _nLoops ;
    int _nInputLists ;
    int _nObjectsMatched;
    int _nOutputPoints ;

    double xmin,xmax,ymin,ymax;

    GloriaFitsTable *_lightCurveTable;


    // Private code for matrix equation solving

    int GaussjSolve(double *alfa,double *beta,int n);

  } ;
  
} // end namespace luiza
#endif



