// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef SelectNewObjects_h
#define SelectNewObjects_h 1

#include "luiza/Processor.h"
#include "gloria/GloriaSkyCatalog.h"
#include "gloria/GloriaAstrometry.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to identify new objects in the sequence of object lists
  /*! This processor matches objects from the input object list with
   *  objects from the previous list (previous frame). If an object is
   *  found which was not present in the previous list, it is
   *  considered as the new objects candidate. Additional cut can be
   *  applied on the signal to noise ratio. If astrometry is stored in
   *  the object list, it is also checked that the object was within
   *  the previous field of view. Selected rows from the input object
   *  list are copied to the output table.
   *
   *  \par Input 
   *  Processor reads object lists from input collection
   *
   *  \param InputCollectionName -
   *             Name of the collection containing input tables.
   *
   *  \par Output
   *  Object list is stored in the output table
   *
   *  \param OutputCollectionName 
   *          Name of the collection where the resulting table
   *          should be stored. If missing, input collection table
   *          will be modified.
   * 
   *  \param PermanentCollection 
   *         Flag output collection, as permanent, not deleted after
   *         loop is finished (default is false)
   *
   *
   *  \par Algorithm parameters
   *  User has to specify matching tolerance and significance cut. 
   *
   *  \param  MatchingTolerance
   *    Tolerance for object position matching, given in degrees
   *    
   *  \param  SignalThreshold -
   *           Threshold for the signal value, given in units of noise
   *           level (above background level). For signal below this
   *           threshold given object is considered a background
   *           fluctuation and is not included in the output list.
   *
   *  \param SignalColumns
   *    Names of the object table columns from which object brightness
   *    and noise level should be taken. These are 'Signal' and
   *    'Noise' by default.
   *
   *  \param  PositionColumns
   *    Names of the object table columns from which object position should
   *    be taken. These are 'RA' and 'DEC' by default (default column
   *    names for Luiza astrometry processor), but can be changed by
   *    user. 
   *
   *  \param ImageSize
   *    Size of the image from which the object list was created. If
   *    not given, it is taken from astometry parameters (assuming
   *    that the reference point was set to the centre of the image).
   *
   *  \param EdgeCut
   *     New objects which are found closer to the image edge than
   *     given number of pixels (on the old or new frame) are not
   *     considered as new objects. 
   *
   */

  class SelectNewObjects : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new SelectNewObjects ; }
    
    
    SelectNewObjects() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! Print end message
    void printEndMessage() const ;

    //! Name of the collection containing input table
    string _inputCollectionName;

    //! Name of the output table  collection (for selected rows)
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;


     //! Position matching tolerance for object and reference stars
    double _matchingTolerance;

     //! Threshold for the new object signal value, in units of noise
    double _signalThreshold;

    //! Names of the signal and noise columns in the object table
    std::vector<std::string> _signalColumns;

    //! Names of the position columns in the object table
    std::vector<std::string> _positionColumns;


    //! Size of the image corresponding to the object list given
    std::vector<int> _imageSize;


    //! Edge cut to remove new objects identified very close to the image edge
    int _edgeCut;

    // Previous frame is stored as a catalog - allow using standard matching tools
    GloriaSkyCatalog *_previousFrame;
    std::vector<long> *_newFlagVec;

    GloriaAstrometry _previousAstro;
    bool _astroStored;
    int _oldSizeX, _oldSizeY;

    int _nLoops ;
    int _nInputTables ;
  } ;
  
} // end namespace luiza
#endif



