// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef SelectTableData_h
#define SelectTableData_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to select data rows from the table
  /*! This processor can be used to select rows of the table
   *  (eg. reconstricted objects in the object list) by specifying
   *  simple cuts on values stored in table columns. Cuts can be
   *  applied on integer or float columns only (no string nor vector
   *  columns can be used for selection). Additional flag allows for
   *  sorting the output table by selected column value. 
   *
   *  \param InputCollectionName -
   *             Name of the collection containing input tables.
   *
   *  \param OutputCollectionName 
   *          Name of the collection where the resulting table
   *          should be stored. If missing, input collection table
   *          will be modified.
   * 
   *  \param PermanentCollection 
   *         Flag output collection, as permanent, not deleted after
   *         loop is finished (default is false)
   *
   *  \param SortColumn
   *          Name of the column which should be used for sorting the
   *          output table. 
   *
   *  \param ReverseSort
   *          Flag to indicate that output table should be sorted in
   *          reverse order.
   *
   *  \param ColumnValueRange
   *          Name of the column and two numbers specifying the
   *          allowed range for the column value. If the higher value
   *          is given first, values outside specified range will be
   *          selected. More that one cut can be specified by giving
   *          many sets of name plus two numbers. 
   *
   *  \param AcceptIfAny
   *          Flag indicating if given row should be selected when any
   *          selection criteria is fulfilled (i.e. \a OR of cuts is
   *          used). Default is \a false i.e. all cuts have to be
   *          fulfilled to select the data row.
   *
   *  \param OutputColumns
   *          Names of columns, which should be stored in output
   *          table. By default (if not given) all columns are
   *          stored. This option can also be used to change column
   *          order in the output table.
   *
   *  \param MinCount
   *          Minimum number of rows in the output table, for which
   *          the processor return value will be set to true.
   *
   */

  class SelectTableData : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new SelectTableData ; }
    
    
    SelectTableData() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! Print end message
    void printEndMessage() const ;

    //! Name of the collection containing input table
    string _inputCollectionName;

    //! Name of the output table  collection (for selected rows)
    string _outputCollectionName;

    //! Flag the output collection to be permanent
    bool _permanentCollection;

    //! Name of the column which should be used for table sorting
    string _sortColumn;

    //! Flag to indicate that output table should be sorted in reverse order.
    bool _reverseSort;

    //! Names of table columns with allowed value ranges
    std::vector<std::string> _columnValueRange;

    //! Names of table columns to be stored to output file
    std::vector<std::string> _outputColumns;

    //! Flag indicating if row should be selected by OR of cuts (when
    //! \a true) or by AND of cuts (when \a false)
    bool _acceptIfAny;
    
    //! Minimum number of selected rows for processor returning true value
    int _minCount;
    
    int _nLoops ;
    int _nInputTables ;
  } ;
  
} // end namespace luiza
#endif



