// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef TestHeaderKeyword_h
#define TestHeaderKeyword_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Simple processor for testing fits header
  /*! This processor returns the true value if the proper value of the  
   *  specified header keyword is found. Can be used to steer data
   *  processing. 
   *
   *  \param ImageCollectionName -
   *             Name of the collection containing input images.
   *
   *  \param TableCollectionName -
   *             Name of the collection containing input tables.
   *            Either image or table collection need to be given.
   *
   *  \param RequiredKeywords
   *          Header keywords which will be checked. If given keyword
   *          is not set, the decision is the same as for the wrong
   *          keyword value.
   * 
   *  \param RequiredValues
   *          Values of the considered header keywords for which the
   *          positive decision is taken.
   * 
   *  \param ValueTolerances
   *          Precision for value comparison. Numerical (floating
   *          point) comparison is used if the value is different than
   *          zero.  Otherwise (zero given) strings are compared.
   * 
   *  \param  RequireAllKeywords
   *          If true (default), positive decision will be taken only,
   *          if all header keywords mach the specified values.
   *          If false, positive decision is given already for one match.
   *
   *  \param  RequireAllData
   *          If true (default), positive decision will be taken only,
   *          if all headers in the collection pass the criteria. If
   *          false, it is enough that one table or image passes the
   *          selection. 
   */

  class TestHeaderKeyword : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new TestHeaderKeyword ; }
    
    
    TestHeaderKeyword() ;
    
    //! Called at the begin of the job before anything is read.
    /*! Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
 
    //! Called for every data collection

    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* ) ;
    
    
    //! Called after data processing for clean up.

    virtual void end() ;
    
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    string _imageCollectionName;
    string _tableCollectionName;

    std::vector<string> _requiredKeywords;
    std::vector<string> _requiredValues;
    std::vector<float>  _valueTolerances;
    
    std::vector<float>  _valueMin;
    std::vector<float>  _valueMax;

    bool _requireAllData;
    bool _requireAllKeywords;

    bool _useTableData;
    
    int _nLoops ;
    int _nImages ;
    int _nTables ;

  } ;
  
} // end namespace luiza
#endif



