#include "CalibrateFrameObjects.h"
#include "gloria/GloriaAverageMap.h"
#include "gloria/GloriaMedianMap.h"


#include "luiza/LuizaExceptions.h"
// #include <iostream>
// #include <sstream>
#include <fstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"
#include "gloria/GloriaSkyCatalog.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

CalibrateFrameObjects aCalibrateFrameObjects;

CalibrateFrameObjects::CalibrateFrameObjects() :
		Processor("CalibrateFrameObjects")
{

	_description = "Processor to reconstruct object light curve.";

	registerInputCollection(TableCollection, "InputCollectionName",
			"Name of the collection containing input object lists",
			_inputCollectionName, std::string(""));

	registerInputCollection(TableCollection, "SkyCatalogName",
			"Name of the collection containing star catalog",
			_skyCatalogName, std::string(""));

	registerInputCollection(ImageCollection, "InputImageCollection",
			"Name of the input image collection - just to get size of images",
			_inputImageCollection, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection  where output table will be stored",
			_outputCollectionName, std::string(""));

	registerOutputCollection(ImageCollection, "CalibrationImageCollection",
			"Name of the image collection to store calibration correction image",
			 _calibrationImageCollection, std::string(""));

	registerOutputCollection(ImageCollection, "ErrorImageCollection",
			"Name of the image collection to store calibration correction image",
			 _errorImageCollection, std::string(""));

	registerOutputCollection(TableCollection, "CalibrationMapCollection",
			"Name of the table collection to store calibration correction map",
			 _calibrationMapCollection, std::string(""));


	registerProcessorParameter(
			"MatchingTolerance",
			"Position matching tolerance for object and reference stars",
			_matchingTolerance, (double)0.02);

	std::vector<std::string > initColumns;

	initColumns.push_back(std::string("RA"));
	initColumns.push_back(std::string("DEC"));

	registerOptionalParameter(
			"PositionColumns",
			"Names of the position columns in the object table",
			_positionColumns, initColumns);

	std::vector<std::string > initBrightness;
	initBrightness.push_back(std::string("Signal"));

	registerOptionalParameter(
			"BrightnessColumns",
			"Names of the object brightness columns",
			_brightnessColumns, initBrightness );


	registerOptionalParameter(
			"MagnitudoGiven",
			"Flag to indicate that brightness is already given in magnitudo units (logarythmic scale)",
		_magnitudoGiven, false);


	registerOptionalParameter(
			"ReferenceMagColumn",
			"Names of the column in the reference star catalog containing star magnitudo",
			_referenceMagColumn, std::string("Mag"));


        std::vector<float > initRefRange;

	initRefRange.push_back(5.0);
	initRefRange.push_back(15.0);

	registerProcessorParameter(
			"ReferenceMagRange",
			"Magnitudo range for reference stars used for calibration",
			_referenceMagRange, initRefRange);


	registerOptionalParameter(
			"FrameDivision",
			"Frame division for background calculation",
			_frameDivision, (int)1 );

	registerOptionalParameter(
			"AutoAdjustDivision",
			"Flag to allow automatic adjustment of map binning for too few entries",
			_autoAdjustDivision, false );

	registerOptionalParameter(
			"CorrectionFitOrder",
			"Order of the polinomial for the calibration fit in the subframe. Set negative for using median.",
			_correctionFitOrder, int(0));

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for calibartion map interpolation",
			_interpolationOrder, (int)1 );


        std::vector<int > imageSize;

	registerOptionalParameter(
			"InputImageSize",
	"Size of the image corresponding to the analysed object list (for map definition)",
			_inputImageSize, imageSize );

	std::vector<std::string > outMagColumns;

	outMagColumns.push_back(std::string("Mag"));
	registerOptionalParameter(
			"OutputMagnitudoColumns",
			"Names of columns for calibrated magnitudo",
			_outputMagnitudoColumns, outMagColumns );


	std::vector<std::string > outErrColumns;
	outErrColumns.push_back(std::string("err_Mag"));

	registerOptionalParameter(
			"OutputErrorColumns",
			"Names of columns for calibration uncertainty",
			_outputErrorColumns, outErrColumns );


}


void CalibrateFrameObjects::init()
{

	streamlog_out(MESSAGE) << "CalibrateFrameObjects::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nLoops = 0;
	_nInputLists = 0;
	_nObjectsMatched = 0;

	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}

	if (_skyCatalogName.size() == 0)
	{
		streamlog_out(WARNING) << "Sky catalog collection name not given! " << std::endl;
		streamlog_out(WARNING) << "Reference stars will be searched for in input table! " << std::endl;
	}

	if (_outputCollectionName.size() == 0)
	{
		streamlog_out(WARNING) << "No output collection name given! " << std::endl;
		streamlog_out(WARNING) << "Input object collection will be modified! " << std::endl;
	}


	if (_brightnessColumns.size() == 0)
	{
		streamlog_out(ERROR) << "Error specifying input names! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}

	_nSignalColumn = _brightnessColumns.size();
	
	
	if (_outputMagnitudoColumns.size() != _brightnessColumns.size() )
	{
		streamlog_out(ERROR) << "Error specifying output column names! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}


	if (_outputErrorColumns.size() == 0)
	{
		streamlog_out(WARNING) << "No column name for calibration uncertainty given! " << std::endl;
		streamlog_out(WARNING) << "Only calibration result will be stored! " << std::endl;
	}


	_addError = true;
	_addAverageError = false;
	
	
	if (_outputErrorColumns.size() != _outputMagnitudoColumns.size() &&
	    _outputErrorColumns.size() != 2 * _outputMagnitudoColumns.size() )
	{
		streamlog_out(WARNING) << "Number of names for uncertainty does not match number of input columns! " << std::endl;
		streamlog_out(WARNING) << "Only calibration result will be stored! " << std::endl;

		_addError =  false;
	}

	if(_outputErrorColumns.size() == 2 * _outputMagnitudoColumns.size())_addAverageError = true;
	
	
	if (_referenceMagRange.size() != 2)
	{
		streamlog_out(ERROR) << "Error specifying reference star magnitudo range! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}

	if (_inputImageCollection.size() == 0 && _inputImageSize.size() != 2 )
	{
 	   streamlog_out(ERROR) << "Will not be able determine image size! " << std::endl;
	   streamlog_out(ERROR) << "Nothing to do! " << std::endl;

	   throw StopProcessingException( this ) ;
	}


}



void CalibrateFrameObjects::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "CalibrateFrameObjects::startAnalysis()" << endl;
}




void CalibrateFrameObjects::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops << std::endl;


	// Check if input collection exists

	if (!data->TableCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}


	// 
	int nTab = data->GetTableCollection(_inputCollectionName).size();

	if (nTab == 0)
	{
		streamlog_out(DEBUG) << "No object list in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nTab << " tables " << endl;




	// Create output collection if not found (and name is specified)

	if (_outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
          data->CreateTableCollection(_outputCollectionName,_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;
	  }

	// Create calibration image collection if not found (and name is specified)

	if (_calibrationImageCollection.size() > 0  &&
	   !data->ImageCollectionExists(_calibrationImageCollection))
	  {
          data->CreateImageCollection(_calibrationImageCollection,false);

	  streamlog_out(DEBUG) << "  Calibration image collection " << _calibrationImageCollection << " created " << endl;
	  }

	// Create error image collection if not found (and name is specified)

	if (_errorImageCollection.size() > 0  &&
	   !data->ImageCollectionExists(_errorImageCollection))
	  {
          data->CreateImageCollection(_errorImageCollection,false);

	  streamlog_out(DEBUG) << "  Error image collection " << _calibrationImageCollection << " created " << endl;
	  }


	// Create calibration map collection if not found (and name is specified)

	if (_calibrationMapCollection.size() > 0  &&
	   !data->TableCollectionExists(_calibrationMapCollection))
	  {
          data->CreateTableCollection(_calibrationMapCollection,false);

	  streamlog_out(DEBUG) << "  Calibration map collection " << _calibrationMapCollection << " created " << endl;
	  }


	// Check if catalog collection exists (if name specified)

	if (_skyCatalogName.size()>0 &&  !data->TableCollectionExists(_skyCatalogName))
	{
		streamlog_out(WARNING) << "Sky catalog collection not found ! " << endl;

		return;
	}


	// Check if catalog collection is not empty (if name specified)

	int nCat;
	GloriaSkyCatalog *catalog;


	if(_skyCatalogName.size()>0)
	  {
           nCat = data->GetTableCollection(_skyCatalogName).size();
 
	   if (nCat == 0)
	       {
		streamlog_out(WARNING) << "No catalog table in catalog collection ! " << endl;
		return;
	       }

	   if (nCat > 1)
	       {
	streamlog_out(WARNING) << "Multiple catalog tables founf in catalog collection ! " << endl;
	streamlog_out(WARNING) << "First catalog table will be used ! " << endl;
	       }

           catalog = (GloriaSkyCatalog *) &(data -> GetTableCollection(_skyCatalogName).at(0));  

  	   // Sort catalog in DEC, if not yet done - should speed up processing

  	   if(!catalog->IsSortedBy("DEC")) catalog->SortTable("DEC");
	  }
	else
	  {
	   nCat = 0;
	   catalog = NULL;
	  }


	// Check if source image collection exists (if name specified)

	if (_inputImageCollection.size()>0 &&  !data->ImageCollectionExists(_inputImageCollection))
	{
		streamlog_out(WARNING) << "Source image collection not found ! " << endl;

		if(_inputImageSize.size() != 2)  return;
	}



	int nImg;

	// Check if source image collection is not empty (if name specified)

	if(_inputImageCollection.size()>0 &&  data->ImageCollectionExists(_inputImageCollection))
	  {
           nImg = data->GetImageCollection(_inputImageCollection).size();
 
	   if (nImg != nTab)
	       {
	streamlog_out(WARNING) << "Number of source images differ from number of object lists!" << endl;
	streamlog_out(WARNING) << "Size of first image will be used for all lists! " << endl;
	       nImg = 1;
	       }
	  }
	else
	   nImg = 0;


  // Loop over object lists found in input collection

  for(int itb=0; itb<nTab; itb++)
    {
      GloriaObjectList *source = ( GloriaObjectList *) &(data -> GetTableCollection(_inputCollectionName).at(itb));  

     if(! source->ColumnExists(_positionColumns[0]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[0] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }


     if(! source->ColumnExists(_positionColumns[1]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[1] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }


     if(_skyCatalogName.size()==0  && ! source->ColumnExists(_referenceMagColumn))
	    {
	      streamlog_out(ERROR) << " Reference magnitudo column " << _referenceMagColumn
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }

 
     for(int isig=0;isig<_nSignalColumn;isig++)
       {
        if(! source->ColumnExists(_brightnessColumns[isig]))
	    {
	      streamlog_out(ERROR) << "  Brightness column " << _brightnessColumns[isig]
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }
       }
     

    _nInputLists++;

    // Lists of reference stars selected

    vector<double> refX;
    vector<double> refY;
    vector<double> *refCorr;

    refCorr = new vector<double>[_nSignalColumn];

    // Input objects

    int nObject = source->GetRowNumber();

    // GloriaObjectList does have X-Y CCD positions uniquely defined
    vector<double> *xvec = &(source->GetXposVector());
    vector<double> *yvec = &(source->GetYposVector());

    // Signal columns - allow  choice of multiple columns
    vector<double>** svec = new vector<double>*[_nSignalColumn];

    for(int isig=0; isig<_nSignalColumn; isig++)
        svec[isig] = &(source->GetVectorOf<double>(_brightnessColumns[isig]));

    // Result vectors
    vector<double> *mvec;
    vector<double> *evec;
    vector<double> *avec;
    vector<long> matchid;

    streamlog_out(DEBUG) << " Searching object list " << itb << " with "<< nObject << " objects " << std::endl;


    // If no catalog is specified: select reference stars from input table

    if(catalog)
      {
      // Match to star catalog if available

      matchid = catalog->MatchObjectTable(source,_matchingTolerance,_positionColumns);


      mvec = &(catalog->GetVectorOf<double>(_referenceMagColumn));
      }
    else
      {
      mvec = &(source->GetVectorOf<double>(_referenceMagColumn));

      for(int iobj=0; iobj<nObject; iobj++)
                         matchid.push_back(iobj);
      }


    for(int iobj=0; iobj<nObject; iobj++)
	{
	  long imatch = matchid.at(iobj);

	  // Skip objects without match

	  if(imatch<0) continue;

	  double mag = mvec->at(imatch);

	  // Skip objects with no or wrong reference magnitudo
	  // (assume that no reference magnitudo corresponds to value outside specified range)

	  if(mag < _referenceMagRange[0] || mag > _referenceMagRange[1]) continue;


	  // Skip negative or zero signal values (unless magnitudo is given)

          for(int isig=0; isig<_nSignalColumn; isig++)
	    {
	    if(! _magnitudoGiven && svec[isig]->at(iobj) <= 0.)  
	      {
  	       streamlog_out(DEBUG) << " Skiping object "<< iobj
	        << ", signal at column " << isig << " not positive:  " << svec[isig]->at(iobj) << std::endl;

	       continue;
	      }
	    }
	  
	  double xpos= xvec->at(iobj);
	  double ypos= yvec->at(iobj);

	  refX.push_back(xpos);
	  refY.push_back(ypos);

	  // Store calibration entries

         for(int isig=0; isig<_nSignalColumn; isig++)
	   {
  	    double sig= svec[isig]->at(iobj);

	    if(! _magnitudoGiven) sig = 18. - 2.5 * log10(sig);

	    refCorr[isig].push_back(mag - sig);
	   }
	 
	  _nObjectsMatched++;
	}

    // Define image size

    double sizeX, sizeY;

    if(_inputImageSize.size() == 2)
      {
	sizeX = _inputImageSize[0];
	sizeY = _inputImageSize[1];
      }


    int ifm = itb;
    if( nImg == 1) ifm = 0;

    if(nImg > 0)
      {
      GloriaFitsImage *img = &(data->GetImageCollection(_inputImageCollection).at(ifm));

      sizeX = img->SizeX();
      sizeY = img->SizeY();
      }

    // Total number of reference objects found

    int nall = refX.size();

    streamlog_out(DEBUG) << nall << " objects selected to calculate calibration map " << endl;

    int divMax = _frameDivision;
    int divMin = divMax;

    if(_autoAdjustDivision) divMin = 1;


    // Create callibration map (average map by default)

    GloriaAverageMap* calibrationMap = new GloriaAverageMap();

    // Cast pointer on GloriaMedianMap (to be used for negative fit order given)

    GloriaMedianMap* calibrationMedian = (GloriaMedianMap *)calibrationMap;

    int nmin=0;
    int nreq = (_correctionFitOrder+1)*(_correctionFitOrder+2) + 1;

    // For median:
    if(_correctionFitOrder < 0) nreq = 3;

    for(int iDiv = divMax; iDiv >= divMin ; iDiv-- )
      {
	if(_autoAdjustDivision && iDiv>1 && iDiv*iDiv*nreq > nall) continue;

	calibrationMap->InitializeMaps(iDiv,iDiv,_nSignalColumn);

        calibrationMap->SetImageSize(sizeX, sizeY);

	if(_correctionFitOrder > 0)
            nmin = calibrationMap->CalculateFitMap(0,&refX,&refY,&refCorr[0],_correctionFitOrder);

	if(_correctionFitOrder == 0)
            nmin = calibrationMap->CalculateMap(0,&refX,&refY,&refCorr[0]);

	if(_correctionFitOrder < 0)
            nmin = calibrationMedian->CalculateMap(0,&refX,&refY,&refCorr[0]);


	streamlog_out(DEBUG) << "Calibration map " << iDiv << "x" << iDiv 
                             << " order " << _correctionFitOrder 
                             << " fit returned " << nmin << endl;

	if(nmin < 0 || nmin >= nreq) break;
      }

    // Check if successful

    if (nmin < 0)
      {
	streamlog_out(ERROR) << "Error in calibration procedure for list " << itb << endl;
        continue;
      }

    if (nmin == 0)
      {
	streamlog_out(ERROR) << "No reference stars in one (or more) of calibration map bins!" << endl;
	streamlog_out(ERROR) << "Calibration procedure failed for list " << itb << endl;
        continue;

      }

    if (nmin < nreq )
      {
	streamlog_out(ERROR) << "Too few reference stars for selected correction fit order!" << endl;
	streamlog_out(ERROR) << "Calibration results uncertain for list " << itb << endl;
      }


    // Calculate calibration maps for remaining signal columns (with same binning)

    
     for(int isig=1; isig<_nSignalColumn; isig++)
       {
	if(_correctionFitOrder > 0)
            calibrationMap->CalculateFitMap(isig,&refX,&refY,&refCorr[isig],_correctionFitOrder);

	if(_correctionFitOrder == 0)
            calibrationMap->CalculateMap(isig,&refX,&refY,&refCorr[isig]);

	if(_correctionFitOrder < 0)
            calibrationMedian->CalculateMap(isig,&refX,&refY,&refCorr[isig]);
       }
    
    // Store calibration images/map (if collection names given)

    if(_calibrationImageCollection.size() > 0 )
     {
      GloriaFitsImage *CalImage = calibrationMap->GetAverageImage(0,_interpolationOrder);

      (data->GetImageCollection(_calibrationImageCollection)).push_back(*CalImage);

       delete CalImage;
     }


    if(_errorImageCollection.size() > 0 )
     {
       // No interpolation for error image (interpolation order is always 0!)
      GloriaFitsImage *ErrImage = calibrationMap->GetSpreadImage(0,0);

      (data->GetImageCollection(_errorImageCollection)).push_back(*ErrImage);

       delete ErrImage;
     }


    if(_calibrationMapCollection.size() > 0 )
     {
       GloriaFitsTable *CalMap = calibrationMap->GetTable();

     (data->GetTableCollection(_calibrationMapCollection)).push_back(*CalMap);

       delete CalMap;
     }


    // Create output table as a copy of the input table (if name is specified)
    // Store it in the output collection and retrive pointer for future use.
    // Otherwise use input table pointer for output.

    GloriaObjectList *result;

    if (_outputCollectionName.size() > 0)
      {
      result = new GloriaObjectList(*source);

      (data->GetTableCollection(_outputCollectionName)).push_back(*result);

      int nOut = (data->GetTableCollection(_outputCollectionName)).size();

      delete result;
      result = ( GloriaObjectList *) &(data->GetTableCollection(_outputCollectionName).at(nOut-1));
      }
    else
      result = source;


    // Check if output columns exist, create if needed

    for(int isig=0; isig<_nSignalColumn; isig++)
       {
       if(!result->ColumnExists(_outputMagnitudoColumns.at(isig)) )
           result->AddColumn(_outputMagnitudoColumns.at(isig),(double) -10., string("mag"), 
                                                       string("Calibrated magnitudo from Luiza"));

       if(_addError && !result->ColumnExists(_outputErrorColumns.at(isig)) )
         result->AddColumn(_outputErrorColumns.at(isig),(double) 0., string("mag"), 
                                                       string("Calibration uncertainty from Luiza"));

       if(_addAverageError && !result->ColumnExists(_outputErrorColumns.at(_nSignalColumn+isig)) )
         result->AddColumn(_outputErrorColumns.at(_nSignalColumn+isig),(double) 0., string("mag"), 
                                                       string("Average frame calibration uncertainty"));
       }
    
    result->AdjustColumns();

    // Apply calculated calibration.
    // Set pointers to output table columns

    xvec = &(result->GetXposVector());
    yvec = &(result->GetYposVector());

    for(int isig=0; isig<_nSignalColumn; isig++)
      {
       svec[isig] = &(result->GetVectorOf<double>(_brightnessColumns[isig]));

       mvec = &(result->GetVectorOf<double>(_outputMagnitudoColumns[isig]));

       if(_addError)
         evec = &(result->GetVectorOf<double>(_outputErrorColumns[isig]));
       else
         evec = NULL;

       if(_addAverageError)
         avec = &(result->GetVectorOf<double>(_outputErrorColumns[isig+_nSignalColumn]));
       else
         avec = NULL;


       // Row number should be the same for source and result, but just in case:

       nObject = result->GetRowNumber();

       for(int iobj=0; iobj<nObject; iobj++)
   	  {
	  double xpos= xvec->at(iobj);
	  double ypos= yvec->at(iobj);
	  double sig= svec[isig]->at(iobj);

	  // negative or zero signal values - no calibration possible

	  if(! _magnitudoGiven && sig <= 0.)   continue;


	  if(! _magnitudoGiven) sig = 18. - 2.5 * log10(sig);

	  double corr = calibrationMap->GetAverage(isig,xpos,ypos,_interpolationOrder);
          mvec->at(iobj)= sig + corr;


	  // No interpolation for error map (interpolation order forced to 0!)

	  if(evec)
              evec->at(iobj)= calibrationMap->GetSpread(isig,xpos,ypos,0);

	  if(avec)
              avec->at(iobj)= calibrationMap->GetSpreadMean(isig);
	}

      }
    
    }
     // end of loop over input object lists 



  // always return true  for ProcessorName
  setReturnValue(true);

}

void CalibrateFrameObjects::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void CalibrateFrameObjects::concludeAnalysis( GloriaDataContainer* data )
{

	streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;

}


void CalibrateFrameObjects::end()
{

	printEndMessage();

}

void CalibrateFrameObjects::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << _nInputLists << " object lists processed, " 
			       << _nObjectsMatched << " reference stars matched" << std::endl;

}


} // namespace luiza
