#include "MatchSkyCatalog.h"


#include "luiza/LuizaExceptions.h"
// #include <iostream>
// #include <sstream>
#include <fstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaSkyCatalog.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

MatchSkyCatalog aMatchSkyCatalog;

MatchSkyCatalog::MatchSkyCatalog() :
		Processor("MatchSkyCatalog")
{

	_description = "Processor to match objects with sky catalog.";

	registerInputCollection(TableCollection, "InputCollectionName",
			"Name of the collection containing input object lists",
			_inputCollectionName, std::string(""));

	registerInputCollection(TableCollection, "SkyCatalogName",
			"Name of the collection containing star catalog",
			_skyCatalogName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where output table will be stored",
			_outputCollectionName, std::string(""));

	registerProcessorParameter(
			"MatchingTolerance",
			"Position matching tolerance for object and catalog stars",
			_matchingTolerance, (double)0.02);

	std::vector<std::string > initColumns;

	initColumns.push_back(std::string("RA"));
	initColumns.push_back(std::string("DEC"));

	registerOptionalParameter(
			"PositionColumns",
			"Names of the position columns in the table",
			_positionColumns, initColumns);


	std::vector<std::string > emptyStringVec;

	registerOptionalParameter(
			"CopyColumns",
			"Names of the catalog columns to be copied to the object table",
			_copyColumns, emptyStringVec);


	registerOptionalParameter(
			"NewColumnNames",
			"Names which should be used for columns copied from the catalog",
			_newColumnNames, emptyStringVec);

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);


}


void MatchSkyCatalog::init()
{

	streamlog_out(MESSAGE) << "MatchSkyCatalog::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nLoops = 0;
	_nInputLists = 0;

	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}

	if (_skyCatalogName.size() == 0)
	{
		streamlog_out(ERROR) << "Sky catalog collection name not given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}

	if (_outputCollectionName.size() == 0)
	{
		streamlog_out(WARNING) << "No output collection name given! " << std::endl;
		streamlog_out(WARNING) << "Input object collection will be modified! " << std::endl;
	}


	if (_copyColumns.size() != 0  && _newColumnNames.size() !=0 &&  
            _copyColumns.size() > _newColumnNames.size() )
	{
		streamlog_out(ERROR) << "New column names not properly specified!" << std::endl;
		streamlog_out(ERROR) << "Data can not be copied from the catalog!" << std::endl;
		streamlog_out(ERROR) << "Nothing to do!" << std::endl;

		throw StopProcessingException( this ) ;
	}

}




void MatchSkyCatalog::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "MatchSkyCatalog::startAnalysis()" << endl;
}



void MatchSkyCatalog::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->TableCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}


	// Check if input collection is not empty

	int nTab = data->GetTableCollection(_inputCollectionName).size();

	if (nTab == 0)
	{
		streamlog_out(DEBUG) << "No object list in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nTab << " tables " << endl;

	// Check if catalog collection exists

	if (!data->TableCollectionExists(_skyCatalogName))
	{
		streamlog_out(WARNING) << "Sky catalog collection not found ! " << endl;

		return;
	}


	// Check if catalog collection is not empty

	int nCat = data->GetTableCollection(_skyCatalogName).size();
 
	if (nCat == 0)
	{
		streamlog_out(WARNING) << "No catalog table in catalog collection ! " << endl;

		return;
	}

	if (nCat > 1 && nCat != nTab)
	{
	streamlog_out(WARNING) << "Multiple catalog tables found in catalog collection ! " << endl;
	streamlog_out(WARNING) << "Number of catalogs differ from the number of frames " << endl;
	streamlog_out(WARNING) << "First catalog table will be used ! " << endl;
	}

	GloriaSkyCatalog *catalog = (GloriaSkyCatalog *) &(data -> GetTableCollection(_skyCatalogName).at(0));  

	// Sort catalog in DEC, if not yet done - should speed up processing
	// (only for single permanent catalog)

	if(data->TableCollectionPermanent(_skyCatalogName) && nCat==1 &&
                                   !catalog->IsSortedBy("DEC")) catalog->SortTable("DEC");
 

	// Create output collection if not found (and name is specified)

	if (_outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
          data->CreateTableCollection(_outputCollectionName,_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;
	  }

	// Loop over object lists found in input collection

        for(int itb=0; itb<nTab; itb++)
	  {
	  GloriaFitsTable *source =  &(data -> GetTableCollection(_inputCollectionName).at(itb));  

          if(! source->ColumnExists(_positionColumns[0]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[0] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }

          if(! source->ColumnExists(_positionColumns[1]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[1] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }


          _nInputLists++;


	  // Create output table as a copy of the input table (if name is specified)
	  // Store it in the output collection and retrive pointer for future use.
	  // Otherwise use input table pointer for output.

	  GloriaFitsTable *result;

	  if (_outputCollectionName.size() > 0)
	    {
	    result = new GloriaFitsTable(*source);

	    (data->GetTableCollection(_outputCollectionName)).push_back(*result);

	    int nOut = (data->GetTableCollection(_outputCollectionName)).size();

	    delete result;
            result = &(data->GetTableCollection(_outputCollectionName).at(nOut-1));
	    }
	  else
	    result = source;

	  // If multiple catalogs found (eg. for frame matching)

 	  if (nCat == nTab  && itb>0)
	    catalog = (GloriaSkyCatalog *) &(data -> GetTableCollection(_skyCatalogName).at(itb));

	  // Whole matching is done in one call

          vector<long> matchIDs =  catalog->MatchObjectTable(result,
                       _matchingTolerance,_positionColumns,_copyColumns,_newColumnNames);

	  int nMatch = 0;
	  for(int iobj=0; iobj<matchIDs.size(); iobj++)
	              if(matchIDs[iobj]>=0)nMatch++;

	  streamlog_out(MESSAGE) << nMatch << " objects out of " << matchIDs.size() 
                                 << " matched to the catalog with " << catalog->GetObjectNumber() << " entries" << endl;


	  }  // end of loop over input object lists 



  // always return true  for ProcessorName
  setReturnValue(true);

}

void MatchSkyCatalog::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void MatchSkyCatalog::concludeAnalysis( GloriaDataContainer* data )
{

	streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;
}


void MatchSkyCatalog::end()
{

	printEndMessage();

}

void MatchSkyCatalog::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() reached after " << _nLoops
                         << " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << " Total of " << _nInputLists 
                         << " object lists read" << std::endl;


}


} // namespace luiza
