#include "ObjectLightCurve.h"


#include "luiza/LuizaExceptions.h"
// #include <iostream>
// #include <sstream>
#include <fstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"
#include "gloria/GloriaSkyCatalog.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

ObjectLightCurve aObjectLightCurve;

ObjectLightCurve::ObjectLightCurve() :
		Processor("ObjectLightCurve")
{

	_description = "Processor to reconstruct object light curve.";

	registerInputCollection(TableCollection, "InputCollectionName",
			"Name of the collection containing input object lists",
			_inputCollectionName, std::string(""));

	registerInputCollection(TableCollection, "ReferenceStarCatalog",
			"Name of the collection containing star catalog",
			_referenceStarCatalog, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where light curve will be stored",
			_outputCollectionName, std::string(""));

  std::vector<float > initPosition;

	registerProcessorParameter(
			"ObjectPosition",
			"Position of object of interest (in degrees or pixels)",
			_objectPosition, initPosition);

	registerProcessorParameter(
			"MatchingTolerance",
			"Position matching tolerance for object and reference stars",
			_matchingTolerance, (double)0.02);

	std::vector<std::string > initColumns;

	initColumns.push_back(std::string("RA"));
	initColumns.push_back(std::string("DEC"));

	registerOptionalParameter(
			"PositionColumns",
			"Names of the position columns in the table",
			_positionColumns, initColumns);


	registerOptionalParameter(
			"BrightnessColumn",
			"Name of the object brightness column",
			_brightnessColumn, std::string("Signal"));


	registerOptionalParameter(
			"MagnitudoGiven",
			"Flag to indicate that brightness is already given in magnitudo units (logarythmic scale)",
		_magnitudoGiven, false);


	registerOptionalParameter(
			"PolarAngleColumn",
			"Name of the column corresponding to position in DEC",
			_polarAngleColumn, std::string("DEC"));


        std::vector<float > initRefStars;

	registerProcessorParameter(
			"ReferenceStars",
			"Positions and brightnesses of reference stars (3 numers per star)",
			_referenceStars, initRefStars);


	std::vector<std::string > emptyStringVec;

	registerOptionalParameter(
			"ReferenceMagColumn",
			"Names of the column in the reference star catalog containing star magnitudo",
			_referenceMagColumn, std::string("Mag"));


	registerOptionalParameter(
			"ReferenceMaxDist",
			"Maximum allowed distance to reference star, when reference stars are read from file",
			_referenceMaxDist, double(0.));


	registerOptionalParameter(
			"MaxReferenceCount",
			"Maximum number of reference stars read from file, closest stars are selected",
			_maxReferenceCount, int(0));


	registerOptionalParameter(
			"CorrectionFitOrder",
			"Order of the polynomial fit of correction dependence on position",
			_correctionFitOrder, int(0));


        registerOptionalParameter ("ReferenceUseMedian",
                              "Flag for using median instead of mean for calibration to reference stars",
                              _referenceUseMedian,  false );

	registerOptionalParameter(
			"EstimateError",
			"Flag to do photometry error estimate, based on reference star spread",
		        _estimateError, false);

	registerOptionalParameter(
			"CopyColumns",
			"Names of the table columns to be copied to the light curve table",
			_copyColumns, emptyStringVec);


	registerOptionalParameter(
			"HeaderRecords",
			"Keywords of header records to be copied to the light curve table",
			_headerRecords, emptyStringVec);



}


void ObjectLightCurve::init()
{

	streamlog_out(MESSAGE) << "ObjectLightCurve::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nLoops = 0;
	_nInputLists = 0;
	_nObjectsMatched = 0;
	_nOutputPoints = 0;

	_lightCurveTable=NULL;

	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}

	if (_outputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No output collection name given! "
				<< std::endl;
		streamlog_out(ERROR) << "Object light curve will not be stored! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}


}



void ObjectLightCurve::initializeRefStars( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "ObjectLightCurve::initializeRefStars()" << endl;


 // Conversion factor deg->rad

 double deg2rad = 3.1415926/180.;
    
 // Correction for spherical geometry (angular positions in degrees!)

 double xcor=1.;
 double ycor=1.;
 if(_polarAngleColumn == _positionColumns[0])ycor = cos(deg2rad*_objectPosition[0]);
 if(_polarAngleColumn == _positionColumns[1])xcor = cos(deg2rad*_objectPosition[1]);



  if(  _referenceStarCatalog.size() == 0 &&  _referenceStars.size() < 3 )
    {
		streamlog_out(ERROR) << "Reference stars not specified! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
    }


  if(  _referenceStarCatalog.size() > 0 &&  _referenceStars.size() >= 3 )
    {
		streamlog_out(ERROR) << "Multiple reference star definition! " << std::endl;
		streamlog_out(ERROR) << "Reference star sets will be combined !!! " << std::endl;
    }



  if(  _referenceStarCatalog.size() > 0)
    {

    GloriaSkyCatalog *stars =  NULL;

    if(data->TableCollectionExists(_referenceStarCatalog) 
       && data->GetTableCollection(_referenceStarCatalog).size() > 0 )
          stars =(GloriaSkyCatalog *) &(data -> GetTableCollection(_referenceStarCatalog).at(0));  
    else
      {
	streamlog_out(ERROR) << "Reference stars collection not found! " << std::endl;
	streamlog_out(ERROR) << "Can not proceed! " << std::endl;

	throw StopProcessingException( this ) ;
      }

    if(! stars->CheckCatalog() )
	    {
	      streamlog_out(ERROR) << "  Reference table does not contain valid star catalog - stop processing " << endl;

	      throw StopProcessingException( this ) ;
             }

    if(! stars->ColumnExists(_referenceMagColumn))
	    {
	      streamlog_out(ERROR) << " Magnitudo column " << _referenceMagColumn
                                   << " not found in reference star catalog - stop processing " << endl;

	      throw StopProcessingException( this ) ;
             }


      // Get table with selected reference stars

      int nStars = stars->GetRowNumber();

      streamlog_out(DEBUG) << " Searching reference star list with "<< nStars << " stars " << std::endl;


      vector<long> match = stars->MatchAllObjects(_objectPosition[0],_objectPosition[1],_referenceMaxDist,_matchingTolerance);

      int nSel = match.size();
      streamlog_out(DEBUG) << nSel <<  " reference stars found within the matching distance "<< endl;


      vector<double> *ravec = &(stars->GetVectorOf<double>(0));
      vector<double> *decvec = &(stars->GetVectorOf<double>(1));
      vector<double> *mvec = &(stars->GetVectorOf<double>(_referenceMagColumn));

      // If the number of reference stars is limitted - take the nearest ones

      if(nSel > _maxReferenceCount && _maxReferenceCount > 0 ) nSel = _maxReferenceCount;

       std::multimap<double, int>::iterator refDistMapInd;

       for(int isel=0; isel < nSel ; isel++)
	 {
	   int iref = match.at(isel);

	   _referenceStars.push_back(ravec->at(iref));  
	   _referenceStars.push_back(decvec->at(iref));  
	   _referenceStars.push_back(mvec->at(iref));  

	 }
    
       // End of     if( _referenceStarCatalog.size() > 0 )

    }



      int nRef = _referenceStars.size()/3;

      streamlog_out(MESSAGE) << "Light Curve will be reconstructed for object at  ("
                             << _positionColumns[0] << ", " << _positionColumns[1] << ")  =  ("
                             <<  _objectPosition[0] << ", " << _objectPosition[1] << ")" << endl;

      if(nRef==0)
	  {
	  streamlog_out(ERROR) << "No reference stars found in input list - nothing to do! " << std::endl;
	  throw StopProcessingException( this ) ;
	  }

      streamlog_out(MESSAGE) << nRef <<  " reference stars selected for analysis: "<< endl;


      for(int iref=0;iref<nRef;iref++)
	    streamlog_out(DEBUG) << "Magnitudo " << _referenceStars[3*iref+2] << "  at  (" 
                                 << _referenceStars[3*iref] << ", " << _referenceStars[3*iref+1] << ")" << endl;

      // Remember coordinate range

      xmin=xmax=_objectPosition[0];
      ymin=ymax=_objectPosition[1];

      for(int iref=0;iref<nRef;iref++)
	{
           if(_referenceStars[3*iref]<xmin)xmin=_referenceStars[3*iref];
           if(_referenceStars[3*iref]>xmax)xmax=_referenceStars[3*iref];

           if(_referenceStars[3*iref+1]<ymin)ymin=_referenceStars[3*iref+1];
           if(_referenceStars[3*iref+1]>ymax)ymax=_referenceStars[3*iref+1];
	}

      xmin-=_matchingTolerance/xcor;
      xmax+=_matchingTolerance/xcor;

      ymin-=_matchingTolerance/ycor;
      ymax+=_matchingTolerance/ycor;


      streamlog_out(DEBUG) << "Window for star searching: " 
                           << _positionColumns[0] << " = " << xmin << " - " << xmax << ",   "
                           << _positionColumns[1] << " = " << ymin << " - " << ymax << endl;



        int npolpar = (_correctionFitOrder+1)*(_correctionFitOrder+2)/2;

        if(nRef < 2*npolpar )
	  {
         streamlog_out(WARNING) << " Number of selected stars too small for required correction fit order: "<< endl;
         streamlog_out(WARNING) << " Conctant correction (no position dependence fit) will be calculated! "<< endl;
	 _correctionFitOrder=0;
	  }

      return;

}


void ObjectLightCurve::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "ObjectLightCurve::startAnalysis()" << endl;
}



void ObjectLightCurve::initializeLCTable( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "ObjectLightCurve::initializeLCTable()" << endl;

   // Create table to store reconstructed light curve

   _lightCurveTable = new GloriaFitsTable;
                              
   // Define columns - header fields go first (eg. HJD)
   // Header fields are always stored as strings (as in FITS)
                              
   for(int irec=0; irec< _headerRecords.size(); irec++)
	      {
	      string keyword = _headerRecords.at(irec);
	      //  string value = source->GetKeywordValue(keyword.c_str());

              string value = "empty";
              _lightCurveTable->AddColumn(keyword,value);
	      }


   // Define columns - selected object table columns go next
   // Check type of column in the input table, if available                     
         
   GloriaFitsTable *source =  NULL;

   if(data->TableCollectionExists(_inputCollectionName) 
       && data->GetTableCollection(_inputCollectionName).size() > 0 )
            source = &(data -> GetTableCollection(_inputCollectionName).at(0));  

   for(int icol=0; icol< _copyColumns.size(); icol++)
	      {
	      string colname = _copyColumns.at(icol);

	      // STORE VALUES AS STRINGS, independent on column type (!)
	      //              string value = "empty";
              // _lightCurveTable->AddColumn(colname,value);


              TableColumnType coltype = StringColumn; 
              if(source) coltype = source->GetColumnType(colname);

	      _lightCurveTable->AddColumn(colname,coltype);
	      }

   // Define column to store lightcurve
 
   _lightCurveTable->AddColumn("Mag",double(-1.0));

   if(_estimateError)
        _lightCurveTable->AddColumn("MagError",double(-1.0));


   // Add the created column to the data collection if name given

   if (_outputCollectionName.size() > 0)
     {
     if( !data->TableCollectionExists(_outputCollectionName))
          {
	   data->CreateTableCollection(_outputCollectionName,true);

           streamlog_out(DEBUG) << "  Permanent output collection " << _outputCollectionName << " created " << endl;
          }

     int ilc = (data->GetTableCollection(_outputCollectionName)).size();

     (data->GetTableCollection(_outputCollectionName)).push_back(*_lightCurveTable);

     // Delete local instance of table 

     delete _lightCurveTable;

     // Store pointer to the table in the GLORIA data container

      _lightCurveTable = &(data -> GetTableCollection(_outputCollectionName).at(ilc));
      }

}


void ObjectLightCurve::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	// Initialize reference star table (if not given directly by user)

	if(isFirstEvent() )initializeRefStars(data);

	// Initialize output table

	if(_lightCurveTable==NULL)initializeLCTable(data);

	_nLoops++;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->TableCollectionExists(_inputCollectionName))
	{
		streamlog_out(MESSAGE) << "Input collection not found ! " << endl;

		return;
	}


	// 
	int nTab = data->GetTableCollection(_inputCollectionName).size();

	if (nTab == 0)
	{
		streamlog_out(DEBUG) << "No object list in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nTab << " tables " << endl;


	// Loop over object lists found in input collection

        for(int itb=0; itb<nTab; itb++)
	  {
	  GloriaFitsTable *source =  &(data -> GetTableCollection(_inputCollectionName).at(itb));  

          if(! source->ColumnExists(_positionColumns[0]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[0] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }


          if(! source->ColumnExists(_positionColumns[1]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[1] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }

          if(! source->ColumnExists(_brightnessColumn))
	    {
	      streamlog_out(ERROR) << "  Brightness column " << _brightnessColumn
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }


          _nInputLists++;



     // Object list analysis starts here
     // ================================

      // Conversion factor deg->rad

      double deg2rad = 3.1415926/180.;
    
      // Correction for spherical geometry (angular positions in degrees!)

      double xcor=1.;
      double ycor=1.;
      if(_polarAngleColumn == _positionColumns[0])ycor = cos(deg2rad*_objectPosition[0]);
      if(_polarAngleColumn == _positionColumns[1])xcor = cos(deg2rad*_objectPosition[1]);

      double rmax= _matchingTolerance * _matchingTolerance;
      double rmatch= rmax;
      double smatch=0.;
      int imatch = -1;

      int nRef = _referenceStars.size()/3;
      int *iref = new int[nRef];
      double *rref = new double[nRef];
      double *sref = new double[nRef];
      double *xref = new double[nRef];
      double *yref = new double[nRef];

      for(int ir=0;ir<nRef;ir++)
	{
	  iref[ir]=-1;
	  rref[ir]=rmax;
	  sref[ir]=0.;
	  xref[ir]=0.;
	  yref[ir]=0.;
	}

      int nObject = source->GetRowNumber();

      vector<double> *xvec = &(source->GetVectorOf<double>(_positionColumns[0]));
      vector<double> *yvec = &(source->GetVectorOf<double>(_positionColumns[1]));
      vector<double> *svec = &(source->GetVectorOf<double>(_brightnessColumn));

      streamlog_out(DEBUG) << " Searching object list " << itb << " with "<< nObject << " objects " << std::endl;


      for(int iobj=0; iobj<nObject; iobj++)
	{
	  //   double xpos= source-> GetValueOf<double>(_positionColumns[0],iobj);
	  //   double ypos= source-> GetValueOf<double>(_positionColumns[1],iobj);
	  //   double sig= source-> GetValueOf<double>(_brightnessColumn,iobj);

	  double xpos= xvec->at(iobj);
	  double ypos= yvec->at(iobj);
	  double sig= svec->at(iobj);

	  // Skip if outside window

          if(xpos<xmin || xpos>xmax || ypos<ymin || ypos>ymax)continue;

	  // Skip negative or zero signal values (unless magnitudo is given)

	 if(! _magnitudoGiven && sig <= 0.)  
	   {
  	     streamlog_out(DEBUG) << " Skiping object "<< iobj
				  << ", signal not positive:  " << sig << std::endl;

	     continue;
	   }


	 if(! _magnitudoGiven) sig = 18. - 2.5 * log10(sig);

	 // Try to match with object position

	 double dx=(xpos - _objectPosition[0])*xcor;
	 double dy=(ypos - _objectPosition[1])*ycor;

         if(dx*dx+dy*dy < rmatch)
	   {
	     imatch=iobj;
             rmatch = dx*dx+dy*dy;
             smatch = sig;
	   }

	 // Try to match with reference stars
	 int irefvec=0;
         for(int ir=0;ir<nRef;ir++)
	   {
	   double dxr=(xpos - _referenceStars[irefvec++])*xcor;
	   double dyr=(ypos - _referenceStars[irefvec++])*ycor;
           double cat = _referenceStars[irefvec++];

           if(dxr*dxr+dyr*dyr < rref[ir])
	     {
	       iref[ir]=iobj;
	       rref[ir]=dxr*dxr+dyr*dyr;
	       sref[ir]=sig - cat;
	       xref[ir]=dx;
	       yref[ir]=dy;
	     }
	   }

	}


      if(imatch < 0 )
       {
       streamlog_out(MESSAGE) << " No matching object found in the input list!" << std::endl;

       delete [] iref;
       delete [] rref;
       delete [] sref;
       delete [] xref;
       delete [] yref;
    
       continue;
       }

      // Check matched reference stars

      double refsum=0.;
      double refrms=0.;
      int refmatch = 0;
      std::vector<double> refvec;

      // Surface fit table definitions

     int npolpar = (_correctionFitOrder+1)*(_correctionFitOrder+2)/2;
     int nfittab = npolpar*npolpar;
     double *mag2sum3;
     double *magsum3;
     double *magval3;
     double mag2sum=0.;

     if(_correctionFitOrder>0)
       {
        mag2sum3 = new double[nfittab];
        magsum3 = new double[npolpar];
        magval3 = new double[npolpar];

        for(int i1=0;i1<npolpar;i1++)
          {
          magsum3[i1]=0.;
          for(int i2=0;i2<npolpar;i2++)
    	     mag2sum3[i1+npolpar*i2]=0.;
          }
       }


      for(int ir=0;ir<nRef;ir++)
	{
	  if(iref[ir]>=0)
	    {
	      streamlog_out(DEBUG) << " Reference star " << ir << " in row " << iref[ir] << " at dr = " << sqrt(rref[ir]) 
                                   << " mag - cat = " << sref[ir] << std::endl;
	      refsum+=sref[ir];
	      refrms+=sref[ir]*sref[ir];
              refvec.push_back(sref[ir]);

	   if(_correctionFitOrder>0)
	     {
              int ival=0;
              for(int i1=0;i1<=_correctionFitOrder;i1++)
       	        for(int i2=0;i2<=_correctionFitOrder-i1;i2++)
	          magval3[ival++]= pow( xref[ir],i1) * pow(yref[ir],i2);

              for(int i1=0;i1<npolpar;i1++)
                {
       	        magsum3[i1]+=sref[ir]*magval3[i1];
       	        for(int i2=0;i2<npolpar;i2++)
       	              mag2sum3[i1+npolpar*i2]+=magval3[i1]*magval3[i2];
       	        }

	      mag2sum+=sref[ir]*sref[ir];
	     }
       
	    refmatch++;
	    }
	}

       streamlog_out(DEBUG) << " Object found in row " << imatch << " with magnitudo " << smatch << std::endl;


       // Error unknown at the moment

       double serror = -1.;

       // Normalize to reference stars if given
       // =====================================

	if(refmatch>0)
	   {

	   // Simplest case: mean

           double refcorr = refsum/refmatch;
	   if(_estimateError && refmatch>1)
	     {
	       refrms/=refmatch;
               refrms-=refcorr*refcorr;
	       if(refrms>0.)serror = sqrt(refrms);
	     }


	   // Possible solution: median

           if(_referenceUseMedian)
	     {
	     std::sort(refvec.begin(),refvec.end());
             int imed=refmatch/2;

             if(refmatch%2 == 1)     
               refcorr=refvec[imed];
             else
               refcorr=0.5*(refvec[imed]+refvec[imed-1]);
	     }

	   // Pi way: correction surface

	   if(_correctionFitOrder>0 && refmatch>=2*npolpar)
	     {
             double *mag2inv = new double[nfittab];
             double *magpar = new double[npolpar];

             for(int i1=0;i1<npolpar;i1++)
                 magpar[i1] = magsum3[i1];

             for(int i2=0;i2<nfittab;i2++)
    	          mag2inv[i2]=mag2sum3[i2];

             GaussjSolve(mag2inv,magpar,npolpar);

	     refcorr=magpar[0];

	     if(_estimateError)
	       {
	       serror = mag2sum;

 	       for(int i1=0;i1<npolpar;i1++)
		 {
		 serror -= 2*magpar[i1]*magsum3[i1];

 	         for(int i2=0;i2<npolpar;i2++)
		    serror += magpar[i1]*magpar[i2]*mag2sum3[i1+npolpar*i2];
		 }

               if(serror>0)serror = sqrt(serror/refmatch);
	       }

             delete [] mag2inv;
             delete [] magpar;
	     }


           smatch -=  refcorr;

   	   streamlog_out(MESSAGE) << "     Magnitudo after normalizing to " << refmatch 
                                  << " reference stars is " << smatch  << std::endl;

	   if(_estimateError && serror>0.)
   	           streamlog_out(MESSAGE) << "     Estimated error is " << serror << std::endl;
	   }


	 delete [] iref;
	 delete [] rref;
	 delete [] sref;
	 delete [] xref;
	 delete [] yref;
    
         if(_correctionFitOrder>0)
           {
	     delete [] mag2sum3;
	     delete [] magsum3;
	     delete [] magval3;
	   }


         if(nRef>0 && refmatch==0)
           {
           streamlog_out(MESSAGE) << " No reference stars matched from the input list!" << std::endl;
           continue;
           }


         _lightCurveTable->AddRow();

         int irow = _lightCurveTable->GetRowNumber()-1;

	 _lightCurveTable->SetLast("Mag",smatch);

	 if(_estimateError && serror>0.)
	          _lightCurveTable->SetLast("MagError",serror);


	   // Add header fields
                              
            for(int irec=0; irec< _headerRecords.size(); irec++)
	      {
	      string keyword = _headerRecords.at(irec);
	      string value = source->GetKeywordValue(keyword.c_str());

              _lightCurveTable->SetLast(keyword,value);
	      }


           // Copy table columns 
                               
           for(int icol=0; icol< _copyColumns.size(); icol++)
	      {
	      string colname = _copyColumns.at(icol);

              if(_lightCurveTable->GetColumnType(colname) == StringColumn)
                 _lightCurveTable->SetLast(colname,(source->GetColumn(colname))->StringVal(imatch));
	      else
		(_lightCurveTable->GetColumn(colname))->CopyCellValue(source->GetColumn(colname),imatch,irow);
	      }

	 _lightCurveTable->AdjustColumns();


     }
     // end of loop over input object lists 



  // always return true  for ProcessorName
  setReturnValue(true);

}

void ObjectLightCurve::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void ObjectLightCurve::concludeAnalysis( GloriaDataContainer* data )
{

	streamlog_out(MESSAGE) << " concludeAnalysis() " << std::endl;

	// Print light curve table if not stored

   if (_outputCollectionName.size() == 0)
     {
      stringstream tabss;

      _lightCurveTable->Print(tabss,20);

      streamlog_out(MESSAGE) << " Reconstructed light curve table " << endl;
      streamlog_out(MESSAGE) << " For object at " 
                     << _positionColumns[0] << " = " << _objectPosition[0] 
	             << _positionColumns[1] << " = " << _objectPosition[1] <<  endl;

      streamlog_out(MESSAGE) <<  tabss.str() << endl;

     }

   streamlog_out(MESSAGE) << " Light curve reconstructed with " <<  _lightCurveTable->GetRowNumber() << " points" << std::endl;

}


void ObjectLightCurve::end()
{

	printEndMessage();

}

void ObjectLightCurve::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

}

//
// Routine for solving matrix equations
//

int ObjectLightCurve::GaussjSolve(double *alfa,double *beta,int n)
{
  int *ipiv;
  int *indxr;
  int *indxc;
  int i,j,k;
  int irow=0;
  int icol=0;
  double abs,big,help,pivinv;

  ipiv = new int[n];
  indxr = new int[n];
  indxc = new int[n];

  for(i=0;i<n;ipiv[i++]=0);

  for(i=0;i<n;i++)
    {
      big=0.;
      for(j=0;j<n;j++)
	{
	  if(ipiv[j]==1)continue;
	  for(k=0;k<n;k++)
	    {
	      if(ipiv[k]!=0)continue;
	      abs=fabs(alfa[n*j+k]);
	      if(abs>big)
		{
		  big=abs;
		  irow=j;
		  icol=k;
		}
	    }
	}
      ipiv[icol]++;

      if(ipiv[icol]>1)
	return 1;

      if(irow!=icol)
	{
	  help=beta[irow];
	  beta[irow]=beta[icol];
	  beta[icol]=help;
	  for(j=0;j<n;j++)
	    {
	      help=alfa[n*irow+j];
	      alfa[n*irow+j]=alfa[n*icol+j];
	      alfa[n*icol+j]=help;
	    }
	}
      indxr[i]=irow;
      indxc[i]=icol;

      if(alfa[n*icol+icol]==0.)
	return 1;

      help=alfa[n*icol+icol];
      pivinv=1./help;
      alfa[n*icol+icol]=1.;
      for(j=0;j<n;alfa[n*icol+(j++)]*=pivinv);
      beta[icol]*=pivinv;

      for(j=0;j<n;j++)
	{
	  if(j==icol)continue;
	  help=alfa[n*j+icol];
	  alfa[n*j+icol]=0.;
	  for(k=0;k<n;k++)
	    alfa[n*j+k]-=alfa[n*icol+k]*help;
	  beta[j]-=beta[icol]*help;
	}
    }

  for(i=n-1;i>=0;i--)
    {
      if(indxr[i]==indxc[i])continue;
      for(j=0;j<n;j++)
	{
	  help=alfa[n*j+indxr[i]];
	  alfa[n*j+indxr[i]]=alfa[n*j+indxc[i]];
	  alfa[n*j+indxc[i]]=help;
	}
    }

  delete [] ipiv;
  delete [] indxr;
  delete [] indxc;

  return 0;
}

} // namespace luiza
