#include "SelectNewObjects.h"


#include "luiza/LuizaExceptions.h"
// #include <iostream>
// #include <sstream>
#include <fstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaSkyCatalog.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

SelectNewObjects aSelectNewObjects;

SelectNewObjects::SelectNewObjects() :
		Processor("SelectNewObjects")
{

	_description = "Processor to select table data.";

	registerInputCollection(TableCollection, "InputCollectionName",
			"Name of the collection containing input tables",
			_inputCollectionName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where output table will be stored",
			_outputCollectionName, std::string(""));

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);


	registerProcessorParameter(
			"MatchingTolerance",
			"Position matching tolerance for object and catalog stars",
			_matchingTolerance, (double)0.02);

	registerProcessorParameter(
			"SignalThreshold",
			"Threshold for the new object signal value, given in units of noise",
			_signalThreshold, (double)5.);

	std::vector<std::string > initSigColumns;

	initSigColumns.push_back(std::string("Signal"));
	initSigColumns.push_back(std::string("Noise"));

	registerOptionalParameter(
			"SignalColumns",
			"Names of the brightness and noise columns in the input table",
			_signalColumns, initSigColumns);


	std::vector<std::string > initPosColumns;

	initPosColumns.push_back(std::string("RA"));
	initPosColumns.push_back(std::string("DEC"));

	registerOptionalParameter(
			"PositionColumns",
			"Names of the position columns in the table",
			_positionColumns, initPosColumns);


	std::vector<int > initSize;

	registerOptionalParameter(
			"ImageSize",
			"Size of the image corresponding to the object list",
			_imageSize, initSize);


	registerOptionalParameter(
			"EdgeCut",
			"Cut to remove new objects identified very close to the image edge",
			_edgeCut, (int) 50);



}


void SelectNewObjects::init()
{

	streamlog_out(MESSAGE) << "SelectNewObjects::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nLoops = 0;
	_nInputTables = 0;

        _previousFrame = new GloriaSkyCatalog();
	_astroStored = false;

	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}


	if (_positionColumns.size() != 2)
	{
		streamlog_out(ERROR) << "Wrong value for position column names! " << std::endl;
		streamlog_out(ERROR) << "Can not proceed! " << std::endl;

		throw StopProcessingException( this ) ;
	}


	if (_outputCollectionName.size() == 0)
	{
		streamlog_out(WARNING) << "No output collection name given! " << std::endl;
		streamlog_out(WARNING) << "Input object collection will be modified! " << std::endl;
	}


	if (_signalColumns.size() != 2)
	  {
		streamlog_out(WARNING) << "Wrong values for signal and noise column names! " << std::endl;
		streamlog_out(WARNING) << "Signal cut will not be used! " << std::endl;
	  }
	else
	  {
	    _previousFrame->AddColumn(_signalColumns[0], double(0.));
	    _previousFrame->AddColumn(_signalColumns[1], double(0.));
	  }

        _previousFrame->AddColumn("NewObjectFlag",(long) 0);
	_newFlagVec = &(_previousFrame->GetVectorOf<long>("NewObjectFlag"));
}



void SelectNewObjects::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "SelectNewObjects::startAnalysis()" << endl;
}



void SelectNewObjects::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->TableCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}


	// Check if input collection is not empty

	int nTab = data->GetTableCollection(_inputCollectionName).size();

	if (nTab == 0)
	{
		streamlog_out(DEBUG) << "No object list in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nTab << " tables " << endl;


	// Create output collection if not found (and name is specified)

	if (_outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
          data->CreateTableCollection(_outputCollectionName,_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;
	  }


	//=====================================================
	//
	//      Loop over tables in input collection
	//


	// For first table processed, use second frame for comparison (if available in this loop)

	int ifirst=0;

	if(_previousFrame->GetObjectNumber() == 0  && nTab>1) ifirst = -1;

        for(int itb0=ifirst; itb0<nTab; itb0++)
	  {
          int itb=abs(itb0);

	  GloriaFitsTable *source =  &(data -> GetTableCollection(_inputCollectionName).at(itb));  

          if(! source->ColumnExists(_positionColumns[0]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[0] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }


          if(! source->ColumnExists(_positionColumns[1]))
	    {
	      streamlog_out(ERROR) << "  Position column " << _positionColumns[1] 
                                   << " not found in input object list - skip " << endl;
	      continue;
	    }


          _nInputTables++;


	  source->AdjustColumns();
	  int nrow = source->GetRowNumber();

	  // Position columns need to be there (checked above)

          vector<double>* RAvec = &(source->GetVectorOf<double>(_positionColumns[0]));
          vector<double>* DECvec = &(source->GetVectorOf<double>(_positionColumns[1]));

	  // Check if astrometry is stored in the table header

          GloriaAstrometry newAstro; 
          bool isAstro = newAstro.IsAstrometryPresent(source);

          if(isAstro) newAstro.InitFromHeader(source);


          int sizeX=0;
          int sizeY=0;

          if(isAstro)
	    {
	      sizeX  = 2*newAstro.crpix[0];
	      sizeY  = 2*newAstro.crpix[1];
	    }

	  if(_imageSize.size() == 2)
	    {
	      sizeX = _imageSize.at(0);
	      sizeY = _imageSize.at(1);
	    }


	  bool positionCut=(sizeX>0 && sizeY>0);

	  vector<double>* Xvec;
          vector<double>* Yvec;

	  if(source->ColumnExists("CCD_X"))
            Xvec = &(source->GetVectorOf<double>("CCD_X"));
	  else
	    positionCut=false;

	  if(source->ColumnExists("CCD_Y"))
            Yvec = &(source->GetVectorOf<double>("CCD_Y"));
	  else
	    positionCut=false;


	  vector<double>* Svec;
          vector<double>* Nvec;
	  bool signalCut=true;

	  if(_signalColumns.size()==2 && source->ColumnExists(_signalColumns[0]))
	    Svec = &(source->GetVectorOf<double>(_signalColumns[0]));
	  else
	    signalCut=false;

	  if(_signalColumns.size()==2 && source->ColumnExists(_signalColumns[1]))
	    Nvec = &(source->GetVectorOf<double>(_signalColumns[1]));
	  else
	    signalCut=false;


	  // Empty output structure

          GloriaFitsTable *result = source->EmptyClone();


          // Vector of new object row IDs

	  vector<int> newObjects;

          // Vector of new objects below threshold - to be marked
 
          vector<int> newCandidates;

	  // If old frame is stored in the catalog structure - search for new objects

	  if(_previousFrame->GetObjectNumber() > 0 )
	    {

	      streamlog_out(DEBUG) << "Searching for new objects in table " << itb 
				   << "   Position cut used: " << positionCut 
				   << "   Signal cut used: " << signalCut 
				   << "   Astrometry used: " << _astroStored << endl;

  	    // Matching with old frame done in one call

            vector<long> oldIDs = 
                  _previousFrame->MatchObjectTable(source,_matchingTolerance,_positionColumns);


	    // Verify new objects

            for(int irow=0; irow<nrow ; irow++)
	       {
	       //  Object mached on previous frame, and not a candidate

	       int oldID = oldIDs[irow];

	       if ( oldID >= 0 && _newFlagVec->at(oldID) != 1 )continue;

	       // Object too close to image edge

	       if(positionCut)
		  {
		    if(Xvec->at(irow) < _edgeCut) continue;
		    if(Yvec->at(irow) < _edgeCut) continue;
		    if(Xvec->at(irow) > sizeX-_edgeCut) continue;
		    if(Yvec->at(irow) > sizeY-_edgeCut) continue;

		   // Object too close to the edge on previous image

		   if(_astroStored && _oldSizeX * _oldSizeY>0)
		     {
                     double iox,ioy;
		     _previousAstro.GetXYFromRaDec(RAvec->at(irow), DECvec->at(irow), &iox, &ioy);


 	             if(iox < _edgeCut) continue;
		     if(ioy < _edgeCut) continue;
		     if(iox > _oldSizeX-_edgeCut) continue;
		     if(ioy > _oldSizeY-_edgeCut) continue;
		     }
		  }

	       // If object is too dim - mark it as a candidate

	       if(signalCut && Svec->at(irow) < _signalThreshold * Nvec->at(irow))
		    newCandidates.push_back(irow);
	       else
		    newObjects.push_back(irow);

	       }

 	     streamlog_out(DEBUG) << "New objects after cuts: " << newObjects.size()
				 << "   Candidate (too dim) objects: " << newCandidates.size() << endl;


	     // Copy new object to the output table

	     for(int inew=0; inew<newObjects.size() ; inew++)
                       result->CopyTableRow(source, newObjects[inew], -1);


	     }   //    if(_previousFrame->GetObjectNumber() > 0 )



          // Fill catalog structure with current frame
	  // Copy whole columns

	  _previousFrame->EmptyTable();

	  _previousFrame->InsertColumn("RA",(source->GetColumn(_positionColumns[0]))->Clone());
	  _previousFrame->InsertColumn("DEC",(source->GetColumn(_positionColumns[1]))->Clone());

	  if(signalCut)
	    {
	    _previousFrame->InsertColumn(_signalColumns[0],(source->GetColumn(_signalColumns[0]))->Clone());
	    _previousFrame->InsertColumn(_signalColumns[1],(source->GetColumn(_signalColumns[1]))->Clone());
	    }

	  // Add column for new object flag

          _previousFrame->AddColumn("NewObjectFlag",(long) 0);
          _previousFrame->AdjustColumns();
	  _newFlagVec = &(_previousFrame->GetVectorOf<long>("NewObjectFlag"));


	  for(int icand=0; icand<newCandidates.size() ; icand++)
  	       _newFlagVec->at(newCandidates[icand])=1;

	  for(int inew=0; inew<newObjects.size() ; inew++)
  	       _newFlagVec->at(newObjects[inew])=2;


	  // Store astrometry information

          if(isAstro)
	    {
             _previousAstro.InitFromHeader(source);
             _astroStored=true;
	    }

	  _oldSizeX = sizeX;
	  _oldSizeY = sizeY;


          // Store new object list

	  if(result->GetRowNumber()>0)
	    {
	     if (_outputCollectionName.size() > 0 )
	       {
               (data->GetTableCollection(_outputCollectionName)).push_back(*result);
   	       delete result;

	       int nOut = (data->GetTableCollection(_outputCollectionName)).size();
               result = &(data->GetTableCollection(_outputCollectionName).at(nOut-1));
	       }
	     else
	       {
		data->GetTableCollection(_inputCollectionName).at(itb) = *result;
   	        delete result;

                result = &(data->GetTableCollection(_inputCollectionName).at(itb));
	       }

	    }

	  // Clear old list, if input collection is to be modified and no search result
	  // Do not do it when starting from second frame (itb0=-1)

	  if(itb0>=0 && _outputCollectionName.size()==0 && 
                   result->GetRowNumber()==0 && source->GetRowNumber()>0)
                           source->ClearTable();


	  } // end of loop over input tables

  // always return true  for ProcessorName
  setReturnValue(true);

}

void SelectNewObjects::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

	string checkName = _inputCollectionName;

	if(_outputCollectionName.size() > 0)checkName = _outputCollectionName;


	if (!data->TableCollectionExists(checkName))
	  {
	  streamlog_out(DEBUG) << "Output collection not found: " << checkName << endl;
  	  return;
	  }

	int nTab = data->GetTableCollection(checkName).size();

	streamlog_out(DEBUG) << nTab << " tables in output collection " << std::endl;

        for(int itb=0; itb<nTab; itb++)
	  {
	  GloriaFitsTable *result =  &(data -> GetTableCollection(checkName).at(itb));  
	  streamlog_out(DEBUG) << result->GetRowNumber() << "  ";
	  }

	streamlog_out(DEBUG) << " new objects identified in these tables " << std::endl;
}


void SelectNewObjects::concludeAnalysis( GloriaDataContainer* data )
{

	streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;
}


void SelectNewObjects::end()
{

	printEndMessage();

}

void SelectNewObjects::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() reached after " << _nLoops
                         << " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << " Total of " << _nInputTables 
                         << " tables analysed" << std::endl;


}


} // namespace luiza
