#include "SelectTableData.h"


#include "luiza/LuizaExceptions.h"
// #include <iostream>
// #include <sstream>
#include <fstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaSkyCatalog.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

SelectTableData aSelectTableData;

SelectTableData::SelectTableData() :
		Processor("SelectTableData")
{

	_description = "Processor to select table data.";

	registerInputCollection(TableCollection, "InputCollectionName",
			"Name of the collection containing input tables",
			_inputCollectionName, std::string(""));

	registerOutputCollection(TableCollection, "OutputCollectionName",
			"Name of the collection where output table will be stored",
			_outputCollectionName, std::string(""));

	registerOptionalParameter(
			"PermanentCollection",
			"Flag output collection, as permanent, not deleted after loop is finished",
			_permanentCollection, false);

	registerOptionalParameter(
			"SortColumn",
			"Name of the column used for table sorting",
			_sortColumn, std::string(""));

	registerOptionalParameter(
			"ReverseSort",
			"Flag to sort output table in reverse order.",
			_reverseSort, false);

	std::vector<std::string > emptyStringVec;

	registerOptionalParameter(
			"ColumnValueRange",
			"Names of the columns and allowed ranges of values (3 fields per cut)",
			_columnValueRange, emptyStringVec);

	registerOptionalParameter(
			"AcceptIfAny",
			"Choose selection based on OR of cuts (when true) or AND of cuts (when false).",
			_acceptIfAny, false);


	registerOptionalParameter(
			"OutputColumns",
			"Names of columns which should be written to output table",
			_outputColumns, emptyStringVec);

	registerOptionalParameter(
			"MinCount",
			"Minimum number of selected rows for processor returning true value",
			_minCount, int(1) );

}


void SelectTableData::init()
{

	streamlog_out(MESSAGE) << "SelectTableData::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nLoops = 0;
	_nInputTables = 0;

	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}


	if (_outputCollectionName.size() == 0)
	{
		streamlog_out(WARNING) << "No output collection name given! " << std::endl;
		streamlog_out(WARNING) << "Input object collection will be modified! " << std::endl;
	}


	if (_columnValueRange.size() == 0  && _sortColumn.size() ==0 && _outputColumns.size() == 0)
	{
		streamlog_out(ERROR) << "No cuts, sorting nor output column specified!" << std::endl;
		streamlog_out(ERROR) << "Nothing to do!" << std::endl;

		throw StopProcessingException( this ) ;
	}

	if (_columnValueRange.size()%3 != 0 )
	{
		streamlog_out(ERROR) << "Wrong definition of cuts!" << std::endl;
		streamlog_out(ERROR) << "Nothing to do!" << std::endl;

		throw StopProcessingException( this ) ;
	}

   // No output yet
	
   setReturnValue(false);
}




void SelectTableData::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(MESSAGE) << "SelectTableData::startAnalysis()" << endl;
}



void SelectTableData::processData(GloriaDataContainer* data)
{
  // Maximum number of selected rows in a table
  
   int maxCount = 0;

   if(_minCount > 0)
      setReturnValue(false);
   else
      setReturnValue(true);
     

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->TableCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}


	// Check if input collection is not empty

	int nTab = data->GetTableCollection(_inputCollectionName).size();

	if (nTab == 0)
	{
		streamlog_out(DEBUG) << "No object list in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nTab << " tables " << endl;


	// Create output collection if not found (and name is specified)

	if (_outputCollectionName.size() > 0  &&
	   !data->TableCollectionExists(_outputCollectionName))
	  {
          data->CreateTableCollection(_outputCollectionName,_permanentCollection);

	  streamlog_out(DEBUG) << "  Output collection " << _outputCollectionName << " created " << endl;
	  }


	// Loop over tables in input collection

        for(int itb=0; itb<nTab; itb++)
	  {
	  GloriaFitsTable *source =  &(data -> GetTableCollection(_inputCollectionName).at(itb));  

          _nInputTables++;


          int ncut = _columnValueRange.size()/3;
  	  streamlog_out(DEBUG) <<  ncut << " cuts will be applied to table " << itb << endl;

	   // Pointer to output table (source table by default)

	  GloriaFitsTable *result = source;

	  // Just copy input table if no cuts...

	  if (ncut==0 && _outputCollectionName.size() > 0)
	    {
	    result = new GloriaFitsTable(*source);

	    (data->GetTableCollection(_outputCollectionName)).push_back(*result);

	    int nOut = (data->GetTableCollection(_outputCollectionName)).size();

	    delete result;
            result = &(data->GetTableCollection(_outputCollectionName).at(nOut-1));
	    }


	  // Check table rows and prepare selection

	  if(ncut > 0)
	    {
  	    bool* isdouble = new bool[ncut];

	    int ndcut = 0;
	    int nlcut = 0;

	    vector<double>** dcolvec = new vector<double>* [ncut];
	    vector<long>** lcolvec = new vector<long>* [ncut];

	    double* dmin = new double[ncut];
	    double* dmax = new double[ncut];

	    long* lmin = new long[ncut];
	    long* lmax = new long[ncut];

 
	    for(int icut = 0; icut < ncut ; icut++)
	      {
	      string colname = _columnValueRange.at(3*icut);

              if(source->GetColumnType(colname)!= DoubleColumn &&
                 source->GetColumnType(colname)!= LongColumn) continue;
                    
	      if(source->GetColumnType(colname)== DoubleColumn)
		{
		  isdouble[icut]=true;
                  dmin[icut] = atof(_columnValueRange.at(3*icut+1).c_str());
                  dmax[icut] = atof(_columnValueRange.at(3*icut+2).c_str());
                  dcolvec[icut] = &(source->GetVectorOf<double>(colname));
		  lcolvec[icut] = NULL;
		  ndcut++;
		}
	      else
		{
		  isdouble[icut]=false;
                  lmin[icut] = atol(_columnValueRange.at(3*icut+1).c_str());
                  lmax[icut] = atol(_columnValueRange.at(3*icut+2).c_str());
                  lcolvec[icut] = &(source->GetVectorOf<long>(colname));
		  dcolvec[icut] = NULL;
		  nlcut++;
		}

	      }

	    if(ndcut+nlcut != ncut)
	      {
	       streamlog_out(ERROR) << "Wrong structure of input table!" << std::endl;
	       streamlog_out(ERROR) << "This processing loop will be skiped !" << std::endl;

	       delete [] isdouble;
               delete [] dcolvec;
               delete [] lcolvec;
               delete [] dmin;
               delete [] dmax;
               delete [] lmin;
               delete [] lmax;

               throw SkipEventException( this ) ;
	      }


           // Table row selection - fill decision vector

	   source->AdjustColumns();
	   int nrow = source->GetRowNumber();

	   vector<bool> keepRow;
	   int nkeep = 0;

	   // Iterate over rows

	   for(int irow=0; irow<nrow ; irow++)
	      {
	      bool cutor = false;
	      bool cutand = true;

	      for(int icut = 0; icut < ncut ; icut++)
	        {
		if(isdouble[icut])
		  {
		    if( (dcolvec[icut]->at(irow)-dmin[icut])*(dmax[icut] - dcolvec[icut]->at(irow))*(dmax[icut]-dmin[icut]) > 0 )
		      cutor = true;
		    else
		      cutand = false;
		  }
		else
		  {
		    if( (lcolvec[icut]->at(irow)-lmin[icut])*(lmax[icut] - lcolvec[icut]->at(irow))*(lmax[icut]-lmin[icut]) > 0 )
		      cutor = true;
		    else
		      cutand = false;
		  }
	        }

	      if( cutand || ( _acceptIfAny && cutor) )
		{
		keepRow.push_back(true);
		nkeep++;
		}
	      else
		keepRow.push_back(false);
	      }


	   // If most of data are to be kept - it is better to copy the table and delete 'bad' rows.
	   // Otherwise copy 'good' rows to empty clone

           if(nkeep > 0.95 * nrow)
	     { 
	     if (_outputCollectionName.size() > 0 )
	        {
	        result = new GloriaFitsTable(*source);

	        (data->GetTableCollection(_outputCollectionName)).push_back(*result);

	        int nOut = (data->GetTableCollection(_outputCollectionName)).size();

	        delete result;
                result = &(data->GetTableCollection(_outputCollectionName).at(nOut-1));
	        }
    

	     for(int irow=nrow-1; irow>=0 ; irow--)
	        if(!keepRow.at(irow)) result->DeleteRow(irow);

	     }
	   else
	     {
	     result = source->EmptyClone();

	     for(int irow=0; irow<nrow ; irow++)
	        if(keepRow.at(irow)) result->CopyTableRow(source, irow, -1);


	     if (_outputCollectionName.size() > 0 )
	       {
               (data->GetTableCollection(_outputCollectionName)).push_back(*result);
   	       delete result;

	       int nOut = (data->GetTableCollection(_outputCollectionName)).size();
               result = &(data->GetTableCollection(_outputCollectionName).at(nOut-1));
	       }
	     else
	       {
		data->GetTableCollection(_inputCollectionName).at(itb) = *result;
   	        delete result;

                result = &(data->GetTableCollection(_inputCollectionName).at(itb));
	       }
	     }

             int nnew = result->GetRowNumber();

	     streamlog_out(DEBUG) << ndcut << "(double) + " << nlcut << "(long) cuts : " 
                                  << nnew << " rows selected out of " << nrow << endl;

             delete [] isdouble;
             delete [] dcolvec;
             delete [] lcolvec;
             delete [] dmin;
             delete [] dmax;
             delete [] lmin;
             delete [] lmax;
	    }
	    // end if(ncut > 0)


	  if(result->GetRowNumber() > maxCount) maxCount = result->GetRowNumber();


	  // Sort final table if required

	  if( _sortColumn.size() != 0)
	     {
	       result->SortTable(_sortColumn, _reverseSort);

 	       streamlog_out(DEBUG) << "Table sorted by " << _sortColumn << endl;
	     }


	  // Select columns in the table, if required

	  if( _outputColumns.size() != 0)
	     {
             streamlog_out(DEBUG) << "Table columns selected: " ;

             vector<GloriaTableColumn*> colstore;

             for(int icol=0; icol < _outputColumns.size() ; icol ++)
	       {
                 streamlog_out(DEBUG) << _outputColumns.at(icol) << " " ;

		 GloriaTableColumn* col = result->RemoveColumn(_outputColumns.at(icol));

                 if(col) colstore.push_back(col);
	       }

	     int nsel = colstore.size();

	     if(nsel!=_outputColumns.size())
	       {
                 streamlog_out(WARNING) << "Some columns not found in input table! " << endl ;
	       }


	     int nleft = result->GetColumnNumber();

	     if(nleft)
	       {
  	       streamlog_out(DEBUG) << endl << "      columns removed: ";

               for(int idel=0; idel< nleft; idel++)
		 {
		   streamlog_out(DEBUG) << result->GetColumnName(0) << " ";
		   result->DeleteColumn(0);
		 }

               streamlog_out(DEBUG) << endl;  
	       }


	     for(int isel=0;isel<nsel;isel++)
               result->InsertColumn(colstore.at(isel));
	     }



	  }  // end of loop over input tables

  // always return true  for ProcessorName

     if(maxCount >= _minCount)
	{
         streamlog_out(DEBUG) << "Minimum required count reached. Setting output flag to TRUE." << endl; 
         setReturnValue(true);
        }
     else
	{
         streamlog_out(DEBUG) << "Minimum required count not reached. Setting output flag to FALSE." << endl; 
         setReturnValue(false);
        }

}

void SelectTableData::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void SelectTableData::concludeAnalysis( GloriaDataContainer* data )
{

	streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;
}


void SelectTableData::end()
{

	printEndMessage();

}

void SelectTableData::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() reached after " << _nLoops
                         << " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << " Total of " << _nInputTables 
                         << " tables analysed" << std::endl;


}


} // namespace luiza
