#include "TestHeaderKeyword.h"
#include "luiza/LuizaExceptions.h"
#include "luiza/Global.h"
#include <iostream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza{



  TestHeaderKeyword aTestHeaderKeyword ;


  TestHeaderKeyword::TestHeaderKeyword() : Processor("TestHeaderKeyword") {

    _description = "Simple processor to test header keywords." ;

   registerInputCollection (  ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection to test",
                              _imageCollectionName,  std::string ("") );

   registerInputCollection (  TableCollection,
                              "TableCollectionName",
                              "Name of the table collection to test",
                              _tableCollectionName,  std::string ("") );

   std::vector<std::string > emptyStringVec;

   registerProcessorParameter("RequiredKeywords",
			"Header keywords which will be checked",
			_requiredKeywords, emptyStringVec);

   registerProcessorParameter("RequiredValues",
			"Expected values of the considered header keywords",
			_requiredValues, emptyStringVec);


   std::vector<float> emptyFloatVec;

   registerOptionalParameter("ValueTolerances",
	"Precision for value comparison. Strings are compared if not given or zero",
			_valueTolerances, emptyFloatVec);

   registerOptionalParameter("RequireAllKeywords",
			     "Positive decision only if all header keywords mach their values",
		 	    _requireAllKeywords, true);


   registerOptionalParameter("RequireAllData",
		     "Positive decision only if all input images/tabels fulfill the criteria",
		 	    _requireAllData, true);

 }

  void TestHeaderKeyword::init() { 

    streamlog_out( MESSAGE ) << "  TestHeaderKeyword::init()" << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    

     if (_imageCollectionName.size() == 0  && _tableCollectionName.size() == 0 )
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}


     if (_imageCollectionName.size() > 0  && _tableCollectionName.size() > 0 )
	{
		streamlog_out(ERROR) << "Can not use image and table collection at the same time! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stopped! " << std::endl;

		throw StopProcessingException( this ) ;
	}


     if (_requiredKeywords.size() == 0)
	{
		streamlog_out(ERROR) << "No keywords given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		throw StopProcessingException( this ) ;
	}

     if (_requiredKeywords.size() != _requiredValues.size() )
	{
		streamlog_out(ERROR) << "Number of values given does not match keyword number! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stopped! " << std::endl;

		throw StopProcessingException( this ) ;
	}


     if (_valueTolerances.size() > 0 && _requiredKeywords.size() != _valueTolerances.size() )
	{
		streamlog_out(ERROR) << "Number of tolerances given does not match keyword number! " << std::endl;
		streamlog_out(ERROR) << "Processing will be stopped! " << std::endl;

		throw StopProcessingException( this ) ;
	}

     // Set all tolerances to zero (string comparison), if not given
     
     if(_valueTolerances.size()==0)
       {
	 for(int iKey=0; iKey<_requiredKeywords.size(); iKey++)
	   _valueTolerances.push_back(0.0);
       }

     // Set minimum and maximum value for numerical comparison

     for(int iKey=0; iKey<_requiredKeywords.size(); iKey++)
       {
	 if(_valueTolerances.at(iKey) > 0 )
	   {
	    float val = atof(_requiredValues.at(iKey).c_str());

	    _valueMin.push_back(val + _valueTolerances.at(iKey));
	    _valueMax.push_back(val + _valueTolerances.at(iKey));

	    streamlog_out(MESSAGE) << "Require header keyword " << _requiredKeywords.at(iKey) << " to be between "
				   << _valueMin.at(iKey) << " and " << _valueMax.at(iKey) << std::endl;
	      
	   }
	 else
	   {
	    _valueMin.push_back(0.0);
	    _valueMax.push_back(0.0);

	    streamlog_out(MESSAGE) << "Require header keyword " << _requiredKeywords.at(iKey) << " be set to "
				   << _requiredValues.at(iKey) << std::endl;
	      
	   }
	   
       }
     
    _useTableData =  _tableCollectionName.size() > 0;
	
    _nImages = 0 ;
    _nTables =0;
    _nLoops = 0 ;
 
   // No output yet
	
   setReturnValue(false);
  }


 void TestHeaderKeyword::startAnalysis( GloriaDataContainer* data ) { 

    streamlog_out( MESSAGE ) << "  TestHeaderKeyword::startAnalysis()" << std::endl ;
    
  }



  void TestHeaderKeyword::processData( GloriaDataContainer* data ) { 

    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    // Check if input collection exists

    bool inputOK = true;
    
    if (_useTableData)
      inputOK = data->TableCollectionExists(_tableCollectionName);
    else
      inputOK = data->ImageCollectionExists(_imageCollectionName);

    if(!inputOK)
	{
	 streamlog_out(DEBUG) << "Input collection not found ! " << endl;
  	 return;
	}


    int nCol = 0;

    if (_useTableData)
      {
       nCol =   data -> GetTableCollection(_tableCollectionName).size();
       streamlog_out(DEBUG) << nCol << " tables found in input collection" << std::endl ;
       _nTables+=nCol;
      }
    else
      {
       nCol =   data -> GetImageCollection(_imageCollectionName).size();
       streamlog_out(DEBUG) << nCol << " images found in input collection" << std::endl ;
       _nImages+=nCol;
      }

    
    // Loop over images/tables collections

    bool OKifAnyData = false;
    bool OKifAllData = true;
    
    for(int iCol=0; iCol<nCol; iCol++)
      {
	GloriaFitsHeader *source;

        if (_useTableData)
	  source = &(data -> GetTableCollection(_tableCollectionName).at(iCol));
        else
	  source = &(data -> GetImageCollection(_imageCollectionName).at(iCol));

	
	bool OKifAny = false;
	bool OKifAll = true;

	for(int iKey=0; iKey < _requiredKeywords.size(); iKey++)
	  {
	  const char *keyName = _requiredKeywords.at(iKey).c_str();

	  if(source->IsKeywordSet(keyName))
	    {
	      if( _valueTolerances.at(iKey) == 0.0 )
		{
	         string keyVal = source->GetKeywordValue(keyName);

		 bool keyPass = ( keyVal == _requiredValues.at(iKey) ||
				  keyVal == "'" + _requiredValues.at(iKey) + "'" );
		 
		 if( keyPass)
		   OKifAny = true;
		 else
		   OKifAll = false;

		 
		 streamlog_out(DEBUG) << _requiredKeywords.at(iKey) << " value is " << keyVal
				      << " required is " << _requiredValues.at(iKey) 
		                      << " outcome is " << keyPass << std::endl ;
		}
	      else
		{
	         float keyVal = source->GetKeywordValueDouble(keyName);

		 bool keyPass = ( keyVal > _valueMin.at(iKey) &&  keyVal < _valueMax.at(iKey));

		 if( keyPass)
		   OKifAny = true;
		 else
		   OKifAll = false;
		  
		 streamlog_out(DEBUG) << _requiredKeywords.at(iKey) << " value is " << keyVal
		                      << " outcome is " << keyPass << std::endl ;
		}
	    }
	  else
	    {
	     streamlog_out(DEBUG) << _requiredKeywords.at(iKey) << " not found " << std::endl ;
      
  	     OKifAll = false;
	    }
	  
	  }


	bool fileVal;
	
	if(_requireAllKeywords)
	  fileVal = OKifAll;
	else
	  fileVal = OKifAny;


	streamlog_out(DEBUG) << " Keyword test result for header " << iCol << " is " << fileVal << std::endl ;

	
	if(fileVal)
	  OKifAnyData = true;
	else
	  OKifAllData = false;

	
      }


    // return proper value

    bool retVal;

    if(_requireAllData)
      retVal = OKifAllData;
    else
      retVal = OKifAnyData;

    streamlog_out(MESSAGE) << "Setting output flag to " << retVal << std::endl ;
    
    setReturnValue( retVal ) ;
 
  }
  
  void TestHeaderKeyword::check(GloriaDataContainer* data) { 
    
    streamlog_out(DEBUG) << "   TestHeaderKeyword::check() " << std::endl ;
    
  }
  
 void TestHeaderKeyword::concludeAnalysis( GloriaDataContainer* data ) { 

    
    streamlog_out( DEBUG ) << "   TestHeaderKeyword::concludeAnalysis()  " << std::endl ;
    
    printEndMessage() ;

 }


  void TestHeaderKeyword::end(){ 

    streamlog_out( DEBUG ) << "   TestHeaderKeyword::end()  " << std::endl ;

  }



 void TestHeaderKeyword::printEndMessage() const {

    streamlog_out(MESSAGE) << " Closing after " << _nLoops << " calls to processData() " << std::endl ;

    if(_useTableData)
      {
	streamlog_out(MESSAGE) << " processed "     << _nTables << " tables "  << std::endl;
      }
    else
      {
        streamlog_out(MESSAGE) << " processed "     << _nImages << " images " << std::endl ;
      }
    
  }
  
}// namespace luiza
