// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef PiFitcImageReader_h
#define PiFitcImageReader_h 1

#include "luiza/DataSourceProcessor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor for reading compressed Pi of the Sky images (.fitc)
  /*! This processor reads FITS images taken by Pi of the Sky,
   *  compressed and written in nonstandard format (.fitc extension).
   *  In is supposed to work exactly as FitsImageReader for standard
   *  files. 
   * 
   *  \par Input 
   *  Names of input FITC files can be given in two ways: directly, or
   *  by specifying names of list files (i.e. files containing image
   *  file names, one per line). 
   *    
   *  \param FitsFileList - List of FITS files to be read
   * 
   *  \param ListFileList - Names of files containing FITS file lists
   *                      (one per line) 
   *
   *  \param ImagesPerLoop - Number of images to be read in each
   *   processor call. Default is one. If zero value is set, all images
   *   from the list are read at the first call (!)
   *
   *  \par Output
   *   Images read as gloria::GloriaFitsImage are added to specified
   *   image collection (named image vector) and put into global 
   *   gloria::GloriaDataContainer
   *
   *  \param ImageCollectionName - Name of the created
   *    GloriaImageCollection (default is "FitsImages");
   *
   *  \param PermanentCollection - Flag indicating that created
   *    collection should be kept in memory for the whole
   *    processing time. This is needed in case of eg. calibration frames
   *    (darks, flats) or collections containing analysis
   *    results. Default is \a false
   *
   *  \param RemoveLine1024 - Flag for removing middle line in each
   *    image. It is added by the camera firmware when reading
   *    double-sided chip. Default is \a true
   *
   * 
   *  \author A.F. Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   */

   class PiFitcImageReader : public DataSourceProcessor {
    
  public:
    
    virtual PiFitcImageReader*  newProcessor();
    
    
    PiFitcImageReader() ;
    
    /** Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /** Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    /** Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! Decoding routine, modified from CFITSIO library
   int pi_fits_rdecomp (unsigned char *c, int clen, unsigned int *array,
			int nx, int nblock);
     
    //! List of FITS files to be read
    std::vector< std::string > _fitsFileList ;

    //! Names of files containing FITS file lists (one per line) - alternative method
    std::vector< std::string > _listFileList ;

    //! Name of the image collection to which images from file should be stored
    std::string _imageCollectionName ;

    //! Flag for collections, which should not be deleted after loop is finished
    bool _permanentCollection;

    //! Flag for removing line 1024
    bool _rml1024;

    //! Number of images to be read per processing loop (0 for all)
    int _imagesPerLoop ;

    //! Loop counter
    int _nLoops ;

    //! Image counter
    int _nImages ;

    //! File counter
    int _nFiles ;

    /**Test method for const.*/
    void printEndMessage() const ;
  } ;

 } // end namespace luiza
#endif
