// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef PiHalvesEqualizer_h
#define PiHalvesEqualizer_h 1

#include "luiza/Processor.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor to correct for gain difference between two halves of the frame
  /*! Processor calculates background map of given image. The frame
   *  is divided into FrameDivision x FrameDivision subframes, then
   *  background level is calculated from median in each subframe. 
   *  Multiplicative gain correction for upper half is calculated to
   *  match background levels in the middle of the frame. Input images
   *  are modified.
   *
   *  \par Input 
   *  Processor takes images from Gloria data collection
   *
   *  \param InputCollectionName -
   *             Name of the collection containing images to be corrected
   *
   *  \par Algorithm parameters
   *  User can modify processor performance by setting following
   *  parameters: 
   *
   *  \param BackgroundSampling  - Background level calculation can be
   *                       sped up by considering only
   *                       1/BackgroundSampling of pixels.
   *
   *  \param FrameDivision - Background is calculated for
   *                         FrameDivision x FrameDivision subimages.
   *                         Even number required (!).
   *
   *  \param InterpolationOrder - polinomial order used to interpolate
   *                         gain corrections along horizontal axis.
   *
   */

  class PiHalvesEqualizer : public Processor {
    
  public:
    
    virtual Processor*  newProcessor() { return new PiHalvesEqualizer ; }
    
    
    PiHalvesEqualizer() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
  protected:

    /**Test method for const.*/
    void printEndMessage() const ;

    //! Name of the image collection containing images to be corrected
    string _inputCollectionName;

    //! Background sampling factor
    /*! Parameter used to speed up background level calculation.
     *  Only every BackgroundSampling pixel is used.
     */
    int _backgroundSampling;

    //! Frame division for background calculation
    /*! Background is calculated for FrameDivision x FrameDivision 
     *  subimages and then interpolated between them
     */
    int _frameDivision;

    //! Polinomial order used to interpolate background map
    int _interpolationOrder; 

    // Other internal parameters

    int _sizeX;
    int _sizeY;
    long _sizeXY;

    int _nLoops ;
    int _nInputImages ;
    int _nOutputImages ;

  } ;
  
} // end namespace luiza
#endif



