// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef PiTableReader_h
#define PiTableReader_h 1

#include "luiza/DataSourceProcessor.h"
// #include "Processor.h"
// #include "data/GloriaDataContainer.h"

using namespace gloria;

namespace luiza {
  
  
  //! Processor for reading Pi of the Sky data tables (mag and ast files)
  /*! This processor reads data tables created in the Pi of the Sky
   *  image processing pipeline (mag and ast files). Tables are stored
   *  as FITS images, so proper conversion is required. 
   *  It is defined as luiza::DataSourceProcessor, so it
   *  can be the first processor in the analysis chain. 
   * 
   *  \par Input 
   *  Names of input files can be given in two ways: directly, or
   *  by specifying names of list files (i.e. text files containing table
   *  file names, one per line). 
   *    
   *  \param TableFileList List of fits files to be read
   * 
   *  \param ListFileList Names of files containing lists of fits file names (one per line)
   *
   *  \param TablesPerLoop Number of tables to be read in each
   *   processor call. Default is one. If zero value is set, all tables
   *   from the list are read at the first call (!)
   *
   *
   *  \par Output
   *   Tables read are added to specified table collection (named
   *   table vector) and put into global gloria::GloriaDataContainer.
   *
   *  \param TableCollectionName Name of the created
   *    GloriaTableCollection (default is "PiTable");
   *
   *  \param PermanentCollection Flag indicating that created
   *    collection should be kept in memory for the whole
   *    processing time. This is needed in case of eg. tables
   *    containing reference stars.
   *
   *  \param SelectedColumns 
   *    Names (and order) of columns which should be read into the
   *    output table. If missing, all columns are read. 
   *
   *  \author A.F. Zarnecki, University of Warsaw, zarnecki@fuw.edu.pl
   */

   class PiTableReader : public DataSourceProcessor {
    
  public:
    
    virtual PiTableReader*  newProcessor();
    
    
    PiTableReader() ;
    
    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    //! Called before first data loop.
    /*! Can be used to create permanent data structures, if needed.
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;
    
    /*! Called for every data loop
     */
    virtual void processData( GloriaDataContainer* data ) ;
    
    
    virtual void check( GloriaDataContainer* data ) ; 
    
    
    //! Called after all data processing loops to finalize the analysis

    virtual void concludeAnalysis( GloriaDataContainer* data ) ;

    /*! Called after data processing for clean up.
     */
    virtual void end() ;
    
    
  protected:

    //! List of fits table files to be read
    std::vector< std::string > _tableFileList ;

    //! Names of files containing lists of table file names (one per line)
    std::vector< std::string > _listFileList ;

    //! Name of the table collection to which tables from file should be stored
    std::string _tableCollectionName ;

    //! Flag for collections, which should not be deleted after loop is finished
    bool _permanentCollection;

    //! Name (and order) of columns which should be copied to the output table
    std::vector< std::string > _selectedColumns ; 

    //! Number of tables to be read per processing loop (0 for all)
    int _tablesPerLoop ;

    //! Loop counter
    int _nLoops ;

    //! Table counter
    int _nTables ;

    //! File counter
    int _nFiles ;

    /**Test method for const.*/
    void printEndMessage() const ;

    GloriaFitsImage* _inputImage; 
    GloriaFitsTable* _inputTable; 
  } ;

 } // end namespace luiza
#endif
