// -*- mode: c++; mode: auto-fill; mode: flyspell-prog; -*-

#ifndef PiTableWriter_h
#define PiTableWriter_h 1

#include "luiza/Processor.h"

#include <fstream>

using namespace gloria;

namespace luiza {


  //! Processor for writing tables in IMAGE file format
  /*! This processor writes tables to output file
   *  in the format of FITS image. This is the standard format for
   *  data from the Pi of the Sky pipeline 
   *
   *  \par Input 
   *  Tables are taken from the data collection.
   *  User has to specify in the steering file:
   *
   *  \param TableCollectionName -
   *    Name of the table collection which should be stored to file(s)
   *
   *  \par Output
   *  GloriaFitsTables from given input collection 
   *  are stored as FITS image files. Output file names consist of file name 
   *  root, table number (starting from 1) and name extension. If the
   *  name root ends with '/', the input file name is appended. 
   *  Tables are stored at each loop unless the collection 
   *  is permanent - then they are stored only at concludeAnalysis()
   *  call. 
   *
   *  \param TextFileNameRoot -
   * Output file name root, which will be appended by the table number 
   * and extension
   *
   *  \param TextFileNameExt - Output file name extension
   *
   *  \param SelectedColumns - names of columns to be stored in the
   *         file. If not specified (empty list) all columns are
   *         written (as found in the first table being processed!)
   *
   *  \param NewColumnNames
   *          Column names which should be stored in output file. This 
   *          allows user to adjust column names to their needs.
   *          If missing, unchanged column names will be used.
   *     
   *  \param PosColumns - names of columns where the star positions
   *  in RA, DEC are stored. Values in these columns will be used to
   *  calculate star position w.r.t. frame center. RA is assumed to be
   *  given in degrees and the relative position D_RA will be given in
   *  hours when writing to file. 
   *
   *  \param ShiftKeywords - names of header records where the
   *  coordinates of the frame center are stored (both in
   *  degrees). These values will be used to transform RA,DEC
   *  positions given in the table to the relative shifts. 
   *
   *  \author A.F. Zarnecki, University of Warsaw
   */

   class PiTableWriter : public Processor {

  public:

    virtual PiTableWriter*  newProcessor();


    PiTableWriter();

    /*! Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;

    /*! Called for before first data loop
     */
    virtual void startAnalysis( GloriaDataContainer* data ) ;


    /*! Called for every data collection
     */
    virtual void processData( GloriaDataContainer* data ) ;


    virtual void check( GloriaDataContainer* data ) ;


    /*! Called after all data loops to conclude analysis
     */
    virtual void concludeAnalysis( GloriaDataContainer* data ) ;


    /*! Called after data processing for clean up.
     */
    virtual void end() ;


  protected:

    //! Name of the table collection which should be stored
    string _tableCollectionName ;

    //! Output file name root
    string _textFileNameRoot;

    //! Output file name extention
    string _textFileNameExt;

    //! Table columns to be stored
    std::vector< std::string > _selectedColumns;

    //! New column names, for columns stored to the file
    std::vector<std::string> _newColumnNames;

    //! Header fields for RA/DEC offsets
    std::vector< std::string > _shiftKeywords;

    //! Column names for RA/DEC 
    std::vector< std::string > _posColumns;

    //! Temporary table to prepare text output
    GloriaFitsImage* _outputImage;

    //! Output file pointer
    ofstream _outputStream;

    //! Loop counter
    int _nLoops ;

    //! Table counter, used to number file name
    int _nTables ;

    //! Counter for images successfully written to file
    int _nWritten ;


    void printEndMessage() const ;
  } ;

 } // end namespace luiza
#endif
