#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "PiFitcImageReader.h"

namespace luiza {

  PiFitcImageReader aPiFitcImageReader ;


  // Constructor

  PiFitcImageReader::PiFitcImageReader() : DataSourceProcessor("PiFitcImageReader") {

    _description = "Processor for reading compressed Pi of the Sky images"
                   "Reads images from given files" ;


  std::vector< std::string > emptyStringVec;

  registerProcessorParameter ("FitsFileList",
                              "List of FITS files to be read",
                              _fitsFileList,  emptyStringVec );


  registerProcessorParameter ("ListFileList",
                              "Names of files containing FITS file lists (one per line)",
                              _listFileList,  emptyStringVec );


  // Although this processor reads images from file it creates image collection as an output
  // This is why we register this collection as output collection
  registerOutputCollection (  ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection to which images from file should be stored",
                              _imageCollectionName,  std::string ("FitsImages") );


  // Remaining parameters are optional - can be skiped in the steering file

  registerOptionalParameter ("PermanentCollection",
                              "Flag for collections, which should not be deleted after loop is finished",
                              _permanentCollection,  false );



  registerOptionalParameter ("RemoveLine1024",
                              "Flag for removing middle line in each image.",
                              _rml1024,  true );



  registerOptionalParameter ("ImagesPerLoop",
                              "Number of images to be read per processing loop (0 for all))",
                              _imagesPerLoop,  static_cast < int > (1) );


  }


  PiFitcImageReader*  PiFitcImageReader::newProcessor() {
    return new PiFitcImageReader ;
  }
  

  void PiFitcImageReader::init() { 

    streamlog_out( MESSAGE ) << "PiFitcImageReader::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    
    _nFiles = 0 ;
    _nImages = 0 ;
    _nLoops = 0 ;
 
    // Check if input file name given

    if(_fitsFileList.size()==0 && _listFileList.size()==0)
      {
	streamlog_out( ERROR ) << "No input files defined for PiFitcImageReader! " << std::endl;
	streamlog_out( ERROR ) << "Input NOT initialized! " << std::endl;

	//        throw StopProcessingException( this ) ;
	return;

      }

  // Open list files and copy file names to the fits file list

    for(int ilist=0; ilist < _listFileList.size(); ilist++)
      {
      std::ifstream listFile;
      listFile.open(_listFileList[ilist].c_str(),std::ios::in);

      if(listFile.fail())
	{
	  streamlog_out( ERROR ) << "Error opening input list file: " 
                                 << _listFileList[ilist] <<std::endl;
          continue;
	}

      std::string fitsFile;

      while(1)
	{
	  listFile >> fitsFile;

          if(!listFile.good())break;

	  _fitsFileList.push_back(fitsFile);  

	}

      listFile.close();
      }

  // As there are file names specified, we are ready to start processing

    _isInputInitialized = true;

    streamlog_out( MESSAGE ) << "Ready to fill "<< _imageCollectionName << " image collection" << std::endl;
    streamlog_out( MESSAGE ) << "Input file list has " << _fitsFileList.size() << " entries" <<std::endl;

 // Print list of files in DEBUG mode

    for(int ifile=0;ifile<_fitsFileList.size();ifile++)
	 streamlog_out( DEBUG )  <<  "      File " <<  std::setw ( 5 ) << ifile+1 << " : " 
                                 << _fitsFileList[ifile] <<std::endl;

    // Set return value to false (no input yet)
    
    setReturnValue( false ) ;
    
 }

  void PiFitcImageReader::processData( GloriaDataContainer* data ) { 

    bool readOK = false;
    
    // Set return value to false (no input)
    
    setReturnValue( false ) ;
    
    // Return without processing if no files left to read

    if(_nFiles >= _fitsFileList.size() ) return;


    // Count only loops where something was done...

    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    
    // Create output collection if not found

    if(! data -> ImageCollectionExists(_imageCollectionName))
         data -> CreateImageCollection(_imageCollectionName,_permanentCollection);

    // Read _imagesPerLoop in each call to processData()
    // If _imagesPerLoop is set to 0: all images are read at first call (!)

    for(int ifile=0; (ifile<_imagesPerLoop || _imagesPerLoop==0) && _nFiles<_fitsFileList.size() ;ifile++)
      {
      // Take next file name

      std::string fileName = _fitsFileList[_nFiles];

      _nFiles++;

      // When this is the last file, set the 'close' flag
    
      if(_nFiles == _fitsFileList.size()) _isInputClosed=true;

  
  fitsfile *fp=NULL;
  int status = 0;

   // Open fits file: look for an image (!)

  fits_open_image(&fp, fileName.data(), READONLY, &status);

  if(status)
      {
      streamlog_out(ERROR) << "Could not open image file " << fileName << endl;
      continue;
      }

  
  //Read image parameters
  
  int bitpix=0;
  int naxis=0;
  long axsizes[2];

  fits_get_img_param(fp, 2, &bitpix, &naxis, axsizes, &status);

  if(naxis!=2 || bitpix != 16)
     {
       streamlog_out(ERROR) << "Image structure not supported for file " << fileName << endl;

       fits_close_file(fp, &status);

       continue;
     }

  // Read fits header
  
  GloriaFitsHeader myhead;
  myhead.ReadFitsHeader(fp);

  // Remove compression information from header
  
  myhead.RemoveKeywords("COMPRES");
  myhead.RemoveKeywords("BLOCKSZ");
  myhead.RemoveKeywords("DIVISOR");

  // Read compressed data
  
  int sizeX = axsizes[0];
  int sizeY = axsizes[1];
  int sizeXY= sizeX * sizeY;

  unsigned short *inPixels = new unsigned short[sizeXY];
  unsigned char *buffer = new unsigned char[2*sizeXY];
  unsigned int *outPixels = new unsigned int[sizeXY];

  for(int ipix=0; ipix<sizeXY; ipix++) inPixels[ipix] = 0;
  
   long firstpixel[2] = {1, 1};
   
   fits_read_pix(fp, TUSHORT, firstpixel, sizeXY,
                                   NULL, inPixels, NULL, &status);

   // Ignore status 108 

    if(status!=0 && status!=108)
      {
       streamlog_out(ERROR) << "Error reading compressed image data from file " << fileName << endl;

       fits_close_file(fp, &status);

       delete [] inPixels;
       delete [] buffer;
       delete [] outPixels;

       continue;
      }

   fits_close_file(fp, &status);

   int npix = 0;

   for(int ipix=0; ipix<sizeXY; ipix++)
     {
     inPixels[ipix] -= 32768;
     buffer[2*ipix] = inPixels[ipix] >> 8;
     buffer[2*ipix+1] = inPixels[ipix] & 0xFF ;

     if(inPixels[ipix]!=0)npix=ipix+1;
     }

   // Decode line by line

   unsigned char *bstart = buffer;

   int nstart = 2*sizeXY;
   
   for(int iy=0;iy<sizeY;iy++)
     {  
     int left=pi_fits_rdecomp(bstart,nstart,outPixels+iy*sizeX,sizeX,32);

     if(left<0)
       {
       streamlog_out(ERROR) << "Error returned when decompressing line " << iy << " of file " << fileName << endl;

       delete [] inPixels;
       delete [] buffer;
       delete [] outPixels;

       continue;
       }

     int nused = nstart - left;

     nstart -=nused;
     bstart +=nused;
     }

   // Create output image (Int type)
   
   GloriaFitsImage *myfit = new GloriaFitsImage(sizeX, sizeY, false);

   myfit->CopyHeader(&myhead);
		      
   myfit->SetInputFileName(fileName);

   long* gPixels = myfit->GetIntImageData();


   if(_rml1024)
     {
     // For removal of line 1024

     for(int iy=0; iy<sizeY; iy++)
       {
       int oshift = 0;
       
       if(iy==1024)oshift = 1023*sizeX;
       if(iy>1024)oshift = -sizeX;
        
       for(int ix=0; ix<sizeX; ix++)
         {
         gPixels[ix+iy*sizeX+oshift] =  outPixels[ix+iy*sizeX];
         }
       }
     }
   else
      for(int ipix=0; ipix<sizeXY; ipix++)
        gPixels[ipix] =  outPixels[ipix];

      // Add image to the collection

      (data->GetImageCollection(_imageCollectionName)).push_back(*myfit);
      delete myfit;

       delete [] inPixels;
       delete [] buffer;
       delete [] outPixels;

       readOK = true;
      _nImages++;
      }


    // return true if image read
    if(readOK) setReturnValue( true ) ;
    
  }
  
  void PiFitcImageReader::check(GloriaDataContainer* data) { 
    
    if(! data -> ImageCollectionExists(_imageCollectionName))
      {
         streamlog_out(ERROR) << " check(): Collection not found !? " << std::endl ;
	 return;
      }


    int nImg = data->GetImageCollection(_imageCollectionName).size();

    streamlog_out(MESSAGE) << " check(): Collection " << _imageCollectionName << " with " << nImg << " images" << std::endl ;
 
  }
  
  void PiFitcImageReader::end(){ 
    
    printEndMessage() ;

  }

  void PiFitcImageReader::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;

    streamlog_out(MESSAGE) << " processed "     << _nFiles << " files " << std::endl 
			   << std::endl ;
    
    streamlog_out(MESSAGE) << " read "     << _nImages << " images " << std::endl 
			   << std::endl ;
  }


int PiFitcImageReader::pi_fits_rdecomp (
	     unsigned char *c,		/* input buffer			*/
	     int clen,			/* length of input		*/
	     unsigned int array[],	/* output array			*/
	     int nx,			/* number of output pixels	*/
	     int nblock)		/* coding block size		*/
{
int i, k, imax;
int nbits, nzero, fs;
unsigned char *cend, bytevalue;
unsigned int b, diff, lastpix;
int fsmax, fsbits, bbits;

static const int nonzero_count[256] = {
0, 
1, 
2, 2, 
3, 3, 3, 3, 
4, 4, 4, 4, 4, 4, 4, 4, 
5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8};


    /* For Pi of the Sky */
    fsbits = 4;
    fsmax = 14;

    bbits = 1<<fsbits;

    /*
     * Decode in blocks of nblock pixels
     */

    /* first 4 bytes of input buffer contain the value of the first */
    /* 4 byte integer value, without any encoding */
    
    lastpix = 0;
    bytevalue = c[0];
    lastpix = lastpix | (bytevalue<<24);
    bytevalue = c[1];
    lastpix = lastpix | (bytevalue<<16);
    bytevalue = c[2];
    lastpix = lastpix | (bytevalue<<8);
    bytevalue = c[3];
    lastpix = lastpix | bytevalue;

    c += 4;  
    cend = c + clen - 4;

    b = *c++;		    /* bit buffer			*/
    nbits = 8;		    /* number of bits remaining in b	*/
    for (i = 0; i<nx; ) {
	/* get the FS value from first fsbits */
	nbits -= fsbits;
	while (nbits < 0) {
	    b = (b<<8) | (*c++);
	    nbits += 8;
	}
	fs = (b >> nbits) - 1;

	b &= (1<<nbits)-1;
	/* loop over the next block */
	imax = i + nblock;
	if (imax > nx) imax = nx;
	if (fs<0) {
	    /* low-entropy case, all zero differences */
	    for ( ; i<imax; i++) array[i] = lastpix;
	} else if (fs==fsmax) {
	    /* high-entropy case, directly coded pixel values */
	    for ( ; i<imax; i++) {
		k = bbits - nbits;
		diff = b<<k;
		for (k -= 8; k >= 0; k -= 8) {
		    b = *c++;
		    diff |= b<<k;
		}
		if (nbits>0) {
		    b = *c++;
		    diff |= b>>(-k);
		    b &= (1<<nbits)-1;
		} else {
		    b = 0;
		}
		/*
		 * undo mapping and differencing
		 * Note that some of these operations will overflow the
		 * unsigned int arithmetic -- that's OK, it all works
		 * out to give the right answers in the output file.
		 */
		if ((diff & 1) == 0) {
		    diff = diff>>1;
		} else {
		    diff = ~(diff>>1);
		}
		array[i] = diff+lastpix;
		lastpix = array[i];
	    }
	} else {
	    /* normal case, Rice coding */
	    for ( ; i<imax; i++) {
		/* count number of leading zeros */
		while (b == 0) {
		    nbits += 8;
		    b = *c++;
		}
		nzero = nbits - nonzero_count[b];
		nbits -= nzero+1;
		/* flip the leading one-bit */
		b ^= 1<<nbits;
		/* get the FS trailing bits */
		nbits -= fs;
		while (nbits < 0) {
		    b = (b<<8) | (*c++);
		    nbits += 8;
		}
		diff = (nzero<<fs) | (b>>nbits);
		b &= (1<<nbits)-1;

		/* undo mapping and differencing */
		if ((diff & 1) == 0) {
		    diff = diff>>1;
		} else {
		    diff = ~(diff>>1);
		}
		array[i] = diff+lastpix;
		lastpix = array[i];
	    }
	}
	if (c > cend) {
	  streamlog_out(ERROR) << "Decompression error: hit end of compressed byte stream" << endl;
	  return -1;
	}
    }
    // Needed for Pi implementation
    return (cend-c);
}

  
}// namespace luiza
