#include "PiHalvesEqualizer.h"
#include "gloria/GloriaMedianMap.h"


#include "luiza/LuizaExceptions.h"
#include <iostream>
#include <iomanip>  
#include <sstream>
#include <cmath>
#include <algorithm>

#include "gloria/GloriaObjectList.h"

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

namespace luiza
{

PiHalvesEqualizer aPiHalvesEqualizer;

PiHalvesEqualizer::PiHalvesEqualizer() :
		Processor("PiHalvesEqualizer")
{

	_description = "Processor to correct for gain differences between two halve of the frame.";

	registerInputCollection(ImageCollection, "InputCollectionName",
			"Name of the collection containing input images",
			_inputCollectionName, std::string(""));

	registerOptionalParameter(
			"BackgroundSampling",
			"Only every BackgroundSampling pixel is considered in background calculation",
			_backgroundSampling, (int)1 );

	registerOptionalParameter(
			"FrameDivision",
			"Frame division for background calculation",
			_frameDivision, (int)8 );

	registerOptionalParameter(
			"InterpolationOrder",
			"Polinomial order used for gain interpolation",
			_interpolationOrder, (int)1 );

}


void PiHalvesEqualizer::init()
{

	streamlog_out(MESSAGE) << "PiHalvesEqualizer::init()  " << name()
			<< std::endl << "  parameters: " << std::endl << *parameters()
			<< std::endl;

	_nInputImages = 0;
	_nOutputImages = 0;
	_nLoops = 0;


	// Check input parameters

	if (_inputCollectionName.size() == 0)
	{
		streamlog_out(ERROR) << "No input collection name given! " << std::endl;
		streamlog_out(ERROR) << "Nothing to do! " << std::endl;

		//        throw StopProcessingException( this ) ;
		return;
	}


	if (_frameDivision%2 != 0 || _frameDivision<0)
	  {
	  streamlog_out(ERROR) << "Wrong frame division specified: " 
                               << _frameDivision << std::endl;

	  streamlog_out(ERROR) << "Positive, even number required! " << std::endl;
	  throw StopProcessingException( this );
  	  return;
	  }

	return;

}

void PiHalvesEqualizer::processData(GloriaDataContainer* data)
{

	// Return if input or output collection not specified
	// Error already printed in init()

	if (_inputCollectionName.size() == 0)
		return;

	_nLoops++;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops
			<< std::endl;

	// Check if input collection exists

	if (!data->ImageCollectionExists(_inputCollectionName))
	{
		streamlog_out(DEBUG) << "Input collection not found ! " << endl;

		return;
	}

	int nImg = data->GetImageCollection(_inputCollectionName).size();

	if (nImg == 0)
	{
		streamlog_out(DEBUG) << "No images in input collection ! " << endl;

		return;
	}

	streamlog_out(DEBUG) << "  Collection " << _inputCollectionName << " with "
			<< nImg << " images " << endl;


	// Loop over images in input collection
	//=====================================

	for (int ifm = 0; ifm < nImg; ifm++)
	  {
          GloriaFitsImage *cur_img = &(data->GetImageCollection(_inputCollectionName).at(ifm));

	  // Skip empty images

	  _sizeX = cur_img->SizeX();
	  _sizeY = cur_img->SizeY();

	  _sizeXY = _sizeX * _sizeY;

	  if(_sizeXY == 0) continue;

	  _nInputImages++;


	  // Calculate background map for this image
	  // Use GloriaMedianMap class

	  GloriaMedianMap* BgMap = new GloriaMedianMap( _frameDivision, _frameDivision);
	  BgMap->CalculateBgMap( cur_img, _backgroundSampling);


	  // Control printout of the bachround map

	  double *BgTable = BgMap->GetBgMap();

  	  streamlog_out(DEBUG) << "  Calculated background map: " << endl << endl;

	  for(int iy=0; iy<_frameDivision; iy++)
	    {
  	     for(int ix=0; ix<_frameDivision; ix++)
	       streamlog_out(DEBUG) << "   " << setw(10) << BgTable[iy*_frameDivision+ix];

               streamlog_out(DEBUG) << endl << endl;
	    }


	  // Calculate gain corrections in columns

	  double *GainCorr = new double[_frameDivision];

	  int iyu1 = _frameDivision/2;
	  int iyu2 = iyu1+1;
	  int iyl1 = iyu1-1;
	  int iyl2 = iyu1-2;

	  // Sum tables for interpolation

          double *sumx= new double[5];
	  double *sumxy= new double[3];

          for(int ipow=0;ipow<=4;ipow++)sumx[ipow]=0.;
	  for(int ipow=0;ipow<=2;ipow++)sumxy[ipow]=0.;

  	  streamlog_out(DEBUG) << "  Calculated gain corrections: " << endl << endl;

	  for(int ix=0; ix<_frameDivision; ix++)
	    {
            double vu1 = BgTable[iyu1*_frameDivision+ix];
            double vu2 = BgTable[iyu2*_frameDivision+ix];
            double vl1 = BgTable[iyl1*_frameDivision+ix];
            double vl2 = BgTable[iyl2*_frameDivision+ix];


	    double vuexp = vu1 * sqrt(vu1/vu2);
	    double vlexp = vl1 * sqrt(vl1/vl2);

            GainCorr[ix] = vlexp/vuexp;

	    streamlog_out(DEBUG) << "   " << setw(10) <<  GainCorr[ix];


	    double xval = 1.;
	    for(int ipow=0;ipow<=4;ipow++)
	      {
		sumx[ipow]+=xval;
		xval*=ix;
	      }

	    double xyval = GainCorr[ix];
	    for(int ipow=0;ipow<=2;ipow++)
	      {
		sumxy[ipow]+=xyval;
		xyval*=ix;
	      }

	    }

          streamlog_out(DEBUG) << endl << endl;

	  // Calculate gain correction polinomial (solving matrix equation 1x1, 2x2 or 3x3)
	  
	  double corr0,corr1,corr2;

	  if(_interpolationOrder==0)
	    {
	      corr0=sumxy[0]/sumx[0];
	      corr1=corr2=0;
  	      streamlog_out(MESSAGE) << "  Gain corrections aplied: " << corr0 << endl;
	    }
	  else
	  if(_interpolationOrder==1)
	    {
	      double wg = sumx[2]*sumx[0]-sumx[1]*sumx[1];
	      corr0 = (sumx[2]*sumxy[0]-sumx[1]*sumxy[1])/wg;
	      corr1 = (sumxy[1]*sumx[0]-sumxy[0]*sumx[1])/wg;
	      corr2=0;
  	      streamlog_out(MESSAGE) << "  Gain corrections aplied: " << corr0 << " + ix * " << corr1 << endl;
	    }
	  else
	    {
	      double wg=sumx[4]*sumx[2]*sumx[0]+sumx[3]*sumx[1]*sumx[2]+sumx[2]*sumx[3]*sumx[1]
                       -sumx[4]*sumx[1]*sumx[1]-sumx[3]*sumx[3]*sumx[0]-sumx[2]*sumx[2]*sumx[2];

                  corr0=sumx[4]*sumx[2]*sumxy[0]+sumx[3]*sumx[1]*sumxy[2]+sumx[2]*sumx[3]*sumxy[1]
                       -sumx[4]*sumx[1]*sumxy[1]-sumx[3]*sumx[3]*sumxy[0]-sumx[2]*sumx[2]*sumxy[2];

	          corr1=sumx[4]*sumxy[1]*sumx[0]+sumx[3]*sumxy[0]*sumx[2]+sumx[2]*sumxy[2]*sumx[1]
                       -sumx[4]*sumxy[0]*sumx[1]-sumx[3]*sumxy[2]*sumx[0]-sumx[2]*sumxy[1]*sumx[2];

	          corr2=sumxy[2]*sumx[2]*sumx[0]+sumxy[1]*sumx[1]*sumx[2]+sumxy[0]*sumx[3]*sumx[1]
                       -sumxy[2]*sumx[1]*sumx[1]-sumxy[1]*sumx[3]*sumx[0]-sumxy[0]*sumx[2]*sumx[2];

		  corr0/=wg;
		  corr1/=wg;
		  corr2/=wg;

  	      streamlog_out(MESSAGE) << "  Gain corrections aplied: " << corr0 
                                     << " + ix * " << corr1 << " + ix * ix * " << corr2 << endl;
	    }

	  // Apply gain correction to upper image half

	  long iStart = _sizeXY/2;

	  if(cur_img ->IsIntImage())
		{
		long *pixels = cur_img->GetIntImageData();
                for(long ipix=iStart; ipix < _sizeXY; ipix++)
		  {
		    double xval= ipix%_sizeX;
		    xval = _frameDivision * xval/_sizeX -0.5;
                    pixels[ipix]*= corr0 + corr1*xval + corr2*xval*xval;
		  }
		}
	      else
		{
		double *pixels = cur_img->GetFloatImageData();
                for(long ipix=iStart; ipix < _sizeXY; ipix++)
		  {
		    double xval= ipix%_sizeX;
		    xval = _frameDivision * xval/_sizeX -0.5;
                    pixels[ipix]*= corr0 + corr1*xval + corr2*xval*xval;
		  }
		}

	  // Done!

	  delete BgMap;   
          delete [] GainCorr,sumx,sumxy;

    } //	for (int ifm = 0; ifm < nImg; ifm++) - loop over images in input collection


  // always return true  for ProcessorName
  setReturnValue(true);

}

void PiHalvesEqualizer::check(GloriaDataContainer* data)
{

	streamlog_out(DEBUG) << " check() " << std::endl;

}

void PiHalvesEqualizer::end()
{

	printEndMessage();

}

void PiHalvesEqualizer::printEndMessage() const
{

	streamlog_out(MESSAGE) << " end() " << " after " << _nLoops
			<< " calls to processData() " << std::endl;

	streamlog_out(MESSAGE) << _nInputImages << " images processed " 
			       << _nOutputImages << " stored to file " << std::endl;

}



} // namespace luiza
