#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "PiTableReader.h"

namespace luiza {

  PiTableReader aPiTableReader ;


  // Constructor

  PiTableReader::PiTableReader() : DataSourceProcessor("PiTableReader") {

    _description = "Processor for reading data tables written as FITS images" 
                   "Reads Pi data tables from given files" ;


  std::vector< std::string > emptyStringVec;

  registerProcessorParameter ("TableFileList",
                              "List of table files to be read",
                              _tableFileList,  emptyStringVec );


  registerProcessorParameter ("ListFileList",
                              "Names of files containing list of table file names (one per line)",
                              _listFileList,  emptyStringVec );


  // Although this processor reads tables from file it creates table collection as an output
  // This is why we register this collection as output collection

  registerOutputCollection (  TableCollection,
                              "TableCollectionName",
                              "Name of the image collection to which images from file should be stored",
                              _tableCollectionName,  std::string ("PiTable") );

  

  // Remaining parameters are optional - can be skiped in the steering file


  registerOptionalParameter ("SelectedColumns",
                              "Name (and order) of columns which should be copied to the output table",
                              _selectedColumns,  emptyStringVec  );

  registerOptionalParameter ("PermanentCollection",
                              "Flag for collections, which should not be deleted after loop is finished",
                              _permanentCollection,  false );

  registerOptionalParameter ("TablesPerLoop",
                              "Number of table files to be read per processing loop (0 for all))",
                              _tablesPerLoop,  static_cast < int > (1) );


  }


  PiTableReader*  PiTableReader::newProcessor() {
    return new PiTableReader ;
  }
  

  void PiTableReader::init() { 

    streamlog_out( MESSAGE ) << "PiTableReader::init()  " << name() 
			     << std::endl 
			     << "  parameters: " << std::endl 
			     << *parameters()  
			     << std::endl ;
    
    _nFiles = 0 ;
    _nTables = 0 ;
    _nLoops = 0 ;

    _isInputInitialized = false;

    // Check if input file name given

    if(_tableFileList.size()==0 && _listFileList.size()==0)
      {
	streamlog_out( ERROR ) << "No input files defined for PiTableReader! " << std::endl;
	streamlog_out( ERROR ) << "Input NOT initialized! " << std::endl;

	//        throw StopProcessingException( this ) ;
	return;

      }

  // Open list files and copy file names to the fits file list

    for(int ilist=0; ilist < _listFileList.size(); ilist++)
      {
      std::ifstream listFile;
      listFile.open(_listFileList[ilist].c_str(),std::ios::in);

      if(listFile.fail())
	{
	  streamlog_out( ERROR ) << "Error opening input list file: " 
                                 << _listFileList[ilist] <<std::endl;
          continue;
	}

      std::string fitsFile;

      while(1)
	{
	  listFile >> fitsFile;

          if(!listFile.good())break;

	  _tableFileList.push_back(fitsFile);  

	}

      listFile.close();
      }


 // Construct empty table, as defined by user

    _inputTable = new GloriaFitsTable;

    _inputImage = new GloriaFitsImage;

 // As there are file names specified, we are ready to start processing

    _isInputInitialized = true;

    streamlog_out( MESSAGE ) << "Ready to fill "<< _tableCollectionName << " table collection" << std::endl;
    streamlog_out( MESSAGE ) << "Input file list has " << _tableFileList.size() << " entries" <<std::endl;

 // Print list of files in DEBUG mode

    for(int ifile=0;ifile<_tableFileList.size();ifile++)
	 streamlog_out( DEBUG )  <<  "      File " <<  std::setw ( 5 ) << ifile+1 << " : " 
                                 << _tableFileList[ifile] <<std::endl;

 
    // Set return value to false (no input yet)
    
    setReturnValue( false ) ;
    
 }


void PiTableReader::startAnalysis( GloriaDataContainer* data ) 
{
  streamlog_out(DEBUG) << " startAnalysis() " << endl;
}


void PiTableReader::processData( GloriaDataContainer* data ) 
{ 
 
    bool readOK = false;
    
    // Set return value to false (no input)
    
    setReturnValue( false ) ;
    
    // Return without processing if no files left to read

    if(_nFiles >= _tableFileList.size() ) return;


    // Count only loops where something was done...

    _nLoops ++ ;

    streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << std::endl ;

    
    // Create output collection if not found

    if(! data -> TableCollectionExists(_tableCollectionName))
         data -> CreateTableCollection(_tableCollectionName,_permanentCollection);

    // Read _tablesPerLoop in each call to processData()
    // If _tablesPerLoop is set to 0: all files are read at first call (!)

    for(int ifile=0; (ifile<_tablesPerLoop || _tablesPerLoop==0) && _nFiles<_tableFileList.size() ;ifile++)
      {
      // Take next file name

      std::string fileName = _tableFileList[_nFiles];

      _nFiles++;

      // When this is the last file, set the 'close' flag
    
      if(_nFiles == _tableFileList.size()) _isInputClosed=true;


      // Clear input table (removing table structure as well)

      _inputTable->EmptyTable();

      // read  data from image file 

      _inputImage->ReadFitsFile(fileName);

      // Skip table if not properly read 
      if(_inputImage-> Error())
         {
	 streamlog_out(ERROR) << "Error reading fits file " << fileName << std::endl ;
	 streamlog_out(ERROR) << _inputImage-> ErrorMessage() << std::endl ;
         continue;
         }

      // Copy table header

      _inputTable->CopyHeader(_inputImage);

      // Set input file name
      
      _inputTable->SetInputFileName(fileName);
 
      // Define table structure

      int ncol = _inputImage->SizeX();
      int nrow = _inputImage->SizeY();

      bool isfloat = _inputImage->IsFloatImage();

      streamlog_out(DEBUG) << "File " << fileName << " with " << ncol << " x " << nrow 
                           << ((isfloat)?" floats":" ints") << " opened" << std::endl ;

      double *fvaltab;
      long *lvaltab;

      if(isfloat) 
        fvaltab = _inputImage->GetFloatImageData();
      else
	lvaltab = _inputImage->GetIntImageData();


      // Get column names and id

      vector<int> selids;
      vector<string> selnames;
      
      map<string, int> colnames;

      for(int icol = 0; icol < ncol ; icol++)
	{
          stringstream keyname;
          keyname.fill('0');
          keyname << "FIELD_" << setw(2) << icol+1;

          string colname = keyname.str();

	  if(_inputImage->IsKeywordSet(keyname.str().c_str()))
	    {
	      colname = _inputImage->GetKeywordValue(keyname.str().c_str());

              int ifpos = colname.find((char)39);
	      int ilpos = colname.rfind((char)39);

              if(ifpos==0 && ilpos+1==colname.size()) colname = colname.substr(1,ilpos-1); 
	    }
         else
            {
	      streamlog_out(WARNING) << "Name for column " << icol << " not found in FITS header" << std::endl ;
	    }

         selids.push_back(icol);
         selnames.push_back(colname);

         colnames.insert( pair<string, int>(colname,icol));
	}


      // Narrow list to selected columns (if specified)

      if(_selectedColumns.size() > 0 )
	{
	selids.clear();
	selnames.clear();
	  
	for(int isel=0; isel < _selectedColumns.size(); isel++)
	  {
	    string colname = _selectedColumns.at(isel);
	    if(colnames.find(colname) != colnames.end())
	      {
               selids.push_back(colnames.find(colname)->second);
               selnames.push_back(colname);
	      }
	    else
	      {
	      streamlog_out(WARNING) << "Column " << colname << " not found in input table" << std::endl ;
	      }
	  }
        }

      
      int nsel=selids.size();

      for(int isel = 0; isel < nsel ; isel++)
	{
	  int icol = selids.at(isel);

          string colname = selnames.at(isel);

	  streamlog_out(DEBUG) << "Reading column " << icol << " with name " << colname << std::endl ;

	  double shift = 0.0;
	  double scale = 1.0;

	  if(colname == string("D_RA") && _inputImage->IsKeywordSet("RA2000") )
	    {
	      shift = _inputImage->GetKeywordValueDouble("RA2000");
              scale = 15.0;
	      colname = "RA";
	      streamlog_out(DEBUG) << "       shift of " << shift*scale << " will be applied to " << colname << std::endl ;
	    }


	  if(colname == string("D_DEC") && _inputImage->IsKeywordSet("DEC2000") )
	    {
	      shift = _inputImage->GetKeywordValueDouble("DEC2000");
	      colname = "DEC";
	      streamlog_out(DEBUG) << "       shift of " << shift << " will be applied to " << colname << std::endl ;
	    }


	  if(isfloat)
	    {
	    _inputTable->AddColumn(colname, DoubleColumn);
	    _inputTable->AdjustColumns();

  	    for(int irow=0; irow<nrow;irow++)
	      {
	       if(isel==0) _inputTable->AddRow();
               double val = scale*(fvaltab[icol+ncol*irow]+shift);
	       _inputTable->SetValue(isel,irow, val);
	      }
	    }
	  else
	    {
	    _inputTable->AddColumn(colname, LongColumn);
	    _inputTable->AdjustColumns();

 	    for(int irow=0; irow<nrow;irow++)
	      {
	       if(isel==0) _inputTable->AddRow();
	       _inputTable->SetValue(isel,irow, lvaltab[icol+ncol*irow]);
	      }
	    }
	}


      // Add table to the collection

      (data->GetTableCollection(_tableCollectionName)).push_back(*_inputTable);

       readOK = true;
      _nTables++;
      }


    // return true if table read
    if(readOK) setReturnValue( true ) ;
    
  }
  
  void PiTableReader::check(GloriaDataContainer* data) { 
    
    if(! data -> TableCollectionExists(_tableCollectionName))
      {
         streamlog_out(ERROR) << " check(): Collection not found !? " << std::endl ;
	 return;
      }


    int nTbl = data->GetTableCollection(_tableCollectionName).size();

    streamlog_out(MESSAGE) << " check(): Collection " << _tableCollectionName << " with " << nTbl << " tables" << std::endl ;
 
  }

void PiTableReader::concludeAnalysis( GloriaDataContainer* data )
{
   streamlog_out(DEBUG) << " concludeAnalysis() " << std::endl;
}


  
void PiTableReader::end()
{ 
    printEndMessage() ;
}

  void PiTableReader::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;

    streamlog_out(MESSAGE) << " processed "     << _nFiles << " files " << std::endl 
			   << std::endl ;
    
    streamlog_out(MESSAGE) << " read "     << _nTables << " tables " << std::endl 
			   << std::endl ;
  }

  
}// namespace luiza
