#include "luiza/Exceptions.h"
#include <iostream>
#include <iomanip>
#include <sstream>

#include "luiza/VerbosityLevels.h"
#include "streamlog/streamlog.h"

#include "gloria/GloriaDataContainer.h"

#include "PiTableWriter.h"

namespace luiza {

  PiTableWriter aPiTableWriter ;


  // Constructor

  PiTableWriter::PiTableWriter() : Processor("PiTableWriter") {

    _description = "Processor for writing data tables as FITS images"
                   "Used to write Pi pipeline data to output files" ;


  registerInputCollection   ( TableCollection,
                              "TableCollectionName",
                              "Name of the table collection which should be stored to file",
                              _tableCollectionName,  std::string ("Objects") );

  registerProcessorParameter ("TextFileNameRoot",
                              "Output file name root, which will be appended by the table number",
                              _textFileNameRoot,  std::string("output/pi_table_") );

  registerProcessorParameter ("TextFileNameExt",
                              "Output file name extention",
                              _textFileNameExt,  std::string(".ast") );

  std::vector< std::string > emptyStringVec;

  registerOptionalParameter ("SelectedColumns",
                             "List of table columns to be stored (empty for all)",
			     _selectedColumns, emptyStringVec );

  registerOptionalParameter( "NewColumnNames",
   			     "Names which should be used for stored columns",
			     _newColumnNames, emptyStringVec);

  std::vector< std::string > shiftKeywordVec;
  shiftKeywordVec.push_back(std::string("CRVAL1"));
  shiftKeywordVec.push_back(std::string("CRVAL2"));
 
  registerOptionalParameter ("ShiftKeywords",
                             "Header keywords, which should be used for RA,DEC positions of the frame center",
			     _shiftKeywords, shiftKeywordVec );


  std::vector< std::string > posColumnVec;
  posColumnVec.push_back(std::string("RA"));
  posColumnVec.push_back(std::string("DEC"));
 
  registerOptionalParameter ("PosColumns",
                             "Table columns, where RA and DEC positions are stored",
			     _posColumns, posColumnVec );

  }


  PiTableWriter*  PiTableWriter::newProcessor() {
    return new PiTableWriter;
  }


  void PiTableWriter::init() {

    streamlog_out( MESSAGE ) << "PiTableWriter::init()  " << name()
			     << endl
			     << "  parameters: " << endl
			     << *parameters()
			     << endl ;
    
    if (_selectedColumns.size() != 0  && _newColumnNames.size() !=0 &&  
            _selectedColumns.size() != _newColumnNames.size() )
	{
	streamlog_out(ERROR) << "New column names not properly specified!" << std::endl;
	streamlog_out(ERROR) << "Names from the first input table will be used!" << std::endl;

	_newColumnNames.clear();
	}

    if (_selectedColumns.size() == 0  && _newColumnNames.size() !=0 )
	{
	streamlog_out(ERROR) << "Can not specify new column names if selected columns not listed!" << std::endl;
	streamlog_out(ERROR) << "Names from the first input table will be used!" << std::endl;

	_newColumnNames.clear();
	}

    _outputImage = NULL;

    _nLoops = 0 ;

    _nTables = 0 ;

    _nWritten = 0 ;

 }

  void PiTableWriter::startAnalysis( GloriaDataContainer* data )
  {
   streamlog_out(MESSAGE) << "Starting data processing data" << endl ;
  }


  void PiTableWriter::processData( GloriaDataContainer* data )
  {
	_nLoops ++ ;

	streamlog_out(MESSAGE) << "Processing data,  pass nr " << _nLoops << endl ;

        if(! data -> TableCollectionExists(_tableCollectionName))
	  {
          streamlog_out(DEBUG) << "Input collection not found ! " << endl ;

          return;
	  }
      

        if( data -> TableCollectionPermanent(_tableCollectionName))
	  {
         if(_isFirstEvent)
             streamlog_out(MESSAGE) << "Permanent collection - skiped ! " << endl ;

          return;
	  }

	// This is not very nice, but effective
	// Use concludeAnalysis() method also to process 'normal' data 

	concludeAnalysis( data );

}


 void PiTableWriter::concludeAnalysis( GloriaDataContainer* data )
  {
       int nTab = data -> GetTableCollection(_tableCollectionName).size();


        if(data -> TableCollectionPermanent(_tableCollectionName) && nTab == 0)
	  {
          streamlog_out(DEBUG) << "No tables in permanent collection ! " << endl ;

          return;
	  }
      

        streamlog_out(DEBUG) << "  Collection " << _tableCollectionName << " with " << nTab << " tables " << endl ;


        for(int itb=0; itb<nTab; itb++)
	  {
	   GloriaFitsTable *source =  &(data -> GetTableCollection(_tableCollectionName).at(itb));  

          _nTables++;

           
	  // If list of columns is empty, all columns will be stored,
	  // as found in the first processed table (!)

          if(_selectedColumns.size() == 0)
 	      for(int icol=0; icol < source->GetColumnNumber(); icol++)
 		   _selectedColumns.push_back(source->GetColumnName(icol));


          int ncol = _selectedColumns.size();
          int nrow = source->GetRowNumber();

	  
          // Create output image (floating point values by default)

	  _outputImage = new GloriaFitsImage(ncol,nrow,true);

          _outputImage->CopyHeader(source);

	  // Pointer to image table

	  double *fvaltab = _outputImage->GetFloatImageData();
	  
	  // Copy selected columns
          // If column is missing, zero values are inserted

          for(int icol=0; icol< _selectedColumns.size();icol++)
	    {
	      
	    string colname = _selectedColumns.at(icol);
   
	    string newname = colname;
	    if(_selectedColumns.size() == _newColumnNames.size()) newname = _newColumnNames.at(icol);
   
	    string colcomment = "not found";
	      
 	    if(source->ColumnExists(colname))
	      {
		
  	      streamlog_out(DEBUG) << "Writing column " << icol << " with name " << colname << endl ;
	      
              colcomment = source->GetColumn(colname)->Comment();
	      
              double shift = 0.0;
              double scale = 1.0;
		  
	      if(colname == _posColumns.at(0) && _outputImage->IsKeywordSet(_shiftKeywords.at(0).c_str()))
	         {
	         shift = _outputImage->GetKeywordValueDouble(_shiftKeywords.at(0).c_str());
                 scale = 15.0;
	         colname = "D_RA";
	         streamlog_out(DEBUG) << "       shift of " << shift << " will be subtracted for " << colname << std::endl ;

		 string racomment = "field center from " + _shiftKeywords.at(0) + " / 15";
		 
		 _outputImage->SetKeywordDouble("RA2000",shift/scale,racomment);
			 
	         }


	      if(colname == _posColumns.at(1) && _outputImage->IsKeywordSet(_shiftKeywords.at(1).c_str()))
	         {
	         shift = _outputImage->GetKeywordValueDouble(_shiftKeywords.at(1).c_str());
	         colname = "D_DEC";
	         streamlog_out(DEBUG) << "       shift of " << shift << " will be subtracted for " << colname << std::endl ;

		 string deccomment = "field center from " + _shiftKeywords.at(1);
		 
		 _outputImage->SetKeywordDouble("DEC2000",shift,deccomment);
			 
		 }

	      for(int irow=0; irow<nrow;irow++)
		{
		  double val;
		  source->GetValue(_selectedColumns.at(icol), irow, &val);
		  fvaltab[icol+ncol*irow]=(val - shift)/scale;
		}
	      }
	    else
	      {
	      streamlog_out(WARNING) << "Column " << icol << " with name " << colname << " not found in input collection" << endl ;
	      

	      for(int irow=0; irow<nrow;irow++)
		  fvaltab[icol+ncol*irow]=0.;
	      }

	    // Store column name in the header
	    
	    stringstream keyname;
            keyname.fill('0');
            keyname << "FIELD_" << setw(2) << icol+1;

	    // Add for proper string recognition:
	    
	    newname = "'" + newname + "'";

	    _outputImage->SetKeyword(keyname.str().c_str(), newname, colcomment );
	    
	    }

	  // Remove column definitions from the header
	  
	  for(int icol=_selectedColumns.size();;icol++)
	    {
	    stringstream keyname;
            keyname.fill('0');
            keyname << "FIELD_" << setw(2) << icol+1;

	    if(! _outputImage->IsKeywordSet(keyname.str().c_str()))break;

	    _outputImage->RemoveKeyword(keyname.str().c_str());
	    }
	  
	  // Construct output file name from given root, table number and .fit extension

          stringstream outname;

          outname << _textFileNameRoot;

	  // If the file name root ends with '/' append it with input file name (if present)

          int fnlen=_textFileNameRoot.size();

          if( _textFileNameRoot.at(fnlen-1) == '/' && source->InputFileName().size() > 0 )
             {
              int flen = source->InputFileName().size();

	      int ilpos = source->InputFileName().rfind('.');

	      if (ilpos != std::string::npos) flen = ilpos;

	      int ifpos = source->InputFileName().rfind('/');

              if (ifpos == std::string::npos)
		outname << source->InputFileName().substr(0,flen) ;
	      else
		outname << source->InputFileName().substr(ifpos+1,flen-ifpos-1) ; 
	     }
	  else
          if(!data->TableCollectionPermanent(_tableCollectionName) || nTab!=1)
	      {
              outname.fill('0'); 
              outname << setw(4) << _nTables;
	      }
           
          outname << _textFileNameExt ;


	  // Write image to file

	  int bitPix = -32;

          _outputImage->WriteFitsFile(outname.str(),bitPix);

	  if(_outputImage->Error())
		{
		streamlog_out(ERROR) << "Error writing table to output image file " << endl ;
		streamlog_out(ERROR) << _outputImage->ErrorMessage() << endl ;

                delete _outputImage;

		return;
		}
	  
          _nWritten++;

          delete _outputImage;
	  }
  

    // always return true  for ProcessorName
    setReturnValue( true ) ;

  }

  void PiTableWriter::check(GloriaDataContainer* data)
  {
	    streamlog_out(DEBUG) << " check() " << endl;
  }

  void PiTableWriter::end()
  {
	  printEndMessage() ;

  }

  void PiTableWriter::printEndMessage() const {

    streamlog_out(MESSAGE) << " end() "
			   << " after " << _nLoops << " calls to processData() " << endl ;

    streamlog_out(MESSAGE) << _nTables << " tables processed " << endl;

    streamlog_out(MESSAGE) << _nWritten << " tables written to output file " << endl
			   << endl ;
  }


}// namespace luiza
