#include "SatelliteSearching.h"
#include "SearchSatellite.h"
#include "FirstFit.h"
#include "Fit2DGauss.h"


double median,trigger;
long *pixel, X, Y;
vector< satellite > satellites;
set<long> checked;
satellite s;
double p[16];
GloriaFitsImage *frame;

namespace luiza
{
	SatelliteSearching aSatelliteSearching ;

	SatelliteSearching::SatelliteSearching() : Processor("SatelliteSearching")
	{

    		_description = "Search satellites on frame and fit 2DGauss curve"
      		"Prints coordinates of beginning and ending points of the satellites." ;

   		registerInputCollection (  ImageCollection,
                              "ImageCollectionName",
                              "Name of the image collection, empty string for all collections",
                              _imageCollectionName,  std::string ("") );

   		// Optional parameter to define window for mean and RMS calculation

  		std::vector<int > initWindow;

  		registerOptionalParameter ("OutputTextFile",
                      "Name of the output text file to store the table",
                             _outputTextFile, std::string(""));
 	}

  	void SatelliteSearching::init() 
	{ 
    		streamlog_out( MESSAGE ) << "SatelliteSearching::init()  " << std::endl 
                             	<< "  processor name: " << name() << std::endl 
			     	<< "  parameters: " << std::endl 
			     	<< *parameters()  
			     	<< std::endl ;
    

    		_nImages = 0 ;
    		_nImgCollections = 0 ;
    		_nLoops = 0 ;
    		_imgTable = NULL;
	}


	void SatelliteSearching::startAnalysis( GloriaDataContainer* data ) 
	{ 

    		streamlog_out( MESSAGE ) << "SatelliteSearching::startAnalysis()  " << std::endl ;
    
    		// Allocate table for statistics

    		_imgTable = new GloriaFitsTable;

    		// Define columns
		_imgTable->AddColumn("File",StringColumn);
		
		_imgTable->AddColumn("X1",DoubleColumn);
		_imgTable->AddColumn("Y1",DoubleColumn);
		_imgTable->AddColumn("X2",DoubleColumn);
		_imgTable->AddColumn("Y2",DoubleColumn);
  	}



	void SatelliteSearching::processData( GloriaDataContainer* data ) 
	{ 
    		_nLoops ++ ;

    		streamlog_out(DEBUG) << "Processing data,  pass nr " << _nLoops << std::endl ;
                             
    		int nCol=0;

   		if(_imageCollectionName.size()==0)
      			nCol = data -> GetImageCollectionNumber();
    		else
      			if(data -> ImageCollectionExists(_imageCollectionName)) nCol=1;
      

    		if(nCol==0)
      		{
			streamlog_out(DEBUG) << "  no images in data collection " << _imageCollectionName << std::endl ;
        		return;
      		}

    		_nImgCollections+=nCol;

    	// Loop over image collections (only one collection, if collection name given)
    	for(int icl=0; icl<nCol; icl++)
      	{
			std::string ColName = 
         (_imageCollectionName.size()==0)?data -> GetImageCollectionName(icl):_imageCollectionName;


		// Number of images in the collection
       		int nImg = data -> GetImageCollection(ColName).size();

       		streamlog_out(DEBUG) << "  Collection " << ColName << " with " << nImg << " images " << std::endl ;

		// Loop over images in the collection
        	for(int ifm=0; ifm<nImg; ifm++)
	  	{
	    		frame =  &(data -> GetImageCollection(ColName).at(ifm));  
			streamlog_out(DEBUG) 
				<< "     Image " << frame->InputFileName()  << " is " << frame->SizeX() 
                          	<< "x" << frame->SizeY() << std::endl ;

           		_nImages++;
		

           		//_imgTable->AddRow((long)_nImages, (long)frame->SizeX(), (long)frame->SizeY());
		
          		int sizeX = frame->SizeX();
          		int sizeY = frame->SizeY();


           		if(frame->IsIntImage())
			{
              			long *pixels = frame->GetIntImageData();
				pixel = pixels;
			}
	

			X = sizeX;
			Y = sizeY;

			MedianValue();

			SearchSatellite::FindPossibleSatellites();
			quantity = satellites.size();

			while(satellites.size()>0)
			{
				if(checked.find(satellites.begin()->start) == checked.end())
				{
					FirstFit::Start();
					Fit2DGauss::Start();
					if(p[0] > minLength)
						WriteResults();
				}
	
				EraseSatellite();
			}
	  	}

      }


    	// always return true  for ProcessorName
    	setReturnValue( true ) ;
    
  	// set ProcessorName.EvenNumberOfEvents == true if this processor has been called 2n 			(n=0,1,2,...) times 
    	setReturnValue("EvenNumberOfEvents",  !( _nLoops % 2 )   ) ;    
}
  

  	void SatelliteSearching::check(GloriaDataContainer* data) 
	{ 
    		streamlog_out(DEBUG) << " check() " << std::endl ;
  	}
  
  	void SatelliteSearching::concludeAnalysis( GloriaDataContainer* data )
	{   
    		printEndMessage() ;

    		if(_outputTextFile.size() > 0 )
         		_imgTable->WriteTextFile(_outputTextFile);
  	}

  	void SatelliteSearching::printEndMessage() const 
	{

    		streamlog_out(MESSAGE) << " concludeAnalysis() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;


   		streamlog_out(MESSAGE) << " processed "     << _nImages << " images in " 
			   << _nImgCollections << " collections " << std::endl 
			   << std::endl ;
    
    		stringstream tabss;

    		_imgTable->Print(tabss,20);

    			streamlog_out(MESSAGE) << " Image statistics collected during processing " << endl;

    		if(_imageCollectionName.size()!=0)
      		{
      			streamlog_out(MESSAGE) << "   for image collection '" << _imageCollectionName << "':" << endl;
      		}
    		else
      			streamlog_out(MESSAGE) << "   for all image collection found: " << endl;

    		streamlog_out(MESSAGE) <<  tabss.str() << endl;


    		// test deprecated method message 
    		message<DEBUG>( " and this is really the final DEBUG message ....") ;
  	}
  
  	void SatelliteSearching::end() 
	{
    		streamlog_out(MESSAGE) << " end() "  
			   << " after " << _nLoops << " calls to processData() " << std::endl ;
  	}

	void SatelliteSearching::MedianValue()
	{	
		vector<long> pixelVector(X*Y);
		for(int i=0; i<X*Y; ++i)
			pixelVector[i] = pixel[i];

		sort(pixelVector.begin(), pixelVector.end());
	
		median = pixelVector[0.5*X*Y];
		trigger = ( pixelVector[0.83*X*Y] - pixelVector[0.17*X*Y] )/2;
	}

	void SatelliteSearching::EraseSatellite()
	{
		satellites.erase(satellites.begin());	
	}

	void SatelliteSearching::WriteResults()
	{
		/*cout << "----------------------------------------------------" << endl;
		cout << "Date: " << image.GetKeywordValue("DATE-OBS") << endl;
		cout << "HJD: " << image.GetKeywordValue("JD") << endl << endl;

		cout << "Path length: " << sqrt(pow(p[1]-p[3],2)+pow(p[2]-p[4],2)) << endl;    
		cout << "Start: (" << p[1] << ";" << p[2] << ")" << endl; 
		cout << "End: (" << p[3] << ";" << p[4] << ")" << endl;
		cout << p[12] << " " << p[13] << " " << p[14] << " " << p[15]  << endl; 
		
		cout << "Line y=ax+b" << endl;
		cout << "a = " << p[5] << endl;
		cout << "b = " << p[6] << endl << endl;
	
		cout << "2D Gauss " << endl;
		cout << "sigma = " << p[9] << endl;
		cout << "A = " << p[10] << endl;
		cout << "C = " << p[11] << endl << endl;*/

                _imgTable->AddRow();

	        _imgTable->SetLast("File",frame->InputFileName());
	        _imgTable->SetLast("X1",p[1]);
		_imgTable->SetLast("Y1",p[2]);
		_imgTable->SetLast("X2",p[3]);
		_imgTable->SetLast("Y2",p[4]);
	
		cout << "Progress: " << (2-satellites.size()/quantity)*100 << "%" << endl;
	}

}// namespace luiza



